/*
 * FASt Term Recognizer
 *
 *  FASTR/alalldyn.h
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for dynamic memory allocations
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*****************************************************************************
******************************************************************************
                               DAGS
******************************************************************************
*****************************************************************************/
/*****************************************************************************
                         FILS DE DAGS
*****************************************************************************/
void     MiseAZeroDag( TabDagMem p_ptDag , int p_typeDag )
{
    int         i;

    p_ptDag->TypeDag = (char)p_typeDag;
    p_ptDag->Marque = _False;
    p_ptDag->Fils = NULL;
    p_ptDag->Frere = NULL;
    switch( p_typeDag )
    {
    case _TypeDagVide : break;
    case _AtomeEntier : p_ptDag->ValDag.Atome.AtomeConjDisj = _False;
                        p_ptDag->ValDag.Atome.ValAtome.Entier = _Nil; break;
    case _AtomeChaine : p_ptDag->ValDag.Atome.AtomeConjDisj = _False;
                        for( i = 0 ; i < _MaxLetIdent + 1 ; i++ )
                        {
                           (p_ptDag->ValDag.Atome.ValAtome.Chaine)[ i ] 
                                  = (char)_Nil;
                        }
                        break;
    case _AtomeRef :    p_ptDag->ValDag.Atome.AtomeConjDisj = _False;
                        p_ptDag->ValDag.Atome.ValAtome.Ref = _Nil; break;
    case _AtomeListe :  p_ptDag->ValDag.Atome.AtomeConjDisj = _False;
                        p_ptDag->ValDag.Atome.ValAtome.Ref = _Nil; break;
    case _Arc : p_ptDag->ValDag.Arc.ArcERNegation = _False; 
                p_ptDag->ValDag.Arc.ValMin = (char)_Nil; 
                p_ptDag->ValDag.Arc.ValMax = (char)_Nil; 
                p_ptDag->ValDag.Arc.Comptage = (char)_ArcOrdinaire; 
                p_ptDag->ValDag.Arc.ValSymb = _Nil; 
                p_ptDag->ValDag.Arc.ValRef = _Nil; 
                *(p_ptDag->ValDag.Arc.Ident) = '\0';
                break;
    case _Oper :p_ptDag->ValDag.OperLog.Oper = (char)_Nil; 
                p_ptDag->ValDag.OperLog.TypeFils1 = (char)_Nil; 
                p_ptDag->ValDag.OperLog.TypeFils2 = (char)_Nil; 
                p_ptDag->ValDag.OperLog.Fils1.ptPereAtome = (TabDagMem)_Nil; 
                p_ptDag->ValDag.OperLog.Fils2.ptPereAtome = (TabDagMem)_Nil; 
                break;
    default :   sprintf( ChAux , "\nType [%d]\n" , p_typeDag );
                printf( ChAux );
                TxtErreur( 44 , "MiseAZeroDag(alalldyn)" );
                break;
    }
}

TabDagMem NouveauDag( int p_typeDag )
{
    TabDagMem    ptDag;

    if( !PremierDagLibre )
    {
      IniElemsAllouables( _TabDag );
    }
    ptDag = PremierDagLibre;
    PremierDagLibre = PremierDagLibre->Frere;

    MiseAZeroDag( ptDag , p_typeDag );
    return( ptDag );
}

char     RendUnDag( TabDagMem *p_ptDag )
{
    if( !(*p_ptDag) )
    {
         return _True;
    }

    if( !((*p_ptDag)->Marque) )
    {
        (*p_ptDag)->Frere = PremierDagLibre;
        PremierDagLibre = (*p_ptDag);
        (*p_ptDag)->Marque = _True;
    }
    else
    {
        TxtErreur( 44 , "RendUnDag(alalldyn)" );
        return _False;
    }

    (*p_ptDag) = NULL;
    return _True;
}

char     MarqueDag( TabDagMem *p_ptDag )
{
    if( !(*p_ptDag) )
    {
         return _True;
    }
    (*p_ptDag)->Marque = _True;
    return _True;
}

char     DemarqueDag( TabDagMem *p_ptDag )
{
    if( !(*p_ptDag) )
    {
         return _True;
    }
    (*p_ptDag)->Marque = _False;

    return _True;
}

/*****************************************************************************
******************************************************************************
                               ARBRES
******************************************************************************
*****************************************************************************/
/*****************************************************************************
                                MOTS
*****************************************************************************/
void     MiseAZeroUnMot( TabMotMem p_ptMot )
{
    *(p_ptMot->ChaineMot) = '\0';
    p_ptMot->PtListeLemmes = NULL;
    p_ptMot->ArcsFils = NULL;
    p_ptMot->Forme = 'N'; /* Normal */
    p_ptMot->MotFrere = NULL;
}

TabMotMem NouveauMot( void )
{
    TabMotMem    ptMot;

    if( !PremierMotLibre )
    {
        IniElemsAllouables( _TabMots );
    }
    ptMot = PremierMotLibre;
    PremierMotLibre = PremierMotLibre->MotFrere;
    
    MiseAZeroUnMot( ptMot );
    return( ptMot );
}

TabMotMem IndiceAMot( int p_indMot )
{
    TabMotMem    ptMot;
    int          indMotPhrase;

    ptMot = MotsPhrase;

    /* saut des mots */
    for( indMotPhrase = 0 ; indMotPhrase < p_indMot && ptMot ; 
	 indMotPhrase++ ) {
      ptMot = ptMot->MotFrere;
    }

    return( ptMot );
}

char     RendMot( TabMotMem *p_ptMot )
{
    if( !(*p_ptMot) )
    {
         return _True;
    }
    
    (*p_ptMot)->MotFrere = PremierMotLibre;
    PremierMotLibre = (*p_ptMot);
    (*p_ptMot) = NULL;
    return _True;
}

/*****************************************************************************
                                LEMMES
*****************************************************************************/
void     MiseAZeroUnLemme( TabLemmeMem p_ptLemme )
{
    *(p_ptLemme->ValTerme.Ident) = '\0';
    p_ptLemme->ValTerme.Ref = _Nil;
    p_ptLemme->ValTerme.RefRacine = _Nil;
    p_ptLemme->ValTerme.RefSemantique = NULL;
    p_ptLemme->ValTerme.RefSelf = NULL;
    p_ptLemme->ValTerme.TypeTerme = (char)_Nil;
    p_ptLemme->ValTerme.ArcsFils = NULL;
    p_ptLemme->ValTerme.PosFicRegle = _Nil;

    p_ptLemme->NbRegles = _Nil;
    p_ptLemme->NbDerives = _Nil;
    p_ptLemme->NbVoisinsSemantiques = _Nil;
    p_ptLemme->NbVoisinsSemantiquesRec = _Nil;
    p_ptLemme->PosFicDerives = _Nil;
    p_ptLemme->PosFicRacine = _Nil;
    p_ptLemme->PosFicLiensSemantiques = _Nil;
    p_ptLemme->PosFicLiensSemantiquesRec = _Nil;

    p_ptLemme->LemmeFrere = NULL;
}

TabLemmeMem NouveauLemme( void )
{
    TabLemmeMem    ptLemme;

    if( !PremierLemmeLibre )
    {
        IniElemsAllouables( _TabLemmes );
    }
    ptLemme = PremierLemmeLibre;
    PremierLemmeLibre = PremierLemmeLibre->LemmeFrere;
    
    MiseAZeroUnLemme( ptLemme );
    return( ptLemme );
}

char     DemarqueEtRendLemme( TabLemmeMem *p_ptLemme )
{
    if( !(*p_ptLemme) )
    {
         return _True;
    }
    
    DemarqueEtRendDag( &((*p_ptLemme)->ValTerme.ArcsFils) );
    RendListeRefPhrase( &((*p_ptLemme)->ValTerme.RefSemantique)  );
    RendListeRefPhrase( &((*p_ptLemme)->ValTerme.RefSelf)  );
    
    (*p_ptLemme)->LemmeFrere = PremierLemmeLibre;
    PremierLemmeLibre = (*p_ptLemme);
    (*p_ptLemme) = NULL;
    return _True;
}

char     DemarqueEtRendListeLemme( TabLemmeMem *p_ptLemme )
{
    TabLemmeMem         ptLemmeCou;
    TabLemmeMem         ptLemmeSuiv;

    if( !(ptLemmeCou = *p_ptLemme) )
    {
         return _True;
    }

    while( ptLemmeCou )
    {
        ptLemmeSuiv = ptLemmeCou->LemmeFrere;
        DemarqueEtRendDag( &(ptLemmeCou->ValTerme.ArcsFils) );
	RendListeRefPhrase( &(ptLemmeCou->ValTerme.RefSemantique)  );
	RendListeRefPhrase( &(ptLemmeCou->ValTerme.RefSelf)  );

        ptLemmeCou->LemmeFrere = PremierLemmeLibre;
        PremierLemmeLibre = ptLemmeCou;

        ptLemmeCou = ptLemmeSuiv;
    }

    (*p_ptLemme) = NULL;
    return _True;
}

/*****************************************************************************
                              TERMES
*****************************************************************************/
void     MiseAZeroUnTerme( TabTermeMem p_ptTerme )
{
    *(p_ptTerme->ValTerme.Ident) = '\0';
    p_ptTerme->ValTerme.Ref = _Nil;
    p_ptTerme->ValTerme.RefRacine = _Nil;
    p_ptTerme->ValTerme.RefSemantique = NULL;
    p_ptTerme->ValTerme.RefSelf = NULL;
    p_ptTerme->ValTerme.TypeTerme = _TypeTermeVide;
    p_ptTerme->ValTerme.ArcsFils = NULL;
    p_ptTerme->ValTerme.PosFicRegle = (long)_RegleNonLiee;

    p_ptTerme->TermeFrere = NULL;
    p_ptTerme->TermeFils = NULL;
}

TabTermeMem NouveauTerme( void )
{
    TabTermeMem    ptTerme;

    if( !PremierTermeLibre )
    {
        IniElemsAllouables( _TabTermes );
    }
    ptTerme = PremierTermeLibre;
    PremierTermeLibre = PremierTermeLibre->TermeFrere;
    
    MiseAZeroUnTerme( ptTerme );
    return( ptTerme );
}

/* le dag est rendu sans etre demarque pour ne pas reveinr sur
   un demarquage initial de la totalite de l'arbre */
char     RendTerme( TabTermeMem *p_hdlTerme )
{
    if( !(*p_hdlTerme) )
    {
         return _True;
    }
    
    RendDag( &((*p_hdlTerme)->ValTerme.ArcsFils) );
    RendListeRefPhrase( &((*p_hdlTerme)->ValTerme.RefSemantique)  );
    RendListeRefPhrase( &((*p_hdlTerme)->ValTerme.RefSelf)  );
    
    (*p_hdlTerme)->TermeFrere = PremierTermeLibre;
    PremierTermeLibre = (*p_hdlTerme);
    (*p_hdlTerme) = NULL;
    return _True;
}

/* le dag est rendu en etant demarque dans le cas d'un terme rendu isolement
   (et non dans un arbre de regle ou de metaregle  */
char     DemarqueEtRendTerme( TabTermeMem *p_hdlTerme )
{
    if( !(*p_hdlTerme) )
    {
         return _True;
    }
    
    DemarqueEtRendDag( &((*p_hdlTerme)->ValTerme.ArcsFils) );
    RendListeRefPhrase( &((*p_hdlTerme)->ValTerme.RefSemantique)  );
    RendListeRefPhrase( &((*p_hdlTerme)->ValTerme.RefSelf)  );
    
    (*p_hdlTerme)->TermeFrere = PremierTermeLibre;
    PremierTermeLibre = (*p_hdlTerme);
    (*p_hdlTerme) = NULL;
    return _True;
}

/*****************************************************************************
                          TERMES DE METAREGLE
*****************************************************************************/
void     MiseAZeroUnRegleMeta( TabRegleMemMeta p_ptRegleMeta )
{
    p_ptRegleMeta->NoTableau = 0;
    p_ptRegleMeta->AvecUnificationExterne = _False;

    p_ptRegleMeta->TermeFils = NULL;
    p_ptRegleMeta->RegleFrere = NULL;
}

TabRegleMemMeta NouveauRegleMeta( void )
{
    TabRegleMemMeta    ptRegleMeta;

    if( !PremierRegleLibreMeta )
    {
        IniElemsAllouables( _TabReglesMeta );
    }
    ptRegleMeta = PremierRegleLibreMeta;
    PremierRegleLibreMeta = PremierRegleLibreMeta->RegleFrere;

    MiseAZeroUnRegleMeta( ptRegleMeta );
    return( ptRegleMeta );
}

char     DemarqueEtRendRegleMeta( TabRegleMemMeta *p_ptRegleMeta )
{
    char                 resul;

    if( !(*p_ptRegleMeta) )
    {
         return _True;
    }

     resul &= DemarqueEtRendArbre( &((*p_ptRegleMeta)->TermeFils) );

    (*p_ptRegleMeta)->RegleFrere = PremierRegleLibreMeta;
    PremierRegleLibreMeta = (*p_ptRegleMeta);
    (*p_ptRegleMeta) = NULL;
    return resul;
}

/*****************************************************************************
                              REGLES
*****************************************************************************/
void     MiseAZeroUneRegle( TabRegleMem p_ptRegle )
{
    p_ptRegle->TermeFils = NULL;
    p_ptRegle->RegleFrere = NULL;

    p_ptRegle->Transformee = NULL;
    p_ptRegle->PosFicRegleSuivante = _Nil;
}

TabRegleMem NouvelleRegle( void )
{
    TabRegleMem    ptRegle;

    if( !PremiereRegleLibre )
    {
        IniElemsAllouables( _TabRegles );
    }
    ptRegle = PremiereRegleLibre;
    PremiereRegleLibre = PremiereRegleLibre->RegleFrere;

    MiseAZeroUneRegle( ptRegle );
    return( ptRegle );
}

char     DemarqueEtRendRegle( TabRegleMem *p_ptRegle )
{
     char           resul = _True;
    
     if( !(*p_ptRegle) )
     {
         return _True;
     }

     resul &= DemarqueArbreDF( &((*p_ptRegle)->TermeFils) );
     resul &= DemarqueArbreDF( &((*p_ptRegle)->Transformee) );
     resul &= ParcoursDFArbre( &((*p_ptRegle)->TermeFils) , RendTerme , _PostOrdre );
     resul &= ParcoursDFArbre( &((*p_ptRegle)->Transformee) , RendTerme , _PostOrdre );

     (*p_ptRegle)->RegleFrere = PremiereRegleLibre;
     PremiereRegleLibre = (*p_ptRegle);
     (*p_ptRegle) = NULL;
     return _True;
}

/*****************************************************************************
                  REFERENCES MULTIPLES DES TERMES DE PHRASES
*****************************************************************************/
void     MiseAZeroUnRefPhrase( TabRefPhraseMem p_ptRefPhrase )
{
    p_ptRefPhrase->Ref = _Nil;
    p_ptRefPhrase->Suivant = NULL;
}

TabRefPhraseMem NouveauRefPhrase( void )
{
    TabRefPhraseMem    ptRefPhrase;

    if( !PremierRefPhraseLibre )
    {
        IniElemsAllouables( _TabRefPhrase );
    }
    ptRefPhrase = PremierRefPhraseLibre;
    PremierRefPhraseLibre = PremierRefPhraseLibre->Suivant;

    MiseAZeroUnRefPhrase( ptRefPhrase );
    return( ptRefPhrase );
}

/*  */
char     RendRefPhrase( TabRefPhraseMem *p_hdlRefPhrase )
{
    if( !(*p_hdlRefPhrase) )
    {
         return _True;
    }
    
    (*p_hdlRefPhrase)->Suivant = PremierRefPhraseLibre;
    PremierRefPhraseLibre = (*p_hdlRefPhrase);
    (*p_hdlRefPhrase) = NULL;
    return _True;
}

/*  */
char     RendListeRefPhrase( TabRefPhraseMem *p_hdlRefPhrase )
{
    TabRefPhraseMem         ptRefPhraseCou;
    TabRefPhraseMem         ptRefPhraseSuiv;

    if( !(ptRefPhraseCou = *p_hdlRefPhrase) )
    {
         return _True;
    }

    while( ptRefPhraseCou )
    {
        ptRefPhraseSuiv = ptRefPhraseCou->Suivant;

        ptRefPhraseCou->Suivant = PremierRefPhraseLibre;
        PremierRefPhraseLibre = ptRefPhraseCou;

        ptRefPhraseCou = ptRefPhraseSuiv;
    }

    (*p_hdlRefPhrase) = NULL;
    return _True;
}

/************************************************************************
*************************************************************************
                       DESALLOCATION GENERALISEE
*************************************************************************
*************************************************************************/

/*****************************************************************************
              MARQUAGE , DEMARQUAGE  ET DESALLOCATION D'UN DAG
*****************************************************************************/
char     MarqueDagDF( TabDagMem  *p_ptDag )
{
     return ParcoursDFDag( p_ptDag , MarqueDag , _PostOrdre , _False );
}
char     MarqueDagTerme( TabTermeMem  *p_hdlTerme )
{
     return MarqueDagDF( &((*p_hdlTerme)->ValTerme.ArcsFils) );
}

char     DemarqueDagDF( TabDagMem  *p_ptDag )
{
     return ParcoursDFDag( p_ptDag , DemarqueDag , _PostOrdre , _False );
}

char     DemarqueDagTerme( TabTermeMem  *p_hdlTerme )
{
     return DemarqueDagDF( &((*p_hdlTerme)->ValTerme.ArcsFils) );
}

char     RendDag( TabDagMem *p_ptDag )
{
     return ParcoursDFDag( p_ptDag , RendUnDag , _PostOrdre , _True );
}

char     DemarqueEtRendDag( TabDagMem *p_ptDag )
{
     char                  resul;

     resul = DemarqueDagDF( p_ptDag ) && RendDag( p_ptDag );
     *p_ptDag = NULL;
     return resul;
}

/*****************************************************************************
                     MARQUAGE D'UN ARBRE DE TERMES
*****************************************************************************/
char     MarqueArbreDF( TabTermeMem *p_hdlTerme )
{
      return ParcoursDFArbre( p_hdlTerme , MarqueDagTerme , _PreOrdre );
}

/*****************************************************************************
                     DEMARQUAGE D'UN ARBRE DE TERMES
*****************************************************************************/
char     DemarqueArbreDF( TabTermeMem *p_hdlTerme )
{
      return ParcoursDFArbre( p_hdlTerme , DemarqueDagTerme , _PreOrdre );
}

/*****************************************************************************
                     DESALLOCATION D'UN ARBRE DE TERMES
*****************************************************************************/
char     DemarqueEtRendArbre( TabTermeMem *p_hdlTerme )
{
     char           resul = _True;

     resul &= DemarqueArbreDF( p_hdlTerme );
     resul &= ParcoursDFArbre( p_hdlTerme , RendTerme , _PostOrdre );
     return resul;
}

/*****************************************************************************
                  DESALLOCATION D'UNE LISTE DE REGLES
*****************************************************************************/
char     DemarqueEtRendListeRegle( TabRegleMem *p_ptRegle )
{
    TabRegleMem         ptRegleSuiv;
    TabRegleMem         ptRegleCou;

    ptRegleCou = *p_ptRegle;
    while( ptRegleCou )
    {
        ptRegleSuiv = ptRegleCou->RegleFrere;
        DemarqueEtRendRegle( &ptRegleCou );
        ptRegleCou = ptRegleSuiv;
    }
    (*p_ptRegle) = NULL;
    return _True;
}

char     DemarqueEtRendListeRegleMeta( TabRegleMemMeta *p_ptRegleMeta )
{
    TabRegleMemMeta      ptRegleMetaSuiv;
    while( *p_ptRegleMeta )
    {
        ptRegleMetaSuiv = (*p_ptRegleMeta)->RegleFrere;
        DemarqueEtRendRegleMeta( p_ptRegleMeta );
        (*p_ptRegleMeta) = ptRegleMetaSuiv;
    }
    return _True;
}

/*****************************************************************************
                     Desallocation des regles memoire
*****************************************************************************/
char     RAZReglesMemoire( void )
{
    int                        nbReglesLoc;

    /* initialisation du tableau des metaregles */
    for( nbReglesLoc = 0 ; nbReglesLoc < NbReglesMem ; nbReglesLoc++ )
    {
          DemarqueEtRendArbre( &((ReglesMem + nbReglesLoc)->TermeFils) );
          MiseAZeroUneRegle( ReglesMem + nbReglesLoc );
    }
    NbReglesMem = 0;

    return _True;
}

/*****************************************************************************
                            Desallocation des metaregles
*****************************************************************************/
char     RAZReglesMeta( void )
{
    int                       noTableau;

    /* initialisation du tableau des metaregles */
    for( noTableau = 0 ; noTableau < _NbMaxTableau ; noTableau++ )
    {
          if( !DemarqueEtRendListeRegleMeta( 
                     &(Tableaux[ noTableau ].PtRegleMeta ) ) )
          {
                  return _False;
          }
          Tableaux[ noTableau ].PtRegleMeta = NULL;
          *(Tableaux[ noTableau ].Ident) = '\0';
    }

    return _True;
}

/************************************************************************
*************************************************************************
                                COPIES
*************************************************************************
*************************************************************************/
/*****************************************************************************
   AJOUT RECURSIF D'UN DAG, RETOURNE L'ADRESSE DE LA RACINE
   NE RECOPIE QUE LA PARTIE DEMARQUEE, LA PARTIE MARQUEE EST SUPPOSEE ETRE
   RETROUVEE DANS LE TABLEAU DE LIENS
*****************************************************************************/
char     CopieDagDF( TabDagMem *p_hdlDest , TabDagMem p_ptSrc ,
		     int p_estAutomate )
{
     TabDagMem               *hdlDestFils;
     TabDagMem                ptSrcCou;
     TabDagMem                ptDestCou;
     TabDagMem                dagVide;
     char                     appelRecursif;
     char                     estDagCouAutomate;
     int                      ind;

/*      fprintf(FicLog,"Src\n"); */
/*      EcrireDagBasique( &p_ptSrc ); */
/*      fprintf(FicLog,"Dest (autom %d)\n",p_estAutomate); */
/*      EcrireDagBasique( p_hdlDest ); */
     
     if( !p_ptSrc && !(*p_hdlDest) )
     {
          return _True;
     }

     /******************************************************
          source marquee => deja parcourue => a retrouver
     *******************************************************/
     if( p_ptSrc && p_ptSrc->Marque )
     {
         for( ind = 0 ; ind < NbLiensNoeudsCopies ; ind++ )
         {
             if( LiensNoeudsCopies[ ind ].PtNoeudAnc == p_ptSrc )
             {
                  if( *p_hdlDest && (*p_hdlDest)->TypeDag != _TypeDagVide
                      && *p_hdlDest != LiensNoeudsCopies[ ind ].PtNoeudNouv )
                  {
                     printf( "Type [%d]\n" , (*p_hdlDest)->TypeDag );
                     TxtErreur( 199 , "CopieDagDF(ALDAGS)" );
                     return _False;
                  }
                  else if( !(*p_hdlDest) )
                  {
                      *p_hdlDest = LiensNoeudsCopies[ ind ].PtNoeudNouv;
                  }
                  else if( (*p_hdlDest)->TypeDag == _TypeDagVide )
                  {
                      dagVide = *p_hdlDest;
                      RestaurePointeurVersDagJete( dagVide , 
                                       LiensNoeudsCopies[ ind ].PtNoeudNouv );
                      if( *p_hdlDest != LiensNoeudsCopies[ ind ].PtNoeudNouv )
                      {
                          TxtErreur( 1 , "revoir restauration pointeurs" );
                      }
                      DemarqueDag( &dagVide );
                      RendUnDag( &dagVide );

                  }
                  return _True;
             }
         }
         TxtErreur( 200 , "CopieDagDF(ALDAGS)" );
         return _False;
     }

     /******************************************************
          source vide => 
          si pas de pteur en dest, on ajoute un dag vide
     *******************************************************/
     if( p_ptSrc && p_ptSrc->TypeDag == _TypeDagVide )
     {
          if( p_estAutomate )
          {
              TxtErreur( 93 , "SubsumeDagDF etape inattendue [1](ALDAGS)" );
              ptSrcCou = p_ptSrc;
          }

          if( !(*p_hdlDest) )
          {
              if( !AjouteDagVideFils( p_hdlDest ) )
              {
                  p_ptSrc->Marque = _True;
                  return _False;
              }
          }

          (*p_hdlDest)->Marque = _True;
          p_ptSrc->Marque = _True;

          LiensNoeudsCopies[ NbLiensNoeudsCopies ].PtNoeudAnc = p_ptSrc;
          /* on fait pointer le noeud nouveau vers le debut de la liste des
             fils. Mais le pointage n'a d'importance que pour le premier 
             de la liste qui pointera, lui vers la tete de liste puisque 
             l'ajout se fait en queue                                      */
          LiensNoeudsCopies[ NbLiensNoeudsCopies ].PtNoeudNouv = *p_hdlDest;
          NbLiensNoeudsCopies++;
          if( NbLiensNoeudsCopies >= NbMaxLiensMemorises )
          {
              TxtErreur( 94 , "CopieDagDF(alalldyn)" );
              return _False;
          }

          return _True;
     }

     /*******************************************
                      SUBSOMPTION
      *******************************************/
     ptDestCou = *p_hdlDest;
     /* on verifie la subsomption : tous les fils de la destination
        doivent se retrouver dans la source */
     /* le cas d'un fils vide ou d'absence de fils revient au meme
        pour la subsomption : c'est ok, mais on considere que ca n'a
        pas ete parcouru */
     while( ptDestCou )
     {
          if( p_estAutomate )
          {
              TxtErreur( 93 , "SubsumeDagDF etape inattendue [2](ALDAGS)" );
              ptSrcCou = p_ptSrc;
          }

          appelRecursif = _False;

          /* arc ou atome deja parcouru : erreur, la source
             correspondante doit elle aussi etre marquee */
          if( (ptDestCou->Marque) )
          {
               TxtErreur( 222 , "CopieDagDF(alalldyn)" );
               return _False;
          }
          ptDestCou->Marque = _True;

          /* arc non encore parcouru */
          if( ptDestCou->TypeDag == _Arc 
	      || ptDestCou->TypeDag == _AtomeListe )
          {
             if( (ptSrcCou 
		  = ExisteArcOuListeFils( p_ptSrc , 
					  ptDestCou ,
					  ptDestCou->TypeDag ) ) )
             {
                 appelRecursif = _True;
             }
          }
          /* arc non encore parcouru */
          else if( ptDestCou->TypeDag == _AtomeRef 
                 || ptDestCou->TypeDag == _AtomeEntier 
                 || ptDestCou->TypeDag == _AtomeChaine )
          /* atome non encore parcouru */
          {
              printf( "Atome de type %d\n" , ptDestCou->TypeDag );
              printf( "Valeur ref %ld\n" , 
                      ptDestCou->ValDag.Atome.ValAtome.Ref );
              TxtErreur( 93 , "SubsumeDagDF etape inattendue(ALDAGS)" );
              ptSrcCou = p_ptSrc;
          }
          /* atome non encore parcouru */
          else if( ptDestCou->TypeDag == _TypeDagVide )
          /* arc vide non encore parcouru */
          {
              ptSrcCou = NULL;
          }
          /* arc vide non encore parcouru */
          else
          {
              printf( "Type [%d]\n" , ptDestCou->TypeDag );
              TxtErreur( 93 , "SubsumeDagDF(ALDAGS)" );
              return _False;
          }

          /* cas d'un arc ou d'un atome commun */
          if( ptSrcCou )
          {
              ptSrcCou->Marque = _True;

              LiensNoeudsCopies[ NbLiensNoeudsCopies ].PtNoeudAnc = ptSrcCou;
              /* on fait pointer le noeud nouveau vers le debut de la
                 liste des fils.  */
              LiensNoeudsCopies[ NbLiensNoeudsCopies ].PtNoeudNouv 
                       = *p_hdlDest;
              NbLiensNoeudsCopies++;
              if( NbLiensNoeudsCopies >= NbMaxLiensMemorises )
              {
                  TxtErreur( 94 , "SubsumeDagDF(ALDAGS)" );
                  return _False;
              }

              /* pour les arcs communs, on rappelle recursivement */
              if( appelRecursif )
              {
                   if( ptDestCou->TypeDag == _Arc 
                       || ptDestCou->TypeDag == _AtomeListe )
                   {
                       if( !CopieDagDF( &(ptDestCou->Fils) , 
                                        ptSrcCou->Fils , p_estAutomate) )
                       {
                            return _False;
                       }
                   }
                   else
                   {
                       printf( "Type [%d]\n" , ptDestCou->TypeDag );
                       TxtErreur( 93 , "SubsumeDagDF(ALDAGS)" );
                       return _False;
                   }
              }
              /* pour les arcs communs, on rappelle recursivement */
          }
          /* cas d'un arc ou d'un atome commun */

          ptDestCou = ptDestCou->Frere;
     }
     /* on a verifie la subsomption */

     if( !p_ptSrc )
     {
          return _True;
     }

     /*******************************************
                COPIE
      *******************************************/
     ptSrcCou = p_ptSrc;
     /* on avance dans les fils du dag tant qu'ils coincident avec le
      dag initial si aucun fils du dag ne coincide avec l'arc courant
      du dag, on l'ajoute */
     while( ptSrcCou )
     {
         appelRecursif = _False;
         estDagCouAutomate = _False;

         /* arc non encore recopie */
         if( ptSrcCou->TypeDag == _Arc 
	     || ptSrcCou->TypeDag == _AtomeListe )
         {
             /* les automates sont systematiquement des copies vers un dag
		vide avec ajout imperatif (lorsque le dag n'a pas encore
		ete traite) */
             if( p_estAutomate || !(*p_hdlDest) 
                 || !(ptDestCou = ExisteArcOuListeFils( (*p_hdlDest) , 
							ptSrcCou ,
							ptSrcCou->TypeDag ) ) )
             {
                 /* arc ou atome non encore recopie */
                 /* s'il est marque c'est qu'il a ete trouve dans la
                    phase de subsomption, il doit donc se retrouver
                    dans la dest */
                 if( (ptSrcCou->Marque) )
                 {
                      if( !p_estAutomate )
                      {
                           TxtErreur( 226 , "CopieDagDF(alalldyn)" );
                            return _False;
                      }
                      else
                      {
                            /* un automate peut deja avoir ete traite lors
			       de la descente sur un fils precedent, dans
			       ce cas il faut le rechercher dans les liens
			       connus */
                            for( ind = 0 ; ind < NbLiensNoeudsCopies ; ind++ )
                            {
                                if( LiensNoeudsCopies[ ind ].PtNoeudAnc 
                                    == ptSrcCou )
                                {
                                     if( !(hdlDestFils 
                                           = AjouteArcOuListeFils( p_hdlDest , 
                        &((LiensNoeudsCopies[ ind ].PtNoeudNouv)->ValDag) ,
                        LiensNoeudsCopies[ ind ].PtNoeudNouv ,
	                _True , ptSrcCou->TypeDag ) ) )
                                     {
                                         return _False;
				     }
                                     break;
                                }
                            }
                            if( ind == NbLiensNoeudsCopies )
                            {
                                 TxtErreur( 200 , "CopieDagDF(ALDAGS)" );
                                 return _False;
                            }
                      }
                 }
                 else
                 {
                      if( !(hdlDestFils = AjouteArcOuListeFils( p_hdlDest , 
                                       &(ptSrcCou->ValDag) , 
                                       NULL , _True , ptSrcCou->TypeDag ) ) )
                      {
                          return _False;
                      }

                      if( ptSrcCou->ValDag.Arc.ValSymb 
                          == CheminLeadingExprReg->ValDag.Arc.ValSymb 
		       || ptSrcCou->ValDag.Arc.ValSymb 
                          == CheminTrailingExprReg->ValDag.Arc.ValSymb )
                      {
                           estDagCouAutomate = _True;
                      }

                      appelRecursif = _True;
                  }
            }
             else
             {
                 hdlDestFils = &(ptDestCou->Fils);
                 /* s'il n'est pas marque c'est qu'il n'a pas ete
                    trouve dans la phase de subsomption, il doit donc
                    se retrouver dans la dest */
                 if( !(ptSrcCou->Marque) )
                 {
                      TxtErreur( 227 , "CopieDagDF(alalldyn)" );
                      return _False;
                 }
             }
         }
         /* arc non encore recopie */
         else if( ptSrcCou->TypeDag == _AtomeRef 
                 || ptSrcCou->TypeDag == _AtomeEntier 
                 || ptSrcCou->TypeDag == _AtomeChaine )
         /* atome non encore recopie */
         {
             if( !(*p_hdlDest)
                 || !(ExisteAtomeFils( (*p_hdlDest) , ptSrcCou->TypeDag ,
                                       ptSrcCou->ValDag.Atome ) ) )
             {
                 /* s'il est marque c'est qu'il a ete trouve dans la
                    phase de subsomption, il doit donc se retrouve
                    dans la dest */
                 if( (ptSrcCou->Marque) )
                 {
                      TxtErreur( 231 , "CopieDagDF(alalldyn)" );
                      return _False;
                 }

		 /* atome ordinaire, on recopie la valeur */
		 if( !AjouteAtomeFils( p_hdlDest , 
				       ptSrcCou->TypeDag ,
				       ptSrcCou->ValDag.Atome , _True ) ) {
		   return _False;
		 }
             }
         }
         /* atome non encore recopie */
         else
         {
              printf( "Type [%d]\n" , ptDestCou->TypeDag );
              TxtErreur( 93 , "CopieDagDF(ALDAGS)" );
              return _False;
         }

         ptSrcCou->Marque = _True;
         (*p_hdlDest)->Marque = _True;

         LiensNoeudsCopies[ NbLiensNoeudsCopies ].PtNoeudAnc = ptSrcCou;
         /* on fait pointer le noeud nouveau vers le debut de la liste des
            fils. Mais le pointage n'a d'importance que pour le premier 
            de la liste qui pointera, lui vers la tete de liste puisque 
            l'ajout se fait en queue                                      */
         LiensNoeudsCopies[ NbLiensNoeudsCopies ].PtNoeudNouv = *p_hdlDest;
         NbLiensNoeudsCopies++;
         if( NbLiensNoeudsCopies >= NbMaxLiensMemorises )
         {
             TxtErreur( 94 , "CopieDagDF(alalldyn)" );
             return _False;
         }

         /* pour les arcs communs, on rappelle recursivement */
         if( appelRecursif )
         {
             if( estDagCouAutomate )
             {
                 if( !CopieDagDF( hdlDestFils , ptSrcCou->Fils , _True ) )
                 {
                      return _False;
                 }
             }
             else
             {
                 if( !CopieDagDF( hdlDestFils , ptSrcCou->Fils , 
                                  p_estAutomate ) )
                 {
                      return _False;
                 }
             }
         }
         
         ptSrcCou = ptSrcCou->Frere;
     }

     return _True;
}

/*****************************************************************************
    AJOUT D'UN DAG, RETOURNE L'ADRESSE DU FILS DU DERNIER DAG AJOUTE
*****************************************************************************/
char     CopieDag( TabDagMem *p_hdlDest , TabDagMem p_ptSrc )
{
     char                    resul;

     NbLiensNoeudsCopies = 0;

     if( !DemarqueDagDF( &p_ptSrc ) || !DemarqueDagDF( p_hdlDest ) )
     {
         return _False;
     }

     PtOperLogU = NULL;                /* liste des operations d'unif
                                          a faire */
     HdlDagU = p_hdlDest;              /* liste des dags sur lesquels
                                          porte l'unif */
     HdlTermeU = NULL;                 /* arbre de termes portant les
                                          dags unifies */
     PtBlocU = NULL;                   /* liste des blocs portant les
                                          dags unifies */

     resul = CopieDagDF( p_hdlDest , p_ptSrc , _False );
     return resul;
}

/*************************************************************************
       RECOPIE D'UNE LISTE DE REFERENCES
*************************************************************************/
char     CopieReferences( TabRefPhraseMem *p_hdlDest , 
			  TabRefPhraseMem p_ptSrc )
{
  TabRefPhraseMem                    ptDest;
  TabRefPhraseMem                    ptDestPrec;
  TabRefPhraseMem                    ptSrcCou;

  ptDest = NULL;
  ptDestPrec = NULL;
  *p_hdlDest = NULL;
  ptSrcCou = p_ptSrc;

  /* parcours de tous les atomes de la liste de references */
  while( ptSrcCou ) {
    /* allocation */
    if( !(ptDest = NouveauRefPhrase()) ) {
      return _False;
    }
    ptDest->Ref = ptSrcCou->Ref;

    /* chainage */
    if( !(*p_hdlDest) ) {
      *p_hdlDest = ptDest;
    }
    if( ptDestPrec ) {
      ptDestPrec->Suivant = ptDest;
    }
    ptDestPrec = ptDest;

    /* passage au suivant a copier */
    ptSrcCou = ptSrcCou->Suivant;
  }

  return _True;
}

/*************************************************************************
       RECOPIE D'UN ARBRE (ET ALLOCATION DES TERMES COPIES)
*************************************************************************/
char     CopieEtUnifieArbreDF( TabTermeMem *p_hdlDest , TabTermeMem p_ptSrc ,
                                int p_avecUnification )
{
     if( !p_ptSrc )
     {
         return _True;
     }

     if( !(*p_hdlDest) )
     {
         *p_hdlDest = NouveauTerme();
         if( !(*p_hdlDest) )
         {
             return _False;
         }

         *(*p_hdlDest) = *p_ptSrc;
         (*p_hdlDest)->ValTerme.ArcsFils = NULL;
         (*p_hdlDest)->TermeFils = NULL;
         (*p_hdlDest)->TermeFrere = NULL;
     }
     else
     {
         if( (*p_hdlDest)->ValTerme.Ref == _Nil )
         {
               (*p_hdlDest)->ValTerme.Ref 
                          = p_ptSrc->ValTerme.Ref;
               (*p_hdlDest)->ValTerme.RefRacine 
                          = p_ptSrc->ValTerme.RefRacine;
	       CopieReferences( &((*p_hdlDest)->ValTerme.RefSelf), 
				p_ptSrc->ValTerme.RefSelf );
	       CopieReferences( &((*p_hdlDest)->ValTerme.RefSemantique), 
				p_ptSrc->ValTerme.RefSemantique );
               (*p_hdlDest)->ValTerme.PosFicRegle 
                          = p_ptSrc->ValTerme.PosFicRegle;
         }
         (*p_hdlDest)->ValTerme.TypeTerme = p_ptSrc->ValTerme.TypeTerme;
     }

     if( !CopieEtUnifieArbreDF( &((*p_hdlDest)->TermeFils) , 
                                 p_ptSrc->TermeFils , 
                                 p_avecUnification )
         || !CopieEtUnifieArbreDF( &((*p_hdlDest)->TermeFrere) , 
                                    p_ptSrc->TermeFrere , 
                                    p_avecUnification ) )
     {
         return _False;
     }

     if( p_avecUnification )
     {
         return Unifier( &((*p_hdlDest)->ValTerme.ArcsFils) ,
                         &(p_ptSrc->ValTerme.ArcsFils) ); 
     }
     else
     {
         return CopieDagDF( &((*p_hdlDest)->ValTerme.ArcsFils) , 
                            p_ptSrc->ValTerme.ArcsFils , _False );
     }
            
}

/*************************************************************************
       RECOPIE D'UN ARBRE (ET ALLOCATION DES TERMES COPIES)
*************************************************************************/
char     CopieEtUnifieArbre( TabTermeMem *p_hdlDest , TabTermeMem p_ptSrc ,
                             int p_avecUnification , 
                             TabTermeMem *p_hdlTermeU )
{
     char              resul;

     NbLiensNoeudsCopies = 0;

     if( !DemarqueArbreDF( &p_ptSrc ) || !DemarqueArbreDF( p_hdlDest ) )
     {
         return _False;
     }

     PtOperLogU = NULL;        /* liste des operations d'unif a faire      */
     HdlDagU = NULL;                   /* liste des dags sur lesquels porte l'unif */
     HdlTermeU = p_hdlTermeU;          /* arbre de termes portant les dags unifies */
     PtBlocU = NULL;                   /* liste des blocs portant les dags unifies */

     resul = CopieEtUnifieArbreDF( p_hdlDest , p_ptSrc , p_avecUnification );
     return resul;
}

/*************************************************************************
       LORS DE LA GENERATION D'UNE METAREGLE, ON RECOPIE EN PARTIE DROITE
      LES SOUS-ARBRES DES TERMES DE LA PARTIE GAUCHE QUI SONT DES FEUILLES
                             EN PARTIE DROITE.
*************************************************************************/
char     CopieTermeEtPartageDags( TabTermeMem *p_hdlDest , TabTermeMem p_ptSrc )
{
    if( !p_ptSrc )
    {
        return _True;
    }

    *p_hdlDest = NouveauTerme();
    if( !(*p_hdlDest) )
    {
        return _False;
    }

    *(*p_hdlDest) = *p_ptSrc;
    /* les pointeurs vers ArcsFils sont donc egaux */
    (*p_hdlDest)->TermeFils = NULL;
    (*p_hdlDest)->TermeFrere = NULL;

    if( !CopieTermeEtPartageDags( &((*p_hdlDest)->TermeFils) , 
				  p_ptSrc->TermeFils )
        || !CopieTermeEtPartageDags( &((*p_hdlDest)->TermeFrere) , 
				     p_ptSrc->TermeFrere ) )
    {
        return _False;
    }
    return _True;
}

/*************************************************************************
       LORS DE LA GENERATION D'UNE METAREGLE, ON RECOPIE EN PARTIE DROITE
      LES SOUS-ARBRES DES TERMES DE LA PARTIE GAUCHE QUI SONT DES FEUILLES
                         EN PARTIE DROITE.
*************************************************************************/
char     CopieSousArbreTermesEgaux( TabTermeMem *p_hdlDest )
{
    int                             indTerme;

    if( (*p_hdlDest)->TermeFils )
    {
        return _True;
    }

    /* recherche du terme de la partie droite de la metaregle ayant
       cet identificateur */
    for( indTerme = 0 ; indTerme < NbLiensIdentTerme ; indTerme++ )
    {
        if( strcmp( LienIdentTerme[ indTerme ].Ident ,
                    (*p_hdlDest)->ValTerme.Ident ) == 0 )
        {
            /* le type du terme change, il passe vraisemblablement au
               type concatenation, on recopie pour cela le type du
               terme d'origine */
            (*p_hdlDest)->ValTerme.TypeTerme
                    = (LienIdentTerme[ indTerme ].PtTerme)->ValTerme.TypeTerme;
            /* on recopie la filiation de ce terme, en mettant les
               dags en commun */
            return CopieTermeEtPartageDags( &((*p_hdlDest)->TermeFils) ,
                      (LienIdentTerme[ indTerme ].PtTerme)->TermeFils );
        }
    }
    return _True;
}

/************************************************************************
*************************************************************************
                                DIVERS
*************************************************************************
*************************************************************************/
/*****************************************************************************
                  CHAINAGE INITIAL DE TOUS LES TERMES LIBRES
*****************************************************************************/
long     IniElemsAllouables( int p_typeElement )
{
    long                indRegle;
    long                indLemme;
    long                indTerme;
    long                indDag;
    long                indMot;
    long                indRefPhrase;
    long                nbElemAlloues;

    if( NbReallocations[ p_typeElement ] == 0 ) {
      nbElemAlloues = NbMaxElemAllouables[ p_typeElement ];
      NbReglesMem = 0;
    }
    else {
      nbElemAlloues = (long)((float)(NbMaxElemAllouables[ p_typeElement ]) 
			     * 0.2);
      NbMaxElemAllouables[ p_typeElement ] += nbElemAlloues;
    }

/*     if( Trace ) { */
/*       fprintf( stderr , "type %d alloues dyn %ld alloues tot %ld\n" ,  */
/* 	       p_typeElement , nbElemAlloues ,  */
/* 	       NbMaxElemAllouables[ p_typeElement ] ); */
/*     } */

    (NbReallocations[ p_typeElement ])++;

    switch( p_typeElement ) {
    case _TabRegles :
	PremiereRegleLibre 
	  = (TabRegleMem)_Alloc( (unsigned long) sizeof( struct RegleMem )
                      * (unsigned long) nbElemAlloues );
	if ( !PremiereRegleLibre ) {
	  TxtErreur( 18 , "IniElemsAllouables(alalldyn)" );
	  exit(0);
	}
	for( indRegle = 0 ; indRegle < nbElemAlloues - 1 ; indRegle++ ) {
	  (PremiereRegleLibre + indRegle)->RegleFrere 
	    = PremiereRegleLibre + indRegle + 1;
	}
	(PremiereRegleLibre + nbElemAlloues - 1)->RegleFrere = NULL;
	break;
    case _TabReglesMeta :
	PremierRegleLibreMeta 
	  = (TabRegleMemMeta)_Alloc( (unsigned long)
				     sizeof( struct RegleMemMeta )
                      * (unsigned long) nbElemAlloues );
	if ( !PremierRegleLibreMeta ) {
	  TxtErreur( 18 , "IniElemsAllouables(alalldyn)" );
	  exit(0);
	}
	for( indRegle = 0 ; indRegle < nbElemAlloues - 1 ; indRegle++ ) {
	  (PremierRegleLibreMeta + indRegle)->RegleFrere 
	    = PremierRegleLibreMeta + indRegle + 1;
	}
	(PremierRegleLibreMeta + nbElemAlloues - 1)->RegleFrere = NULL;
	break;
    case _TabLemmes :
        PremierLemmeLibre = (TabLemmeMem)_Alloc( (unsigned long) 
						 sizeof( struct LemmeMem )
			   * (unsigned long) nbElemAlloues );
	if ( !PremierLemmeLibre ) {
	  TxtErreur( 18 , "IniElemsAllouables(alalldyn)" );
	  exit(0);
	}
	for( indLemme = 0 ; indLemme < nbElemAlloues - 1 ; indLemme++ ) {
	  (PremierLemmeLibre + indLemme)->LemmeFrere 
	    = PremierLemmeLibre + indLemme + 1;
	}
	(PremierLemmeLibre + nbElemAlloues - 1)->LemmeFrere = NULL;
	break;
    case _TabRefPhrase :
        PremierRefPhraseLibre = (TabRefPhraseMem)_Alloc( (unsigned long) 
					    sizeof( struct RefPhraseMem )
			       	 * (unsigned long) nbElemAlloues );
	if ( !PremierRefPhraseLibre ) {
	  TxtErreur( 18 , "IniElemsAllouables(alalldyn)" );
	  exit(0);
	}
	for( indRefPhrase = 0 ; indRefPhrase < nbElemAlloues - 1 ; 
	     indRefPhrase++ ) {
	  (PremierRefPhraseLibre + indRefPhrase)->Suivant 
	    = PremierRefPhraseLibre + indRefPhrase + 1;
	}
	(PremierRefPhraseLibre + nbElemAlloues - 1)->Suivant = NULL;
	break;
    case _TabTermes :
        PremierTermeLibre = (TabTermeMem)_Alloc( (unsigned long) 
					 sizeof( struct TermeMem )
				   * (unsigned long) nbElemAlloues );
	if ( !PremierTermeLibre ) {
	  TxtErreur( 18 , "IniElemsAllouables(alalldyn)" );
	  exit(0);
	}
	for( indTerme = 0 ; indTerme < nbElemAlloues - 1 ; indTerme++ ) {
	  (PremierTermeLibre + indTerme)->TermeFrere 
	    = PremierTermeLibre + indTerme + 1;
	}
	(PremierTermeLibre + nbElemAlloues - 1)->TermeFrere = NULL;
	break;
    case _TabDag :
        PremierDagLibre = (TabDagMem)_Alloc( (unsigned long) 
					     sizeof( struct DagMem )
				    * (unsigned long) nbElemAlloues );
	if ( !PremierDagLibre ) {
	  TxtErreur( 18 , "IniElemsAllouables(alalldyn)" );
	  exit(0);
	}
	for( indDag = 0 ; indDag < nbElemAlloues - 1 ; indDag++ ) {
	  (PremierDagLibre + indDag)->Frere = PremierDagLibre + indDag + 1;
	}
	(PremierDagLibre + nbElemAlloues - 1)->Frere = NULL;
	break;
    case _TabMots :
        PremierMotLibre = (TabMotMem)_Alloc( (unsigned long) 
						 sizeof( struct MotMem )
			   * (unsigned long) nbElemAlloues );
	if ( !PremierMotLibre ) {
	  TxtErreur( 18 , "IniElemsAllouables(alalldyn)" );
	  exit(0);
	}
	for( indMot = 0 ; indMot < nbElemAlloues - 1 ; indMot++ ) {
	  (PremierMotLibre + indMot)->MotFrere 
	    = PremierMotLibre + indMot + 1;
	}
	(PremierMotLibre + nbElemAlloues - 1)->MotFrere = NULL;
	break;
    }
    return nbElemAlloues;
}

int      NbReglesLibres( void )
{
    int              valRet = 0;  
    TabRegleMem      ptRegle = PremiereRegleLibre;

    while( ptRegle )
    {
        valRet++;
        ptRegle = ptRegle->RegleFrere;
        if( valRet > NbMaxElemAllouables[ _TabRegles ] )
        {
            TxtErreur( 45 , "NbReglesLibres(alalldyn)" );
            return valRet;
        }
    }
    return valRet;
}

int      NbReglesMetaLibres( void )
{
    int                  valRet = 0;  
    TabRegleMemMeta      ptRegleMeta = PremierRegleLibreMeta;

    while( ptRegleMeta )
    {
        valRet++;
        ptRegleMeta = ptRegleMeta->RegleFrere;
        if( valRet > NbMaxElemAllouables[ _TabReglesMeta ] )
        {
            TxtErreur( 45 , "NbReglesMetaLibres(alalldyn)" );
            return valRet;
        }
    }
    return valRet;
}

int      NbLemmesLibres( void )
{
    int            valRet = 0;  
    TabLemmeMem    ptLemme = PremierLemmeLibre;

    while( ptLemme )
    {
        valRet++;
        ptLemme = (TabLemmeMem)ptLemme->LemmeFrere;
        if( valRet > NbMaxElemAllouables[ _TabLemmes ] )
        {
            TxtErreur( 45 , "NbLemmesLibres(alalldyn)" );
            return valRet;
        }
    }
    return valRet;
}

int      NbTermesLibres( void )
{
    int            valRet = 0;  
    TabTermeMem    ptTerme = PremierTermeLibre;

    while( ptTerme )
    {
        valRet++;
        ptTerme = ptTerme->TermeFrere;
        if( valRet > NbMaxElemAllouables[ _TabTermes ] )
        {
            TxtErreur( 45 , "NbTermesLibres(alalldyn)" );
            return valRet;
        }
    }
    return valRet;
}

int      NbDagLibres( void )
{
    int            valRet = 0;  
    TabDagMem      ptDag = PremierDagLibre;

    while( ptDag )
    {
        valRet++;
        ptDag = ptDag->Frere;
        if( valRet > NbMaxElemAllouables[ _TabDag ] )
        {
            TxtErreur( 45 , "NbDagLibres(alalldyn)" );
            return valRet;
        }
    }
    return valRet;
}

int      NbRefPhraseLibres( void )
{
    int            valRet = 0;  
    TabRefPhraseMem      ptRefPhrase = PremierRefPhraseLibre;

    while( ptRefPhrase )
    {
        valRet++;
        ptRefPhrase = ptRefPhrase->Suivant;
        if( valRet > NbMaxElemAllouables[ _TabRefPhrase ] )
        {
            TxtErreur( 45 , "NbRefPhraseLibres(alalldyn)" );
            return valRet;
        }
    }
    return valRet;
}

int      NbMotsLibres( void )
{
    int            valRet = 0;  
    TabMotMem      ptMot = PremierMotLibre;

    while( ptMot )
    {
        valRet++;
        ptMot = ptMot->MotFrere;
        if( valRet > NbMaxElemAllouables[ _TabMots ] )
        {
            TxtErreur( 45 , "NbMotsLibres(alalldyn)" );
            return valRet;
        }
    }
    return valRet;
}

int      PhotoMemoire( FILE *p_ptFic )
{
     int              nbRegles;
     int              nbReglesMeta;
     int              nbLemmes;
     int              nbTermes;
     int              nbDags;
     int              nbMots;
     int              nbRefs;

    fprintf( p_ptFic , 
"Used: Rules %-8d Mots %-8d Meta %-8d Lemmas %-8d Terms %-8d Dags %-8d Refs %-8d\n" ,
             (nbRegles = NbMaxElemAllouables[ _TabRegles ] - NbReglesLibres()) , 
             (nbMots = NbMaxElemAllouables[ _TabMots ] - NbMotsLibres()) , 
             (nbReglesMeta = NbMaxElemAllouables[ _TabReglesMeta ] - NbReglesMetaLibres()) , 
             (nbLemmes = NbMaxElemAllouables[ _TabLemmes ] - NbLemmesLibres()) , 
             (nbTermes = NbMaxElemAllouables[ _TabTermes ] - NbTermesLibres()) , 
             (nbDags = NbMaxElemAllouables[ _TabDag ] - NbDagLibres()) , 
             (nbRefs = NbMaxElemAllouables[ _TabRefPhrase ] - NbRefPhraseLibres()) );
    fprintf( p_ptFic , 
"      Rules [%-5.2f%%] Mots [%-5.2f%%] Meta [%-5.2f%%] Lemmas [%-5.2f%%] Terms [%-5.2f%%] Dags [%-5.2f%%] Refs [%-5.2f%%]\n" ,
             100.0 * (float)nbRegles / (float)NbMaxElemAllouables[ _TabRegles ],
             100.0 * (float)nbMots / (float)NbMaxElemAllouables[ _TabMots ],
             100.0 * (float)nbReglesMeta / (float)NbMaxElemAllouables[ _TabReglesMeta ],
             100.0 * (float)nbLemmes / (float)NbMaxElemAllouables[ _TabLemmes ],
             100.0 * (float)nbTermes / (float)NbMaxElemAllouables[ _TabTermes ],
             100.0 * (float)nbDags / (float)NbMaxElemAllouables[ _TabDag ],
             100.0 * (float)nbRefs / (float)NbMaxElemAllouables[ _TabRefPhrase ] );
    return _True;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                        fin de alalldyn.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
