/*
 * FASt Term Recognizer
 *
 *  FASTR/alanalut.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains auxiliary code for the parser
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*************************************************************************
      VARIABLES EXTERNES DE RECHERCHE D'UNE COLLOCATION
*************************************************************************/
int                         PremMotColloc;
int                         DernMotColloc;
int                         LongueurTerme;
int                         EstPermute;
int                         NbPrintedWords;

/*************************************************************************
     ANALYSE D'UNE EXPRESSION REGULIERE PAR DESCENTE RECURSIVE ET 
        NON DETERMINISTE CAR L'AUTOMATE N'EST PAS DETERMINISE 
                (NI MINIMISE NI SANS TRANS VIDE)
**************************************************************************/
char     AnalyserER( TabDagMem *p_hdlArc , int p_indMotCou , 
		     TabMotMem p_ptMotCou , int p_indMotMax ,
		     int p_profondeur , int p_typeArcPrec )
{
    int              tabCats[ _NbMaxHomographe ];
    long             tabRefs[ _NbMaxHomographe ];
    int              cats;
    int              catCou;
    long             refCou;
    long             refArc;
    int              refCatArc;
    TabDagMem        ptFilsCou;
    TabLemmeMem      ptLemme;
    int              indCat;
    char             trouve;
    char             estNeg;
    char             courtCircuitInterd;
    char             courtCircuitOblig;
    char             typeArc;
    int              arcParcouru;

    if( Trace ) {
      fprintf( FicLog , "Recursive parsing RE from %d to %d depth %d\n" ,
	       p_indMotCou , p_indMotMax , p_profondeur );
    }

    /* on peut depasser l'indice du mot maximal d'une unite si on est
       en fin d'ER et qu'il ne reste que des arcs n'attendant pas
       de lemmes a reconnaitre */
    if( p_indMotCou > p_indMotMax )
    {
        return _False;
    }
    if( p_profondeur > 100 )
    {
        fprintf( FicLog , 
                 "\n************ ====> Boucle infinie\n" );
        TxtErreur( 197 , "AnalyserER(alanalut)" );
        return _False;
    }

    cats = 0;
    /* on ne peut stocker les categories que si l'on n'a pas
       depasse la fin de la phrase */
    if( p_indMotCou < p_indMotMax ) {
      /* cats est l'indice du tableau tabCats qui stocke l'ensemble des
	 categories du lemme courant */
      ptLemme = p_ptMotCou->PtListeLemmes;
      /* on stocke l'ensemble des categories des lemmes courants */
      while( ptLemme && cats < _NbMaxHomographe ) {
        /* seuls les mots flechis sont pris en compte ici (les mots
           derives ne sont pas consideres dans les expressions
           regulieres) */
        if( ptLemme->ValTerme.TypeTerme == _MotFlechi ) {
	  catCou = CategorieLexicale( &(ptLemme->ValTerme.ArcsFils) );
	  refCou = ReferenceLexicale( &(ptLemme->ValTerme.ArcsFils) );
	  trouve = _False;
	  
	  /* on recherche si cette categorie est deja connue */
	  for( indCat = 0 ; indCat < cats ; indCat++ ) {
	    if( tabCats[ indCat ] == catCou && tabRefs[ indCat ] == refCou ) {
	      trouve = _True;
	      break;
	    }
	  }
	  
	  /* categorie non encore trouvee */
	  if( !trouve ) {
	    tabCats[ cats ] = catCou;
	    tabRefs[ cats ] = refCou;
	    cats++;
	  }
	}
        /* seuls les mots flechis sont pris en compte ici */
	
	ptLemme = ptLemme->LemmeFrere;
      }
    }

    if( cats >= _NbMaxHomographe )
    {
        TxtErreur( 198 , "AnalyserER(alanalut)" );
        exit( 0 );
        return _False;
    }

    courtCircuitInterd = _False;
    courtCircuitOblig = _False;
    if( p_typeArcPrec == _ArcDebComptage && NbBoucEnCours > 0 )
    {
        /* par construction, le court-circuit est le dernier 
           fils de la liste des freres courants */
        if( TBoucEnCours[ NbBoucEnCours ].NbFaites 
               < TBoucEnCours[ NbBoucEnCours ].Min ) 
        {
            courtCircuitInterd = _True;
        }
        if( TBoucEnCours[ NbBoucEnCours ].NbFaites 
               >= TBoucEnCours[ NbBoucEnCours ].Max
            || p_indMotCou >= p_indMotMax )
        {
            courtCircuitOblig = _True;
        }
    }

    ptFilsCou = *p_hdlArc;
    /* essai sur tous les arcs d'une fraternite sauf 
       sur le dernier dans le cas ou le parcours de 
       court-circuit n'est pas autorise */
    /* le court-circuit est le dernier fils par construction */
    if( courtCircuitOblig )
    {
       while( ptFilsCou && ptFilsCou->Frere )
       {
           ptFilsCou = ptFilsCou->Frere;
       }
    }

    arcParcouru = 0;
    /* essai sur tous les arcs d'une fraternite */
    while( ptFilsCou 
           && !(courtCircuitInterd && !(ptFilsCou->Frere)) )
    {
        /* cas d'un arc */
        if( ptFilsCou->TypeDag == _Arc )
        {
            refCatArc = (int)ptFilsCou->ValDag.Arc.ValSymb;
            refArc = ptFilsCou->ValDag.Arc.ValRef;
            typeArc = ptFilsCou->ValDag.Arc.Comptage; 
	    arcParcouru++;

            if( typeArc == _ArcSuiteComptage )
            {
                (TBoucEnCours[ NbBoucEnCours ].NbFaites)++;
            }
            else if( typeArc == _ArcFinComptage )
            {
                if( TBoucEnCours[ NbBoucEnCours ].NbFaites 
                       > TBoucEnCours[ NbBoucEnCours ].Max 
                    || TBoucEnCours[ NbBoucEnCours ].NbFaites 
                           < TBoucEnCours[ NbBoucEnCours ].Min ) 
                {
                    if( Trace )
                    {
                         fprintf( FicLog , 
                                  "Loop exit %d %d/(%d-%d)\n" , 
                                  NbBoucEnCours , 
                                  TBoucEnCours[ NbBoucEnCours ].NbFaites , 
                                  TBoucEnCours[ NbBoucEnCours ].Min , 
                                  TBoucEnCours[ NbBoucEnCours ].Max );
                    }
                    return _False;
                }
                if( Trace )
                {
                     fprintf( FicLog , "Loop end %d\n" , NbBoucEnCours );
                }
            }
            else if( typeArc == _ArcDebComptage )
            {
                if( p_typeArcPrec != _ArcSuiteComptage ) 
                {
                    NbBoucEnCours++;
                    if( NbBoucEnCours > _MaxBoucEnCours )
                    {
                        TxtErreur( 329 , "AnalyserER(alanalut)" );
                        exit( 0 );
                    }
                    TBoucEnCours[ NbBoucEnCours ].Min 
                       = ptFilsCou->ValDag.Arc.ValMin;
                    TBoucEnCours[ NbBoucEnCours ].Max
                       = ptFilsCou->ValDag.Arc.ValMax;
                    TBoucEnCours[ NbBoucEnCours ].NbFaites = 0;
                }
                if( Trace )
                {
                    if( p_typeArcPrec != _ArcSuiteComptage ) 
                    {
                        fprintf( FicLog , "Loop start %d (%d-%d)\n" , 
                              NbBoucEnCours , ptFilsCou->ValDag.Arc.ValMin ,
                              ptFilsCou->ValDag.Arc.ValMax );
                    }
                    else
                    {
                        fprintf( FicLog , "Loop cont'd %d  %d/(%d-%d)\n" , 
                              NbBoucEnCours , 
                              TBoucEnCours[ NbBoucEnCours ].NbFaites ,
                              ptFilsCou->ValDag.Arc.ValMin ,
                              ptFilsCou->ValDag.Arc.ValMax );
                    }
               }
            }

            /* arc vide, on passe au suivant en restant sur le meme mot */
            if( refCatArc == _Nil )
            {
                if( Trace )
                {
                   fprintf( FicLog , "Empty cat position %d\n" ,
                                          p_indMotCou );
                }
                if( AnalyserER( &(ptFilsCou->Fils) , 
                                p_indMotCou , p_ptMotCou , 
				p_indMotMax , p_profondeur + 1 ,
                                (int)typeArc ) )
                {
                    return _True;
                }
            }
            /* arc vide, on passe au suivant en restant sur le meme mot */
            else 
            /* arc plein on verifie si un des lemmes du mot courant 
               a sa categorie */
            {
                char     trouve;

                estNeg = ptFilsCou->ValDag.Arc.ArcERNegation;
                trouve = _False;

	        /* parcours des categories non redondantes des lemmes 
                   courants contenues dans tabCats */
                for( indCat = 0 ; indCat < cats ; indCat++ )
                {
		  if( Trace ){
		    fprintf( FicLog , "Cat curr word %d rank %d\n" , 
			     tabCats[ indCat ] , indCat );
		  }

		  /* un des homographes a la categorie requise */
		  if( ( !estNeg && tabCats[ indCat ] == refCatArc
			&& (refArc == _Nil || tabRefs[ indCat ] == refArc) )
		      || ( estNeg && tabCats[ indCat ] != refCatArc 
			&& (refArc == _Nil || tabRefs[ indCat ] != refArc) )
		      || refCatArc == RefCatLexVide
		      || tabCats[ indCat ] == RefCatLexVide
		      || tabCats[ indCat ] == _Nil )
		    {
		      trouve = _True;

		      if( Trace ) {
			if( !estNeg ) {
			  fprintf( FicLog ,
       "Cat expected (%d/%ld,%s) encountered %d/%ld position %d\n" ,
				   refCatArc , refArc , 
				   ptFilsCou->ValDag.Arc.Ident , 
				   tabCats[ indCat ] , tabRefs[ indCat ] , 
				   p_indMotCou );
			}
			else{
			  fprintf( FicLog ,
       "Cat negated (%d/%ld,%s) encountered %d/%ld position %d\n" ,
				   refCatArc , refArc , 
				   ptFilsCou->ValDag.Arc.Ident , 
				   tabCats[ indCat ] , tabRefs[ indCat ] , 
				   p_indMotCou );
			}
		      }
		      if( !estNeg ) {
			/* poursuite recursive sur l'arc et le
			   mot suivants */
			if( AnalyserER( &(ptFilsCou->Fils) , 
					p_indMotCou + 1 , 
					p_ptMotCou->MotFrere ,	p_indMotMax , 
                                        p_profondeur + 1 , (int)typeArc ) ) {
			  return _True;
			}
		      }
		      /* on n'avance pas sur les negations, elles
			 sont en sequence suivies d'une categorie vide */
		      else {
			/* poursuite recursive sur l'arc et le
			   mot suivants */
			if( AnalyserER( &(ptFilsCou->Fils) , 
					p_indMotCou , p_ptMotCou , 
					p_indMotMax , p_profondeur + 1  , 
					(int)typeArc ) ) {
			  return _True;
			}
		      }

		      /* on a trouve, inutile de poursuyivre le 
			 parcours des categories */
		      break;
		    }
		  /* un des homographes a la categorie requise */
                }
	        /* parcours des categories non redondantes des lemmes 
                   courants contenues dans tabCats */

                if( !trouve && Trace ) {
		  if( !estNeg ) {
		    fprintf( FicLog , 
          "Cat expected (%d/%ld,%s) not encountered position %d\n" ,
			     refCatArc , refArc , 
			     ptFilsCou->ValDag.Arc.Ident , 
			     p_indMotCou );
		  }
		  else {
		    fprintf( FicLog ,
          "Cat rejected (%d/%ld,%s) not accepted position %d\n" ,
			     refCatArc , refArc , 
			     ptFilsCou->ValDag.Arc.Ident , 
			     p_indMotCou  );
		  }
                }
            }
            /* arc plein on verifie si un des lemmes du mot courant a sa
	       categorie */
        }
        /* cas d'un arc */
        else if( ptFilsCou->TypeDag == _AtomeChaine )
        /* cas d'un atome */
        {
            typeArc = _ArcOrdinaire;

	    /* on a atteint la fin de l'empan */
            if( p_indMotCou >= p_indMotMax )
	    {
                if( Trace )
                {
                    fprintf( FicLog , "End of parse RE\n" );
                }
                return _True;
            }
            if( Trace )
            {
                fprintf( FicLog , "Early exit of parse RE\n" );
            }
            /* ne doit pas avoir de freres: impasse terminale */
            return _False;
        }
        /* cas d'un atome */
        else
        /* autres cas: erreur */
        {
            fprintf( FicLog , 
                     "\n************ ====> Arc de type inconnu\n" );
            TxtErreur( 197 , "AnalyserER(alanalut)" );
            return _False;
        }
        /* autres cas: erreur */

        ptFilsCou = ptFilsCou->Frere;
    }
    /* essai sur tous les arcs d'une fraternite */

    /* descente en backtracking, abandon de la boucle courante */
    if( arcParcouru
	&& typeArc == _ArcDebComptage 
        && p_typeArcPrec != _ArcSuiteComptage 
        && p_typeArcPrec != _ArcFinComptage )
    {
        if( Trace )
        {
             fprintf( FicLog , "Backtrack loop %d\n" , 
                      NbBoucEnCours );
        }
        NbBoucEnCours--;
        if( NbBoucEnCours < 0 || arcParcouru > 1 )
        {
	   fprintf( FicLog , 
              "\n******** ====> Backtrack sur indice de boucle %d < 0 (%d)\n" ,
              NbBoucEnCours , arcParcouru );
           TxtErreur( 70 , "AnalyserER(alanalut)" );
           exit( 0 );
           return _False;
        }
    }

    return _False;
}

/*************************************************************************
  CHECKS WHETHER A CATEGORY IS A MAJOR CATEGORY
**************************************************************************/
int      MajorCategory( int p_catLexTerme )
{
    int                            indCategory;
    
    for( indCategory = 0 ; indCategory < NbMajorCategories ; indCategory++ ) {
      if( p_catLexTerme == (int)RefMajorCategories[ indCategory ] ) {
	return _True;
      }
    }
    return _False;
}

/*************************************************************************
  VERIFIE QUE TOUS LES LEMMES D'UNE REGLE SONT PRESENTS DANS UNE PHRASE
**************************************************************************/
char     LemmePresentDansPhrase( TabTermeMem *p_hdlTerme )
{
    TabMotMem                      ptMotCou;
    TabLemmeMem                    ptLemmeAux;
    long                           refTerme;
    long                           refTermeRacine;
    int                            catLexTerme;

    if( (*p_hdlTerme)->ValTerme.TypeTerme != _UniteLexicale )
    {
        return _True;
    }

    catLexTerme = CategorieLexicale( &((*p_hdlTerme)->ValTerme.ArcsFils) );
    if( !MajorCategory( catLexTerme ) || catLexTerme == _Nil )
    {
        return _True;
    }
    refTerme = (*p_hdlTerme)->ValTerme.Ref;
    refTermeRacine = (*p_hdlTerme)->ValTerme.RefRacine;

/*     printf("terme %ld %ld\n",refTerme,refTermeRacine); */
    if( (refTerme == _Nil || refTerme == _RefConjonctionNeg )
        && ( refTermeRacine == _Nil || refTermeRacine == _RefConjonctionNeg ) )
    {
        return _True;
    }

    ptMotCou = MotsPhrase;
    while( ptMotCou )
    {
      if( (ptLemmeAux = ptMotCou->PtListeLemmes) )
      {
        /* parcours de tous les homographes d'un mot */
        while( ptLemmeAux )
        {
/* 	  printf("Mot %ld %ld\n",ptLemmeAux->ValTerme.Ref, */
/* 		 ptLemmeAux->ValTerme.RefRacine); */
            if( refTerme == ptLemmeAux->ValTerme.Ref
                && refTermeRacine == ptLemmeAux->ValTerme.RefRacine  )
            {
                return _True;
            }

            ptLemmeAux = ptLemmeAux->LemmeFrere;
        }
        /* parcours de tous les homographes d'un mot */
      }
      ptMotCou = ptMotCou->MotFrere;
    }
    return _False;
}

/*************************************************************************
  VERIFIE QUE TOUS LES LEMMES D'UNE REGLE SONT PRESENTS DANS UNE PHRASE
**************************************************************************/
char     NoteDebutEtFinColloc( TabTermeMem *p_hdlTerme )
{
    TabMotMem                      ptMotCou;
    TabLemmeMem                    ptLemmeAux;
    long                           refTerme;
    long                           refTermeRacine;
    int                            indMotPhrase;
    
    if( (*p_hdlTerme)->ValTerme.TypeTerme != _UniteLexicale )
    {
        return _True;
    }

    refTerme = (*p_hdlTerme)->ValTerme.Ref;
    refTermeRacine = (*p_hdlTerme)->ValTerme.RefRacine;
    if( (refTerme == _Nil || refTerme == _RefConjonctionNeg )
        && ( refTermeRacine == _Nil || refTermeRacine == _RefConjonctionNeg ) )
    {
        return _True;
    }

    ptMotCou = MotsPhrase;
    indMotPhrase = 0;
    while( ptMotCou )
    {
      if( (ptLemmeAux = ptMotCou->PtListeLemmes) )
      {
        /* parcours de tous les homographes d'un mot */
        while( ptLemmeAux )
        {
            /* seuls les mots pleins sont consideres pour les collocations */
            if( refTerme == ptLemmeAux->ValTerme.Ref 
                && refTermeRacine == ptLemmeAux->ValTerme.RefRacine )
            {
                if( PremMotColloc == _Nil )
                {
                    PremMotColloc = indMotPhrase;
                }
                else if( PremMotColloc > indMotPhrase )
                {
                    PremMotColloc = indMotPhrase;
                }
                else
                {
                    EstPermute = _True;
                }
                if( DernMotColloc == _Nil )
                {
                    DernMotColloc = indMotPhrase;
                }
                else if( DernMotColloc < indMotPhrase )
                {
                    DernMotColloc = indMotPhrase;
                }
                LongueurTerme++;
                return _True;
            }

            ptLemmeAux = ptLemmeAux->LemmeFrere;
        }
        /* parcours de tous les homographes d'un mot */
      }
      ptMotCou = ptMotCou->MotFrere;
      indMotPhrase++;
    }

    return _False;
}

/*************************************************************************
  VERIFIE QUE TOUS LES LEMMES D'UNE REGLE SONT PRESENTS DANS UNE PHRASE
**************************************************************************/
char     TousLemmesPresentsDansPhrase( TabRegleMem p_ptRegle )
{
    return ParcoursDFArbre( &(p_ptRegle->TermeFils) , LemmePresentDansPhrase , 
                            _PostOrdre );
}

/*************************************************************************
  VERIFIE QUE TOUS LES LEMMES D'UNE REGLE SONT PRESENTS DANS UNE PHRASE
**************************************************************************/
void     ChercheEtImprimeCollocation( TabRegleMem p_ptRegle ,
                                      TabCorrIdent_PosFic p_tCorrespondance ,
                                      FILE *p_ptFicSortie )
{
    int                               indMotPhrase;
    ChIdent                           ident;
    TabMotMem                         ptMotCou;

    PremMotColloc = _Nil;
    DernMotColloc = _Nil;
    LongueurTerme = 0;
    EstPermute = _False;
    ParcoursDFArbre( &(p_ptRegle->TermeFils) , NoteDebutEtFinColloc , 
                     _PostOrdre );
    if( (FormatDeSortie == _SortieInist 
	  || FormatDeSortie == _SortieInistNum
	  || FormatDeSortie == _SortieLexter
	  || FormatDeSortie == _SortieInistSGML
	  || FormatDeSortie == _SortieInistSGMLWithSem 
	  || FormatDeSortie == _SortieAmaryllis 
	  || FormatDeSortie == _SortieInistAvecLemmes 
	  || FormatDeSortie == _SortieQA )
        && LongueurTerme < DernMotColloc - PremMotColloc + 1 )
    {
         fprintf( p_ptFicSortie , "%06ld\t" , NbLignesTot );
         if( p_ptRegle && Identificateur( ident , 
                                         p_ptRegle->TermeFils->ValTerme.Ref , 
                                         FicRefIdent , p_tCorrespondance ) )
         {
              fprintf( p_ptFicSortie ,"%s\t", ident );
         }

	 ptMotCou = IndiceAMot( PremMotColloc );
	 /* ecriture de la collocation */
         for( indMotPhrase = PremMotColloc ; 
	      indMotPhrase <= DernMotColloc && ptMotCou ; 
              indMotPhrase++ )
         {
              if( indMotPhrase == DernMotColloc 
		  || FormatDeSortie == _SortieLexter )
              {
                  fprintf( p_ptFicSortie ,"%s", 
                           ptMotCou->ChaineMot );
              }
              else
              {
                  fprintf( p_ptFicSortie ,"%s ", 
                           ptMotCou->ChaineMot );
              }
	      ptMotCou = ptMotCou->MotFrere;
         }
         fprintf( p_ptFicSortie , "\t" );
         for( indMotPhrase = 0 ; indMotPhrase < LongueurTerme ;
              indMotPhrase++ )
         {
              fprintf( p_ptFicSortie , "X" );
         }
         fprintf( p_ptFicSortie , ",%d" , 
                  DernMotColloc - PremMotColloc + 1 );
         if( EstPermute )
         {
             fprintf( p_ptFicSortie , ",Perm\n" );
         }
         else
         {
             fprintf( p_ptFicSortie , ",Ins\n" );
         }
    }
}

/*************************************************************************
           INDENTITE DE D'UN TERME ET D'UN MOT DU CORPUS
**************************************************************************/

int EgaliteReferenceSimple_Multiple( long p_refSimple , TabRefPhraseMem p_refMultiple )
{
  while( p_refMultiple ) {
    if( p_refMultiple->Ref == p_refSimple ) {
      return _True;
    }
    p_refMultiple = p_refMultiple->Suivant;
  }
  return _False;
}

int EgaliteReferenceMultiple_Multiple( TabRefPhraseMem p_refMultiple1 , 
									   TabRefPhraseMem p_refMultiple2 )
{
  TabRefPhraseMem      	               p_refAux;

  while( p_refMultiple1 ) {
    p_refAux = p_refMultiple2;
    while( p_refAux ) {
      if( p_refMultiple1->Ref == p_refAux->Ref ) {
	return _True;
      }
      p_refAux = p_refAux->Suivant;
    }
    p_refMultiple1 = p_refMultiple1->Suivant;
  }
  return _False;
}


int      MotsCompatibles( long p_refTerme , long p_refTermeRacine , 
			  TabRefPhraseMem p_refTermeSemantique , 
			  TabRefPhraseMem p_refTermeSelf ,
		          long p_refLemme , long p_refLemmeRacine , 
			  TabRefPhraseMem p_refLemmeSemantique , 
			  TabRefPhraseMem p_refLemmeSelf ,
                          int p_catLexTerme , int p_catLexLemme ,
		          int p_typeLemme )
{
  /* en cas de morphologie robuste, on egalise un mot avec un
     paradigme obtenu par lemmatisation brutale que s'il s'agit
     d'un nom d'un verbe ou d'un adjectif et non d'une
     ponctuation, d'une coordination, d'un determinant.... */ 
  if( SuffStripping )
  {
     return MajorCategory( p_catLexTerme) || p_catLexTerme == _Nil;
  }

  if( p_typeLemme != _MotFlechi ) {
    return _False;
  }

  /* 1. terme de regle inconnu */
  /* sa reference et sa reference de racine sont inconnues */
  if( ( p_refTerme == _Nil || p_refTerme == _RefConjonctionNeg )
      && ( p_refTermeRacine == _Nil 
	   || p_refTermeRacine == _RefConjonctionNeg ) 
      && !p_refTermeSemantique
      && !p_refTermeSelf ) {
     /* en cas de terme non connu, l'egalite se verifie sur les 
	categories lexicales : une des deux vides ou les deux egales */
    if( ( ( p_catLexTerme == _Nil || p_catLexLemme == _Nil
	    || p_catLexTerme == RefCatLexVide 
	    || p_catLexLemme == RefCatLexVide )
	  || p_catLexTerme == p_catLexLemme ) ) {
/* printf("1\n"); */
      return _True;
    }
    else {
      return _False;
    }
  }
  
  /* 2.1 le terme est un lien par la racine. (Cette racine n'est pas
     vide.)  */
  /* La racine du terme et celle du lemme doivent etre identiques */
  /* ou la racine du terme doit etre egale au lemme ou a son champ self */
  else if( ( p_refTerme == _Nil || p_refTerme == _RefConjonctionNeg ) 
	   && ! p_refTermeSelf  ) {
    if( p_refTermeRacine != _Nil 
	&& p_refTermeRacine != _RefConjonctionNeg
	&& ( p_refTermeRacine == p_refLemme 
	     || p_refTermeRacine == p_refLemmeRacine 
	     || EgaliteReferenceSimple_Multiple ( p_refTermeRacine ,
					     p_refLemmeSelf ) ) ) {
/* printf("2\n"); */
      return _True;
    }

    /* 2.2 le terme est un lien par synonymie. (Ce lien n'est pas
       vide.)  */
    /* Le lien du terme et celui du lemme doivent etre identiques */
    /* ou le lien du terme doit etre egale au lemme ou a son champ self */
    else if( p_refTermeSemantique
	     && ( EgaliteReferenceSimple_Multiple( p_refLemme ,
						   p_refTermeSemantique )
		  || EgaliteReferenceMultiple_Multiple( p_refTermeSemantique ,
							p_refLemmeSemantique )
		  || EgaliteReferenceMultiple_Multiple ( p_refTermeSemantique ,
							 p_refLemmeSelf ) ) ) {
/* printf("3\n"); */
      return _True;
    }
    else {
      return _False;
    }
  }
  
  /* 3.a le terme est un element lexical connu par sa reference */
  else if( p_refTerme !=  _Nil && p_refTerme != _RefConjonctionNeg ) {
    if( p_refTerme == p_refLemme || p_refTerme == p_refLemmeRacine 
	|| EgaliteReferenceSimple_Multiple ( p_refTerme ,
					     p_refLemmeSemantique ) 
	|| EgaliteReferenceSimple_Multiple ( p_refTerme ,
					     p_refLemmeSelf ) ) {
/* printf("4\n"); */
      return _True;
    }
    else {
      return _False;
    }
  }

  /* 3.b le terme est un element lexical connu par son trait self */
  else if( p_refTermeSelf ) {
    if( EgaliteReferenceSimple_Multiple ( p_refLemme ,
					  p_refTermeSelf ) 
	|| EgaliteReferenceMultiple_Multiple ( p_refTermeSelf ,
					       p_refLemmeSelf ) ) {
/* printf("5\n"); */
      return _True;
    }
    else {
      return _False;
    }
  }

  /* cas residuel inattendu */
  else {
    printf( "Unexpected term/lemma comparison\n" );
    TxtErreur( 3111 , "MotsCompatibles(alanalut)" );
    return _False;
  }
}

/*************************************************************************
     ANALYSE D'UNE EXP REG ENTRE DEUX BORNES
**************************************************************************/
int      LaunchREParsing( int p_indMotPhrase , int *p_ptIndDebER ,
			  TabDagMem *p_hdlFilsLeadingER , int  p_leadingER )
{
  int                   analyseEROk;
  TabMotMem             ptMotDebER;

  if( Trace ) {
    fprintf( FicLog , "Parsing RE from %d to %d\n" ,
	     IndiceMotCou , p_indMotPhrase );
  }

  NbBoucEnCours = 0;
  analyseEROk = _False;

  /* ER precedent une unite lexicale */
  if( p_leadingER ) {

    /* ER en tete de variation */
    if( IndiceMotCou == _Nil ) {
      *p_ptIndDebER = 0;
      ptMotDebER = MotsPhrase;
      while( ((*p_ptIndDebER) <= p_indMotPhrase) && ptMotDebER ) {
	analyseEROk = AnalyserER( p_hdlFilsLeadingER , *p_ptIndDebER , 
				  ptMotDebER , p_indMotPhrase , 1 , 
				  _ArcOrdinaire );
	if( analyseEROk ) {
	  break;
	}
	(*p_ptIndDebER)++;
	ptMotDebER = ptMotDebER->MotFrere;
      }
    }

    /* ER apres un mot */
    else {
      *p_ptIndDebER = IndiceMotCou;
      analyseEROk = AnalyserER( p_hdlFilsLeadingER , *p_ptIndDebER ,
				IndiceAMot( *p_ptIndDebER ) , 
				p_indMotPhrase , 1 , _ArcOrdinaire );
    }
  }
  /* ER en fin de meta-regle ne pas modifier le debut de l'ER */
  else {
      analyseEROk = AnalyserER( p_hdlFilsLeadingER , *p_ptIndDebER ,
				IndiceAMot( *p_ptIndDebER ) , p_indMotPhrase , 
				1 , _ArcOrdinaire );
  }
		  
  if( analyseEROk ) {
    if( Trace ) {
      fprintf( FicLog , "Successfull parse of RE\n" );
    }
  }
  else {
    *p_ptIndDebER = _Nil;
    if( Trace ) {
      fprintf( FicLog , "Unsuccessfull parse of RE\n" );
    }
  }
  return analyseEROk;
}

/*************************************************************************
     RECHERCHE SI UN MOT FLECHI EST PRESENT DANS LA PHRASE
               EN FONCTION DE IndiceMotCou
**************************************************************************/

int      MotFlechiPresentDansLaPhrase( TabTermeMem p_ptTerme , 
                                       TabLemmesMots p_tLemmesMots )
{
    int                            indMotPhrase;
    int                            indDebutER;
    int                            indFinER;
    int                            indDebutTrailingER;
    int                            empan;
    TabLemmeMem                    ptLemmeAux;
    int                            lemmeTrouve;
    int                            catLexTerme;
    long                           refTerme;
    long                           refTermeRacine;
    TabRefPhraseMem                refTermeSemantique;
    TabRefPhraseMem                refTermeSelf;
    TabDagMem                     *hdlFilsLeadingER;
    TabDagMem                     *hdlFilsTrailingER;
    int                            nbHomographes;
    TabMotMem                      ptMotCou;
    int                            indIni;
    int                            indFin;

    catLexTerme = CategorieLexicale( &(p_ptTerme->ValTerme.ArcsFils) );
    hdlFilsLeadingER = ExisteCheminFils( CheminLeadingExprReg , 
                                  &(p_ptTerme->ValTerme.ArcsFils) );
    hdlFilsTrailingER = ExisteCheminFils( CheminTrailingExprReg , 
                                  &(p_ptTerme->ValTerme.ArcsFils) );
    refTerme = p_ptTerme->ValTerme.Ref;
    refTermeRacine = p_ptTerme->ValTerme.RefRacine;
    refTermeSemantique = p_ptTerme->ValTerme.RefSemantique;
    refTermeSelf = p_ptTerme->ValTerme.RefSelf;

    /* si on a une expression reguliere suivie ou non d'un ancrage
       lexical le lemme suivant, s'il existe, peut etre cherche plus
       loin que le suivant strict pour permettre, dans l'intervalle
       entre les deux la reconnaissance d'une ER basee sur les
       categories lexicales. */
    if( hdlFilsLeadingER || hdlFilsTrailingER ) {
        empan = NbMots - IndiceMotCou;
    }
    /* si on est sur le premier mot, on peut rechercher jusqu'a la fin de
       la phrase */
    else if( IndiceMotCou == (int)_Nil )
    {
      if( FormatDEntree != _TxtFenTaggePartielt ) {
        empan = NbMots;
      }
      /* in case of a window, the occurrence must begin at
	 the first word of the window */
      else {
	IndiceMotCou = 0;
        empan = 1;
      }
    }
    else
    {
        empan = 1;
    }

    if( Trace )
    {
      fprintf( 
	      FicLog , 
	      "\nSearch for word ref %ld/%ld cat %d (%s) from %d to %d\n" ,
	      refTerme , refTermeRacine , catLexTerme ,
	      p_ptTerme->ValTerme.Ident , 
	      ((IndiceMotCou == (int)_Nil)? 0 : IndiceMotCou ) ,
	      ((IndiceMotCou 
		== (int)_Nil)? NbMots : IndiceMotCou + empan) - 1 );
    }

    nbHomographes = 0;

    /* parcours de tous les mots de la phrase */
    indIni =  ((IndiceMotCou == (int)_Nil)? 0 : IndiceMotCou );
    indFin =  ((IndiceMotCou == (int)_Nil)? NbMots : IndiceMotCou + empan);
    ptMotCou = IndiceAMot( indIni );
    /* saut des premiers mots */
    for( indMotPhrase = 0 ; 
	 indMotPhrase < PremMotColloc && ptMotCou ; 
	 indMotPhrase++ ) {
      ptMotCou = ptMotCou->MotFrere;
    }
    for( indMotPhrase = indIni ;
         indMotPhrase < indFin && indMotPhrase < NbMots && ptMotCou
              && nbHomographes < _NbMaxHomographe ; indMotPhrase++ ) {

      /* mot connu eventuellement precede d'une ER */
      if( (ptLemmeAux = ptMotCou->PtListeLemmes) ) 
      {
        /* parcours de tous les homographes d'un mot */
        while( ptLemmeAux )
        {
            /* On trouve un lemme de meme reference que le terme de
               la regle */
	    if( MotsCompatibles( 
			 refTerme , refTermeRacine ,
			 refTermeSemantique , refTermeSelf ,
			 ptLemmeAux->ValTerme.Ref ,
			 ptLemmeAux->ValTerme.RefRacine ,
			 ptLemmeAux->ValTerme.RefSemantique ,
			 ptLemmeAux->ValTerme.RefSelf ,
			 catLexTerme , 
			 CategorieLexicale( &(ptLemmeAux->ValTerme.ArcsFils) ),
			 ptLemmeAux->ValTerme.TypeTerme ) )
            {
                if( Trace )
                {
                    fprintf( FicLog , 
    "Word encountered ref %ld/%ld cat %d (%s): \n\
     ref %ld/%ld cat %d type %d (%d) position %d\n" ,
    refTerme , refTermeRacine , catLexTerme , ptMotCou->ChaineMot , 
    ptLemmeAux->ValTerme.Ref , ptLemmeAux->ValTerme.RefRacine ,
    CategorieLexicale( &(ptLemmeAux->ValTerme.ArcsFils) ),
    ptLemmeAux->ValTerme.TypeTerme , _MotFlechi , indMotPhrase );
                }

		lemmeTrouve = _False;
		indDebutER = _Nil;
                if( hdlFilsLeadingER && indMotPhrase >= IndiceMotCou )
                {
		  lemmeTrouve = LaunchREParsing( indMotPhrase , 
						 &indDebutER ,
						 hdlFilsLeadingER ,
						 _True );
                }
		else 
		{
		  lemmeTrouve = _True;
		}

		/* cas d'ER terminale */
		indFinER = _Nil;
		if( lemmeTrouve && hdlFilsTrailingER ) {
		  lemmeTrouve = _False;
		  for( indFinER  = indMotPhrase + 1 ;
		       indFinER <= NbMots ;
		       indFinER++ ) {
		    indDebutTrailingER = indMotPhrase + 1;
		    if( LaunchREParsing( indFinER , 
					 &indDebutTrailingER ,
					 hdlFilsTrailingER , _False ) ) {
		      lemmeTrouve = _True;
		      break;
		    }
		  }
		}

 		/* si lemme trouve: on le stocke */
		if( lemmeTrouve )
		{
		  /* tLemmesMots est tableau de pointeurs vers des paires
		     lemme, mot de la phrase, dont la reference est celle du
		     terme attendu (pour les caract lex-synt elles seront
		     verifiees lors de l'unification) */
		  p_tLemmesMots[ nbHomographes ].PtLemme = ptLemmeAux;
		  p_tLemmesMots[ nbHomographes ].IndMot = indMotPhrase;
		  p_tLemmesMots[ nbHomographes ].IndDebutER = indDebutER;
		  p_tLemmesMots[ nbHomographes ].IndFinER = indFinER;

		  nbHomographes++;
		  if( nbHomographes >= _NbMaxHomographe )
		  {
		    break;
		  }
		}
 		/* si lemme trouve: on le stocke */
	    }
            /* On trouve un lemme de meme reference que le terme de
               la regle */

            ptLemmeAux = ptLemmeAux->LemmeFrere;
        }
        /* parcours de tous les homographes d'un mot */
      }
      /* mot connu */
      ptMotCou = ptMotCou->MotFrere;
    }
    /* parcours de tous les mots de la phrase */

   /* on retourne le nombre d'homographes dont la ref
       est egale a celle attendue */
    return nbHomographes;
}

/*************************************************************************
     SAUVEGARDE DE L'ENVIRONNEMENT POUR RESTAURATION AU BACKTRACKING
**************************************************************************/

void     SauverEnvironnement( TabBlocInitial p_ptBlocIniCou ,
               TabBlocInitial p_ptBlocIniPere )
{
    TabTermeMem                *hdlTermeAux;

    switch( p_ptBlocIniPere->typeBloc )
    {
      case _RegleNormale :
      case _RegleElision :
      case _RegleMeta :
          /* lien automatique avec le bloc suivant */
          break;
      case _Regle :
          /* au retour en arriere T2 sert a stocker la metaregle suivante et
             a desallouer la regle courante */
          break;
      case _Feuille :           /* paradigme defini en intension */
      case _UniteLexicale :     /* lemme dont on connait la reference */
      case _Liste :
      case _Concatenation :
          /* on sauve le pointeur Ti+1 du pere */
          hdlTermeAux = (TabTermeMem *)
                  ((TabChar)p_ptBlocIniPere + sizeof(struct BlocInitial));
          hdlTermeAux += p_ptBlocIniPere->noChoix;
          p_ptBlocIniCou->ptSauve1 = (TabChar)(*hdlTermeAux);
          break;
      default :
          if( Trace )
          {
              fprintf( FicLog ,
               "Unexpected block type: %d\n",p_ptBlocIniPere->typeBloc);
          }
          sprintf( ChAux , "SauverEnvironnement(alanalut) [%d]\n" , 
                           p_ptBlocIniPere->typeBloc);
          TxtErreur( 1 , ChAux);
          break;
    }

    /* sauvegarde des 2 variables globales */
    p_ptBlocIniCou->indiceMotCou = IndiceMotCou;
    p_ptBlocIniCou->debutMotsReconnusCou = DebutMotsReconnus;
    p_ptBlocIniCou->finMotsReconnusCou = FinMotsReconnus;
}

/*************************************************************************
     RESTAURATION DE L'ENVIRONNEMENT
**************************************************************************/

void     RestaurerEnvironnement( TabBlocInitial p_ptBlocIniCou ,
               TabBlocInitial p_ptBlocIniPere )
{
    TabTermeMem                *hdlTermeAux;

    switch( p_ptBlocIniPere->typeBloc )
    {
    case _RegleNormale :
    case _RegleElision :
    case _RegleMeta :
        /* lien automatique avec le bloc suivant */
        break;
    case _Regle :
        /* au retour en arriere T2 sert a stocker la metaregle suivante et
           a desallouer la regle courante */
        break;
    case _Feuille :             /* paradigme defini en intension */
    case _UniteLexicale :       /* lemme dont on connait la reference */
    case _Liste :
    case _Concatenation :
        /* on restaure le pointeur Ti+1 du pere */
        hdlTermeAux = (TabTermeMem *)
                ((TabChar)p_ptBlocIniPere + sizeof(struct BlocInitial));
        hdlTermeAux += p_ptBlocIniPere->noChoix - 1;
        *hdlTermeAux = (TabTermeMem)p_ptBlocIniCou->ptSauve1;

        if( p_ptBlocIniPere->typeBloc == _Concatenation )
        {
            /* on decremente les choix du pere */
            (p_ptBlocIniPere->noChoix)--;
        }
        break;
    default :
        sprintf( ChAux , "RestaurerEnvironnement(alanalut) [%d]" , 
                 p_ptBlocIniPere->typeBloc );
        TxtErreur( 2 , ChAux );
    break;
    }

    /* restauration des 2 variables globales */
    IndiceMotCou = p_ptBlocIniCou->indiceMotCou;
    DebutMotsReconnus = p_ptBlocIniCou->debutMotsReconnusCou;
    FinMotsReconnus = p_ptBlocIniCou->finMotsReconnusCou;
}

/*************************************************************************
      ecriture des chemins semantiques en cas de sortie riche
**************************************************************************/
int   ImprimeCheminsSemantiques( struct ValTerme *p_ptValTerme , 
				 FILE *p_ptFicSortie )
{
   int                     indCheminSemantique;
   TabDagMem              *hdlDag;

   indCheminSemantique = 0;
   /* parcours de tous les chemins semantiques */
   fprintf( p_ptFicSortie, "<sem>" );
   while( CheminsSemantique[indCheminSemantique ] ) {
     hdlDag = ExisteCheminFils( CheminsSemantique[ indCheminSemantique ] , 
				&(p_ptValTerme->ArcsFils) );

     /* parcours de toutes les listes d'une disjonction */
     /* la chaine du lemme pointe par un lien semantique
	est la deuxieme composante de la liste */
     while( hdlDag 
	    && *hdlDag && (*hdlDag)->TypeDag == _AtomeListe
	    && ((*hdlDag)->ValDag.Atome.AtomeConjDisj) 
	    != _ConjonctionNeg 
	    && (*hdlDag)->Fils && ((*hdlDag)->Fils)->Frere ) {
       EcrireDag( CheminsSemantique + indCheminSemantique , 
		  _False , p_ptFicSortie , _SortieWithoutLF );
       fprintf( p_ptFicSortie, ">='%s'" , 
		(((*hdlDag)->Fils)->Frere)->ValDag.Atome.ValAtome.Chaine );

       hdlDag = &((*hdlDag)->Frere);
     }
     /* parcours de toutes les listes d'une disjonction */

     indCheminSemantique++;
   }
   fprintf( p_ptFicSortie, "</sem>" );
   return _True;
}

/*************************************************************************
      ecriture des chaines analysees correctement ou generees
**************************************************************************/
void     EcrireChaineAnalysee( int p_estGeneration ,
                               FILE *p_ptFicSortie )
{
     int                     ind;
     TabMotMem               ptMot;

     for( ind = DebutMotsReconnus ; ind <= FinMotsReconnus ; ind++ )
     {
       if( ind > 0 && FormatDeSortie != _SortieLexter ) {
	 fprintf( p_ptFicSortie, " " );
	 if( Trace ) {
	   fprintf( FicLog , " " );
	 }
       }
       if( FormatDeSortie == _SortieInist 
	   || FormatDeSortie == _SortieLexter 
	   || FormatDeSortie == _SortieInistSGML 
	   || FormatDeSortie == _SortieInistNum 
	   || FormatDeSortie == _SortieInistSGMLWithSem
	   || FormatDeSortie == _SortieAmaryllis
	   || FormatDeSortie == _Txt 
	   || FormatDeSortie == _SortieInistAvecLemmes 
	   || FormatDeSortie == _SortieQA ) {
	 ptMot = IndiceAMot( ind );
	 fprintf( p_ptFicSortie, "%s" , ptMot->ChaineMot );
	 if( FormatDeSortie == _SortieInistAvecLemmes
	     || FormatDeSortie == _SortieQA
	     || FormatDeSortie == _SortieInistNum ) {
	   int catLex 
	     = CategorieLexicale( 
			 &(ptMot->PtListeLemmes->ValTerme.ArcsFils) );
	   fprintf( p_ptFicSortie, "[%s][%s]" , 
                  ChaineLemme( &(ptMot->PtListeLemmes->ValTerme.ArcsFils) ) ,
	          DicoTraitsCatLex[ catLex ].ChaineDico );
	 }
	 
	 /* displays the semantic features of each lemma */
	 if( FormatDeSortie == _SortieInistSGMLWithSem ) {
	   TabLemmeMem             ptLemme;
	   
	   ptMot = IndiceAMot( ind );
	   ptLemme = ptMot->PtListeLemmes;
	   
	   /* on ne parcourt que le premier lemme du mot
	      en supposant que les suivants sont, non pas
	      des homographes, mais des voisins morphologiques
	      ou semantiques 
	      while( ptLemme ) { */
	   ImprimeCheminsSemantiques( &(ptLemme->ValTerme) ,
				      p_ptFicSortie );
	 }
	 /* displays the semantic features of each lemma */
       }
       if( Trace ) {
	 ptMot = IndiceAMot( ind );
	 fprintf( FicLog , "%s" , ptMot->ChaineMot );
       }
     }

     if( FormatDeSortie == _SortieInistNum 
	 || FormatDeSortie == _SortieQA  ) {
       fprintf( p_ptFicSortie , "\t" );
       for( ind = DebutMotsReconnus ; ind <= FinMotsReconnus ; ind++ ) {
	 if( ind > 0 ) {
	   fprintf( p_ptFicSortie, " " );
	   if( Trace ) {
	     fprintf( FicLog , " " );
	   }
	 }
	 fprintf( p_ptFicSortie, "%d" , ind );
       }
     }

     if( FormatDeSortie == _Txt )
     {
          fprintf( p_ptFicSortie , "\n" );
     }
     else
     {
          fprintf( p_ptFicSortie , "\t" );
     }
     if( Trace )
     {
          fprintf( FicLog , "\t" );
     }

     NbMaxGenerations++;
     if(NbMaxGenerations == 200 && p_estGeneration)
     {
           TxtAlerte( 12 , 0 );
           exit(0);
     }
}

/*************************************************************************
      ecriture des regles utilisees
**************************************************************************/
char     EcrireReglesUtilisees( TabBlocInitial p_ptBlocIni , 
                                TabCorrIdent_PosFic p_tCorrespondance ,
                                int p_typeIndexation , int p_regles ,
                                int p_metaRegles , FILE *p_ptFicSortie )
{
     long                            ptMem = 0;
     TabRegleMem                     *hdlRegleDest;
     ChIdent                         ident;
     char                            noChoixPere;
     char                           *ptChIdent;
     char                           *ptChType;

     ptMem = p_ptBlocIni->ptBlocPrecedent;
     if( ptMem )
     {
         if( EcrireReglesUtilisees( (TabBlocInitial)(Pile + ptMem) , 
                                     p_tCorrespondance ,
                                     p_typeIndexation , p_regles , 
                                     p_metaRegles , p_ptFicSortie ) )
         {
             /* en extraction  terminologique, 
                on n'ecrit que le terme de base */
             if( FormatDeSortie != _Txt )
             {
                 return _True;
             }
         }
     }

     /* parcours remontant */
     if( p_ptBlocIni->typeBloc == _Regle )
     {
          hdlRegleDest = (TabRegleMem *)((TabChar)p_ptBlocIni 
                                         + sizeof(struct BlocInitial));
          if( p_typeIndexation ==  _Variations
              || p_typeIndexation ==  _Analyse
              || p_typeIndexation ==  _IndexationVariations
              || p_typeIndexation ==  _Indexation )
          {
              noChoixPere = p_ptBlocIni->noChoix;

              if( p_metaRegles )
              {
                 if( noChoixPere > 1 )
                 {
                    if( ChaineMetaRegle( 
                           (*hdlRegleDest)->TermeFils->ValTerme.ArcsFils , 
                           noChoixPere - 2 , &ptChIdent , &ptChType ) )
                    {
                         if( FormatDeSortie == _Txt )
                         {
                             fprintf( p_ptFicSortie ,"[%s,%d,%s]" , ptChIdent ,
                                      noChoixPere - 1 , ptChType );
                         }
                         else
                         {
                             fprintf( p_ptFicSortie , "%s,%d,%s" , ptChIdent , 
                                      noChoixPere - 1 , ptChType );
                         }
                         if( Trace )
                         {
                              fprintf( FicLog , "%s,%d,%s" , ptChIdent , 
                                    noChoixPere - 1 , ptChType );
                         }
                    }
                    else
                    {
                         if( FormatDeSortie == _Txt )
                         {
                              fprintf( p_ptFicSortie ,"[?,?]" );
                         }
                         else
                         {
                              fprintf( p_ptFicSortie , "?,?" );
                         }
                         if( Trace )
                         {
                              fprintf( FicLog , "?,?" );
                         }
                    }
                 }
                 else
                 {
                    if( FormatDeSortie != _Txt )
                    {
                         fprintf( p_ptFicSortie , "0" );
                    }
                    if( Trace )
                    {
                         fprintf( FicLog , "0" );
                    }
                 }
              }
          }

	  /* ecriture du terme dont l'occurrence ou la variante
	     est trouvee */
          if( p_regles && Identificateur( ident , 
                                   (*hdlRegleDest)->TermeFils->ValTerme.Ref , 
                                   FicRefIdent , p_tCorrespondance ) ) {
	   
	    /* ecriture de l'identifieur du terme */
	    if( FormatDeSortie == _Txt ) {
	      fprintf( p_ptFicSortie ,"<%s> ", ident );
	    }
	    /* on ecrit les mots du terme avec leurs traits
	       semantiques */
	    else if( FormatDeSortie == _SortieInistSGMLWithSem ) {
	      TabTermeMem            ptTermeCou;
	      TabChar                chMotCou;

	      ptTermeCou = (*hdlRegleDest)->TermeFils;  /* the head word */
	      if( ptTermeCou ) {
		ptTermeCou = ptTermeCou->TermeFils; /* the 1st daughter node */
	      }
	      while( ptTermeCou ) {
		if( (chMotCou = 
		     ChaineLemme( &(ptTermeCou->ValTerme.ArcsFils) ) ) ) {
		  fprintf( p_ptFicSortie , "%s", chMotCou );
		}

		/* prints semantic paths */
		if( FormatDeSortie == _SortieInistSGMLWithSem ) {
		  ImprimeCheminsSemantiques( &(ptTermeCou->ValTerme) ,
					     p_ptFicSortie );
		}

		fprintf( p_ptFicSortie , " " );
		ptTermeCou = ptTermeCou->TermeFrere;
	      }
	      fprintf( p_ptFicSortie , "\t" );
	      
	    }
	    /* ecriture de l'identifieur du terme */
	    else {
	      /* SortieAmaryllis ecriture de l'identifieur du terme 
		 deux fois */
	      if( FormatDeSortie == _SortieAmaryllis ) {
		fprintf( p_ptFicSortie ,"%s\t", ident );
	      }

	      fprintf( p_ptFicSortie ,"%s\t", ident );
	    }
	    
	    if( Trace ) {
	      fprintf( FicLog , "%s\t", ident );
	    }
          }
          return _True;
     }
     return _False;
}

/*************************************************************************
           PARCOURS EN POST-ORDRE DE L'ARBRE D'ANALYSE 
         pour recopie des avant unification destructrice
**************************************************************************/
char     RecopieDagUnBloc( TabBlocInitial p_ptBlocIni )
{
   TabBlocInitial                ptBlocFils;
   TabTermeMem                   ptTerme;
   TabLemmeMem                   ptLemme;

   p_ptBlocIni->dagCopie = NULL;

   /***********************************************************
   * recopies eventuelles de flexions                         *
   ************************************************************/
   switch( p_ptBlocIni->typeBloc )
   {
   case _MotFlechi :
       /* lemme flechi : on recopie les dags */
       ptLemme = p_ptBlocIni->T1.PtLemme;
       return CopieDagDF( &(p_ptBlocIni->dagCopie) , 
                          ptLemme->ValTerme.ArcsFils , _False );

   case _Concatenation :
   case _Liste :
   case _Feuille :            /* paradigme defini en intension */
   case _UniteLexicale :      /* lemme dont on connait la reference */
   case _Regle :
       /* terme de regle : on recopie les dags */
       ptTerme = p_ptBlocIni->T1.PtTerme;
       return CopieDagDF( &(p_ptBlocIni->dagCopie) , 
                                  ptTerme->ValTerme.ArcsFils , _False );

   case _RegleNormale :
   case _RegleElision :
   case _RegleMeta :
       /* on pointe vers les valeurs du fils unique qui est le terme suivant
          dans l'arbre a plat dans la pile (la racine de la regle) */
       ptBlocFils = (TabBlocInitial)(Pile + p_ptBlocIni->ptBlocSuivant);
       p_ptBlocIni->dagCopie = ptBlocFils->dagCopie;
       return _True;

   default :
       TxtErreur( 6 , "RecopieDagUnBloc(alanalut)" );
       return _False;
   }
}

/*************************************************************************
           PARCOURS EN POST-ORDRE DE L'ARBRE D'ANALYSE 
                pour demarquage eventuelles de dags
**************************************************************************/
char     EffectueFonctionSurDagUnBloc(TabBlocInitial p_ptBlocIni , 
                                      int p_estDagTerme ,
                                      char (*p_fonctionDag)(
           TabDagMem                          *p_ptFils
                                                    ) )
{
   TabTermeMem                   ptTerme;
   TabLemmeMem                   ptLemme;

   switch( p_ptBlocIni->typeBloc )
   {
   case _MotFlechi :
       /* lemme flechi : on demarque les dags */
       if( p_estDagTerme )
       {
           ptLemme = p_ptBlocIni->T1.PtLemme;
           return (*p_fonctionDag)( &(ptLemme->ValTerme.ArcsFils) );
       }
       else
       {
           return (*p_fonctionDag)( &(p_ptBlocIni->dagCopie) );
       }

   case _Concatenation :
   case _Liste :
   case _Feuille :            /* paradigme defini en intension */
   case _UniteLexicale :      /* lemme dont on connait la reference */
   case _Regle :
       /* terme de regle : on demarque les dags */
       if( p_estDagTerme )
       {
           ptTerme = p_ptBlocIni->T1.PtTerme;
           return (*p_fonctionDag)( &(ptTerme->ValTerme.ArcsFils) );
       }
       else
       {
           return (*p_fonctionDag)( &(p_ptBlocIni->dagCopie) );
       }

   case _RegleNormale :
   case _RegleElision :
   case _RegleMeta :
       return _True;

   default :
       TxtErreur( 6 , "EffectueFonctionSurDagUnBloc(alanalut)" );
       return _False;
   }
}

/*************************************************************************
           PARCOURS EN POST-ORDRE DE L'ARBRE D'ANALYSE 
                pour desallocations eventuelles de dags
**************************************************************************/
char     DesalloueDagUnBloc( TabBlocInitial p_ptBlocIni )
{
   switch( p_ptBlocIni->typeBloc )
   {
   case _MotFlechi :
   case _Concatenation :
   case _Liste :
   case _Feuille :            /* paradigme defini en intension */
   case _UniteLexicale :      /* lemme dont on connait la reference */
   case _Regle :
       /* terme de regle ou lemme flechi : on rend les dags copies */
       /* sans les demarquer au prealable, car c'est fait initialement 
          pour tous les blocs */
       return RendDag( &(p_ptBlocIni->dagCopie) );

   case _RegleNormale :
   case _RegleElision :
   case _RegleMeta :
       p_ptBlocIni->dagCopie = NULL;
       return _True;

   default :
       TxtErreur( 6 , "DesalloueDagUnBloc(alanalut)" );
       return _False;
   }
}

/*************************************************************************
     PARCOURS EN POST-ORDRE DE L'ARBRE D'ANALYSE POUR UNIFICATION
**************************************************************************/
char     NoteUnificationBloc( TabBlocInitial p_ptBlocIni ,
			      TabDagMem *p_hdlOperLogiques , 
			      int p_estGeneration )
{
   TabLemmesMots                  tLemmeMot;
   TabBlocInitial                 ptBlocFils;
   char                           resul;

   resul = _True;

   /***********************************************************
   * recopies eventuelles de flexions                         *
   ************************************************************/
   switch( p_ptBlocIni->typeBloc )
   {
   case _MotFlechi :
   case _Concatenation :
   case _Liste :
       break;

   case _Feuille :            /* paradigme defini en intension */
   case _UniteLexicale :      /* lemme dont on connait la reference */
       ptBlocFils = (TabBlocInitial)(Pile + p_ptBlocIni->ptBlocSuivant);
       if( ptBlocFils->typeBloc != _MotFlechi )
       /* terme optionnel pour lequel n'existe pas de fils */
       {
          TxtErreur( 4 , "NoteUnificationBloc(alanalut)" );
          resul = _False;
       }
       else
       {
          if( !p_estGeneration )
          {
            tLemmeMot = (TabLemmesMots)
	      ((TabChar)p_ptBlocIni + sizeof(struct BlocInitial));
          }

          /* on unifie les valeurs du terme avec les valeurs du mot flechi */
          if( VerificationAccord )
          {
              resul = AjoutOperFils( p_hdlOperLogiques , '=' , 
                            _Arc , (TabChar)&(p_ptBlocIni->dagCopie) , 
                            _Arc , (TabChar)&(ptBlocFils->dagCopie) ) != NULL;
          }
       }
       break;

   case _RegleNormale :
   case _RegleElision :
   case _RegleMeta :
       break;

   case _Regle :
       /* on recupere les valeurs du fils unique qui est le terme suivant
          dans l'arbre a plat */
       ptBlocFils = (TabBlocInitial)(Pile + p_ptBlocIni->ptBlocSuivant);

       if( ptBlocFils->typeBloc != _RegleNormale
       && ptBlocFils->typeBloc != _RegleMeta
       && ptBlocFils->typeBloc != _RegleElision )
       {
           sprintf( ChAux , "NoteUnificationBloc(alanalut) [%d]" , 
                            ptBlocFils->typeBloc );
           TxtErreur( 5 , ChAux );
       }
       else
       {
          if( VerificationAccord )
          {
              /* on unifie les valeurs de la tete de regle et 
                 celle du terme appelant */
              resul = AjoutOperFils( p_hdlOperLogiques , '=' , 
                            _Arc , (TabChar)&(p_ptBlocIni->dagCopie) , 
                            _Arc , (TabChar)&(ptBlocFils->dagCopie) ) != NULL;
          }
       }
       break;

   default :
       TxtErreur( 6 , "NoteUnificationBloc(alanalut)" );
       resul = _False;
       break;
   }
   return resul;
}

/*************************************************************************
     LANCEMENT DE L'UNIFICATION
**************************************************************************/
char     Unification( TabBlocInitial p_ptBlocIni , int p_estGeneration )
{
     char                            valRet;
     char                            ok;

     TabBlocInitial                  ptBlocCou;
     long                            ptMem;
     TabDagMem                       operLogiques;

     /* initialisation du tableau des noeuds copies pour les recopies
        de dags */
     NbLiensNoeudsCopies = 0;

     /* analyse sans verification des contraintes : pas d'unification
        on releve uniquement les mots utilises */
     if( !VerificationAccord )
     {
         /* note les mots du type des blocs */
         /* parcours remontant */
         operLogiques = NULL;
         NoteUnificationBloc( p_ptBlocIni , &operLogiques , p_estGeneration );
         ptMem = p_ptBlocIni->ptBlocPrecedent;
         while( ptMem != _BasPile )
         {
             ptBlocCou = (TabBlocInitial)(Pile + ptMem);
             NoteUnificationBloc( ptBlocCou , &operLogiques , 
				  p_estGeneration );
             ptMem = ptBlocCou->ptBlocPrecedent;
         }

         return _True;
     }
     else
     {
         if( Trace )
         {
/*              fprintf( FicLog , "\n*********** DAGS before copy\n" ); */
/*              EcritDagTousBloc( p_ptBlocIni , _True ); */
/*              fprintf( FicLog , "\n*********** End DAGS DAGS before copy\n" ); */
         }

         /* demarquage des dags des termes avant copie */
         /* parcours remontant */
         ok = EffectueFonctionSurDagUnBloc( p_ptBlocIni , _True ,
					    DemarqueDagDF );
         ptMem = p_ptBlocIni->ptBlocPrecedent;
         while( ptMem != _BasPile && ok )
         {
             ptBlocCou = (TabBlocInitial)(Pile + ptMem);
             ok = EffectueFonctionSurDagUnBloc( ptBlocCou , _True ,
						DemarqueDagDF );
             ptMem = ptBlocCou->ptBlocPrecedent;
         }
         if( !ok )
         {
            return _False;
         }
             
         /* recopie des dags pour eviter leur sauvegarde prealable
            parcours remontant */
         ok = RecopieDagUnBloc( p_ptBlocIni );
         ptMem = p_ptBlocIni->ptBlocPrecedent;
         while( ptMem != _BasPile && ok )
         {
             ptBlocCou = (TabBlocInitial)(Pile + ptMem);
             ok = RecopieDagUnBloc( ptBlocCou );
             ptMem = ptBlocCou->ptBlocPrecedent;
         }
         if( !ok )
         {
            return _False;
         }

         if( Trace )
         {
/*              fprintf( FicLog , "\n*********** DAGS after copy\n" ); */
/*              EcritDagTousBloc( p_ptBlocIni , _False ); */
/*              fprintf( FicLog , "\n*********** End DAGS after copy\n" ); */
         }

         /* note les unifications sur tous les blocs en fonction de
            leur type parcours remontant */
         operLogiques = NULL;
         ok = NoteUnificationBloc( p_ptBlocIni , &operLogiques ,
                                           p_estGeneration );
         ptMem = p_ptBlocIni->ptBlocPrecedent;
         while( ptMem != _BasPile && ok )
         {
             ptBlocCou = (TabBlocInitial)(Pile + ptMem);
             ok = NoteUnificationBloc( ptBlocCou , &operLogiques ,
				       p_estGeneration );
             ptMem = ptBlocCou->ptBlocPrecedent;
         }
         if( !ok )
         {
            return _False;
         }

         /* unification donnee par la liste des operations d'egalite */
         valRet = UnifierContraintes( &operLogiques , NULL ,
                                      NULL , p_ptBlocIni );
         if( Trace )
         {
            fprintf( FicLog ,"Unification resul\n" );
         }

         if( Trace )
         {
/*              fprintf( FicLog ,  */
/*                       "\n*********** DAGS after unification\n" ); */
/*              EcritDagTousBloc( p_ptBlocIni , _False ); */
/*              fprintf( FicLog ,  */
/*                       "\n*********** Fin DAGS after unification\n" ); */
         }

         /* demarquage des dags copies apres unification pour preparer
            desallocation */
         /* parcours remontant */
         ok = EffectueFonctionSurDagUnBloc( p_ptBlocIni , _False ,
					    DemarqueDagDF );
         ptMem = p_ptBlocIni->ptBlocPrecedent;
         while( ptMem != _BasPile && ok )
         {
             ptBlocCou = (TabBlocInitial)(Pile + ptMem);
             ok &= EffectueFonctionSurDagUnBloc( ptBlocCou , _False ,
				        	 DemarqueDagDF );
             ptMem = ptBlocCou->ptBlocPrecedent;
         }
         if( !ok )
         {
            return _False;
         }
             
         /* desallocation de ce qui reste des dags copies apres unification */
         /* parcours remontant */
         ok = DesalloueDagUnBloc( p_ptBlocIni );
         ptMem = p_ptBlocIni->ptBlocPrecedent;
         while( ptMem != _BasPile && ok )
         {
             ptBlocCou = (TabBlocInitial)(Pile + ptMem);
             ok = DesalloueDagUnBloc( ptBlocCou );
             ptMem = ptBlocCou->ptBlocPrecedent;
         }

         return valRet;
     }
}

/*************************************************************************
     UNIFICATION PAR APPEL DE FONCTION EXTERNE
**************************************************************************/
int      PrintExternalUnificationDF( TabBlocInitial p_ptBlocIni ,
				     TabCorrIdent_PosFic p_tCorrespondance )
{
     long                       ptMem;
     TabRegleMem               *hdlRegleDest;
     char                      *ptChIdent;
     char                      *ptChType;
     ChIdent                    ident;

     ptMem = p_ptBlocIni->ptBlocPrecedent;
     if( ptMem ) {
       PrintExternalUnificationDF( (TabBlocInitial)(Pile + ptMem) ,
				   p_tCorrespondance );
     }

     /* lexical leave, all the features are written */
     if( p_ptBlocIni->typeBloc == _Feuille
	 || p_ptBlocIni->typeBloc == _UniteLexicale ) {
       if( NbPrintedWords ) {
	 fprintf( FicUnification , "\n\n" );
       }
       fprintf( FicUnification , "Word [#%d]", ++NbPrintedWords);
       EffectueFonctionSurDagUnBloc( p_ptBlocIni , _True ,
				     PrintDagUnification );
     }
     /* inflected word, onley features that are not provided
        by the lexical leave are provided */
     else if( p_ptBlocIni->typeBloc == _MotFlechi ) {
       EffectueFonctionSurDagUnBloc( p_ptBlocIni , _True ,
				     PrintDagUnificationWithoutRepetition );
     }
     else if( p_ptBlocIni->typeBloc == _Regle ) {
       /* variante */
       if( p_ptBlocIni->noChoix > 1 ) {
         hdlRegleDest = (TabRegleMem *)((TabChar)p_ptBlocIni 
					+ sizeof(struct BlocInitial));
	 if( ChaineMetaRegle( (*hdlRegleDest)->TermeFils->ValTerme.ArcsFils , 
			      p_ptBlocIni->noChoix - 2 , &ptChIdent , 
			      &ptChType ) ) {
	   fprintf( FicUnification , "Variant [%s,%d,%s]\n\n" , ptChIdent , 
		    p_ptBlocIni->noChoix - 1 , ptChType );
	   
	 }
	 else {
	   fprintf( FicUnification , "Unknown metarule\n\n" );
	 }
       }
       /* occurrence simple */
       else {
	 fprintf( FicUnification , "Occurrence [0]\n\n" );
       }
       if( Identificateur( ident , 
			   (*hdlRegleDest)->TermeFils->ValTerme.Ref , 
			   FicRefIdent , p_tCorrespondance ) ) {
	 fprintf( FicUnification ,"Term [%s]\n\n", ident );
       }
       else {
	 fprintf( FicUnification , "Unknown term\n\n" );
       }
      }
     return _True;
}

int      PrintExternalUnification( TabBlocInitial p_ptBlocIni ,
				   TabCorrIdent_PosFic p_tCorrespondance )
{
/*    TabBlocInitial                  ptBlocCou; */
/*    long                            ptMem; */
/*    TabDagMem                       operLogiques; */

  if( !OuvreFichier( NomFicUnification , &FicUnification , "w" ) ) {
    return _False;
  }
  NbPrintedWords = 0;
  
  PrintExternalUnificationDF( p_ptBlocIni , p_tCorrespondance );

  fprintf( FicUnification , "\n" );
  fclose( FicUnification );

  /* analyse sans verification des contraintes : pas d'unification
        on releve uniquement les mots utilises */
  /* note les mots du type des blocs */
  /* parcours remontant */
/*    operLogiques = NULL; */
/*    NoteUnificationBloc( p_ptBlocIni , &operLogiques , _False ); */
/*    ptMem = p_ptBlocIni->ptBlocPrecedent; */
/*    while( ptMem != _BasPile ) { */
/*      ptBlocCou = (TabBlocInitial)(Pile + ptMem); */
/*      NoteUnificationBloc( ptBlocCou , &operLogiques , _False ); */
/*      ptMem = ptBlocCou->ptBlocPrecedent; */
/*    } */

  return system( NomFicExtUnifier );
}

/*************************************************************************
           ECRITURE DU DAG COPIE SUR TOUS LES BLOCS POUR TRACE
**************************************************************************/
char     EcritDagTousBloc( TabBlocInitial p_ptBlocIni , int p_estDagTerme )
{
     char                            ok;
     TabBlocInitial                  ptBlocCou;
     long                            ptMem;

     fprintf( FicLog ,
              "\n--->Courant : ty(%d) Ref(%ld) Chx(%d/%d)",
              p_ptBlocIni->typeBloc,
              (p_ptBlocIni->T1.PtTerme)->ValTerme.Ref,
              p_ptBlocIni->noChoix,
              p_ptBlocIni->nbChoix);

     ok = EffectueFonctionSurDagUnBloc( p_ptBlocIni , p_estDagTerme ,
                                        EcrireDagBasique );
     ptMem = p_ptBlocIni->ptBlocPrecedent;
     while( ptMem != _BasPile && ok )
     {
         ptBlocCou = (TabBlocInitial)(Pile + ptMem);
         /* terme de regle ou lemme flechi : on recopie les dags */
         fprintf( FicLog ,
              "\n--->Courant : ty(%d) Ref(%ld) Chx(%d/%d)",
              ptBlocCou->typeBloc,
              (ptBlocCou->T1.PtTerme)->ValTerme.Ref,
              ptBlocCou->noChoix,
              ptBlocCou->nbChoix);

         ok = EffectueFonctionSurDagUnBloc( ptBlocCou , p_estDagTerme ,
					    EcrireDagBasique );
         ptMem = ptBlocCou->ptBlocPrecedent;
     }
     if( !ok )
     {
        return _False;
     }
     return _True;
}

/*************************************************************************
    FONCTIONS POUR NOTER LA REFERENCE D'UN TERME DE METAREGLE EN PARTIE 
   GAUCHE APRES UNIFICATION ET POUR LA TRANSMETTRE AU TERME DE LA PARTIE 
                        DROITE AYANT LE MEME IDENTITICATEUR
**************************************************************************/
char     NoteLienIdentTerme( TabTermeMem *p_hdlTerme )
{
    char            trouve;
    int             indTerme;

    if( !(*p_hdlTerme) )
    {
         return _True;
    }
    
    /* lecture de l'identificateur et lien avec le terme */
    trouve = _False;
    for( indTerme = 0 ; indTerme < NbLiensIdentTerme ; indTerme++ )
    {
         if( strcmp( LienIdentTerme[ indTerme ].Ident , 
                     (*p_hdlTerme)->ValTerme.Ident ) == 0 )
         {
                 trouve = _True;
                 break;
         }
    }

    if( !trouve )
    {
        LienIdentTerme[ NbLiensIdentTerme ].PtValTerme 
                  = &((*p_hdlTerme)->ValTerme);
        LienIdentTerme[ NbLiensIdentTerme ].PtTerme 
                  = (*p_hdlTerme);
        strcpy( LienIdentTerme[ NbLiensIdentTerme ].Ident , 
                (*p_hdlTerme)->ValTerme.Ident );
        NbLiensIdentTerme++;
        if( NbLiensIdentTerme >= NbMaxTermesDansUneRegle )
        {
             TxtErreur( 104 , "NoteLienIdentTerme(alanalut)" );
             return _False;
        }
    }
    return _True;
}


/*************************************************************************
     TRAIT sem, self ou root a valeurs numeriques (references)
**************************************************************************/
int      TraitAReferences( TabDagMem *p_hdlDagTerme , 
			   TabDagMem p_traitRecherche ,
			   TabLong p_tabRef , 
			   TabRefPhraseMem  *p_tabRefPhraseMem ,
			   int p_valeursMultiplesPossibles )
{
    TabDagMem         *hdlDag;
    TabChar            chaine;
    TabLong            ptRef;
    TabLong            ptCategorie;
    TabLong            ptNoFlexion; 
    TabRefPhraseMem    ptTabRefPhraseCou;

    /* ce trait existe sur ce terme */
    if( ( hdlDag = ExisteCheminFils( p_traitRecherche ,
				     p_hdlDagTerme ) ) ) {
      /* disjonctions autorisees (Synonymie) */
      if( p_valeursMultiplesPossibles ) {
	/* disjonction : on parcourt l'ensemble des fils de
	   l'entree courante */
	if( (*hdlDag)  && (*hdlDag)->TypeDag != _TypeDagVide
	    && ((*hdlDag)->ValDag.Atome.AtomeConjDisj) == _DisjonctionPos ) { 
	  while( hdlDag && (*hdlDag) ) {
	    if( !LienUniteLexicaleATraits( hdlDag , 
					   (TabDagMem *)NULL ,
					   &ptRef , &chaine ,
					   &ptCategorie , &ptNoFlexion , 
					   _False ) ) {
	      EcrireDagBasique( hdlDag );
	      TxtErreur( 311 , "TraitAReferences(alanalut)" );
	      return _False;
	    }

	    /* introduction de la nouvelle valeur en tete 
	       des references */
	    ptTabRefPhraseCou = *p_tabRefPhraseMem;
	    if( !(*p_tabRefPhraseMem = NouveauRefPhrase()) ) {
	      return _False;
	    }
	    (*p_tabRefPhraseMem)->Ref = *ptRef;
	    (*p_tabRefPhraseMem)->Suivant = ptTabRefPhraseCou;

	    hdlDag = &((*hdlDag)->Frere);
	  }
	}
	/* disjonction : on parcourt l'ensemble des fils de
	   l'entree courante */
	
	/* non disjonction : on ne regarde que la premiere valeur */
	/* les negations ne sont pas prevues a ce point */
	else if( (*hdlDag) && (*hdlDag)->TypeDag != _TypeDagVide
		 && ((*hdlDag)->ValDag.Atome.AtomeConjDisj) == _False ) { 
	  if( !LienUniteLexicaleATraits( hdlDag , 
					 (TabDagMem *)NULL ,
					 &ptRef , &chaine ,
					 &ptCategorie , &ptNoFlexion , 
					 _False ) ) {
	    EcrireDagBasique( hdlDag );
	    TxtErreur( 311 , "TraitAReferences(alanalut)" );
	    return _False;
	  }

	  /* introduction de la nouvelle valeur en tete 
	     des references */
	  ptTabRefPhraseCou = *p_tabRefPhraseMem;
	  if( !(*p_tabRefPhraseMem = NouveauRefPhrase()) ) {
	    return _False;
	  }
	  (*p_tabRefPhraseMem)->Ref = *ptRef;
	  (*p_tabRefPhraseMem)->Suivant = ptTabRefPhraseCou;
	}
	
	else if( (*hdlDag) && (*hdlDag)->TypeDag != _TypeDagVide ) {
	  TxtErreur( 309 , "TraitAReferences(alanalut)" );
	  return _False;
	}
      }
      /* disjonctions autorisees (Synonymie) */

      /* seule une valeur unique est autorisee */
      else {
	if( (*hdlDag) && (*hdlDag)->TypeDag != _TypeDagVide
	    && ((*hdlDag)->ValDag.Atome.AtomeConjDisj) == _False ) { 
	  if( !LienUniteLexicaleATraits( hdlDag , 
					 (TabDagMem *)NULL ,
					 &ptRef , &chaine ,
					 &ptCategorie , &ptNoFlexion , 
					 _False ) ) {
	    EcrireDagBasique( hdlDag );
	    TxtErreur( 311 , "TraitAReferences(alanalut)" );
	    return _False;
	  }
	  *p_tabRef = *ptRef;
	}
	
	else if( (*hdlDag) && (*hdlDag)->TypeDag != _TypeDagVide ) {
	  TxtErreur( 309 , "TraitAReferences(alanalut)" );
	  return _False;
	}
      }
      /* seule une valeur unique est autorisee */
    }
    /* ce trait existe sur ce terme */

    return _True;
}

/*************************************************************************
     transformation des traits sem, self ou root 
     en valeurs numeriques (references)
**************************************************************************/
char     GarnitReferenceRegleTransformee( TabTermeMem *p_hdlTerme )
{
    int                               indTerme;
    int                               indTraitSem;
    long                              ref;

    if( !(*p_hdlTerme) )
    {
         return _True;
    }

    /* recherche du terme de la partie droite de la metaregle ayant
       cet identificateur */
    for( indTerme = 0 ; indTerme < NbLiensIdentTerme ; indTerme++ )
    {
        if( strcmp( LienIdentTerme[ indTerme ].Ident ,
                    (*p_hdlTerme)->ValTerme.Ident ) == 0 )
        {
            /* on releve le pteur vers le dag de ce terme */
            (*p_hdlTerme)->ValTerme.Ref 
	      = (LienIdentTerme[ indTerme ].PtValTerme)->Ref;
            (*p_hdlTerme)->ValTerme.RefRacine 
	      = (LienIdentTerme[ indTerme ].PtValTerme)->RefRacine;
	    CopieReferences(
		  &((*p_hdlTerme)->ValTerme.RefSelf), 
		  (LienIdentTerme[ indTerme ].PtValTerme)->RefSelf );
	    CopieReferences(
		  &((*p_hdlTerme)->ValTerme.RefSemantique), 
		  (LienIdentTerme[ indTerme ].PtValTerme)->RefSemantique );
            (*p_hdlTerme)->ValTerme.PosFicRegle
                       = (LienIdentTerme[ indTerme ].PtValTerme)->PosFicRegle;
            return _True;
        }
    }

   /* chemin racine */
   if( !TraitAReferences( &((*p_hdlTerme)->ValTerme.ArcsFils) ,
			  CheminRacine ,
			  &((*p_hdlTerme)->ValTerme.RefRacine) ,
			  NULL ,
			  _False ) ) {
     return _False;
   }

   /* parcours de tous les traits semantiques declares
      dans le header */
   indTraitSem = 0;
   while( CheminsSemantique[ indTraitSem ] && indTraitSem < _NbMaxTypesLiensSem ) {
     if( !TraitAReferences( &((*p_hdlTerme)->ValTerme.ArcsFils) ,
			    CheminsSemantique[ indTraitSem ]  ,
			    NULL ,
			    &((*p_hdlTerme)->ValTerme.RefSemantique) ,
			    _True ) ) {
       return _False;
     }
     indTraitSem++;
   }

   /* trait self */
   if( !TraitAReferences( &((*p_hdlTerme)->ValTerme.ArcsFils) ,
			  CheminSelf ,
			  NULL ,
			  &((*p_hdlTerme)->ValTerme.RefSelf) ,
			  _True ) ) {
     return _False;
   }

   /* reference lexicale standard */
   if( ( ref = ReferenceLexicale( &((*p_hdlTerme)->ValTerme.ArcsFils) ) ) 
       != _Nil ) {
     (*p_hdlTerme)->ValTerme.Ref = ref;
   }

   return _True;
}

/*************************************************************************
     GENERATION  D'UNE METAREGLE
*************************************************************************/
char     GenererMetaRegle( TabRegleMem p_ptRegle , int p_noMetaRegle )
{
     int                       indMetaRegle;
     TabRegleMemMeta           ptRegleMeta;
     TabTermeMem               sauveMetaRegle;
     TabTermeMem               sauveRegle;
     char                      trouve;
     char                     *ptCh;
     int                       noTableau;
     struct TermeMem           termeU;
     TabTermeMem               ptTermeU;

     if( !DagANoMetaregle( p_ptRegle->TermeFils->ValTerme.ArcsFils , 
                           &ptCh , &noTableau , &trouve ) || !trouve )
     {
          return _False;
     }

     ptRegleMeta = Tableaux[ noTableau ].PtRegleMeta;
     indMetaRegle = 0;
     while( ptRegleMeta && indMetaRegle < p_noMetaRegle )
     {
          ptRegleMeta = ptRegleMeta->RegleFrere;
          indMetaRegle++;
     }

     if( indMetaRegle != p_noMetaRegle || !ptRegleMeta )
     {
          return _False;
     }

     /* generation de la transformee par parcours parallele de la
        regle d'origine et de la regle de gauche de la metaregle */

     /* 1 duplication de la metaregle, car les valeur de ses dags
        seront modifiees */
     sauveMetaRegle = NULL;
     if( !CopieEtUnifieArbre( &sauveMetaRegle , ptRegleMeta->TermeFils , 
                              _False , NULL ) )
     {
         return _False;
     }

     /* 2 duplication de la regle, car les valeur de ses dags seront
        modifiees */
     sauveRegle = NULL;
     if( !CopieEtUnifieArbre( &sauveRegle , p_ptRegle->TermeFils , 
                              _False , NULL ) )
     {
         return _False;
     }

     /* 3 unification entre l'arbre gauche de la metaregle et la regle
        initiale avec recopie dans l'arbre de la metaregle des infos
        de la regle qui n'y sont pas encore */
     /* on cree un terme dont les deux fils sont les deux arbres
        unifies pour la restauration des pointeurs lors de
        l'unification */
     ptTermeU = &termeU;
     MiseAZeroUnTerme( ptTermeU );
     ptTermeU->TermeFils = sauveMetaRegle;
     ptTermeU->TermeFrere = sauveRegle;
     if( CopieEtUnifieArbre( &(sauveMetaRegle->TermeFils) , sauveRegle , 
                             _True , &ptTermeU ) )
     {
         /* 4 on recopie les references des termes de gauche dans ceux
            de droite */
         NbLiensIdentTerme = 0;
         if( !ParcoursDFArbre( &(sauveMetaRegle->TermeFils) , 
                               NoteLienIdentTerme , _PostOrdre ) )
         {
             return _False;
         }
         if( !ParcoursDFArbre( &(sauveMetaRegle->TermeFrere) ,
                               GarnitReferenceRegleTransformee , _PostOrdre ) )
         {
             return _False;
         }

         /* 5 on recopie les sous-arbres des termes de gauche dans
            ceux de droite en s'assurant de leur compatibilite */
         if( !ParcoursDFArbre( &(sauveMetaRegle->TermeFrere) , 
                               CopieSousArbreTermesEgaux , _PostOrdre ) )
         {
             return _False;
         }

        /* 6 recopie dans la transformee de la regle de l'arbre de
            droite de la metaregle */
         if( !DemarqueArbreDF( &sauveMetaRegle ) 
             || !DemarqueArbreDF( &sauveRegle )
             || !MarqueArbreDF( &(sauveMetaRegle->TermeFrere) ) )
         {
             return _False;
         }
         /* on n'utilise pas DemarqueEtRendArbre qui demarquerait tout
            le terme source de la metaregle, sans conserver ce qui
            appartient a la destination */
         if( !ParcoursDFArbre( &(sauveMetaRegle->TermeFils) , RendTerme , 
                               _PostOrdre ) 
             || !ParcoursDFArbre( &(sauveRegle) , RendTerme , _PostOrdre ) )
         {
             return _False;
         }
         p_ptRegle->Transformee = sauveMetaRegle->TermeFrere;
	 p_ptRegle->TransformeeAvecUnificationExterne 
	   = ptRegleMeta->AvecUnificationExterne;
         sauveMetaRegle->TermeFrere = NULL;
         sauveMetaRegle->TermeFils = NULL;
         if( !RendTerme( &sauveMetaRegle ) )
         {
             return _False;
         }
     }
     else
     {
         if( Trace )
         {
            fprintf( FicLog , "\nMetarule unification failure\n" );
         }
         /* 4 desallocation des dexu copies */
         if( !DemarqueArbreDF( &sauveMetaRegle ) 
             || !DemarqueArbreDF( &sauveRegle ) )
         {
             return _False;
         }
         /* on separe demarquage et retour pour ne pas risquer de
            rendre deux fois un pointeur commun */
         if( !ParcoursDFArbre( &sauveMetaRegle , RendTerme , _PostOrdre ) 
             || !ParcoursDFArbre( &sauveRegle , RendTerme , _PostOrdre ) )
         {
             return _False;
         }
         p_ptRegle->Transformee = NULL;
     }
     return _True;
}

/*************************************************************************
     AFFICHAGE DU NUMERO  D'UNE METAREGLE
*************************************************************************/
char     ChaineMetaRegle( TabDagMem p_ptDag , int p_noMetaRegle ,
                          TabChar *p_hdlIdent , TabChar *p_hdlType )
{
     char                 trouve;
     int                  noTableau;
     int                  indMetaRegle;
     TabRegleMemMeta      ptRegleMeta;

     if( !DagANoMetaregle( p_ptDag , p_hdlIdent , 
			   &noTableau , &trouve ) || !trouve )
     {
          return _False;
     }
     else
     {
         ptRegleMeta = Tableaux[ noTableau ].PtRegleMeta;
         indMetaRegle = 0;
         while( ptRegleMeta && indMetaRegle < p_noMetaRegle )
         {
              ptRegleMeta = ptRegleMeta->RegleFrere;
              indMetaRegle++;
         }

         if( indMetaRegle != p_noMetaRegle || !ptRegleMeta )
         {
              return _False;
         }
         *p_hdlType = ptRegleMeta->TermeFils->ValTerme.Ident;
         return _True;
     }
}

/*************************************************************************
     NOMBRE EFFECTIF DE METAREGLES, EN COMPTANT LES METAREGLES MULTIPLES
           POUR UNE MEME IDENTIFICATEUR (LES COORDINATIONS PAR EXEMPLE)
*************************************************************************/
char     DagANbMetaregle( TabDagMem p_ptDag )
{
     char                      valRet = 0;
     TabRegleMemMeta           ptRegleMeta;
     char                      trouve;
     char                     *ptCh;
     int                       noTableau;

     if( !DagANoMetaregle( p_ptDag , &ptCh , &noTableau , &trouve ) 
	 || !trouve )
     {
       return 0;
     }

     ptRegleMeta = Tableaux[ noTableau ].PtRegleMeta;
     while( ptRegleMeta )
     {
          ptRegleMeta = ptRegleMeta->RegleFrere;
          valRet++;
     }
     if( Trace ) {
       fprintf( FicLog , "# metarules for this label: %d\n" , valRet );
     }
     return valRet;
}

/*************************************************************************
     NO DE TABLEAU ET REFERENCE D'UNE METAREGLE D'APRES UN DAG
     S'IL N'Y A PAS DE METAREGLE DU TABLEAU AYANT L'IDENTIFICATEUR
     DONNE PAR LE DAG, ON RETOURNE L'INDICE DU TABLEAU IMMEDIATEMENT APRES
*************************************************************************/
char     DagANoMetaregle( TabDagMem p_ptDag , char **p_hdlCh , 
			  int *p_noTableau , 
                          char *p_trouve )
{
    TabDagMem            *hdlFilsEtiquette;

    *p_trouve = _False;
    *p_noTableau = 0;
    *p_hdlCh = NULL;

    hdlFilsEtiquette = ExisteCheminFils( CheminEtiquetteMeta , &p_ptDag );
    if( !hdlFilsEtiquette || !(*hdlFilsEtiquette) )
    {
        if( Trace ) {
             fprintf( FicLog , "No metarule label\n" );
        }
        return _True;
    }

    if( (*hdlFilsEtiquette)->TypeDag != _AtomeChaine )
    {
        TxtErreur( 101 , "DagANoMetaregle(alanalut)" );
        return _False;
    }

    (*p_hdlCh) = (*hdlFilsEtiquette)->ValDag.Atome.ValAtome.Chaine;
    if( !(*(*p_hdlCh)) )
    {
        TxtErreur( 102 , "DagANoMetaregle(alanalut)" );
        return _False;
    }

    if( Trace ) {
      fprintf( FicLog , "Metarule label: [%s]\n" , (*p_hdlCh) );
    }

    /* no du tableau en fonction de la chaine de la regle */
    *p_trouve = _False;
    for( (*p_noTableau) = 0 ; (*p_noTableau) < _NbMaxTableau ; 
	 (*p_noTableau)++ )
    {
          if( !(*(Tableaux[ (*p_noTableau) ].Ident)) )
          {
               break;
          }
          if( strcmp( Tableaux[ (*p_noTableau) ].Ident , (*p_hdlCh) ) == 0 )
          {
               *p_trouve = _True;
               break;
          }
    }
    if( (*p_noTableau) == _NbMaxTableau )
    {
        TxtErreur( 103 , "DagANoMetaregle(alanalut)" );
        return _False;
    }
    return _True;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                        fin de alanalut.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

