/*
 * FASt Term Recognizer
 *
 *  FASTR/alanalys.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for the parser
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*************************************************************************
AJOUT DES POINTEURS EN FIN DE BLOC LORSQUE LE TERME T1 EST DEJA RECOPIE
**************************************************************************/
void     AjouterTermesSuivants( TabBlocInitial p_ptBlocIni ,
               char *p_backTrackSucces , char *p_backTrackEchec ,
               TabRegleMem p_ptRegle , int p_estGeneration , 
               int p_typeIndexation )
{
  int                      ind;
  TabBlocInitial           ptBlocPere;
  TabTermeMem              ptTermeAux;
  TabTermeMem             *hdlTermeDest;
  TabRegleMem             *hdlRegleDest;
  unsigned long            ptMem;

  /* Termes generes ou reconnus dans la phrase      */
  struct LemmesMots        tLemmesMots[ _NbMaxHomographe ];
  TabLemmesMots            ptLemmeMotDest;

  char                     regleTrouvee;
  int                      nbLemmesTrouves;
  char                     nbMetaRegles;
  long                     refRegle;

  *p_backTrackSucces = _False;
  *p_backTrackEchec = _False;

  switch( p_ptBlocIni->typeBloc ) {
  case _Regle :
    regleTrouvee = _True;

    /* ce n'est pas la regle initiale */
    if( !p_ptRegle ) {
      refRegle = (p_ptBlocIni->T1.PtTerme)->ValTerme.PosFicRegle;
      if( Trace ) {
	fprintf( FicLog , "--> Rule %ld\n" , refRegle );
      }

      /* la regle est une regle fichier connue */
      if( refRegle >= 0L ) {
	if( !ReleverUneRegleBDic( (unsigned long)refRegle , &p_ptRegle ) ) {
	  if( Trace ) {
	    fprintf( FicLog ,"Unknown rule %ld\n", refRegle );
	  }
	  regleTrouvee = _False;
	}
      }
      /* la regle est une regle memoire connue */
      else if( refRegle != _RegleNonLiee ) {
	if( - refRegle >= (long)NbReglesMem ) {
	  TxtErreur( 7 , "AjouterTermesSuivants(alanalys)" );
	  exit(0);
	}
	p_ptRegle = ReglesMem - refRegle;

	if( Trace ) {
	  fprintf( FicLog , "Memory rule %ld\n", - refRegle );
	}
      }
      /* la regle est inconnue */
      else {
	sprintf( ChAux , "AjouterTermesSuivants(alanalys) [%ld]" , 
		 (p_ptBlocIni->T1.PtTerme)->ValTerme.Ref );
	TxtErreur( 135 , ChAux );
	regleTrouvee = _False;
      }
    }

    if( !regleTrouvee ) {
      *p_backTrackEchec = _True;
      p_ptBlocIni->ptBlocSuivant = HautPile + sizeof(struct BlocInitial);
      p_ptBlocIni->noChoix = 0;
      p_ptBlocIni->nbChoix = 0;
    }
    else {
      ptMem = HautPile + sizeof(struct BlocInitial);
      hdlRegleDest = (TabRegleMem *)(Pile + ptMem);
      *hdlRegleDest = p_ptRegle;
      p_ptBlocIni->ptBlocSuivant = ptMem + sizeof(TabRegleMem);
      p_ptBlocIni->noChoix = 0;
      p_ptBlocIni->nbChoix = 1;

      /* metaregles , on compte les positives */
      if( Trace ) {
	fprintf( FicLog ,"Total number of metarules: %d\n",
		 DagANbMetaregle( p_ptRegle->TermeFils->ValTerme.ArcsFils ) );
      }
      if( ( nbMetaRegles 
	    = DagANbMetaregle( p_ptRegle->TermeFils->ValTerme.ArcsFils ) )
	  > 0 ) {
	p_ptBlocIni->nbChoix += nbMetaRegles;
      }

      /* si variations seules, on ne tente pas l'effacement de la regle
	 initiale telle quelle */
      if( p_typeIndexation == _Variations ) {
	p_ptBlocIni->noChoix = 1;
      }
    }
    break;
  case _RegleNormale :
  case _RegleMeta :
  case _RegleElision :
    p_ptBlocIni->ptBlocSuivant = HautPile + sizeof(struct BlocInitial);
    p_ptBlocIni->noChoix = 0;

    /* si la transformee n'a pu etre produite, on met le nombre de
       choix a 0 */
    ptBlocPere = (TabBlocInitial)(Pile + p_ptBlocIni->ptBlocPere );
    hdlRegleDest = (TabRegleMem *)(Pile
				   + (p_ptBlocIni->ptBlocPere + sizeof(struct BlocInitial)) );
    if( ptBlocPere->noChoix > 1
	&& (*hdlRegleDest)->Transformee == NULL ) {
      p_ptBlocIni->nbChoix = 0;
      *p_backTrackEchec = _True;
    }
    else
      p_ptBlocIni->nbChoix = 1;

    break;

  case _Concatenation :
  case _Liste :
    ind = 0;
    ptMem = HautPile + sizeof(struct BlocInitial);
    ptTermeAux = (p_ptBlocIni->T1.PtTerme)->TermeFils;   /* fils de T1 */
    while( ptTermeAux ) {
      hdlTermeDest = (TabTermeMem *)(Pile + ptMem);
      *hdlTermeDest = ptTermeAux;
      ind++;
      ptMem += sizeof(TabTermeMem);
      ptTermeAux = ptTermeAux->TermeFrere;
    }
    p_ptBlocIni->ptBlocSuivant = ptMem;
    p_ptBlocIni->noChoix = 0;
    p_ptBlocIni->nbChoix = ind;
    break;

  case _Feuille :            /* paradigme defini en intension      */
  case _UniteLexicale :      /* lemme dont on connait la reference */
    /* si IndiceMotCou == _Nil, on recherche tous les mots de la phrase
       ayant un lemme flechi dont la reference est celle de l'unite lexicale
       sinon on recherche si le mot suivant de la phrase a un homographe de
       meme reference que l'unite lexicale                         */
    /* si la recherche reussit, on recopie dans la pile des paires formees de
       1) le pointeur vers le lemme flechi
       2) l'indice du mot auquel il correspond                     */

    ptTermeAux = p_ptBlocIni->T1.PtTerme;
    /* cas de l'analyse */
    if( !p_estGeneration ) {
      nbLemmesTrouves =  
	MotFlechiPresentDansLaPhrase( ptTermeAux , tLemmesMots );
    }
    else {
      nbLemmesTrouves =  MotFlechiProduit( ptTermeAux , tLemmesMots );
    }

    /* mot trouve dans la phrase */
    if( nbLemmesTrouves ) {
      /* tLemmesMots est tableau de pointeurs vers des paires
	 lemme, mot de la phrase, dont la reference est celle du
	 terme attendu (pour les caract lex-synt elles seront
	 verifiees lors de l'unification) */

      ptMem = HautPile + sizeof(struct BlocInitial);

      /* on ajoute les pointeurs vers les homographes et les mots
	 apres le bloc */
      ptLemmeMotDest = (TabLemmesMots)(Pile + ptMem);
      ptMem += sizeof(struct LemmesMots) * nbLemmesTrouves;
      if( ptMem >= TaillePileEffacement ) {
	sprintf( ChAux , "AjouterTermesSuivants(alanalys) [%d]",
		 nbLemmesTrouves );
	TxtErreur( 133 , ChAux );
	exit(0);
      }
      for( ind = 0 ; ind < nbLemmesTrouves ; ind++ ) {
	ptLemmeMotDest[ ind ] = tLemmesMots[ ind ];
      }

      p_ptBlocIni->ptBlocSuivant = ptMem;
      p_ptBlocIni->noChoix = 0;
      p_ptBlocIni->nbChoix = nbLemmesTrouves;
      if( Trace ) {
	fprintf( FicLog ,"Word %ld/%ld #homographs %d\n" ,
		 ptTermeAux->ValTerme.Ref , 
		 ptTermeAux->ValTerme.RefRacine , nbLemmesTrouves );
      }
    }
    /* mot trouve dans la phrase */
    else
      /* mot non trouve dans la phrase */
      {
	*p_backTrackEchec = _True;
	p_ptBlocIni->ptBlocSuivant = ptMem;
	p_ptBlocIni->noChoix = 0;
	p_ptBlocIni->nbChoix = 0;
	if( Trace ) {
	  fprintf( FicLog , "Word %ld/%ld unmatched\n" ,
		   ptTermeAux->ValTerme.Ref , 
		   ptTermeAux->ValTerme.RefRacine );
	}
      }
    /* mot non trouve dans la phrase */
    break;
  case _MotFlechi :
    /* pas d'ajouts en fin de bloc */
    p_ptBlocIni->ptBlocSuivant = HautPile + sizeof(struct BlocInitial);
    p_ptBlocIni->noChoix = 1;
    p_ptBlocIni->nbChoix = 1;
    *p_backTrackSucces = _True;
    break;
  default :
    sprintf( ChAux , "AjouterTermesSuivants(alanalys) [%ld,%d]" ,
	     (p_ptBlocIni->T1.PtTerme)->ValTerme.Ref , 
	     p_ptBlocIni->typeBloc );
    TxtErreur( 9 , ChAux );
    exit(0);
  } /* switch */
}

/*************************************************************************
        backTrack succes :
        - s'il n'y a plus de concatenation incomplete , on note la
          reussite et on lance le backTrackEchec sur le 1er choix restant
          eventuel
        - sinon on cherche a completer l'effacement de la 1ere concatenation
          incomplete trouvee.
**************************************************************************/
char     RechercheConcIncomplete( TabBlocInitial p_ptBlocIniCou ,
               int p_estGeneration , int p_typeIndexation ,
               char *p_uneConcTrouvee , unsigned long *p_hautPilePere ,
               unsigned long *p_hautPilePrec , 
               TabCorrIdent_PosFic p_tCorrespondance , char *p_cutProlog ,
               long *p_nbUnifs , long *p_nbUnifsSucces ,
               FILE *p_ptFicSortie ,
	       int p_transformeeAvecUnificationExterne )
{
  unsigned long          hautPileAux;
  TabBlocInitial         ptBlocIniAux;
  char                   unifOk;

  *p_uneConcTrouvee = _True;
  *p_cutProlog = _False;
  unifOk = _False;
  hautPileAux = p_ptBlocIniCou->ptBlocPrecedent;

  /* recherche d'une concatenation incomplete */
  while( _True ) {
    /* fin de pile */
    if( hautPileAux == _BasPile ) {
      /* plus de concatenation ; echec, on revient en arriere */
      (*p_nbUnifs)++;
      if( !Unification( p_ptBlocIniCou , p_estGeneration ) ) {
	if( Trace ) {
	  fprintf( FicLog , "Unsuccessfull unification\n" );
	}
      }
      /* plus de concatenation ; echec unif externe, on revient en arriere */
      else if ( p_transformeeAvecUnificationExterne 
		&& !PrintExternalUnification( p_ptBlocIniCou ,
					      p_tCorrespondance ) ) {
	if( Trace ) {
	  fprintf( FicLog , "Unsuccessfull external unification\n" );
	}
      }
      /* plus de concatenation ; succes ; on l'affiche */
      else {
	(*p_nbUnifsSucces)++;
	unifOk = _True;
	if( Trace ) {
	  fprintf( FicLog ,"Successfull unification\n");
	  fprintf( FicLog , "%06ld\t" , NbLignesTot );
	}

	if( FormatDeSortie == _Txt ) {
	  EcrireChaineAnalysee( p_estGeneration , p_ptFicSortie );
	}
	else {
	  /* en extraction terminologique, on ecrit 
	     1 le No de ligne, 
	     2 l'identificateur de regle, 3 le terme transforme
	     4 la transformation                            */
	  if( FormatDeSortie != _SortieAmaryllis ) { 
	    if( NbLignesTot >= 0 ) {
	      fprintf( p_ptFicSortie , "%06ld\t" , NbLignesTot );
	    }
	    else {
	      fprintf( p_ptFicSortie , "%06ld\t" , -NbLignesTot );
	    }
	  }
	}

	/* en indexation ou en analyse on ecrit la regle qui a    */
	/*  abouti au succes                                      */
	if( !p_estGeneration ) {
	  if( FormatDeSortie == _Txt ) {
	    EcrireReglesUtilisees( p_ptBlocIniCou ,
				   p_tCorrespondance , 
				   p_typeIndexation ,
				   _True , _True , 
				   p_ptFicSortie );
	    fprintf( p_ptFicSortie , "\n" );
	  }
	  else {
	    EcrireReglesUtilisees( p_ptBlocIniCou ,
				   p_tCorrespondance , 
				   p_typeIndexation ,
				   _True , _False , 
				   p_ptFicSortie );
	    if( FormatDeSortie == _SortieInistNum 
		|| FormatDeSortie == _SortieQA ) {
	      EcrireReglesUtilisees( p_ptBlocIniCou ,
				     p_tCorrespondance , 
				     p_typeIndexation ,
				     _True , _False , 
				     p_ptFicSortie );
	    }
	  }
	}

	if( FormatDeSortie != _Txt ) {
	  EcrireChaineAnalysee( p_estGeneration , p_ptFicSortie );
	  EcrireReglesUtilisees( p_ptBlocIniCou ,
				 p_tCorrespondance , 
				 p_typeIndexation ,
				 _False , _True , 
				 p_ptFicSortie );
	  fprintf( p_ptFicSortie , "\n" );

	  /* ecriture des mots de la phrase lue dans le
	     fichier de sortie */
	  if( TraceTexteEnSortie == _SortieApresChaqueIndex
	      /* phrase apres chaque occurrence */) {
	    TabMotMem           ptMotCou;
	    ptMotCou = MotsPhrase;
	    while( ptMotCou ) {
	      if( FormatDeSortie != _SortieLexter ) {
		fprintf( p_ptFicSortie , "%s " , 
			 ptMotCou->ChaineMot );
	      }
	      else {
		fprintf( p_ptFicSortie , "%s" , 
			 ptMotCou->ChaineMot );
	      }
	      ptMotCou = ptMotCou->MotFrere;
	    }
	    fprintf( p_ptFicSortie , "\n" );
	  }

	  if( Trace ) {
	    fprintf( FicLog , "\n" );
	  }
	}

	/* en indexation on ne met que la premiere unification    */
	/* reussie d'une regle                                    */
	if( p_typeIndexation == _Indexation 
	    || p_typeIndexation == _IndexationVariations
	    || p_typeIndexation == _Variations ) {
	  *p_cutProlog = _True;
	}
      }
      /* plus de concatenation ; succes ; on l'affiche */

      *p_uneConcTrouvee = _False;
      break;
    }
    /* fin de pile */

    ptBlocIniAux = (TabBlocInitial)(Pile + hautPileAux);

    if( Trace ) {
      fprintf( FicLog ,
       "Backtrack after success. Block type %d ref %ld #altern. (%d/%d)\n",
	       ptBlocIniAux->typeBloc,
	       (ptBlocIniAux->T1.PtTerme)->ValTerme.Ref,
	       ptBlocIniAux->noChoix,
	       ptBlocIniAux->nbChoix );
    }

    if( ptBlocIniAux->noChoix < ptBlocIniAux->nbChoix
	&& ptBlocIniAux->typeBloc == _Concatenation ) {
      /* on a trouve une concatenation incomplete, elle devient le
	 pere du bloc suivant */
      *p_hautPilePrec = HautPile;
      HautPile = p_ptBlocIniCou->ptBlocSuivant;

      /* le pere est la concatenation incomplete */
      *p_hautPilePere = hautPileAux;

      break;
    }
    /* concatenation incomplete trouvee */
    hautPileAux = ptBlocIniAux->ptBlocPrecedent;
  }
  /* recherche d'une concatenation incomplete */
    
  return unifOk;
}

/*************************************************************************
        backTrack echec :
          recherche d'une liste incomplete en depilant les blocs au fur
             et a mesure
**************************************************************************/
void     DepileJusquaChoixIncomplet( TabBlocInitial *p_ptBlocIniCou ,
               TabBlocInitial *p_ptBlocIniPere , int p_estGeneration ,
               int p_cutProlog ,
               unsigned long *p_hautPilePere , unsigned long *p_hautPilePrec)
{
  TabBlocInitial        ptBlocIniAux;
  char                  typeBlocCou;
  char                  indChoix;
  unsigned long         ptMemAux;
  TabRegleMem           *hdlRegleAux;
  TabLemmesMots          ptLemmeMotAux;

  ptBlocIniAux = *p_ptBlocIniCou;

  while( _True ) {
    /* depilement d'un bloc */
    ptBlocIniAux = (TabBlocInitial)(Pile + HautPile);
    *p_ptBlocIniPere = (TabBlocInitial)
      (Pile + ptBlocIniAux->ptBlocPere);
    if( Trace && ptBlocIniAux ) {
      long   ref = 0L;

      if( ptBlocIniAux->T1.PtTerme ) {
	ref = (ptBlocIniAux->T1.PtTerme)->ValTerme.Ref;
      }
          
      fprintf( FicLog ,
	       "Backtrack after failure. Block type %d ref %ld #altern. (%d/%d)\n",
	       ptBlocIniAux->typeBloc,
	       ref,
	       ptBlocIniAux->noChoix,
	       ptBlocIniAux->nbChoix );
    }

    /* si on est sur un choix incomplet on s'arrete , sauf si on est 
       en cut, auquel cas on redescend toute la pile                  */
    if( ptBlocIniAux->noChoix < ptBlocIniAux->nbChoix ) {
      typeBlocCou = ptBlocIniAux->typeBloc;

      if( typeBlocCou == _Liste
	  || typeBlocCou == _Feuille
	  || typeBlocCou == _UniteLexicale
	  || typeBlocCou == _Regle ) {
	/* si on est sur une regle, on desalloue la transformee
	   courante */
	if( typeBlocCou == _Regle ) {
	  /* le pointeur vers regle d'origine est au-dessus du bloc */
	  ptMemAux = HautPile + sizeof(struct BlocInitial);
	  hdlRegleAux = (TabRegleMem *)(Pile + ptMemAux);

	  /* on desalloue la transformation courante */
	  DemarqueEtRendArbre( &((*hdlRegleAux)->Transformee) );
	  (*hdlRegleAux)->Transformee = NULL;
	  (*hdlRegleAux)->TransformeeAvecUnificationExterne = _False;
	  if( Trace ) {
	    fprintf( FicLog ,"Desallocation of transformed rule\n");
	  }
	}

	/* pour un cut, on ne s'arrete que lorsque 
	   on est revenu en bas de pile */
	if( p_cutProlog && HautPile == _BasPile ) {
	  break;
	}

	/* montee dans l'arbre d'analyse */
	/* chainage des blocs suivants et precedents */
	*p_hautPilePrec = HautPile;
	HautPile = ptBlocIniAux->ptBlocSuivant;

	/* chainage des blocs peres et fils */
	/* le pere est le bloc courant */
	*p_hautPilePere = *p_hautPilePrec;

	/* on a trouve un choix incomplet, on s'arrete sauf 
	   si on est en succes sur indexation */
	if( !p_cutProlog ) {
	  break;
	}
      }
    }
    /* si on est sur un choix incomplet on s'arrete */
    else
      /* si on est sur un choix complet et s'il s'agit d'une
         regle, on la desalloue ; s'il s'agit d'une unite
         lexicale en generation, on desalloue tous ses homographes */
      {
	if( ptBlocIniAux->typeBloc == _Regle
	    && ptBlocIniAux->nbChoix > 0 ) {
	  /* la regle d'origine est au-dessus du bloc */
	  ptMemAux = HautPile + sizeof(struct BlocInitial);
	  hdlRegleAux = (TabRegleMem *)(Pile + ptMemAux);

	  /* on desalloue la transformation courante */
	  DemarqueEtRendArbre( &((*hdlRegleAux)->Transformee) );
	  (*hdlRegleAux)->TransformeeAvecUnificationExterne = _False;
	  (*hdlRegleAux)->Transformee = NULL;

	  /* on restitue la memoire occupee par cette regle s'il
	     ne s'agit pas d'une regle memoire et si ce n'est pas
	     la regle initiale */
	  if( (ptBlocIniAux->T1.PtTerme) 
	      && (ptBlocIniAux->T1.PtTerme)->ValTerme.PosFicRegle >= 0L
	      && HautPile != _BasPile ) {
	    DemarqueEtRendRegle( hdlRegleAux );
	    if( Trace ) {
	      fprintf( FicLog ,"Desallocation secondary rule\n");
	    }
	  }
	}
	if( p_estGeneration
	    && (ptBlocIniAux->typeBloc == _UniteLexicale
		|| ptBlocIniAux->typeBloc == _Feuille) ) {
	  /* 1er terme de la liste */
	  ptLemmeMotAux = (TabLemmesMots)
	    ((TabChar)ptBlocIniAux + sizeof(struct BlocInitial));

	  for( indChoix = 0 ; indChoix < ptBlocIniAux->noChoix ;
	       indChoix++ ) {
	    DemarqueEtRendLemme( &(ptLemmeMotAux[ (int)indChoix ].PtLemme) );
	  }
	}

	/* fin de pile */
	if( HautPile == _BasPile ) {
	  break;
	}
	/* fin de pile */
      }
    /* si on est sur un choix complet */
    /* depilement d'un bloc */

    /* restauration de l'environnement */
    RestaurerEnvironnement( ptBlocIniAux , *p_ptBlocIniPere );

    HautPile = ptBlocIniAux->ptBlocPrecedent;
  }
  *p_ptBlocIniCou = ptBlocIniAux;
}

/*************************************************************************
                ATTRIBUTION AUTOMATIQUE DE METAREGLES
**************************************************************************/
void     AttributionAutomatiqueMetaregles( TabTermeMem p_ptFilsRegle )
{
  int                                   nbFils;
  TabTermeMem                           ptTerme;
  struct ValAtomeMem                    atome;

  ptTerme = p_ptFilsRegle;
  nbFils = 0;
  if( ptTerme && ptTerme->ValTerme.TypeTerme == _Concatenation ) {
    ptTerme = ptTerme->TermeFils;
    while( ptTerme ) {
      nbFils++;
      ptTerme = ptTerme->TermeFrere;
    }
  }

  if( nbFils && TabListeMetaReglesStd[ nbFils ]
      && (TabListeMetaReglesStd[ nbFils ])->TypeDag == _AtomeChaine ) {
    atome = (TabListeMetaReglesStd[ nbFils ])->ValDag.Atome;
    AjouteCheminEtAtomeFils( &(p_ptFilsRegle->ValTerme.ArcsFils) ,
			     CheminEtiquetteMeta , _AtomeChaine , 
			     atome , _True );
  }
}

/*************************************************************************
      BOUCLE D'EFFACEMENT : EFFECTUE L'EFFACEMENT + LE BACKTRACKING
      TANT QUE LA PILE N'EST PAS VIDE
**************************************************************************/
char     EffacementUneRegle( TabRegleMem p_ptRegleRacine  , int p_estGeneration ,
                             int p_typeIndexation , char *p_estVariante ,
                             TabCorrIdent_PosFic p_tCorrespondance ,
                             long *p_nbUnifs , long *p_nbUnifsSucces , 
                             long *p_nbMetaRegles , FILE *p_ptFicSortie )
{
  unsigned long         hautPilePere;
  unsigned long         hautPilePrec;
  unsigned long         ptMemAux;

  TabBlocInitial        ptBlocIniPere;
  TabBlocInitial        ptBlocIniCou;

  TabRegleMem           ptReglePere;
  TabRegleMem          *hdlRegleAux;

  char                  backTrackSucces;
  char                  backTrackEchec;
  char                  cutProlog;  /* cut en cas de succes en indexation */
  char                  unifOk;  /* vrai si unification sur succes reussie */
  char                  uneConcTrouvee;
  char                  noChoixPere;
  TabTermeMem          *hdlTermeAux;
  TabLemmesMots         ptLemmeMot;

  int                   indDebutER;
  int                   indFinER;
  int                   transformeeAvecUnificationExterne;

  cutProlog = _False;
  unifOk = _False;
  *p_estVariante = _False;

  hautPilePrec = (long)NULL;
  hautPilePere = (long)NULL;
  HautPile = _BasPile;

  /* initialisation : on met un terme bidon d'appel de la regle initiale
     dans le bas de la pile */
  ptBlocIniCou = (TabBlocInitial)(Pile + HautPile);
  ptBlocIniCou->ptBlocPrecedent = hautPilePrec;
  ptBlocIniCou->ptBlocPere = hautPilePere;

  ptBlocIniCou->typeBloc = _Regle;
  ptBlocIniCou->T1.PtTerme = NULL;

  /* le pointeur vers le fils de la tete est recopie dans T2 */
  AjouterTermesSuivants( ptBlocIniCou , &backTrackSucces ,
			 &backTrackEchec , p_ptRegleRacine , 
			 p_estGeneration , p_typeIndexation );

  /* montee dans l'arbre d'analyse */
  /* chainage des blocs suivants et precedents */
  hautPilePrec = HautPile;
  HautPile = ptBlocIniCou->ptBlocSuivant;
  /* chainage des blocs peres et fils : le pere est le bloc pdent
     dans ce cas */
  hautPilePere = hautPilePrec;

  /* boucle infinie : l'effacement n'est pas recursif */
  while( _True ) {
    if( HautPile >= TaillePileEffacement ) {
      sprintf( ChAux , "EffacementUneRegle(alanalys) [%ld]",
	       p_ptRegleRacine->TermeFils->ValTerme.Ref );
      TxtErreur( 133 , ChAux );
      exit(0);
    }

    ptBlocIniPere = (TabBlocInitial)(Pile + hautPilePere);
    ptBlocIniCou = (TabBlocInitial)(Pile + HautPile);
    ptBlocIniCou->ptBlocPrecedent = hautPilePrec;
    ptBlocIniCou->ptBlocPere = hautPilePere;

    SauverEnvironnement( ptBlocIniCou , ptBlocIniPere );

    if( Trace && ptBlocIniPere ) {
      long        ref = 0L;

      if( ptBlocIniPere->T1.PtTerme ) {
	ref = (ptBlocIniPere->T1.PtTerme)->ValTerme.Ref;
      }
      fprintf( FicLog ,
	       "*Mother: ty(%d) ref(%ld) alt.(%d/%d)\n",
	       ptBlocIniPere->typeBloc,
	       ref,
	       ptBlocIniPere->noChoix,
	       ptBlocIniPere->nbChoix);
    }

    /* choix en fonction du type de bloc du pere */
    switch( ptBlocIniPere->typeBloc ) {
    case _Regle :
      /* le pointeur vers regle d'origine est dans le T2 du pere */
      ptMemAux = hautPilePere + sizeof(struct BlocInitial);
      hdlRegleAux = (TabRegleMem *)(Pile + ptMemAux);
      ptReglePere = *hdlRegleAux;

      /* incrementation des choix du pere */
      noChoixPere = ++(ptBlocIniPere->noChoix);

      if( noChoixPere >= 2 ) {
	if( Trace ) {
	  fprintf( FicLog ,"Variant # (%d/%d) of rule %ld\n",
		   noChoixPere - 1 ,
		   DagANbMetaregle( 
			   ptReglePere->TermeFils->ValTerme.ArcsFils ) ,
		   ptReglePere->TermeFils->ValTerme.Ref);
	}

	/* les metaregles sont numerotees de 0 a NbMetaRegles - 1
	   pour la fonction GenererMetaRegle                    */
	if( !GenererMetaRegle( ptReglePere , (int)(noChoixPere) - 2 ) ) {
	  TxtErreur( 183 , "EffacementUneRegle(alanalys)" );
	  exit(0);
	}
	(*p_nbMetaRegles)++;
	*p_estVariante = _True;
	DebutMotsReconnus = (int)_Nil;
	FinMotsReconnus = (int)_Nil;

	if( Trace ) {
	  EcrireRegleMem( &ptReglePere , _False , FicLog , _Sortie );
	}
	ptBlocIniCou->typeBloc = _RegleMeta;
      }
      else {
	/* le T2 du pere pointe toujours vers la regle d'origine */
	ptBlocIniCou->typeBloc = _RegleNormale;
      }

      /* la tete est recopiee dans T1 */
      ptBlocIniCou->T1.PtRegle = ptReglePere;

      break;

    case _RegleNormale :
    case _RegleMeta :
    case _RegleElision :
      /* l'incrementation du nombre de choix ou de concatenation du pere
	 est bidon , il passe toujours de 0 a 1 dans ce cas.        */
      ++(ptBlocIniPere->noChoix);

      /* On recopie le fils de la regle du bloc pere dans le terme
	 courant */
      ptReglePere = ptBlocIniPere->T1.PtRegle;
      if( ptBlocIniPere->typeBloc == _RegleNormale ) {
	ptBlocIniCou->T1.PtTerme = ptReglePere->TermeFils;
	transformeeAvecUnificationExterne = _False;
      }
      else {
	if( ptReglePere->Transformee == NULL ) {
	  TxtErreur( 11 , "EffacementUneRegle(alanalys)" );
	  exit(0);
	}
	ptBlocIniCou->T1.PtTerme = ptReglePere->Transformee;
	transformeeAvecUnificationExterne =
	  ptReglePere->TransformeeAvecUnificationExterne;
      }

      /* le type du bloc courant est celui du terme courant */
      ptBlocIniCou->typeBloc 
	= (ptBlocIniCou->T1.PtTerme)->ValTerme.TypeTerme;
      break;

    case _Concatenation :
    case _Liste :
      /* incrementation du numero de choix, ou de l'indice de
	 concatenation */
      ++(ptBlocIniPere->noChoix);

      /* recherche du ieme terme de la liste dans le pere */
      /* 1er terme de la liste */
      hdlTermeAux = (TabTermeMem *)
	((TabChar)ptBlocIniPere + sizeof(struct BlocInitial));

      /* ieme terme de la liste */
      hdlTermeAux += ptBlocIniPere->noChoix - 1;

      /* on recopie le ieme terme de la liste dans T1 */
      ptBlocIniCou->T1.PtTerme = *hdlTermeAux;

      /* le type du bloc courant est celui de T1 */
      ptBlocIniCou->typeBloc = (*hdlTermeAux)->ValTerme.TypeTerme;

      break;

    case _Feuille :             /* paradigme defini en intension */
    case _UniteLexicale :       /* lemme dont on connait la reference */
      /* incrementation de l'indice d'homographe du pere */
      ++(ptBlocIniPere->noChoix);

      /* recherche du ieme terme de la liste dans le pere */
      /* 1er terme de la liste */
      ptLemmeMot = (TabLemmesMots)
	((TabChar)ptBlocIniPere + sizeof(struct BlocInitial));

      /* on recopie le ieme terme de la liste dans T1 */
      ptBlocIniCou->T1.PtLemme 
	= ptLemmeMot[ ptBlocIniPere->noChoix - 1 ].PtLemme;
      IndiceMotCou = ptLemmeMot[ ptBlocIniPere->noChoix - 1 ].IndMot + 1;
      indDebutER = ptLemmeMot[ ptBlocIniPere->noChoix - 1 ].IndDebutER;
      indFinER = ptLemmeMot[ ptBlocIniPere->noChoix - 1 ].IndFinER;

      /* on note les mots de l'expression reguliere precedente
	 s'il y an a une */
      if( indDebutER != _Nil && indDebutER < IndiceMotCou - 1 ) {
	if( DebutMotsReconnus == _Nil ) {
	  DebutMotsReconnus = indDebutER;
	}
      }

      /* on note le mot courant */
      if( DebutMotsReconnus == _Nil ) {
	DebutMotsReconnus = IndiceMotCou - 1;
      }
      FinMotsReconnus = IndiceMotCou - 1;

      /* on note les mots de l'expression reguliere suivante
	 s'il y an a une */
      if( indFinER != _Nil && indFinER - 1 > IndiceMotCou - 1 ) {
	FinMotsReconnus = indFinER - 1;
      }

      /* le type du bloc courant est celui de T1 */
      ptBlocIniCou->typeBloc = _MotFlechi;

      break;

    default :
      sprintf( ChAux , "EffacementUneRegle(alanalys) [%ld,%d]" ,
	       (ptBlocIniPere->T1.PtTerme)->ValTerme.Ref ,
	       ptBlocIniPere->typeBloc );
      TxtErreur( 12 , ChAux );
      break;
    }
    /* switch sur le type de bloc pere */

    if( Trace ) {
      fprintf( FicLog ,
	 "Current block: ty(%d/moth %d) ref(%ld) alt.(%d/%d) (moth %d/%d)\n",
	       ptBlocIniCou->typeBloc,
	       ptBlocIniPere->typeBloc,
	       (ptBlocIniCou->T1.PtTerme)->ValTerme.Ref,
	       ptBlocIniCou->noChoix,
	       ptBlocIniCou->nbChoix,
	       ptBlocIniPere->noChoix,
	       ptBlocIniPere->nbChoix);
    }

    /* chainage des blocs precedents et suivants */
    /* les pointeurs vers les fils sont recopies dans T2 */
    AjouterTermesSuivants( ptBlocIniCou , &backTrackSucces ,
			   &backTrackEchec , NULL , p_estGeneration ,
			   p_typeIndexation );

    if( !backTrackSucces && !backTrackEchec ) {
      /* montee dans l'arbre d'analyse */
      /* chainage des blocs suivants et precedents */
      hautPilePrec = HautPile;
      HautPile = ptBlocIniCou->ptBlocSuivant;

      /* chainage des blocs peres et fils : le pere est le bloc precedent
	 dans ce cas */
      hautPilePere = hautPilePrec;
    }

    /* backTracking eventuel si le terme courant est une _UniteLexicale
       ou un _MotFlechi */
    if( backTrackSucces ) {
      unifOk |= RechercheConcIncomplete( 
					ptBlocIniCou , p_estGeneration , 
					p_typeIndexation , &uneConcTrouvee ,
					&hautPilePere , &hautPilePrec ,
					p_tCorrespondance , &cutProlog ,
					p_nbUnifs , p_nbUnifsSucces ,
					p_ptFicSortie , 
					transformeeAvecUnificationExterne );
      if( !uneConcTrouvee ) {
	backTrackEchec = _True;
      }
    }
    /* backTrack succes */

    /* backTrack echec : on cherche a prendre le choix suivant de la 1ere
       liste incomplete trouvee. */
    /* ou succes en indexation : on redescend toute la pile */
    if( backTrackEchec || cutProlog ) {
      DepileJusquaChoixIncomplet( 
				 &ptBlocIniCou , &ptBlocIniPere , 
				 p_estGeneration ,
				 cutProlog , &hautPilePere , &hautPilePrec ) ;

      /* fin de pile */
      if( HautPile == _BasPile ) {
	/* plus de choix : echec definitif ; on l'affiche */
	if( Trace ) {
	  fprintf( FicLog ,"Unsuccessfull parse\n");
	}
	return unifOk;
      }
      /* fin de pile */
    }
    /* recherche d'une concatenation incomplete */
  }
  /* boucle infinie : l'effacement n'est pas recursif */
}

/*************************************************************************
         EFFACEMENT D'UNE LISTE DE REGLES
**************************************************************************/
char     EffacementListeRegle( TabRegleMem p_ptListeRegle , 
                               int p_typeIndexation , int p_estRegleMem ,
                               TabCorrIdent_PosFic p_tCorrespondance , 
                               long *p_nbRegles , long *p_nbReglesTot ,
                               long *p_nbReglesSucces , 
                               long *p_nbUnifs , long *p_nbUnifsSucces , 
                               long *p_nbMetaRegles , 
                               long *p_nbMetaReglesSucces , 
                               FILE *p_ptFicSortie )
{
  TabRegleMem              ptRegle;
  TabRegleMem              ptRegleSuiv;
  TabTermeMem              ptTerme;
  char                     aEffacer;
  char                     estVariante;
  int                      nbFils;

  ptRegle = p_ptListeRegle;
  while( ptRegle ) {
    aEffacer = _True;

    /* seules les regles d'arite entre AriteMin et AriteMax sont prises */
    if( AriteMin > 0 || AriteMax > 0 ) {
      ptTerme = ptRegle->TermeFils;
      nbFils = 0;
      if( ptTerme ) {
	ptTerme = ptTerme->TermeFils;
	while( ptTerme ) {
	  nbFils++;
	  ptTerme = ptTerme->TermeFrere;
	}
      }
      if( AriteMin > 0 && nbFils < AriteMin ) {
	aEffacer = _False;
      }
      if( AriteMax > 0 && nbFils > AriteMax ) {
	aEffacer = _False;
      }
      if( Trace && !aEffacer ) {
	fprintf( FicLog , 
		 "Incorrect arity for rule %ld %d vs[%d-%d]\n" , 
		 ptRegle->TermeFils->ValTerme.Ref , nbFils ,
		 AriteMin , AriteMax );
      }
    }
      
    /* pour optimiser les temps d'analyse en 
       extraction terminologique, on verifie
       que tous les lemmes d'une regle sont presents dans la phrase.
       Cette optimisation tombe en defaut dans de nombreux cas :
       - metaregles elidant certains lemmes
       - regles ou metaregles comportant des disjonctions ... etc */
    if( aEffacer 
	&& ( LexicalisationMultiple || p_typeIndexation == _Collocations ) ) {
      {
	aEffacer = TousLemmesPresentsDansPhrase( ptRegle );
	if( Trace ) {
	  fprintf( FicLog , "A effacer %d rule %ld\n" , aEffacer ,
		   ptRegle->TermeFils->ValTerme.Ref );
	}
      }
    }
      
    if( aEffacer && p_typeIndexation == _Collocations ) {
      ChercheEtImprimeCollocation( ptRegle , p_tCorrespondance ,
				   p_ptFicSortie );
      aEffacer = _False;
    }

    DebutMotsReconnus = (int)_Nil;
    FinMotsReconnus = (int)_Nil;
    IndiceMotCou = (int)_Nil; 
    (*p_nbReglesTot)++;
    ptRegleSuiv = ptRegle->RegleFrere;
    if( aEffacer ) {
      if( Trace ) {
	fprintf( FicLog , "Rule %ld\n" , 
		 ptRegle->TermeFils->ValTerme.Ref );
	fprintf( FicLog ,"Memory before rule parsing\n");
	PhotoMemoire( FicLog );
      }
      
      /* en indexation sur les variations, on garnit
	 automatiquement les metaregles a + en attribuant la bonne
	 liste de metaregles en fonction de la structure */
      if( !p_estRegleMem && ( p_typeIndexation == _IndexationVariations
			      || p_typeIndexation == _Variations
			      || p_typeIndexation == _Analyse ) ) {
	AttributionAutomatiqueMetaregles( ptRegle->TermeFils );
      }

      /* si variations seules, on ne tente l'effacement de la regle
	 initiale que sur ses transformations si elle en a */
      if( !( p_typeIndexation == _Variations 
	     && DagANbMetaregle( ptRegle->TermeFils->ValTerme.ArcsFils ) 
	     <= 0 ) ) {
	if( EffacementUneRegle( ptRegle , _False , p_typeIndexation , 
				&estVariante , p_tCorrespondance , 
				p_nbUnifs , p_nbUnifsSucces ,
				p_nbMetaRegles , p_ptFicSortie ) ) {

	  (*p_nbReglesSucces)++;
	  if( estVariante ) {
	    (*p_nbMetaReglesSucces)++;
	  }
	}

	(*p_nbRegles)++;
      }
    }
    else if( Trace ) {
      fprintf( FicLog , "Parsing rule %ld dismissed\n" , 
	       ptRegle->TermeFils->ValTerme.Ref );
    }
    ptRegle = ptRegleSuiv;
      
    if( Trace ) {
      fprintf( FicLog ,"Memory after rule parsing\n");
      PhotoMemoire( FicLog );
    }

    if( _ClickDansAnnulerInfo() ) {
      return _False;
    }
  }

  return _True;
}

/*************************************************************************
            INDEXATION D'UN TEXTE
**************************************************************************/
void     IndexationOuAnalyse( char *p_nomFicEntree , char *p_nomFicSortie , 
                              int p_typeIndexation )
{
  TabLemmeMem                ptLemmeCou;
  TabRegleMem                ptRegleCou;
  char                       car;
  long                       nbMotsTot = 0L;
  int                        nbPhrases;
  char                       chMotsIndexes[ 64 ];
  TabCorrIdent_PosFic        tCorrespondance;
  long                       nbRegles;
  long                       nbReglesSucces;
  long                       nbReglesSuccesCou;
  long                       nbReglesTot;
  long                       nbUnifs;
  long                       nbUnifsSucces;
  long                       nbMetaRegles;
  long                       nbMetaReglesSucces;
  int                        traceCou;
  FILE                      *ptFicSortie;
  FILE                      *ptFicEntree;
  TabMotMem                  ptMotCou;
  TabMotMem                  ptMotAux;

  if( strcmp( p_nomFicEntree , "stdin" ) == 0 ) {
    ptFicEntree = stdin;
  }
  else {
    if( !OuvreFichier( p_nomFicEntree , &ptFicEntree , _FopenRT ) ) {
      return;
    }
    fseek( ptFicEntree , 0L , 0 );
  }
  FicBufferiseCourant = ptFicEntree;

  if( strcmp( p_nomFicSortie , "stdout" ) == 0 ) {
    ptFicSortie = stdout;
  }
  else {
    if( !OuvreFichier( p_nomFicSortie , &ptFicSortie , _FopenWT ) ) {
      return;
    }
  }

  _NouveauDialogueInfo( 128 , _True );
  TxtTitresDialogueInfo( 16 , 0 , 0 );
  tCorrespondance = NULL;
  traceCou = Trace;
  Trace = _False;
  if( !AlloueEtChargeCorrespondances( FicRefIdent , &tCorrespondance ) ) {
    if( ptFicEntree != stdin ) {
      fclose( ptFicEntree );
    }
    return;
  }
  Trace = traceCou;
  _FinDialogueInfo();

  if( !BDicSurDisque ) {
    AlloueBuffersLongsEtChargeDico();
  }

  _NouveauDialogueInfo( 128 , _True );
  if( p_typeIndexation == _Indexation ) {
    TxtTitresDialogueInfo( 1 , 2 , 3 );
  }
  if( p_typeIndexation == _Analyse ) {
    TxtTitresDialogueInfo( 4 , 5 , 6 );
  }
  if( p_typeIndexation == _IndexationVariations ) {
    TxtTitresDialogueInfo( 30 , 2 , 3 );
  }
  if( p_typeIndexation == _Variations ) {
    TxtTitresDialogueInfo( 35 , 2 , 3 );
  }

  if( HeaderEnSortie ) {
    _LaDate( ChAux );
    fprintf( FicLog , "%s" , ChAux);
  }
 
  IndCar = (long)TailleBuffStd;
  IndCarM = 0L;
  NbLignesTot = NoPremLigne - 1L;
  nbRegles = 0L;
  nbReglesSucces = 0L;
  nbReglesTot = 0L;
  nbUnifs = 0L;
  nbUnifsSucces = 0L;
  nbMetaRegles = 0L;
  nbMetaReglesSucces = 0L;
  nbPhrases = 100;
  car = _LireC( ptFicEntree );

  if( Trace ) {
    fprintf( FicLog ,"\n---\n\n");
    fprintf( FicLog ,"Initial memory\n");
    PhotoMemoire( FicLog );
    fprintf( FicLog ,"\n---\n\n");
  }

  while( car != EOF ) {
/*              fprintf(FicLog , "Nouvelle Phase\n"); */
/*              PhotoMemoire( FicLog ); */
/*              fprintf(FicLog , "\n"); */

    nbReglesSuccesCou = nbReglesSucces;

    if( FormatDeSortie == _SortieQA ) {
      fprintf( ptFicSortie , "<SENT>\n" );
    }

    if( FormatDEntree != _TxtTagge ) {
      LirePhrase( &car , ptFicEntree );
      PhraseALemme();
    }
    else {
      if( !LirePhraseTagge( &car , ptFicEntree ) ) {
	break;
      }
    }
    nbMotsTot += (long)NbMots;

/* for( indMot = 0 ; indMot < NbMots ; indMot++ ) {  */
/* fprintf( ptFicSortie , "Mot: %s" ,   */
/*  (Mots + indMot)->ChaineMot );  */
/* }  */
/* fprintf( ptFicSortie , "\n" );  */

    /* ecriture des lemmes de la phrase lue dans le fichier de mesages */
    if( Trace ) {
      fprintf( FicLog ,"\n---\n\n");
      fprintf( FicLog , 
	       "Current words (%d word(s)):\n" ,
	       NbMots );
      EcrirePhraseEtiquetee( FicLog , _EtiquetageFin );

      fprintf( FicLog ,"\n---\n\n");
      fprintf( FicLog , "Current Rules\n" );
      ptRegleCou = ReglesPhrase;
      while( ptRegleCou ) {
	EcrireRegleMem( &ptRegleCou , _False , FicLog , _Sortie );
	ptRegleCou = ptRegleCou->RegleFrere;
      }
      if( NbReglesMem > 0 ) {
	fprintf( FicLog ,"\n---\n\n");
	fprintf( FicLog , "Current memory rules\n" );
	ptRegleCou = ReglesMem;
	while( ptRegleCou ) {
	  EcrireRegleMem( &ptRegleCou , _False , FicLog , _Sortie );
	  ptRegleCou = ptRegleCou->RegleFrere;
	}
      }
      fprintf( FicLog , "\n" );
    }

    /* ecriture de la phrase etiquetee */
    if( p_typeIndexation == _Etiquetage 
	|| p_typeIndexation == _EtiquetageFin ) {
      EcrirePhraseEtiquetee( ptFicSortie , p_typeIndexation );
    }
    /* ecriture de la phrase etiquetee */
    else
      /* indexation */
      {
	/* ecriture des mots de la phrase lue dans le fichier de sortie */
	if( TraceTexteEnSortie == _SortieSystematique 
	    && p_typeIndexation != _Etiquetage && NbMots > 0 ) {
	  ptMotCou = MotsPhrase;
	  while( ptMotCou ) {
            if( FormatDeSortie != _SortieLexter )
            {
	      fprintf( ptFicSortie , "%s " , 
		       ptMotCou->ChaineMot );
	    }
	    else {
	      fprintf( ptFicSortie , "%s" , 
		       ptMotCou->ChaineMot );
	    }
	    ptMotCou = ptMotCou->MotFrere;
	  }
	  fprintf( ptFicSortie , "\n" );
	}
	/* sortie Amaryllis : les tags xml sont ressortis systemantiquement */
	else if( FormatDeSortie == _SortieAmaryllis ) {
	  /* ecrit les phrases dont le premier mot est un XML */
	  ptMotCou = MotsPhrase;
	  ptLemmeCou = ptMotCou->PtListeLemmes;
	  if( CategorieLexicale( &((ptLemmeCou)->ValTerme.ArcsFils) ) 
	      == RefDeterminantNumeral ) {
	    while( ptMotCou ) {
	      fprintf( ptFicSortie , "%s " , 
		       ptMotCou->ChaineMot );
	      ptMotCou = ptMotCou->MotFrere;
	    }
	    fprintf( ptFicSortie , "\n" );
	  }
	}

	if( FormatDeSortie == _SortieQA ) {
	  fprintf( ptFicSortie , "<TERMS>\n" );
	}

	if( !EffacementListeRegle( ReglesPhrase , p_typeIndexation , 
				   _False , tCorrespondance , 
				   &nbRegles , &nbReglesTot , 
				   &nbReglesSucces ,
				   &nbUnifs , &nbUnifsSucces , 
				   &nbMetaRegles , &nbMetaReglesSucces ,
				   ptFicSortie ) ) {
	  break;
	}

	if( NbReglesMem > 0 ) {
	  if( !EffacementListeRegle( ReglesMem , p_typeIndexation , 
				     _True , tCorrespondance , 
				     &nbRegles  , &nbReglesTot , 
				     &nbReglesSucces ,
				     &nbUnifs , &nbUnifsSucces ,
				     &nbMetaRegles , 
				     &nbMetaReglesSucces , 
				     ptFicSortie ) ) {
	    break;
	  }
	}

	/* ecriture des mots de la phrase lue dans le fichier de sortie */
	if( TraceTexteEnSortie == _SortieSiIndex 
	    && nbReglesSuccesCou != nbReglesSucces 
	    /* succes d'au moins une regle */) {
	  ptMotCou = MotsPhrase;
	  while( ptMotCou ) {
	      if( FormatDeSortie != _SortieLexter ) {
		fprintf( ptFicSortie , "%s " , 
			 ptMotCou->ChaineMot );
	      }
	      else {
		fprintf( ptFicSortie , "%s" , 
			 ptMotCou->ChaineMot );
	      }
	      ptMotCou = ptMotCou->MotFrere;
	  }
	  fprintf( ptFicSortie , "\n" );
	}
      }
    /* indexation */

    if( FormatDeSortie == _SortieQA ) {
      fprintf( ptFicSortie , "</TERMS>\n" );
      fprintf( ptFicSortie , "</SENT>\n" );
    }

    if( nbPhrases++ > 10 ) {
      nbPhrases = 0;
      sprintf( ChAux , "%ld" , nbMotsTot );
      TxtEcritItemInfo( 1 , ChAux );
      sprintf( ChAux , "%ld" , nbRegles );
      TxtEcritItemInfo( 2 , ChAux );
    }

    /* on desalloue les regles chargees */
    if( Trace ) {
      PhotoMemoire( FicLog );
      fprintf( FicLog ,"Desallocation of active rules\n");
    }
    DemarqueEtRendListeRegle( &ReglesPhrase );

    if( Trace ) {
      PhotoMemoire( FicLog );
      fprintf( FicLog ,"Desallocation of current lemmas\n\n");
    }

    /* on desalloue les termes des homographes de mots de la phrase */
    ptMotCou = MotsPhrase;
    while( ptMotCou ) {
      ptMotAux = ptMotCou->MotFrere;

      DemarqueEtRendListeLemme( &(ptMotCou->PtListeLemmes) );
      DemarqueEtRendDag( &(ptMotCou->ArcsFils) );
      RendMot( &ptMotCou );

      ptMotCou = ptMotAux;
    }

    if( Trace ) {
      PhotoMemoire( FicLog );
    }
  }

  if( Trace ) {
    fprintf( FicLog ,"\n---\n\n");
    fprintf( FicLog ,"Final memory\n");
    PhotoMemoire( FicLog );
    fprintf( FicLog ,"\n---\n\n");
  }

  if( HeaderEnSortie ) {
    TxtChaineRsrc( 134 , 2 , chMotsIndexes );
    _LaDate( ChAux );
    fprintf( FicLog , "%s%s %ld\n" , 
	     ChAux ,  chMotsIndexes , nbMotsTot );
    TxtChaineRsrc( 134 , 3 , chMotsIndexes );
    fprintf( FicLog , "%s (%ld/%ld/%ld)\n" , 
	     chMotsIndexes , nbReglesTot , nbRegles , nbReglesSucces );
    TxtChaineRsrc( 134 , 15 , chMotsIndexes );
    fprintf( FicLog , "%s (%ld/%ld)\n" , 
	     chMotsIndexes , nbUnifs , nbUnifsSucces );
    TxtChaineRsrc( 134 , 16 , chMotsIndexes );
    fprintf( FicLog , "%s (%ld/%ld)\n" , 
	     chMotsIndexes , nbMetaRegles , nbMetaReglesSucces );
  }

  _FinDialogueInfo();

  if( ptFicEntree && ptFicEntree != stdin ) {
    fclose( ptFicEntree );
    FicBufferiseCourant = NULL;
  }
  if( tCorrespondance ) {
    _Free( tCorrespondance );
  }
  if( !BDicSurDisque ) {
    DesalloueBuffersLongsEtDechargeDico();
  }

  if( strcmp( p_nomFicSortie , "stdout" ) != 0 && ptFicSortie
      && ptFicSortie != stdout  ) {
    fclose( ptFicSortie );
  }
}

/*************************************************************************
            GENERATION D'UNE REGLE
**************************************************************************/
void     GenerationUneRegle( ChIdent p_chMot )
{
  unsigned long                   indAbsLemmeCou;
  unsigned long                   indAbsRegleCou;
  unsigned long                   longFic;
  struct LemmeMem                 lemmeCou;
  char                            estVariante;
  TabRegleMem                     ptRegleCou;
  TabRegleMem                     ptRegleAux;
  int                             nbReglesLoc;
  long                            refRegleCherchee;
  int                             erreur;
  TabCorrIdent_PosFic             tCorrespondance;
  long                            nbUnifs = 0L;
  long                            nbUnifsSucces = 0L;
  long                            nbMetaRegles = 0L;

  /* description du terme releve dans le dico */
  char                            chaineAnnexe[_MaxLetIdent];

  sprintf( ChAux , "%s" , p_chMot );

  tCorrespondance = NULL;
  AlloueEtChargeCorrespondances( FicRefIdent , 
				 &tCorrespondance );

  TxtEcritItemInfo( 2 , ChAux );

  if( Recherche( p_chMot , &refRegleCherchee , 
		 FicRefIdent , &erreur ) == _Nil
      || erreur
      || refRegleCherchee == _Nil ) {
    return;
  }

  NbMaxGenerations = 0;

  /* regles fichier */
  indAbsLemmeCou = (long)( _TailleAmorce + _TailleSeparation );
  fseek( FicBDicLemmes , 0L , 2 );
  fgetpos( FicBDicLemmes , &longFic );
  *(lemmeCou.ValTerme.Ident) = '\0';
  while( indAbsLemmeCou < longFic ) {
    if( RestructurerBDicLemme( &indAbsLemmeCou , lemmeCou.ValTerme.Ident  ,
			       &lemmeCou , chaineAnnexe ) == _Nil 
	|| EgaliteChaine( lemmeCou.ValTerme.Ident , DernLemme ) ) {
      DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
      break;
    }

    indAbsRegleCou = lemmeCou.ValTerme.PosFicRegle;
    if( RestructurerLesRegles( &indAbsRegleCou , &ptRegleCou , 
			       lemmeCou.NbRegles , _True ) == _Nil ) {
      DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
      DemarqueEtRendListeRegle( &ptRegleCou );
      break;
    }

    ptRegleAux = ptRegleCou;
    while( ptRegleCou ) {
      /* generation de la regle trouvee */
      if( ptRegleCou->TermeFils->ValTerme.Ref == refRegleCherchee ) {
	sprintf( ChAux , "%s" , lemmeCou.ValTerme.Ident );
	TxtEcritItemInfo( 1 , ChAux );
	EffacementUneRegle( ptRegleCou , _True , _False , 
			    &estVariante , tCorrespondance , 
			    &nbUnifs , &nbUnifsSucces , 
			    &nbMetaRegles , FicLog );
	DemarqueEtRendListeRegle( &ptRegleAux );
	return;
      }
      ptRegleCou = ptRegleCou->RegleFrere;
    }
    DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
    DemarqueEtRendListeRegle( &ptRegleAux );
  }
  /* regles memoire */
  for( nbReglesLoc = 0 ; nbReglesLoc < NbReglesMem ; nbReglesLoc++ ) {
    ptRegleCou = ReglesMem + nbReglesLoc;
    /* generation de la regle trouvee */
    if( ptRegleCou->TermeFils->ValTerme.Ref == refRegleCherchee ) {
      EffacementUneRegle( ptRegleCou , _True , _False , 
			  &estVariante , tCorrespondance , 
			  &nbUnifs , &nbUnifsSucces , &nbMetaRegles ,
			  FicLog );
      return;
    }
  }
  if( tCorrespondance ) {
    _Free( tCorrespondance );
  }
}

/*************************************************************************
            GENERATION D'UNE REGLE
**************************************************************************/
void     GenerationParFichier( char *p_nomFic )
{
  ChIdent                         chMot;
  char                            carLu;
  FILE                           *ptFicEntree;

  if( !OuvreFichier( p_nomFic , &ptFicEntree , _FopenRT ) ) {
    return;
  }

  _NouveauDialogueInfo( 128 , _True );
  TxtTitresDialogueInfo( 7 , 8 , 9 );

  carLu = ' ';
  while( carLu != EOF ) {
    while( carLu != '.' && carLu != EOF ) {
      fprintf( FicLog , "%c", carLu );
      carLu = LireCar( ptFicEntree );
    }
    fprintf( FicLog , "\n" );
    while( carLu != '<' && carLu != EOF ) {
      carLu = LireCar( ptFicEntree );
    }
    while( carLu == '<' ) {
      ReconnaitreCaractere( ptFicEntree , (int)'<' , &carLu , _True );
      LireIdentificateur ( ptFicEntree , &carLu , chMot , 
			   _MaxLetIdent , _True );
      GenerationUneRegle( chMot );
      if( !ReconnaitreCaractere( ptFicEntree , (int)'>' , 
				 &carLu , _True ) ) {
	carLu = EOF;
      }
    }
    fprintf( FicLog , "\n" );

    if( _ClickDansAnnulerInfo() ) {
      carLu = EOF;
    }
  }

  _FinDialogueInfo();
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  fin de alanalys.c
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
