/*
 * FASt Term Recognizer
 *
 *  FASTR/albarbre.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for B-Tree management
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/****************************************************************************
                      UTILITAIRES
****************************************************************************

   determination du rang du premier caractere qui differencie les chaines 1 et 2
   a partir du rang p_indiceCar (peut peut-etre se transformer en macro)
   utilise pour donner l'indiceCar a un rang superieur du barbre        */

int      PremCarDiscrim( int p_indiceCar , char *p_chaine1 , char *p_chaine2  )
{
    if( EgaliteChaine( p_chaine1 , p_chaine2 ) )
    {
        return p_indiceCar;
    }

    if ( (int)strlen ( p_chaine1 ) > p_indiceCar + 1
         && (int)strlen ( p_chaine2 ) > p_indiceCar + 1 )
    {
            while ( p_chaine1[ p_indiceCar ] == p_chaine2[ p_indiceCar ]
                    && p_chaine1[ p_indiceCar ] != '\0'
                    && p_chaine2[ p_indiceCar ] != '\0' )
            {
                            p_indiceCar++;
            }
    }

    return p_indiceCar;
}

/* egalite de 2 chaines entre 2 bornes                                    */

char     EgalitePartielle( char *p_chaineSource , char *p_chaineDest ,
                           int p_noCarDeb , int p_noCarFin )
{
    int     ind = 0;

    /* saut des 1ers caracteres de la source non recopies           */
    while( ind < p_noCarDeb && p_chaineSource[ ind ] != '\0'
                            && p_chaineDest[ ind ] != '\0' )
    {
             ind++;
    }

    /* recopie des caracteres choisis                               */
    while( ind <= p_noCarFin && p_chaineSource[ ind ] != '\0'
            && p_chaineDest[ ind ] != '\0'
            && p_chaineDest[ ind ] == p_chaineSource[ ind ] )
    {
             ind++;
    }

    return( ind > p_noCarFin 
            || ( ( p_chaineDest[ ind ] == '\0'
                   || p_chaineSource[ ind ] == '\0' )
                 &&  p_chaineDest[ ind - 1 ] == p_chaineSource[ ind - 1] ));
}

/* recopie s'ils existent, les caracteres de la chaine source de p_noCarDeb
   a p_noCarFin dans la chaine de destination                           */

void     RecopieChaine( char *p_chaineSource , char *p_chaineDest , 
                        int p_noCarDeb ,
                        int p_noCarFin )
{
    int      indSrc = 0;
    int      indDest = 0;

    /* saut des 1ers caracteres de la source non recopies           */
    while( indSrc < p_noCarDeb && p_chaineSource[ indSrc ] != '\0' )
    {
             indSrc++;
    }

    /* recopie des caracteres choisis                               */
    while( indSrc <= p_noCarFin && p_chaineSource[ indSrc ] != '\0')
    {
            p_chaineDest[ indDest++ ] = p_chaineSource[ indSrc++ ];
    }
    /* rajout de blancs en fin de chaine                            */
    while( indDest <= p_noCarFin - p_noCarDeb )
    {
            p_chaineDest[ indDest++ ] = ' ';
    }
}

/*****************************************************************************
                        sauver le BArbre
*****************************************************************************/

char     SauveBArbre( void )
{
    FILE                           *ptFic;
    char                           nomFicBArbre[ _NbMaxLetNomFic ];

    strcpy( nomFicBArbre , NomFicDico );
    strcat( nomFicBArbre , ".BTR" );

    if( !OuvreFichier( nomFicBArbre , &ptFic , _FopenWB ) )
    {
         return _False;
    }

    fseek(ptFic , 0L , 0 );
    fwrite( (char *)B_ArbInf , (unsigned long)sizeof(struct PorteNormale) ,
                               (unsigned long)TailleBArbInf  , ptFic );

    fwrite( (char *)B_ArbSup , (unsigned)sizeof(struct PorteSuperieure) ,
                               (unsigned long)TailleBArbSup  , ptFic );

    fclose( ptFic );
    return _True;
}

/*****************************************************************************
                        lire le BArbre
*****************************************************************************

   chargement et dechargement du BArbre en debut et en fin de session  */

char     LireBArbre( void )
{
    FILE                             *ptFic;
    char                             nomFicBArbre[ _NbMaxLetNomFic ];

    strcpy( nomFicBArbre , NomFicDico );
    strcat( nomFicBArbre , ".BTR" );

    if( !OuvreFichier( nomFicBArbre , &ptFic , _FopenRB ) )
    {
         return _False;
    }

    fseek(ptFic , 0L , 0 );
    fread( (void *)B_ArbInf , (unsigned long)sizeof(struct PorteNormale) ,
                              (unsigned long)TailleBArbInf , ptFic );

    fread( (void *)B_ArbSup , (unsigned long)sizeof(struct PorteSuperieure) ,
                              (unsigned long)TailleBArbSup , ptFic );

    fclose( ptFic );

    return _True;
}

/****************************************************************************
  procedure de debugging pemettant d'obsever le barbre d'acces aux Lemme
****************************************************************************/

void     EcrireBArbre( void )
{
    long                             ind,deb,fin;
    int                              i;


    _Cs();
    for( i = 1 ; i <= ProfondeurBArbre ; i++ )
    {
        sprintf( ChAux , "TPremIndB_Arb[%d] %ld , TNbTermesB_Arb[%d] %ld \n" ,
                         i , TPremIndB_Arb[ i ] , i , TNbTermesB_Arb[ i ] );
        printf( ChAux );
    }
    TxtAttente();

    _Cs();
    printf("Barbre inferieur\n");
    sprintf( ChAux , "\ndepart : (0 - %ld) ->",TailleBArbInf);scanf("%ld",&deb);
    printf( ChAux );
    sprintf( ChAux , "\nfin    : (0 - %ld) ->",TailleBArbInf);scanf("%ld",&fin);
    printf( ChAux );
    sprintf( ChAux , "Barbre inferieur  : Numero , chaine , indiceCar\n");
    printf( ChAux );
    for ( ind = deb ; ind < fin ; ind++ )
    {
        sprintf( ChAux , " %ld (%c,%c,%c,%c) %d\n",ind,
             B_ArbInf[ind].ChaineBAr[0],B_ArbInf[ind].ChaineBAr[1],
             B_ArbInf[ind].ChaineBAr[2],
             B_ArbInf[ind].ChaineBAr[3],B_ArbInf[ind].IndCar);
        printf( ChAux );
    }
    printf("******************************\n");
    TxtAttente();
    _Cs();
    printf("Barbre superieur\n");
    sprintf( ChAux , "\ndepart : (0 - %ld) ->",TailleBArbSup);scanf("%ld",&deb);
    printf( ChAux );
    sprintf( ChAux , "\nfin    : (0 - %ld) ->",TailleBArbSup);scanf("%ld",&fin);
    printf( ChAux );
    sprintf( ChAux , "Barbre superieur  : Numero , chaine , indiceCar\n");
    printf( ChAux );
    for ( ind = deb ; ind <= fin ; ind++ )
    {
        sprintf( ChAux , " %ld (%c,%c,%c,%c) ",ind,
            B_ArbSup[ ind ].ChaineBAr[0],B_ArbSup[ ind ].ChaineBAr[1],
            B_ArbSup[ ind ].ChaineBAr[2],
            B_ArbSup[ ind ].ChaineBAr[3]);
        printf( ChAux );
        sprintf( ChAux , " %ld\n",B_ArbSup[ ind ].PosFicLemme);
        printf( ChAux );
    }
    printf("******************************\n");
    TxtAttente();

}

/***************************************************************************
                        creation du BArbre
****************************************************************************/
char FaireBArbre ( long p_indLemmeDepart , long p_indLemmeFin , 
		   int p_niveauCou ,
                   char *p_chaineDepart , char *p_chaineFin , 
		   int p_indiceCar , 
		   long p_indB_Arb , unsigned long *p_tindLemme )
{
    long           indLocB_Arb;          /* compteusr                    */
    int            nouvIndCar;           /* indice du 1er car discrim pour
                                            le rang sup                  */
    long           reste;
    long           nbTermeTranche;       /* division euclidienne du nombre
                                            de termes pour avoir une repartition
                                            egale entre chaque portes    */
    long           decalageOrigineNiveau;
    long           refAbsCourante;
    long           refAbsSuivante;       /* Lemme limites pour la table precedente
                                            du rang superieur            */

    char          *ptChaine;             /* chaine courante a garnir ( chaine
                                            du barbre )                  */
    ChIdent        chAux;
    ChIdent        premChaineNonDerLemme;
    ChIdent        chaineSuivante;
    ChIdent        chaineCourante;       /* chaines de caracteres des termes
                                            limites pour le rang superieur*/
    char           valRet;

    /* ProfondeurBArbre = Ent (ln(taille lexique) / ln(longueur des tables + 1)) */

    if( p_niveauCou < ProfondeurBArbre 
        && p_indB_Arb + (long)NbPortesBArbre - 1L > TailleBArbInf )
    {
        sprintf( ChAux , "FaireBArbre(albarbre) [%ld]" , 
                         (p_indB_Arb + NbPortesBArbre - 1) );
        TxtErreur( 121 , ChAux );
    }
    else if( p_indB_Arb + (long)NbPortesBArbre - 1L > TailleBArbSup )
    {
        sprintf( ChAux , "FaireBArbre(albarbre) [%ld]" ,
                         (p_indB_Arb + NbPortesBArbre - 1) );
        TxtErreur( 122 , ChAux );
    }

    /* nombre de termes dans chaque tranche du niveau considere du BArbre*/
    nbTermeTranche = ( p_indLemmeFin - p_indLemmeDepart )
                     / (long)(NbPortesBArbre + 1) ;
    reste = ( p_indLemmeFin - p_indLemmeDepart )
                     % (long)(NbPortesBArbre + 1);

    /* (reste) parts de (nbtranche + 1) termes et (10 - reste) parts de
       (nbtranche termes)                                  */
    nbTermeTranche++ ;

    refAbsSuivante = p_indLemmeDepart;
    strcpy( chaineSuivante , p_chaineDepart );

    /* le BArbre est cree par niveaux croissants en remplissant chaque
       niveau                                                         */
    for( indLocB_Arb = 0L ; indLocB_Arb <= (long)NbPortesBArbre ; indLocB_Arb++)
    {
        refAbsCourante = refAbsSuivante;
        strcpy( chaineCourante , chaineSuivante );

        if( p_niveauCou == 1 )
        {
           strcpy( chAux , chaineCourante );
           TxtEcritItemInfo( 1 , chAux );
        }
        if( _ClickDansAnnulerInfo() )
        {
            return _False;
        }

        /* on augmente le nombre de terme par tranches afin de
           repartir le reste de la division euclidienne sur toutes
           les tranches                                           */
        if ( indLocB_Arb == reste )
        {
                nbTermeTranche--;
        }

        if( nbTermeTranche == 0L || indLocB_Arb == (long)NbPortesBArbre
            || refAbsCourante >= p_indLemmeFin - nbTermeTranche)
        {
            /* cas de tranche vide, ou de derniere tranche
                on recopie le terme final                         */
            refAbsSuivante = p_indLemmeFin;
            strcpy( chaineSuivante , p_chaineFin );
        }
        else
        {
            /* cas de tranche non vide, on s'arrete a la fin de
                la tranche suivante                               */
            refAbsSuivante = refAbsCourante + nbTermeTranche;

            /* on cherche un dernier terme n'ayant pas les memes
                chaines entre p_indiceCar et p_indiceCar + _NbMaxLetBArb - 1
                que le premier terme de la tranche                */
            strcpy(  chaineSuivante,
                     BufferLemmesCourant + 
		     _IRelLemme( p_tindLemme[ refAbsSuivante ] ) );
             while( refAbsSuivante < p_indLemmeFin
                    && EgalitePartielle( chaineSuivante ,
                            chaineCourante , p_indiceCar ,
                            p_indiceCar + _NbMaxLetBArb - 1 ) == _True )
            {
                ++refAbsSuivante;
                strcpy( chaineSuivante,
                        BufferLemmesCourant + 
			_IRelLemme( p_tindLemme[ refAbsSuivante ] ) );
            }

            /* puis on inclut tous les termes ayant les memes
                chaines entre p_indiceCar et p_indiceCar + _NbMaxLetBArb - 1
                que le dernier terme de la tranche                */
            while(  !EgaliteChaine( chaineSuivante , DernLemme )
                    && EgalitePartielle( chaineSuivante ,
                          BufferLemmesCourant 
                             + _IRelLemme( p_tindLemme[ refAbsSuivante - 1L ] ) ,
                          p_indiceCar ,
                          p_indiceCar + _NbMaxLetBArb - 1 ) == _True )
            {
                ++refAbsSuivante;
                strcpy( chaineSuivante,
                        BufferLemmesCourant + 
			_IRelLemme( p_tindLemme[ refAbsSuivante ] ) );
            }
        }

        /* ce n'est pas la premiere tranche                        */
        /* on garnit la porte du barbre                            */
        if( indLocB_Arb > 0 )
        {
            if (p_niveauCou < ProfondeurBArbre )
            {
                ptChaine = B_ArbInf[ p_indB_Arb + indLocB_Arb - 1 ].ChaineBAr;
                if( EgaliteChaine( chaineSuivante , DernLemme )
                    && ! EgaliteChaine( chaineCourante , DernLemme ) )
                {
                    /* on recherche le dernier terme de la tranche n'ayant pas
                       pour chaine la chaine du lemme artificiel maximal */
                    while( refAbsSuivante >= p_indLemmeDepart
                           && EgaliteChaine( DernLemme ,
                                  BufferLemmesCourant
                                      + _IRelLemme( p_tindLemme[ refAbsSuivante ] ) ) )
                    {
                        refAbsSuivante--;
                    }
                    strcpy( premChaineNonDerLemme ,
                            BufferLemmesCourant
                                 + _IRelLemme( p_tindLemme[ refAbsSuivante ] ) );
                    nouvIndCar = PremCarDiscrim ( p_indiceCar ,
                                     chaineCourante , premChaineNonDerLemme );
                }
                else
                {
                    nouvIndCar = PremCarDiscrim ( p_indiceCar ,
                                     chaineCourante , chaineSuivante );
                }
                B_ArbInf[ p_indB_Arb + indLocB_Arb - 1 ].IndCar = nouvIndCar;

            }
            else
            {
                ptChaine
                    = B_ArbSup[ p_indB_Arb + indLocB_Arb - 1 ].ChaineBAr;
                B_ArbSup[ p_indB_Arb + indLocB_Arb - 1].PosFicLemme
                    = *(p_tindLemme + refAbsCourante);
            }
            RecopieChaine( chaineCourante , ptChaine ,
                           p_indiceCar ,
                           p_indiceCar + _NbMaxLetBArb - 1 );
        }
        else
            nouvIndCar = p_indiceCar;

        if (p_niveauCou < ProfondeurBArbre )
        {
            /* appel recursif si on n'est pas encore au niveau max du
                   BArbre                                         */
            decalageOrigineNiveau
                  = p_indB_Arb - TPremIndB_Arb[p_niveauCou];
            valRet = FaireBArbre ( refAbsCourante , refAbsSuivante ,
                          p_niveauCou + 1 , chaineCourante , chaineSuivante ,
                          nouvIndCar , TPremIndB_Arb[p_niveauCou + 1]
                             + ( decalageOrigineNiveau + indLocB_Arb
                             + ( decalageOrigineNiveau / (long)NbPortesBArbre ))
                                       * (long)NbPortesBArbre , p_tindLemme );
           if( valRet == _False )
           {
                 return _False;
           }
        }

    }  /* fin de la boucle for */
    return _True;
}

/****************************************************************************
            Lancement de la fabrication du BArbre
***************************************************************************/

char IniBArbre ()
{
    ChIdent                 chaineAnnexe;
    ChIdent                 chAux;
    struct LemmeMem         lemme;

    unsigned long           indBDimo;  /* pointeur dans le fichier des lemmes*/
    unsigned long          *tindLemme; /* tableau contenant l'indice absolu de
                                          debut de
                                          chacune des Lemme du fichier pour 
                                          permettre
                                          la construction du BArbre          */
    unsigned long           longFic;
    long                    indtLemme;  /* indice dans le tableau des 
                                           pointeurs vers les
                                           debuts de Lemme                   */

    long                    tailTindLemme;
    long                    valRet;


    _NouveauDialogueInfo( 128 , _True );
    TxtTitresDialogueInfo( 11 , 12 , 0 );

    /* attention aux mots a racines multiple */
    tailTindLemme = NbLemmesDictionnaire() + 2000L;
    tindLemme = (unsigned long *)_Alloc( (unsigned long)tailTindLemme
                                * (unsigned long)sizeof( unsigned long ) );
    if( !tindLemme )
    {
         TxtErreur( 81 , "IniBArbre(albarbre)" );
         return _False;
    }
    *tindLemme = (indBDimo = (long)( _TailleAmorce + _TailleSeparation));

    if( !FicBDicLemmes )
    {
         TxtErreur( 15 , "IniBArbre(albarbre)" );
         _Free( tindLemme );
         return _False;
    }

    fseek( FicBDicLemmes , 0L , 2 );
    fgetpos( FicBDicLemmes , &longFic );
    indtLemme = 1L;
    while( indBDimo < longFic )
    {
           if( RestructurerBDicLemme( &indBDimo , lemme.ValTerme.Ident  ,
                                      &lemme , chaineAnnexe ) == _Nil 
               || EgaliteChaine( lemme.ValTerme.Ident , DernLemme ) )
           {
                   DemarqueEtRendDag( &(lemme.ValTerme.ArcsFils) );
                   break;
           }
           DemarqueEtRendDag( &(lemme.ValTerme.ArcsFils) );

           if( indtLemme >= tailTindLemme )
           {
                sprintf( ChAux , "\nNb lemmes attendus %ld ind courant %ld\n" ,
                        tailTindLemme , indtLemme );
                printf( ChAux );
                TxtErreur( 16 , "IniBArbre(albarbre)" );
                _Free( tindLemme );
                _FinDialogueInfo();
                return _False;
           }
           *(tindLemme + indtLemme++) = indBDimo;

           if( (indtLemme % 1000L) == 0 )
           {
              strcpy( chAux , lemme.ValTerme.Ident );
              TxtEcritItemInfo( 1 , chAux );
           }
           if( _ClickDansAnnulerInfo() )
           {
               _Free( tindLemme );
               _FinDialogueInfo();
               return _False;
           }
    }
    /* on ne tient pas compte de la derniere valeur qui est
       l'extremite du fichier                                        */
    indtLemme -= 2L;

    TxtTitresDialogueInfo( 13 , 14 , 0 );
    if( FaireBArbre ( 0L , indtLemme , 1 , PremLemme ,
                      DernLemme , 0 , 0L , tindLemme ) == _False )
    {
       _Free( tindLemme );
       _FinDialogueInfo();
       return _False;
    }

    /* on garnit le 1er terme du barbre superieur                    */
    B_ArbSup[ 0 ].PosFicLemme = (unsigned long)(_TailleSeparation 
                                                + _TailleAmorce);

    /*      sauvegardes des initialisations                         */
    TxtTitresDialogueInfo( 15 , 0 , 0 );
    valRet =  SauveBArbre();

    _Free( tindLemme );
    _FinDialogueInfo();

    return (char)valRet;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                        fin de albarbre.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
