/*
 * FASt Term Recognizer
 *
 *  FASTR/alchaine.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for displaying term strings
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*****************************************************************************
                       lecture des chaines de termes
******************************************************************************/
int      LireChainesDesTermes( FILE *p_ptFicTermes , 
                               unsigned long *p_nbChainesTermes ,
                               int p_chaineDeCategories ,
                               TabChaineTermes  p_tabChainesAuxTermes ,
			       TabChar p_chainesTermes ,
			       unsigned long p_tailleChainesTermes )
{
     char                      carLu;
     char                      chaineOk;
     char                      identTerme[ _MaxLetIdent + 1 ];
     TabChaineTermes           ptChaineTerme;

     unsigned long             indChaineCou;
     unsigned long             indChaineIni;

     unsigned long             indChaineTermesCou;
     unsigned long             indChaineTermesAux;
     int                       valcomp;
     char                      trouve;
     int                       ind;

     carLu = getc( p_ptFicTermes );
     indChaineCou = 0L;
     indChaineIni = 0L;

     while( carLu != EOF && carLu != '\n' 
            && (*p_nbChainesTermes) < (unsigned long)NbMaxChainesTermes 
	    && indChaineCou < p_tailleChainesTermes )
     {
         chaineOk = _True;
         /* lecture de l'identificateur du terme */
         for( ind = 0 ; ind < _MaxLetIdent ; ind++ )
         {
             if( !_Alphanum( carLu ) && !(_CarBlanc(carLu))
                 && carLu != '-' && carLu != '.' )
             {
                  chaineOk = _False;
                  break;
             }
             if( _CarBlanc(carLu) )
             {
                  break;
             }
             identTerme[ ind ] = carLu;
             carLu = getc( p_ptFicTermes );
         }
         if( chaineOk )
         {
	   SautBlanc( p_ptFicTermes , &carLu );
	   identTerme[ ind ] = '\0';
         }
         else
         {
	   TxtErreur( 276 , "LireChainesDesTermes(alchaine)" );
	   return _False;
         }

	 indChaineIni = indChaineCou;

         if( p_chaineDeCategories )
         {
             /* lecture de la chaine du terme */
             while( chaineOk && carLu != EOF && !(_CarBlanc(carLu))
                    && carLu != '\n' )
             {
                 carLu = getc( p_ptFicTermes );
             }
             if( _CarBlanc(carLu) )
             {
	       SautBlanc( p_ptFicTermes , &carLu );
             }
             else
             {
	       TxtErreur( 277 , "LireChainesDesTermes(alchaine)" );
	       return _False;
             }
             
             /* lecture de la chaine lemmatisee du terme */
             while( chaineOk && carLu != EOF 
                    && !(_CarBlanc(carLu)) && carLu != '\n' )
             {
	       carLu = getc( p_ptFicTermes );
             }
             if( carLu == '\t' )
             {
	       SautBlanc( p_ptFicTermes , &carLu );
             }
             else
             {
	       TxtErreur( 278 , "LireChainesDesTermes(alchaine)" );
	       return _False;
             }
             
             /* lecture de la chaine du terme */
             while( chaineOk && carLu != EOF && carLu != '\n' 
		    && indChaineCou < p_tailleChainesTermes ) {
                 if( MinusculisationAutomatique )
                 {
                     p_chainesTermes[ indChaineCou ] = _ToLower( carLu );
                 }
                 else
                 {
                     p_chainesTermes[ indChaineCou ] = carLu;
                 }
		 indChaineCou++;
                 carLu = getc( p_ptFicTermes );
             }
             if( chaineOk )
             {
                 if( carLu != EOF && carLu != '\n' )
                 {
                      chaineOk = _False;
                 }
                 p_chainesTermes[ indChaineCou ] = '\0';
		 indChaineCou++;
             }
             else
             {
	       TxtErreur( 279 , "LireChainesDesTermes(alchaine)" );
	       return _False;
             }

             /* lecture jusqu'a la fin de la ligne */
             while( carLu != EOF && carLu != '\n' )
             {
                 carLu = getc( p_ptFicTermes );
             }
             if( carLu == '\n' )
             {
                 carLu = getc( p_ptFicTermes );
             }
         }
         else
         {
             /* lecture de la chaine du terme */
             while( chaineOk && carLu != EOF && carLu != '\n'
		    && carLu != '\t' 
		    && indChaineCou < p_tailleChainesTermes ) {
                 p_chainesTermes[ indChaineCou ] = carLu;
		 indChaineCou++;
                 carLu = getc( p_ptFicTermes );
             }
             if( chaineOk )
             {
                 if( carLu != '\t' && carLu != '\n' )
                 {
                      chaineOk = _False;
                 }
                 p_chainesTermes[ indChaineCou ] = '\0';
		 indChaineCou++;
             }
             else
             {
	       TxtErreur( 280 , "LireChainesDesTermes(alchaine)" );
	       return _False;
             }
             
             /* lecture jusqu'a la fin de la ligne */
             while( carLu != EOF && carLu != '\n' )
             {
                 carLu = getc( p_ptFicTermes );
             }
             if( carLu == '\n' )
             {
                 carLu = getc( p_ptFicTermes );
             }
         }

         if( chaineOk )
         {
             /* recherche du point d'insertion du terme pour */
             /* un classement par ordre croissant */
             trouve = _False;
             ptChaineTerme = p_tabChainesAuxTermes 
                             + (*p_nbChainesTermes);
             for( indChaineTermesCou = (*p_nbChainesTermes) ; 
                  indChaineTermesCou > 0 ; 
                  indChaineTermesCou-- )
             {
                 valcomp =  strcmp( identTerme , (ptChaineTerme - 1)->ident );
                 if( valcomp == 0 )
                 {
                      trouve = _True;
                      break;
                 }
                 if( valcomp > 0 )
                 {
                      break;
                 }
                 ptChaineTerme--;
             }
             if( !trouve )
             {
                 for( indChaineTermesAux = (*p_nbChainesTermes) ; 
                      indChaineTermesAux > indChaineTermesCou ; 
                      indChaineTermesAux-- )
                 {
                     strcpy( (p_tabChainesAuxTermes 
                                 + indChaineTermesAux)->ident ,
                             (p_tabChainesAuxTermes 
                                 + indChaineTermesAux - 1)->ident );
                     (p_tabChainesAuxTermes 
                                 + indChaineTermesAux)->chaine =
                             (p_tabChainesAuxTermes 
                                 + indChaineTermesAux - 1)->chaine;
                 }
                 strcpy( (p_tabChainesAuxTermes 
                                 + indChaineTermesCou)->ident ,
                         identTerme );
                 (p_tabChainesAuxTermes 
                                 + indChaineTermesCou)->chaine =
                         p_chainesTermes + indChaineIni;
		 /* printf("[%s] [%s]\n",identTerme,p_chainesTermes + 
		    indChaineIni); */
                 (*p_nbChainesTermes)++;
		 if( *p_nbChainesTermes >= (unsigned long)NbMaxChainesTermes 
		                           - 1 
		     || indChaineCou >= p_tailleChainesTermes ) {
		   TxtErreur( 153 , "LireChainesDesTermes(alchaine)" );
		   return _False;
		 }
                 if( (*p_nbChainesTermes) % 100L == 0 )
                 {
                     sprintf( ChAux , "%ld" , (*p_nbChainesTermes) );
                     TxtEcritItemInfo( 1 , ChAux );
                 }
             }
         }
     }
     if( carLu != EOF && carLu != '\n' )
     {
          TxtErreur( 281 , "LireChainesDesTermes(alchaines)" );
     }
     fclose( p_ptFicTermes );
     return _True;
}

/*****************************************************************************
      remplacement d'une chaine par un identificateur ou l'inverse 
******************************************************************************/
void     LireIdentOuChaineEtRemplacer( FILE *p_ptFicOccSource , 
                               FILE *p_ptFicOccCible ,
                               char *p_carLu , int p_identVersChaine ,
                               unsigned long p_nbChainesTermes ,
                               TabChaineTermes  p_tabChainesAuxTermes )
{
     char                      chaineOk;
     char                      identTerme[ _MaxLetIdent + 1 ];
     char                      chaineTerme[ _NbCarChaine + 1 ];
     TabChaineTermes           ptChaineTerme;
     unsigned long             indChaineTermesCou;
     char                      trouve;
     int                       ind;

     chaineOk = _True;
     if( p_identVersChaine )
     {
         /* lecture de l'identificateur du terme */
         for( ind = 0 ; ind < _MaxLetIdent && chaineOk ; ind++ )
         {
             if( !_Alphanum( (*p_carLu) )  && (*p_carLu) != '\t'
                 && (*p_carLu) != '.' && (*p_carLu) != '\n' )
             {
                  /* chaine incorrecte (c'est une phrase du texte,
                     on la recopie telle quelle) */
                  chaineOk = _False;
                  identTerme[ ind ] = '\0';
                  break;
             }
             if( (*p_carLu) == '\t' || (*p_carLu) == '\n' )
             {
                  break;
             }
             if( MinusculisationAutomatique )
             {
                 identTerme[ ind ] = _ToLower( (*p_carLu) );
             }
             else
             {
                 identTerme[ ind ] = (*p_carLu);
             }
             (*p_carLu) = getc( p_ptFicOccSource );
         }
         identTerme[ ind ] = '\0';
         /* lecture de l'identificateur du terme */

         if( chaineOk && ((*p_carLu) != '\t' && (*p_carLu) != '\n') )
         {
             chaineOk = _False;
         }
     }
     else
     {
         /* lecture de la chaine du terme */
         for( ind = 0 ; ind < _NbCarChaine && chaineOk ; ind++ )
         {
             if( !_Alphanum( (*p_carLu) ) && (*p_carLu) != ' ' 
                 && (*p_carLu) != '\t' && (*p_carLu) != '\n' )
             {
                  /* chaine incorrecte (c'est une phrase du texte,
                     on la recopie telle quelle) */
                  chaineOk = _False;
                  chaineTerme[ ind ] = '\0';
                  break;
             }
             if( (*p_carLu) == '\t' || (*p_carLu) == '\n' )
             {
                  break;
             }
             chaineTerme[ ind ] = (*p_carLu);
             (*p_carLu) = getc( p_ptFicOccSource );
         }
         chaineTerme[ ind ] = '\0';
         /* lecture de la chaine du terme */

         if( chaineOk && ((*p_carLu) != '\t') && ((*p_carLu) != '\n') )
         {
             chaineOk = _False;
         }
     }

     if( !chaineOk )
     {
         /* chaine incorrecte (on recopie la chaine telle quelle) */
         if( p_identVersChaine )
         {
             fprintf( p_ptFicOccCible , "%s" , identTerme );
         }
         else
         {
             fprintf( p_ptFicOccCible , "%s" , chaineTerme );
         }
     }
     else
     {
         /* chaine correcte recherche de l'ident ou de la chaine du terme */
         /* recherche du terme pour retrouver sa chaine */
         trouve = _False;
         ptChaineTerme = p_tabChainesAuxTermes;
         for( indChaineTermesCou = 0 ; 
              indChaineTermesCou < p_nbChainesTermes ; 
              indChaineTermesCou++ )
         {
             if( p_identVersChaine )
             {
                 if( strcmp( identTerme , ptChaineTerme->ident ) == 0 )
                 {
                     trouve = _True;
                     break;
                 }
             }
             else
             {
                 if( strcmp( chaineTerme , ptChaineTerme->chaine ) == 0 )
                 {
                     trouve = _True;
                     break;
                 }
             }
             ptChaineTerme++;
         }
         if( !trouve )
         {
             if( !TraitementBatch )
             {
                if( p_identVersChaine )
                {
                    sprintf( ChAux , 
                             "[%s] LireChainesDesTermes(alchaines)" , 
                             identTerme );
                    TxtErreur( 245 , ChAux );
                }
                else
                {
                    sprintf( ChAux , 
                             "[%s] LireChainesDesTermes(alchaines)" , 
                             chaineTerme );
                    TxtErreur( 246 , "LireChainesDesTermes(alchaines)" );
                }
             }
             if( p_identVersChaine )
             {
                 fprintf( p_ptFicOccCible , "%s" , identTerme );
             }
             else
             {
                 fprintf( p_ptFicOccCible , "%s" , identTerme );
             }
         }
         else
         {
             if( p_identVersChaine )
             {

                 /* les termes initiaux sont distingues par une
                    majuscule en tete */
                 if( *(ptChaineTerme->ident) == '0' )
                 {
                     fprintf( p_ptFicOccCible , "%s", 
                              (ptChaineTerme->chaine) );
                 }
                 else
                 {
                     fprintf( p_ptFicOccCible , "%s", 
                              ptChaineTerme->chaine );
                 }
             }
             else
             {
                 fprintf( p_ptFicOccCible , "%s", 
                          ptChaineTerme->ident );
             }
         }
     }
}

/****************************************************************************
        remplace les identificateurs de termes par leur chaine
*****************************************************************************/
void     ChainesTermes( char *p_nomFicOccSource , char *p_nomFicTermes , 
                        char *p_nomFicOccCible , int p_noCol ,
                        int p_identVersChaine , int p_ficTermesMultiples ,
                        int p_chaineDeCategories )
{
     char                    carLu;
     FILE                   *ptFicOccSource;
     FILE                   *ptFicOccCible;
     FILE                   *ptFicChaineTerme;

     unsigned long           tailletTabChainesAuxTermes;
     TabChaineTermes         tabChainesAuxTermes;

     unsigned long           tailleChainesTermes;
     TabChar                 chainesTermes;

     unsigned long           nbChainesTermes;
     unsigned long           nbOccurencesTermes;
     int                     ind;
     int                     noEtape;
     char                    chaineOk;

     if( strcmp( p_nomFicOccSource , "stdin" ) == 0 )
     {
         ptFicOccSource = stdin;
     }
     else
     {
        if( !OuvreFichier( p_nomFicOccSource , &ptFicOccSource , "r" ) )
        {
            return;
        }
     }

     if( strcmp( p_nomFicOccCible , "stdout" ) == 0 )
     {
         ptFicOccCible = stdout;
     }
     else
     {
        if( !OuvreFichier( p_nomFicOccCible , &ptFicOccCible , "w" ) )
        {
            return;
        }
     }

     /* la taille des chaines est allouee dynamiquement */
     tailletTabChainesAuxTermes = (unsigned long) sizeof( struct ChaineTerme )
                              * (unsigned long) NbMaxChainesTermes;
     tabChainesAuxTermes 
            = (TabChaineTermes)_Alloc( tailletTabChainesAuxTermes );
     tailleChainesTermes = (unsigned long)_NbCarChaine
                              * (unsigned long) NbMaxChainesTermes;
     chainesTermes 
            = (TabChar)_Alloc( tailleChainesTermes );
     if( !tabChainesAuxTermes || !chainesTermes )
     {
          TxtErreur( 81 , "LireChainesDesTermes(alchaines)" );
          return;
     }

     /*************************************************
                   lecture des chaines de termes
     **************************************************/
     nbChainesTermes = 0L;
     if( !p_ficTermesMultiples )
     {
         TxtTitresDialogueInfo( 38 , 39 , 0 );
         if( !OuvreFichier( p_nomFicTermes , &ptFicChaineTerme , "r" ) )
         {
             return;
         }

         if( !LireChainesDesTermes( ptFicChaineTerme , &nbChainesTermes ,
				    p_chaineDeCategories ,
				    tabChainesAuxTermes ,
				    chainesTermes , 
				    tailleChainesTermes ) ) {
	   return;
	 }
     }
     else
     {
         TxtTitresDialogueInfo( 38 , 39 , 45 );

         for( noEtape = 1 ; noEtape < 100 ; noEtape++ )
         {
             sprintf( ChAux , "%s%d" , p_nomFicTermes , noEtape );
             ptFicChaineTerme = fopen( ChAux  , _FopenRT );
             if( !ptFicChaineTerme )
             {
                break;
             }
             TxtEcritItemInfo( 2 , ChAux );
             if( !LireChainesDesTermes( ptFicChaineTerme , &nbChainesTermes ,
					p_chaineDeCategories ,
					tabChainesAuxTermes ,
					chainesTermes , 
					tailleChainesTermes  ) ) {
	       fclose( ptFicChaineTerme );
	       return;
	     }
             fclose( ptFicChaineTerme );
         }
     }
     
     /*************************************************
         Ecriture des ou des ident chaines de termes
     **************************************************/
     TxtTitresDialogueInfo( 40 , 39 , 0 );

     /* remplacement des identificateurs de termes par leur chaine */
     nbOccurencesTermes = 0L;
     if( p_noCol != 0 )
     {
         carLu = getc( ptFicOccSource );
         while( carLu != EOF )
         {
             chaineOk = _True;
             ind = 1;
             while( ind < p_noCol )
             {
                 while( carLu != EOF && carLu != '\t' && carLu != '\n' )
                 {
                     fputc( carLu , ptFicOccCible );
                     carLu = getc( ptFicOccSource );
                 }
                 if( carLu =='\t' )
                 {
                     fputc( carLu , ptFicOccCible );
                     carLu = getc( ptFicOccSource );
                 }
                 else
                 {
                     chaineOk = _False;
                     break;
                 }
                 ind++;
             }

             if( chaineOk )
             {
                 LireIdentOuChaineEtRemplacer(
                        ptFicOccSource , ptFicOccCible , 
                        &carLu , p_identVersChaine , 
                        nbChainesTermes , 
			tabChainesAuxTermes );
             }

             /* lecture jusqu'a la fin de la ligne */
             while( carLu != EOF && carLu != '\n' )
             {
                 fputc( carLu , ptFicOccCible );
                 carLu = getc( ptFicOccSource );
             }
             if( carLu == '\n' )
             {
                 fputc( carLu , ptFicOccCible );
                 carLu = getc( ptFicOccSource );
             }

             nbOccurencesTermes++;
             if( nbOccurencesTermes % 100L == 0 )
             {
                 sprintf( ChAux , "%ld" , nbOccurencesTermes );
                 TxtEcritItemInfo( 1 , ChAux );
             }
         }
     }
     else
     {

         carLu = getc( ptFicOccSource );
         while( carLu != EOF )
         {
             if( !p_identVersChaine || _Num( carLu ) )
             {
                 LireIdentOuChaineEtRemplacer( 
                         ptFicOccSource , ptFicOccCible , 
                         &carLu , p_identVersChaine , 
                         nbChainesTermes , 
			 tabChainesAuxTermes );
             }
             else
             {
                 while( carLu != '\n' && carLu != '\t' )
                 {
                      fputc( carLu , ptFicOccCible );
                      carLu = getc( ptFicOccSource );
                 }
             }

             fputc( carLu , ptFicOccCible );
             carLu = getc( ptFicOccSource );

             nbOccurencesTermes++;
             if( nbOccurencesTermes % 100L == 0 )
             {
                 sprintf( ChAux , "%ld" , nbOccurencesTermes );
                 TxtEcritItemInfo( 1 , ChAux );
             }
         }
     }
 
     fclose( ptFicOccSource );
     fclose( ptFicOccCible );
     _Free( tabChainesAuxTermes );
     _Free( chainesTermes );
}


/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                        fin de alchaines.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
