/*
 * FASt Term Recognizer
 *
 *  FASTR/alcompil.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for rule compiling
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*****************************************************************************
           ANALYSE DU TERME DE TETE D'UNE REGLE DE MOT SIMPLE
*****************************************************************************/
char    CompileLemme( TabLemmeMem p_ptLemme , char *p_carLu , FILE *p_ptFic  )
{
    TabDagMem            operLogFils;
    struct ValAtomeMem   atome;

    if( !ReconnaitreChaine ( p_ptFic , "Word" , p_carLu ) )
    {
          return _False;
    }

    /* lecture de la racine principale */
    if( !LireChaineAvecBalise( p_ptFic , p_carLu , p_ptLemme->ValTerme.Ident ,
			       _MaxLetIdent , (int)'\'' , _True ) 
        || !ReconnaitreCaractere( p_ptFic , (int)':' , p_carLu , _True ) )
    {
        return _False;
    }

    /* lecture des contraintes sur les dags */
    operLogFils = NULL;
    if( !LireContraintesTerme( p_carLu , p_ptFic , &operLogFils , 
                               &(p_ptLemme->ValTerme.ArcsFils) ) )
    {
        return _False;
    }

    /* unification des contraintes sur les dags */
    if( !UnifierContraintes( &operLogFils , &(p_ptLemme->ValTerme.ArcsFils) , 
                             NULL , NULL ) )
    {
        return _False;
    }
    
    /* Transformation de la chaine du mot en un trait lemme dans le
       cas ou le lemme n'est pas pecifie par la regle (cas des mots
       flechis) */
    if( !ExisteCheminFils( CheminLemme ,
                           &(p_ptLemme->ValTerme.ArcsFils) ) )
    {
        atome.AtomeConjDisj = _False; 
        strcpy( atome.ValAtome.Chaine , p_ptLemme->ValTerme.Ident );
        if( !AjouteCheminEtAtomeFils( &(p_ptLemme->ValTerme.ArcsFils) ,
                                      CheminLemme , _AtomeChaine , 
				      atome , _True ) )
        {
            return _False;
        }
    }

    if( !ReconnaitreCaractere( p_ptFic , (int)'.' , p_carLu , _True ) )
    {
          return _False;
    }

    return _True;
}

/*****************************************************************************
                        ANALYSE D'UNE REGLE
*****************************************************************************/
char    CompileRegle( TabRegleMem *p_ptRegle , char *p_carLu , FILE *p_ptFic ,
                       struct ValTerme **p_ptValTermeDeLexicalisation , 
                       int p_estRegleMemoire )
{
    TabDagMem            operLogFils;
    TabDagMem           *hdlFilsLexicalisation;
    ChIdent              chaineTermeLexicalisation;
    int                  indTerme;

    if( !ReconnaitreChaine ( p_ptFic , "Rule" , p_carLu ) )
    {
        return _False;
    }

    /**************************************************************
         allocation (regle memoire : recherche d'une case libre)
     **************************************************************/
    if( p_estRegleMemoire )
    {
        /* on prend la regle memoire suivante dans le tableau qui est 
           disponible   et on la chaine avec les autres          */
        if( NbReglesMem >= NbMaxReglesMem )
        {
            TxtErreur( 30 , "CompileRegle(alcompil)" );
            return _False;
        }

        (*p_ptRegle) = ReglesMem + NbReglesMem;
        MiseAZeroUneRegle( (*p_ptRegle) );

        if( NbReglesMem > 0 )
        {
            (ReglesMem[ NbReglesMem - 1 ]).RegleFrere = (*p_ptRegle);
        }
        (*p_ptRegle)->RegleFrere = NULL;
        NbReglesMem++;
    }
    else
    {
        (*p_ptRegle) = NouvelleRegle();
    }

    /**************************************************************
                   lecture de la structure de la regle
     **************************************************************/
    NbLiensIdentTerme = 0;
    if( !((*p_ptRegle)->TermeFils 
           = CompileListeOuConcat( p_carLu , p_ptFic , NULL , _False )) )
    {
         return _False;
    }
    if( !ReconnaitreCaractere( p_ptFic , (int)':' , p_carLu , _True ) )
    {
        return _False;
    }

    /**************************************************************
                   lecture et unification des contraintes
     **************************************************************/
    operLogFils = NULL;
    /* lecture des contraintes sur les dags */
    if( !LireContraintesRegle( p_carLu , p_ptFic , &operLogFils ) )
    {
        return _False;
    }
    if( !ReconnaitreCaractere( p_ptFic , (int)'.' , p_carLu , _True ) )
    {
          return _False;
    }

    /* unification des contraintes sur les dags */
    if( !UnifierContraintes( &operLogFils , NULL , 
                             &((*p_ptRegle)->TermeFils) , 
                             NULL ) )
    {
        TxtChaineErreur( 188 , ChErreurCompil );
        return _False;
    }

    /**************************************************************
      on remplace terminaux et non terminaux par leurs references
     **************************************************************/
    /* mots inconnus */
    if( !GarnirReferenceRegle( (*p_ptRegle) , p_estRegleMemoire ) )
    {
         return _False;
    }

    /**************************************************************
          rattachement : recherche du terme correspondant eventuel
     **************************************************************/
    *p_ptValTermeDeLexicalisation = NULL;
    hdlFilsLexicalisation
             = ExisteCheminFils( CheminLexicalisation ,
                            &((*p_ptRegle)->TermeFils->ValTerme.ArcsFils) );
 
    if( hdlFilsLexicalisation && (*hdlFilsLexicalisation)
        && (*hdlFilsLexicalisation)->TypeDag == _AtomeChaine )
    {
        strcpy( chaineTermeLexicalisation , 
                (*hdlFilsLexicalisation)->ValDag.Atome.ValAtome.Chaine );

        /* recherche du terme de la regle ayant cet identificateur */
        for( indTerme = 0 ; indTerme < NbLiensIdentTerme ; indTerme++ )
        {
            if( strcmp( LienIdentTerme[ indTerme ].Ident ,
                        chaineTermeLexicalisation ) == 0 )
            {
                /* on releve le pteur vers le dag de ce terme */
                *p_ptValTermeDeLexicalisation 
                         = LienIdentTerme[ indTerme ].PtValTerme;
                break;
            }
        }

        if( !(*p_ptValTermeDeLexicalisation) )
        {
             TxtChaineErreur( 163 , ChAux );
             sprintf( ChErreurCompil , "%s : [%s]" , ChAux , 
                                       chaineTermeLexicalisation );
             return _False;
        }
    }

    /**************************************************************
          suppression des chemins de lemme et d'inflection 
                  en cas de dictionnaire reduit
     **************************************************************/
    if( BDicReduit )
    {
        if( !SupprimeEtiquetteLemmeEtFlexionDF( (*p_ptRegle) ) )
        {
            return _False;
        }
    }
    return _True;
}

/*****************************************************************************
            ANALYSE D'UNE METAREGLE
*****************************************************************************/
char    CompileRegleMeta( TabRegleMemMeta p_ptRegleMeta , char *p_carLu , 
                          FILE *p_ptFic )
{
    TabDagMem          operLogFils;
    TabTermeMem        ptFils;

    operLogFils = NULL;
    NbLiensIdentTerme = 0;
    MiseAZeroUnRegleMeta( p_ptRegleMeta );

    if( !ReconnaitreChaine ( p_ptFic , "Metarule" , p_carLu ) )
    {
        return _False;
    }

    /**************************************************************
                   lecture de la structure de la meta-regle
     **************************************************************/
    if( !(ptFils = CompileTerminalOuNonTerminal( p_carLu , p_ptFic , 
                                                 NULL , _False , _False ,
						 NULL ) ) )
    {
        return _False;
    }
    p_ptRegleMeta->TermeFils = ptFils;

    if( *p_carLu == '*' ) {
      ReconnaitreCaractere( p_ptFic , (int)'*' , p_carLu , _True ) ;
      p_ptRegleMeta->AvecUnificationExterne = _True;
    }

    if( !ReconnaitreCaractere( p_ptFic , (int)'(' , p_carLu , _True ) )
    {
        return _False;
    }

    if( !(ptFils->TermeFils 
          = CompileListeOuConcat( p_carLu , p_ptFic , NULL , _False  )) )
    {
         return _False;
    }

    if( !ReconnaitreCaractere( p_ptFic , (int)')' , p_carLu , _True )
        || !ReconnaitreCaractere( p_ptFic , (int)'=' , p_carLu , _True ) )
    {
        return _False;
    }

    if( !(ptFils->TermeFrere 
          = CompileListeOuConcat( p_carLu , p_ptFic , &operLogFils , _True )) )
    {
         return _False;
    }

    if( !ReconnaitreCaractere( p_ptFic , (int)':' , p_carLu , _True ) )
    {
        return _False;
    }

    /**************************************************************
                   lecture et unification des contraintes
     **************************************************************/
    /* lecture des contraintes sur les dags */
    if( !LireContraintesRegle( p_carLu , p_ptFic , &operLogFils ) )
    {
        return _False;
    }

    /* unification des contraintes sur les dags */
    if( !UnifierContraintes( &operLogFils , NULL , 
                             &(p_ptRegleMeta->TermeFils) , NULL ) )
    {
        TxtChaineErreur( 189 , ChErreurCompil );
        return _False;
    } 

    if( !ReconnaitreCaractere( p_ptFic , (int)'.' , p_carLu , _True ) )
    {
          return _False;
    }

    return _True;
}

/*****************************************************************************
      ANALYSE D'UN IDENTIFICATEUR DE TERME OU DE TETE DE REGLE
*****************************************************************************/
char    CompileIdentificateurCategorie( struct ValTerme *p_ptValTerme , 
                           TabDagMem *p_hdlOperLog ,
                           int p_extraireCategorie )
{
    int                 i;
    ChIdent             chCat;
    char               *ch;
    int                 catLex;
    TabDagMem           filsOperLog;
    TabDagMem           ptAtome;
    int                 indTerme;
    char                trouve;

    /* lien entre l'identificateur et le terme */
    trouve = _False;
    for( indTerme = 0 ; indTerme < NbLiensIdentTerme ; indTerme++ )
    {
         if( strcmp( LienIdentTerme[ indTerme ].Ident , 
                     p_ptValTerme->Ident ) == 0 )
         {
             /* le pointeur p_hdlOperLog est nul dans le cas d'une
                regle ou d'une partie gauche de metaregle, il ne peut
                y avoir repetition d'un identificateur dans ce cas */
             if( !p_hdlOperLog )
             {
                 TxtErreur( 105 , "CompileIdentificateurCategorie(alcompil)" );
                 return _False;
             }
             else
             /* dans le cas d'une partie droite de metaregle, il est
                necessaire de noter les egalites de termes entre
                partie droite et partie gauche */
             {
                 if( !AjoutOperFils( p_hdlOperLog , '=' , 
                     _Arc , (TabChar)(&(LienIdentTerme[ indTerme ]
                                        .PtValTerme->ArcsFils)) , 
                     _Arc , (TabChar)(&(p_ptValTerme->ArcsFils)) ) )
                 {
                      return _False;
                 }
                 trouve = _True;
             }
         }
    }

    if( !trouve )
    {
        LienIdentTerme[ NbLiensIdentTerme ].PtValTerme = p_ptValTerme;
        strcpy( LienIdentTerme[ NbLiensIdentTerme ].Ident , 
                p_ptValTerme->Ident );
        NbLiensIdentTerme++;
        if( NbLiensIdentTerme >= NbMaxTermesDansUneRegle )
        {
             TxtErreur( 104 , "CompileIdentificateurCategorie(alcompil)" );
             return _False;
        }
    }

    if( p_extraireCategorie )
    {
        /* extraction de la categorie lexicale de ce terme */
        i = 0;
        ch = p_ptValTerme->Ident;
        while( *(ch + i) && _Alpha(*(ch + i)) )
        {
            *(chCat + i) = *(ch + i);
            i++;
        }
        *(chCat + i) = '\0';
        if( !(strcmp( chCat , "X" ) == 0 || *chCat == '\0') )
        {
             if( !IdentATraitsCatLex( chCat , _CatLex , &catLex ) )
             {
                 return _False;
             }

             filsOperLog = NULL;
             if( !(ptAtome = NouveauDag( _AtomeEntier )) )
             {
                 return _False;
             }
             ptAtome->ValDag.Atome.ValAtome.Ref = (long)catLex;
             ptAtome->TypeDag = _AtomeRef;
             
             if( !AjoutOperCheminAtome( &filsOperLog ,'=' , 
                                        &(p_ptValTerme->ArcsFils) ,
                                        CheminCategorie , ptAtome , _Arc ) )
             {
                 return _False;
             }
             if( !UnifierContraintes( &filsOperLog , 
                                      &(p_ptValTerme->ArcsFils) , 
                                      NULL , NULL ) )
             {
                 return _False;
             }
        }
    }

    return _True;
}

/*****************************************************************************
         ANALYSE D'UNE LISTE OU D'UNE CONC ENTRE ()
             EST APPELE APRES RENCONTRE D'UNE (
*****************************************************************************/
TabTermeMem CompileListeOuConcat( char *p_carLu , FILE *p_ptFic , 
                                  TabDagMem *p_hdlOperLog , 
                                  int p_ERAutorisee )
{
    TabTermeMem             ptTerme = NULL;
    TabTermeMem             racine = NULL;
    TabTermeMem             ptTermePrec = NULL;

    if( !(racine 
          = CompileTerminalOuNonTerminal( p_carLu , p_ptFic , 
					  p_hdlOperLog , _True , 
					  p_ERAutorisee , NULL ) ) )
    {
        return NULL;
    }
    
    if( *p_carLu == '=' )
    {
         ReconnaitreCaractere( p_ptFic , (int)'=' , p_carLu , _True );
         racine->ValTerme.TypeTerme = _Liste;
    }
    else
    {
         if( !ReconnaitreChaine ( p_ptFic , "->" , p_carLu ) )
         {
             return NULL;
         }
         racine->ValTerme.TypeTerme = _Concatenation;
    }

    ptTermePrec = NULL;
    while( *p_carLu != ')' && *p_carLu != '.' 
           && *p_carLu != ':' && *p_carLu != EOF )
    {
        if( racine->ValTerme.TypeTerme == _Liste && ptTermePrec )
        {
             if( !ReconnaitreCaractere( p_ptFic , (int)'+' , 
                                        p_carLu , _True ) )
             {
                  return NULL;
             }
        }

	if( !( ptTerme 
	       = CompileTerme( p_carLu , p_ptFic , p_hdlOperLog , 
			       p_ERAutorisee , ptTermePrec ) ) ) {
	  return NULL;
	}

	if( ptTermePrec ) {
	  ptTermePrec->TermeFrere = ptTerme;
	  ptTerme->TermeFrere = NULL;
        }
        else {
	  racine->TermeFils = ptTerme;
        }
	ptTermePrec = ptTerme;
    }

    return racine;
}

/*****************************************************************************
                        ANALYSE D'UN TERME
*****************************************************************************/
TabTermeMem CompileTerme( char *p_carLu , FILE *p_ptFic , 
			  TabDagMem *p_hdlOperLog ,
			  int p_ERAutorisee , 
			  TabTermeMem p_ptTermePrec )
{
    TabTermeMem                  valRet;

    switch( *p_carLu )
    {
    case '(' : ReconnaitreCaractere( p_ptFic , (int)'(' , p_carLu , _True );
               valRet = CompileListeOuConcat( p_carLu , p_ptFic ,
                                              p_hdlOperLog , p_ERAutorisee );
	       if( !valRet ) {
		 return NULL;
	       }
               if( !ReconnaitreCaractere( p_ptFic , (int)')' , 
                                          p_carLu , _True ) ) {
		 return NULL;
               }
               break;
    default : if( !( _Alphanum(*p_carLu) || *p_carLu == '['
                     || *p_carLu == '<' || *p_carLu == '{' ) )
              {
                  TxtChaineErreur( 162 , ChErreurCompil );
                  return NULL;
              }
              else
              {
                  valRet 
                    = CompileTerminalOuNonTerminal( p_carLu , p_ptFic , 
						    p_hdlOperLog , _True , 
						    p_ERAutorisee ,
						    p_ptTermePrec );
              }
              break;
    }

    return valRet;
}

/*****************************************************************************
                        ANALYSE D'UN TERME SIMPLE
*****************************************************************************/
TabTermeMem CompileTerminalOuNonTerminal( char *p_carLu , FILE *p_ptFic , 
                           TabDagMem *p_hdlOperLog , 
                           int p_extraireCategorie , 
                           int p_ERAutorisee ,
                           TabTermeMem p_ptTermePrec )
{
    TabTermeMem             ptTerme = NULL;
    struct DagMem           premDagER;
    TabDagMem              *hdlArcFin;

    /* terme avec ER */
    if( p_ERAutorisee 
        && ( *p_carLu == '[' || *p_carLu == '<' || *p_carLu == '{' ) ) {
      MiseAZeroDag( &premDagER , _Arc );
      if( !CompileER( p_carLu , p_ptFic , &premDagER ) ) {
	/* &(ptTerme->ValTerme.ArcsFils) */
	DemarqueEtRendTerme( &ptTerme );
	return NULL;
      }

      /* ER finale non suivie d'un terme */
      if( *p_carLu == ':' ) {
	if( !p_ptTermePrec ) {
	  TxtChaineErreur( 326 , ChErreurCompil );
	  return NULL;
	}
	  
	if( !(hdlArcFin = CopieChemin( &(p_ptTermePrec->ValTerme.ArcsFils) , 
				       CheminTrailingExprReg ) ) ) {
	  return NULL;
	}
	*(hdlArcFin) = premDagER.Fils;

	return p_ptTermePrec;
      }
      /* ER suivie d'un terme */
      else if( _Alpha( *p_carLu ) ) {
	ptTerme = NouveauTerme();
	ptTerme->ValTerme.TypeTerme = _Feuille;
	if( !(hdlArcFin = CopieChemin( &(ptTerme->ValTerme.ArcsFils) , 
				       CheminLeadingExprReg ) ) ) {
	  return NULL;
	}
	*(hdlArcFin) = premDagER.Fils;

	LireIdentificateur( p_ptFic , p_carLu , ptTerme->ValTerme.Ident , 
			   _MaxLetIdent , _True );
      }
      else {
	 TxtChaineErreur( 282 , ChErreurCompil );
	 return NULL;
      }
    }
    /* terme avec ER */
    /* terme sans ER */
    else {
      ptTerme = NouveauTerme();
      ptTerme->ValTerme.TypeTerme = _Feuille;

      LireIdentificateur( p_ptFic , p_carLu , ptTerme->ValTerme.Ident , 
			 _MaxLetIdent , _True );
    }
    /* terme sans ER */

    if( !CompileIdentificateurCategorie( &(ptTerme->ValTerme) ,
                                         p_hdlOperLog , p_extraireCategorie ) )
    {
         DemarqueEtRendTerme( &ptTerme );
         return NULL;
    }

    return ptTerme;
}

/*****************************************************************************
            Ecriture d'une regle erronnee
*****************************************************************************/
void    MessageErreurCompil( long p_posFicInitial , char *p_carLu , 
                           FILE *p_ptFic )
{
   long              tailleRegle;
   long              indBuff;
   unsigned long     posFicFinal;
   long              tailleLue;

   
   /* erreur de compilation -> arret */
   if( p_ptFic ) {
     /* on met 2 !! sur l'erreur */
     
     /* on revient eventuellemnt d'un pas en arriere sur le fichier 
	pour recuperer le caractere initial */
     if( p_posFicInitial > 0L )
       {
	 p_posFicInitial--;
       }
     fgetpos( p_ptFic , &posFicFinal );
     tailleRegle = (long)( posFicFinal - p_posFicInitial );
     if( tailleRegle > 256L )
       {
	 tailleRegle = 256L;
	 p_posFicInitial = posFicFinal - 256L;
       }
     
     /* on relit la regle en la stockant dans le buffer
	en s'arretant a un ; ou a la fin du fichier             */
     fseek( p_ptFic , p_posFicInitial , 0 );
     for( indBuff = 0L ; indBuff < TailleBuffStd / 2 ; indBuff++ )
       {
	 BuffTexte[ indBuff ] = getc( p_ptFic );
	 if( BuffTexte[ indBuff ] == EOF || BuffTexte[ indBuff ] == '.' )
	   {
	     break;
	   }
       }
     tailleLue = indBuff + 1L;
     BuffTexte[ tailleLue ] = '\0';
     
     /* on insere 2 !! sur l'erreur si elle est dans le domaine lu */
     for( indBuff = tailleLue + 2L ;
	  indBuff >= tailleRegle && indBuff >= 2L ;
	  indBuff-- )
       {
	 BuffTexte[ indBuff ] = BuffTexte[ indBuff - 2L ];
       }
     BuffTexte[ indBuff ] = '!';
     BuffTexte[ indBuff - 1L ] = '!';
   }
   else {
     BuffTexte[ 0 ] = '\0';
   }
   
   /* securite sur les fins de chaines */
   BuffTexte[ TailleBuffStd - 1L ] = '\0';
   ChErreurCompil[ 512L - 1L ] = '\0';
   TxtErreurCompil( BuffTexte , ChErreurCompil );
   if( ArretSurErreurCompil )
     {
       sprintf( ChAux , "\n%s\n%s\n" , ChErreurCompil , BuffTexte );
       printf( ChAux );
       TxtAttente();
     }
   fprintf( FicLog , "\n%s\n%s\n\n" , ChErreurCompil ,
	    BuffTexte );
   
   /* on passe a la regle suivante */
   /* en revenant en arriere jusqu'a la fin de la regle fausse ; */
   if( p_ptFic ) {
     fseek( p_ptFic , p_posFicInitial + tailleLue  - 1L , 0 );
     *p_carLu = getc( p_ptFic);
     if( *p_carLu == '.' )
       {
	 ReconnaitreCaractere( p_ptFic , (int)'.' , p_carLu , _True );
       }
   }
}

/**************************************************************************
                        Lecture des regles grammaticales
***************************************************************************/
char    CompilationRegles( char *p_nomFic , FILE *p_ptFic , char *p_carLu ,
                           int p_estInsertionFichier , 
                           int p_estVerifSyntaxe , 
                           int p_estRegleMemoire , int *p_nbReglesFausses ,
                           long *p_nbReglesMots , long *p_nbReglesTermes )
{
    char                       carLu;
    TabRegleMem                ptRegle;
    TabLemmeMem                ptLemme;
    struct ValTerme           *ptValTermeLexicalisation;
    unsigned long              posFic;
    char                       chRegleMemoire[ _NbMaxLetChOut ];
    char                       chLog[ _NbMaxLetChOut ];

    (*p_nbReglesMots) = 0L;
    (*p_nbReglesTermes) = 0L;

    _NouveauDialogueInfo( 128 , _True );
    TxtTitresDialogueInfo( 27 , 28 , 29 );
    sprintf( ChAux , "%ld" , (*p_nbReglesMots) );
    TxtEcritItemInfo( 1 , ChAux );
    sprintf( ChAux , "%ld" , (*p_nbReglesTermes) );
    TxtEcritItemInfo( 2 , ChAux );

    if( !p_ptFic )
    {
        if( !OuvreFichier( p_nomFic , &PtFicRegles , _FopenRT ) )
        {
             return _False;
        }
        carLu = getc( PtFicRegles );
        SautBlanc( PtFicRegles , &carLu );
    }
    else
    {
        PtFicRegles = p_ptFic;
        carLu = *p_carLu;
    }

    if( p_estRegleMemoire && NbReglesMem > 0 )
    {
       if( !RAZReglesMemoire() )
       {
           TxtErreur( 29 , "CompilationRegles(alcompil)" );
           return _False;
        }
    }
    TxtChaineRsrc( 134 , 4 , chRegleMemoire );
    
    /* parcours de toutes les regles
    if( p_estInsertionFichier )
    {
       if( !SauveArchivesDictionnaire() )
       {
           TxtErreur( 29 , "CompilationRegles(alcompil)" );
           return _False;
        }
    }
    */

    *p_nbReglesFausses = 0;
    /* parcours de toutes les regles */
    while( carLu != EOF && carLu != '#' )
    {
        if( p_estVerifSyntaxe )
        {
             _Echo( '.' );
        }
        /* on stocke l'emplacement fichier du debut de la regle
           pour le message d'erreur de compilation eventuel */
        fgetpos( PtFicRegles , &posFic );
        *ChErreurCompil = '\0';

        /* regle de mot simple */
        if( carLu == 'W' && !p_estRegleMemoire )
        {
	    if( Trace )
            {
                fprintf( FicLog , "\nCompilation lemme\n" );
                PhotoMemoire( FicLog );
            }
            (*p_nbReglesMots)++;

            /* erreur de syntaxe */
            ptLemme = NouveauLemme();
            if( !CompileLemme( ptLemme , &carLu , PtFicRegles ) )
            {
                (*p_nbReglesFausses)++;
                MessageErreurCompil( posFic , &carLu , PtFicRegles );
                DemarqueEtRendLemme( &ptLemme );
            }
            /* pas d'erreur de syntaxe */
            else
            {
	         *chLog = '\0';
    
                 /* mot deja connu */
                 if( !GarnirReferenceLemme( ptLemme ) )
                 {
                      TxtChaineErreur( 185 , ChAux );
                      sprintf( ChAux , "%s [%s]\n" , ChAux ,
                               ptLemme->ValTerme.Ident );
                      strcat( ChErreurCompil , ChAux );
                      MessageErreurCompil( posFic , &carLu , PtFicRegles );

                      (*p_nbReglesFausses)++;
                      DemarqueEtRendLemme( &ptLemme );
                      if( !p_estInsertionFichier || p_estVerifSyntaxe )
                      {
                           break;
                      }
                 }

                 /* complete les listes des traits morphologiques */
                 else if( !CompleteTraitsMorphologiquesLemme( ptLemme ,
								 chLog ) )
                 {
                      MessageErreurCompil( posFic , &carLu , PtFicRegles );

                      (*p_nbReglesFausses)++;
                      DemarqueEtRendLemme( &ptLemme );
                      if( !p_estInsertionFichier || p_estVerifSyntaxe )
                      {
                           break;
                      }
                 }

		 /* complete les listes des traits semantiques */
                 else if( !CompleteTraitsLiensSemantiquesLemme( ptLemme ,
								   chLog ) )
                 {
                      MessageErreurCompil( posFic , &carLu , PtFicRegles );

                      (*p_nbReglesFausses)++;
                      DemarqueEtRendLemme( &ptLemme );
                      if( !p_estInsertionFichier || p_estVerifSyntaxe )
                      {
                           break;
                      }
                 }

                 /* mot inconnu dont la racine eventuelle est connue */
                 else
                 {
                       /* insertion d'une regle OK */
                      if( p_estInsertionFichier && ptLemme )
                      {
                           ptLemme->NbRegles = _Nil;
			   if( !InsertionLemme( ptLemme , _True , chLog ) ) {
			     MessageErreurCompil( posFic , 
						  &carLu , PtFicRegles );

			     (*p_nbReglesFausses)++;
			     DemarqueEtRendLemme( &ptLemme );
			   }
                           DemarqueEtRendLemme( &ptLemme );
                      }
                      /* pas d'insertion : on rend le lemme */
                      else if( ptLemme )
                      {
                           DemarqueEtRendLemme( &ptLemme );
                      }

                      if( Trace )
                      {
                          fprintf( FicLog , "\nApres compil mot\n" );
                          PhotoMemoire( FicLog );
                      }
                 }
                 /* mot inconnu dont la racine eventuelle est connue */
            }
            /* pas d'erreur de syntaxe */

            if( (*p_nbReglesMots) % 100L == 1L )
            {
                sprintf( ChAux , "%ld" , (*p_nbReglesMots) );
                TxtEcritItemInfo( 1 , ChAux );
            }
            if( p_estInsertionFichier 
                && (*p_nbReglesMots) % FrequenceMAJBArbre == 0 )
            {
               _FinDialogueInfo();
               if( !IniBArbre() )
               {
                   p_nbReglesFausses++;
               }
               _NouveauDialogueInfo( 128 , _True );
               TxtTitresDialogueInfo( 27 , 28 , 29 );
               sprintf( ChAux , "%ld" , (*p_nbReglesMots) );
               TxtEcritItemInfo( 1 , ChAux );
               sprintf( ChAux , "%ld" , (*p_nbReglesTermes) );
               TxtEcritItemInfo( 2 , ChAux );
            }
        }
        /* regle de mot simple */
        /* regle de mot compose */
        else if( carLu == 'R' )
        {
            if( Trace )
            {
                fprintf( FicLog , "\nCompil regle\n" );
                PhotoMemoire( FicLog );
            }
            (*p_nbReglesTermes)++;

            /* erreur de syntaxe */
            if( !CompileRegle( &ptRegle , &carLu , PtFicRegles ,
                                &ptValTermeLexicalisation , 
                                p_estRegleMemoire ) )
            {
                (*p_nbReglesFausses)++;
                MessageErreurCompil( posFic , &carLu , PtFicRegles );
                if( !p_estRegleMemoire )
                {
                    DemarqueEtRendRegle( &ptRegle );
                }
                else
                {
                    DemarqueEtRendArbre( &(ptRegle->TermeFils) );
                    NbReglesMem--;
                    if( NbReglesMem > 0 )
                    {
                         (ReglesMem[ NbReglesMem - 1 ]).RegleFrere = NULL;
                    }
                }
                if( !p_estInsertionFichier || p_estVerifSyntaxe )
                {
                      break;
                }
            }
            /* pas d'erreur de syntaxe */
            else
            {
                if( Trace )
                {
                   EcrireRegleMem( &ptRegle , _False , FicLog , _Sortie );
                   PhotoMemoire( FicLog );
                }

                /* insertion d'une regle OK */
                if( p_estInsertionFichier ) {
		  if( ptRegle ) {
		    
		    /* recherche automatique de terme
		       de rattachelment si non fourni */
		    if( !ptValTermeLexicalisation ) {
		      if( LexicalisationImperativeDesRegles ) {
                        /* en terminologie, on sait que toutes les
			   regles correspondent a des termes lexicaux */
			ptValTermeLexicalisation 
			  = LexicalisationImperative( ptRegle->TermeFils );
		      }
		      else {
			ptValTermeLexicalisation 
			  = LexicalisationSansRisque( ptRegle->TermeFils );
		      }
		    }

		    if( !InsertionDesallocRegle( &ptRegle ,
						 ptValTermeLexicalisation ) ) {
		      (*p_nbReglesFausses)++;
		      MessageErreurCompil( posFic , &carLu , PtFicRegles );
		      DemarqueEtRendRegle( &ptRegle );
		      if( !p_estInsertionFichier || p_estVerifSyntaxe ) {
			break;
		      }
		    }

		    if( Trace ) {
		      fprintf( FicLog , "\nApres compil regle\n" );
		      PhotoMemoire( FicLog );
		    }
		  }
		  else {
		    DemarqueEtRendRegle( &ptRegle );
		  }
                }

                /* pas d'insertion : on rend la regle */
                if( p_estVerifSyntaxe )
                {
                    DemarqueEtRendRegle( &ptRegle );
                }

                /* regle memoire on le signale */
                if( p_estRegleMemoire )
                {
                    if( ptRegle->TermeFils )
                    {
                        sprintf( ChAux , "Rule [%s] --> %s\n" , 
                                 ptRegle->TermeFils->ValTerme.Ident , 
                                 chRegleMemoire );
                    }
                    else
                    {
                        sprintf( ChAux , "%s\n" , chRegleMemoire );
                    }
                    fprintf( FicLog , ChAux );
                }
            }
            /* pas d'erreur de syntaxe */

            if( (*p_nbReglesTermes) % 100L == 1L )
            {
                 sprintf( ChAux , "%ld" , (*p_nbReglesTermes) );
                 TxtEcritItemInfo( 2 , ChAux );
            }
        }
        /* regle de mot compose */
        else
        {
             TxtChaineErreur( 187 , ChAux );
             strcpy( ChErreurCompil , ChAux );
             MessageErreurCompil( posFic , &carLu , PtFicRegles );
             break;
        }
    }
    /* parcours de toutes les regles */

    _FinDialogueInfo();
    if( p_estInsertionFichier )
    {
       if( !IniBArbre() )
       {
          p_nbReglesFausses++;
       }

       MapReferencesTermesSimples( _False , _False );
    }

    if( p_carLu )
    {
        *p_carLu = carLu;
    }
    if( !p_ptFic )
    {
        fclose(PtFicRegles);
    }

    fprintf( FicLog , "\nCompiled: %ld lemma(s) %ld rule(s)\n" ,
                           (*p_nbReglesMots) , (*p_nbReglesTermes)  );
    fprintf( FicLog , "Compiling error(s): %d\n\n" , *p_nbReglesFausses );

    return *p_nbReglesFausses == 0;
}

/*****************************************************************************
            Lecture des metaregles
*****************************************************************************/
char    CompilationReglesMeta( FILE *p_ptFic , char *p_carLu )
{
    TabRegleMemMeta            ptRegleMeta;
    TabRegleMemMeta            ptRegleMetaAux;
    TabRegleMemMeta            ptRegleMetaPrec;
    unsigned long              posFic;
    int                        noTableau;
    char                       trouve;
    char                      *ptCh;
    char                       resulCompil;

    resulCompil = _True;
 
    if( !RAZReglesMeta() )
    {
       TxtErreur( 31 , "CompilationReglesMeta(alcompil)" );
       return _False;
    }

    while( (*p_carLu) != EOF  && (*p_carLu) != '#' )
    {
        ptRegleMeta = NouveauRegleMeta();

        /* on stocke l'emplacement fichier du debut de la regle
           pour le message d'erreur de compilation eventuel */
        fgetpos( p_ptFic , &posFic );
        *ChErreurCompil = '\0';

        if( !CompileRegleMeta( ptRegleMeta , p_carLu , p_ptFic ) )
        {
            MessageErreurCompil( posFic , p_carLu , p_ptFic );
            DemarqueEtRendRegleMeta( &ptRegleMeta );
            resulCompil = _False;
        }
        else
        {
           /* references lexicales des termes de la metaregle */
           if( !GarnirReferenceRegleMeta( ptRegleMeta ) )
           {
               TxtChaineErreur( 165 , ChAux );
               if( ptRegleMeta->TermeFils )
               {
                    sprintf( ChErreurCompil ,
                             "%s [%d]" , ChAux , ptRegleMeta->NoTableau );
               }
               else
               {
                    sprintf( ChErreurCompil , "%s " , ChAux );
               }
               MessageErreurCompil( posFic , p_carLu , p_ptFic );
               resulCompil = _False;
           }

           /* insertion de la metaregle dans la liste du tableau
              correspondant par No de regle croissant */
           if( !DagANoMetaregle( 
                   ptRegleMeta->TermeFils->TermeFils->ValTerme.ArcsFils , 
                   &ptCh , &noTableau , &trouve ) )
           {
               return _False;
           }
           ptRegleMeta->NoTableau = noTableau;

           /* cas d'un identificateur de metaregle non encore connu */
           if( !ptCh )
           {
               TxtChaineErreur( 256 , ChErreurCompil );
               MessageErreurCompil( posFic , p_carLu , p_ptFic );
               resulCompil = _False;
           }
           else if( !trouve )
           {
               strcpy( Tableaux[ noTableau ].Ident , ptCh );
               Tableaux[ noTableau ].PtRegleMeta = ptRegleMeta;
           }
           else
           /* cas d'un identificateur de metaregle deja connu :
              insertion en queue */
           {
               ptRegleMetaAux = Tableaux[ noTableau ].PtRegleMeta;
               ptRegleMetaPrec = Tableaux[ noTableau ].PtRegleMeta;
               while( ptRegleMetaAux  )
               {
                   ptRegleMetaPrec = ptRegleMetaAux;
                   ptRegleMetaAux = ptRegleMetaAux->RegleFrere;
               }

               /* insertion de la nouvelle regle */
               ptRegleMeta->RegleFrere = ptRegleMetaAux;
               if( ptRegleMetaPrec == NULL )
               {
                   Tableaux[ noTableau ].PtRegleMeta = ptRegleMeta;
               }
               else
               {
                   ptRegleMetaPrec->RegleFrere = ptRegleMeta;
               }
           }
        }
    }
    return  resulCompil;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                        fin de alcompil.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
