/*
 * FASt Term Recognizer
 *
 *  FASTR/aldags.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for directed acyclic gralphs management
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*****************************************************************************
    EXISTENCE D'UN ARC VIDE BIDON POUR POINTAGE COMMUN
*****************************************************************************/
TabDagMem     ExisteDagVideFils( TabDagMem p_filsArc )
{
  if( p_filsArc ) {
    if( p_filsArc->TypeDag == _TypeDagVide ) {
      return p_filsArc;
    }         
  }
  return NULL;
}

/*****************************************************************************
    EXISTENCE D'UN ATOME DE DAG AVEC UNE ETIQUETTE DONNEE
*****************************************************************************/
TabDagMem     ExisteAtomeFils( TabDagMem p_filsArc , 
                               int p_typeAtome , struct ValAtomeMem p_atome )
{
  TabDagMem                filsCou;

  filsCou = p_filsArc;
  while( filsCou ) {
    if( filsCou->TypeDag == (char)p_typeAtome ) {
      if( filsCou->ValDag.Atome.AtomeConjDisj == p_atome.AtomeConjDisj
	  && (  ( p_typeAtome == _AtomeRef
		  && filsCou->ValDag.Atome.ValAtome.Ref
		  == p_atome.ValAtome.Ref)
		||( p_typeAtome == _AtomeChaine
		    && strcmp( filsCou->ValDag.Atome.ValAtome.Chaine , 
			       p_atome.ValAtome.Chaine ) == 0 )
		||( p_typeAtome == _AtomeEntier
		    && filsCou->ValDag.Atome.ValAtome.Entier 
		       == p_atome.ValAtome.Entier ) )  ) {
	return filsCou;
      }
    }         
    filsCou = filsCou->Frere;
  }
  return NULL;
}

/*****************************************************************************
    EXISTENCE D'UN ARC DE DAG AVEC UNE ETIQUETTE DONNEE
*****************************************************************************/
TabDagMem     ExisteArcOuListeFils( TabDagMem p_filsArc , 
				    TabDagMem p_arcSource , int p_typeDag )
{
  TabDagMem                         filsCou;
  TabDagMem                         petitFilsSourceCou;
  TabDagMem                         petitFilsDestCou;
  int                               listeEgales;

  filsCou = p_filsArc;
  /* tries all the dags in the target */
  while( filsCou ) {

    /* a target dag with the appropriate type */
    if( filsCou->TypeDag == (char)p_typeDag ) {

      /* 2 dags with the same type */
      if( p_typeDag == _AtomeListe ) /* liste */ {

	petitFilsDestCou = filsCou->Fils;
	petitFilsSourceCou = p_arcSource->Fils;
	listeEgales = _True;

	/* comparison of the list values 1 by 1 */
	while( petitFilsDestCou && petitFilsSourceCou ) {
	  
	  /* different types */
	  if( petitFilsDestCou->TypeDag != petitFilsSourceCou->TypeDag ) {
	    listeEgales = _False;
	    break;
	  }

	  /* comparison of values in case of identical types */
	  if( petitFilsDestCou->TypeDag == _AtomeEntier ) {
	    if( petitFilsDestCou->ValDag.Atome.ValAtome.Entier 
		!= petitFilsSourceCou->ValDag.Atome.ValAtome.Entier ) {
	      listeEgales = _False;
	      break;
	    }
	  }
	  else if( petitFilsDestCou->TypeDag == _AtomeRef ) {
	    if( petitFilsDestCou->ValDag.Atome.ValAtome.Ref 
		!= petitFilsSourceCou->ValDag.Atome.ValAtome.Ref ) {
	      listeEgales = _False;
	      break;
	    }
	  }
	  else if( petitFilsDestCou->TypeDag == _AtomeChaine ) {
	    if( strcmp( petitFilsDestCou->ValDag.Atome.ValAtome.Chaine , 
			petitFilsSourceCou->ValDag.Atome.ValAtome.Chaine ) 
		!= 0 ) {
	      listeEgales = _False;
	      break;
	    }
	  }
	  else {
	    TxtErreur( 302 , "ExisteArcOuListeFils(aldags)" );
	  }
	  
	  petitFilsDestCou = petitFilsDestCou->Fils;
	  petitFilsSourceCou = petitFilsSourceCou->Fils;

	}
	/* comparison of the list values 1 by 1 */

	if( listeEgales ) {
	  return filsCou;
	}
      }
      else if( p_typeDag == _Arc )  /* arc */ {
	if( filsCou->ValDag.Arc.ValSymb == p_arcSource->ValDag.Arc.ValSymb ) {
	  return filsCou;
	}
      }
      else {
	TxtErreur( 303 , "ExisteArcOuListeFils(aldags)" );
      }
    }
    /* a target dag with the appropriate type */

    filsCou = filsCou->Frere;
  }
  /* tries all the dags in the target */
  return NULL;
}

/*****************************************************************************
    EXISTENCE D'UN CHEMIN DE DAG DONNE ET RENVOI DE L'ADRESSE
         DU FILS DU DEERNIER ARC TROUVE (SI OK)
*****************************************************************************/
TabDagMem    *ExisteCheminFils( TabDagMem p_ptSrc , TabDagMem *p_hdlDest )
{
  TabDagMem                 *hdlDest;
  TabDagMem                  ptSrcCou;
  TabDagMem                  ptDestCou;

  hdlDest = p_hdlDest;
  ptSrcCou = p_ptSrc;
  /* on avance dans les fils du dag tant qu'ils coincident avec le
     chemin et qu'ils sont de type _Arc */
  /* si aucun fils du dag ne coincide avec le dag courant du
     chemin, on renvoie faux */
  while( _True ) {
    if( ptSrcCou && ptSrcCou->TypeDag == _Arc ) {
      if( (ptDestCou = ExisteArcOuListeFils( (*hdlDest) , 
					     ptSrcCou , _Arc ) ) ) {
	hdlDest = &(ptDestCou->Fils);
      }
      else
	{
	  return NULL;
	}
    }
    else
      {
	break;
      }
    ptSrcCou = ptSrcCou->Fils;
  }
  return hdlDest;
}

/*****************************************************************************
    AJOUT D'UN DAG VIDE UTILISE POUR POINTAGE COMMUN
*****************************************************************************/
TabDagMem     AjouteDagVideFils( TabDagMem *p_hdlArc )
{
  if( (*p_hdlArc) ) {
    TxtErreur( 93 , "AjouteDagVideFils(aldags)" );
    return NULL;
  }

  (*p_hdlArc) = NouveauDag( _TypeDagVide );
  if( !(*p_hdlArc) ) {
    return NULL;
  }

  return (*p_hdlArc);
}

/*****************************************************************************
    AJOUT D'UNE VALEUR DE DAG D'UN TYPE DONNE
*****************************************************************************/
TabDagMem     AjouteAtomeFils( TabDagMem *p_hdlArc , 
                               int p_typeAtome , struct ValAtomeMem p_atome ,
                               int p_estAjoutEnQueue )
{
  TabDagMem                nouvFils;
  TabDagMem                filsCou;

  if( (*p_hdlArc) ) {
    /* si le dag existant n'est pas le dag vide, on ne paut
       ajouter l'atome */
    if( (*p_hdlArc)->TypeDag == _TypeDagVide ) {
      nouvFils = (*p_hdlArc);
      (*p_hdlArc)->TypeDag = p_typeAtome;
      (*p_hdlArc)->ValDag.Atome = p_atome;

      return (*p_hdlArc);
    }
    else {
      /* seuls des atomes ou listes entre eux peuvent etre des
	 fils multiples */
      if( (*p_hdlArc)->TypeDag != _AtomeChaine
	  && (*p_hdlArc)->TypeDag != _AtomeListe
	  && (*p_hdlArc)->TypeDag != _AtomeRef
	  && (*p_hdlArc)->TypeDag != _AtomeEntier ) {
	TxtErreur( 92 , "AjouteAtomeFils(aldags)" );
	return NULL;
      }

      nouvFils = NouveauDag( p_typeAtome );
      if( !nouvFils ) {
	return NULL;
      }
      nouvFils->ValDag.Atome = p_atome;

	/* ajout du fils dans la liste des arcs */
      if( p_estAjoutEnQueue ) {
	filsCou = (*p_hdlArc);
	while( filsCou->Frere ) {
	  filsCou = filsCou->Frere;
	}
	filsCou->Frere = nouvFils;

	return nouvFils;
      }
      /* ajout en tete de la liste des fils */ 
      else {
	filsCou = (*p_hdlArc)-> Fils;
	(*p_hdlArc)-> Fils = nouvFils;
	nouvFils->Frere = filsCou;
	/* RestaurePointeurVersDagJete( *p_hdlArc , nouvFils );
	   if( *p_hdlArc != nouvFils ) {
	   TxtErreur( 1 , "revoir restauration pointeurs 1" );
	   } */
      }

      return nouvFils;
    }
  }
  else {
    (*p_hdlArc) = NouveauDag( p_typeAtome );
    if( !(*p_hdlArc) ) {
      return NULL;
    }
    (*p_hdlArc)->ValDag.Atome = p_atome;
    
    return (*p_hdlArc);
  }
}

/*****************************************************************************
    AJOUT D'UN ARC DE DAG AVEC UNE ETIQUETTE DONNEE
*****************************************************************************/
TabDagMem    *AjouteArcOuListeFils( TabDagMem *p_hdlArc , 
                             union ValueDag *p_valArc , TabDagMem p_arcAjoute ,
                             int p_estAjoutEnQueue , int p_typeDag )
{
  TabDagMem                       nouvFils;
  TabDagMem                       filsCou;

  /* some other daughter dags already exist */
  if( (*p_hdlArc) ) {
    if( (*p_hdlArc)->TypeDag == _TypeDagVide ) {
      nouvFils = (*p_hdlArc);
    }
    else
      {
	/* seuls des arcs ou des listes entre eux peuvent etre
	   des fils multiples */
	if( ( p_typeDag == _Arc && (*p_hdlArc)->TypeDag != _Arc )
	    || ( p_typeDag == _AtomeListe 
		 && (*p_hdlArc)->TypeDag != _AtomeListe ) ) {
	  printf( "%d/%d\n" , p_typeDag , (*p_hdlArc)->TypeDag );
	  TxtErreur( 93 , "AjouteArcOuListeFils(aldags)" );
	  return NULL;
	}

	if( !p_arcAjoute ) {
	  nouvFils = NouveauDag( p_typeDag );
	  if( !nouvFils ) {
	    return NULL;
	  }
	}
	else
	  {
	    nouvFils = p_arcAjoute;
	  }
             
	/* ajout du fils dans la liste des arcs */
	if( p_estAjoutEnQueue ) {
	  filsCou = (*p_hdlArc);
	  while( filsCou->Frere ) {
	    filsCou = filsCou->Frere;
	  }
	  filsCou->Frere = nouvFils;
	}
	else /* ajout en tete de la liste des fils */
	  {
	    filsCou = (*p_hdlArc);

	    RestaurePointeurVersDagJete( *p_hdlArc , nouvFils );
	    if( *p_hdlArc != nouvFils ) {
	      TxtErreur( 1 , "revoir restauration pointeurs 2" );
	    }

	    nouvFils->Frere = filsCou;
	  }
      }
  }
  else
    {
      if( !p_arcAjoute ) {
	nouvFils = NouveauDag( p_typeDag );
	if( !nouvFils ) {
	  return NULL;
	}
      }
      else
	{
	  nouvFils = p_arcAjoute;
	}
      (*p_hdlArc) = nouvFils;
    }

  /* the value of the dag is provided, not the dag itself */
  /* its value must be copied */
  if( !p_arcAjoute ) {
    nouvFils->TypeDag = (char)p_typeDag;
    if( p_typeDag == _Arc ) {
      nouvFils->ValDag.Arc.ValSymb = p_valArc->Arc.ValSymb;
      nouvFils->ValDag.Arc.ValRef = p_valArc->Arc.ValRef;
      strcpy( nouvFils->ValDag.Arc.Ident , p_valArc->Arc.Ident );
      nouvFils->ValDag.Arc.ValMin = p_valArc->Arc.ValMin;
      nouvFils->ValDag.Arc.ValMax = p_valArc->Arc.ValMax;
      nouvFils->ValDag.Arc.Comptage = p_valArc->Arc.Comptage;
      nouvFils->ValDag.Arc.ArcERNegation = p_valArc->Arc.ArcERNegation;
    }
    else if( p_typeDag == _AtomeListe ) { 
      nouvFils->ValDag.Atome.AtomeConjDisj =
	p_valArc->Atome.AtomeConjDisj;
      /* nothing to copy, values are useless */
    }
    else {
      printf( "Type: %d\n" , p_typeDag );
      TxtErreur( 307 , "AjouteArcOuListeFils(aldags)" );
      return NULL;
    }
  }

  return &(nouvFils->Fils);
}

/*****************************************************************************
 AJOUT D'UNE OPERATION LOGIQUE ENTRE DAGS ET VALEURS AVEC UNE ETIQUETTE DONNEE
*****************************************************************************/
TabDagMem     AjoutOperFils( TabDagMem *p_filsOperLog , int p_oper ,
                              int p_typeFils1 , TabChar p_fils1 ,
                              int p_typeFils2 , TabChar p_fils2 )
{
  TabDagMem                    nouvFils;
  TabDagMem                    filsCou;
  TabDagMem                   *hdlAux;
  TabDagMem                    ptAux;
     
  nouvFils = NouveauDag( _Oper );
  if( !nouvFils ) {
    return NULL;
  }

  /* rajout du fils en queue de liste */
  if( !(*p_filsOperLog) ) {
    (*p_filsOperLog) = nouvFils;
  }
  else
    {
      filsCou = (*p_filsOperLog);
      while( filsCou->Frere ) {
	filsCou = filsCou->Frere;
      }
      filsCou->Frere = nouvFils;
    }

  nouvFils->ValDag.OperLog.Oper = (char)p_oper;
  nouvFils->ValDag.OperLog.TypeFils1 = (char)p_typeFils1;
  nouvFils->ValDag.OperLog.TypeFils2 = (char)p_typeFils2;
  if( p_typeFils1 == _Arc ) {
    hdlAux = (TabDagMem *)p_fils1;
    nouvFils->ValDag.OperLog.Fils1.hdlPereChemin = hdlAux;
  }
  else
    {  /* atome */
      ptAux = (TabDagMem)p_fils1;
      nouvFils->ValDag.OperLog.Fils1.ptPereAtome = ptAux;
    }
  if( p_typeFils2 == _Arc ) {
    hdlAux = (TabDagMem *)p_fils2;
    nouvFils->ValDag.OperLog.Fils2.hdlPereChemin = hdlAux;
  }
  else
    {  /* atome */
      ptAux = (TabDagMem)p_fils2;
      nouvFils->ValDag.OperLog.Fils2.ptPereAtome = ptAux;
    }

  return nouvFils;
}

/*****************************************************************************
    AJOUT D'UN CHEMIN DANS UN DAG, RETOURNE L'ADRESSE DU FILS DU
    DERNIER DAG AJOUTE
*****************************************************************************/
TabDagMem    *CopieChemin( TabDagMem *p_hdlDest , TabDagMem p_ptSrc )
{
  TabDagMem               *hdlDest;
  TabDagMem                ptSrcCou;
  TabDagMem                ptDestCou;

  hdlDest = p_hdlDest;
  ptSrcCou = p_ptSrc;
  /* on avance dans les fils du dag tant qu'ils coincident avec le chemin 
     et qu'ils sont de type _Arc                                          */
  /* si aucun fils du dag ne coincide avec le dag courant du chemin, on 
     l'ajoute                                                             */
  while( _True ) {
    if( ptSrcCou && ptSrcCou->TypeDag == _Arc ) {
      if( !(ptDestCou = ExisteArcOuListeFils( (*hdlDest) , 
					      ptSrcCou , 
					      ptSrcCou->TypeDag ) ) ) {
	hdlDest = AjouteArcOuListeFils( hdlDest , 
					&(ptSrcCou->ValDag) , 
					NULL , _True , 
					ptSrcCou->TypeDag );
	if( !hdlDest ) {
	  return NULL;
	}
      }
      else
	{
	  hdlDest = &(ptDestCou->Fils);
	}
    }
    else
      {
	break;
      }
    ptSrcCou = ptSrcCou->Fils;
  }
  return hdlDest;
}

/*****************************************************************************
    AJOUT D'UN CHEMIN DANS UN DAG, RETOURNE L'ADRESSE DU FILS DU
    DERNIER DAG AJOUTE
*****************************************************************************/
TabDagMem     AjouteCheminEtAtomeFils( TabDagMem *p_hdlDest ,
                                TabDagMem p_ptSrc , int p_typeAtome ,
                                struct ValAtomeMem p_atome ,
                                int p_estAjoutEnQueue )
{
  TabDagMem               *hdlDest;

  if( ( hdlDest = CopieChemin( p_hdlDest , p_ptSrc ) ) ) {
    return AjouteAtomeFils( hdlDest , p_typeAtome , 
			    p_atome , p_estAjoutEnQueue );
  }
  else  {
    return NULL;
  }
}

/*****************************************************************************
    UNIFICATION DE DEUX CHEMINS DE DAG
*****************************************************************************/
TabDagMem     AjoutEgaliteCheminChemin( TabDagMem *p_filsOperLog ,
				   int p_oper ,
                                   TabDagMem *p_hdlArc1 , 
                                   TabDagMem *p_hdlArc2 , 
                                   TabDagMem p_filsArcChemin1 , 
                                   TabDagMem p_filsArcChemin2 )
{
  TabDagMem                            *hdlFilsIni1;
  TabDagMem                            *hdlFilsIni2;

  hdlFilsIni1 = CopieChemin( p_hdlArc1 , p_filsArcChemin1 );
  hdlFilsIni2 = CopieChemin( p_hdlArc2 , p_filsArcChemin2 );
  return AjoutOperFils( p_filsOperLog , p_oper , 
			_Arc , (TabChar)hdlFilsIni1 , 
			_Arc , (TabChar)hdlFilsIni2 );
}

/*****************************************************************************
    AJOUT D'UNE CONTRAINTE D'EGALITE ENTRE UN CHEMIN ET UNE VALEUR DE DAG
*****************************************************************************/
TabDagMem     AjoutOperCheminAtome( TabDagMem *p_filsOperLog , int p_oper ,
				    TabDagMem *p_hdlArc , 
				    TabDagMem p_filsArcChemin1 , 
				    TabDagMem p_ptAtome , 
				    int p_typeDagDernierArcChemin )
{
  TabDagMem              *teteCheminCopieSurDagTerme;

  /* recopie le chemin de l'equation sur le dag du terme
     correspondant */
  teteCheminCopieSurDagTerme = CopieChemin( p_hdlArc , p_filsArcChemin1 );
  return AjoutOperFils( p_filsOperLog , p_oper , 
			p_typeDagDernierArcChemin , 
			(TabChar)teteCheminCopieSurDagTerme , 
			p_ptAtome->TypeDag , (TabChar)p_ptAtome );
}

/*****************************************************************************
     PASSAGE D'UN CHEMIN DE REGLE DONT LE PREMIER TRAIT EST UN IDENTIFICATEUR
     DE TERME AU POINTEUR VERS LE DAG DE CE TERME + MODIFICATION DU CHEMIN 
     EN SUPPRIMANT CE PREMIER TRAIT DESORMAIS SUPERFLU
*****************************************************************************/
TabDagMem    *TeteDeCheminADagDeTerme( TabDagMem *p_hdlArcCheminLu )
{
  TabDagMem                        filsCou;
  struct ValTerme                 *ptValTerme;
  int                              indTerme;

  filsCou = (*p_hdlArcCheminLu);
  /* recherche du premier trait du chemin et de l'identificateur associe */
  if( filsCou && filsCou->TypeDag == _Arc ) {
    if( filsCou->ValDag.Arc.ValSymb != _Nil ) {
      TxtErreur( 93 , "TeteDeCheminADagDeTerme(aldags)" );
      return NULL;
    }
  }
  else
    {
      TxtErreur( 93 , "TeteDeCheminADagDeTerme(aldags)" );
      return NULL;
    }

  /* recherche du terme de la regle ayant cet identificateur */
  for( indTerme = 0 ; indTerme < NbLiensIdentTerme ; indTerme++ ) {
    if( strcmp( LienIdentTerme[ indTerme ].Ident ,
		filsCou->ValDag.Arc.Ident ) == 0 ) {
      /* on supprime du premier trait du chemin */
      (*p_hdlArcCheminLu) = filsCou->Fils;
      DemarqueDag( &filsCou );
      RendUnDag( &filsCou );

      /* on releve le pteur vers le dag de ce terme */
      ptValTerme = LienIdentTerme[ indTerme ].PtValTerme;
      return &(ptValTerme->ArcsFils);
    }
  }
  TxtChaineErreur( 100 , ChAux );
  sprintf( ChErreurCompil , "%s : [%s]" , ChAux ,
	   filsCou->ValDag.Arc.Ident );
  return NULL;
}

/*****************************************************************************
    LECTURE D'UN ATOME
*****************************************************************************/
char     LireUnAtome( char *p_carLu , FILE *p_ptFic ,
                      TabDagMem *p_unPereAtome , int p_estEtiquette ,
                      int p_traitCatLex )
{
  ChIdent            ident;
  long               ref;
  long               valRet;
  int                erreur;
  int                trouve;
  int                traitCatLex;

  if( !( _Alphanum(*p_carLu) || *p_carLu == '\'' 
	 || *p_carLu == '-'  /* signe - */ ) ) {
    printf( "[%c]\n" , *p_carLu );
    TxtChaineErreur( 171 , ChErreurCompil );
    return _False;
  }

  if( _Alpha(*p_carLu) ) {
    LireIdentificateur( p_ptFic , p_carLu , 
			ident , _MaxLetIdent , _True );
    *p_unPereAtome = NouveauDag( _AtomeRef );
    if( !(*p_unPereAtome) ) {
      return _False;
    }

    if( p_estEtiquette ) {
      valRet 
	= InsereIdentFichier( ident , &ref , FicRefIdent , 
			      &erreur , &trouve );
      (*p_unPereAtome)->ValDag.Atome.ValAtome.Ref = ref;
      if( valRet == _Nil || erreur ) {
	return _False;
      }
    }
    else
      {
	if( !IdentATraitsCatLex( ident , p_traitCatLex , &traitCatLex ) ) {
	  printf( "Identifier [%s]\n" , ident );
	  TxtErreur( 234 , "LireUnAtome(aldags)" );
	  return _False;
	}
	(*p_unPereAtome)->ValDag.Atome.ValAtome.Ref = (long)traitCatLex;
      }
  }
  else if( _Num(*p_carLu) || *p_carLu == '-' ) {
    *p_unPereAtome = NouveauDag( _AtomeEntier );
    if( !(*p_unPereAtome) ) {
      return _False;
    }
    (*p_unPereAtome)->ValDag.Atome.ValAtome.Entier 
      = LireEntier( p_ptFic , p_carLu );
  }
  else if( *p_carLu == '\'' ) {
    *p_unPereAtome = NouveauDag( _AtomeChaine );
    if( !(*p_unPereAtome) ) {
      return _False;
    }
    if( !LireChaineAvecBalise( 
			      p_ptFic , p_carLu , 
			      (*p_unPereAtome)->ValDag.Atome.ValAtome.Chaine ,
			      _MaxLetIdent , '\'' , _True ) ) {
      return _False;
    }
  }

  return _True;
}

/*****************************************************************************
    LECTURE D'UNE LISTE
*****************************************************************************/
char     LireUneListe( char *p_carLu , FILE *p_ptFic ,
                       TabDagMem *p_unPereAtome )
{
  TabDagMem        filsValCou;
  TabDagMem        nouvFils;
  char             premierTerme;
/*   TabDagMem       *hdlDernierArcChemin; */
/*   int              typeDagDernierArcChemin; */

  ReconnaitreCaractere( p_ptFic , (int)'(' , p_carLu , _True );

  *p_unPereAtome = NouveauDag( _AtomeListe );
  if( !(*p_unPereAtome) ) {
    return _False;
  }

  /* il aura 1 fils le premier fils de la liste courante */
  premierTerme = _True;
  while( *p_carLu != ')' && *p_carLu != EOF ) {
    nouvFils = NULL;
    if( *p_carLu == '(' ) {
      if( !LireUneListe( p_carLu , p_ptFic , &(nouvFils) ) ) {
	return _False;
      }
    }
/*     else if( *p_carLu == '<' ) { */
/*       if( !(hdlDernierArcChemin  */
/* 	    = LireUnChemin( p_carLu , p_ptFic , */
/* 			    &nouvFils , _True ,  */
/* 			    &typeDagDernierArcChemin ) ) ) { */
/* 	return _False; */
/*       }   */
/*     } */
    else {
      if( !LireUnAtome( p_carLu , p_ptFic , &(nouvFils) ,
			_False , _CatLex ) ) {
	return _False;
      }
    }

    /* premier terme de la liste */
    if( premierTerme ) {
      (*p_unPereAtome)->Fils = nouvFils;
      premierTerme = _False;
    }
    /* ajout en  queue de liste */
    else {
      filsValCou = (*p_unPereAtome)->Fils;
      while( filsValCou->Frere ) {
	filsValCou = filsValCou->Frere;
      }
      filsValCou->Frere = nouvFils;
    }

    if( *p_carLu == ',' ) {
      ReconnaitreCaractere( p_ptFic , (int)',' , p_carLu , _True );
    }
    else
      {
	return ReconnaitreCaractere( p_ptFic , (int)')' , p_carLu , _True );
      }
  }

  return ReconnaitreCaractere( p_ptFic , (int)')' , p_carLu , _True );
}

/*****************************************************************************
      ANALYSE D'UNE EXPRESSION REGULIERE : CAS D'UN CHOIX
*****************************************************************************/
TabDagMem    CompileERStarOption( char *p_carLu , FILE *p_ptFic ,  
                                  TabDagMem p_ptArcDeb , 
				  TabDagMem p_ptArcFin )
{
  TabDagMem                ptArcFinAux = NULL;
  TabDagMem                ptArcDebAux = NULL;
  TabDagMem                ptArcFinParal = NULL;
  char                     carIni;
  unsigned char            valMin;
  unsigned char            valMax;

  if( _Num(*p_carLu) ) {
    valMin = (unsigned char)LireEntier( p_ptFic , p_carLu );
    if( !ReconnaitreCaractere( p_ptFic , (int)'-' , p_carLu , _True ) ) {
      return NULL;
    }
    valMax = (unsigned char)LireEntier( p_ptFic , p_carLu );
    if( valMax < valMin ) {
      TxtErreur( 235 , "CompileERStarOption(aldags)" );
      return NULL;
    }
    carIni = '0';
    SautBlanc( p_ptFic , p_carLu );
  }
  else {
    carIni = *p_carLu;
    *p_carLu = LireCar( p_ptFic );
    SautBlanc( p_ptFic , p_carLu );
  }

  if( !p_ptArcDeb || !p_ptArcFin ) {
    TxtErreur( 196 , "CompileERStarOption(aldags)" );
    return NULL;
  }

  if( !(p_ptArcDeb->Fils) || p_ptArcDeb->Frere 
      || p_ptArcFin->Fils || p_ptArcFin->Frere ) {
    TxtErreur( 318 , "CompileERStarOption(aldags) 2" );
    return NULL;
  }

  if( p_ptArcDeb == p_ptArcFin ) {
    return p_ptArcFin;
  }

  /* On ajoute un maillon en tete qui est charge de contenir le 
     pointeur vers le dernier dag ajoute pour ?, * et 0
     et de declencher le comptage pour carIni == '0' */
  if( carIni == '?' || carIni == '*'  || carIni == '0' ) {
    if( !( ptArcDebAux = NouveauDag( _Arc ) ) ) {
      return NULL;
    }
    ptArcDebAux->ValDag.Arc.ValSymb = _Nil;
    ptArcDebAux->ValDag.Arc.ValRef = _Nil;
    strcpy( ptArcDebAux->ValDag.Arc.Ident  , "#" );
    if( carIni == '0' ) {
      ptArcDebAux->ValDag.Arc.Comptage = _ArcDebComptage;
      ptArcDebAux->ValDag.Arc.ValMin = valMin;
      ptArcDebAux->ValDag.Arc.ValMax = valMax;
    }

    /* on chaine l'arc de debut et l'arc de debut nouveau*/
    ptArcDebAux->Fils = p_ptArcDeb->Fils;
    p_ptArcDeb->Fils = ptArcDebAux;
  }

  /* On ajoute un dag vide en queue qui est le lien vers lequel pointe 
     le parcours de court-circuit */
  if( carIni == '?' || carIni == '+' || carIni == '*' || carIni == '0' ) {
    if( !( ptArcFinAux = NouveauDag( _Arc ) ) ) {
      return NULL;
    }
    ptArcFinAux->ValDag.Arc.ValSymb = _Nil;
    ptArcFinAux->ValDag.Arc.ValRef = _Nil;
    strcpy( ptArcFinAux->ValDag.Arc.Ident  , "#" );
    if( carIni == '0' ) {
      ptArcFinAux->ValDag.Arc.Comptage = _ArcFinComptage;
    }

    /* on chaine l'arc de fin et l'arc de fin nouveau */
    if( carIni == '?' ) {
      p_ptArcFin->Fils = ptArcFinAux;
    }
  }

  /* on ajoute un maillon parallele au dernier maillon servant 
     a revenir pointer au debut dans un parcours de court-circuit 
     avec un parcours minimum */
  if( carIni == '+'  && p_ptArcFin->ValDag.Arc.ValSymb != _Nil ) {
    if( !( ptArcFinParal = NouveauDag( _Arc ) ) ) {
      return NULL;
    }
    ptArcFinParal->ValDag.Arc.ValSymb = p_ptArcFin->ValDag.Arc.ValSymb;
    ptArcFinParal->ValDag.Arc.ValRef = p_ptArcFin->ValDag.Arc.ValRef;
    strcpy( ptArcFinParal->ValDag.Arc.Ident  , 
	    p_ptArcFin->ValDag.Arc.Ident );

    /* on chaine l'arc de fin avec cet arc de fin parallele */
    p_ptArcFin->Frere = ptArcFinParal;
    /* on chaine l'arc de fin parallele et l'arc de fin nouveau */
    ptArcFinParal->Fils = ptArcFinAux;
  }

  if( carIni == '?' ) {
    /* on introduit le court-circuit du dernier frere du deuxieme
      maillon vers la fin (le premier maillon est intentionnellement
      vide pour qu'il n'ait pas de freres) */
    TabDagMem                *ptArcDebCou;

    /* on parcourt tous les freres du 2eme maillon pour s'arreter
       sur le dernier */
    ptArcDebCou =  &((ptArcDebAux->Fils)->Frere);
    while( *ptArcDebCou ) {
      ptArcDebCou  =  &((*ptArcDebCou)->Frere);
    }
    *ptArcDebCou = ptArcFinAux;

    return ptArcFinAux;
  }

  if( carIni == '*' || carIni == '0' ) {
    /* on introduit le retour en sautant l'arc de deb vide */
    p_ptArcFin->Fils = ptArcDebAux;
    if( carIni == '0' ) {
      p_ptArcFin->ValDag.Arc.Comptage = _ArcSuiteComptage;
    }

    /* on introduit le court-circuit pour les boucles on demande de
       passer au moins une fois par ptArcDebAux; donc le court-circuit
       se fait sur les fils de ptArcDebAux */
    {
      TabDagMem  ptCou = ptArcDebAux->Fils;
      if( !ptCou ) {
	TxtErreur( 319 , "CompileERStarOption(aldags)" );
	return NULL;
      }
      while( ptCou->Frere ) {
	ptCou = ptCou->Frere;
      }
      ptCou->Frere = ptArcFinAux;
    }

    return ptArcFinAux;
  }

  if( carIni == '+' ) {
    if( ptArcFinParal ) {
      /* on introduit le retour via l'arc ident a l'arc de deb */
      p_ptArcFin->Fils = p_ptArcDeb->Fils;
    }
    else
      {
	/* on introduit le retour en sautant l'arc de deb vide */
	p_ptArcFin->Frere = p_ptArcDeb->Fils;
      }
 
    return ptArcFinAux;
  }
  TxtErreur( 320 , 
	     "CompileERStarOption(aldags)" );
  return ptArcFinAux;
}

/*****************************************************************************
      LECTURE D'UN LEMME DANS UNE ER
*****************************************************************************/
long     LireLemmeER( char *p_carLu , FILE *p_ptFic , int p_categorie )
{
  int                             noTypeFlexion = _Nil;
  TabLemmeMem                     ptLemme;
  struct ValAtomeMem              atome;
  char                            estMotCourant;
  unsigned long                   posFicLemme;
  long                            resul;
  
  if( !(ptLemme = NouveauLemme()) ) {
    return _Nil;
  }
  
  if( !LireChaineAvecBalise( 
			    p_ptFic , p_carLu , ptLemme->ValTerme.Ident , 
			    _MaxLetIdent , '\'' , _True ) ) {
    DemarqueEtRendLemme( &ptLemme );
    return _Nil;
  }
  
  if( *p_carLu == '/' ) {
    ReconnaitreCaractere( p_ptFic , (int)'/' , p_carLu , _True );
    noTypeFlexion = (int)LireEntier( p_ptFic , p_carLu );
  }
  
  atome.AtomeConjDisj = _False; 
  strcpy( atome.ValAtome.Chaine , ptLemme->ValTerme.Ident );
  if( !AjouteCheminEtAtomeFils( 
			       &(ptLemme->ValTerme.ArcsFils) ,
			       CheminLemme , _AtomeChaine , 
			       atome , _True ) ) {
    DemarqueEtRendLemme( &ptLemme );
    return _Nil;
  }
  
  atome.AtomeConjDisj = _False; 
  atome.ValAtome.Ref = (long)noTypeFlexion;
  if( !AjouteCheminEtAtomeFils( 
			       &(ptLemme->ValTerme.ArcsFils) ,
			       CheminFlexion , _AtomeEntier , 
			       atome , _True ) ) {
    DemarqueEtRendLemme( &ptLemme );
    return _Nil;
  }
  
  atome.AtomeConjDisj = _False; 
  atome.ValAtome.Ref = (long)p_categorie;
  if( !AjouteCheminEtAtomeFils( 
			       &(ptLemme->ValTerme.ArcsFils) ,
			       CheminCategorie ,_AtomeRef , 
			       atome , _True ) ) {
    DemarqueEtRendLemme( &ptLemme );
    return _Nil;
  }

  if( ReferenceTermeLexical( ptLemme->ValTerme.Ident , 
			     &(ptLemme->ValTerme) , &estMotCourant ,
			     &posFicLemme ) != 0L ) {
    resul = ptLemme->ValTerme.Ref;
    DemarqueEtRendLemme( &ptLemme );
    return resul;
  }
  else {
    TxtChaineErreur( 264 , ChErreurCompil );
    resul = _Nil;
    DemarqueEtRendLemme( &ptLemme );
    return resul;
  }
}

/*****************************************************************************
      ANALYSE D'UNE EXPRESSION REGULIERE : CAS D'UNE ALTERNATIVE
        CHOIX ENTRE DES CATEGORIES SIMPLES OU LEURS NEGATIONS
*****************************************************************************/
TabDagMem    CompileERAlternative( char *p_carLu , FILE *p_ptFic , 
                             TabDagMem p_ptArcDeb )
{
  ChIdent            ident;
  int                categorie;
  TabDagMem          ptDebNou;
  TabDagMem          ptArcDebCou;
  TabDagMem          ptArcDebCommun;
  TabDagMem          ptArcFinCommun;
  char               estNeg;

  if( !ReconnaitreCaractere( p_ptFic , (int)'[' , p_carLu , _True ) ) {
    return NULL;
  }

  if( p_ptArcDeb->Frere || p_ptArcDeb->Fils ) {
    TxtErreur( 321 , "CompileERAlternative(aldags)" );
    return NULL;
  }

  ptArcFinCommun = NULL;
  ptArcDebCommun = NULL;
  if( *p_carLu == '^' ) {
    estNeg = _True;
    ReconnaitreCaractere( p_ptFic , (int)'^' , p_carLu , _True );

    ptArcDebCou = p_ptArcDeb;
  }
  else
    {
      estNeg = _False;

      if( !( ptArcDebCommun = NouveauDag( _Arc ) ) ) {
	return NULL;
      }
      ptArcDebCommun->ValDag.Arc.ValSymb = _Nil;
      ptArcDebCommun->ValDag.Arc.ValRef = _Nil;
      strcpy( ptArcDebCommun->ValDag.Arc.Ident  , "#" );
      p_ptArcDeb->Fils = ptArcDebCommun;

      ptArcDebCou = ptArcDebCommun;
    }

  while( _Alpha(*p_carLu) ) {
    LireIdentificateur( p_ptFic , p_carLu , ident ,  
			_MaxLetIdent , _True );
    if( !IdentATraitsCatLex( ident , _CatLex , &categorie ) ) {
      return NULL;
    }

    ptDebNou = NouveauDag( _Arc );
    if( !ptDebNou ) {
      return NULL;
    }
    ptDebNou->TypeDag = _Arc;
    ptDebNou->ValDag.Arc.ValSymb = (long)categorie;
    ptDebNou->ValDag.Arc.ValRef = _Nil;
    ptDebNou->ValDag.Arc.ArcERNegation = estNeg;
    strcpy( ptDebNou->ValDag.Arc.Ident , ident );

    /* cas ou la reference est donnee */
    if( *p_carLu == '/' ) {
      ReconnaitreCaractere( p_ptFic , (int)'/' , p_carLu , _True );
      if( _Num( *p_carLu ) ) {
	ptDebNou->ValDag.Arc.ValRef = LireEntier( p_ptFic , p_carLu );
      }
      else if( *p_carLu == '\'' ) {
	ptDebNou->ValDag.Arc.ValRef = LireLemmeER( p_carLu , 
						   p_ptFic ,categorie );
	if( ptDebNou->ValDag.Arc.ValRef == _Nil ) {
	  return NULL;
	}
      }
    }

    /* les alternatives positives sont des disjonctions */
    if( !estNeg ) {
      if( ptArcDebCou == ptArcDebCommun ) {
	ptArcDebCommun->Fils = ptDebNou;
      }
      else
	{
	  ptArcDebCou->Frere = ptDebNou;
	}
      ptArcDebCou = ptDebNou;

      if( !ptArcFinCommun ) {
	if( !( ptArcFinCommun = NouveauDag( _Arc ) ) ) {
	  return NULL;
	}
	ptArcFinCommun->ValDag.Arc.ValSymb = _Nil;
	ptArcFinCommun->ValDag.Arc.ValRef = _Nil;
	strcpy( ptArcFinCommun->ValDag.Arc.Ident  , "#" );
      }
      ptDebNou->Fils = ptArcFinCommun;
    }
    else
      /* les alternatives negatives sont des conjonctions, elles sont
	 decrites par une sequence de filtres negatifs suivie d'une
	 categorie vide sur laquelle on avance, si les filtres
	 precedents ont ete passes avec succes */
      {
	if( ptArcDebCou == p_ptArcDeb ) {
	  p_ptArcDeb->Fils = ptDebNou;
	}
	else
	  {
	    ptArcDebCou->Fils = ptDebNou;
	  }
	ptArcDebCou = ptDebNou;
      }
  }

  if( !ReconnaitreCaractere( p_ptFic , (int)']' , p_carLu , _True ) ) {
    return NULL;
  }
    
  if( !estNeg && !ptArcFinCommun ) {
    DemarqueDag( &ptArcDebCommun );
    RendUnDag( &ptArcDebCommun );
  }
  if( estNeg && ptArcDebCou != p_ptArcDeb ) {
    if( !( ptArcFinCommun = NouveauDag( _Arc ) ) ) {
      return NULL;
    }
    ptArcFinCommun->ValDag.Arc.ValSymb = RefCatLexVide;
    ptArcFinCommun->ValDag.Arc.ValRef = _Nil;
    strcpy( ptArcFinCommun->ValDag.Arc.Ident  , "X" );
    ptArcDebCou->Fils = ptArcFinCommun;
  }

  /* repetition */
  if( *p_carLu == '+' || *p_carLu == '*' 
      || *p_carLu == '?' || _Num(*p_carLu) ) {
    if( ptArcFinCommun ) {
      if( !( ptDebNou 
	     = CompileERStarOption( p_carLu , p_ptFic ,
				    p_ptArcDeb , ptArcFinCommun ) ) ) {
	return NULL;
      }
      ptArcFinCommun = ptDebNou;
    }
    else
      {
	*p_carLu = LireCar( p_ptFic );
	SautBlanc( p_ptFic , p_carLu );
      }
  }

  if( ptArcFinCommun ) {
    return ptArcFinCommun;
  }
  else
    {
      return p_ptArcDeb;
    }
}

/*****************************************************************************
      ANALYSE D'UNE EXPRESSION REGULIERE : CAS D'UN CHOIX
*****************************************************************************/
TabDagMem    CompileERChoix( char *p_carLu , FILE *p_ptFic , 
                             TabDagMem p_ptArcDeb )
{
  TabDagMem          ptDebNou;
  TabDagMem          ptArcDebCou;
  TabDagMem          ptArcDebPrec;
  TabDagMem          ptArcFinCommun;

  if( !ReconnaitreCaractere( p_ptFic , (int)'{' , p_carLu , _True ) ) {
    return NULL;
  }

  if( p_ptArcDeb->Frere || p_ptArcDeb->Fils ) {
    TxtErreur( 322 , "CompileERChoix(aldags)" );
    return NULL;
  }

  ptArcFinCommun = NULL;
  ptArcDebCou = NULL;
  ptArcDebPrec = NULL;
  while( _Alpha(*p_carLu)  || *p_carLu == '<' 
	 || *p_carLu == '[' || *p_carLu == '{' ) {
    if( !( ptArcDebCou = NouveauDag( _Arc ) ) ) {
      return NULL;
    }
    ptArcDebCou->ValDag.Arc.ValSymb = _Nil;
    ptArcDebCou->ValDag.Arc.ValRef = _Nil;
    strcpy( ptArcDebCou->ValDag.Arc.Ident  , "#" );

    ptDebNou = ptArcDebCou;

    if( _Alpha(*p_carLu)  || *p_carLu == '<' ) {
      if( !( ptDebNou = CompileERSequence( p_carLu , p_ptFic , 
					   ptArcDebCou ) ) ) {
	return NULL;
      }
    }
    else if( *p_carLu == '{' ) {
      if( !( ptDebNou = CompileERChoix( p_carLu , p_ptFic , 
					ptArcDebCou ) ) ) {
	return NULL;
      }
    }
    else if( *p_carLu == '[' ) {
      if( !( ptDebNou = CompileERAlternative( p_carLu , p_ptFic , 
					      ptArcDebCou ) ) ) {
	return NULL;
      }
    }

    if( ptDebNou != ptArcDebCou ) {
      if( ptArcDebPrec ) {
	ptArcDebPrec->Frere = ptArcDebCou;
      }
      else
	{
	  p_ptArcDeb->Fils = ptArcDebCou;
	}

      if( !ptArcFinCommun ) {
	if( !( ptArcFinCommun = NouveauDag( _Arc ) ) ) {
	  return NULL;
	}
	ptArcFinCommun->ValDag.Arc.ValSymb = _Nil;
	ptArcFinCommun->ValDag.Arc.ValRef = _Nil;
	strcpy( ptArcFinCommun->ValDag.Arc.Ident  , "#" );
      }
      ptDebNou->Fils = ptArcFinCommun;

      ptArcDebPrec = ptArcDebCou;
      ptArcDebCou = NULL;
    }
    else
      {
	DemarqueDag( &ptArcDebCou );
	RendUnDag( &ptArcDebCou );
      }

    if( *p_carLu == '|' ) {
      ReconnaitreCaractere( p_ptFic , (int)'|' , p_carLu , _True );
    }
  }

  if( !ReconnaitreCaractere( p_ptFic , (int)'}' , p_carLu , _True ) ) {
    return NULL;
  }
    
  if( !ptArcDebPrec ) {
    DemarqueDag( &ptArcDebPrec );
    RendUnDag( &ptArcFinCommun );
  }

  /* repetition */
  if( *p_carLu == '+' || *p_carLu == '*' 
      || *p_carLu == '?'  || _Num(*p_carLu) ) {
    if( ptArcFinCommun ) {
      if( !( ptDebNou = CompileERStarOption( p_carLu , p_ptFic ,
					     p_ptArcDeb , ptArcFinCommun ) ) ) {
	return NULL;
      }
      ptArcFinCommun = ptDebNou;
    }
    else
      {
	*p_carLu = LireCar( p_ptFic );
	SautBlanc( p_ptFic , p_carLu );
      }
  }

  if( ptArcFinCommun ) {
    return ptArcFinCommun;
  }
  else
    {
      return p_ptArcDeb;
    }
}

/*****************************************************************************
      ANALYSE D'UNE EXPRESSION REGULIERE : CAS D'UNE SEQUENCE
*****************************************************************************/
TabDagMem    CompileERSequence( char *p_carLu , FILE *p_ptFic ,   
                             TabDagMem p_ptArcDeb )
{
  ChIdent            ident;
  int                categorie;
  char               delim;
  char               estNeg;
  TabDagMem          ptDebCou;
  TabDagMem          ptDebNou;
  TabDagMem          ptDebNouAux;
  TabDagMem         *hdlFin;

  if( p_ptArcDeb->Frere || p_ptArcDeb->Fils ) {
    TxtErreur( 323 , "CompileERSequence(aldags)" );
    return NULL;
  }

  delim = ( *p_carLu == '<' );
  if( delim 
      && !ReconnaitreCaractere( p_ptFic , (int)'<' , p_carLu , _True ) ) {
    return NULL;
  }

  ptDebCou = p_ptArcDeb;
  ptDebNou = ptDebCou;
  while( _Alpha(*p_carLu) || *p_carLu == '<' || *p_carLu == '^' 
	 || *p_carLu == '{' || *p_carLu == '[' ) {
    if( *p_carLu == '<' ) {
      if( !( ptDebNou = CompileERSequence( p_carLu , p_ptFic , 
					   ptDebCou ) ) ) {
	return NULL;
      }
    }
    else if( *p_carLu == '{' ) {
      if( !( ptDebNou = CompileERChoix( p_carLu , p_ptFic ,  
					ptDebCou ) ) ) {
	return NULL;
      }
    }
    else if( *p_carLu == '[' ) {
      if( !( ptDebNou = CompileERAlternative( p_carLu , p_ptFic ,  
					      ptDebCou ) ) ) {
	return NULL;
      }
    }
    else
      {
	estNeg = _False;
	if( *p_carLu == '^' ) {
	  estNeg = _True;
	  ReconnaitreCaractere( p_ptFic , (int)'^' , p_carLu , _True );
	}
	LireIdentificateur( p_ptFic , p_carLu , ident ,  
			    _MaxLetIdent , _True );
	if( !IdentATraitsCatLex( ident , _CatLex , &categorie ) ) {
	  return NULL;
	}
	ptDebNou = NouveauDag( _Arc );
	if( !ptDebNou ) {
	  return NULL;
	}
	ptDebNou->TypeDag = _Arc;
	ptDebNou->ValDag.Arc.ValSymb = (long)categorie;
	ptDebNou->ValDag.Arc.ValRef = _Nil;
	ptDebNou->ValDag.Arc.ArcERNegation = estNeg;
	strcpy( ptDebNou->ValDag.Arc.Ident , ident );

	/* cas ou la reference est donnee */
	if( *p_carLu == '/' ) {
	  ReconnaitreCaractere( p_ptFic , (int)'/' , p_carLu , _True );
	  if( _Num( *p_carLu ) ) {
	    ptDebNou->ValDag.Arc.ValRef 
	      = LireEntier( p_ptFic , p_carLu );
	  }
	  else if( *p_carLu == '\'' ) {
	    ptDebNou->ValDag.Arc.ValRef 
	      = LireLemmeER( p_carLu , p_ptFic , categorie );
	    if( ptDebNou->ValDag.Arc.ValRef == _Nil ) {
	      return NULL;
	    }
	  }
	}

	if( !( hdlFin = AjouteArcOuListeFils( &(ptDebCou->Fils) , 
					      &(ptDebNou->ValDag) , ptDebNou , 
					      _True , _Arc ) ) ) {
	  return NULL;
	}
	if( hdlFin != &(ptDebNou->Fils) ) {
	  TxtErreur( 324 , "CompileERSequence(aldags)" );
	  return NULL;
	}

	if( estNeg ) {
	  if( !( ptDebNouAux = NouveauDag( _Arc ) ) ) {
	    return NULL;
	  }
	  ptDebNouAux->ValDag.Arc.ValSymb = RefCatLexVide;
	  ptDebNouAux->ValDag.Arc.ValRef = _Nil;
	  strcpy( ptDebNouAux->ValDag.Arc.Ident  , "X" );
	  if( !( hdlFin = AjouteArcOuListeFils( &(ptDebNou->Fils) , 
						&(ptDebNou->ValDag) , 
						ptDebNouAux , 
						_True , _Arc ) ) ) {
	    return NULL;
	  }
	  ptDebNou = ptDebNouAux;
	}
	if( hdlFin != &(ptDebNou->Fils) ) {
	  TxtErreur( 325 , "CompileERSequence(aldags)" );
	  return NULL;
	}

	if( *p_carLu == '+' || *p_carLu == '*'
	    || *p_carLu == '?' || _Num(*p_carLu)  ) {
	  if( !( ptDebNou = CompileERStarOption( p_carLu , p_ptFic ,
						 ptDebCou , ptDebNou ) ) ) {
	    return NULL;
	  }
	}
      }

    ptDebCou = ptDebNou;
  }

  if( delim && !ReconnaitreCaractere( p_ptFic , (int)'>' , p_carLu , _True ) ) {
    return NULL;
  }
    
  /* repetition */
  if( *p_carLu == '+' || *p_carLu == '*' 
      || *p_carLu == '?' || _Num(*p_carLu) ) {
    if( ptDebNou != p_ptArcDeb ) {
      if( !( ptDebNou = CompileERStarOption( p_carLu , p_ptFic ,
					     p_ptArcDeb , ptDebNou ) ) ) {
	return NULL;
      }
    }
    else
      {
	*p_carLu = LireCar( p_ptFic );
	SautBlanc( p_ptFic , p_carLu );
      }
  }

  return ptDebNou;
}

/*****************************************************************************
      ANALYSE D'UNE EXPRESSION REGULIERE : LANCEMENT
*****************************************************************************/
TabDagMem    *CompileER( char *p_carLu , FILE *p_ptFic , 
			 TabDagMem p_filsArc )
{
  TabDagMem                   ptArcFin = NULL;
  TabDagMem                  *hdlArcFin;
  struct ValAtomeMem          atome;

  if( *p_carLu == '<' ) {
    if( !( ptArcFin = CompileERSequence( p_carLu , p_ptFic , p_filsArc ) ) ) {
      return NULL;
    }
  }
  else if( *p_carLu == '{' ) {
    if( !( ptArcFin = CompileERChoix( p_carLu , p_ptFic , p_filsArc ) ) ) {
      return NULL;
    }
  }

  hdlArcFin = &(ptArcFin->Fils);
  strcpy( atome.ValAtome.Chaine , "$" );
  if( !( ptArcFin = AjouteAtomeFils( hdlArcFin , _AtomeChaine , 
				     atome , _True ) ) ) {
    return NULL;
  }

  return &(ptArcFin->Fils);
}

/*****************************************************************************
    LECTURE D'UN CHEMIN DE DAG. Retourne le pointeur vers la derniere
    etiquette du chemin: sert a typer le chemin. Un chemin se
    terminant par le trait categorie ne peut etre egalise qu'avec un
    atome de type categorie.
*****************************************************************************/
TabDagMem    *LireUnChemin( char *p_carLu , FILE *p_ptFic ,
                            TabDagMem *p_hdlArc , int p_regle ,
			    int *p_typeDagDernierArcChemin )
{
  ChIdent            ident;
  int                trait;
  TabDagMem         *hdlCou;
  TabDagMem         *hdlFin;
  TabDagMem         *hdlPrecFin;
  struct DagMem      unDag;

  if( !ReconnaitreCaractere( p_ptFic , (int)'<' , p_carLu , _True ) ) {
    return NULL;
  }
  if( !(_Alpha(*p_carLu)) ) {
    TxtChaineErreur( 172 , ChErreurCompil );
    return NULL;
  }

  hdlCou = p_hdlArc;
  hdlFin = NULL;
  hdlPrecFin = NULL;
  while( _Alpha(*p_carLu) ) {
    LireIdentificateur( p_ptFic , p_carLu , ident , _MaxLetIdent , _True );
    if( p_regle && !hdlFin ) {
      /* identificateur de terme en tete de regle, le
	 rattachement sera fait dans TeteDeCheminADagDeTerme */
      trait = (int)_Nil;
    }
    else
      {
	if( !IdentATraitsCatLex( ident , _Trait , &trait ) ) {
	  return NULL;
	}
      }

    MiseAZeroDag( &unDag , _Arc );
    strcpy( unDag.ValDag.Arc.Ident , ident );
    unDag.ValDag.Arc.ValSymb = (long)trait;
    if( !( hdlFin = AjouteArcOuListeFils( hdlCou , &(unDag.ValDag) , 
					  NULL , _True , _Arc ) ) ) {
      return NULL;
    }
    *p_typeDagDernierArcChemin = _Arc;
    hdlPrecFin = hdlCou;
    hdlCou = hdlFin;
  }

  if( !ReconnaitreCaractere( p_ptFic , (int)'>' , p_carLu , _True ) ) {
    return NULL;
  }

  /* retourne le pointeur vers le dernier arc du chemin donc ne peut
     pas lire un chemin vide */
  return hdlPrecFin;
}

/*****************************************************************************
    LECTURE D'UNE EGALITE OU INEGALITE CHEMIN / ATOME (OU DISJ ATOMES)
*****************************************************************************/

char     LireEquationCheminAtome( char *p_carLu  , FILE *p_ptFic , 
				  TabDagMem *p_filsOperLog , int p_oper ,
				  TabDagMem *p_hdlDagTermeTeteChemin , 
				  TabDagMem p_filsArcChemin ,
				  TabDagMem *p_hdlDernierArcChemin ,
				  int p_typeDagDernierArcChemin )
{
  TabDagMem                    ptAtome;
  char                         estEtiquette;
  char                         traitCatLex;

  /* atom = list */
  if( *p_carLu == '(' && ( p_oper == (int)'=' ) ) {
    if( !LireUneListe( p_carLu , p_ptFic , &ptAtome ) ) {
      return _False;
    }

    /* first atom of a disjunction */
    if( *p_carLu == '|' ) {
      ptAtome->ValDag.Atome.AtomeConjDisj = _DisjonctionPos;
    }

    /* equality path/atom(list) */
    if( !AjoutOperCheminAtome( p_filsOperLog , p_oper , 
			       p_hdlDagTermeTeteChemin ,
			       p_filsArcChemin , ptAtome , 
			       p_typeDagDernierArcChemin ) ) {
      return _False;
    }
	  
    /* additional equalities path/atom(list) in case of a disjunction */
    while( *p_carLu == '|' ) {
      ReconnaitreCaractere( p_ptFic , '|' , p_carLu , _True );
      if( !LireUneListe( p_carLu , p_ptFic , &ptAtome ) ) {
	return _False;
      }
      ptAtome->ValDag.Atome.AtomeConjDisj = _DisjonctionPos;
      if( !AjoutOperCheminAtome( p_filsOperLog , p_oper , 
				 p_hdlDagTermeTeteChemin ,
				 p_filsArcChemin , ptAtome , 
				 p_typeDagDernierArcChemin ) ) {
	return _False;
      }
    }
  }
  /* atom = list */
  /* atom = atomic value */
  else  if( *p_carLu != '(' ) {
    /* cas particulier du chemin d'etiquette */
    if( (*p_hdlDernierArcChemin)->ValDag.Arc.ValSymb
	== CheminEtiquette->ValDag.Arc.ValSymb ) {
      estEtiquette = _True; 
      traitCatLex = _Trait; /* ss importance ici */
    }
    /* cas particulier du chemin d'etiquette */
    /* chemin != etiquette */
    else {
      estEtiquette = _False;

      /* cas particulier du chemin de categorie */
      if( (*p_hdlDernierArcChemin)->ValDag.Arc.ValSymb
	  == CheminCategorie->ValDag.Arc.ValSymb ) {
	traitCatLex = _CatLex;
      }
      else {
	traitCatLex = _Trait;
      }
    }
    /* chemin != etiquette */

    if( !LireUnAtome( p_carLu , p_ptFic , &ptAtome ,
		      estEtiquette , traitCatLex ) ) {
      return _False;
    }

    /* inequality */
    if( p_oper == (int)'!' ) {
      ptAtome->ValDag.Atome.AtomeConjDisj = _ConjonctionNeg;
    }

    /* first atom of a disjunction */
    if( *p_carLu == '|' ) {
      if( p_oper == (int)'!' ) {
	TxtErreur( 291 , "LireEquationCheminAtome(aldags)" );
	return _False;
      } 
      ptAtome->ValDag.Atome.AtomeConjDisj = _DisjonctionPos;
    }

    /* equality path/atom */
    if( !AjoutOperCheminAtome( p_filsOperLog , p_oper , 
			       p_hdlDagTermeTeteChemin ,
			       p_filsArcChemin , ptAtome , 
			       p_typeDagDernierArcChemin ) ) {
      return _False;
    }

    /* additional equalities path/atom in case of a disjunction */
    while( *p_carLu == '|' ) {
      ReconnaitreCaractere( p_ptFic , '|' , p_carLu , _True );
      if( !LireUnAtome( p_carLu , p_ptFic , &ptAtome ,
			estEtiquette , traitCatLex ) ) {
	return _False;
      }
      ptAtome->ValDag.Atome.AtomeConjDisj = _DisjonctionPos;
      if( !AjoutOperCheminAtome( p_filsOperLog , p_oper , p_hdlDagTermeTeteChemin ,
				 p_filsArcChemin , ptAtome , 
				 p_typeDagDernierArcChemin ) ) {
	return _False;
      }
    }
  }
  /* atom = atomic value */
  else {
    TxtErreur( 290 , "LireEquationCheminAtome(aldags)" );
    return _False;
  }    
  
  return _True;
}

/*****************************************************************************
    LECTURE DES CONTRAINTES SUR LES DAGS
*****************************************************************************/
char     LireContraintesTerme( char *p_carLu , FILE *p_ptFic ,
                               TabDagMem *p_filsOperLog ,
                               TabDagMem *p_hdlArc )
{
  TabDagMem            filsArcChemin1;
  TabDagMem           *hdlChemin1Aux; /* en cas de partie
					 commune pour le chemin 1 */
  TabDagMem            filsArcChemin2;
  TabDagMem           *hdlDernierArcChemin1;
  char                 estAvecPartieInitialeCommune;
  char                 oper;
  int                  typeDagDernierArcChemin;

  filsArcChemin1 = NULL;
  filsArcChemin2 = NULL;
  while( *p_carLu == '<' || *p_carLu == '-' ) {
    if( filsArcChemin2 ) {
      TxtErreur( 241 , "LireContraintesTerme(aldags)" );
      return _False;
    }
    hdlChemin1Aux = &filsArcChemin1;
      
    /* lecture d'une partie initiale commune eventuelle */
    if( *p_carLu == '-' ) {
      estAvecPartieInitialeCommune = _True;
      while( *p_carLu == '-' ) {
	if( !(*hdlChemin1Aux) ) {
	  TxtErreur( 242 , "LireContraintesTerme(aldags)" );
	  return _False;
	}
	hdlChemin1Aux = &((*hdlChemin1Aux)->Fils);
	ReconnaitreCaractere( p_ptFic , (int)'-' , p_carLu , _True );
      }
    }
    /* lecture d'une partie initiale commune eventuelle */
    else {
      estAvecPartieInitialeCommune = _False;
    }
      
    /* lecture de la nouvelle partie du chemin */
    if( *p_carLu == '<' ) {
      if( (*hdlChemin1Aux) ) {
	DemarqueEtRendDag( hdlChemin1Aux );
      }
	
      if( !(hdlDernierArcChemin1 
	    = LireUnChemin( p_carLu , p_ptFic ,
			    hdlChemin1Aux , _False , 
			    &typeDagDernierArcChemin ) ) ) {
	return _False;
      }  
    }
    /* lecture de la nouvelle partie du chemin */

    if( !estAvecPartieInitialeCommune ) {
      filsArcChemin1 = *hdlChemin1Aux;
    }

    /* unknown operator */
    if( *p_carLu != '=' && *p_carLu != '!' ) {
      TxtErreur( 143 , "LireContraintesTerme(aldags)" );
      return _False;
    }

    oper = *p_carLu;
    ReconnaitreCaractere( p_ptFic , (int)oper , p_carLu , _True );

    /* path and equality */
    if( *p_carLu == '<' && (oper == '=' ) ) {
      if( !LireUnChemin( p_carLu , p_ptFic ,
			 &filsArcChemin2 , _False ,
			 &typeDagDernierArcChemin ) ) {
	DemarqueEtRendDag( &filsArcChemin1 );
	return _False;
      }
      if( !AjoutEgaliteCheminChemin( p_filsOperLog , (int)oper , p_hdlArc , 
				     p_hdlArc , filsArcChemin1 , 
				     filsArcChemin2 ) ) {
	DemarqueEtRendDag( &filsArcChemin1 );
	DemarqueEtRendDag( &filsArcChemin2 );
	return _False;
      }
      DemarqueEtRendDag( &filsArcChemin2 );
    }
    /* path and equality */
    /* atom and ( equality or inequality ) */
    else {
      if( !LireEquationCheminAtome( p_carLu  , p_ptFic , 
				    p_filsOperLog , (int)oper ,
				    p_hdlArc , filsArcChemin1 ,
				    hdlDernierArcChemin1 , 
				    typeDagDernierArcChemin ) ) {
	DemarqueEtRendDag( &filsArcChemin1 );
	return _False;
      }
    }
    /* atom and ( equality or inequality ) */
  }

  DemarqueEtRendDag( &filsArcChemin1 );
  return _True;
}

/*****************************************************************************
    LECTURE DES CONTRAINTES SUR LES DAGS DANS UNE REGLE ( AVEC
    NECESSITE DE RETROUVER LE TERME D'ANCRAGE DES DAGS CORRESPONDANTS
    EN FCT DES IDENTIFICATEURS )
*****************************************************************************/
char     LireContraintesRegle( char *p_carLu , FILE *p_ptFic ,
                               TabDagMem *p_filsOperLog )
{
  TabDagMem            filsArcChemin1;
  TabDagMem            filsArcChemin2;
  TabDagMem           *hdlDagTermeTeteChemin1;
  TabDagMem           *hdlDagTermeTeteChemin2;
  TabDagMem           *hdlDernierArcChemin1;
  char                 oper;
  int                  typeDagDernierArcChemin;

  while( *p_carLu == '<' ) {
    filsArcChemin1 = NULL;
    filsArcChemin2 = NULL;
    hdlDagTermeTeteChemin1 = NULL;
    hdlDagTermeTeteChemin2 = NULL;

    if( !(hdlDernierArcChemin1 
	  = LireUnChemin( p_carLu , p_ptFic ,
			  &filsArcChemin1 , _True ,
			  &typeDagDernierArcChemin ) ) ) {
      return _False;
    }
    if( *p_carLu == '=' || *p_carLu == '!' ) {
      oper = *p_carLu;
      ReconnaitreCaractere( p_ptFic , (int)oper , p_carLu , _True );
      if( *p_carLu == '<' && ( oper == '=' ) ) {
	if( !LireUnChemin( p_carLu , p_ptFic , &filsArcChemin2 , 
			   _True , &typeDagDernierArcChemin ) ) {
	  return _False;
	}
	if( !( hdlDagTermeTeteChemin1 
	       = TeteDeCheminADagDeTerme( &filsArcChemin1 ) )
	    || !( hdlDagTermeTeteChemin2 
		  = TeteDeCheminADagDeTerme( &filsArcChemin2 ) )
	    || !AjoutEgaliteCheminChemin( p_filsOperLog , (int)oper ,
					  hdlDagTermeTeteChemin1 , 
					  hdlDagTermeTeteChemin2 , 
					  filsArcChemin1 , 
					  filsArcChemin2 ) ) {
	  DemarqueEtRendDag( &filsArcChemin1 );
	  DemarqueEtRendDag( &filsArcChemin2 );
	  return _False;
	}
	DemarqueEtRendDag( &filsArcChemin1 );
	DemarqueEtRendDag( &filsArcChemin2 );
      }
      else
	{
	  if( !( hdlDagTermeTeteChemin1 
		 = TeteDeCheminADagDeTerme( &filsArcChemin1 ) ) ) {
	    DemarqueEtRendDag( &filsArcChemin1 );
	    return _False;
	  }
		 
	  if( !LireEquationCheminAtome( p_carLu  , p_ptFic , 
					p_filsOperLog , (int)oper , 
					hdlDagTermeTeteChemin1 , 
					filsArcChemin1 ,
					hdlDernierArcChemin1 ,
					typeDagDernierArcChemin ) ) {
	    DemarqueEtRendDag( &filsArcChemin1 );
	    return _False;
	  }
	  DemarqueEtRendDag( &filsArcChemin1 );
	}
    }
    else
      {
	TxtChaineErreur( 143 , ChErreurCompil );
	return _False;
      }
  }

  return _True;
}

/*****************************************************************************
    UNIFICATION ENTRE 2 DAGS
*****************************************************************************/
char     UnificationArcArc( TabDagMem *p_hdlPereArc1 ,
			    TabDagMem *p_hdlPereArc2 )
{
  TabDagMem                           filsCou2;
  TabDagMem                           filsCou1;
  TabDagMem                           filsSuivant2;
  char                                resul;

  resul =_True;

  filsCou2 = *p_hdlPereArc2;
  /* 2 listes d'arcs  */
  if( filsCou2->TypeDag == _Arc ) {
    /* parcours de tous les fils du dag 2 */
    while( filsCou2 ) {
      /* on note le fils suivant du dag 2 pour le passage en fin de
	 boucle */
      filsSuivant2 = filsCou2->Frere;

      /* cas d'un arc commun aux deux dags */
      if( (filsCou1 = ExisteArcOuListeFils( *p_hdlPereArc1 , 
					    filsCou2 , _Arc ) ) ) {
	if( Unifier( &(filsCou1->Fils) , &(filsCou2->Fils) ) ) {
	}
	else
	  {
	    return _False;
	  }

	/* apres l'unification des arcs, on desalloue arc2 */
	/* arc 2 et arc 1 pointent vers le meme fils (en
	   principe!) */
	if( filsCou1->Fils != filsCou2->Fils ) {
	  TxtErreur( 93 , 
		     "UnificationArcArc 1(aldags)" );
	  return _False;
	}
              
	if( filsSuivant2 ) {
	  RestaurePointeurVersDagJete( *p_hdlPereArc2 , 
				       filsSuivant2 );
	  if( *p_hdlPereArc2 != filsSuivant2 ) {
	    TxtErreur( 1 , "revoir restauration pointeurs 3" );
	  }
	}
	else
	  {
	    RestaurePointeurVersDagJete( *p_hdlPereArc2 , 
					 *p_hdlPereArc1 );
	    if( *p_hdlPereArc2 != *p_hdlPereArc1 ) {
	      TxtErreur( 1 , "revoir restauration pointeurs 4" );
	    }
	  }

	DemarqueDag( &filsCou2 );
	RendUnDag( &filsCou2 );
      }
      /* cas d'un arc commun aux deux dags */
      else
	/* arc present uniquement dans le dag 2 */
	{
	  if( filsSuivant2 ) {
	    RestaurePointeurVersDagJete( *p_hdlPereArc2 , filsSuivant2 );
	    if( *p_hdlPereArc2 != filsSuivant2 ) {
	      TxtErreur( 1 , "revoir restauration pointeurs 5" );
	    }
	  }
	  else
	    {
	      RestaurePointeurVersDagJete( *p_hdlPereArc2 , 
					   *p_hdlPereArc1 );
	      if( *p_hdlPereArc2 != *p_hdlPereArc1 ) {
		TxtErreur( 1 , "revoir restauration pointeurs 6" );
	      }
	    }

	  /* on ajoute l'arc de 2 qui n'existe pas dans 1 en queue de 1 */
	  filsCou1 = *p_hdlPereArc1;
	  while( filsCou1->Frere ) {
	    filsCou1 = filsCou1->Frere;
	  }
	  filsCou1->Frere = filsCou2;
	  filsCou2->Frere = NULL;
	}
      /* arc present uniquement dans le dag 2 */

      /* passage au fils suivant du dag 2 */
      filsCou2 = filsSuivant2;
    }
    /* parcours de tous les fils du dag 2 */
  }
  /* 2 listes d'arcs  */
  else
    {
      TxtErreur( 93 , "UnificationArcArc 3(aldags)" );
      return _False;
    }

  if( *p_hdlPereArc2 != *p_hdlPereArc1 ) {
    TxtErreur( 1 , "revoir algo unification" );
    return _False;
  }

  *p_hdlPereArc2 = *p_hdlPereArc1;

  return resul;
}

/*****************************************************************************
    RETURN A BOOLEAN VALUE INDICATING WHETHER TWO ATOMS ARE EQUUAL
    IN CASE OF A LIST AND A SELECTOR, DICARDS THE UNUSEFUL ELEMENTS
    OF THE LIST
*****************************************************************************/
int      AreEqual( TabDagMem p_filsCou1 , TabDagMem p_filsCou2 )
{
  /* 2 atoms of different types */
  if( p_filsCou2->TypeDag != p_filsCou1->TypeDag  ) {
    return _False;
  }

  /* 2 genuine atoms */
  else if( p_filsCou1->TypeDag == _AtomeEntier 
	   || p_filsCou1->TypeDag == _AtomeRef
	   || p_filsCou1->TypeDag == _AtomeChaine ) {
    return ( ( p_filsCou1->TypeDag == _AtomeEntier 
	       && p_filsCou1->ValDag.Atome.ValAtome.Entier 
	       == p_filsCou2->ValDag.Atome.ValAtome.Entier )
	     || ( p_filsCou1->TypeDag == _AtomeRef
		  && p_filsCou1->ValDag.Atome.ValAtome.Ref
		  == p_filsCou2->ValDag.Atome.ValAtome.Ref )
	     || ( p_filsCou1->TypeDag == _AtomeChaine
		  && strcmp( p_filsCou1->ValDag.Atome.ValAtome.Chaine , 
			     p_filsCou2->ValDag.Atome.ValAtome.Chaine ) 
		  == 0 ) );
  }

  /* 2 lists */
  else if( p_filsCou2->TypeDag == p_filsCou1->TypeDag 
	   && p_filsCou1->TypeDag == _AtomeListe ) {
    /* unification of the 2 lists */
    {
      TabDagMem         fils1 = p_filsCou1->Fils;
      TabDagMem         fils2 = p_filsCou1->Fils;

      while( fils1 && fils2 ) {
	if( !AreEqual( fils1 , fils2 ) ) {
	  return _False;
	}
	fils1 = fils1->Frere;
	fils2 = fils2->Frere;
      }
      return !fils1 && !fils2;
    }
  }
  /* 2 lists */

  /* combinaison illegale */
  else {
    fprintf( FicLog , "Dag #1 [%d] #2 [%d]\n" , 
	     p_filsCou1->TypeDag , p_filsCou2->TypeDag );
    printf( "Dag #1 [%d] #2 [%d]\n" , 
	    p_filsCou1->TypeDag , p_filsCou2->TypeDag );
    TxtErreur( 300 , "AreEqual(aldags)" );
    return _False;
  }
}

/*****************************************************************************
    UNIFICATION ENTRE DEUX VALEURS
*****************************************************************************/
char     UnificationAtomeAtome( TabDagMem *p_hdlPereAtome1 , 
                                TabDagMem *p_hdlPereAtome2 )
{
  TabDagMem            filsCou1;
  TabDagMem            filsCou2;
  TabDagMem            filsIni1;
  TabDagMem            filsIni2;
  TabDagMem            filsSuivant1;
  TabDagMem            filsSuivant2;
  char                 resul;
  struct DagMem        dagTrace1;
  struct DagMem        dagTrace2;
  char                 disjConj1;
  char                 disjConj2;

  int                  estAtome1;
  int                  estAtome2;
  int                  estListe1;
  int                  estListe2;
    
  resul = _True;

  /* current nodes */
  filsCou1 = *p_hdlPereAtome1;
  filsCou2 = *p_hdlPereAtome2;

  /* trace memory */
  if( Trace ) {
    dagTrace1 = *filsCou1;
    dagTrace2 = *filsCou2;
  }

  /* sister node of the second node (in case of loops for lists) */
  filsSuivant2 = filsCou2->Frere;

  /* boolean values characterizing the atoms */
  estAtome1 = ( filsCou1->TypeDag == _AtomeRef
                || filsCou1->TypeDag == _AtomeChaine
                || filsCou1->TypeDag == _AtomeEntier );
  estAtome2 = ( filsCou2->TypeDag == _AtomeRef
                || filsCou2->TypeDag == _AtomeChaine
                || filsCou2->TypeDag == _AtomeEntier );
  estListe1 = filsCou1->TypeDag == _AtomeListe;
  estListe2 = filsCou2->TypeDag == _AtomeListe;

  disjConj1 = filsCou1->ValDag.Atome.AtomeConjDisj;
  disjConj2 = filsCou2->ValDag.Atome.AtomeConjDisj;

  /* verifies the atomicity of the dags */
  if( !(estAtome1 || estListe1) 
      || !(estAtome2 || estListe2) ) {
    char ch1[ 64 ];
      
    fprintf( FicLog , "Dag #1 [%d] #2 [%d]\n" , 
	     filsCou1->TypeDag , filsCou2->TypeDag );
    printf( "Dag #1 [%d] #2 [%d]\n" , 
	    filsCou1->TypeDag , filsCou2->TypeDag );
    filsCou1 = &dagTrace1;
    filsCou2 = &dagTrace2;
    ChaineDagMem( &filsCou1 , ch1 );
    fprintf( FicLog , "\n%s\n" , ch1 );
    printf( "\n%s\n" , ch1 );
    ChaineDagMem( &filsCou2 , ch1 );
    fprintf( FicLog , "%s\n" , ch1 );
    printf( "%s\n" , ch1 );
    TxtErreur( 293 , "UnificationAtomeAtome(aldags)" );
    return _False;
  }
      
  /* 2 atomes negatifs ou deux atomes de disjonction */
  /* on les met bout a bout */
  if( ( disjConj1 == _ConjonctionNeg && disjConj2 == _ConjonctionNeg )
      || ( disjConj1 == _DisjonctionPos && disjConj2 == _DisjonctionPos ) ) {
    /* verifies the atomicity of the elements in the list */
    /* and their homogeneity (only pos disjunctions or only negative
       conjunctions) */
    while( filsCou1->Frere ) {
      filsCou1 = filsCou1->Frere;
      if( ( filsCou1->TypeDag != _AtomeRef
	    && filsCou1->TypeDag != _AtomeChaine
	    && filsCou1->TypeDag != _AtomeEntier
	    && filsCou1->TypeDag != _AtomeListe )
	  || filsCou1->ValDag.Atome.AtomeConjDisj != disjConj1 ) {
	TxtErreur( 294 , "UnificationAtomeAtome(aldags)" );
	return _False;
      }
    }
    while( filsCou2 ) {
      filsCou2 = filsCou2->Frere;
      if( filsCou2
	  && ( (  filsCou2->TypeDag != _AtomeRef
		  && filsCou2->TypeDag != _AtomeChaine
		  && filsCou2->TypeDag != _AtomeEntier
		  && filsCou2->TypeDag != _AtomeListe )
	       || filsCou2->ValDag.Atome.AtomeConjDisj != disjConj1 ) ) {
	TxtErreur( 295 , "UnificationAtomeAtome(aldags)" );
	return _False;
      }
    }

    /* chains the two pos disj or neg conj */
    /* memorizes the old pointer before merging */
    filsSuivant2 = *p_hdlPereAtome2;
    RestaurePointeurVersDagJete( *p_hdlPereAtome2 , 
				 *p_hdlPereAtome1 );
    /*       if( *p_hdlPereAtome2 != *p_hdlPereAtome1 ) {  */
    /* 	TxtErreur( 298 , "UnificationAtomeAtome(aldags)" );  */
    /* 	return _False;  */
    /*       } */
    filsCou1->Frere = filsSuivant2;

    /* on a fini, il ne reste qu'une liste d'atomes
       neggatifs */
    filsCou1 = NULL;
    filsCou2 = NULL;
  }
  /* 2 atomes negatifs ou deux atomes de disjonction */

  /* 2 atomes positifs */
  /* on les compare et s'ils sont egaux, on en elimine 1 
     et on passe au suivant (s'il s'agit d'une liste) */
  else if( disjConj1 == _False && disjConj2 == _False ) {

    /* 2 listes d'atomes  parcours de tous les fils */
    while( resul && filsCou1 && filsCou2 ) {

      /* on note le fils suivant du dag 2 pour le passage en fin de
	 boucle */
      filsSuivant2 = filsCou2->Frere;

      if( !AreEqual( filsCou1 , filsCou2 ) ) {
	resul = _False;
      }

      /* si on est sur une liste, on continue */
      if( filsSuivant2 ) {
	RestaurePointeurVersDagJete( *p_hdlPereAtome2 , 
				     filsSuivant2 );
	/*	  if( *p_hdlPereAtome2 != *p_hdlPereAtome1 ) { */
	if( *p_hdlPereAtome2 != filsSuivant2 ) { 
	  TxtErreur( 301 , "UnificationAtomeAtome(aldags)" ); 
	  return _False; 
	} 
      }
      else { 
	RestaurePointeurVersDagJete( *p_hdlPereAtome2 ,  
				     *p_hdlPereAtome1 ); 
	if( *p_hdlPereAtome2 != *p_hdlPereAtome1 ) { 
	  TxtErreur( 297 , "UnificationAtomeAtome(aldags)" ); 
	  return _False; 
	} 
      } 
	
      DemarqueEtRendDag( &filsCou2 );

      /* on est sur des atomes, on passe aux suivants s'ils
	 existent (dans le cas d'une liste) */
      filsCou1 = filsCou1->Frere;
      filsCou2 = filsSuivant2;
      /*	filsSuivant1 = filsCou1->Frere;*/
    }
  }
  /* 2 atomes positifs */

  /* 1 atome positif et une disjonction */
  /* on recherche les valeurs de  l'atome positif compatibles avec
     les valeurs des atomes negatifs et on retourne toutes
     les valeurs de la disjonction qui conviennent */
  else if( ( disjConj1 == _DisjonctionPos && disjConj2 == _False ) || 
	   ( disjConj2 == _DisjonctionPos && disjConj1 == _False ) ) {

    /* disjonction = positif */
    if( disjConj1 == _DisjonctionPos ) {
      resul = _False;
      /* scans the first list */
      while( filsCou1 ) {
	filsSuivant1 = filsCou1->Frere;
	if( ( filsCou1->TypeDag != _AtomeRef
	      && filsCou1->TypeDag != _AtomeChaine
	      && filsCou1->TypeDag != _AtomeEntier
	      && filsCou1->TypeDag != _AtomeListe )
	    || filsCou1->ValDag.Atome.AtomeConjDisj != _DisjonctionPos ) {
	  printf( "Expected list of disjunctions %d\n" , filsCou1->TypeDag );
	  TxtErreur( 299 , "UnificationAtomeAtome(aldags)" );
	  return _False;
	}

	/* cas d'egalite entre une valeur de la disjonction
	   et une valeur positive */
	if( AreEqual( filsCou1 , filsCou2 ) ) {
	  /* do nothing: inspect the next atom in the list */
	  resul = _True;
	}
	  
	filsCou1 = filsSuivant1;
      }
      /* scans the first list */

      filsCou2 = *p_hdlPereAtome2;

      /* the remaining list replaces the positive atom */
      RestaurePointeurVersDagJete( *p_hdlPereAtome2 , 
				   *p_hdlPereAtome1 );
      if( *p_hdlPereAtome2 != *p_hdlPereAtome1 ) { 
	TxtErreur( 297 , "UnificationAtomeAtome(aldags)" ); 
	return _False; 
      } 

      DemarqueEtRendDag( &filsCou2 );

      if( !resul ) {
	return _False;
      }
    }
    /* disjonction = positif */

    /* positif = disjonction */
    else if( disjConj2 == _DisjonctionPos ) {

      resul = _False;
      /* scans the second list */
      while( filsCou2 ) {
	filsSuivant2 = filsCou2->Frere;
	if( ( filsCou2->TypeDag != _AtomeRef
	      && filsCou2->TypeDag != _AtomeChaine
	      && filsCou2->TypeDag != _AtomeEntier
	      && filsCou2->TypeDag != _AtomeListe )
	    || filsCou2->ValDag.Atome.AtomeConjDisj != _DisjonctionPos ) {
	  printf( "Expected list of disjunctions %d\n" , filsCou2->TypeDag );
	  TxtErreur( 299 , "UnificationAtomeAtome(aldags)" );
	  return _False;
	}

	/* cas d'egalite entre une valeur de la disjonction
	   et une valeur positive */
	if( AreEqual( filsCou1 , filsCou2 ) ) {
	  /* do nothing: inspect the next atom in the list */
	  resul = _True;
	}
	  
	filsCou2 = filsSuivant2;
      }
      /* scans the second list */

      filsCou1 = *p_hdlPereAtome1;

      /* the remaining list replaces the positive atom */
      RestaurePointeurVersDagJete( *p_hdlPereAtome1 , 
				   *p_hdlPereAtome2 );
      if( *p_hdlPereAtome2 != *p_hdlPereAtome1 ) { 
	TxtErreur( 297 , "UnificationAtomeAtome(aldags)" ); 
	return _False; 
      } 

      DemarqueEtRendDag( &filsCou1 );
    }
    /* disjonction = positif */
  }
  /* 1 atome positif et une disjonction */

  /* 1 atome positif et un atome negatif */
  /* on verifie que l'atome positif est different de toutes
     les valeurs des atomes negatifs et on ne garde que
     l'atome positif en cas de succes */
  else if( ( disjConj1 == _ConjonctionNeg && disjConj2 == _False ) || 
	   ( disjConj2 == _ConjonctionNeg && disjConj1 == _False ) ) {

    /* negatif = positif */
    if( disjConj1 == _ConjonctionNeg ) {

      /* scans the first list */
      filsIni1 = filsCou1;
      while( filsCou1 ) {
	filsSuivant1 = filsCou1->Frere;
	if( ( filsCou1->TypeDag != _AtomeRef
	      && filsCou1->TypeDag != _AtomeChaine
	      && filsCou1->TypeDag != _AtomeEntier
	      && filsCou1->TypeDag != _AtomeListe )
	    || filsCou1->ValDag.Atome.AtomeConjDisj != _ConjonctionNeg ) {
	  fprintf( FicLog , "Expected list of negations\n" );
	  TxtErreur( 299 , "UnificationAtomeAtome(aldags)" );
	  return _False;
	}

	/* cas d'egalite entre une valeur interdite
	   et une valeur positive */
	if( AreEqual( filsCou1 , filsCou2 ) ) {
	  resul = _False;
	  /* break; */
	}
	filsCou1 = filsSuivant1;
      }
      /* scans the first list */
      /* suppresses first list in case of success */
      if( resul ) {
	  RestaurePointeurVersDagJete( *p_hdlPereAtome1 , 
				       *p_hdlPereAtome2 );
	  DemarqueEtRendDag( &filsIni1 );

/* 	filsCou1 = filsIni1; */
/* 	while( filsCou1 ) { */
/* printf("rend1\n"); */
/* 	  filsSuivant1 = filsCou1->Frere; */
/* 	  RestaurePointeurVersDagJete( *p_hdlPereAtome1 ,  */
/* 				       *p_hdlPereAtome2 ); */
/* printf("rend2\n"); */
/* 	  DemarqueEtRendDag( &filsCou1 ); */
/* printf("rend3\n"); */
	  
/* 	  filsCou1 = filsSuivant1; */
/* 	} */
      }
      /* suppresses the first list */
    }
    /* negatif = positif */

    /* positif = negatif */
    else if( filsCou2->ValDag.Atome.AtomeConjDisj == _ConjonctionNeg ) {

      /* scans the second list */
      filsIni2 = filsCou2;
      while( filsCou2 ) {
	filsSuivant2 = filsCou2->Frere;
	if( ( filsCou2->TypeDag != _AtomeRef
	      && filsCou2->TypeDag != _AtomeChaine
	      && filsCou2->TypeDag != _AtomeEntier )
	    || filsCou2->ValDag.Atome.AtomeConjDisj 
	    != _ConjonctionNeg ) {
	  TxtErreur( 1 , "liste de negations attendue" );
	  return _False;
	}
	  
	/* cas d'egalite entre une valeur interdite
	   et une valeur positive */
	if( AreEqual( filsCou1 , filsCou2 ) ) {
	  resul = _False;
	  /* break; */
	}
	filsCou2 = filsSuivant2;
      }
      /* scans the second list */
      /* suppresses the second list in case of success */
      if( resul ) {
	  RestaurePointeurVersDagJete( *p_hdlPereAtome2 , 
				       *p_hdlPereAtome1 );
	  DemarqueEtRendDag( &filsIni2 );

/* 	filsCou2 = filsIni2; */
/* 	while( filsCou2 ) { */
/* 	  filsSuivant2 = filsCou2->Frere; */
/* 	  RestaurePointeurVersDagJete( *p_hdlPereAtome2 ,  */
/* 				       *p_hdlPereAtome1 ); */
/* 	  DemarqueEtRendDag( &filsCou2 ); */
	  
/* 	  filsCou2 = filsSuivant2; */
/* 	} */
      }
      /* suppresses the second list */
    }

    /* on a fini, il ne reste que l'atome positif */
    filsCou1 = NULL;
    filsCou2 = NULL;
  }
  /* 1 atomes positif et un atome negatif */

  /* combinaison illegale */
  else {
    TxtErreur( 292 , "UnificationAtomeAtome(aldags)" );
    return _False;
  }

  if( filsCou1 || filsCou2 || !resul ) {
    if( Trace ) {
      char ch1[64];
      fprintf(FicLog,
	      "Failure of unification atom/atom:");
      filsCou1 = &dagTrace1;
      filsCou2 = &dagTrace2;
      ChaineDagMem( &filsCou1 , ch1 );
      fprintf( FicLog , "\n%s\n" , ch1 );
      ChaineDagMem( &filsCou2 , ch1 );
      fprintf( FicLog , "%s\n" , ch1 );
    }
    return _False;
  }
  return _True;

}

/*****************************************************************************
    UNIFICATION RECURSIVE DE DEUX DAGS
*****************************************************************************/
char     Unifier( TabDagMem *p_hdlFils1 , TabDagMem *p_hdlFils2  )
{
  char                       typeFils1;
  char                       typeFils2;
  TabDagMem                  filsVide;

  /*     printf("Arcs:\n"); */
  /*     EcrireDagBasiqueStdOut( p_hdlFils1 );  */
  /*     printf("\n"); */
  /*     EcrireDagBasiqueStdOut( p_hdlFils2 );  */
  /*     printf("\n"); */
 
  /* dans le cas ou on unifie 2 fils de fin de chemin,
     on rajoute un arc vide bidon pour garder le pointage multiple.
     Cet arc vide sera remplace par un atome ou un arc a la premiere
     unification venant modifier l'un des deux fils                */
  if( !(*p_hdlFils1) && !(*p_hdlFils2) ) {
    filsVide = NouveauDag( _TypeDagVide );

    if( !filsVide ) {
      return _False;
    }
    (*p_hdlFils1) = filsVide;
    (*p_hdlFils2) = filsVide;
    return _True;
  }
  if( !(*p_hdlFils1) || !(*p_hdlFils2) ) {
    if( !(*p_hdlFils1) ) {
      *p_hdlFils1 = *p_hdlFils2;
      return _True;
    }
    if( !(*p_hdlFils2) ) {
      *p_hdlFils2 = *p_hdlFils1;
      return _True;
    }
  }

  if( (*p_hdlFils1) == (*p_hdlFils2) ) {
    return _True;
  }

  /* unification selon le type des fils */
  typeFils1 = (*p_hdlFils1)->TypeDag;
  typeFils2 = (*p_hdlFils2)->TypeDag;
    
  if( typeFils1 == _TypeDagVide && typeFils2 == _TypeDagVide ) {
    filsVide = *p_hdlFils2;

    RestaurePointeurVersDagJete( *p_hdlFils2 , *p_hdlFils1 );
    if( *p_hdlFils2 != *p_hdlFils1 ) {
      TxtErreur( 1 , "revoir restauration pointeurs 12" );
    }

    /* on n'utilise pas p_hdlFils2 ici pour ne pas modifier sa valeur
       (il pointe vers l'atome commun)      */
    DemarqueDag( &filsVide );
    RendUnDag( &filsVide );

    return _True;
  }
  if( typeFils1 == _TypeDagVide || typeFils2 == _TypeDagVide ) {
    if( typeFils1 == _TypeDagVide ) {
      filsVide = *p_hdlFils1;

      RestaurePointeurVersDagJete( *p_hdlFils1 , *p_hdlFils2 );
      if( *p_hdlFils1 != *p_hdlFils2 ) {
	TxtErreur( 1 , "revoir restauration pointeurs 13" );
      }

      /* on n'utilise pas p_hdlFils1 ici pour ne pas modifier
	 sa valeur (il pointe vers l'atome commun) */
      DemarqueDag( &filsVide );
      RendUnDag( &filsVide );

      return _True;
    }
    if( typeFils2 == _TypeDagVide ) {
      filsVide = *p_hdlFils2;

      RestaurePointeurVersDagJete( *p_hdlFils2 , *p_hdlFils1 );
      if( *p_hdlFils2 != *p_hdlFils1 ) {
	TxtErreur( 1 , "revoir restauration pointeurs 14" );
      }

      /* on n'utilise pas p_hdlFils1 ici pour ne pas modifier
	 sa valeur (il pointe vers l'atome commun) */
      DemarqueDag( &filsVide );
      RendUnDag( &filsVide );

      return _True;
    }
  }

  if( ( typeFils1 == _AtomeEntier || typeFils1 == _AtomeRef
	|| typeFils1 == _AtomeChaine || typeFils1 == _AtomeListe )
      &&
      ( typeFils2 == _AtomeEntier || typeFils2 == _AtomeRef
	|| typeFils2 == _AtomeChaine || typeFils2 == _AtomeListe ) ) {
    return UnificationAtomeAtome( p_hdlFils1 , p_hdlFils2 );
  }
  else if( typeFils1 != typeFils2 ) {
    if( Trace ) {
      fprintf(FicLog,"Echec unification arc/atome\n");
      EcrireDagBasique( p_hdlFils1 ); 
      EcrireDagBasique( p_hdlFils2 );
    }
    return _False;
  }
  else if( typeFils1 == _Arc && typeFils2 == _Arc
	   /* || typeFils1 == _AtomeListe && typeFils2 == _AtomeListe */ ) {
    return UnificationArcArc( p_hdlFils1 , p_hdlFils2 );
  }

  TxtErreur( 93 , "Unifier(aldags)" );
  return _False;
}

/*****************************************************************************
    UNIFICATION DES CONTRAINTES SUR LES DAGS
*****************************************************************************/
char     UnifierContraintes( TabDagMem *p_filsOperLog , 
                             TabDagMem *p_dagUnifie ,
                             TabTermeMem *p_hdlTerme , 
                             TabBlocInitial p_ptBloc )
{
  TabDagMem                filsOperCou;
  char                     typeFils1;
  char                     typeFils2;
  char                     resul;
  char                     oper;
  struct OperLogMem        operLogMemCou;
  TabDagMem               *hdlPereChemin1;
  TabDagMem               *hdlPereChemin2;

  /* on suppose que tous les chemins necessaires ont ete crees lors de la 
     saisie des contraintes. Donc les operations pointent vers des noeuds
     existants qui sont soit des atomes soit des arcs
     L'unification arc/atome est possible
     - si l'arc est libre
     - ou s'il a un atome egale a cette atome
     - tous les autres cas sont des echecs d'unification
     L'unification arc arc est une operation recursive qui consiste a 
     faire pointer les deux peres vers le meme arc et a subsumer les arcs
     fils. La subsomption se realise recursivement ainsi
     - si un des deux arcs a un atome, l'autre doit avoir la meme atome
     ou etre libre, l'arc subsumant a cette atome
     - sinon - si l'etiquette d'un fils est dans un arc et pas dans l'autre, 
     on l'ajoute au arc subsumant
     - si l'etiquette d'un fils est dans les deux arcs, on subsume
     ces arcs fils et on ajoute cette etiquette et ce fils 
     subsumant a l'arc subsumant.
  */

  PtOperLogU = (*p_filsOperLog);        /* liste des operations
					   d'unif a faire */
  HdlDagU = p_dagUnifie;                /* liste des dags sur
					   lesquels porte l'unif */
  HdlTermeU = p_hdlTerme;               /* arbre de termes portant
					   les dags unifies */
  PtBlocU = p_ptBloc;                   /* liste des blocs portant
					   les dags unifies */

  resul = _True;
  filsOperCou = (*p_filsOperLog);

  while( filsOperCou && resul ) {
    /* recherche des fils de l'operation */
    if( filsOperCou->TypeDag != _Oper ) {
      printf( "Type dag: %d\n " , filsOperCou->TypeDag );
      TxtErreur( 93 , "UnifierContraintes(aldags)" );
      return _False;
    }
 
    oper = filsOperCou->ValDag.OperLog.Oper;
    if( oper != '=' && oper != '!' ) {
      sprintf( ChAux , "\nOperation : [%c]\n\n" , 
	       filsOperCou->ValDag.OperLog.Oper );
      printf( ChAux );
      TxtErreur( 93 , "UnifierContraintes(aldags)" );
      return _False;
    }

    operLogMemCou = filsOperCou->ValDag.OperLog;
    typeFils1 = operLogMemCou.TypeFils1;
    typeFils2 = operLogMemCou.TypeFils2;
    hdlPereChemin1 = operLogMemCou.Fils1.hdlPereChemin;
    hdlPereChemin2 = operLogMemCou.Fils2.hdlPereChemin;

    if( ( typeFils1 == _Arc || typeFils1 == _AtomeListe )
	&& ( typeFils2 == _AtomeEntier || typeFils2 == _AtomeRef
	     || typeFils2 == _AtomeChaine 
	     || typeFils2 == _AtomeListe ) ) {
      /* 	  if( Trace ) { */
      /* 	    fprintf( FicLog , "\nAvant unification Arc Atome %d/%d" , */
      /* 		    typeFils1 , typeFils2 ); */
      /* 	    EcrireDagBasique( hdlPereChemin1 ); */
      /* 	    EcrireDagBasique( &(operLogMemCou.Fils2.ptPereAtome) ); */
      /* 	  } */
      resul &= Unifier( hdlPereChemin1 , 
			&(operLogMemCou.Fils2.ptPereAtome) );
      /* 	  if( Trace ) { */
      /* 	    fprintf( FicLog , "\nApres unification Arc Atome %d/%d" ,  */
      /* 		    typeFils1 , typeFils2 ); */
      /* 	    EcrireDagBasique( hdlPereChemin1 ); */
      /* 	    EcrireDagBasique( &(operLogMemCou.Fils2.ptPereAtome) ); */
      /* 	  } */
      hdlPereChemin1 = NULL;
      operLogMemCou.Fils2.ptPereAtome = NULL;
    }
    else if( typeFils1 == _Arc && typeFils2 == _Arc ) {
      if( oper == '!' ) {
	TxtErreur( 14 , "UnifierContraintes(aldags)" );
	return _False;
      }

/*       if( Trace ) { */
/* 	fprintf( FicLog , "\nAvant unification Arc Arc %d/%d" , */
/* 		 typeFils1 , typeFils2 ); */
/* 	EcrireDagBasique( hdlPereChemin1 ); */
/* 	EcrireDagBasique( hdlPereChemin2 ); */
/*       } */
      resul &= Unifier( hdlPereChemin1 , hdlPereChemin2 );
/*       if( Trace ) { */
/* 	fprintf( FicLog , "\nApres unification Arc Arc %d/%d" , */
/* 		 typeFils1 , typeFils2 ); */
/* 	EcrireDagBasique( hdlPereChemin1 ); */
/* 	EcrireDagBasique( hdlPereChemin2 ); */
/*       } */
      hdlPereChemin1 = NULL;
      hdlPereChemin2 = NULL;
    }
    else
      {
	printf("Types of arcs: %d/%d\n",typeFils1,typeFils2);
	TxtErreur( 93 , "UnifierContraintes(aldags)" );
	return _False;
      }
        
    filsOperCou = filsOperCou->Frere;
  }

  DemarqueEtRendDag( p_filsOperLog );
  return resul;
}

/*****************************************************************************
          RESTAURATION DES POINTEURS D'UN DAG APRES UNIFICATION
*****************************************************************************/
char     RestaurePointeursDag( TabDagMem *p_hdlDag )
{
  char                      typeDag = (*p_hdlDag)->TypeDag;
     
  switch( typeDag ) {
  case _TypeDagVide :
  case _AtomeEntier :
  case _AtomeRef :
  case _AtomeChaine :
    return _True;
  case _Oper :
    if( (*p_hdlDag)->ValDag.OperLog.TypeFils1 == _Arc ) {
      if( (*p_hdlDag)->ValDag.OperLog.Fils1.hdlPereChemin 
	  == &(PtDagJeteU->Fils) ) {
	(*p_hdlDag)->ValDag.OperLog.Fils1.hdlPereChemin 
	  = &(PtDagGardeU->Fils);
	NbRestaurationsU++;

      }
      else if( (*p_hdlDag)->ValDag.OperLog.Fils1.hdlPereChemin 
	       == &(PtDagJeteU->Frere) ) {
	(*p_hdlDag)->ValDag.OperLog.Fils1.hdlPereChemin 
	  = &(PtDagGardeU->Frere);
	NbRestaurationsU++;
      }
    }
    else
      {
	if( (*p_hdlDag)->ValDag.OperLog.Fils1.ptPereAtome == PtDagJeteU ) {
	  (*p_hdlDag)->ValDag.OperLog.Fils1.ptPereAtome = PtDagGardeU;
	  NbRestaurationsU++;
	}
      }
    if( (*p_hdlDag)->ValDag.OperLog.TypeFils2 == _Arc ) {
      if( (*p_hdlDag)->ValDag.OperLog.Fils2.hdlPereChemin 
	  == &(PtDagJeteU->Fils) ) {
	(*p_hdlDag)->ValDag.OperLog.Fils2.hdlPereChemin 
	  = &(PtDagGardeU->Fils);
	NbRestaurationsU++;
      }
      else if( (*p_hdlDag)->ValDag.OperLog.Fils2.hdlPereChemin 
	       == &(PtDagJeteU->Frere) ) {
	(*p_hdlDag)->ValDag.OperLog.Fils2.hdlPereChemin 
	  = &(PtDagGardeU->Frere);
	NbRestaurationsU++;
      }
    }
    else
      {
	if( (*p_hdlDag)->ValDag.OperLog.Fils2.ptPereAtome == PtDagJeteU ) {
	  (*p_hdlDag)->ValDag.OperLog.Fils2.ptPereAtome = PtDagGardeU;
	  NbRestaurationsU++;
	}
      }
    /* absence intentionnelle de break !!!!!!!! attention si chgt
       d'ordre */
  case _AtomeListe :
  case _Arc :
    if( (*p_hdlDag)->Fils == PtDagJeteU ) {
      (*p_hdlDag)->Fils = PtDagGardeU;
      NbRestaurationsU++;
    }
    if( (*p_hdlDag)->Frere == PtDagJeteU ) {
      (*p_hdlDag)->Frere = PtDagGardeU;
      NbRestaurationsU++;
    }
    break;
  default :
    sprintf( ChAux , "Dag de type inconnu : %d\n" , typeDag );
    printf( ChAux );
    TxtErreur( 93 , "RestaurePointeursDag(aldags)" );
    return _False;
  }

  return _True;
}

/*****************************************************************************
          RESTAURATION DES POINTEURS D'UN DAG APRES UNIFICATION
*****************************************************************************/
char     RestaurePointeurTerme( TabTermeMem *p_hdlTerme )
{
  if( (*p_hdlTerme)->ValTerme.ArcsFils == PtDagJeteU ) {  
    (*p_hdlTerme)->ValTerme.ArcsFils = PtDagGardeU;
    NbRestaurationsU++;
  }
  return  ParcoursDFDag( &((*p_hdlTerme)->ValTerme.ArcsFils) , RestaurePointeursDag , 
			 _PreOrdre , _False );
}

/*****************************************************************************
          RESTAURATION DES POINTEURS D'UN DAG APRES UNIFICATION
*****************************************************************************/
char     RestaurePointeurVersDagJete( TabDagMem p_dagAncien , 
                                      TabDagMem p_dagNouveau )
{
  TabBlocInitial              ptBlocCou;
  long                        ptMem;

  PtDagJeteU = p_dagAncien;
  PtDagGardeU = p_dagNouveau;
  NbRestaurationsU = 0;

  if( !ParcoursDFDag( &PtOperLogU , RestaurePointeursDag , 
		      _PreOrdre , _False ) ) {
    return _False;
  }

  if( HdlDagU ) {
    if( *HdlDagU == PtDagJeteU ) { 
      *HdlDagU = PtDagGardeU;
      NbRestaurationsU++;
    }
    if( !ParcoursDFDag( HdlDagU , RestaurePointeursDag , 
			_PreOrdre , _False ) ) {
      return _False;
    }
  }

  if( HdlTermeU && !ParcoursDFArbre( HdlTermeU , RestaurePointeurTerme , 
				     _PreOrdre ) ) {
    return _False;
  }

  if( PtBlocU ) {
    /* parcours remontant */
    if( PtBlocU->dagCopie == PtDagJeteU ) { 
      PtBlocU->dagCopie = PtDagGardeU;
      NbRestaurationsU++;
    }
    if( !ParcoursDFDag( &(PtBlocU->dagCopie) , RestaurePointeursDag , 
			_PreOrdre , _False ) ) {
      return _False;
    }

    ptMem = PtBlocU->ptBlocPrecedent;
    while( ptMem != _BasPile ) {
      ptBlocCou = (TabBlocInitial)(Pile + ptMem);
      if( ptBlocCou->dagCopie == PtDagJeteU ) { 
	ptBlocCou->dagCopie = PtDagGardeU;
	NbRestaurationsU++;
      }
      if( !ParcoursDFDag( &(ptBlocCou->dagCopie) , 
			  RestaurePointeursDag , _PreOrdre , _False ) ) {
	return _False;
      }
      ptMem = ptBlocCou->ptBlocPrecedent;
    }
  }
  return _True;
}

/*****************************************************************************
       COMPTE LE NOMBRE DE REFERENCES VERS UN DAG EN FAISANT UNE MISE A JOUR
       BIDON DU DAG SUR LUI-MEME ET EN COMPTANT LE NB DE M A J
*****************************************************************************/
int      NombreReferencesDag( TabDagMem p_ptDag )
{
  RestaurePointeurVersDagJete( p_ptDag , p_ptDag );
  return NbRestaurationsU;
}

/*****************************************************************************
       SUPPRESSION DES CHEMINS INUTILES DANS LES DAGS (FLEXION ET LEMME)
*****************************************************************************/
char     SupprimeCheminEtValeur( TabDagMem p_chemin , TabDagMem *p_hdlDag )
{
  TabDagMem                    ptCou;
  TabDagMem                    ptFrere;
  TabDagMem                   *hdlCou;

  /* descente synchrone du dag et du chemin */
  if( p_chemin ) {
    if( p_chemin->TypeDag == _Arc ) {
      /* on lance recursivement la suppression pour l'effectuer
	 de la fin vers le debut */
      if( (ptCou = ExisteArcOuListeFils( *p_hdlDag , 
					 p_chemin , _Arc ) ) ) {
	/* c'est l'arc ptCou qui sera supprime a la redescente
	   cela ne peut se faire que s'il n'y a pas d'autres
	   dags pointant vers lui */
	if( NombreReferencesDag( ptCou ) == 1 ) {
	  if( !SupprimeCheminEtValeur( p_chemin->Fils , 
				       &(ptCou->Fils) ) ) {
	    return _False;
	  }
	}
	else
	  {
	    return _False;
	  }
      }
      else
	{
	  return _False;
	}
    }
    else
      {
	return _False;
      }
  }
  /* descente synchrone du dag et du chemin */
  else
    /* on est en fin de chemin, s'il a une valeur atome, on la supprime */
    {
      if( *p_hdlDag ) {
	/* s'il y a un atome en fin de chemin, on le supprime */
	if( ( (*p_hdlDag)->TypeDag == _AtomeRef
	      ||(*p_hdlDag)->TypeDag == _AtomeEntier
	      || (*p_hdlDag)->TypeDag == _AtomeChaine )
	    && !((*p_hdlDag)->Fils) && !((*p_hdlDag)->Frere) ) {
	  ptCou = (*p_hdlDag);
	  if( NombreReferencesDag( ptCou ) == 1 ) {
	    RestaurePointeurVersDagJete( (*p_hdlDag) , NULL );
	    if( (*p_hdlDag) ) {
	      TxtErreur( 1 , "revoir restauration pointeurs 15" );
	    }
	    DemarqueDag( &ptCou );
	    RendUnDag( &ptCou );
	  }
	}
	/* s'il y a un atome en fin de chemin, on le supprime */
	else
	  /* s'il y a autre chose qu'un atome en fin de chemin, */
	  /* on le laisse et on arrete */
	  {
	    return _False;
	  }
      }
      /* s'il n'y a rien en fin de chemin, on ne fait rien */
      return _True;
    }
    
  /* en retour d'appel recursif, on desalloue les termes de la fin
     vers le debut si le terme courant a encore des fils cela
     signifie qu'il avait initialement plusieurs fils, on ne peut
     donc le desalloouer cette situation ne devrait normalement se
     produire que a la base du chemin */
  if( !(ptCou->Fils) ) {
    hdlCou = p_hdlDag;
    while( *hdlCou ) {
      /* on sauve le fere pour maintenir le chainage en cas de
	 fils multiples */
      ptFrere = (*hdlCou)->Frere;
      if( *hdlCou == ptCou ) {
	RestaurePointeurVersDagJete( ptCou , NULL );
	if( (*hdlCou) ) {
	  TxtErreur( 1 , "revoir restauration pointeurs 16" );
	}
	DemarqueDag( &ptCou );
	RendUnDag( &ptCou );
                   
	*hdlCou = ptFrere;
	return _True;
      }
      hdlCou = &((*hdlCou)->Frere);
    }
    /* fils non retrouve */
    TxtErreur( 93 , "SupprimeCheminEtValeur(aldags)" );
  }
  /* on arrete la suppression car un arc avait plusieurs fils et n'a
     pu etre supprime */
  return _False;
}

/*****************************************************************************
       SUPPRESSION DES CHEMINS INUTILES DANS LES TERMES (FLEXION ET LEMME)
*****************************************************************************/
char     SupprimeEtiquetteLemmeEtFlexion( TabTermeMem *p_hdlTerme )
{
  SupprimeCheminEtValeur( CheminLemme , 
			  &((*p_hdlTerme)->ValTerme.ArcsFils) );
  SupprimeCheminEtValeur( CheminFlexion , 
			  &((*p_hdlTerme)->ValTerme.ArcsFils) );
  SupprimeCheminEtValeur( CheminEtiquette , 
			  &((*p_hdlTerme)->ValTerme.ArcsFils) );
  SupprimeCheminEtValeur( CheminLexicalisation , 
			  &((*p_hdlTerme)->ValTerme.ArcsFils) );
  return _True;
    
}

/*****************************************************************************
       SUPPRESSION DES CHEMINS INUTILES DANS LES REGLES (FLEXION ET LEMME)
*****************************************************************************/
char     SupprimeEtiquetteLemmeEtFlexionDF( TabRegleMem p_ptRegle )
{
  PtOperLogU = NULL;
  HdlDagU = NULL;
  HdlTermeU = &(p_ptRegle->TermeFils);
  PtBlocU = NULL;

  if( !ParcoursDFArbre( HdlTermeU , SupprimeEtiquetteLemmeEtFlexion , 
			_PreOrdre ) ) {
    return _False;
  }
  return _True;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  fin de aldags.c
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
