/*
 * FASt Term Recognizer
 *
 *  FASTR/aldeflex.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for morphological analysis
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*****************************************************************************
 recherche toutes les racines possibles pour le mot, en parcourant la liste
                          des suffixes
*****************************************************************************/
void     DeSuffixage( char *p_chaineSuffixee ,
                      int p_taillesRacPossibles[ _MaxLetIdent ] , 
                      int p_longueur )
{
    int                 indCh;
    int                 nbLetSuff;
    int                 indTaillesRacPoss;
    int                 rang;
    int                 trouve;
    int                 derCh;

    ChIdent             chaineSuffixeeEnvers;

    indTaillesRacPoss = 0;
    if( SuffStripping )
    {
        /* si egalite approximative, la chaine est proposee dimninuee
           de la taille du desuffixage (a condition que la chaine
           restante comporte au moins deux caracteres) */
        if( p_longueur >= TailSuffStripping + 2 )
        {
            p_taillesRacPossibles[ indTaillesRacPoss++ ] =
                    p_longueur - TailSuffStripping;
        }
    }
    else if( !MorphoFlex && !MorphoDer )
    {
      /* si ni morpho flex ni morpho deriv, seule la chaine entiere
         est proposee */
    }
    else
    {
        /* morpho flex ou deriv on retire tous les suffixes potentiels
           et on propose les chaines de longueur correspondante */
        for( indCh = 0 ; indCh < p_longueur ; indCh++ )
        {
            chaineSuffixeeEnvers[ indCh ] = 
                    p_chaineSuffixee[ p_longueur - 1 - indCh ];
        }
        chaineSuffixeeEnvers[ p_longueur ] = '\0';

        for( nbLetSuff = _Min( _TailSuffixage , (p_longueur - 1) ) ;
             nbLetSuff >= 1 ; nbLetSuff-- )
        {
            /* parcours des suffixes a l'envers precedents */
            derCh = (int)chaineSuffixeeEnvers[ nbLetSuff ];
            chaineSuffixeeEnvers[ nbLetSuff ] = '\0';

            /* recherche d'un suffixe de long. nbLetSuff */
            PtDernierNoeudCommun( chaineSuffixeeEnvers , &rang , 
                                  &trouve , derCh );

            if( trouve == _True && p_longueur - nbLetSuff > 0 )
            {
                p_taillesRacPossibles[ indTaillesRacPoss++ ] =
                        p_longueur - nbLetSuff;
            }
        }
    }

    p_taillesRacPossibles[ indTaillesRacPoss++ ] = p_longueur;
    p_taillesRacPossibles[ indTaillesRacPoss++ ] = _Nil;
}

/*****************************************************************************
                creation des flexions a partir d'une racine
*****************************************************************************

/* retourne 0 si il n'y a plus de suffixages a produire, un sinon
   traite d'abord les flexions puis les derivations */
/* pour un mot ayant 3 flexions et 2 derivations, l'ordre dans lequel 
   les suffixages sont tentes est le suivant (flexion,derivation):
   (0,-1)->(1,-1)True, (1,-1)->(2,-1)True, (2,-1)->(3,-1)True, 
   (3,-1)->(-1,1)True, (-1,1)->(-1,2)True, (-1,2)->(-1,-1)False) */
/* si pas de flexion ni de derivation (0,-1)->(1,-1)True 
   (1,-1)->(-1,-1)False*/

char     ProduireSuffixages( char *p_racine ,char *p_racineAux ,
           int p_catLexCou , int p_nbReglesCou ,
           int p_noTypeFlexion , int p_noTypeDerivation ,
           int p_nbIndFlexions , int p_nbIndDerivations ,
           int *p_indiceDansTypeFlexion , int *p_indiceDansTypeDerivation ,
           char *p_chaineSuffixee ,
           int p_morphoFlex , int p_morphoDer )
{
     char               *ptCh;               /* chaine auxiliaire      */
     struct Suffixe     *ptsuff;             /* pointeur auxiliaire    */
     int                 rangPremSuffixe;

     /* mot sans flexions ni derivations -> une seule graphie: la racine */
     if( p_noTypeFlexion == _Nil && p_noTypeDerivation == _Nil )
     {
        if (*p_indiceDansTypeFlexion == 0)
        {
              strcpy( p_chaineSuffixee , p_racine );
              *p_indiceDansTypeFlexion = 1;
              *p_indiceDansTypeDerivation = _Nil;
              return _True;
        }
        else
        {
              *p_chaineSuffixee = '\0';
              *p_indiceDansTypeFlexion = _Nil;
              *p_indiceDansTypeDerivation = _Nil;
              return _False;
        }
     }

     /* passage des flex aux deriv */
     if( *p_indiceDansTypeFlexion >= p_nbIndFlexions 
         && *p_indiceDansTypeFlexion != _Nil)
     {
        /* en morpho deriv, on passe a la morpho deriv
           d'un lemme lorsqu'on a fini sa morpho flexionnelle */
        if( p_morphoDer )
        {
            *p_indiceDansTypeFlexion = _Nil;
            *p_indiceDansTypeDerivation = 0;
        }
        /* si non, on arrete l'analyse morphologique apres la flexion */
        else
        {
            *p_indiceDansTypeFlexion = _Nil;
            *p_indiceDansTypeDerivation = _Nil;
            return _False;
        }
     }

     /* fin des deriv */
     if( *p_indiceDansTypeDerivation >= p_nbIndDerivations 
         && *p_indiceDansTypeDerivation != _Nil)
     {
        *p_indiceDansTypeFlexion = _Nil;
        *p_indiceDansTypeDerivation = _Nil;
        return _False;
     }

     /* on parcourt les flexions puis les derivations jusqu'a pouvoir
        en construire une */
     while( ( *p_indiceDansTypeFlexion != _Nil
	      && *p_indiceDansTypeFlexion < p_nbIndFlexions
              && p_morphoFlex ) /* flexions */
            ||
             ( *p_indiceDansTypeFlexion == _Nil
               && p_nbIndDerivations != _Nil
               && *p_indiceDansTypeDerivation < p_nbIndDerivations 
               && p_morphoDer ) /* derivations */ )
     {

        /* recherche de la position dans le dico des suffixes */
        if( *p_indiceDansTypeFlexion < p_nbIndFlexions  
            && *p_indiceDansTypeFlexion != _Nil)
        {
           rangPremSuffixe 
               = DicoTraitsCatLex[ p_catLexCou ].RangPremSuffixeFlexion;

           ptsuff = (Suffixes +  rangPremSuffixe 
                           + (p_noTypeFlexion - 1) * p_nbIndFlexions
                           + *p_indiceDansTypeFlexion);
            ++(*p_indiceDansTypeFlexion);
        }
        else
        {
           rangPremSuffixe 
               = DicoTraitsCatLex[ p_catLexCou ].RangPremSuffixeDerivation;

           ptsuff = (Suffixes +  rangPremSuffixe 
                           + (p_noTypeDerivation - 1) * p_nbIndDerivations
                           + *p_indiceDansTypeDerivation);
           ++(*p_indiceDansTypeDerivation);
        }

        /* pas de racine auxiliaire */
        if( *p_racineAux == '\0' )
        {
            strcpy( p_chaineSuffixee , p_racine );
            /* on retourne la chaine suffixee si ce n'est pas un
               suffixe non approprie, sinon on passe au suffixe
               suivant */
            if( *(ptCh = ptsuff->Suff) == '*'|| *ptCh == '?')
            {
            }
            else if( *ptCh == '!' )
            {
                strcat( p_chaineSuffixee , 
                        p_racine + strlen( p_racine ) - 1 );
                strcat( p_chaineSuffixee , ptCh + 1 );
                p_chaineSuffixee[ _MaxLetIdent - 1 ] = '\0';
		/* secu de taille */
                return _True;
            }
            else
            {
                strcat( p_chaineSuffixee , ptCh );
                p_chaineSuffixee[ _MaxLetIdent - 1 ] = '\0';
		/* secu de taille */
                return _True;
            }
        }
        else
        {
            strcpy( p_chaineSuffixee , p_racineAux );
            if( *(ptCh = ptsuff->Suff) == '*'|| *ptCh != '?'
                    || ( ptCh[1] - '1' !=  _RacineAnnexe1 - p_nbReglesCou) )
            /* le caractere suivant le ? dans le suffixe indique
               le numero de la racine annexe choisie            */
            {
            }
            else
            {
                strcat( p_chaineSuffixee , ptCh + 2 );
                p_chaineSuffixee[ _MaxLetIdent - 1 ] = '\0'; /* secu de taille */
                return _True;
            }
        }

        if( *p_indiceDansTypeFlexion >= p_nbIndFlexions 
            && *p_indiceDansTypeFlexion != _Nil)
        {
           *p_indiceDansTypeFlexion = _Nil;
           *p_indiceDansTypeDerivation = 0;
        }
    }
    *p_indiceDansTypeFlexion = _Nil;
    *p_indiceDansTypeDerivation = _Nil;
    *p_chaineSuffixee = '\0';
    return _False;
}

/*****************************************************************************
                        mots ----> lemmes flechi
    1) dans le dictionnaire memoire 2) dans le dictionnaire binaire
*****************************************************************************/

char     MotALemme( TabMotMem p_ptMot  )
{
    struct EntreeDucs               *listePt[ _NbMaxHomographe ];

    int                              nbHomographe;
    TabLemmeMem                      ptLemme;
    TabLemmeMem                      ptLemmeAux;

    TabRegleMem                      ptRegle;

    struct ValAtomeMem               atome;

    ptLemme = NULL;
    ptRegle = NULL ;
    atome.AtomeConjDisj = _False;

    /* cas particulier d'un nombre */
    if( PretraitementDesNombres && _Num(p_ptMot->ChaineMot[ 0 ]) )
    {
        if( Trace )
        {
             fprintf( FicLog , "Numeric word\n" );
        }

        ptLemme = NouveauLemme();

        ptLemme->ValTerme.Ref = (long)_RefNum;
        ptLemme->ValTerme.TypeTerme  = _MotFlechi;
        if( strcmp( p_ptMot->ChaineMot , "1" ) == 0 )
        {
            atome.ValAtome.Ref = RefSingulier;
        }
        else
        {
            atome.ValAtome.Ref = RefPluriel;
        }
        if( !AjouteCheminEtAtomeFils( &(ptLemme->ValTerme.ArcsFils) ,
                                      CheminNombre , _AtomeRef , 
				      atome , _True ) )
        {
            return _False;
        }
        
        atome.ValAtome.Ref = RefDeterminantNumeral;
        if( !AjouteCheminEtAtomeFils( &(ptLemme->ValTerme.ArcsFils) ,
                                      CheminCategorie , _AtomeRef , 
				      atome , _True ) )
        {
            return _False;
        }
        
        strncpy( atome.ValAtome.Chaine , p_ptMot->ChaineMot , 
		 _MaxLetIdent - 1 );
	atome.ValAtome.Chaine[ _MaxLetIdent - 1 ] = '\0';
        if( !AjouteCheminEtAtomeFils( &(ptLemme->ValTerme.ArcsFils) ,
                                      CheminLemme , _AtomeChaine , 
				      atome , _True ) )
        {
            return _False;
        }
    }
    /* cas particulier d'un nombre */
    else if( FormatDEntree != _TxtTagge &&
	     FormatDEntree != _TxtTaggePartiellement &&
	     FormatDEntree != _TxtFenTaggePartielt &&
	     strchr( Ponctuation , *(p_ptMot->ChaineMot) ) )
    /* cas particulier d'une ponctuation */
    {
        if( Trace )
        {
            fprintf( FicLog , "Punctuation\n" );
        }

        ptLemme = NouveauLemme();

        ptLemme->ValTerme.Ref
            = _DebutRefPonctuation 
              + (long)(strchr( Ponctuation , 
                       *(p_ptMot->ChaineMot) ) - Ponctuation);
        ptLemme->ValTerme.TypeTerme  = _MotFlechi;

        if( strchr( PonctuationFinale , *(p_ptMot->ChaineMot) ) )
        {
            atome.ValAtome.Ref = RefPonctuationFinale;
        }
        else
        {
            atome.ValAtome.Ref = RefPonctuation;
        }
        if( !AjouteCheminEtAtomeFils( &(ptLemme->ValTerme.ArcsFils) ,
                                      CheminCategorie , _AtomeRef , 
				      atome , _True ) )
        {
            return _False;
        }

        strncpy( atome.ValAtome.Chaine , p_ptMot->ChaineMot  , 
		 _MaxLetIdent - 1 );
	atome.ValAtome.Chaine[ _MaxLetIdent - 1 ] = '\0';
        if( !AjouteCheminEtAtomeFils( &(ptLemme->ValTerme.ArcsFils) ,
                                      CheminLemme , _AtomeChaine , 
				      atome , _True ) )
        {
            return _False;
        }
    }
    /* cas particulier d'une ponctuation */
    else
    /* mot ni entier ni ponctuation */
    {
       ChIdent  chaineloc;

       strncpy( chaineloc , p_ptMot->ChaineMot  , 
		 _MaxLetIdent - 1 );
       chaineloc[ _MaxLetIdent - 1 ] = '\0';

       /* on regarde si le mot est dans le dictionnaire resident        */
       ListePointeursDucsParChaine( chaineloc , listePt );

       /* cas ou le mot se trouve dans le dico resident */
       if( listePt[0] )
       {
           if( Trace )
           {
                fprintf( FicLog , "Matched (frequent word): [%s]/%ld\n" ,
			 listePt[ 0 ]->Lemme.ValTerme.Ident , 
			 listePt[ 0 ]->Lemme.ValTerme.Ref );
           }

           nbHomographe = 0;
           ptLemme = NouveauLemme();
           ptLemmeAux = ptLemme;

           while( listePt[ nbHomographe ] )
           {
               if( nbHomographe )
               {
                  ptLemmeAux->LemmeFrere = NouveauLemme();
                  ptLemmeAux = ptLemmeAux->LemmeFrere;
               }

               *ptLemmeAux = listePt[ nbHomographe ]->Lemme;
               ptLemmeAux->ValTerme.TypeTerme  = _MotFlechi;
               ptLemmeAux->ValTerme.ArcsFils = NULL;
               ptLemmeAux->LemmeFrere = NULL;
               if( !CopieDag( &(ptLemmeAux->ValTerme.ArcsFils) , 
                       listePt[ nbHomographe ]->Lemme.ValTerme.ArcsFils ) )
               {
                   return _False;
               }

               nbHomographe++;
           }
       }
       /* cas ou le mot se trouve dans le dico resident */
       else
       /* recherche dans le dictionnaire binaire sur fichier */
       {
            if( !DeflexionLemme( p_ptMot , &ptLemme , &ptRegle ) )
            {
                return _False;
            }
       }
       /* recherche dans le dictionnaire binaire sur fichier */
    }
    /* mot ni entier ni ponctuation */

    p_ptMot->PtListeLemmes = ptLemme;
    /* si les flexions ont des regles associees , on les ajoute en queue de
       ReglesPhrase */
    if( ptRegle )
    {
          ReporteReglesSurReglesPhrase( ptRegle );
    }
    return _True;
}

/*****************************************************************************
 en cas de corpus partiellement etiquete, on ne retient que les lemmes
 dont la structure de traits est compatible avec cele fournie dans
 l'etiquetage (et portee par le mot)
*****************************************************************************/
char     FiltreLemme( TabMotMem p_ptMot  )
{
    TabDagMem         ptDagAux;
    TabLemmeMem      *hdlLemme;
    TabLemmeMem       ptLemmeAux;
    TabTermeMem       ptTerme;

    hdlLemme = &(p_ptMot->PtListeLemmes);
    /* pour chaque lemme, on tente l'unification avec les traits
       presents sur le mot flechi */
    while( (*hdlLemme) )
    {
       /* copie du dag du mot flechi (les tags partiels) */
       ptDagAux = NULL;
       if( !CopieDag( &ptDagAux , p_ptMot->ArcsFils ) )
       {
           return _False;
       }

       /* placement du dag du lemme dans un terme pour la restauration
          des pointeurs */
       if( !(ptTerme = NouveauTerme()) )
       {
           return _False;
       }
       ptTerme->ValTerme.ArcsFils = (*hdlLemme)->ValTerme.ArcsFils;

       /* les 4 bases de restauration des pointeurs */
       PtOperLogU = NULL; 
       HdlDagU = &ptDagAux; 
       HdlTermeU = &ptTerme;
       PtBlocU = NULL;   

       /* echec d'unification: on elimine le lemme considere */
       if( !Unifier( &ptDagAux , &(ptTerme->ValTerme.ArcsFils) ) )
       {
	 /* on rend (1) le dag du copie sur le mot : ptDagAux
	            (2) le dag du lemme courant [avec le lemme
		        qui ne peut plus etre conserve car non
			unifiable avec le mot courant] : 
	                ptTerme->ValTerme.ArcsFils */
	 /* on commence donc par tout demarquer, puis par marquer
	    celui qui est rendu en 2eme, le dag du lemme */
         if( !DemarqueDagDF( &ptDagAux ) 
             || !DemarqueDagDF( &(ptTerme->ValTerme.ArcsFils) )
             || !MarqueDagDF( &(ptTerme->ValTerme.ArcsFils) ) ) {
	   return _False;
         }
	 
	 /* on rend le dag copie sur le mot courant */
         if( !RendDag( &ptDagAux ) ) {
             return _False;
         }

	 /* on ecarte le lemme du chainage des lemmes pour le detruire */
	 ptLemmeAux = (*hdlLemme);
	 (*hdlLemme) = (*hdlLemme)->LemmeFrere;

	 /* on coupe le lien entre le lemme courant et
	    son dag, puisque le dag va etre desalloue lors
	    de la desallocation du terme auxiliaire ptTerme */
	 ptLemmeAux->ValTerme.ArcsFils = NULL;
	 ptLemmeAux->LemmeFrere = NULL;
	 DemarqueEtRendLemme( &ptLemmeAux );
       }
       /* echec d'unification: on elimine le lemme considere */
       else
       /* reussite d'unification: on garde le lemme unifie et on pase
          au suivant */
       {
           (*hdlLemme)->ValTerme.ArcsFils = ptTerme->ValTerme.ArcsFils;
           ptTerme->ValTerme.ArcsFils = NULL;
           hdlLemme = &((*hdlLemme)->LemmeFrere);
       }
       /* reussite d'unification: on garde le lemme unifie et on pase
          au suivant */

       /* desallocation du terme */
       DemarqueEtRendTerme( &ptTerme );
    }
    DemarqueEtRendDag( &(p_ptMot->ArcsFils) );
    return _True;
}

/*****************************************************************************
 transfert des regles trouvees pour un mot vers la phrase, sans repetitions
*****************************************************************************/
void     ReporteReglesSurReglesPhrase( TabRegleMem p_ptRegle )
{
    TabRegleMem                      ptRegleAux;
    TabRegleMem                      ptRegleNouv;
    TabRegleMem                      ptRegleNouvPrec;

    char                             trouve;

    if( p_ptRegle )
    {
        if( ReglesPhrase )
        {
           /* suppression des regles deja presentes dans la liste     */
           ptRegleNouvPrec = p_ptRegle;
           ptRegleNouv = p_ptRegle;
           while( ptRegleNouv )
           {
              ptRegleAux = ReglesPhrase;
              trouve = _False;
              while( ptRegleAux )
              {
                 if( ptRegleAux->TermeFils->ValTerme.Ref
                     == ptRegleNouv->TermeFils->ValTerme.Ref )
                 {
                     trouve = _True;
                     break;
                 }
                 ptRegleAux = ptRegleAux->RegleFrere;
              }
              if( trouve )
              {
                   if( ptRegleNouv == p_ptRegle )
                   {
                       p_ptRegle = p_ptRegle->RegleFrere;
                       DemarqueEtRendRegle( &ptRegleNouv );
                       ptRegleNouv = p_ptRegle;
                   }
                   else
                   {
                       ptRegleNouv = ptRegleNouv->RegleFrere;
                       DemarqueEtRendRegle( &(ptRegleNouvPrec->RegleFrere) );
                       ptRegleNouvPrec->RegleFrere = ptRegleNouv;
                   }
              }
              else
              {
                   ptRegleNouvPrec = ptRegleNouv;
                   ptRegleNouv = ptRegleNouv->RegleFrere;
              }
           }
           /* ajout des regles relevees en queues de celles existantes */

           /* ajout des regles relevees en queues de celles existantes */
           ptRegleAux = ReglesPhrase;
           while( ptRegleAux->RegleFrere )
           {
              ptRegleAux = ptRegleAux->RegleFrere;
           }

           ptRegleAux->RegleFrere = p_ptRegle;
           /* ajout des regles relevees en queues de celles existantes */
        }
        else
        {
            ReglesPhrase = p_ptRegle;
        }
    }
}

/*****************************************************************************
          Recolte les regles attachees a un lemme trouve en memoire
*****************************************************************************/
void     AjouteReglesARetourner( TabLemmeMem p_ptLemmeCou , 
                                 TabRegleMem *p_hdlRegle )
{
    unsigned long         indAbsRegles;         /* pointeur absolu
                                                   vers les regles courante*/
    TabRegleMem           ptRegleCou;
    TabRegleMem           ptRegleAux;

    indAbsRegles = p_ptLemmeCou->ValTerme.PosFicRegle;
    if( p_ptLemmeCou->NbRegles > 0
         && RestructurerLesRegles( &indAbsRegles , &ptRegleCou , 
                                   p_ptLemmeCou->NbRegles , _True ) != _Nil
         && ptRegleCou )
    {
       if( !( *p_hdlRegle ) )
       {
          *p_hdlRegle = ptRegleCou;
       }
       else
       {
          ptRegleAux = (*p_hdlRegle);
          while( ptRegleAux->RegleFrere )
          {
              ptRegleAux = ptRegleAux->RegleFrere;
          }
          ptRegleAux->RegleFrere = ptRegleCou;
       }
    }
}

/*****************************************************************************
          Recherche de tous les lemmes possibles attaches a un mot flechi
*****************************************************************************/
char     RechercheLemmes( TabChar p_chaineMotCou ,
                          ChIdent p_chPartielleMotCou , 
                          ChIdent p_chaineUlCou , ChIdent p_chaineAnnexe ,
                          TabLemmeMem p_ptLemmeCou , TabMotMem p_ptMot , 
                          TabLemmeMem *p_hdlLemme , 
                          int p_catLex , int p_nbReglesCou ,
                          int *p_unSuffixageTrouve ,
                          int p_noTypeFlexion , int p_noTypeDerivation , 
                          int p_nbIndFlexions , int p_nbIndDerivations ,
                          int p_morphoFlex , int p_morphoDer )
{
    /* indice parcours des flexions ou des deriv. d'1 Lemme */
    int                   indFlexion;        
    int                   indDerivation;        

    char                  racineSuffixee[ 2 * _MaxLetIdent ];
    TabLemmeMem           ptLemmeAux;

    /* DESUFFIXAGE LINGUISTIQUE => AJOUT DE TRAITS MORPHOS      */
    if( p_morphoFlex || p_morphoDer )
    {
        *p_unSuffixageTrouve = _False;
        indDerivation = _Nil;
        /* on fait appel a la construction des flexions par racine et
           suffixe */
        if( p_morphoFlex )
        {
            indFlexion = 0;
        }
        else
        {
            indFlexion = p_nbIndFlexions;
        }

        /* PARCOURS DES FLEXIONS ET DERIVATIONS DU LEMME COURANT      */
        while( ProduireSuffixages( p_chaineUlCou , p_chaineAnnexe ,
                p_catLex , p_nbReglesCou ,
                p_noTypeFlexion , p_noTypeDerivation , 
                p_nbIndFlexions , p_nbIndDerivations , 
                &indFlexion , &indDerivation ,
                racineSuffixee , p_morphoFlex , p_morphoDer ) )
        {
            if( Trace ) 
            {
              fprintf( FicLog , 
          "Calculated inflected word [%s/%s] #infl [%d/%d] #deriv [%d/%d]\n", 
                       racineSuffixee , p_chaineUlCou ,
		       indFlexion , p_noTypeFlexion ,
		       indDerivation , p_noTypeDerivation );
            }

            /* flexion dont la chaine convient */
            if(EgaliteChaine( racineSuffixee , p_chaineMotCou ))
            {
                /* garnissage des traits                  */
                if( AlloueEtGarnitLemme( p_ptLemmeCou , p_catLex , 
					 indFlexion , indDerivation ,
					 &ptLemmeAux , 
					 (int)p_ptMot->Forme , p_chaineUlCou ))
                {
                  *p_unSuffixageTrouve = _True;

                  /* on ajoute le terme correspondant a
                     retourner */
                  if( !( *p_hdlLemme ) )
                  {
                      *p_hdlLemme = ptLemmeAux;
                  }
                  else
                  {
                      ptLemmeAux->LemmeFrere = *p_hdlLemme;
                      *p_hdlLemme = ptLemmeAux;
                  }

                  /* on ajoute le terme correspondant a
                     retourner */
                  if( Trace )
                  {
		    fprintf( FicLog , 
     "Matched (regular lemma): [%s]/[%s]/%ld #infl [%d/%d] #deriv [%d/%d]\n" ,
                             p_chaineUlCou , racineSuffixee , 
			     ptLemmeAux->ValTerme.Ref , 
			     indFlexion , p_noTypeFlexion ,
			     indDerivation , p_noTypeDerivation  );
                  }
                }
                else
                {
                  if( Trace )
                  {
                    fprintf( FicLog , 
         "\nMatched (lemma) Incomp features lemma/suff: [%s]/[%s]\n",
                             p_chaineUlCou , racineSuffixee);
                  }
                }
            }
            /* flexion dont la chaine convient */
        }
        /* PARCOURS DES FLEXIONS ET DERIVATIONS DU LEMME COURANT      */
        /* on fait appel a la construction des flexions par racine et
           suffixe */
    }
    /* DESUFFIXAGE LINGUISTIQUE => AJOUT DE TRAITS MORPHOS      */
    else if( ( SuffStripping
                && (strncmp( p_chPartielleMotCou , p_chaineUlCou , 
                             strlen( p_chPartielleMotCou ) ) == 0
                    || strncmp( p_chPartielleMotCou , p_chaineUlCou , 
                              strlen( p_chaineUlCou ) - TailSuffStripping ) 
                                      == 0 ) )
              || ( !p_morphoFlex && !p_morphoDer 
                   && strcmp( p_chPartielleMotCou , p_chaineUlCou ) == 0 ) )
    /* DESUFFIXAGE UNIFORME => PAS DE FLEXION OU DE DERIVATION     */
    {
        int              catLexCou;

        if( SuffStripping )
        {
            catLexCou = _Nil;
        }
        else
        {
            catLexCou = p_catLex;
        }
        *p_unSuffixageTrouve = _True;

        /* garnissage des traits                  */
        if( !AlloueEtGarnitLemme( p_ptLemmeCou , catLexCou , _Nil , _Nil , 
                                  &ptLemmeAux , _Nil ,
                                  p_chaineUlCou ) )
        {
            fprintf( FicLog , "Lemme cree\n" );
            EcrireLemmeMem( &ptLemmeAux , FicLog , _Log , 0 );
            ptLemmeAux = p_ptLemmeCou;
            fprintf( FicLog , "Lemme donne\n" );
            EcrireLemmeMem( &ptLemmeAux , FicLog , _Log , 0 );
            DemarqueEtRendDag( &(p_ptLemmeCou->ValTerme.ArcsFils) );
            TxtErreur( 88 , "RechercheLemmes(aldeflex)" );
            return _False;
        }

        /* on ajoute le terme correspondant a retourner */
        if( !( *p_hdlLemme ) )
        {
            *p_hdlLemme = ptLemmeAux;
        }
        else
        {
            ptLemmeAux->LemmeFrere = *p_hdlLemme;
            *p_hdlLemme = ptLemmeAux;
        }
        /* on ajoute le terme correspondant a retourner */
    }
    /* DESUFFIXAGE UNIFORME => PAS DE FLEXION OU DE DERIVATION     */
    return _True;
}

/****************************************************************************
        mots ----> lemmes flechis (dans le dictionnaire binaire)
*****************************************************************************
      passage d'un mot aux flexions possibles
La recherche d'un Lemme correspondant a un mot donne se fait en 2 temps :
        1 racine incluse dans le mot
        2 flexion permettant de completer le mot.
La recherche dans le BArbre fournit pt : le pointeur le plus proche
precedent la forme candidate.                                   */

char     DeflexionLemme( TabMotMem  p_ptMot,
                         TabLemmeMem *p_hdlLemme , TabRegleMem *p_hdlRegle )
{
    char                  nbReglesCou;       /* permet de choisir la
                                                rac annexe */
    char                  egale;             /* ulr de meme racine */
    int                   nbLettresRacine;
    long                  refAux;

    unsigned long         indAbsLemme;       /* pt fic vers le Lemme courante*/
    unsigned long         indAbsLemmeAux;    /* pt fic vers le Lemme annexe */

    ChIdent               chaineMotCou; 
    TabChar               lemmeMotCou; 
    TabChar               chLemmeCou;        /* pt chaine du Lemme cour.  */
    ChIdent               chPartielleMotCou; /* sous-chaine du mot fourni */
    ChIdent               chaineUlCou;       /*la chaine du Lemme cour.  */
    ChIdent               chaineAnnexe;   /* autre racine evntuelle Lemme ct */

    ChIdent               chaineAux1;
    ChIdent               chaineAux2;

    struct LemmeMem       lemmeCou;                   /* Lemme courant */

    int                   taillesRacPossibles[ _MaxLetIdent ];
    int                   indTaillesRacPoss;

    int                   unSuffixageTrouve; 

    /* no de flexion ou de derivation determinant le choix des
       suffixes a ajouter */
    int                   noTypeFlexion;
    int                   noTypeDerivation;

    /* nombre de flexions de la cat lex du mot courant */
    int                   nbIndFlexions;
    int                   nbIndDerivations;

    int                   catLex;
    int                   catLexMotCou;
    struct ValAtomeMem    atome;
    int                   lemmePpalTrouve;
    int                   premiereBoucle;

    /* chaine du mot courant donnee par ptMot */
    strncpy( chaineMotCou , p_ptMot->ChaineMot , _MaxLetIdent - 1 );
    chaineMotCou[ _MaxLetIdent - 1 ] = '\0';

    lemmeMotCou = ChaineLemme( &(p_ptMot->ArcsFils) );
    catLexMotCou = CategorieLexicale( &(p_ptMot->ArcsFils) );

    /* lieux de deflexion possible */
    if( !lemmeMotCou )
    {
        DeSuffixage( chaineMotCou , taillesRacPossibles , 
                     (int)strlen(chaineMotCou) );
    }
    else
    /* corpus tagge pour lequel le lemme du mot est donne: pas de deflexion */
    {
        taillesRacPossibles[ 0 ] = (int)strlen(chaineMotCou);
        taillesRacPossibles[ 1 ] = 0;
    }

    indTaillesRacPoss = 0; 

    /* lorsque MorphoDer vaut _MorphoDerSiInconnu, on tente d'abord
       pour un mot une analyse flexionnelle. Si aucune flexion n'est
       trouvee, alors l'analyse derivationnelle est essayeee */
    /* d'ou la necessite de faire 2 boucles */
    premiereBoucle = _True;
    /* BOUCLE SUR LES SOUS-CHAINES DE LA RACINE                     */
    while( _True )
    {
        nbLettresRacine = taillesRacPossibles[ indTaillesRacPoss++ ];

        if( nbLettresRacine <=0 )
        {
           /* en mode flexion et deriv en une fois, on s'arrete
              lorsqu'on atteint la derniere racine possible */
           /* en mode de derivation seulement si on ne trouve pas de
              flexion, la deuxieme boucle n'est faite que si l'on n'a
              pas trouve de flexion */
           if( MorphoDer != _MorphoDerSiInconnu || (*p_hdlLemme) )
           {
               break;
           }
	   else
           /* en mode derivation a posteriori, on ne tente la
              derivation que si la flexion a echoue */
           {
	     /* passage de la premiere boucle a la deuxieme */
               if( premiereBoucle )
               {
                   indTaillesRacPoss = 0;
                   nbLettresRacine 
                         = taillesRacPossibles[ indTaillesRacPoss++ ];
                   premiereBoucle = _False;
               }
               if( nbLettresRacine <=0 )
               {
                   break;
               }
           }
        }

        /* chercher les Lemmes dont la racine est incluse dans le mot
           chaine, extraire une racine de nbletttres-racine de la
           chaine proposee */
        if( !lemmeMotCou )
        {
            strncpy( chPartielleMotCou , chaineMotCou , nbLettresRacine );
            chPartielleMotCou[ nbLettresRacine ] = '\0';
        }
        else
        /* corpus tagge pour lequel le lemme du mot est donne: pas de 
	   deflexion */
        {
            strcpy( chPartielleMotCou , lemmeMotCou );
        }

        /* recherche d'un Lemme ayant pour racine cette chaine partielle */
        indAbsLemme = IndiceInsertionDicLemmes( chPartielleMotCou , &egale );

        /* indAbsLemme est le pteur d'un Lemme de meme racine ou de
           racine superieure, plusieurs Lemmes successifs peuvent
           correspondre a la meme racine. Comparaison de la chaine du
           Lemme obtenu avec la sous-chaine de la racine proposee; si
           elles ne sont pas egales, on passe a la chaine sup. Rappel
           : lorsqu'on a accede a un Lemme elle est entierement
           incluse dans BufferLemmesCourant */

        /* BOUCLE SUR LES LEMMES DE MEME CHAINE                    */
         while( ( chLemmeCou = BufferLemmesCourant + _IRelLemme(indAbsLemme) )
                &&
                ( ( !SuffStripping
                    && EgaliteChaine( chLemmeCou , chPartielleMotCou ) )
                  || ( SuffStripping 
                       /* en retirant au plus TailSuffStripping caracteres
                          a chLemmeCou, on doit trouver chLemmeCou */
                       && strlen( chLemmeCou ) >= strlen( chPartielleMotCou )
                       && strlen( chLemmeCou ) <=  
                           strlen( chPartielleMotCou ) + 2 * TailSuffStripping
                       && strncmp( chPartielleMotCou , chLemmeCou , 
                                   strlen( chPartielleMotCou ) ) >= 0 )
                     )
              )
        {
    
            if( RestructurerBDicLemme( &indAbsLemme , chaineUlCou  ,
                                       &lemmeCou , chaineAnnexe ) == _Nil
               || EgaliteChaine( chaineUlCou , DernLemme ) )
            {
                     DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
                     break;
            }

            nbReglesCou = lemmeCou.NbRegles;

            /* cas d'un verbe ou un adjectif a racines multiples    */
            if( nbReglesCou <= _RacineAnnexe1 
                && nbReglesCou >= _RacineAnnexeMax )
            {
                refAux = lemmeCou.ValTerme.Ref;
                indAbsLemmeAux 
                       = IndiceInsertionDicLemmes( chaineUlCou , &egale );
                DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
                lemmePpalTrouve = _False;

                do
                {
                    if( RestructurerBDicLemme( &indAbsLemmeAux , chaineAux1 ,
                                               &lemmeCou , chaineAux2 ) == _Nil
                        || EgaliteChaine( chaineAux2 , DernLemme ) )
                    {
                         DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
                         break;
                    }

                    if( lemmeCou.ValTerme.Ref < -1L )
                    {
                        sprintf( ChAux , 
                                 "DeflexionLemme(aldeflex) [%s],[%s]/[%ld|%ld]" ,
                                 chaineUlCou , chaineAnnexe , refAux , 
                                 lemmeCou.ValTerme.Ref );
                        TxtErreur( 106 , ChAux );
                        return _False;
                    }

                    if( EgaliteChaine( chaineUlCou , chaineAux1 )
                        && refAux == lemmeCou.ValTerme.Ref )
                    {
                        if( Trace )
                        {
                            sprintf( ChAux , 
				     "Matched auxiliary root: [%s/%s]\n" , 
				     /* main stem */ chaineAux1 , 
				     /* aux stem */ chLemmeCou );
                            fprintf( FicLog , ChAux );
                        }

                        lemmePpalTrouve = _True;
                        break;
                    }
                }
                while( _True );
            }
            /* cas d'un verbe ou un adjectif a racines multiples    */
            else
            /* cas d'absence de racines multiples    */
            {
                chaineAnnexe[ 0 ] = '\0';
                lemmePpalTrouve = _True;
            }
            /* cas d'absence de racines multiples    */

            /* MOT DONT LE LEMME N'EST PAS CONNU: DEFLEXION NORMALE */
            if( !lemmeMotCou )
            {
                /* IL EXISTE UNE SUFFIXATION DU LEMME COURANT DONT LA CHAINE
                   EST CELLE DU MOT */
                if( lemmePpalTrouve )
                {
                    noTypeFlexion 
                           = NoTypeSuffixage( &(lemmeCou.ValTerme.ArcsFils) , 
                                              _True );
                    noTypeDerivation
                           = NoTypeSuffixage( &(lemmeCou.ValTerme.ArcsFils) , 
                                              _False );
                    catLex = CategorieLexicale( &(lemmeCou.ValTerme.ArcsFils) );
    
                    /* mot ayant des flexions on note le nb de flexions de
                       sa categorie */
                    if( catLex == _Nil 
                        || noTypeFlexion <= 0
                        || noTypeFlexion
                           > DicoTraitsCatLex[ catLex ].NbNoTypeFlexion )
                    {
                        noTypeFlexion = _Nil;
                        nbIndFlexions = _Nil;
                    }
                    else
                    {
                        nbIndFlexions 
                            = DicoTraitsCatLex[ catLex ].NbFlexions;
                    }
                    /* mot ayant des derivations on note le nb de
                       derivations de sa categorie */
                    if( catLex == _Nil 
                        || noTypeDerivation <= 0
                        || noTypeDerivation
                           > DicoTraitsCatLex[ catLex ].NbNoTypeDerivation )
                    {
                        noTypeDerivation = _Nil;
                        nbIndDerivations = _Nil;
                    }
                    else
                    {
                        nbIndDerivations 
                            = DicoTraitsCatLex[ catLex ].NbDerivations;
                    }
    
                    if( premiereBoucle )
                    {
		      if( Trace ) {
			fprintf( FicLog , 
			   "infl morph [%s] #infl [%d/%d]\n", 
				 chaineMotCou ,
				 nbIndFlexions , noTypeFlexion );
		      }

                        RechercheLemmes( chaineMotCou , chPartielleMotCou , 
                                         chaineUlCou , chaineAnnexe ,
                                         &lemmeCou , p_ptMot , p_hdlLemme ,
                                         catLex , nbReglesCou ,
                                         &unSuffixageTrouve ,
                                         noTypeFlexion , noTypeDerivation , 
                                         nbIndFlexions , nbIndDerivations ,
                                         MorphoFlex , 
                                         (int)(MorphoDer == _ToujoursMorphoDer) );
                    }
                    else
                    {
		      if( Trace ) {
			fprintf( FicLog , 
			   "der morph [%s] #deriv [%d/%d]\n", 
				 chaineMotCou ,
				 nbIndDerivations , noTypeDerivation );
		      }

                        RechercheLemmes( chaineMotCou , chPartielleMotCou , 
                                         chaineUlCou , chaineAnnexe ,
                                         &lemmeCou , p_ptMot , p_hdlLemme ,
                                         catLex , nbReglesCou ,
                                         &unSuffixageTrouve ,
                                         noTypeFlexion , noTypeDerivation , 
                                         nbIndFlexions , nbIndDerivations ,
                                         _False , /* la morpho flex a ete faite
                                                     a la 1ere boucle */
                                         (int)(MorphoDer == _MorphoDerSiInconnu) );
                    }
                }
                /* IL EXISTE UNE SUFFIXATION DU LEMME COURANT DONT LA CHAINE
                   EST CELLE DU MOT */
            }
            /* MOT DONT LE LEMME N'EST PAS CONNU: DEFLEXION NORMALE */
            else
            /* MOT DONT LE LEMME EST CONNU: LE LEMME EST COPIE TEL QUEL */
            {
	        unSuffixageTrouve = _False;
		catLex = CategorieLexicale( &(lemmeCou.ValTerme.ArcsFils) );

                /* LE LEMME COURANT A LA MEME CHAINE ET LA MEME CATEGORIE
		   QUE LE LEMME DU MOT */
                if( lemmePpalTrouve && (catLexMotCou == catLex
		    || catLexMotCou == _Nil || catLex == _Nil ) )
                {
                    TabLemmeMem           ptLemmeAux;

                    /* garnissage des traits                  */
                    if( AlloueEtGarnitLemme( &lemmeCou , catLex , 
                            _Nil , _Nil , &ptLemmeAux , 
                            (int)_Nil , chaineUlCou ) )
                    {
                      unSuffixageTrouve = _True;
    
                      /* on ajoute le terme correspondant a
                         retourner */
                      if( !( *p_hdlLemme ) )
                      {
                          *p_hdlLemme = ptLemmeAux;
                      }
                      else
                      {
                          ptLemmeAux->LemmeFrere = *p_hdlLemme;
                          *p_hdlLemme = ptLemmeAux;
                      }
    
                      /* on ajoute le terme correspondant a
                         retourner */
                      if( Trace )
                      {
                        fprintf( FicLog , 
                                 "Matched (uninflected lemma): [%s]\n",
                                 chaineUlCou );
                      }
                    }
                    else
                    {
                      if( Trace )
                      {
                        fprintf( FicLog , 
		         "Matched (lemma) Incomp features lemma/suff: [%s]\n",
				 chaineUlCou );
                      }
                    }
                }
                /* LE LEMME COURANT A LA MEME CHAINE QUE LE LEMME DU MOT */
            }
            /* MOT DONT LE LEMME EST CONNU: LE LEMME EST COPIE TEL QUEL */

            /* si une flexion du lemme courant convient, on recupere
               les regles a retourner */
            if( unSuffixageTrouve )
            {
                AjouteReglesARetourner( &lemmeCou , p_hdlRegle );
            }

            DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
        }
        /* BOUCLE SUR LES LEMMES DE MEME CHAINE                    */
    }
    /* BOUCLE SUR LES SOUS-CHAINES DE LA RACINE              */
    
    /* SI PAS DE MOT TROUVE, ON CREE UN MOT DE CHAINE EGALE AU MOT DU
       TEXTE ET DE TRAITS VIDES SAUF CATLEX : NOM (ET DE REFERENCE
       NULLE) */
    if( !(*p_hdlLemme) )
    {
      if( Trace ) {
	fprintf( FicLog , "No matched lemma: [%s]\n", chPartielleMotCou );
      }

       /* garnissage des traits                  */
       (*p_hdlLemme) = NouveauLemme();
       (*p_hdlLemme)->ValTerme.TypeTerme  = _MotFlechi;

       /* MOT DONT LE LEMME N'EST PAS CONNU ET SANS FLEXION:
          FLEXION PAR DEFAUT DES MOTS INCONNUS */
       if( !lemmeMotCou )
       {
	 if( Trace ) {
	   fprintf( FicLog , "Default features: [%s]\n", chPartielleMotCou );
	 }

          if( RefCatMotsInc != RefCatLexVide 
	      && CategorieLexicale( &(p_ptMot->ArcsFils) ) 
	         == _Nil )
          {
             atome.ValAtome.Ref = RefCatMotsInc;
             atome.AtomeConjDisj = _False;
             if( !AjouteCheminEtAtomeFils( 
			       &((*p_hdlLemme)->ValTerme.ArcsFils) ,
                               CheminCategorie , _AtomeRef , 
			       atome , _True ) )
             {
                 TxtErreur( 88 , "DeflexionLemme(aldeflex)" );
                 return _False;
             }
          } 
   
          strcpy( atome.ValAtome.Chaine , chaineMotCou );
          atome.AtomeConjDisj = _False;
          if( !AjouteCheminEtAtomeFils( &((*p_hdlLemme)->ValTerme.ArcsFils) ,
                                        CheminLemme , _AtomeChaine , 
					atome , _True ) )
          {
              TxtErreur( 88 , "DeflexionLemme(aldeflex)" );
              return _False;
          }
       
          atome.ValAtome.Entier = _Nil;
          atome.AtomeConjDisj = _False;
          if( !AjouteCheminEtAtomeFils( &((*p_hdlLemme)->ValTerme.ArcsFils) ,
					CheminReference , _AtomeEntier , 
					atome , _True ) )
          {
              TxtErreur( 88 , "DeflexionLemme(aldeflex)" );
              return _False;
          }	
      }
       /* MOT DONT LE LEMME N'EST PAS CONNU ET SANS FLEXION:
          FLEXION PAR DEFAUT DES MOTS INCONNUS */
    }
    return _True;
}

/************************************************************************
            PRODUCTION DE TOUTES LES FLEXIONS D'UN MOT
*************************************************************************/
int      MotFlechiProduit( TabTermeMem p_ptTermeDonne ,
                           TabLemmesMots p_tLemesMotsProduits )
{
     unsigned long           indAbsCou;
     unsigned long           longFic;
     struct LemmeMem         lemmeCou;
     TabLemmeMem             ptLemmeAux;
     TabLemmeMem             ptLemmeProduit;
     char                    racineSuffixee[ 2 * _MaxLetIdent ];
     int                     nbHomographes;

     /* indice parcours des flexions ou des deriv. d'1 Lemme */
     int                     indFlexion;        
     int                     indDerivation;        

     /* no de flexion ou de derivation determinant le choix des
        suffixes a ajouter */
     int                     noTypeFlexion;
     int                     noTypeDerivation;

     /* nombre de flexions de la cat lex du mot courant */
     int                     nbIndFlexions;
     int                     nbIndDerivations;

     int                     catLex;
     struct EntreeDucs      *listePt[ _NbMaxHomographe ];

     /* description du terme releve dans le dico */
     ChIdent                 chaineAnnexe;

     nbHomographes = 0;

     /* c'est un paradigme */
     if( p_ptTermeDonne->ValTerme.Ref == _Nil )
     {
         ptLemmeAux = (p_tLemesMotsProduits[ nbHomographes ].PtLemme 
                       = NouveauLemme());
         p_tLemesMotsProduits[ nbHomographes ].IndMot = 0;

         ptLemmeAux->ValTerme = p_ptTermeDonne->ValTerme;
         *(ptLemmeAux->ValTerme.Ident) = '\0';
         catLex = CategorieLexicale( &(p_ptTermeDonne->ValTerme.ArcsFils) );
         if( catLex != _Nil )
         {
            strcat( ptLemmeAux->ValTerme.Ident , 
                    DicoTraitsCatLex[ catLex].ChaineDico );
         }
         ptLemmeAux->ValTerme.TypeTerme = _MotFlechi;
         nbHomographes = 1;
     }
     else
     /* ce n'est pas un paradigme */
     {
       /* on regarde si le mot est dans le dictionnaire resident        */
       ListePointeursDucsParReference( p_ptTermeDonne->ValTerme.Ref , 
                                       listePt );

       /* cas ou le mot se trouve dans le dico resident */
       if( listePt[0] )
       {
            if( Trace )
            {
                 fprintf( FicLog , "Matched (frequent word): [%s]\n",
                                  listePt[ 0 ]->Lemme.ValTerme.Ident);
            }

            nbHomographes = 0;

            while( listePt[ nbHomographes ]
                   && nbHomographes < _NbMaxHomographe )
            {
                ptLemmeAux 
                   = ( p_tLemesMotsProduits[ nbHomographes ].PtLemme 
                              = NouveauLemme() );
                p_tLemesMotsProduits[ nbHomographes ].IndMot = 0;

                ptLemmeAux->ValTerme
                   = listePt[ nbHomographes ]->Lemme.ValTerme;
                ptLemmeAux->ValTerme.TypeTerme  = _MotFlechi;
                strcpy( ptLemmeAux->ValTerme.Ident ,
                        listePt[ nbHomographes ]->Lemme.ValTerme.Ident );

                nbHomographes++;
            }
       }
       else
       /* le mot n'est pas ds le dico resident */
       {
         /* recherche sequentielle dans le lexique pour retrouver le mot
            dont la reference est egale a celle duterme */
        indAbsCou = (long)( _TailleAmorce + _TailleSeparation );
        fseek( FicBDicLemmes , 0L , 2 );
        fgetpos( FicBDicLemmes , &longFic );
        *(lemmeCou.ValTerme.Ident) = '\0';

        /* parcours de tous les mots du lexique */
        while( indAbsCou < longFic )
        {
            if( RestructurerBDicLemme( &indAbsCou , lemmeCou.ValTerme.Ident  ,
                                       &lemmeCou  , chaineAnnexe ) == _Nil
                || EgaliteChaine( lemmeCou.ValTerme.Ident , DernLemme ) )
            {
                     DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
                     break;
            }

            /* mot retrouve : on genere toutes les flexions possibles */
            if( lemmeCou.ValTerme.Ref == p_ptTermeDonne->ValTerme.Ref )
            {
                noTypeFlexion 
                      = NoTypeSuffixage( &(lemmeCou.ValTerme.ArcsFils) , 
                                         _True );
                noTypeDerivation 
                      = NoTypeSuffixage( &(lemmeCou.ValTerme.ArcsFils) , 
                                         _False );
                catLex = CategorieLexicale( &(lemmeCou.ValTerme.ArcsFils) );
               /* mot ayant des flexions on note le nb de flexions de
                   sa categorie */
                if( catLex == _Nil 
                    || noTypeFlexion <= 0
                    || noTypeFlexion
                       > DicoTraitsCatLex[ catLex ].NbNoTypeFlexion )
                {
                    noTypeFlexion = _Nil;
                    nbIndFlexions = _Nil;
                }
                else
                {
                    nbIndFlexions 
                        = DicoTraitsCatLex[ catLex ].NbFlexions;
                }
                /* mot ayant des derivations on note le nb de
                   derivations de sa categorie */
                if( catLex == _Nil 
                    || noTypeDerivation <= 0
                    || noTypeDerivation
                       > DicoTraitsCatLex[ catLex ].NbNoTypeDerivation )
                {
                    noTypeDerivation = _Nil;
                    nbIndDerivations = _Nil;
                }
                else
                {
                    nbIndDerivations 
                        = DicoTraitsCatLex[ catLex ].NbDerivations;
                }

                /* production de toutes les suffixations */
                indFlexion = 0;
                indDerivation = _Nil;
                while( ProduireSuffixages( lemmeCou.ValTerme.Ident , 
                            chaineAnnexe , catLex , lemmeCou.NbRegles ,
                            noTypeFlexion , noTypeDerivation , 
                            nbIndFlexions , nbIndDerivations , 
                            &indFlexion , &indDerivation , racineSuffixee ,
                            MorphoFlex , MorphoDer ) 
                       != _False)
               {
                  if( !AlloueEtGarnitLemme( &lemmeCou , catLex , 
                                            indFlexion , indDerivation , 
                                            &ptLemmeProduit , (int)'N' , 
                                            lemmeCou.ValTerme.Ident ) )
                  {
                      DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
                      return _False;
                  }

                  strcpy( ptLemmeProduit->ValTerme.Ident , 
                          racineSuffixee );

                  /* les mots produits sont chaines par les freres */
                  ptLemmeAux 
                      = (p_tLemesMotsProduits[ nbHomographes ].PtLemme 
                                 = NouveauLemme());
                  p_tLemesMotsProduits[ nbHomographes ].IndMot = 0;
                  nbHomographes++;

                  ptLemmeAux->ValTerme = ptLemmeProduit->ValTerme;
               }
               /* production de toutes les flexions */

               DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
               break;
            }
            /* mot retrouve : on genere toutes les flexions possibles */

            DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
        }
        /* parcours de tous les mots du lexique */
       }
       /* le mot n'est pas ds le dico resident */
     }
     /* if( !p_paradigme ) */

     if( !nbHomographes )
     {
        sprintf( ChAux , "MotFlechiProduit(aldeflex) [%ld]" , 
                         p_ptTermeDonne->ValTerme.Ref );
        TxtErreur( 144 , ChAux );
     }
     return nbHomographes;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                         fin de aldeflex.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
