/*
 * FASt Term Recognizer
 *
 *  FASTR/alecrire.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for displaying data structures
 */

#include "al-ext.h"
#include "al-var.h"
#ifndef MACGRAPH
#include "al-graph.h"
#endif

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*************************************************************************
      VARIABLES LOCALES (GRAPHISME)
*************************************************************************/
int                             ProfondeurDagPrec;
int                             ProfondeurDag;
int                             ProfondeurTermePrec;
int                             ProfondeurTerme;
int                             Largeur;
int                             Indentation;
int                             LieuDAffichage;
FILE                           *FichierDAffichage;
int                             CompteurAtomesRegle;
TabDagMem                       PrecedinglyPrintedDag;

/*************************************************************************
            ECRITURE DES CHAINES D'UN TERME MEMOIRE
*************************************************************************/
void     EcritChaineDag( char *p_ch1 )
{
     int                             ind;

    /* pour les dags si la profondeur de parcours augmennte, on ne 
       retourne pas a la ligne */
     if( ProfondeurDag <= ProfondeurDagPrec && *p_ch1 )
     { 
        if( LieuDAffichage != _SortieWithoutLF ) {
	  fprintf( FichierDAffichage , "\n" );
        }
        if( LieuDAffichage == _Ecran || LieuDAffichage == _Log )
	{
            for( ind = 3 ; ind < ProfondeurDag ; ind++ )
            {
                fprintf( FichierDAffichage , "      " );
            }
            fprintf( FichierDAffichage , "{%d}" , ProfondeurDag - 3 );
        }
        else
	{
            for( ind = 0 ; ind < Indentation ; ind++ )
            {
                fprintf( FichierDAffichage , " " );
            }
            for( ind = 3 ; ind < ProfondeurTerme ; ind++ )
            {
                fprintf( FichierDAffichage , "      " );
            }
            for( ind = 3 ; ind < ProfondeurDag ; ind++ )
            {
                fprintf( FichierDAffichage , "- " );
            }
         }
	/* des freres avec deux chemins consecutifs identiques 
	   (e.g. une liste) */
	if( ProfondeurDag > 3 && *p_ch1 == '>' ) {
	  fprintf( FichierDAffichage , "<" );
	}
    }

    ProfondeurDagPrec = ProfondeurDag;
    fprintf( FichierDAffichage , "%s" , p_ch1 );
}

/*************************************************************************
            ECRITURE DES CHAINES D'UN TERME MEMOIRE
*************************************************************************/
void     EcritChaineTerme( TabTermeMem p_ptTermeFils , char *p_ch1 )
{
     int                             ind;
     int                             top;
     int                             left;
     int                             topLien;
     int                             leftLien;

    for( ind = 3 ; ind < ProfondeurTerme ; ind++ )
    {
        fprintf( FichierDAffichage , "      " );
    }

    ProfondeurTermePrec = ProfondeurTerme;
    fprintf( FichierDAffichage , "%s" ,  p_ch1 );

    if( (LieuDAffichage == _Ecran) && _MachineGraphique )
    {
         if( Largeur * 120 + 100 <= _LargeurEcran )
         {
             top = ProfondeurTerme * 50;
             left = Largeur * 120;
         }
         else
         {
             top = (ProfondeurTerme - (Largeur - 4)) * 50;
             left = 4 * 120;
         }
         DessineTerme( top , left , 30 , 100 , p_ch1 );

         if( p_ptTermeFils )
         {
             ind = 0;
             while( p_ptTermeFils )
             {
                 if( (Largeur + ind) * 120 + 100 <= _LargeurEcran )
                 {
                     topLien = (ProfondeurTerme + 1) * 50;
                     leftLien = (Largeur + ind) * 120;
                 }
                 else
                 {
                     topLien = (ProfondeurTerme + 1 
				- ((Largeur + ind) - 4)) * 50;
                     leftLien = 4 * 120;
                 }
                 _CrayonNormal();
                 _Ligne( left + 100 , top + 30 , leftLien , topLien );
                 ind++;
                 p_ptTermeFils = p_ptTermeFils->TermeFrere;
             }
         }
    }
}

/*************************************************************************
            IMPRESSION D'UN TERME MEMOIRE
*************************************************************************/
char     EcrireValeurTermeMem( struct ValTerme *pt_valTerme ,
                               TabTermeMem ptTermeFils )
{
     char                            ch1[ 256 ];
     ChIdent                         typeTerme;
     int                             catLexLemme;


    if( !pt_valTerme )
    {
        return _True;
    }

    if( LieuDAffichage == _Sortie || LieuDAffichage == _SortieWithoutLF )
    {
       if( ( catLexLemme   
                = CategorieLexicale( &(pt_valTerme->ArcsFils) ) ) 
                     == RefCatMotsInc || catLexLemme == _Nil )
       {
           strcpy( typeTerme , "X" );
       }
       else
       {
           strcpy( typeTerme , DicoTraitsCatLex[ catLexLemme ].ChaineDico );
       }
              
       switch( pt_valTerme->TypeTerme )
       {
           case _Concatenation : sprintf( ch1  , "%s%d ->" , typeTerme ,
                                    ++CompteurAtomesRegle  ); break;
           case _Liste : sprintf( ch1  , "%s%d =" , typeTerme ,
                                    ++CompteurAtomesRegle ); break;
           case _RegleNormale : sprintf( ch1  , "regle" ); break;
           case _RegleMeta : sprintf( ch1  , "meta-R" ); break;
           case _RegleElision : sprintf( ch1  , "elis" ); break;
           case _Feuille : sprintf( ch1  , "%s%d" , typeTerme ,
                                    ++CompteurAtomesRegle ); break;
           case _UniteLexicale : sprintf( ch1  , "%s%d" , typeTerme ,
                                    ++CompteurAtomesRegle ); break;
           case _Regle : sprintf( ch1  , "regle" ); break;
           case _MotFlechi : sprintf( ch1  , "mot-flechi" ); break;
       }
    }
    else
    {
       switch( pt_valTerme->TypeTerme )
       {
           case _Concatenation : strcpy( typeTerme , "conc" ); break;
           case _Liste : strcpy( typeTerme , "liste" ); break;
           case _RegleNormale : strcpy( typeTerme , "regle" ); break;
           case _RegleMeta : strcpy( typeTerme , "meta-R" ); break;
           case _RegleElision : strcpy( typeTerme , "elis" ); break;
           case _Feuille : strcpy( typeTerme , "feuille" ); break;
           case _UniteLexicale : strcpy( typeTerme , "unit-lex" ); break;
           case _Regle : strcpy( typeTerme , "regle" ); break;
           case _MotFlechi : strcpy( typeTerme , "mot-flechi" ); break;
       }
    }
    if( ( LieuDAffichage != _Sortie && LieuDAffichage != _SortieWithoutLF )
	|| Trace )
    {
       sprintf( ch1 , "Terme[%s %.31s(Ref:%ld Rac:%ld" , typeTerme , 
                       pt_valTerme->Ident , 
                       pt_valTerme->Ref , pt_valTerme->RefRacine  );
       {
	 TabRefPhraseMem refCou; 

	 refCou = pt_valTerme->RefSemantique;
	 while( refCou ) {
	   sprintf( ch1 , "%s Sem:%ld" , ch1 , refCou->Ref );
	   refCou = refCou->Suivant;
	 }

	 refCou = pt_valTerme->RefSelf;
	 while( refCou ) {
	   sprintf( ch1 , "%s Self:%ld" , ch1 , refCou->Ref );
	   refCou = refCou->Suivant;
	 }
       }
       sprintf( ch1 , "%s)]" , ch1 );
    }

    EcritChaineTerme( ptTermeFils , ch1 );

    ProfondeurDagPrec = 3;
    ProfondeurDag = 3;
    EcrireDag( &(pt_valTerme->ArcsFils) , _True , 
               FichierDAffichage , LieuDAffichage );
    fprintf( FichierDAffichage , "\n" );

    return _True;
}

char     EcrireTermeMem( TabTermeMem *p_ptTerme )
{
    if( !(*p_ptTerme) )
    {
        return _True;
    }
    
    EcrireValeurTermeMem( &((*p_ptTerme)->ValTerme) , 
                          (*p_ptTerme)->TermeFils );

    return _True;
}

char     EcrireRefTermeMotMem( TabTermeMem *p_ptTerme )
{
    if( !(*p_ptTerme) )
    {
        return _True;
    }
    
    if( (*p_ptTerme)->ValTerme.TypeTerme == _UniteLexicale )
    {
       fprintf( FicLog , "%ld " , (*p_ptTerme)->ValTerme.Ref  );
    }

    return _True;
}

/*************************************************************************
            IMPRESSION D'UN DAG MEMOIRE
*************************************************************************/
char     ChaineDagMem( TabDagMem *p_hdlFils , char *p_ch )
{
    ChIdent             ident;
    ChIdent             dagSpec;
    ChIdent             prefixe;
    int                 type;
    char                modeRiche;

    if( LieuDAffichage == _Ecran || LieuDAffichage == _Log )
    {
          modeRiche = _True; 
    }
    else
    {
          modeRiche = _False; 
    }


    if( !(*p_hdlFils) )
    {
        sprintf( p_ch , ">" );
    }
    else
    {
      /* definition du prefixe avant la valeur (pour les atomes) */
      if( (*p_hdlFils)->TypeDag == _AtomeEntier 
	  || (*p_hdlFils)->TypeDag == _AtomeRef
	  || (*p_hdlFils)->TypeDag == _AtomeChaine ) {
            if( (*p_hdlFils)->ValDag.Atome.AtomeConjDisj == _ConjonctionNeg  )
            {
	      if( modeRiche ) {
		strcpy( prefixe , ": -" );
	      }
	      else {
		strcpy( prefixe , "> ! " );
	      }
	    }
	    else if( (*p_hdlFils)->ValDag.Atome.AtomeConjDisj 
		     == _DisjonctionPos  ) {
	      if( modeRiche ) {
		strcpy( prefixe , ": ++" );
	      }
	      else {
		strcpy( prefixe , "> =+ " );
	      }
	    }
	    else {
	      if( modeRiche ) {
		strcpy( prefixe , ": +" );
	      }
	      else {
		strcpy( prefixe , "> = " );
	      }
	    }
      }
      else if( (*p_hdlFils)->TypeDag == _AtomeListe ) {
            if( (*p_hdlFils)->ValDag.Atome.AtomeConjDisj == _ConjonctionNeg  )
            {
		strcpy( prefixe , "!" );
	    }
	    else if( (*p_hdlFils)->ValDag.Atome.AtomeConjDisj 
		     == _DisjonctionPos  ) {
		strcpy( prefixe , "+" );
	    }
	    else {
		strcpy( prefixe , "" );
	    }
      }

      /* ecriture de la valeur */
        switch( (*p_hdlFils)->TypeDag )
        {
        case _TypeDagVide :
            if( modeRiche )
                   sprintf( p_ch , " (nil)>" );
            else
                   sprintf( p_ch , "> = _Nil" );
            break;
        case _AtomeEntier :
	  if( modeRiche )
	    sprintf( p_ch , "%se (%ld)>" , prefixe ,
		     (*p_hdlFils)->ValDag.Atome.ValAtome.Entier );
	  else
	    sprintf( p_ch , "%s%ld" , prefixe , 
		     (*p_hdlFils)->ValDag.Atome.ValAtome.Entier );
	  break;
        case _AtomeRef :
            TraitsCatLexAIdent( ident , &type , 
                                (int)(*p_hdlFils)->ValDag.Atome.ValAtome.Ref );
	    if( modeRiche )
	      sprintf( p_ch , "%sr (%ld)>" , prefixe , 
		       (*p_hdlFils)->ValDag.Atome.ValAtome.Ref );
	    else
	      sprintf( p_ch , "%s%s" , prefixe , ident );
            break;
        case _AtomeChaine :
	  if( modeRiche )
	    sprintf( p_ch , "%sch (%.31s)>" , prefixe , 
		     (*p_hdlFils)->ValDag.Atome.ValAtome.Chaine );
	  else
	    sprintf( p_ch , "%s'%s'" , prefixe , 
		     (*p_hdlFils)->ValDag.Atome.ValAtome.Chaine );
	  break;
        case _AtomeListe :
	  if( ProfondeurDag <= ProfondeurDagPrec )
	    {
	      sprintf( p_ch , "%s<list" , prefixe  );
	    }
	  else
	    {
	      sprintf( p_ch , " %slist" , prefixe  );
	    }
            break;
        case _Arc :
            /* cas part des dags d'automate charges du comptage */
            if( (*p_hdlFils)->ValDag.Arc.Comptage == _ArcDebComptage )
            {
                sprintf( dagSpec , "range[%d-%d]" ,
                        (*p_hdlFils)->ValDag.Arc.ValMin , 
                        (*p_hdlFils)->ValDag.Arc.ValMax );
            }
            else if( (*p_hdlFils)->ValDag.Arc.Comptage == _ArcFinComptage )
            {
                sprintf( dagSpec , "endRange/" );
            }
            else if( (*p_hdlFils)->ValDag.Arc.Comptage == _ArcSuiteComptage )
            {
                sprintf( dagSpec , "contRange/" );
            }
            else 
            {
                strcpy( dagSpec , "" );
            }

            if( TraitsCatLexAIdent( ident , &type , 
                                    (int)((*p_hdlFils)->ValDag.Arc.ValSymb) ) )
            {
                if( type  == _Trait )
                {
                    if( ProfondeurDag <= ProfondeurDagPrec )
                    {
                        if( modeRiche )
                            sprintf( p_ch , "<%s%.31s/%d " , dagSpec , ident, 
				     (int)((*p_hdlFils)->ValDag.Arc.ValSymb) );
                        else
                            sprintf( p_ch , "<%s%.31s" , dagSpec , ident );
                    }
                    else
                    {
                        if( modeRiche )
                            sprintf( p_ch , "%s%.31s/%d " , dagSpec , ident, 
				     (int)((*p_hdlFils)->ValDag.Arc.ValSymb) );
                        else
                            sprintf( p_ch , " %s%.31s" , dagSpec , ident );
                    }
		}
                else
                {
                    if( ProfondeurDag <= ProfondeurDagPrec )
                    {
                        if( modeRiche )
                            sprintf( p_ch , "<%s%.31s/%d " , dagSpec , ident, 
				     (int)((*p_hdlFils)->ValDag.Arc.ValSymb ) );
                        else
                            sprintf( p_ch , "<%s%.31s" , dagSpec , ident );
                    }
                    else
                    {
                        if( modeRiche )
                            sprintf( p_ch , " %s%.31s/%d " , dagSpec , ident, 
                                     (int)((*p_hdlFils)->ValDag.Arc.ValSymb ) );
                        else
                            sprintf( p_ch , " %s%.31s" , dagSpec , ident );
                    }
		 }
            }
            else
            {
                if( ProfondeurDag <= ProfondeurDagPrec )
                {
                    if( modeRiche )
                        sprintf( p_ch , "<%s%ld%s " , 
                                 dagSpec , (*p_hdlFils)->ValDag.Arc.ValSymb ,
				 (*p_hdlFils)->ValDag.Arc.Ident );
                    else
                        sprintf( p_ch , "<%s%s" , dagSpec ,
				 (*p_hdlFils)->ValDag.Arc.Ident );
                }
                else
                {
                    if( modeRiche )
                        sprintf( p_ch , "%s%ld%s " , dagSpec , 
                                 (*p_hdlFils)->ValDag.Arc.ValSymb ,
				 (*p_hdlFils)->ValDag.Arc.Ident );
                    else
                        sprintf( p_ch , " %s%s" , dagSpec ,
				 (*p_hdlFils)->ValDag.Arc.Ident );
                }
            }
            break;
        case _Oper :
            sprintf( p_ch , "O<%c>" , (*p_hdlFils)->ValDag.OperLog.Oper );
            break;
        default :
            sprintf( ChAux , "Dag de type inconnu : %d\n" , 
                     (*p_hdlFils)->TypeDag );
            printf( ChAux );
            TxtErreur( 93 , "ChaineDagMem(alecrire)" );
            return _False;
        }
    }
    return _True;
}

/*************************************************************************
            IMPRESSION D'UN DAG MEMOIRE
*************************************************************************/
char     EcrireDagMem( TabDagMem *p_hdlFils )
{
    char                ch1[ 64 ];

    if( !(*p_hdlFils) )
    {
        return _True;
    }
    
    if( !ChaineDagMem( p_hdlFils , ch1 ) )
    {
       return _False;
    }

    EcritChaineDag( ch1 );

    if( (*p_hdlFils)->TypeDag == _Oper )
    {
        switch( (*p_hdlFils)->ValDag.OperLog.TypeFils1 )
        {
           case _Arc :
           if( !EcrireDagMemAvecOuSansR( 
                         (*p_hdlFils)->ValDag.OperLog.Fils1.hdlPereChemin , 
                           _False ) )
           {
                return _False;
           }
           break;
           case _AtomeRef : case _AtomeEntier : 
           case _AtomeChaine : case _AtomeListe : 
           if( !EcrireDagMemAvecOuSansR( 
                         (&(*p_hdlFils)->ValDag.OperLog.Fils1.ptPereAtome) , 
                           _False ) )
           {
                return _False;
           }
           break;
        }
        switch( (*p_hdlFils)->ValDag.OperLog.TypeFils2 )
        {
           case _Arc :
           if( !EcrireDagMemAvecOuSansR( 
                         (*p_hdlFils)->ValDag.OperLog.Fils2.hdlPereChemin , 
                           _False ) )
           {
                return _False;
           }
           break;
           case _AtomeRef : case _AtomeEntier : 
           case _AtomeChaine : case _AtomeListe : 
           if( !EcrireDagMemAvecOuSansR( 
                         (&(*p_hdlFils)->ValDag.OperLog.Fils2.ptPereAtome) ,  
                           _False ) )
           {
                return _False;
           }
           break;
        }
    }

    return _True;
}

/*************************************************************************
            IMPRESSION D'UN DAG MEMOIRE
*************************************************************************/
char     EcrireDagMemAvecReentrance( TabDagMem *p_hdlFils )
{
    char                ch1[ 64 ];
    char                ch2[ 64 ];
    int                 ind;
    char                aEcrire;
    char                trouve;

    if( !(*p_hdlFils) )
    {
        return _True;
    }

    *ch1 = '\0';
    trouve = _False;
    for( ind = 0 ; ind < NbLiensNoeudsMemVar ; ind++ )
    {
        if( (LiensNoeudsMemVar + ind)->PtMem == *p_hdlFils )
        {
            trouve = _True;
            aEcrire = !((LiensNoeudsMemVar + ind)->Ecrit);
            (LiensNoeudsMemVar + ind)->Ecrit = _True;
            break;
        }
    }
    if( !trouve )
    {
        EcrireDagMem( p_hdlFils );
        TxtErreur( 98 , "EcrireDagMem(alecrire)" );
        return _False;
    }

    if( !aEcrire )
    {
        if( (LiensNoeudsMemVar + ind)->NoVar > 0 )
        {
           sprintf( ch1 , "> = (%d*)" , (LiensNoeudsMemVar + ind)->NoVar );
           EcritChaineDag( ch1 );
           return _ArretParcoursFils;
        }
        else
        {
            sprintf( ChAux , "reference multiple inattendue : %d\n" , 
                      (*p_hdlFils)->TypeDag );
            printf( ChAux );
            TxtErreur( 93 , "EcrireDagMem(alecrire)" );
            return _False;
        }
    }
    else
    {
        if( (LiensNoeudsMemVar + ind)->NoVar > 0 )
        {
           sprintf( ch1 , "(%d*)" , (LiensNoeudsMemVar + ind)->NoVar );
        }
    }
    
    if( !ChaineDagMem( p_hdlFils , ch2 ) )
    {
       return _False;
    }

    strcat( ch1 , ch2 );
    EcritChaineDag( ch1 );

    if( (*p_hdlFils)->TypeDag == _Oper )
    {
        switch( (*p_hdlFils)->ValDag.OperLog.TypeFils1 )
        {
           case _Arc :
           if( !EcrireDagMemAvecOuSansR( 
                         (*p_hdlFils)->ValDag.OperLog.Fils1.hdlPereChemin , 
                           _False ) )
           {
                return _False;
           }
           break;
           case _AtomeRef : case _AtomeEntier : 
           case _AtomeChaine : case _AtomeListe : 
           if( !EcrireDagMemAvecOuSansR( 
                         (&(*p_hdlFils)->ValDag.OperLog.Fils1.ptPereAtome) , 
                           _False ) )
           {
                return _False;
           }
           break;
        }
        switch( (*p_hdlFils)->ValDag.OperLog.TypeFils2 )
        {
           case _Arc :
           if( !EcrireDagMemAvecOuSansR( 
                         (*p_hdlFils)->ValDag.OperLog.Fils2.hdlPereChemin , 
                           _False ) )
           {
                return _False;
           }
           break;
           case _AtomeRef : case _AtomeEntier : 
           case _AtomeChaine : case _AtomeListe : 
           if( !EcrireDagMemAvecOuSansR( 
                         (&(*p_hdlFils)->ValDag.OperLog.Fils2.ptPereAtome) , 
               _False ) )
           {
                return _False;
           }
           break;
        }
    }

    return _True;
}

/*************************************************************************
      APPEL RECURSIF DE L'IMPRESSION D'UN DAG MEMOIRE EN PROFONDEUR
*************************************************************************/
char     EcrireDagMemAvecOuSansR( TabDagMem *p_hdlDag , 
				  int p_avecReentrance )
{
        char     resul;

        if( p_avecReentrance )
        {
            resul = ParcoursDFDag( p_hdlDag , EcrireDagMemAvecReentrance , 
                                   _PreOrdre , _False );
        }
        else
        {
            resul = ParcoursDFDag( p_hdlDag , EcrireDagMem , 
                                   _PreOrdre , _False );
        }
        return resul;
}

/*************************************************************************
        APPEL INITIAL DE L'IMPRESSION D'UN DAG MEMOIRE EN PROFONDEUR
*************************************************************************/
char     EcrireDag( TabDagMem *p_hdlDag , int p_avecReentrance , 
                    FILE *p_ptFicAffichage ,
                    int p_lieuDAffichage )
{
        char     resul;

        LieuDAffichage = p_lieuDAffichage;
        FichierDAffichage = p_ptFicAffichage;

        if( p_avecReentrance )
        {
            resul = ParcoursDFDag( p_hdlDag , EcrireDagMemAvecReentrance , 
                                   _PreOrdre , _False );
        }
        else
        {
            resul = ParcoursDFDag( p_hdlDag , EcrireDagMem , 
                                   _PreOrdre , _False );
        }
	/*        fprintf( FichierDAffichage , "\n" ); */
        return resul;
}

/*************************************************************************
         ECRITURE STANDARD DE DAG SANS REENTRANCE POUR TRACE
**************************************************************************/
char     PrintDagUnification( TabDagMem *p_hdlDag )
{
  ProfondeurDagPrec = 3;
  ProfondeurDag = 3;
  Largeur = 0;
  Indentation = 0;
  NbLiensNoeudsMemVar = 0;
  PrecedinglyPrintedDag = *p_hdlDag;
  ParcoursDFDag( p_hdlDag , NoteLiensNoeudsMemVar , _PreOrdre , _False );
  RenumeroteLiensNoeudsMemVar();
  if( EcrireDag( p_hdlDag , _True , FicUnification , _Sortie ) ) {
    return _True;
  }
  return _False;
}

/*************************************************************************
         ECRITURE STANDARD DE DAG SANS REENTRANCE POUR TRACE
**************************************************************************/
char     PrintDagUnificationWithoutRepetition( TabDagMem *p_hdlDag )
{
  char                                     resul = _True;
  int                                      nbFreres = 0;
  TabDagMem                               *hdlFrereCourant;
  TabDagMem                                frereSuivant;

  ProfondeurDagPrec = 3;
  ProfondeurDag = 3;
  Largeur = 0;
  Indentation = 0;
  NbLiensNoeudsMemVar = 0;
  PrecedinglyPrintedDag = *p_hdlDag;
  ParcoursDFDag( p_hdlDag , NoteLiensNoeudsMemVar , _PreOrdre , _False );
  RenumeroteLiensNoeudsMemVar();

  LieuDAffichage = _Sortie;
  FichierDAffichage = FicUnification;

  hdlFrereCourant = p_hdlDag;
  while( hdlFrereCourant && *hdlFrereCourant && nbFreres < 100 ) {

    if( !ExisteArcOuListeFils( (*hdlFrereCourant)->Frere , 
			       *hdlFrereCourant , _Arc ) ) {
      frereSuivant = (*hdlFrereCourant)->Frere;
      (*hdlFrereCourant)->Frere = NULL;
      resul &= EcrireDag( hdlFrereCourant , _True , FicUnification , _Sortie );
      (*hdlFrereCourant)->Frere = frereSuivant;
    }

    hdlFrereCourant = &((*hdlFrereCourant)->Frere);
    nbFreres++;
  }

  if( resul ) {
    return _True;
  }
  return _False;
}

/*************************************************************************
         ECRITURE STANDARD DE DAG SANS REENTRANCE POUR TRACE
**************************************************************************/
char     EcrireDagBasique( TabDagMem *p_hdlDag )
{
    ProfondeurDagPrec = 3;
    ProfondeurDag = 3;
    Largeur = 0;
    Indentation = 0;
    if( EcrireDag( p_hdlDag , _False , FicLog , _Sortie ) ) {
      fprintf( FicLog , "\n\n" );
      return _True;
    }
    return _False;
}

/*************************************************************************
         ECRITURE STANDARD DE DAG SANS REENTRANCE POUR TRACE
**************************************************************************/
char     EcrireDagBasiqueStdOut( TabDagMem *p_hdlDag )
{
    ProfondeurDagPrec = 3;
    ProfondeurDag = 3;
    Largeur = 0;
    Indentation = 0;
    if( EcrireDag( p_hdlDag , _False , stdout , _Sortie ) ) {
      printf( "\n\n" );
      return _True;
    }
    return _False;
}

/*************************************************************************
        RELEVE LES POINTEURS DES DAGS POUR NUMEROTATION EN AFFICHAGE
*************************************************************************/
char     NoteLiensNoeudsMemVar( TabDagMem *p_hdlDag )
{
    int                      ind;
    
    if( !(*p_hdlDag) )
    {
        return _True;
    }
    
    for( ind = 0 ; ind < NbLiensNoeudsMemVar ; ind++ )
    {
        if( (LiensNoeudsMemVar + ind)->PtMem == *p_hdlDag )
        {
            ((LiensNoeudsMemVar + ind)->NbRef)++;
            return _ArretParcoursFils;
        }
    }
    (LiensNoeudsMemVar + NbLiensNoeudsMemVar)->PtMem = *p_hdlDag;
    (LiensNoeudsMemVar + NbLiensNoeudsMemVar)->NbRef = 1;
    NbLiensNoeudsMemVar++;
    if( NbLiensNoeudsMemVar >= NbMaxLiensMemorises )
    {
         TxtErreur( 94 , "NoteLiensNoeudsMemVar(alecrire)" );
         return _False;
    }
    return _True;
}

/*************************************************************************
       RELEVE LES POINTEURS DES DAGS POUR NUMEROTATION EN AFFICHAGE
*************************************************************************/
char     NoteDFLiensNoeudsMemVar( TabTermeMem *p_hdlTerme )
{
    char     resul;

    if( !(*p_hdlTerme) )
    {
        return _True;
    }
    
    resul = ParcoursDFDag( &((*p_hdlTerme)->ValTerme.ArcsFils) , 
                           NoteLiensNoeudsMemVar , _PreOrdre , _False );

    return resul;
}

/*************************************************************************
       NUMEROTATION DES POINTEURS MULTIPLES DE AVANT AFFICHAGE
*************************************************************************/
void     RenumeroteLiensNoeudsMemVar( void )
{
    int                      ind;
    char                     indVar;
    
    indVar = 1;
    for( ind = 0 ; ind < NbLiensNoeudsMemVar ; ind++ )
    {
        if( (LiensNoeudsMemVar + ind)->NbRef > 1 )
        {
            (LiensNoeudsMemVar + ind)->NoVar = indVar;
            indVar++;
        }
        else
        {
            (LiensNoeudsMemVar + ind)->NoVar = 0;
        }
        (LiensNoeudsMemVar + ind)->Ecrit = _False;
    }
}

/*************************************************************************
            IMPRESSION D'UN LEMME MEMOIRE
*************************************************************************/
/*************************************************************************
#define    _Ecran                   0
#define    _Log                     1
#define    _Sortie                  2
*************************************************************************/
char     EcrireLemmeMem( TabLemmeMem *p_hdlLemme , FILE *p_ptFicAffichage , 
                         int p_lieuDAffichage , int p_indentation )
{
    char                 ch[ 64 ];

    if( !(*p_hdlLemme) )
    {
        return _True;
    }

    if( _MachineGraphique && (p_lieuDAffichage == _Ecran) )
    {
         _EcranGraphique();
    }

    /***  IMPRESSION DE L'IDENTIFICATEUR DE LEMME ****/
    if( _MachineGraphique && ( p_lieuDAffichage == _Ecran ) )
    {
         _EcranGraphique();
         TxtChaineRsrc( 134 , 1 , ch );
         strcat( ChAux , ch );
         DessineTerme( 0 * 50 , 0 * 120 , 30 , 100 , ChAux );
        _AttendClick();
        _FinEcranGraphique();
    }
    else if(p_lieuDAffichage == _Ecran || p_lieuDAffichage == _Log || Trace ) 
    {
        if( (*p_hdlLemme)->ValTerme.TypeTerme == _MotFlechi )
        {
                strcpy( ch , "Infl. word" );
        }
        else if( (*p_hdlLemme)->ValTerme.TypeTerme == _MotDerive )
        {
                strcpy( ch , "Der. word" );
        }
        else if( (*p_hdlLemme)->ValTerme.TypeTerme == _MotRacine )
        {
                strcpy( ch , "Root" );
        }
        else
        {
                strcpy( ch , "Lemma" );
        }
        sprintf( ChAux , "%s[%.31s(%ld Rac:%ld" , 
                      ch , (*p_hdlLemme)->ValTerme.Ident , 
                      (*p_hdlLemme)->ValTerme.Ref , 
                      (*p_hdlLemme)->ValTerme.RefRacine );
        {
	  TabRefPhraseMem refCou; 
	  
	  refCou = (*p_hdlLemme)->ValTerme.RefSemantique;
	  while( refCou ) {
	    sprintf( ChAux , "%s Sem:%ld" , ChAux , refCou->Ref );
	    refCou = refCou->Suivant;
	  }

	  refCou = (*p_hdlLemme)->ValTerme.RefSelf;
	  while( refCou ) {
	    sprintf( ChAux , "%s Self:%ld" , ChAux , refCou->Ref );
	    refCou = refCou->Suivant;
	  }
	}
        sprintf( ChAux , "%s) # rules: %d]" , 
		 ChAux , (*p_hdlLemme)->NbRegles );

	fprintf( p_ptFicAffichage , "\n%s -> ", ChAux );
    }

    /***                    IMPRESSION DU DAG DE LEMME             ****/
    NbLiensNoeudsMemVar = 0;
    ParcoursDFDag( &((*p_hdlLemme)->ValTerme.ArcsFils) , 
                   NoteLiensNoeudsMemVar , _PreOrdre , _False );
    RenumeroteLiensNoeudsMemVar();
    ProfondeurDagPrec = 3;
    ProfondeurDag = 3;
    Largeur = 0;
    Indentation = p_indentation;
    if( ((*p_hdlLemme)->ValTerme.ArcsFils) )
    {
        EcrireDag( &((*p_hdlLemme)->ValTerme.ArcsFils) , _True , 
                   p_ptFicAffichage , p_lieuDAffichage );
    }
    fprintf( p_ptFicAffichage , ".\n" );

    if( _MachineGraphique && (p_lieuDAffichage == _Ecran) )
    {
       _FinEcranGraphique();
    }

    return _True;
}

/*************************************************************************
            IMPRESSION D'UNE REGLE
*************************************************************************/
char     EcrireRegleMem( TabRegleMem *p_hdlRegle , int p_avecTransformees ,  
                         FILE *p_ptFicAffichage , int p_lieuDAffichage )
{
    ChIdent          ident;
    char             ch[ 64 ];
    char            *ptChIdent;
    char            *ptChType;
    int              indRegleMeta;

    if( !(*p_hdlRegle) )
    {
        return _True;
    }

    LieuDAffichage = p_lieuDAffichage;
    FichierDAffichage = p_ptFicAffichage;
    CompteurAtomesRegle = 0;

    fprintf( FichierDAffichage , "\nRule: " );

    /***  IMPRESSION DE L'ARBRE DE REGLE ****/
    NbLiensNoeudsMemVar = 0;
    ParcoursDFArbre( &((*p_hdlRegle)->TermeFils) , 
                     NoteDFLiensNoeudsMemVar , _PreOrdre );
    RenumeroteLiensNoeudsMemVar();
    ProfondeurTermePrec = 3;
    ProfondeurTerme = 3;
    Largeur = 0;
    Indentation = 0;
    ParcoursDFArbre( &((*p_hdlRegle)->TermeFils) , 
                     EcrireTermeMem , _PreOrdre );
    
    if( (*p_hdlRegle)->Transformee )
    {
        /***   IMPRESSION DE LA REGLE TRANSFORMEE ****/
        fprintf( FichierDAffichage , "Transformed rule:\n" );
        NbLiensNoeudsMemVar = 0;
        ParcoursDFArbre( &((*p_hdlRegle)->Transformee) , 
                         NoteDFLiensNoeudsMemVar , _PreOrdre );
        RenumeroteLiensNoeudsMemVar();
        ProfondeurTermePrec = 3;
        ProfondeurTerme = 3;
        Largeur = 0;
        Indentation = 0;
        ParcoursDFArbre( &((*p_hdlRegle)->Transformee) , 
                         EcrireTermeMem , _PreOrdre );
        fprintf( FichierDAffichage , "\n" );
    }

    if( _MachineGraphique && (LieuDAffichage == _Ecran) )
    {
        _AttendClick();
        _EffaceEcranGraphique();
    }

    if( p_avecTransformees )
    {
        /***   IMPRESSION DES TRANSFORMATIONS DE REGLE ****/
        indRegleMeta = 0;
        while( !((*p_hdlRegle)->Transformee) 
               && GenererMetaRegle( (*p_hdlRegle) , indRegleMeta ) )
        {
            if( ChaineMetaRegle( (*p_hdlRegle)->TermeFils->ValTerme.ArcsFils , 
                                  indRegleMeta , &ptChIdent , &ptChType) )
            {
                   fprintf( FichierDAffichage , "Metarule %s/%s (%d) :\n" , 
                                          ptChIdent , ptChType ,
                                          indRegleMeta + 1 );
            }
            else
            {
                   fprintf( FichierDAffichage , "Metarule ???? (%d) :\n" , 
                                          indRegleMeta + 1 );
            }

            if( _MachineGraphique && (LieuDAffichage == _Ecran) )
            {
                TxtChaineRsrc( 134 , 6 , ch );
                DessineTerme( 2 * 50 , 0 * 120 , 30 , 100 , ident );
            }

            ParcoursDFArbre( &((*p_hdlRegle)->Transformee) , 
                             GarnirChaineTermeMem , _PreOrdre );

            NbLiensNoeudsMemVar = 0;
            ParcoursDFArbre( &((*p_hdlRegle)->Transformee) , 
                             NoteDFLiensNoeudsMemVar , _PreOrdre );
            RenumeroteLiensNoeudsMemVar();
            ProfondeurTermePrec = 3;
            ProfondeurTerme = 3;
            Largeur = 0;
            Indentation = 0;
            ParcoursDFArbre( &((*p_hdlRegle)->Transformee) , 
                             EcrireTermeMem , _PreOrdre );
            fprintf( FichierDAffichage , "\n" );
            DemarqueEtRendArbre( &((*p_hdlRegle)->Transformee) );
            (*p_hdlRegle)->Transformee = NULL;
            
            if( _MachineGraphique && (LieuDAffichage == _Ecran) )
            {
                 _AttendClick();
                 _EffaceEcranGraphique();
                 strcat( ChAux , (*p_hdlRegle)->TermeFils->ValTerme.Ident );
                 DessineTerme( 0 * 50 , 0 * 120 , 30 , 100 , ChAux );
             }

            indRegleMeta++;
        }
    }

    return _True;
}

/*************************************************************************
            IMPRESSION D'UNE LISTE DE REGLES
*************************************************************************/
void     EcrireListeRegleMem( TabRegleMem p_ptListeRegle , 
                              int p_avecTransformees ,  
                              FILE *p_ptFicAffichage , int p_lieuDAffichage )
{
  if( !ParcoursListeRegles( &p_ptListeRegle , GarnirRefRegleMem ) )
  {
      return;
  }

  if( _MachineGraphique && p_lieuDAffichage == _Ecran )
  {
       _EcranGraphique();
  }

  while( p_ptListeRegle )
  {
       EcrireRegleMem( &p_ptListeRegle , p_avecTransformees , 
                       p_ptFicAffichage , p_lieuDAffichage  );
       p_ptListeRegle = p_ptListeRegle->RegleFrere;
  }

  if( _MachineGraphique && p_lieuDAffichage == _Ecran )
  {
       _FinEcranGraphique();
  }
}

/*************************************************************************
            ECRITURE D'UNE LISTE DE REGLES
**************************************************************************/
void     EcrireReglesDUnMot( ChIdent p_chMot , FILE *p_ptFicAffichage , 
                             int p_lieuDAffichage )
{
    unsigned long                   indAbsLemmeCou;
    unsigned long                   indAbsRegleCou;
    unsigned long                   longFic;
    struct LemmeMem                 lemmeCou;
    TabLemmeMem                     ptLemmeCou;
    char                            egale;
    TabRegleMem                     ptListeRegle;

    /* description du terme releve dans le dico */
    char                            chaineAnnexe[_MaxLetIdent];

    fseek( FicBDicLemmes , 0L , 2 );
    fgetpos( FicBDicLemmes , &longFic );
    *(lemmeCou.ValTerme.Ident) = '\0';
    indAbsLemmeCou = IndiceInsertionDicLemmes( p_chMot , &egale );

    if( EgaliteChaine( BufferLemmesCourant + _IRelLemme( indAbsLemmeCou ) , 
                       p_chMot ) == _True )
    {
         fprintf( FicLog , "Lemme trouve : %s\n" , p_chMot );
    }
    else
    {
         fprintf( FicLog , "Lemme non trouve : %s\n" , p_chMot );
         return;
    }

    while( indAbsLemmeCou < longFic
           && (EgaliteChaine( BufferLemmesCourant 
                    + _IRelLemme( indAbsLemmeCou ) , p_chMot ) == _True) )
    {
        if( RestructurerBDicLemme( &indAbsLemmeCou , lemmeCou.ValTerme.Ident  ,
                                   &lemmeCou , chaineAnnexe ) == _Nil 
            || EgaliteChaine( lemmeCou.ValTerme.Ident , DernLemme ) )
        {
            DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
            break;
        }
        ptLemmeCou = &lemmeCou;

        EcrireLemmeMem( &ptLemmeCou , p_ptFicAffichage , 
                        p_lieuDAffichage , 0 );

        indAbsRegleCou = lemmeCou.ValTerme.PosFicRegle;
        if( RestructurerLesRegles( &indAbsRegleCou , &ptListeRegle , 
                                   lemmeCou.NbRegles , _True ) == _Nil )
        {
            DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
            DemarqueEtRendListeRegle( &ptListeRegle );
            break;
        }

        /* generation de toutes les regles d'un mot */
        if( ptListeRegle )
        {
            EcrireListeRegleMem( ptListeRegle , _True ,
                                 p_ptFicAffichage , p_lieuDAffichage );
            DemarqueEtRendListeRegle( &ptListeRegle );
        }
        DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
    }
}

/*************************************************************************
            IMPRESSION D'UNE META-REGLE
*************************************************************************/
char     EcrireMetaRegle( TabRegleMemMeta *p_hdlRegleMeta , int p_indRegle )
{
    fprintf( FichierDAffichage , "Metarule %.31s (%d) : " ,
             Tableaux[ (*p_hdlRegleMeta)->NoTableau ].Ident , p_indRegle );

    NbLiensNoeudsMemVar = 0;
    ParcoursDFArbre( &((*p_hdlRegleMeta)->TermeFils) , 
                     NoteDFLiensNoeudsMemVar , _PreOrdre );
    RenumeroteLiensNoeudsMemVar();

    /***   IMPRESSION DE L'IDENTIFICATEUR DE REGLE ****/
    ProfondeurTermePrec = 0;
    ProfondeurTerme = 0;
    Largeur = 0;
    Indentation = 0;
    EcrireTermeMem( &((*p_hdlRegleMeta)->TermeFils) );

    /***  IMPRESSION DES 2 ARBRES DE META-REGLE ****/
    ProfondeurTermePrec = 2;
    ProfondeurTerme = 2;
    Largeur = 0;
    Indentation = 0;
    ParcoursDFArbre( &((*p_hdlRegleMeta)->TermeFils->TermeFils) , 
                     EcrireTermeMem , _PreOrdre );
    fprintf( FichierDAffichage , "\n" );

    ProfondeurTermePrec = 5;
    ProfondeurTerme = 5;
    Largeur = 0;
    Indentation = 0;
    ParcoursDFArbre( &((*p_hdlRegleMeta)->TermeFils->TermeFrere) , 
                     EcrireTermeMem , _PreOrdre );
    fprintf( FichierDAffichage , "\n" );

    if( _MachineGraphique && (LieuDAffichage == _Ecran) )
    {
        _AttendClick();
        _EffaceEcranGraphique();
    }

    return _True;
}

/*************************************************************************
            IMPRESSION D'UN TABLEAU DE META-REGLES
*************************************************************************/
void     EcrireListeRegleMeta( void )
{
    int                             noTableau;
    TabRegleMemMeta                 ptRegleMeta;
    int                             indRegle;

    LieuDAffichage = _Sortie;
    FichierDAffichage = stdout;

    for( noTableau = 0 ; noTableau < _NbMaxTableau ; noTableau++ ) { 

      if( *(Tableaux[ noTableau ].Ident) ) {
	fprintf( FichierDAffichage , "Table [%s]\n\n" , 
		 Tableaux[ noTableau ].Ident );
	ptRegleMeta = Tableaux[ noTableau ].PtRegleMeta;
	indRegle = 1;

	while( ptRegleMeta ) {
	  EcrireMetaRegle( &ptRegleMeta , indRegle++ );
	  ptRegleMeta = ptRegleMeta->RegleFrere;
	}
      }
      else {
	break;
      }
    }
}

/*************************************************************************
            IMPRESSION DES TRAITS ET DES CATEGORIES LEXICALES
*************************************************************************/
void     EcrireTraitsCatLex( void )
{
    int                  ind;

    fprintf( FicLog , "Traits et categories lexicales\n" );
    for( ind = 0 ; ind < NbMaxTraitsCatLex ; ind++ )
    {
        if( *(DicoTraitsCatLex[ ind ].ChaineDico) )
        {
             fprintf( FicLog , "%.31s [%d]\n" , 
                      DicoTraitsCatLex[ ind ].ChaineDico , ind );
        }
    }
    fprintf( FicLog , "\n\n" );
}

/*************************************************************************
            IMPRESSION DES TRAITS ET DES CATEGORIES LEXICALES
*************************************************************************/
void     EcrireListeLemme( FILE *p_ptFicSortie , int p_typeIndexation ,
                           TabLemmeMem p_ptLemme )
{
    int                catLexLemme;
    int                typeTermePrec;
    TabLemmeMem        ptLemmeCou;
    TabLemmeMem        ptLemmeAux;

    typeTermePrec = _Nil;
    ptLemmeCou = p_ptLemme;
    /* parcours de tous les lemmes attaches au mot courant */
    while( ptLemmeCou )
    {
       /* etiquetage avec la categorie lexicale seule */
       if( p_typeIndexation == _Etiquetage )
       {
           char connu = _False;
    
           catLexLemme = CategorieLexicale( 
                           &(ptLemmeCou->ValTerme.ArcsFils) );
           {
              ptLemmeAux = p_ptLemme;
              while( ptLemmeAux && ptLemmeAux != ptLemmeCou )
              {
                  connu |= ( catLexLemme
                    == CategorieLexicale( 
                           &(ptLemmeAux->ValTerme.ArcsFils) ) );
                  ptLemmeAux = ptLemmeAux->LemmeFrere;
              }
           }
           if( !connu )
           {
              if( catLexLemme == RefCatMotsInc )
              {
                 fprintf( p_ptFicSortie , "\\X" );
              }
              else
              {
                 fprintf( p_ptFicSortie , "\\%s" , 
                       DicoTraitsCatLex[ catLexLemme ].ChaineDico );
              }
           }
       } 
       /* etiquetage avec la categorie lexicale seule */
       else 
       /* p_typeIndexation == _EtiquetageFin */
       /* etiquetage avec les traits complets */
       {
           switch( ptLemmeCou->ValTerme.TypeTerme )
           {
              case _MotFlechi:
                EcrireLemmeMem( &ptLemmeCou , p_ptFicSortie , 
                                _Sortie , 0 );
                break;
              case _MotRacine:
                fprintf( p_ptFicSortie , "{{" );
                EcrireLemmeMem( &ptLemmeCou , p_ptFicSortie , 
                                _Sortie , 2 );
                fprintf( p_ptFicSortie , "}}\n" );
                break;
              case _MotSemantLie:
                if( typeTermePrec != _MotSemantLie )
                {
                   fprintf( p_ptFicSortie , "<<" );
                }
                EcrireLemmeMem( &ptLemmeCou , p_ptFicSortie , 
                                _Sortie , 2 );
                if( !ptLemmeCou->LemmeFrere 
                    || ptLemmeCou->LemmeFrere->ValTerme.TypeTerme
                       != _MotSemantLie )
                {
                   fprintf( p_ptFicSortie , ">>\n" );
                }
                else
                {
                   fprintf( p_ptFicSortie , "  +" );
                }
                break;
              case _MotDerive:
                if( typeTermePrec != _MotDerive )
                {
                   fprintf( p_ptFicSortie , "  [[" );
                }
                EcrireLemmeMem( &ptLemmeCou , p_ptFicSortie , 
                                _Sortie , 4 );
                if( !ptLemmeCou->LemmeFrere 
                    || ptLemmeCou->LemmeFrere->ValTerme.TypeTerme
                       != _MotDerive )
                {
                   fprintf( p_ptFicSortie , "  ]]\n" );
                }
                else
                {
                   fprintf( p_ptFicSortie , "    +" );
                }
                break;
              case _MotSemantLieRec:
                if( typeTermePrec != _MotSemantLieRec )
                {
                   fprintf( p_ptFicSortie , "  ||" );
                }
                EcrireLemmeMem( &ptLemmeCou , p_ptFicSortie , 
                                _Sortie , 4 );
                if( !ptLemmeCou->LemmeFrere 
                    || ptLemmeCou->LemmeFrere->ValTerme.TypeTerme
                       != _MotSemantLieRec )
                {
                   fprintf( p_ptFicSortie , "  ||\n" );
                }
                else
                {
                   fprintf( p_ptFicSortie , "    +" );
                }
                break;
              default:
		sprintf( ChAux , "[%d] EcrireListeLemme(alecrire)" ,
			 ptLemmeCou->ValTerme.TypeTerme );
                TxtErreur( 269 , ChAux );
                return;
           }
           typeTermePrec = ptLemmeCou->ValTerme.TypeTerme;
       }
       /* etiquetage avec les traits complets */

       ptLemmeCou = ptLemmeCou->LemmeFrere;
    }
    /* parcours de tous les lemmes attaches au mot courant */
}

/*************************************************************************
            IMPRESSION DES TRAITS ET DES CATEGORIES LEXICALES
*************************************************************************/
void     EcrirePhraseEtiquetee( FILE *p_ptFicSortie , int p_typeIndexation )
{
    TabLemmeMem                ptLemmeCou;
    TabMotMem                  ptMotCou;

    /* parcours de tous les mots de la phrase */
    ptMotCou = MotsPhrase;
    while( ptMotCou )
    {
        fprintf( p_ptFicSortie , "\n'%s'" , ptMotCou->ChaineMot );
        ptLemmeCou = ptMotCou->PtListeLemmes;

        /* cas d'un mot sans lemme */
        if( !ptLemmeCou && p_typeIndexation == _EtiquetageFin )
        {
           fprintf( p_ptFicSortie , ".\n" );
        }

        EcrireListeLemme( p_ptFicSortie , p_typeIndexation , ptLemmeCou );

	ptMotCou = ptMotCou->MotFrere;
    }
    /* parcours de tous les mots de la phrase */
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            fin de alecrire.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
