/*
 * FASt Term Recognizer
 *
 *  FASTR/alentete.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for reading the language environment file
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*****************************************************************************
                 LECTURE D'UN CHEMINSPECIFIQUE DANS L'EN-TETE
*****************************************************************************/
char     LireUnCheminSpecifique( TabDagMem *p_chemin , char *p_chChemin , 
                                 FILE *p_ptFicHeader , char *p_carLu ,
                                 unsigned long *p_posFic , 
				 int p_cheminsMultiples , int p_nbMaxChemins )
{
    int    rangChemin = 0;
    int    typeDernierArc;

    fgetpos( p_ptFicHeader , p_posFic );

    /* le nom du chemin */
    if( !ReconnaitreChaine ( p_ptFicHeader , p_chChemin , p_carLu ) ) {
         return _False;
    }

    /* le mot path ou path(s) */
    if( !p_cheminsMultiples ) {
      if( !ReconnaitreChaine ( p_ptFicHeader , "path" , p_carLu ) ) {
	return _False;
      }
    }
    else {
      if( !ReconnaitreChaine ( p_ptFicHeader , "path(s)" , p_carLu ) ) {
	return _False;
      }
    }

    /* les : avant le(s) chemin(s) */
    if( !ReconnaitreCaractere( p_ptFicHeader , (int)':' , 
                                  p_carLu , _True ) )
    {
         return _False;
    }

    /* la lecture du chemin proprement dite */
    if( !p_cheminsMultiples ) {
      if( !LireUnChemin( p_carLu , p_ptFicHeader , p_chemin , _False ,
			 &typeDernierArc ) ) {
	return _False;
      }
    }
    /* possibilite de chemins multiples */
    else {
      /* lecture des chemins successifs */
      while( *p_carLu == '<' && rangChemin < _NbMaxTypesLiensSem ) {
	if( !LireUnChemin( p_carLu , p_ptFicHeader , 
			   p_chemin + rangChemin , _False ,
			   &typeDernierArc ) ) {
	  return _False;
	}
       rangChemin++;
      }

      /* parametre de taille insuffisant */
      if( rangChemin >= p_nbMaxChemins ) {
	TxtChaineErreur( 284 , ChErreurCompil );
	return _False;
      }
      /*  mise a zero des chemins restants */
      else {
	while( rangChemin < p_nbMaxChemins ) {
	  *(p_chemin + rangChemin) = NULL;
          rangChemin++;
	}
      }
    }
    /* possibilite de chemins multiples */
	
    if( !ReconnaitreCaractere( p_ptFicHeader , (int)'.' , p_carLu , _True ) )
    {
         return _False;
    }
    return _True;
}

/*****************************************************************************
                 LECTURE D'UNE LISTE DE CARACTERES SPECIAUX
*****************************************************************************/
char     LireUneListeDeCaracteres( char *p_listeCar , char *p_chRubrique , 
                                   FILE *p_ptFicHeader , char *p_carLu , 
                                   int p_carFinal , unsigned long *p_posFic )
{
    int                        ind;

    fgetpos( p_ptFicHeader , p_posFic );
    if( !ReconnaitreChaine ( p_ptFicHeader , p_chRubrique , p_carLu )
        || !ReconnaitreCaractere( p_ptFicHeader , (int)':' , p_carLu , _True ) )
    {
         return _False;
    }
    ind = 0;
    while( *p_carLu != EOF && *p_carLu != (char)p_carFinal 
           && ind < _MaxCarSpeciaux )
    {
        p_listeCar[ ind ] = *p_carLu;
        *p_carLu = LireCar( p_ptFicHeader );
        while( *p_carLu == ' ' )
        {
            /* ne pas utiliser SautBlanc qui peut sauter des
               caracteres speciaux tels que les ""         */
            *p_carLu = LireCar( p_ptFicHeader );
        }
        ind++;
    }
    if( ind == _MaxCarSpeciaux )
    {
         TxtChaineErreur( 158 , ChErreurCompil );
         return _False;
    }
    p_listeCar[ ind ] = '\0';
    if( !ReconnaitreCaractere( p_ptFicHeader , (char)p_carFinal , 
                               p_carLu , _True ) )
    {
         return _False;
    }
    return _True;
}

/*****************************************************************************
                 LECTURE DE L'EN-TETE DE LA GRAMMAIRE
*****************************************************************************/
char     LitEnTeteGrammaireEtErreur( void )
{
    FILE                           *ptFicHeader;
    unsigned long                   posFic;
    char                            carLu = ' ';

    if( !OuvreFichier( NomFicHeader , &ptFicHeader , _FopenRT ) )
    {
         return _False;
    }

    if( !LitEnTeteGrammaire( ptFicHeader , &posFic ) )
    {
         MessageErreurCompil( posFic , &carLu , ptFicHeader );
         fclose( ptFicHeader );
         return _False;
    }
    else
    {
         fclose( ptFicHeader );
         return _True;
    }
}

char     LitEnTeteGrammaire( FILE *p_ptFicHeader , unsigned long *p_posFic )
{
    char                        carLu;
    ChIdent                     ident;
    ChIdent                     versionfastrCou;
    ChIdent                     datefastrCou;
    int                         catLex;
    int                         nbReglesFausses;
    int                        *catLexSuffixees;
    int                         nbCatLexSuffixee;
    int                         indCatLexSuffixee;
    int                         indSuffixagesTot;
    int                         nbSuffixagesTot;  /* un suffixage == 1
                                                     flexion ou 1 deriv. */
    int                         nbSuffixesTot;
    int                         nbNoSuffixages;
    int                         ind;
    char                        estFlexion;

    IniTraitsCatLex();
    catLexSuffixees = (int *)_Alloc( NbMaxTraitsCatLex * sizeof(int) );

    /*---------------------------------------------------------------
                   lecture des categories lexicales
      ---------------------------------------------------------------*/
    fgetpos( p_ptFicHeader , p_posFic );
    carLu = LireCar( p_ptFicHeader );
    SautBlanc( p_ptFicHeader , &carLu );

    if( !LireVersionEtDate ( p_ptFicHeader , &carLu ,
			     versionfastrCou , datefastrCou ) )
    {
         return _False;
    }

    if( !ReconnaitreChaine ( p_ptFicHeader , "Categories" , &carLu ) 
        || !ReconnaitreCaractere( p_ptFicHeader , (int)':' , &carLu , _True ) )
    {
         return _False;
    }
    while( carLu != '.' && carLu != EOF )
    {
         LireIdentificateur( p_ptFicHeader , &carLu , ident , _MaxLetIdent ,
                             _True );
         if( !ChercheOuInsereTraitsCatLex( ident , _CatLex , &catLex ) )
	 {
             return _False;
         }
    }
    /* on ajoute une categorie vide en plus */
    if( !ChercheOuInsereTraitsCatLex( "X" , _CatLex , &catLex ) )
    {
        return _False;
    }
    RefCatLexVide = (long)catLex;
    if( !ReconnaitreCaractere( p_ptFicHeader , (int)'.' , &carLu , _True ) )
    {
         return _False;
    }

    /*---------------------------------------------------------------
                   lecture de categories particulieres
      ---------------------------------------------------------------*/
    fgetpos( p_ptFicHeader , p_posFic );
    if( !ReconnaitreChaine ( p_ptFicHeader , "Unknown category" , &carLu ) 
        || !ReconnaitreCaractere( p_ptFicHeader , (int)':' , &carLu , _True ) )
    {
         return _False;
    }
    LireIdentificateur( p_ptFicHeader , &carLu , ident , _MaxLetIdent ,
                        _True );
    if( !ChercheOuInsereTraitsCatLex( ident , _CatLex , &catLex ) )
    {
         return _False;
    }
    RefCatMotsInc = (long)catLex;
    if( !ReconnaitreCaractere( p_ptFicHeader , (int)'.' , &carLu , _True ) )
    {
         return _False;
    }

    /*---------------------------------------------------------------
                   the major categories (non empty words)
      ---------------------------------------------------------------*/
    fgetpos( p_ptFicHeader , p_posFic );
    NbMajorCategories = 0;
    if( !ReconnaitreChaine ( p_ptFicHeader , "Major" , &carLu )
        || !ReconnaitreChaine ( p_ptFicHeader , "categories" , &carLu )
        || !ReconnaitreCaractere( p_ptFicHeader , (int)':' , &carLu , _True ) )
    {
         return _False;
    }
    while( carLu != '.' && carLu != EOF 
	   && NbMajorCategories < _MaxMajorCategories )
    {
         LireIdentificateur( p_ptFicHeader , &carLu , ident , _MaxLetIdent ,
                             _True );
	 if( !ChercheOuInsereTraitsCatLex( ident , _CatLex , &catLex ) ) {
	   return _False;
	 }
	 RefMajorCategories[ NbMajorCategories ] = (long)catLex;
	 NbMajorCategories++;
    }
    if( NbMajorCategories == _MaxMajorCategories )
    {
         TxtChaineErreur( 283 , ChErreurCompil );
         return _False;
    }
    if( !ReconnaitreCaractere( p_ptFicHeader , (int)'.' , &carLu , _True ) )
    {
         return _False;
    }

    fgetpos( p_ptFicHeader , p_posFic );
    if( !ReconnaitreChaine ( p_ptFicHeader , "Numeral" , &carLu ) 
        || !ReconnaitreCaractere( p_ptFicHeader , (int)':' , &carLu , _True ) )
    {
         return _False;
    }
    LireIdentificateur( p_ptFicHeader , &carLu , ident , _MaxLetIdent ,
                        _True );
    if( !ChercheOuInsereTraitsCatLex( ident , _CatLex , &catLex ) )
    {
         return _False;
    }
    RefDeterminantNumeral = (long)catLex;
    if( !ReconnaitreCaractere( p_ptFicHeader , (int)'.' , &carLu , _True ) )
    {
         return _False;
    }

    fgetpos( p_ptFicHeader , p_posFic );
    if( !ReconnaitreChaine ( p_ptFicHeader , "Punctuation" , &carLu ) 
        || !ReconnaitreCaractere( p_ptFicHeader , (int)':' , &carLu , _True ) )
    {
         return _False;
    }
    LireIdentificateur( p_ptFicHeader , &carLu , ident , _MaxLetIdent ,
                        _True );
    if( !ChercheOuInsereTraitsCatLex( ident , _CatLex , &catLex ) )
    {
         return _False;
    }
    RefPonctuation = (long)catLex;
    if( !ReconnaitreCaractere( p_ptFicHeader , (int)'.' , &carLu , _True ) )
    {
         return _False;
    }

    fgetpos( p_ptFicHeader , p_posFic );
    if( !ReconnaitreChaine ( p_ptFicHeader , "Final" , &carLu ) 
        || !ReconnaitreChaine ( p_ptFicHeader , "punctuation" , &carLu ) 
        || !ReconnaitreCaractere( p_ptFicHeader , (int)':' , &carLu , _True ) )
    {
         return _False;
    }
    LireIdentificateur( p_ptFicHeader , &carLu , ident , _MaxLetIdent ,
                        _True );
    if( !ChercheOuInsereTraitsCatLex( ident , _CatLex , &catLex ) )
    {
         return _False;
    }
    RefPonctuationFinale = (long)catLex;
    if( !ReconnaitreCaractere( p_ptFicHeader , (int)'.' , &carLu , _True ) )
    {
         return _False;
    }

    /*---------------------------------------------------------------
                   lecture des traits
      ---------------------------------------------------------------*/
    fgetpos( p_ptFicHeader , p_posFic );
    if( !ReconnaitreChaine ( p_ptFicHeader , "Features" , &carLu )
        || !ReconnaitreCaractere( p_ptFicHeader , (int)':' , &carLu , _True ) )
    {
         return _False;
    }
    while( carLu != '.' && carLu != EOF )
    {
         LireIdentificateur( p_ptFicHeader , &carLu , ident , _MaxLetIdent ,
                             _True );
         if( !ChercheOuInsereTraitsCatLex( ident , _Trait , &catLex ) )
         {
             return _False;
         }
    }
    if( !ReconnaitreCaractere( p_ptFicHeader , (int)'.' , &carLu , _True ) )
    {
         return _False;
    }

    /*---------------------------------------------------------------
                   lecture d'atomes particuliers
      ---------------------------------------------------------------*/
    fgetpos( p_ptFicHeader , p_posFic );
    if( !ReconnaitreChaine ( p_ptFicHeader , "Singular" , &carLu )
        || !ReconnaitreCaractere( p_ptFicHeader , (int)':' , &carLu , _True ) )
    {
         return _False;
    }
    LireIdentificateur( p_ptFicHeader , &carLu , ident , _MaxLetIdent ,
                        _True );
    if( !IdentATraitsCatLex( ident , _Trait , &ind ) )
    {
         return _False;
    }
    RefSingulier = (long)ind;
    if( !ReconnaitreCaractere( p_ptFicHeader , (int)'.' , &carLu , _True ) )
    {
         return _False;
    }

    fgetpos( p_ptFicHeader , p_posFic );
    if( !ReconnaitreChaine ( p_ptFicHeader , "Plural" , &carLu )
        || !ReconnaitreCaractere( p_ptFicHeader , (int)':' , &carLu , _True ) )
    {
         return _False;
    }
    LireIdentificateur( p_ptFicHeader , &carLu , ident , _MaxLetIdent ,
                        _True );
    if( !IdentATraitsCatLex( ident , _Trait , &ind ) )
    {
         return _False;
    }
    RefPluriel = (long)ind;
    if( !ReconnaitreCaractere( p_ptFicHeader , (int)'.' , &carLu , _True ) )
    {
         return _False;
    }

    /*---------------------------------------------------------------
                   lecture des chemins predefinis
      ---------------------------------------------------------------*/
    if( !LireUnCheminSpecifique( &CheminCategorie , "Category" , 
                                 p_ptFicHeader , &carLu , 
				 p_posFic , _False , 0 )
     || !LireUnCheminSpecifique( &CheminEtiquette , "Label" , 
                                 p_ptFicHeader , &carLu , 
				 p_posFic , _False , 0 )
     || !LireUnCheminSpecifique( &CheminLexicalisation , "Lexicalisation" , 
                                 p_ptFicHeader , &carLu , 
				 p_posFic , _False , 0 )
     || !LireUnCheminSpecifique( &CheminLemme , "Lemma" , 
                                 p_ptFicHeader , &carLu , 
				 p_posFic , _False , 0 )
     || !LireUnCheminSpecifique( &CheminFlexion , "Inflection" , 
                                 p_ptFicHeader , &carLu ,
				  p_posFic , _False , 0 )
     || !LireUnCheminSpecifique( &CheminDerivation , "Derivation" , 
                                 p_ptFicHeader , &carLu , 
				 p_posFic , _False , 0 )
     || !LireUnCheminSpecifique( &CheminForme , "Form" , 
                                 p_ptFicHeader , &carLu , 
				 p_posFic , _False , 0 )
     || !LireUnCheminSpecifique( &CheminNombre , "Number" , 
                                 p_ptFicHeader , &carLu , 
				 p_posFic , _False , 0 )
     || !LireUnCheminSpecifique( &CheminReference , "Reference" , 
                                 p_ptFicHeader , &carLu , 
				 p_posFic , _False , 0 )
     || !LireUnCheminSpecifique( &CheminEtiquetteMeta , "Meta label" , 
                                 p_ptFicHeader , &carLu , 
				 p_posFic , _False , 0 )
     || !LireUnCheminSpecifique( &CheminRacinesAnnexes , "Auxiliary stem" , 
                                 p_ptFicHeader , &carLu , 
				 p_posFic , _False , 0 )
     || !LireUnCheminSpecifique( &CheminRacine , "Root" , 
                                 p_ptFicHeader , &carLu , 
				 p_posFic , _False , 0 )
     || !LireUnCheminSpecifique( &CheminLeadingExprReg , "Leading reg ex" , 
                                 p_ptFicHeader , &carLu ,
				 p_posFic , _False , 0 )
     || !LireUnCheminSpecifique( &CheminTrailingExprReg , "Trailing reg ex" , 
                                 p_ptFicHeader , &carLu ,
				 p_posFic , _False , 0 )
     || !LireUnCheminSpecifique( CheminsSemantique , "Semantic" , 
                                 p_ptFicHeader , &carLu , 
				 p_posFic , _True , _NbMaxTypesLiensSem )
     || !LireUnCheminSpecifique( &CheminSelf , "Self" , 
                                 p_ptFicHeader , &carLu ,
				 p_posFic , _False , 0 ) )
    {
         return _False;
    }

    /*---------------------------------------------------------------
        lecture des caracteres speciaux acceptes dans les nombres
      ---------------------------------------------------------------*/
    if( !LireUneListeDeCaracteres( CarSpeciauxMots , "Non alphanumerical" ,
                                   p_ptFicHeader , &carLu , 'z' , p_posFic )
        || !LireUneListeDeCaracteres( Ponctuation , "Punctuation" ,
                                   p_ptFicHeader , &carLu , 'z' , p_posFic )
        || !LireUneListeDeCaracteres( PonctuationFinale , "Final punctuation" ,
                                   p_ptFicHeader , &carLu , 'z' , p_posFic ) )
    {
        return _False;
    }

    /*---------------------------------------------------------------
        lecture des caracteristiques flexionnelles et derivationnelles
      ---------------------------------------------------------------*/
    /* initialisation des Suffixages (flexions & derivations) */
    nbSuffixagesTot = 0;
    for( indSuffixagesTot = 0 ; indSuffixagesTot < _NbMaxSuffixages ; 
         indSuffixagesTot++ )
    {
       *( TabDagsSuffixage + indSuffixagesTot ) = NULL;
    }
    indSuffixagesTot = 0;

    /* initialisation des suffixes */
    nbSuffixesTot = 0;
    *(Suffixes[ 0 ].Suff) = '#';

    estFlexion = _True; /* _False correspond a derivation */
    while( _True ) {
      fgetpos( p_ptFicHeader , p_posFic );
      if( ( estFlexion 
            && !ReconnaitreChaine ( p_ptFicHeader , "Inflected" , &carLu ) )
       || ( !estFlexion 
            && !ReconnaitreChaine ( p_ptFicHeader , "Derived" , &carLu ) ) )
      {
          return _False;
      }
      if( !ReconnaitreChaine ( p_ptFicHeader , "categories" , &carLu )
          || !ReconnaitreCaractere( p_ptFicHeader , (int)':' , &carLu , _True ) )
      {
          return _False;
      }

      /*---------------------------------------------------------------
           lecture des categories flechies et de leur nbre de flexions
        ou lecture des categories derivees et de leur nbre de derivations
        ---------------------------------------------------------------*/
      nbCatLexSuffixee = 0;
      while( carLu != '.' && carLu != EOF )
      {
           if( nbCatLexSuffixee >= NbMaxTraitsCatLex )
           {
               TxtChaineErreur( 96 , ChErreurCompil );
               return _False;
           }
           LireIdentificateur( p_ptFicHeader , &carLu , ident , _MaxLetIdent ,
                               _True );
           if( !IdentATraitsCatLex( ident , _CatLex , &catLex ) )
           {
               return _False;
           }
           if( !ReconnaitreCaractere( p_ptFicHeader , (int)'(' , &carLu , _True ) )
           {
                return _False;
           }

           /* lecture du nb de flexions ou du nb de derivations de la
              categorie concernee */
           catLexSuffixees[ nbCatLexSuffixee++ ] = catLex;
           if( estFlexion )
           {
              DicoTraitsCatLex[ catLex ].NbFlexions 
                          = (int)LireEntier( p_ptFicHeader , &carLu );
              nbSuffixagesTot += DicoTraitsCatLex[ catLex ].NbFlexions;
           }
           else
           {
              DicoTraitsCatLex[ catLex ].NbDerivations 
                          = (int)LireEntier( p_ptFicHeader , &carLu );
              nbSuffixagesTot += DicoTraitsCatLex[ catLex ].NbDerivations;
           }

           if( nbSuffixagesTot >= _NbMaxSuffixages )
           {
               TxtChaineErreur( 97 , ChErreurCompil );
               return _False;
           }
           if( !ReconnaitreCaractere( p_ptFicHeader , (int)')' , &carLu , _True ) )
           {
                return _False;
           }
      }
      if( !ReconnaitreCaractere( p_ptFicHeader , (int)'.' , &carLu , _True ) )
      {
           return _False;
      }

      /*---------------------------------------------------------------
                     lecture des contraintes de flexions
        ---------------------------------------------------------------*/
      fgetpos( p_ptFicHeader , p_posFic );
      for( indCatLexSuffixee = 0 ; indCatLexSuffixee < nbCatLexSuffixee ; 
           indCatLexSuffixee++ )
      {
          catLex = catLexSuffixees[ indCatLexSuffixee ];

          /* on note la position du premier dags pour l'utiliser lors
             de l'analyse du mot */
          if( estFlexion )
          {
              DicoTraitsCatLex[ catLex ].RangPremDagFlexion 
                    = indSuffixagesTot;
          }
          else
          {
              DicoTraitsCatLex[ catLex ].RangPremDagDerivation 
                    = indSuffixagesTot;
          }

          /* on lit les dags */
          if( estFlexion )
          {
             if( !LireContraintesSuffixages( p_ptFicHeader , &carLu , 
                               DicoTraitsCatLex[ catLex ].NbFlexions , 
                               DicoTraitsCatLex[ catLex ].ChaineDico ,
                               &indSuffixagesTot ) )
             {
                 return _False;
             }
          }
          else
          {
             if( !LireContraintesSuffixages( p_ptFicHeader , &carLu , 
                               DicoTraitsCatLex[ catLex ].NbDerivations , 
                               DicoTraitsCatLex[ catLex ].ChaineDico ,
                               &indSuffixagesTot ) )
             {
                 return _False;
             }
          }
      }

      /*---------------------------------------------------------------
               lecture des suffixes de flexion ou de derivation
        ---------------------------------------------------------------*/
      fgetpos( p_ptFicHeader , p_posFic );
      for( indCatLexSuffixee = 0 ; indCatLexSuffixee < nbCatLexSuffixee ; 
           indCatLexSuffixee++ )
      {
          catLex = catLexSuffixees[ indCatLexSuffixee ];
          /* on lit les suffixes */
          if( estFlexion )
          {
             DicoTraitsCatLex[ catLex ].RangPremSuffixeFlexion 
                               = nbSuffixesTot;
             if( !LireSuffixesSuffixages( p_ptFicHeader , p_posFic , &carLu , 
                               DicoTraitsCatLex[ catLex ].NbFlexions , 
                               DicoTraitsCatLex[ catLex ].ChaineDico ,
                               &nbNoSuffixages , &nbSuffixesTot ) )
             {
                 return _False;
             }
             DicoTraitsCatLex[ catLex ].NbNoTypeFlexion = nbNoSuffixages;
          }
          else
          {
             DicoTraitsCatLex[ catLex ].RangPremSuffixeDerivation 
                               = nbSuffixesTot;
             if( !LireSuffixesSuffixages( p_ptFicHeader , p_posFic , &carLu , 
                               DicoTraitsCatLex[ catLex ].NbDerivations , 
                               DicoTraitsCatLex[ catLex ].ChaineDico ,
                               &nbNoSuffixages , &nbSuffixesTot ) )
             {
                 return _False;
             }
             DicoTraitsCatLex[ catLex ].NbNoTypeDerivation = nbNoSuffixages;
          }
      }
      if( estFlexion )
      {
         estFlexion = _False; /* _False correspond a derivation */
      }
      else
      {
         break; /* 1 la flexion, 2  a derivation */
      }
    }
    ConstruitALGSuffixes();
    SautBlanc( p_ptFicHeader , &carLu );

    /**************************************************************
     Lecture des conversions d'etiquettes en gene autom de regles
    ***********************************************************/
/*    fgetpos( p_ptFicHeader , p_posFic );*/
/*    if( !ReconnaitreChaine ( p_ptFicHeader , "Lexical" , &carLu )*/
/*        || !ReconnaitreChaine ( p_ptFicHeader , "category" , &carLu )*/
/*        || !ReconnaitreChaine ( p_ptFicHeader , "conversion" , &carLu )*/
/*        || !ReconnaitreCaractere( p_ptFicHeader , (int)':' , &carLu , _True ) )*/
/*    {*/
/*        return _False;*/
/*    }*/
/**/
/*    ind = 0;*/
/*    while( carLu != EOF && carLu != '#' && ind < _DerCatLex )*/
/*    {*/
/*         LireIdentificateur( p_ptFicHeader , &carLu , */
/*                             DicoTransfert[ ind ].chIni ,*/
/*                             _MaxLetDicoEnum - 1 , _True );*/
/*         if( carLu == '#' )*/
/*         {*/
/*             break;*/
/*         }*/
/*         LireIdentificateurAvecCarSpec( p_ptFicHeader , &carLu , */
/*                             DicoTransfert[ ind ].chACG ,*/
/*                             _MaxLetDicoEnum - 1 , "?" , _True );*/
/*         ind++;*/
/*     }*/
/*     if( ind == _DerCatLex )*/
/*     {*/
/*         while( carLu != '#' )*/
/*         {*/
/*              carLu = LireCar( p_ptFicHeader );*/
/*         }*/
/*         return _False;*/
/*     }*/
/*     *(DicoTransfert[ ind ].chIni) = '\0';*/
/*     *(DicoTransfert[ ind ].chACG) = '\0';*/
/*     if( !ReconnaitreCaractere( p_ptFicHeader , (int)'#' , &carLu , _True ) )*/
/*     {*/
/*         return _False;*/
/*     }*/

    /*---------------------------------------------------------------
                   lecture des mots courants
      ---------------------------------------------------------------*/
    fgetpos( p_ptFicHeader , p_posFic );
    if( !LireDuc( p_ptFicHeader , &carLu , p_posFic ) )
    {
        return _False;
    }

    /*---------------------------------------------------------------
             lecture des regles memoire
      ---------------------------------------------------------------*/
    fgetpos( p_ptFicHeader , p_posFic );
    if( OuvreAutomReglesMem )
    {
        long           nbReglesMots;
        long           nbReglesTermes;

        if( !CompilationRegles( NULL , p_ptFicHeader , &carLu ,
                                _False , _False , _True , &nbReglesFausses ,
                                &nbReglesMots , &nbReglesTermes ) )
        {
             return _False;
        }
    }
    else
    {
        while( carLu != '#' && carLu != EOF )
        {
            carLu = LireCar( p_ptFicHeader );
        }
    }

    if( !ReconnaitreCaractere( p_ptFicHeader , (int)'#' , &carLu , _True ) )
    {
         return _False;
    }

    /*---------------------------------------------------------------
             lecture des metaregles
      ---------------------------------------------------------------*/
    fgetpos( p_ptFicHeader , p_posFic );
    for( ind = 0 ; ind < _NbMaxTableau ; ind++ )
    {
        *( TabListeMetaReglesStd + ind ) = NULL;
    }
    if( OuvreAutomMetaregl )
    {
        if( !ReconnaitreChaine ( p_ptFicHeader , "Default" , &carLu )
            || !ReconnaitreChaine ( p_ptFicHeader , "metarules" , &carLu )
            || !ReconnaitreCaractere( p_ptFicHeader , (int)':' , 
				      &carLu , _True ) )
        {
              return _False;
        }
        while( carLu == '[' )
        {
             ReconnaitreCaractere( p_ptFicHeader , (int)'[' , &carLu , _True );
             ind = (int)LireEntier( p_ptFicHeader , &carLu );
             if( !ReconnaitreCaractere( p_ptFicHeader , (int)']' , 
					&carLu , _True ) )
             {
                 return _False;
             }
             if( ind < _NbMaxTableau && !(*( TabListeMetaReglesStd + ind )) )
             {
                 if( !LireUnAtome( &carLu , p_ptFicHeader , 
                                   TabListeMetaReglesStd + ind ,
                                   _False , _Trait ) )
                 {
                         return _False;
                 }
             }
        }
        if( !CompilationReglesMeta( p_ptFicHeader , &carLu ) )
        {
             return _False;
        }
    }
    else
    {
        while( carLu != '#' && carLu != EOF )
        {
            carLu = LireCar( p_ptFicHeader );
        }
    }

    if( !ReconnaitreCaractere( p_ptFicHeader , (int)'#' , &carLu , _True ) )
    {
         return _False;
    }

    _Free( catLexSuffixees );
    return _True;
}

/*****************************************************************************
                 RAZ DES DONNEES DE L'EN-TETE DE LA GRAMMAIRE
*****************************************************************************/
void     RAZDonneesEnTete( void )
{
    int                         ind;

    IniTraitsCatLex();
    RefCatMotsInc = 0L;
    RefDeterminantNumeral = 0L;
    RefPonctuation = 0L;

    DemarqueEtRendDag( &CheminCategorie );
    DemarqueEtRendDag( &CheminEtiquette );
    DemarqueEtRendDag( &CheminRacinesAnnexes );
    DemarqueEtRendDag( &CheminLexicalisation );
    DemarqueEtRendDag( &CheminLemme );
    DemarqueEtRendDag( &CheminFlexion );
    DemarqueEtRendDag( &CheminForme );
    DemarqueEtRendDag( &CheminNombre );
    DemarqueEtRendDag( &CheminReference );
    DemarqueEtRendDag( &CheminEtiquetteMeta );
    DemarqueEtRendDag( &CheminRacine );
    DemarqueEtRendDag( &CheminLeadingExprReg );
    DemarqueEtRendDag( &CheminTrailingExprReg );

    CheminCategorie = NULL;
    CheminEtiquette = NULL;
    CheminRacinesAnnexes = NULL;
    CheminLexicalisation = NULL;
    CheminLemme = NULL;
    CheminFlexion = NULL;
    CheminForme = NULL;
    CheminNombre = NULL;
    CheminReference = NULL;
    CheminEtiquetteMeta = NULL;
    CheminRacine = NULL;
    CheminLeadingExprReg = NULL;
    CheminTrailingExprReg = NULL;

    CarSpeciauxMots[ 0 ] = '\0';
    Ponctuation[ 0 ] = '\0';
    PonctuationFinale[ 0 ] = '\0';
    for( ind = 0 ; ind < _NbMaxSuffixages ; ind++ )
    {
          DemarqueEtRendDag( TabDagsSuffixage + ind );
          *(TabDagsSuffixage + ind) = NULL;
    }
    for( ind = 0 ; ind < _TailDucs ; ind++ )
    {
          DemarqueEtRendDag( &(Ducs[ ind ].Lemme.ValTerme.ArcsFils) );
          Ducs[ ind ].Lemme.ValTerme.ArcsFils = NULL;
    }
    RAZReglesMemoire();
    for( ind = 0 ; ind < _NbMaxTableau ; ind++ )
    {
          DemarqueEtRendDag( TabListeMetaReglesStd + ind );
          *(TabListeMetaReglesStd + ind) = NULL;
    }
    RAZReglesMeta();
}

/*****************************************************************************
        PASSAGE D'UNE REPRESENTATION MNEMONIQUE A UNE VALEUR
            POUR LES TRAITS OU LES CATEGORIES LEXICALES
*****************************************************************************/
void     IniTraitsCatLex( void )
{
    int           i;

    for( i = 0 ; i < NbMaxTraitsCatLex ; i ++)
    {
        *(DicoTraitsCatLex[i].ChaineDico) = '\0';
        DicoTraitsCatLex[i].TypeDico = (int)_Nil;
        DicoTraitsCatLex[i].NbFlexions = 0;
        DicoTraitsCatLex[i].NbDerivations = 0;
        DicoTraitsCatLex[i].NbNoTypeFlexion = 0;
        DicoTraitsCatLex[i].NbNoTypeDerivation = 0;
        DicoTraitsCatLex[i].RangPremSuffixeFlexion = (int)_Nil;
        DicoTraitsCatLex[i].RangPremSuffixeDerivation = (int)_Nil;
        DicoTraitsCatLex[i].RangPremDagFlexion = (int)_Nil;
        DicoTraitsCatLex[i].RangPremDagDerivation = (int)_Nil;
        DicoTraitsCatLex[i].Indice = i;
        DicoTraitsCatLex[i].Suivant = NULL;
    }
    /* on met une premiere categorie bidon, pour ne pas avoir de
       categorie nulle */
    *(DicoTraitsCatLex[0].ChaineDico) = ' ' + 1;
    *(DicoTraitsCatLex[0].ChaineDico + 1) = '\0';
}

char     ChercheOuInsereTraitsCatLex( ChIdent p_ident , int p_type , 
                                      int *p_catLex )
{
    struct DicoTraitsCatLex        *ptIni;
    struct DicoTraitsCatLex        *ptPrec;
    int                             indIns;

    if( !(*p_ident) )
    {
        *p_catLex = (int)_Nil;
         return _False;
    }

    ptIni = DicoTraitsCatLex + Hash( p_ident , NbMaxTraitsCatLex );
    while( ptIni && *(ptIni->ChaineDico) )
    {
        if( strncmp( p_ident , ptIni->ChaineDico , _MaxLetIdent  ) == 0
            && (char)p_type == ptIni->TypeDico )
        {
            *p_catLex = ptIni->Indice;
             return _True;
        }
        ptPrec = ptIni;
        ptIni = ptIni->Suivant;
    }
    if( !ptIni )
    {
        for( indIns = 0 ; indIns < NbMaxTraitsCatLex ; indIns++ )
        {
             if( !(*(DicoTraitsCatLex[indIns].ChaineDico)) )
             {
                 break;
             }
        }
        if( indIns == NbMaxTraitsCatLex )
        {
             TxtChaineErreur( 232 , ChAux );
             sprintf( ChErreurCompil , "%s [%d]" , ChAux , NbMaxTraitsCatLex );
            *p_catLex = (int)_Nil;
             return _False;
        }
        ptIni = ptPrec;
        ptIni->Suivant = DicoTraitsCatLex + indIns;
        ptIni = DicoTraitsCatLex + indIns;
    }
    strncpy( ptIni->ChaineDico , p_ident , _MaxLetIdent );
    ptIni->TypeDico = (char)p_type;
    (ptIni->ChaineDico)[ _MaxLetIdent - 1 ] = '\0';
    *p_catLex = ptIni->Indice;
    return _True;
}

char     IdentATraitsCatLex( ChIdent p_ident , int p_type , int *p_catLex )
{
    struct DicoTraitsCatLex        *ptIni;

    if( !(*p_ident) )
    {
        *p_catLex = (int)_Nil;
         return _False;
    }

    ptIni = DicoTraitsCatLex + Hash( p_ident , NbMaxTraitsCatLex );
    p_ident[ _MaxLetIdent - 1 ] = '\0';
    
    while( ptIni )
    {
        if( strncmp( p_ident , ptIni->ChaineDico , _MaxLetIdent  ) == 0
            && (char)p_type == ptIni->TypeDico )
        {
            *p_catLex = ptIni->Indice;
             return _True;
        }
        ptIni = ptIni->Suivant;
    }

    TxtChaineErreur( 167 , ChAux );
    sprintf( ChErreurCompil , "%s [%s]" , ChAux , p_ident );
    *p_catLex = (int)_Nil;
    return _False;
}

char     TraitsCatLexAIdent( ChIdent p_ident , int *p_type , int p_catLex )
{
    if( p_catLex > NbMaxTraitsCatLex )
    {
         return _False;
    }
 
    if( !( *(DicoTraitsCatLex[ p_catLex ].ChaineDico) ) )
    {
         TxtChaineErreur( 168 , ChAux );
         sprintf( ChErreurCompil , "%s [%d]" , ChAux , p_catLex );
         (*p_ident) = '\0';
         return _False;
    }
    else
    {
         strncpy( p_ident , DicoTraitsCatLex[ p_catLex ].ChaineDico , _MaxLetIdent );
         *p_type = (int)DicoTraitsCatLex[ p_catLex ].TypeDico;
         p_ident[ _MaxLetIdent - 1 ] = '\0';
         return _True;
    }
}

/*****************************************************************************
        ecriture de la liste des suffixes possibles d'une taille donnee
        dans l'ordre inverse de leur ecriture normale
*****************************************************************************/

void     EcrireListeSuffixes( int p_nbLetSuff , char *p_listeSuffEnvers )
{
    int                 indSuffDonnee;
    int                 indCh;          /* parcours des car  d'un suffixe  */
    int                 indListeSuff;

    char                suffEnvers[ _TailSuffixage ];
    char               *ptChaineSuf;
    char                trouve;

    indListeSuff = 0;
    for( indSuffDonnee = 0 ;
         indSuffDonnee < NbMaxSuffixes ;
         indSuffDonnee++ )
    {
        ptChaineSuf = Suffixes[ indSuffDonnee ].Suff;

        if( *ptChaineSuf == '#' )
            break;
        if( *ptChaineSuf == '?' )
            ptChaineSuf = ptChaineSuf + 2;

        if( (int)strlen( ptChaineSuf ) == p_nbLetSuff && *ptChaineSuf != '*')
        {
            /* ecriture du suffixe a l'envers                            */
            for( indCh = 0 ; indCh < p_nbLetSuff ; indCh++ )
            {
                suffEnvers[ indCh ] = ptChaineSuf[ p_nbLetSuff - 1 - indCh ];
            }
            suffEnvers[ p_nbLetSuff ] = '\0';

            /* parcours des suffixes a l'envers    precedents                */
            trouve = _False;
            for( indCh = 0 ; indCh < indListeSuff ; indCh += p_nbLetSuff + 1 )
            {
                if( EgaliteChaine( suffEnvers ,
                                   &p_listeSuffEnvers[ indCh ] ) == _True )
                {
                    trouve = _True;
                    break;
                }
            }
            if( trouve == _False )
            {
                for( indCh = 0 ; indCh <= p_nbLetSuff ; indCh++ )
                {
                    p_listeSuffEnvers[ indListeSuff++ ]
                             = suffEnvers[ indCh ];
                    if( indListeSuff >= _TailListeSuffixes )
                    {
                        TxtErreur( 112 , "EcrireListeSuffixes(alentete)" );
                    }
                }
            }
        }
    }
    if( indSuffDonnee == NbMaxSuffixes )
    {
        TxtErreur( 111 , "EcrireListeSuffixes(alentete)" );
    }
    p_listeSuffEnvers[ indListeSuff ] = '#';
}

/*****************************************************************************
        ajout d'un nouveau suffixe au graphe des suffixes
*****************************************************************************/

void     AjouterSuffixe( char *p_tSuff , int p_longueur)
{
   struct UnNoeudSuff              *ptDerNoeudCommun;
   struct UnNoeudSuff              *ptNouveauFils;
   int                             indDerCarCommun;
   int                             tailDecalage;
   int                             trouve;
   int                             indCou;

    ptDerNoeudCommun = PtDernierNoeudCommun( p_tSuff , &indDerCarCommun , 
                                             &trouve , '\0' );
    if( trouve == _True )
    {
        sprintf( ChAux , "AjouterSuffixe(alentete) [%s(%d)]",
                         p_tSuff , indDerCarCommun );
        TxtErreur( 113 , ChAux );
        return;
    }

    /* un suffixe plus long existe deja */
    if( indDerCarCommun == p_longueur - 1 )
    {
        TxtErreur( 114 , "AjouterSuffixe(alentete)" );
        ptDerNoeudCommun->Terminal = _True;
        return;
    }

    /* pointeur vers le premier nouveau fils                                */
    if( ptDerNoeudCommun->NbFils > 0)
    {
        ptNouveauFils = ptDerNoeudCommun->PtFils + ptDerNoeudCommun->NbFils;
    }
    else
    {
        ptNouveauFils = NoeudsSuff + NbNoeudsSuff;
        ptDerNoeudCommun->PtFils = ptNouveauFils;
    }
    (ptDerNoeudCommun->NbFils)++;

    /* ajout de ce suffixe a partir de PtDerNoeudSuff                        */
    NbNoeudsSuff += ( tailDecalage = p_longueur - 1 - indDerCarCommun );
    if( NbNoeudsSuff >= _NbMaxNoeudsSuffixe )
    {
        sprintf( ChAux , "AjouterSuffixe(alentete) [%s]" , p_tSuff );
        TxtErreur( 115 , ChAux );
        return;
    }

    /* reactualisation de poineturs vers les fils de tous ceux qui
       pointent vers des noeuds qui sont decales a cause de
       l'insertion du nouveau suffixe( sauf le noeud dont les fils
       sont augmentes */
    for( indCou = 0 ; indCou < NbNoeudsSuff ; indCou++ )
    {
        if( NoeudsSuff[ indCou].PtFils >= ptNouveauFils
            && &NoeudsSuff[ indCou] != ptDerNoeudCommun)
        {
            NoeudsSuff[ indCou ].PtFils = NoeudsSuff[ indCou ].PtFils + tailDecalage;
        }
    }

    /* decal. des noueuds de graphe suivant les fils du dern. noeud
       commun */
    for(    indCou = NbNoeudsSuff ;
            NoeudsSuff + indCou >= ptNouveauFils + tailDecalage ;
            indCou-- )
    {
        NoeudsSuff[ indCou ] = NoeudsSuff[ indCou - tailDecalage ];
    }

    /* recopie de la fin du suffixe sur la palce laissee libre a
        partir de ptNouveauFils */
    for( indCou = 0 ; indCou < tailDecalage ; indCou++ )
    {
        ptNouveauFils[ indCou ].Car = p_tSuff[ indDerCarCommun + indCou + 1 ];
        if( indCou < tailDecalage - 1 )
        {
            ptNouveauFils[ indCou ].NbFils = 1;
            ptNouveauFils[ indCou ].Terminal = _False;
            ptNouveauFils[ indCou ].PtFils = ptNouveauFils + ( indCou + 1 );
        }
        else
        {
            ptNouveauFils[ indCou ].NbFils = 0;
            ptNouveauFils[ indCou ].Terminal = _True;
            ptNouveauFils[ indCou ].PtFils = NULL;
        }
    }
}

/*****************************************************************************
        Recherche du derier caractere commun entre le graphe des suffixes
        et une chaine donnee
*****************************************************************************

    renvoit le dernier noeud commun dans le graphe NoeudsSuff
char        *p_rang;           rang du der car commun
char        *p_trouve;       booleen indiquant si le suffixe est trouve        */

struct UnNoeudSuff *PtDernierNoeudCommun( char *p_tSuff , 
										  int *p_rang , int *p_trouve , 
										  int p_derCarRacine )
{
    int                           nbFils;
    int                           nbFilsCou;
    int                           longueurSuff;
    struct UnNoeudSuff            *ptFils;
    struct UnNoeudSuff            *ptPere;

    longueurSuff = strlen( p_tSuff );
    ptPere = NoeudsSuff;
    ptFils = ptPere->PtFils;
    nbFils = ptPere->NbFils;

    for ( *p_rang = 0 ; *p_rang < longueurSuff ; (*p_rang)++ )
    {
        *p_trouve = _False;
        for( nbFilsCou = 0 ; nbFilsCou < nbFils ; nbFilsCou++ )
        {
            if(  ptFils[ nbFilsCou ].Car == p_tSuff[ *p_rang ]
                 || ( p_derCarRacine
                      && ptFils[ nbFilsCou ].Car == '!'
                      && p_tSuff[ *p_rang ] == p_derCarRacine ) )
             {
                *p_trouve = _True;
                break;
             }
             else
             {
                *p_trouve = _False;
             }
        }

        if( !(*p_trouve) )
        {
            (*p_rang)--;
            break;
        }

        if( *p_rang == longueurSuff - 1     /* p_trouve _True    et fin du suffixe */
            && ptFils[ nbFilsCou ].Terminal == _False )
        {
            *p_trouve = _False;
        }

        ptPere = ptFils + nbFilsCou;
        ptFils = ptPere->PtFils;
        nbFils = ptPere->NbFils;
    }
    return( ptPere );
}

/*****************************************************************************
        passage du fichier des suffixes a la memoire
*****************************************************************************/

void     ConstruitALGSuffixes( void )
{
    int                             nbLetSuff;
    int                             indListeSuff;
    int                             indSuff;

    char                            listeSuffixeEnvers[ _TailListeSuffixes ];
    char                            suffCou[ _TailSuffixage ];
    struct UnNoeudSuff              *ptSuff;

    /* initialisation de NoeudsSuff                                        */
    for( NbNoeudsSuff = 0 ; NbNoeudsSuff < _NbMaxNoeudsSuffixe ; NbNoeudsSuff++ )
    {
        ptSuff = NoeudsSuff + NbNoeudsSuff;
        ptSuff->Car = 0;
        ptSuff->NbFils = 0;
        ptSuff->Terminal = _False;
        ptSuff->PtFils = NULL;
    }
    NbNoeudsSuff = 1;

    /* stocke les suffixes par ordre de taille croissante, a l'envers, en
       s'assurant de ne pas les repeter                                    */
    for( nbLetSuff = 1 ; nbLetSuff <= _TailSuffixage ; nbLetSuff++ )
    {
        EcrireListeSuffixes( nbLetSuff , listeSuffixeEnvers );

        indListeSuff = 0;
        while( listeSuffixeEnvers[ indListeSuff ] != '#')
        {
            indSuff = 0;
            while( listeSuffixeEnvers[ indListeSuff ] != '\0' )
            {
                suffCou[ indSuff++ ] = listeSuffixeEnvers[ indListeSuff++ ];
            }
            suffCou[ indSuff ] = '\0';
            indListeSuff++;

            if( indSuff != nbLetSuff )
            {
                sprintf( ChAux , "ConstruitALGSuffixes(alentete) [%d,%d]" ,
                                 nbLetSuff , indSuff );
                TxtErreur( 116 , ChAux );
            }

            AjouterSuffixe( suffCou , nbLetSuff );
        }
    }
}

/*****************************************************************************
        passage des contraintes de flexions sur fichier a la memoire
*****************************************************************************/
char     LireContraintesSuffixages( FILE *p_ptFic , char *p_carLu, 
                                  int p_nbSuffixages , ChIdent p_identCatLex ,
                                  int *p_indSuffixagesTot )
{
    int                indSuffixages;
    TabDagMem          operLogFils;
    ChIdent            ident;

    for( indSuffixages = 0 ; indSuffixages < p_nbSuffixages ; indSuffixages++ )
    {
         LireIdentificateur( p_ptFic , p_carLu , ident , _MaxLetIdent ,
                             _True );
         if( strcmp( ident , p_identCatLex ) != 0 )
         {
             TxtChaineErreur( 161 , ChAux );
             sprintf( ChErreurCompil , "%s [%s] [%s]" , 
                      ChAux , ident , p_identCatLex );
             return _False;
         }
         if( !ReconnaitreCaractere( p_ptFic , (int)'[' , p_carLu , _True ) )
         {
              return _False;
         }
         if( indSuffixages != (int)LireEntier( p_ptFic , p_carLu ) - 1 )
         {
             TxtChaineErreur( 160 , ChErreurCompil );
             return _False;
         }
         if( !ReconnaitreCaractere( p_ptFic , (int)']' , p_carLu , _True ) )
         {
              return _False;
         }

         /* lecture des contraintes de la flexion */
         operLogFils = NULL;
         *(TabDagsSuffixage + *p_indSuffixagesTot) = NULL;
         if( !LireContraintesTerme( p_carLu , p_ptFic ,
                                    &operLogFils , 
                                    TabDagsSuffixage + *p_indSuffixagesTot ) )
         {
             return _False;
         }

         if( !UnifierContraintes( &operLogFils , 
                                  (TabDagsSuffixage + *p_indSuffixagesTot) , 
                                  NULL , NULL ) )
         {
             TxtErreur( 63 , "LireContraintesSuffixages(alentete)" );
             EcrireDagBasique( TabDagsSuffixage + *p_indSuffixagesTot );
             return _False;
         }
         /* lecture des contraintes de la flexion */

         if( !ReconnaitreCaractere( p_ptFic , (int)'.' , p_carLu , _True ) )
         {
               return _False;
         }

         (*p_indSuffixagesTot)++;
    }
    return _True;
}

/*****************************************************************************
        passage du fichier des suffixes a la memoire
*****************************************************************************/
char     LireSuffixesSuffixages( FILE *p_ptFic , unsigned long *p_posFic , 
                               char *p_carLu, int p_nbSuffixages , 
                               ChIdent p_identCatLex , int *p_nbNoSuffixages , 
                               int *p_nbSuffixesTot )
{
    int             indFlex;
    ChIdent         identCatLex;

    *p_nbNoSuffixages = 0;
    while( _True )
    {
        fgetpos( p_ptFic , p_posFic );

        if( *p_carLu == EOF || *p_carLu == '#')
        {
            if( !ReconnaitreCaractere( p_ptFic , (int)'#' , p_carLu , _True ) )
            {
                 return _False;
            }
            *(Suffixes[ *p_nbSuffixesTot ].Suff) = '#';
            break;
        }
        SautBlanc( p_ptFic , p_carLu );

        LireIdentificateur( p_ptFic , p_carLu , identCatLex , _MaxLetIdent ,
                            _True );
        if( strcmp( p_identCatLex , identCatLex ) != 0 )
        {
            TxtChaineErreur( 159 , ChAux );
            sprintf( ChErreurCompil , "%s [%s] [%s]" , 
                     ChAux , p_identCatLex , identCatLex );
            return _False;
        }
        if( !ReconnaitreCaractere( p_ptFic , (int)'[' , p_carLu , _True ) )
        {
             return _False;
        }
        if( *p_nbNoSuffixages != (int)LireEntier( p_ptFic , p_carLu ) - 1 )
        {
            TxtChaineErreur( 190 , ChErreurCompil );
            return _False;
        }
        if( !ReconnaitreCaractere( p_ptFic , (int)']' , p_carLu , _True ) )
        {
             return _False;
        }

        for( indFlex = 0 ; indFlex < p_nbSuffixages ; indFlex++ )
        {
            LireSuffixe( p_ptFic , p_carLu ,
                         Suffixes[ *p_nbSuffixesTot ].Suff ,
                         _TailSuffixage );
            (*p_nbSuffixesTot)++;
            if( (*p_nbSuffixesTot) >= NbMaxSuffixes )
            {
                TxtChaineErreur( 166 , ChErreurCompil );
                return _False;
            }
        }

        (*p_nbNoSuffixages)++;
    }
    return _True;
}

/*****************************************************************************
           insertion d'un identificateur : indice auquel se fait l'insertion
           + indice de chainage eventuel avec le precedent
*****************************************************************************/

struct EntreeDucs *IndicesDucs( ChIdent p_identificateur , 
							   char *p_trouvee , struct EntreeDucs* *ptPrec , 
							   struct EntreeDucs* *p_ptTrouve )
{
        struct EntreeDucs              *ptIni,
                                       *ptEntree;
        short                          NbSuiv = -1;

        ptEntree = &Ducs[ Hash( p_identificateur , _TailDucs ) ];
        *p_trouvee = _False;
        *p_ptTrouve = NULL;

        /* article non vide                                             */
        if( (ptEntree->Lemme).ValTerme.Ident[ 0 ] != '\0' )
        {
            /* recherche de l'indice d'insertion et insertion       */

            /* recherche du dernier de la chaine & stockage des indices
               des Suivants                                         */
            do
            {
                if( EgaliteChaine( p_identificateur , 
                                   (ptEntree->Lemme).ValTerme.Ident ) == _True )
                {
                    *p_trouvee = _True;
                    *p_ptTrouve = ptEntree;
                }
                *ptPrec = ptEntree;
            }
            while( (ptEntree = ptEntree->Suivant) != NULL
                   && NbSuiv < _TailDucs);

            if( NbSuiv == _TailDucs )
            {
                TxtErreur( 47 , "IndiceDucs(alentete)" );
                 exit(0);
            }

            /* recherche de la 1ere case libre apres la fin de la chaine
               en parcourant sequentiellement la suite des cases    */

            /* on parcourt cycliquement le tableau                                  */
            if ( *ptPrec != &Ducs[_TailDucs - 1] )
                ptIni = *ptPrec + 1;
            else
                ptIni = Ducs;

            ptEntree = ptIni;

            while( (ptEntree->Lemme).ValTerme.Ident[ 0 ] != '\0' )
            {
                /* on parcourt cycliquement le tableau                                  */
                if ( ptEntree != &Ducs[_TailDucs - 1] )
                    ptEntree = ptEntree + 1;
                else
                    ptEntree = Ducs;

                if( ptEntree == ptIni )
                {
                    TxtErreur( 48 , "IndicesDucs(alentete)" );
                    exit(0);
                }
            }

            return( ptEntree );

        }
        /* article non vide                                             */
        else
        /* article vide                                                 */
        {
                /* *p_trouvee = _False; */
                *ptPrec = NULL;
                return( ptEntree );
        }
}

/*****************************************************************************
           recherche d'un identificateur dont on connait la reference
*****************************************************************************/

int      IdentificateurDucs( ChIdent p_identificateur , long p_reference)
{
    int                         i;

    *p_identificateur = '\0';
    for( i = 0 ; i < _TailDucs ; i++ )
    {
       if( (Ducs[ i ].Lemme).ValTerme.Ref == p_reference )
       {
            strcpy( p_identificateur , (Ducs[ i ].Lemme).ValTerme.Ident );
            return _True;
       }
    }
    /* sprintf( ChAux , "IdentificateurDucs(alentete) [%ld]", p_reference );
    TxtErreur( 49 , ChAux ); */
    return _False;
}

/*****************************************************************************
           insertion d'un identificateur : indice auquel se fait l'insertion +
           insertion de la chaine et chainage avec le precedent si besoin
*****************************************************************************/

struct EntreeDucs *InsertionDucs( ChIdent p_identificateur , char *p_trouvee )
{
        struct EntreeDucs      *ptPrec,
                               *ptTrouve,
                               *ptEntree;

        ptEntree = IndicesDucs(p_identificateur , p_trouvee , &ptPrec , &ptTrouve);

        if(ptPrec)
        /* article non vide                                             */
        {
                /* l'identificateur est rajoute en queue                                */
                ptPrec->Suivant = ptEntree;
                strcpy( (ptEntree->Lemme).ValTerme.Ident , p_identificateur );
                return( ptEntree );

        }
        /* article non vide                                             */
        else
        /* article vide                                                 */
        {
                strcpy( (ptEntree->Lemme).ValTerme.Ident , p_identificateur );
                return( ptEntree );
        }
}

/**************************************************************************
                      acces aux entrees par hashcode
       recherche d'un identificateur : pointeurs auxquels il se trouve
**************************************************************************/

void     ListePointeursDucsParChaine( ChIdent  p_identificateur ,
                     struct EntreeDucs *p_listePt[ _NbMaxHomographe ])
{
        int                             NbHomgraphe = 0;
        struct EntreeDucs               *ptEntree;

        ptEntree = &Ducs[ Hash( p_identificateur , _TailDucs ) ];
        p_listePt[0] = NULL;
        if( ptEntree == NULL )
        {
              return;
        }

        while( EgaliteChaine( (ptEntree->Lemme).ValTerme.Ident , p_identificateur ) == _False )
        {
                if( (ptEntree = ptEntree->Suivant) == NULL )
                {
                        return;
                }
        }

        while( EgaliteChaine( (ptEntree->Lemme).ValTerme.Ident , 
                              p_identificateur ) == _True )
        {

                p_listePt[ NbHomgraphe++] = ptEntree;
                p_listePt[ NbHomgraphe] = NULL;

                if( (ptEntree = ptEntree->Suivant) == NULL )
                        return;
        }
}

/**************************************************************************
                      acces aux entrees par hashcode
       recherche d'une reference : pointeurs auxquels elle se trouve
**************************************************************************/
void     ListePointeursDucsParReference( long p_refDonnee ,
                     struct EntreeDucs *p_listePt[ _NbMaxHomographe ])
{
    int                         NbHomgraphe = 0;
    int                         i;

    p_listePt[0] = NULL;
    if( p_refDonnee == -1L )
         return;

    /* PARCOURS TOTAL DU TABLEAU                           */
    for( i = 0 ; i < _TailDucs ; i++ )
    {
       if( (Ducs[ i ].Lemme).ValTerme.Ref == p_refDonnee )
       {
            p_listePt[ NbHomgraphe++] = (Ducs + i);
            p_listePt[ NbHomgraphe] = NULL;
       }
    }
}

/***************************************************************************
    lire le dico des mots simples courts ou ususels dans un fichier texte
***************************************************************************/
char     LireDuc( FILE *p_ptFic , char *p_carLu , unsigned long *p_posFic )
{
    int                  i;

    struct LemmeMem      lemmeMem;
    struct EntreeDucs   *ptEntree;
    char                 chaineTrouvee;
    TabDagMem           *hdlFilsReference;
    TabDagMem           *hdlFilsLemme;
    char                 resul;
    struct ValAtomeMem   atome;

    resul = _True;
    
    /* INITIALISATIONS DE TABLEAUX                           */
    for( i = 0 ; i < _TailDucs ; i++ )
    {
            Ducs[ i ].Suivant = NULL;
            MiseAZeroUnLemme( &(Ducs[ i ].Lemme) );
    }

    while( _True )
    {
        fgetpos( p_ptFic , p_posFic );

        if( *p_carLu == EOF || *p_carLu == '#' )
        {
            if( !ReconnaitreCaractere( p_ptFic , (int)'#' , p_carLu , _True ) )
            {
                 return _False;
            }
            return _True;
        }

        MiseAZeroUnLemme( &lemmeMem );
        if( !CompileLemme( &lemmeMem , p_carLu , p_ptFic ) )
        {
              return _False;
        }
        else
        {
            (lemmeMem.ValTerme.Ident)[ _MaxLetIdent - 1 ] = '\0';
            hdlFilsReference = ExisteCheminFils( CheminReference , 
                                         &(lemmeMem.ValTerme.ArcsFils) );
            hdlFilsLemme = ExisteCheminFils( CheminLemme , 
                                         &(lemmeMem.ValTerme.ArcsFils) );
            
            /* ajout de la chaine du lemme pour gerer la negation par
               unification */
            atome.AtomeConjDisj = _False;
            strcpy( atome.ValAtome.Chaine , lemmeMem.ValTerme.Ident );
            if( !hdlFilsLemme )
            {
               if( !AjouteCheminEtAtomeFils( &(lemmeMem.ValTerme.ArcsFils) ,
					     CheminLemme , _AtomeChaine , 
					     atome , _True ) )
               {
                   return _False;
               }
            }

            if( hdlFilsReference && (*hdlFilsReference) 
                &&  (*hdlFilsReference)->TypeDag == _AtomeEntier )
            {
                lemmeMem.ValTerme.Ref = 
                            (*hdlFilsReference)->ValDag.Atome.ValAtome.Entier;

                /* verification que les regles memoire ont bien une ref
                   superieure a la ref numerique et inferieure a celle 
                   d'une ponctuation                                     */
                if( lemmeMem.ValTerme.Ref != _Nil
                    && (lemmeMem.ValTerme.Ref >= _DebutRefPonctuation
                       || lemmeMem.ValTerme.Ref <= _RefNum) )
                {
                    TxtChaineErreur( 192 , ChAux );
                    sprintf( ChAux , "%s [%s]\n" , ChAux , 
                             lemmeMem.ValTerme.Ident );
                    strcat( ChErreurCompil , ChAux );
                    MessageErreurCompil( *p_posFic , p_carLu , p_ptFic );
                    resul = _False;
                }
            }
            else
            {
                lemmeMem.ValTerme.Ref = _Nil;
                return _False;
            }

            ptEntree = InsertionDucs( lemmeMem.ValTerme.Ident , 
                                      &chaineTrouvee );
            MiseAZeroUnLemme( &(ptEntree->Lemme) );
            ptEntree->Lemme = lemmeMem;
        }
    }
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            fin de alentete.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
