/*
 * FASt Term Recognizer
 *
 *  FASTR/alinsere.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for inserting new data in binary dictionaries
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*************************************************************************
      VARIABLES LOCALES utilisees dans les parcours d'arbre
*************************************************************************/
struct UnLien
{
     long         IndAbs;
     long         Reference;
};
struct UnLien         *LiensIdentPosFic;
int                   NbLiensRegles;
long                  NbLiensMorphologiques;
long                  NbLiensSemantiques;
unsigned long         TailleFicLiensMorphologiquesEtSemantiques;

/*************************************************************************
      LIEN MORPHOLOGIQUE DANS LE FICHIER DES LIENS
*************************************************************************/
struct InterFileLink
{
  long           PtMorphOrSemNeighbor;   /* pt fichier vers le lemme derive */
  long           PtNextLink;   /* le pteur de lien morpho suivant */
};

/*************************************************************************
  RATTACHEMENT SIMPLIFIE D'UNE REGLE AU 1ER TERME DE LA CONCATENATION
*************************************************************************/
struct ValTerme  *LexicalisationSansRisque( TabTermeMem p_ptTerrme )
{
    TabTermeMem        ptTerme;

    if( p_ptTerrme->ValTerme.TypeTerme != _Concatenation )
    {
        return NULL;
    }
    else
    {
        /* recherche du premier fils qui est un terme syntaxique */
        ptTerme = p_ptTerrme; /* la tete de la conc normalement */

        if( ptTerme )
             ptTerme = ptTerme->TermeFils;

        while( ptTerme )
        {
            if( ptTerme->ValTerme.TypeTerme == _UniteLexicale
                && ptTerme->ValTerme.Ref >= _PremRefLemme )
            {
                return &(ptTerme->ValTerme);
            }
            ptTerme = ptTerme->TermeFrere;
        }
        return NULL;
    }
}

/*************************************************************************
      RATTACHEMENT IMPERATIF AU PREMIER TERME RENCONTRE EN FAISANT 
                UN PARCOURS EN LARGEUR DABORD DE LA REGLE
*************************************************************************/
struct ValTerme  *LexicalisationImperative( TabTermeMem p_ptTerrme )
{
    struct ValTerme   *ptValTerme;

    if( p_ptTerrme->ValTerme.TypeTerme == _UniteLexicale
        && p_ptTerrme->ValTerme.Ref >= _PremRefLemme )
    {
        return &(p_ptTerrme->ValTerme);
    }

    if( p_ptTerrme->TermeFrere )
    {
        /* recherche du premier frere qui est un terme syntaxique */
        ptValTerme = LexicalisationImperative( p_ptTerrme->TermeFrere );
        if( ptValTerme )
        {
            return ptValTerme;
        }
    }

    if( p_ptTerrme->TermeFils )
    {
        /* recherche du premier frere qui est un terme syntaxique */
        ptValTerme = LexicalisationImperative( p_ptTerrme->TermeFils );
        if( ptValTerme )
        {
            return ptValTerme;
        }
    }

    return NULL;
}

/*************************************************************************
      VERIFIE S'IL Y A DES REFERENCES A CHANGER
**************************************************************************/
char     PasDeReferenceAChanger( TabTermeMem *p_ptTerme )
{
    if( !(*p_ptTerme) )
    {
        return _True;
    }

    if( (*p_ptTerme)->ValTerme.TypeTerme == _Regle
        && (*p_ptTerme)->ValTerme.Ref != _Nil )
    {
        return _False;
    }
    else
    {
        return _True;
    }
}

/*************************************************************************
      Remplacement des identificateurs de regles par l'indice de
         leur position dans le fichier des regles
      Procedure auxiliaire utilisee dans le parcours de l'arborescence
**************************************************************************/
char     ChangeReference( TabTermeMem *p_ptTerme )
{
    int                         i;
    ChIdent                     ident;

    if( !(*p_ptTerme) )
    {
        return _True;
    }

    if( (*p_ptTerme)->ValTerme.TypeTerme == _Regle
        && (*p_ptTerme)->ValTerme.Ref != _Nil )
    {
      for( i = 0 ; i < NbLiensRegles ; i++ )
      {
         if( LiensIdentPosFic[ i ].Reference == (*p_ptTerme)->ValTerme.Ref )
         {
            (*p_ptTerme)->ValTerme.PosFicRegle = LiensIdentPosFic[ i ].IndAbs;
            return _True;
         }
      }

      Identificateur( ident , (*p_ptTerme)->ValTerme.Ref , 
                      FicRefIdent , NULL );
      sprintf( ChAux , "ChangeReference(alinsere) [%ld,%s]" ,
                 (*p_ptTerme)->ValTerme.Ref , ident );
      TxtErreur( 137 , ChAux );
      return _True;
    }

    return _True;
}

/*************************************************************************
      Remplacement des identificateurs de regles par l'indice de
         leur position dans le fichier des regles
**************************************************************************/
void     EditionLiensTermesAppel( void )
{

   unsigned long                  indAbsCou;
   unsigned long                  indAbsIni;
   unsigned long                  longFic;

   TabRegleMem                     ptRegleCou;
   TabChar                         tCharRegle;

   long                            longTChar;
   long                            nbReglesTot;
   long                            nbReglesRelevees;
   
   int                             indRegleMem;
   
    /* 100 L pour des regles ayant le meme identificateur */
    /* et n'apparaissant qu'une fois dans le fichier d'identificateurs */
    nbReglesTot = NbIdentReglesDictionnaire() + 200L;

   /* LiensTermesFichier contiendra les liens entre les identificateurs
      des regles et leur position absolue dans le fichier des regles
      Cette procedure devra etre effectuee apres toute insertion dans le
      fichier des regles .                                             */
    LiensIdentPosFic = (struct UnLien *)
                    _Alloc( (unsigned long)sizeof( struct UnLien)
                            * (unsigned long)nbReglesTot );
    tCharRegle = AlloueBuffer( TailleBuffReglesCourt );
    if ( !LiensIdentPosFic || !tCharRegle )
    {
        TxtErreur( 81 ,"EditionLiensTermesAppel(alinsere)" );
        return;
    }

    _NouveauDialogueInfo( 128 , _True );
    TxtTitresDialogueInfo( 18 , 19 , 0 );

    /* Etape 1 on releve les identificateur des regles et la position
       initiale du terme de rattachement */
    NbLiensRegles = 0;
    nbReglesRelevees = 0;
    /* regles fichier */
    indAbsCou = 0L;
    fseek( FicBDicRegles , 0L , 2 );
    fgetpos( FicBDicRegles , &longFic );
    IndBuffRegleMin = 0L;
    IndBuffRegleMax = 0L;

    while( indAbsCou < longFic )
    {
        if( nbReglesRelevees++ % 100L == 0 )
        {
            sprintf( ChAux , "%ld" , nbReglesRelevees );
            TxtEcritItemInfo( 1 , ChAux );
        }

        if( _ClickDansAnnulerInfo() )
        {
           _Free( LiensIdentPosFic );
           _Free( tCharRegle );
           _FinDialogueInfo();
           return;
        }

        indAbsIni = indAbsCou;

        if( RestructurerUneRegle( &indAbsCou , &ptRegleCou ) == _Nil 
             || !ptRegleCou )
        {
             TxtErreur( 107 , "EditionLiensTermesAppel(alinsere)" );
              break;
        }

        LiensIdentPosFic[ NbLiensRegles ].IndAbs = (long)indAbsIni;
        if( !(ptRegleCou->TermeFils) )
        {
             TxtErreur( 107 , "EditionLiensTermesAppel(alinsere)" );
             return;
        }
        LiensIdentPosFic[ NbLiensRegles ].Reference
                          = ptRegleCou->TermeFils->ValTerme.Ref;
        NbLiensRegles++;
        if( NbLiensRegles >= nbReglesTot - 1L )
        {
             TxtErreur( 107 , "EditionLiensTermesAppel(alinsere)" );
             return;
        }

        DemarqueEtRendRegle( &ptRegleCou );
    }
    /* regles memoire */
    for( indRegleMem = 0 ; indRegleMem < NbReglesMem ; indRegleMem++ )
    {
         TxtChaineRsrc( 131 , 24 , ChAux );
         TxtEcritItemInfo( 1 , ChAux );

         ptRegleCou = ReglesMem + indRegleMem;
         LiensIdentPosFic[ NbLiensRegles ].IndAbs = (long)(-indRegleMem);
         if( !(ptRegleCou->TermeFils) )
         {
              TxtErreur( 107 , "EditionLiensTermesAppel(alinsere)" );
              return;
         }
         LiensIdentPosFic[ NbLiensRegles ].Reference
                                     = ptRegleCou->TermeFils->ValTerme.Ref;
         NbLiensRegles++;
         if( NbLiensRegles >= nbReglesTot - 1L )
         {
               TxtErreur( 107 , "EditionLiensTermesAppel(alinsere)" );
               return;
         }
    }

    TxtTitresDialogueInfo( 20 , 21 , 0 );

    /* Etape 2 on remplace les identificateurs par la reference de la regle
       initiale du terme de rattachement */
    nbReglesRelevees = 0L;
    /* regles fichier */
    indAbsCou = 0L;
    IndBuffRegleMin = 0L;
    IndBuffRegleMax = 0L;
    while( indAbsCou < longFic )
    {
       if( nbReglesRelevees++ % 100L == 0 )
       {
           sprintf( ChAux , "%ld" , nbReglesRelevees );
           TxtEcritItemInfo( 1 , ChAux );
       }

       if( _ClickDansAnnulerInfo() )
       {
          _Free( LiensIdentPosFic );
          _Free( tCharRegle );
          _FinDialogueInfo();
          return;
       }

       indAbsIni = indAbsCou;

       if(  RestructurerUneRegle( &indAbsCou , &ptRegleCou ) == _Nil 
            || !ptRegleCou )
       {
            TxtErreur( 107 , "EditionLiensTermesAppel(alinsere)" );
            break;
       }

       /* au Moins 1 Reference A Changer */
       if( !( ParcoursDFArbre( &(ptRegleCou->TermeFils) , 
                               PasDeReferenceAChanger ,
                               _PreOrdre ) ) )
       {
             ParcoursDFArbre( &(ptRegleCou->TermeFils) , ChangeReference ,
                              _PreOrdre );

             /* insertion de la regle */
             longTChar = DestructurerUneRegle( ptRegleCou , tCharRegle );
             if( (unsigned long)longTChar != indAbsCou - indAbsIni )
             {
                  TxtErreur( 82 , "EditionLiensTermesAppel(alinsere)" );
             }

             LireEcrire( indAbsIni , longTChar , tCharRegle , _Ecrire , 
                         FicBDicRegles );
       }

       DemarqueEtRendRegle( &ptRegleCou );
    }

    /* regles memoire */
    for( indRegleMem = 0 ; indRegleMem < NbReglesMem ; indRegleMem++ )
    {
         TxtChaineRsrc( 131 , 24 , ChAux );
         TxtEcritItemInfo( 1 , ChAux );

         ptRegleCou = ReglesMem + indRegleMem;
         ParcoursDFArbre( &(ptRegleCou->TermeFils) , ChangeReference ,
                          _PreOrdre );
    }

    _FinDialogueInfo();

    _Free( LiensIdentPosFic );
    _Free( tCharRegle );
}

/*************************************************************************
      Liens entre lemmes dans structures de traits et lemmes en fichier
**************************************************************************/
void     EditionLiensMorphologiquesEtSemantiques( void )
{
    char            nomFicDicLiens[ 256 ];
    unsigned long   tailleDico;
    unsigned long   nbLus;

    if( Trace ) {
      fprintf( FicLog , "Resetting morphological and semantic links\n");
      PhotoMemoire( FicLog );
    }

    _NouveauDialogueInfo( 128 , _True );
    TxtTitresDialogueInfo( 18 , 19 , 0 );
    /* RAZ des liens morphologiques et semantiques dans le fichier des
       lemmes */
    MapLitEtModifieLemmes( ResetsMorphAndSemLinks );
    /* effacement du fichier des liens morphologiques qui va etre
       reconstruit */
    strcpy( nomFicDicLiens , NomFicDico );
    strcat( nomFicDicLiens , ".LIN" );
    if( FicBDicLiens )
    {
        fclose( FicBDicLiens );
    }
    FicBDicLiens = fopen( nomFicDicLiens , _FopenWPB );
    if( !FicBDicLiens )
    {
        sprintf( ChAux , 
		"EditionLiensMorphologiquesEtSemantiques(alinsere) [%s]" , 
		NomFicDico );
        TxtErreur( 147 , ChAux );
        _FinDialogueInfo();
        return;
    }

    if( Trace ) {
      fprintf( FicLog , "Creating morphological and semantic links\n");
      PhotoMemoire( FicLog );
    }

    /* creation des liens morphologiques, ils ne sont pas necessairement
       sequentiels dans le fichier de liens */
    TxtTitresDialogueInfo( 20 , 21 , 0 );
    TailleFicLiensMorphologiquesEtSemantiques = 0L;
    MapLitEtModifieLemmes( GarnitLiensMorphologiques ); 
    MapLitEtModifieLemmes( GarnitLiensSemantiques ); 

    /* creation d'un fichier auxiliaire de liens en vue de la
       sequentialisation des liens morphologiques */
    TxtTitresDialogueInfo( 57 , 58 , 0 );
    FicBDicLiensAux = fopen( NomFicAux , _FopenWPB );
    if( !FicBDicLiensAux )
    {
        sprintf( ChAux , 
		"EditionLiensMorphologiquesEtSemantiques(alinsere) [%s]" , 
		NomFicAux );
        TxtErreur( 147 , ChAux );
        _FinDialogueInfo();
        return;
    }

    if( Trace ) {
      fprintf( FicLog , "Sequentializing morphological and semantic links\n");
      PhotoMemoire( FicLog );
    }

    /* sequentialisation des liens morphologiques dans le fichier de
       liens (leur taille est diminuee par 2 puisqu'ils sont ensuite
       necessairmeent contigus */
    TailleFicLiensMorphologiquesEtSemantiques = 0L;
    NbLiensMorphologiques = 0L;
    NbLiensSemantiques = 0L;
    MapLitEtModifieLemmes( SequentializeMorphAndSemanticLinks ); 

    fclose( FicBDicLiensAux );
    fclose( FicBDicLiens );

    if( remove( nomFicDicLiens ) 
        || rename(  NomFicAux , nomFicDicLiens ) )
    {
         TxtErreur( 53 , "EditionLiensMorphologiquesEtSemantiques(alinsere)" );
         _FinDialogueInfo();
         return;
    }

    FicBDicLiens = fopen( nomFicDicLiens , _FopenRPB );
    if( !FicBDicLiens )
    {
        sprintf( ChAux , 
		 "EditionLiensMorphologiquesEtSemantiques(alinsere) [%s]" , 
                 nomFicDicLiens );
        TxtErreur( 147 , ChAux );
        _FinDialogueInfo();
        return;
    }

    if( Trace ) {
      fprintf( FicLog , "Saving morphological and semantic links\n");
      PhotoMemoire( FicLog );
    }

    /* calcul de la taille du dictionnaire */
    fseek( FicBDicLiens , 0L , 2 );
    fgetpos( FicBDicLiens , &tailleDico );
    fseek( FicBDicLiens , 0L , 0 );

    /* allocation du buffer de liens */
    if( BufferLiensMorphologiquesEtSemantiques )
    {
        _Free( BufferLiensMorphologiquesEtSemantiques );
    }
    BufferLiensMorphologiquesEtSemantiques 
      = (char *)_Alloc( tailleDico + 10L );
    if( BufferLiensMorphologiquesEtSemantiques )
    {
         nbLus = LireEcrire( 0L , tailleDico , 
			     BufferLiensMorphologiquesEtSemantiques , 
                             _Lire, FicBDicLiens );
         if( nbLus != tailleDico )
         {
               sprintf( ChAux ,       
             "EditionLiensMorphologiquesEtSemantiques(alinsere) [%ld/->%ld]",
                        tailleDico , nbLus );
               TxtErreur( 58 , ChAux );
         }
    }
    else {
      TxtErreur( 328 , "EditionLiensMorphologiquesEtSemantiques(alinsere)" );
    }

    fprintf( FicLog , "%ld morphological link(s) updated\n" , 
	     NbLiensMorphologiques );
    fprintf( FicLog , "%ld semantic link(s) updated\n\n" , 
	     NbLiensSemantiques );
    _FinDialogueInfo();

    if( Trace ) {
      PhotoMemoire( FicLog );
    }
}

/*************************************************************************
   Des fonctions appliquees sur les lemmes au moyen de MapLitEtModifieLemmes
*************************************************************************/
char     ResetsMorphAndSemLinks( TabLemmeMem p_ptLemme ,
				 unsigned long p_posFicLemme ,
				 TabChar p_tCharRegles )
{
    p_ptLemme->NbDerives = _Nil;
    p_ptLemme->PosFicDerives = _Nil;
    p_ptLemme->PosFicRacine = _Nil;
    p_ptLemme->NbVoisinsSemantiques = _Nil;
    p_ptLemme->NbVoisinsSemantiquesRec = _Nil;
    p_ptLemme->PosFicLiensSemantiques = _Nil;
    p_ptLemme->PosFicLiensSemantiquesRec = _Nil;
    return _True;
}

/**************************************************************************
         GARNIT DES LES MORPHOLOGIQUES EN FONCTION DES TRAITS
***************************************************************************/
char     GarnitLiensMorphologiques( TabLemmeMem p_ptLemme ,
				    unsigned long p_posFicLemme ,
				    TabChar p_tCharRegles )
{
    TabDagMem             *hdlDagListe;
    struct LemmeMem        lemmeRacine;
    TabLong                ptReference;
    TabChar                ptCh;
    TabLong                ptCategorie;
    TabLong                ptNoFlexion;  
    char                   estMotCourant;
    unsigned long          posFicLemmeRacineIni;
    unsigned long          posFicLemmeRacine;
    long                   tailleLemmeRacine;
    long                   longTChar;
    ChIdent                chaineAnnexe;
    struct InterFileLink   unLien;
    TabDagMem              ptDag;

    /****************** La racine du lemme derive **************************/
    MiseAZeroUnLemme( &lemmeRacine );
    hdlDagListe = ExisteCheminFils( CheminRacine ,  
				    &(p_ptLemme->ValTerme.ArcsFils) );

    /****************** Lemme derive sans racine ***************************/
    if( !hdlDagListe || !(*hdlDagListe) )
    {
       return _True;
    }
    /************* loads the list describing the root lemma *************/
    if( !LienUniteLexicaleATraits( hdlDagListe , 
				   &(lemmeRacine.ValTerme.ArcsFils) ,
				   &ptReference , &ptCh ,
				   &ptCategorie , &ptNoFlexion , _True ) ) {
      TxtErreur( 262 , "GarnitLiensMorphologiques(alinsere)" );
      return _False;
    }

    /****************** Lemme derive avec une racine ***********************/
    /***** recherche de la racine dans le fichier des lemmes ***********/
    strcpy( lemmeRacine.ValTerme.Ident , ptCh );
    if( ( tailleLemmeRacine =
             ReferenceTermeLexical( 
                            lemmeRacine.ValTerme.Ident , 
                            &(lemmeRacine.ValTerme) , 
                            &estMotCourant , 
			    &posFicLemmeRacine ) ) <= 0L )
    {
      int catLexTerme;

      /* racine non retrouvee */
      TxtChaineErreur( 316 , ChAux );
      catLexTerme = CategorieLexicale( &(lemmeRacine.ValTerme.ArcsFils) );
      sprintf( ChErreurCompil , "%s/%s" , 
	       lemmeRacine.ValTerme.Ident ,
	       DicoTraitsCatLex[ catLexTerme ].ChaineDico );
      /* affiche une erreur sans afficher de regle */
      /* MessageErreurCompil( _Nil , &c , NULL ); */
      DemarqueEtRendDag( &(lemmeRacine.ValTerme.ArcsFils) );
      TxtErreur( 316 , ChErreurCompil );
      return _False;
    }
    /***** racine retrouvee  ***********/
    else {
      /* on remet a 0 le dag de la racine qui portait temporairement 
	 la structure des traits de la racine */
      DemarqueEtRendDag( &(lemmeRacine.ValTerme.ArcsFils) );
      
      /* la racine est un terme du dico memoire */
      if( estMotCourant ) {
	int catLexTerme;

	TxtChaineErreur( 332 , ChAux );
	catLexTerme = CategorieLexicale( &(lemmeRacine.ValTerme.ArcsFils) );
	sprintf( ChErreurCompil , "%s/%s" , 
		 lemmeRacine.ValTerme.Ident ,
	       DicoTraitsCatLex[ catLexTerme ].ChaineDico );
	/* MessageErreurCompil( _Nil , &c , NULL ); */
	TxtErreur( 332 , ChErreurCompil );
	return _False;
      }
      /* la racine n'est pas un terme du dico memoire */
      else {
	/* la racine est bien un terme du dico memoire */
	/* on recupere le lemme de la racine */
	/* lemme de la racine non extrait */
	posFicLemmeRacineIni = posFicLemmeRacine;
	if( RestructurerBDicLemme( &posFicLemmeRacine , 
				   lemmeRacine.ValTerme.Ident  ,    
				   &lemmeRacine , chaineAnnexe )  == _Nil ) {
	  TxtErreur( 266 , "GarnitLiensMorphologiques(alinsere)" );
	  return _False;
	}
	/* lemme de la racine non extrait */
      
	/* complements the ref and infl number from the lexicon */
	ptDag = (*hdlDagListe)->Fils; /* reference */
	ptDag->ValDag.Atome.ValAtome.Entier 
	  = lemmeRacine.ValTerme.Ref;
	ptDag = ptDag->Frere; /* lemma */
	ptDag = ptDag->Frere; /* cat */
	ptDag = ptDag->Frere; /* infl */
	ptDag->ValDag.Atome.ValAtome.Ref 
	  = NoTypeSuffixage( &(lemmeRacine.ValTerme.ArcsFils) , 
			     _True );
	
	/************* Lemme qui est sa propre racine ******************/
	if( lemmeRacine.ValTerme.Ref == p_ptLemme->ValTerme.Ref ) {
	  /* on ne fait rien, pas de chainage necessaire */
	  p_ptLemme->ValTerme.RefRacine = p_ptLemme->ValTerme.Ref;
	}
	/******** Lemme qui n'est pas sa propre racine *****************/
	else {
	  /* modification des pointeurs et insertion */
	  /* racine n'ayant pas encore de derives */
	  if( lemmeRacine.PosFicDerives == _Nil ) {
	    lemmeRacine.PosFicDerives 
	      = TailleFicLiensMorphologiquesEtSemantiques;
	    lemmeRacine.NbDerives = 1;
	    unLien.PtNextLink = _Nil;
	  }
	  /* racine ayant derives, on chaine le nouveau lien en tete
             des liens existants */
	  else {
	    (lemmeRacine.NbDerives)++;
	    unLien.PtNextLink = lemmeRacine.PosFicDerives;
	  }
	  /* racine ayant derives */
	  
	  unLien.PtMorphOrSemNeighbor = p_posFicLemme;
	  lemmeRacine.PosFicDerives 
	    = TailleFicLiensMorphologiquesEtSemantiques;
	  TailleFicLiensMorphologiquesEtSemantiques 
	    += LireEcrire( TailleFicLiensMorphologiquesEtSemantiques , 
			   (unsigned long)sizeof(struct InterFileLink) ,
			   (TabChar)(&unLien) , 
			   _Ecrire , FicBDicLiens );
	
	  /* on garnit le pointeur fichier de la racine sur le derive */
	  p_ptLemme->PosFicRacine  = posFicLemmeRacineIni;
	  p_ptLemme->ValTerme.RefRacine = lemmeRacine.ValTerme.Ref;
	
	  /* on recrit la racine sur elle meme car elle n'a pas change
             de taille */
	  longTChar = DestructurerLemme( lemmeRacine.ValTerme.Ident , 
					 &lemmeRacine , p_tCharRegles , 
					 chaineAnnexe );
	  if( longTChar != tailleLemmeRacine ) {
	    TxtErreur( 164 , "GarnitLiensMorphologiques(alinsere)" );
	  }
	  else {
	    LireEcrire( posFicLemmeRacineIni , longTChar , 
			p_tCharRegles , _Ecrire , FicBDicLemmes );
	  }
	}
	/******** Lemme qui n'est pas sa propre racine *****************/
      }
      /* la racine n'est pas un terme du dico memoire */
      DemarqueEtRendDag( &(lemmeRacine.ValTerme.ArcsFils) );
    }
    /***** racine retrouvee  ***********/

    return _True;
}

/**************************************************************************
         GARNIT DES LES SEMANTIQUES EN FONCTION DES TRAITS
***************************************************************************/
char     GarnitLiensSemantiques( TabLemmeMem p_ptLemme ,
				 unsigned long p_posFicLemme ,
				 TabChar p_tCharRegles )
{
    TabDagMem             *hdlDagListe;
    TabDagMem              ptDag;
    struct LemmeMem        semantLinkedLemma;
    TabLong                ptReference;
    TabChar                ptCh;
    TabLong                ptCategorie;
    TabLong                ptNoFlexion;  
    char                   estMotCourant;
    unsigned long          posFicSemantLinkedLemmaIni;
    unsigned long          posFicSemantLinkedLemma;
    ChIdent                chaineAnnexe;
    int                    indChemin;
    long                   sizeOfSemantLinkedLemma;
    struct InterFileLink   unLien;
    long                   longTChar;
    int                    indLinkedWord;
    int                    nbLinkedWord;
    long                   linkedWordRef[ _NbMaxLiensSem ];

    /****************** scans all the semantic features *********************/
    indChemin = 0;
    while( CheminsSemantique[ indChemin ] ) {
      nbLinkedWord = 0;
      if( Trace ) {
	EcrireDag( CheminsSemantique + indChemin  , _False , 
		   FicLog , _Sortie );
	fprintf( FicLog , "\n" );
      }

      hdlDagListe = ExisteCheminFils( CheminsSemantique[ indChemin ] ,  
				      &(p_ptLemme->ValTerme.ArcsFils) );

      /****************** semantic links can be disunctions *****************/
      /**************** scans all the links in a disjunction ****************/
      while( hdlDagListe && *hdlDagListe ) {

	/* loads the list describing the target lexical item */
	MiseAZeroUnLemme( &semantLinkedLemma );
	if( !LienUniteLexicaleATraits( hdlDagListe , 
				       &(semantLinkedLemma.ValTerme.ArcsFils) ,
				       &ptReference , &ptCh ,
				       &ptCategorie , &ptNoFlexion , 
				       _True ) ) {
	  TxtErreur( 262 , "GarnitLiensSemantiques(alinsere)" );
	  return _False;
	}

	if( Trace ) {
	  fprintf( FicLog , "Chaine lien semantique [%s]\n" , ptCh );
	}

	/***** retrieves the semantically linked lemma from the
               lemma dictionary (if it is found in this file) ********/
	strcpy( semantLinkedLemma.ValTerme.Ident , ptCh );
	if( ( sizeOfSemantLinkedLemma =
             ReferenceTermeLexical( 
                     semantLinkedLemma.ValTerme.Ident , 
		     &(semantLinkedLemma.ValTerme) , 
		     &estMotCourant , 
		     &posFicSemantLinkedLemma ) ) <= 0L ) {
	    int catLexTerme;

	    /* semantically linked lemma not retrieved  */
	    TxtChaineErreur( 317 , ChAux );
	    catLexTerme 
	      = CategorieLexicale( &(semantLinkedLemma.ValTerme.ArcsFils) );
	    sprintf( ChErreurCompil , "%s/%s" ,
		     semantLinkedLemma.ValTerme.Ident ,
		     DicoTraitsCatLex[ catLexTerme ].ChaineDico );
	    /* MessageErreurCompil( _Nil , &c , NULL ); */
	    DemarqueEtRendDag( &(semantLinkedLemma.ValTerme.ArcsFils) );
	    TxtErreur( 317 , ChErreurCompil );
	    return _False;
	}
	/* semantically linked lemma retrieved */
	else {
	  /* desalllocation of the feature str of the semantically
	     linked lemma */
	  DemarqueEtRendDag( &(semantLinkedLemma.ValTerme.ArcsFils) );
      
	  /* semantically linked lemma in memory dictionary */
	  if( estMotCourant ) {
	    int catLexTerme;

	    TxtChaineErreur( 333 , ChAux );
	    catLexTerme 
	      = CategorieLexicale( &(semantLinkedLemma.ValTerme.ArcsFils) );
	    sprintf( ChErreurCompil , "%s/%s" ,
		     semantLinkedLemma.ValTerme.Ident ,
		     DicoTraitsCatLex[ catLexTerme ].ChaineDico );
	    /* MessageErreurCompil( _Nil , &c , NULL ); */
	    TxtErreur( 333 , ChErreurCompil );
	    return _False;
	  }
	  /* semantically linked lemma not in memory dictionary */
	  else {
	    if( Trace ) {
	      fprintf( FicLog , "Extraction dico [%s]\n" , 
		       semantLinkedLemma.ValTerme.Ident );
	    }
	  
	    for( indLinkedWord = 0 ; indLinkedWord < nbLinkedWord ; 
		 indLinkedWord++ ) {
	      if( semantLinkedLemma.ValTerme.Ref
		  == linkedWordRef[ indLinkedWord ] ) {
		sprintf( ChAux , 
			 "[%s of %s] GarnitLiensMorphologiques(alinsere)" , 
			 semantLinkedLemma.ValTerme.Ident , 
			 p_ptLemme->ValTerme.Ident );
		fprintf( FicLog , "Linking error\n" );
		TxtErreur( 330 , ChAux );
		return _False;
	      }
	    }
	    linkedWordRef[ nbLinkedWord++ ] = semantLinkedLemma.ValTerme.Ref;
	    if( nbLinkedWord >= _NbMaxLiensSem ) {
	      TxtErreur( 331 , "GarnitLiensSemantiques(alinsere)" );
	      return _False;
	    }

	    /*semantically linked lemma downloaded from memory dictionary */
	    posFicSemantLinkedLemmaIni = posFicSemantLinkedLemma;
	    if( RestructurerBDicLemme( &posFicSemantLinkedLemma , 
				       semantLinkedLemma.ValTerme.Ident  ,    
				       &semantLinkedLemma , chaineAnnexe )  
		== _Nil ) {
	      TxtErreur( 304 , "GarnitLiensSemantiques(alinsere)" );
	      return _False;
	    }
	    
	    /* complements the ref and infl number from the lexicon */
	    ptDag = (*hdlDagListe)->Fils; /* reference */
	    ptDag->ValDag.Atome.ValAtome.Entier 
	      = semantLinkedLemma.ValTerme.Ref;
	    ptDag = ptDag->Frere; /* lemma */
	    ptDag = ptDag->Frere; /* cat */
	    ptDag = ptDag->Frere; /* infl */
	    ptDag->ValDag.Atome.ValAtome.Ref 
	      = NoTypeSuffixage( &(semantLinkedLemma.ValTerme.ArcsFils) , 
				 _True );
	    
	    /********* Lemme qui est son propre lien semantique  *************/
	    if( semantLinkedLemma.ValTerme.Ref == p_ptLemme->ValTerme.Ref ) {
	      /* on ne fait rien d'autre, car le lemme est lie a lui-meme
		 il n'est donc pas necessaire de faire un chainage */
	    }
	    /********* Lemme qui n'est pas son propre lien semantique  *******/
	    else {
	      /* modification des pointeurs et insertion : liens
		 semantiques reciproques des lemmes semantiquement lies */
	      /* lemme semantiquement lie sans autre lien semantique */
	      if( semantLinkedLemma.NbVoisinsSemantiquesRec
		  == _Nil ) {
		semantLinkedLemma.NbVoisinsSemantiquesRec = 1;
		unLien.PtNextLink = _Nil;
	      }
	      /* lemme semantiquement lie ayant des liens semantiques 
	         reciproques */
	      else {
		(semantLinkedLemma.NbVoisinsSemantiquesRec)++;
		unLien.PtNextLink 
		  = semantLinkedLemma.PosFicLiensSemantiquesRec;
	      }
	      unLien.PtMorphOrSemNeighbor 
		= p_posFicLemme;
	      semantLinkedLemma.PosFicLiensSemantiquesRec 
		= TailleFicLiensMorphologiquesEtSemantiques;
	      TailleFicLiensMorphologiquesEtSemantiques +=
		LireEcrire( TailleFicLiensMorphologiquesEtSemantiques , 
			    (unsigned long)sizeof(struct InterFileLink) ,
			    (TabChar)(&unLien) , 
			    _Ecrire , FicBDicLiens );

	      /* modification des pointeurs et insertion : lemme
	         courant */
	      /* lemme courant sans autre lien semantique */
	      if( p_ptLemme->NbVoisinsSemantiques == _Nil ) {
		p_ptLemme->NbVoisinsSemantiques = 1;
		unLien.PtNextLink = _Nil;
	      }
	      /* lemme courant ayant des liens semantiques */
	      else {
		(p_ptLemme->NbVoisinsSemantiques)++;
		unLien.PtNextLink 
		  = p_ptLemme->PosFicLiensSemantiques;
	      }
	      unLien.PtMorphOrSemNeighbor 
		= posFicSemantLinkedLemmaIni;
	      p_ptLemme->PosFicLiensSemantiques 
		= TailleFicLiensMorphologiquesEtSemantiques;
	      TailleFicLiensMorphologiquesEtSemantiques +=
		LireEcrire( TailleFicLiensMorphologiquesEtSemantiques , 
			    (unsigned long)sizeof(struct InterFileLink) ,
			    (TabChar)(&unLien) , 
			    _Ecrire , FicBDicLiens );

	      /* on recrit la racine sur elle meme car elle n'a pas
                 change de taille */
	      longTChar = DestructurerLemme( 
				semantLinkedLemma.ValTerme.Ident , 
				&semantLinkedLemma , p_tCharRegles , 
				chaineAnnexe );
	      if( longTChar != sizeOfSemantLinkedLemma ) {
		TxtErreur( 164 , "GarnitLiensMorphologiques(alinsere)" );
	      }
	      else {
		LireEcrire( posFicSemantLinkedLemmaIni , 
			    longTChar , 
			    p_tCharRegles , _Ecrire , FicBDicLemmes );
	      }
	    }
	    /********* Lemme qui n'est pas son propre lien semantique  *******/
	  }
	  /* semantically linked lemma not in memory dictionary */
	}
	/* semantically linked lemma retrieved */
	DemarqueEtRendDag( &(semantLinkedLemma.ValTerme.ArcsFils) );
	hdlDagListe = &((*hdlDagListe)->Frere);
      }
      /**************** scans all the links in a disjunction *************/
	
      indChemin++;
    }
    /****************** scans all the semantic features ******************/

    return _True;
}

/**************************************************************************
         REMET CONTIGUS DES LIENS MORPHOLOGIQUES CHAINES
***************************************************************************/
char     SequentializeMorphAndSemanticLinks( TabLemmeMem p_ptLemme ,
					     unsigned long p_posFicLemme ,
					     TabChar p_tCharRegles )
{
    struct InterFileLink   unLien;
    int                    indDerive;
    int                    indLienSem;
    long                   posFileMorphLink;
    long                   posFileSemLink;

    posFileMorphLink = p_ptLemme->PosFicDerives;
    for( indDerive = 0 ; indDerive < p_ptLemme->NbDerives ;  
	 indDerive++ )
    {
       LireEcrire( posFileMorphLink , 
                  (unsigned long)sizeof(struct InterFileLink) ,
                  (TabChar)(&unLien) , 
                  _Lire , FicBDicLiens );

       if( ( unLien.PtNextLink == _Nil 
             && indDerive < p_ptLemme->NbDerives - 1 )
           || ( unLien.PtNextLink != _Nil 
               && indDerive == p_ptLemme->NbDerives - 1 ) )
       {
           TxtErreur( 250 , 
                     "SequentializeMorphAndSemanticLinks(alinsere)" );
           return _False;;
       }

       /* chainage */
       if( indDerive != p_ptLemme->NbDerives - 1 )
       {
           posFileMorphLink = unLien.PtNextLink;
       }
       if( indDerive == 0 )
       {
           /* on garnit le pointeur fichier de la racine sur le derive */
           p_ptLemme->PosFicDerives  
	     = TailleFicLiensMorphologiquesEtSemantiques;
       }

       TailleFicLiensMorphologiquesEtSemantiques 
           += LireEcrire( TailleFicLiensMorphologiquesEtSemantiques , 
                          (unsigned long)sizeof(long) ,
                          (TabChar)(&(unLien.PtMorphOrSemNeighbor)) , 
                          _Ecrire , FicBDicLiensAux );
       NbLiensMorphologiques++;

   }

    posFileSemLink = p_ptLemme->PosFicLiensSemantiques;
    for( indLienSem = 0 ; indLienSem < p_ptLemme->NbVoisinsSemantiques ;  
	 indLienSem++ )
    {
       LireEcrire( posFileSemLink , 
                  (unsigned long)sizeof(struct InterFileLink) ,
                  (TabChar)(&unLien) , 
                  _Lire , FicBDicLiens );

       if( ( unLien.PtNextLink == _Nil 
             && indLienSem < p_ptLemme->NbVoisinsSemantiques - 1 )
           || ( unLien.PtNextLink != _Nil 
               && indLienSem == p_ptLemme->NbVoisinsSemantiques - 1 ) )
       {
           TxtErreur( 250 , 
                     "SequentializeMorphAndSemanticLinks(alinsere)" );
           return _False;;
       }

       /* chainage */
       if( indLienSem != p_ptLemme->NbVoisinsSemantiques - 1 )
       {
           posFileSemLink = unLien.PtNextLink;
       }
       if( indLienSem == 0 )
       {
           /* on garnit le pointeur fichier de la racine sur le derive */
           p_ptLemme->PosFicLiensSemantiques  
	     = TailleFicLiensMorphologiquesEtSemantiques;
       }

       TailleFicLiensMorphologiquesEtSemantiques 
           += LireEcrire( TailleFicLiensMorphologiquesEtSemantiques , 
                          (unsigned long)sizeof(long) ,
                          (TabChar)(&(unLien.PtMorphOrSemNeighbor)) , 
                          _Ecrire , FicBDicLiensAux );
       NbLiensSemantiques++;
   }

    posFileSemLink = p_ptLemme->PosFicLiensSemantiquesRec;
    for( indLienSem = 0 ; indLienSem < p_ptLemme->NbVoisinsSemantiquesRec ;  
	 indLienSem++ )
    {
       LireEcrire( posFileSemLink , 
                  (unsigned long)sizeof(struct InterFileLink) ,
                  (TabChar)(&unLien) , 
                  _Lire , FicBDicLiens );

       if( ( unLien.PtNextLink == _Nil 
             && indLienSem < p_ptLemme->NbVoisinsSemantiquesRec - 1 )
           || ( unLien.PtNextLink != _Nil 
               && indLienSem == p_ptLemme->NbVoisinsSemantiquesRec - 1 ) )
       {
           TxtErreur( 250 , 
                     "SequentializeMorphAndSemanticLinks(alinsere)" );
           return _False;;
       }

       /* chainage */
       if( indLienSem != p_ptLemme->NbVoisinsSemantiquesRec - 1 )
       {
           posFileSemLink = unLien.PtNextLink;
       }
       if( indLienSem == 0 )
       {
           /* on garnit le pointeur fichier de la racine sur le derive */
           p_ptLemme->PosFicLiensSemantiquesRec  
	     = TailleFicLiensMorphologiquesEtSemantiques;
       }

       TailleFicLiensMorphologiquesEtSemantiques 
           += LireEcrire( TailleFicLiensMorphologiquesEtSemantiques , 
                          (unsigned long)sizeof(long) ,
                          (TabChar)(&(unLien.PtMorphOrSemNeighbor)) , 
                          _Ecrire , FicBDicLiensAux );
       NbLiensSemantiques++;
   }
   return _True;
}

/*****************************************************************************
                 adds the self feature
*****************************************************************************/
TabDagMem      *AddsSelfFeature( TabDagMem *p_hdlDagFilsTerm , long p_ref , 
								TabChar p_ident , int p_catLex , int p_indFlexion )
{
  TabDagMem                      *hdlDagFilsCheminSelf;
  TabDagMem                      *hdlDagFilsListeCou;

  /* le trait self existe deja */
  if( ( hdlDagFilsCheminSelf 
	 = ExisteCheminFils( CheminSelf , p_hdlDagFilsTerm ) ) ) {
    return hdlDagFilsCheminSelf;
  }

  /* ajout du trait self qui point vers le quadruplet */
  else {
    /* copies the self path */
    if( ( hdlDagFilsCheminSelf 
	  = CopieChemin( p_hdlDagFilsTerm , CheminSelf ) ) ) {
      /* builds the self list */
      *hdlDagFilsCheminSelf = NouveauDag( _AtomeListe );
      if( !(*hdlDagFilsCheminSelf) ) {
	return NULL;
      }
      (*hdlDagFilsCheminSelf)->ValDag.Atome.ValAtome.Entier = _Nil;
      hdlDagFilsListeCou = &((*hdlDagFilsCheminSelf)->Fils);

      /* builds the self list */
      /* the reference */
      *hdlDagFilsListeCou = NouveauDag( _AtomeEntier );
      if( !(*hdlDagFilsListeCou) ) {
	return NULL;
      }
      (*hdlDagFilsListeCou)->ValDag.Atome.ValAtome.Entier = p_ref;
      hdlDagFilsListeCou = &((*hdlDagFilsListeCou)->Frere);

      /* the lemma */
      *hdlDagFilsListeCou = NouveauDag( _AtomeChaine );
      if( !(*hdlDagFilsListeCou) ) {
	return NULL;
      }
      strcpy( (*hdlDagFilsListeCou)->ValDag.Atome.ValAtome.Chaine , p_ident );
      hdlDagFilsListeCou = &((*hdlDagFilsListeCou)->Frere);

      /* the category */
      *hdlDagFilsListeCou = NouveauDag( _AtomeRef );
      if( !(*hdlDagFilsListeCou) ) {
	return NULL;
      }
      (*hdlDagFilsListeCou)->ValDag.Atome.ValAtome.Ref = p_catLex;
      hdlDagFilsListeCou = &((*hdlDagFilsListeCou)->Frere);

      /* the inflection */
      *hdlDagFilsListeCou = NouveauDag( _AtomeEntier );
      if( !(*hdlDagFilsListeCou) ) {
	return NULL;
      }
      (*hdlDagFilsListeCou)->ValDag.Atome.ValAtome.Entier = p_indFlexion;

      return hdlDagFilsCheminSelf;
    }
    else  {
      return  NULL;
    }
  }
  /* ajout du trait self qui point vers le quadruplet */
}

/*************************************************************************
   Des fonctions appliquees sur les lemmes au moyen de MapLitEtModifieLemmes
*************************************************************************/
TabDagMem     *AddsRootFeature( TabLemmeMem p_ptLemme )
{
    TabDagMem                 *hdlDagFilsCheminRacine;
    TabDagMem                 *hdlDagFilsCheminSelf;

  /* le trait racine existe deja */
  if( ( hdlDagFilsCheminRacine 
	 = ExisteCheminFils( CheminRacine , 
			     &(p_ptLemme->ValTerme.ArcsFils) ) ) ) {
    return hdlDagFilsCheminRacine;
  }

  /* ajout du trait racine qui point vers le self */
  else {
    /* on commence par ajouter le chemin self s'il n'existe pas car il
       va coreferer avec le chemin racine automatique */
    if( !(hdlDagFilsCheminSelf 
	  = AddsSelfFeature( &(p_ptLemme->ValTerme.ArcsFils) ,
			     p_ptLemme->ValTerme.Ref ,
			     p_ptLemme->ValTerme.Ident ,
			     CategorieLexicale( 
					&(p_ptLemme->ValTerme.ArcsFils) ) ,
			     NoTypeSuffixage( &(p_ptLemme->ValTerme.ArcsFils) ,
					      _True ) ) ) ) {
      return NULL;
    }

    /****************** Lemme derive sans racine ***************************/
    /* on lui attribue lui-meme comme racine */
    /* on commence par copier le chemin de la racine */
    if( (hdlDagFilsCheminRacine
	 = CopieChemin( &(p_ptLemme->ValTerme.ArcsFils) ,
			CheminRacine ) ) ) {
      if( *hdlDagFilsCheminRacine ) {
	TxtErreur( 312 , "AddsRootFeature(alinsere)" );
	return NULL;
      }
      
      /* puis on le prolonge par l'extremite du chemin self */
      *hdlDagFilsCheminRacine = *hdlDagFilsCheminSelf;

      return hdlDagFilsCheminRacine;
    }
    else {
      return NULL;
    }
  }
}

/*************************************************************************
   Des fonctions appliquees sur les lemmes au moyen de MapLitEtModifieLemmes
*************************************************************************/
TabDagMem     *AddsSemanticFeatures( TabLemmeMem p_ptLemme , 
				     TabDagMem p_ptCheminSemantique )
{
    TabDagMem                 *hdlDagFilsCheminSemantique;
    TabDagMem                 *hdlDagFilsCheminSelf;

  /* le trait semantique existe deja */
  if( ( hdlDagFilsCheminSemantique 
	 = ExisteCheminFils( p_ptCheminSemantique , 
			     &(p_ptLemme->ValTerme.ArcsFils) ) ) ) {
    return hdlDagFilsCheminSemantique;
  }

  /* ajout du trait semantique qui point vers le self */
  else {
    /* on commence par ajouter le chemin self s'il n'existe pas car il
       va coreferer avec les chemins semantiques automatiques */
    if( !(hdlDagFilsCheminSelf 
	  = AddsSelfFeature( &(p_ptLemme->ValTerme.ArcsFils) ,
			     p_ptLemme->ValTerme.Ref ,
			     p_ptLemme->ValTerme.Ident ,
			     CategorieLexicale( 
					&(p_ptLemme->ValTerme.ArcsFils) ) ,
			     NoTypeSuffixage( &(p_ptLemme->ValTerme.ArcsFils) ,
					      _True ) ) ) ) {
      return NULL;
    }

     /****************** Lemme sans lien semantique *********************/
     /* on lui attribue lui-meme comme lien semantique */
     /* on commence par copier le chemin semantique */
     if( (hdlDagFilsCheminSemantique
           = CopieChemin( &(p_ptLemme->ValTerme.ArcsFils) ,
                          p_ptCheminSemantique ) ) ) {
       if( *hdlDagFilsCheminSemantique ) {
              TxtErreur( 312 , "AddsSemanticFeatures(alinsere)" );
              return NULL;
       }
	 
       /* puis on le prolonge par l'extremite du chemin self */
       *hdlDagFilsCheminSemantique = *hdlDagFilsCheminSelf;

       return hdlDagFilsCheminSemantique;
     }
     else {
       return NULL;
     }
  }
}

/*************************************************************************
     insertion d'une regle evetuellement rattachee a un terme
*************************************************************************/
char     InsertionDesallocRegle( TabRegleMem *p_hdlRegleAInserer ,
                struct ValTerme *p_ptValTermeLexicalisation )
{
    unsigned long                   indAbsIni;
    unsigned long                   indAbsCou;
    unsigned long                   tailFicRegles;

    long                            indAbsReglePrec;

    struct LemmeMem                 lemmeCou;
    int                             nbReglesCou;
    long                            longTChar;

    long                            buff[ 1 ];/* buffer de lect
                                                 reference absolue de
                                                 l'U.L.R*/
    TabChar                         tCharLemmes;
    TabChar                         tCharRegles;
    char                            trouvee;  /* indique s'il y a une
                                                 autre U.LR de meme
                                                 chaine deja saisie */
    char                            ch[ 256 ];
    ChIdent                         chaineAnnexe;

    (*p_hdlRegleAInserer)->RegleFrere = NULL;

    tCharLemmes = AlloueBuffer( TailleBuffLemmesCourt );
    tCharRegles = AlloueBuffer( TailleBuffReglesCourt );
    if( !tCharLemmes || !tCharRegles )
    {
        TxtErreur( 81 , "InsertionDesallocRegle(alinsere)" );
        return _False;
    }

    /* si la regle n'est pas une regle lexicale et si elle n'a pas de
       terme de rattachement, on la rattache a un terme bidon de la
       forme *xxxx ou xxx est un entier unique */
    if( !p_ptValTermeLexicalisation )
    {
       /* reference du Lemme incrementee a chaque ajout dans le
          fichier, et se trouvant en tete du fichier FicBDicLemmes */
       LireEcrire( 0L , (long) _TailleAmorce , (TabChar)buff ,
                   _Lire , FicBDicLemmes );

       MiseAZeroUnLemme( &lemmeCou );
       sprintf( (lemmeCou.ValTerme.Ident + 1) , "%ld" , buff[ 0 ] );
       *(lemmeCou.ValTerme.Ident) = '*';

       TxtChaineRsrc( 134 , 1 , ch );
       if( (*p_hdlRegleAInserer)->TermeFils )
       {
           sprintf( ChAux , "%s [%s]" , ch , 
                    (*p_hdlRegleAInserer)->TermeFils->ValTerme.Ident );
       }
       else
       {
           sprintf( ChAux , "%s " , ch );
       }
       fprintf( FicLog , ChAux );
       TxtChaineRsrc( 134 , 9 , ch );
       sprintf( ChAux , " %s --> [%s] \n" , ch , lemmeCou.ValTerme.Ident );
       fprintf( FicLog , ChAux );

       /* insertion du nouveau lemme */
       lemmeCou.NbRegles = _Nil;
       if( !InsertionLemme( &lemmeCou , _True , "" ) ) {
	 return _False;
       }

       /* insertion du nouveau terme */
       p_ptValTermeLexicalisation = &(lemmeCou.ValTerme);
    }
    else
    {
       /* la regle a un mot de rattachement : on la rattache a ce mot */
       TxtChaineRsrc( 134 , 1 , ch );
       if( (*p_hdlRegleAInserer)->TermeFils )
       {
           sprintf( ChAux , "%s [%s]" , ch , 
                    (*p_hdlRegleAInserer)->TermeFils->ValTerme.Ident );
       }
       else
       {
           sprintf( ChAux , "%s " , ch );
       }
       fprintf( FicLog , ChAux );
       TxtChaineRsrc( 134 , 10 , ch );
       sprintf( ChAux ,  " %s --> [%s] \n" , ch , 
                          p_ptValTermeLexicalisation->Ident );
       fprintf( FicLog , ChAux );
    }

    /* recherche du lieu d'insertion                                 */
    indAbsIni = IndiceInsertionDicLemmes( p_ptValTermeLexicalisation->Ident ,
                                          &trouvee );
    indAbsCou = indAbsIni;

    /* BOUCLE SUR LES LEMMES DE MEME RACINE                    */
    /* on recherche le terme de rattachement ds le dico */
    while( EgaliteChaine( BufferLemmesCourant + _IRelLemme( indAbsCou ) ,
                          p_ptValTermeLexicalisation->Ident ) == _True )
    {
          indAbsIni = indAbsCou;
          if( RestructurerBDicLemme( &indAbsCou , lemmeCou.ValTerme.Ident  ,
                                     &lemmeCou , chaineAnnexe ) == _Nil
              || EgaliteChaine( lemmeCou.ValTerme.Ident , DernLemme ) )
          {
                 DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
                 /* lemme non retrouve */
                 break;
          }

          /* le lemme a la bonne reference et on n'est pas sur une
             racine annexe */
          nbReglesCou = lemmeCou.NbRegles;
          if( lemmeCou.ValTerme.Ref == p_ptValTermeLexicalisation->Ref 
              && !( nbReglesCou <= _RacineAnnexe1 
                    && nbReglesCou >= _RacineAnnexeMax ) )
          {
              /* incrementation du nombre de regles */
              if( nbReglesCou > 0 )
              {
                  indAbsReglePrec = lemmeCou.ValTerme.PosFicRegle;
                  (lemmeCou.NbRegles)++;
              }
              else
              {
                  /* regle unique */
                  indAbsReglePrec = (long)_Nil;
                  lemmeCou.NbRegles = 1;
              }

              /* recherche du lieu d'insertion de la regle */
              fseek( FicBDicRegles , 0L , 2 );
              fgetpos( FicBDicRegles , &tailFicRegles );

              /* insertion du lemme et reecriture par dessus le lemme
                 initial */
              /* pour modifier son nombre de regles */
              lemmeCou.ValTerme.PosFicRegle = (long)tailFicRegles;
              longTChar = DestructurerLemme( lemmeCou.ValTerme.Ident , 
                                             &lemmeCou ,
                                             tCharLemmes , chaineAnnexe );
              if( (unsigned long)longTChar != indAbsCou - indAbsIni )
              {
                    TxtErreur( 164 , "InsertionDesallocRegle(alinsere)" );
                    DemarqueEtRendRegle( p_hdlRegleAInserer );
		    DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
                    _Free( tCharLemmes );
                    _Free( tCharRegles );
                    return _False;
              }
              LireEcrire( indAbsIni , longTChar , tCharLemmes , _Ecrire , 
                          FicBDicLemmes );
              DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );

              /* insertion de la regle en queue du fichier de regle et
                 chainage */
              /* avec les autres regles du lemme */
              (*p_hdlRegleAInserer)->PosFicRegleSuivante = indAbsReglePrec;
              longTChar = DestructurerUneRegle( (*p_hdlRegleAInserer) , 
                                                tCharRegles );
              if( longTChar > 0L && longTChar < TailleBuffReglesCourt )
              {
                    LireEcrire( tailFicRegles , longTChar , tCharRegles , 
                                _Ecrire , FicBDicRegles );
                    DemarqueEtRendRegle( p_hdlRegleAInserer );
                    _Free( tCharLemmes );
                    _Free( tCharRegles );
                    return _True;
              }
              else
              {
                    TxtErreur( 164 , "InsertionDesallocRegle(alinsere)" );
                    DemarqueEtRendRegle( p_hdlRegleAInserer );
                    _Free( tCharLemmes );
                    _Free( tCharRegles );
                    return _False;
              }
          }

          DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
    } /* on recherche le terme de rattachement ds le dico */

    /* Lexicalisation word not found in dictionary */
    TxtChaineRsrc( 134 , 11 , ch );

    if( p_ptValTermeLexicalisation )
    {
        if( (*p_hdlRegleAInserer)->TermeFils )
        {
            sprintf( ChErreurCompil , "%s[%s] [%s]\n", ch , 
                     (*p_hdlRegleAInserer)->TermeFils->ValTerme.Ident,
                     p_ptValTermeLexicalisation->Ident );
        }
        else
        {
            sprintf( ChErreurCompil , "%s [%s]\n", ch ,
                     p_ptValTermeLexicalisation->Ident );
        }
    }

    DemarqueEtRendListeRegle( p_hdlRegleAInserer );
    _Free( tCharLemmes );
    _Free( tCharRegles );
    return _False;
}

/*************************************************************************
                    insertion d'un Lemme + tTermeAux + ch
*************************************************************************/
char     InsertionLemme( TabLemmeMem p_ptLemme , int p_avecTrace ,
                         TabChar p_chLogRacine )
{
    unsigned long               indAbs;
    long                        longTChar;
    TabChar                     tCharLemme;
    char                        trouvee;/* indique s'il y a une autre Lemme
                                           de meme chaine deja saisie    */
    long                        refMotsSimples;
    long                        refMotsSimplesAutomatique;
    int                         indChaineAnnexe;
    ChIdent                     chaineAnnexe;
    TabDagMem                  *hdlFilsRacAnnexes;
    TabDagMem                   ptAtomesRacAnnexes;
    struct ValAtomeMem          atome;

    ChIdent                     ch;
    int                         indChemin;

    tCharLemme = AlloueBuffer( TailleBuffLemmesCourt );
    if( !tCharLemme )
    {
        TxtErreur( 81 , "InsertionLemme(alinsere)" );
        return _False;
    }

    /* recherche du lieu d'insertion                                 */
    indAbs = IndiceInsertionDicLemmes( p_ptLemme->ValTerme.Ident , &trouvee );

    /* reference du Lemme incrementee a chaque ajout dans le fichier,
       et se trouvant en tete du fichier FicBDicLemmes                      */
    LireEcrire( 0L , (long)_TailleAmorce , 
                     (TabChar)(&refMotsSimplesAutomatique) ,_Lire , 
                     FicBDicLemmes );
    refMotsSimplesAutomatique++;

    /* si le mot porte deja une ref, on la lui laisse. On maintient cependant 
      l'incrementation du compteur en tete de fichier pour connaitre le nb 
      de lemmes presents */
    if( (refMotsSimples = 
         ReferenceLexicale( &(p_ptLemme->ValTerme.ArcsFils) )) != _Nil )
    {
       p_ptLemme->ValTerme.Ref = refMotsSimples;
    }
    else
    {
       p_ptLemme->ValTerme.Ref = refMotsSimplesAutomatique;
      /* ajout de la chaine du lemme pour gerer la negation par
         unification */
      atome.AtomeConjDisj = _False; 
      atome.ValAtome.Ref = refMotsSimplesAutomatique;
      if( !AjouteCheminEtAtomeFils( &(p_ptLemme->ValTerme.ArcsFils) ,
                            CheminReference , _AtomeEntier , atome , _True ) )
      {
          return _False;
      }
    }

    /* reference du Lemme incrementee a chaque ajout dans le fichier,
       et se trouvant en tete du fichier FicBDicLemmes                      */
    LireEcrire( 0L , (long)_TailleAmorce , 
                     (TabChar)(&refMotsSimplesAutomatique) ,_Ecrire , 
                FicBDicLemmes );

    if( p_avecTrace )
    {
        fprintf( FicLog , "Stem insertion [%s] (%ld) %s\n", 
                          p_ptLemme->ValTerme.Ident , 
                          p_ptLemme->ValTerme.Ref , p_chLogRacine );
    }

    if( AutomaticSelfMorphologicalRoot || AutomaticSelfSemanticLinks )
    {
      AddsSelfFeature( &(p_ptLemme->ValTerme.ArcsFils) ,
			     p_ptLemme->ValTerme.Ref ,
			     p_ptLemme->ValTerme.Ident ,
			     CategorieLexicale( 
					&(p_ptLemme->ValTerme.ArcsFils) ) ,
			     NoTypeSuffixage( &(p_ptLemme->ValTerme.ArcsFils) ,
					      _True ) );
    }

    /* la racine n'est pas donnee le lemme est sa propre racine */
    if( AutomaticSelfMorphologicalRoot )
    {
         if( !AddsRootFeature( p_ptLemme ) )
         {
         }
    }

    /* semantic links are not provided, they are automaically added */
    if( AutomaticSelfSemanticLinks ) {
      /****************** scans all the semantic features *****************/
      indChemin = 0;
      while( CheminsSemantique[ indChemin ] ) {
	
	if( !AddsSemanticFeatures( p_ptLemme ,
				   CheminsSemantique[ indChemin ]  ) ) {
	}
	indChemin++;
      }
    }

    /* insertion du lemme */
    *chaineAnnexe = '\0';
    longTChar = DestructurerLemme( p_ptLemme->ValTerme.Ident , p_ptLemme ,
                                   (TabChar)tCharLemme , chaineAnnexe );

    if( longTChar > 0L && longTChar < TailleBuffLemmesCourt ) {
        Inserer( indAbs , indAbs , longTChar , tCharLemme , FicBDicLemmes );
    }
    else {
      _Free( tCharLemme );
      return _False;
    }      

    /********************************************
           racines annexes eventuelles
     ********************************************/
    hdlFilsRacAnnexes = ExisteCheminFils( CheminRacinesAnnexes ,
                                         &(p_ptLemme->ValTerme.ArcsFils) );
    if( hdlFilsRacAnnexes && (*hdlFilsRacAnnexes) )
    {
        if( (*hdlFilsRacAnnexes)->TypeDag != _AtomeListe )
        {
             TxtErreur( 86 , "InsertionLemme(alinsere)" );
             _Free( tCharLemme );
             return _False;
        }

        indChaineAnnexe = 0;
        ptAtomesRacAnnexes = (*hdlFilsRacAnnexes)->Fils;
        while( ptAtomesRacAnnexes )
        {
            if( ptAtomesRacAnnexes->TypeDag != _AtomeChaine )
            {
                TxtErreur( 87 , "InsertionLemme(alinsere)" );
                _Free( tCharLemme );
                return _False;
            }

            strcpy( chaineAnnexe , 
                    ptAtomesRacAnnexes->ValDag.Atome.ValAtome.Chaine );
            /* recherche du lieu d'insertion                            */
            TxtChaineRsrc( 134 , 13 , ch );
            sprintf( ChAux , "%s [%s]\n", ch , chaineAnnexe );
            fprintf( FicLog , ChAux );
            indAbs = IndiceInsertionDicLemmes( chaineAnnexe , &trouvee );
            p_ptLemme->NbRegles = _RacineAnnexe1 - indChaineAnnexe;

            /* reference du Lemme identique a celle du Lemme principal
               sauf le nombre de fils                                    */
            longTChar = DestructurerLemme( p_ptLemme->ValTerme.Ident , 
                                           p_ptLemme ,
                                           (TabChar) tCharLemme , chaineAnnexe );
            if( longTChar > 0L && longTChar < TailleBuffLemmesCourt )
            {
                Inserer( indAbs , indAbs , longTChar , 
                         tCharLemme , FicBDicLemmes );
            }
            ptAtomesRacAnnexes = ptAtomesRacAnnexes->Frere;
            indChaineAnnexe++;
        }
    }
    _Free( tCharLemme );
    return _True;
}

/*************************************************************************
           classement sequentiel des regles par mots croissants
*************************************************************************/
char     ClasseRegles( int p_avecRecomptage )
{
    char                            nomFicDicRegles[ 256 ];
    unsigned long                   tailFicRegles;
    int                             nbReglesLoc;

    unsigned long                   indAbsRegleCou;
    unsigned long                   indAbsLemmeCou;
    unsigned long                   indAbsLemmeIni;
    unsigned long                   longFicLemme;
 
    TabRegleMem                     ptRegleCou;
    TabRegleMem                     ptRegleAux;

    long                            longTChar;
    long                            nbReglesRelevees;

    TabChar                         tCharLemmes;
    TabChar                         tCharRegles;
    
    struct LemmeMem                 lemmeCou;
    ChIdent                         chaineAnnexe;

    FILE                            *pficAux;

    /* sauvegarde des fichiers existants 
    if( !SauveArchivesDictionnaire() )
    {
       TxtErreur( 29 , "ClasseRegles(alinsere)" );
       return _False;
    }
    */

    strcpy( nomFicDicRegles , NomFicDico );
    strcat( nomFicDicRegles , ".RUL" );
    if( !FicBDicRegles )
    {
        sprintf( ChAux , "ClasseRegles(alinsere) [%s]" , nomFicDicRegles );
        TxtErreur( 147 , ChAux );
        _FinDialogueInfo();
        return _False;
    }
    
    pficAux = fopen( NomFicAux , _FopenWPB );
    tCharLemmes = AlloueBuffer( TailleBuffLemmesCourt );
    tCharRegles = AlloueBuffer( TailleBuffReglesCourt );
    if( !pficAux || !tCharLemmes || !tCharRegles )
    {
        TxtErreur( 81 , "ClasseRegles(alinsere)" );
        _FinDialogueInfo();
        return _False;
    }

    _NouveauDialogueInfo( 128 , _True );
    TxtTitresDialogueInfo( 25 , 26 , 0 );

    /* regles fichier */
    indAbsLemmeCou = (long)( _TailleAmorce + _TailleSeparation );
    fseek( FicBDicLemmes , 0L , 2 );
    fgetpos( FicBDicLemmes , &longFicLemme );
    *(lemmeCou.ValTerme.Ident) = '\0';
    tailFicRegles = 0L;
    nbReglesRelevees = 0L;
    while( indAbsLemmeCou < longFicLemme )
    {
        indAbsLemmeIni = indAbsLemmeCou;
        if( RestructurerBDicLemme( &indAbsLemmeCou , lemmeCou.ValTerme.Ident  ,
                                   &lemmeCou , chaineAnnexe ) == _Nil 
            || EgaliteChaine( lemmeCou.ValTerme.Ident , DernLemme ) )
        {
             DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
             break;
        }

        indAbsRegleCou = lemmeCou.ValTerme.PosFicRegle;
        if( RestructurerLesRegles( &indAbsRegleCou , &ptRegleCou , 
                                   lemmeCou.NbRegles , 
                                   !p_avecRecomptage ) == _Nil )
        {
            TxtErreur( 0 , 
                       "ClasseRegles(alinsere) incomplete restructuration" );
            DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
            DemarqueEtRendListeRegle( &ptRegleCou );
            break;
        }

        nbReglesLoc = 0;
        ptRegleAux = ptRegleCou;
        while( ptRegleCou )
        {
              if( nbReglesRelevees++ % 100L == 0 )
              {
                  sprintf( ChAux , "%ld" , nbReglesRelevees );
                  TxtEcritItemInfo( 1 , ChAux );
              }
              nbReglesLoc++;

              if( nbReglesLoc == 1 )
              {
                  /*  modification du pteur de premiere regle du lemme */
                  lemmeCou.ValTerme.PosFicRegle = (long)tailFicRegles;
              }

              /* insertion de la regle en queue du fichier de regle et
                 chainage */
              /* avec les autres regles du lemme */
              ptRegleCou->PosFicRegleSuivante = _Nil;
              longTChar = DestructurerUneRegle( ptRegleCou , tCharRegles );
              if( ptRegleCou->RegleFrere )
              {
                   ptRegleCou->PosFicRegleSuivante = 
                         (long)tailFicRegles + longTChar;
                   DestructurerUneRegle( ptRegleCou , tCharRegles );
              }
              if( longTChar > 0L && longTChar < TailleBuffReglesCourt )
              {
                    LireEcrire( tailFicRegles , longTChar , 
                                tCharRegles , _Ecrire , 
                                pficAux );
                    tailFicRegles += (unsigned long)longTChar;
              }
              else
              {
                    TxtErreur( 164 , "ClasseRegles(alinsere)" );
                    _FinDialogueInfo();
                    _Free( tCharLemmes );
                    _Free( tCharRegles );
                    return _False;
              }

              ptRegleCou = ptRegleCou->RegleFrere;
        }

        if( p_avecRecomptage )
        {
             if( nbReglesLoc > 0 )
             {
                 lemmeCou.NbRegles = nbReglesLoc;
             }
             else if( nbReglesLoc == 0 && lemmeCou.NbRegles > 0 )
             {
                 lemmeCou.NbRegles = _Nil;
             }
        }
        else
        {
            if( lemmeCou.NbRegles > 0 && lemmeCou.NbRegles != nbReglesLoc )
            {
               sprintf( ChAux , 
                        "ClasseRegles(alinsere) attendues %d , trouvees %d\n" ,
                       lemmeCou.NbRegles , nbReglesLoc );
               TxtErreur( 164 , ChAux );
              _FinDialogueInfo();
              _Free( tCharLemmes );
              _Free( tCharRegles );
              return _False;
            }
        }

        /* insertion du lemme et reecriture par dessus le lemme initial */
        /* pour modifier son pteur de regle 1 et eventt son nombre de regles */
        longTChar = DestructurerLemme( lemmeCou.ValTerme.Ident , &lemmeCou ,
                                       tCharLemmes , chaineAnnexe );
        if( (unsigned long)longTChar != indAbsLemmeCou - indAbsLemmeIni )
        {
             TxtErreur( 164 , "ClasseRegles(alinsere)" );
             _Free( tCharLemmes );
             _Free( tCharRegles );
             _FinDialogueInfo();
             return _False;
        }
        LireEcrire( indAbsLemmeIni , longTChar , tCharLemmes , _Ecrire , 
                     FicBDicLemmes );

        DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
        DemarqueEtRendListeRegle( &ptRegleAux );
    }

    fclose( pficAux );
    fclose( FicBDicRegles );
    if( remove( nomFicDicRegles ) 
        || rename(  NomFicAux , nomFicDicRegles ) )
    {
         TxtErreur( 53 , "ClasseRegles(alinsere)" );
         _Free( tCharLemmes );
         _Free( tCharRegles );
         _FinDialogueInfo();
         return _False;
    }
    FicBDicRegles = fopen( nomFicDicRegles , _FopenRPB );
    if( !FicBDicRegles )
    {
        sprintf( ChAux , "ClasseRegles(alinsere) [%s]" , nomFicDicRegles );
        TxtErreur( 147 , ChAux );
        _Free( tCharLemmes );
        _Free( tCharRegles );
        _FinDialogueInfo();
        return _False;
    }

    fprintf( FicLog , "%ld rule(s) reordered\n\n" , nbReglesRelevees );
    _Free( tCharLemmes );
    _Free( tCharRegles );
    _FinDialogueInfo();
    return _True;
}

/*************************************************************************
        rajout d'une regle par chainage sur un lemme d'ancrage
*************************************************************************/
char     ModifieChainageLemmeEtReglesAncrage(  long  p_refLemmeAncrage ,
                long p_indAbsRegleAjoutee , TabLong p_posLemmeFic , 
                TabChar p_tCharLemmes , TabChar p_tCharRegles )
{
    long                            indAbsLemmeAncrageIni;
    long                            indAbsLemmeAncrageCou;
    long                            indAbsRegleAncrageCou;
    long                            indAbsRegleAncrageIni;
    ChIdent                         chaineAnnexeAncrage;
    struct LemmeMem                 lemmeAncrage;
    TabRegleMem                     ptRegleAncrageCou;

    long                            longTChar;

    /* chargement du lemme d'ancrage */
    indAbsLemmeAncrageCou = *(p_posLemmeFic + p_refLemmeAncrage);
    if( indAbsLemmeAncrageCou == _Nil )
    {
         sprintf( ChAux , 
                  "ModifieChainageLemmeEtReglesAncrage(alinsere) [%ld]" , 
                  indAbsLemmeAncrageCou );
         TxtErreur( 141 , ChAux );
         return _False;
    }

    indAbsLemmeAncrageIni = indAbsLemmeAncrageCou;
    if( indAbsLemmeAncrageIni == _Nil
        || RestructurerBDicLemme( (unsigned long *)&indAbsLemmeAncrageCou , 
                                  lemmeAncrage.ValTerme.Ident  ,
                                  &lemmeAncrage , chaineAnnexeAncrage ) 
                              == _Nil 
        || p_refLemmeAncrage != lemmeAncrage.ValTerme.Ref )
    {
         sprintf( ChAux , 
                  "ModifieChainageLemmeEtReglesAncrage(alinsere) [%ld]" , 
                  indAbsLemmeAncrageIni );
         TxtErreur( 141 , ChAux );
         return _False;
    }

    indAbsRegleAncrageCou = lemmeAncrage.ValTerme.PosFicRegle;
    /* le lemmme d'ancrage n'a pas encore de regle attachee */
    if( indAbsRegleAncrageCou == _Nil )
    {
        /* ecriture du lemme d'ancrage dont le pointeur de regle est modifie */
        lemmeAncrage.ValTerme.PosFicRegle = p_indAbsRegleAjoutee;
        longTChar = DestructurerLemme( lemmeAncrage.ValTerme.Ident , 
                                      &lemmeAncrage ,
                                       p_tCharLemmes , chaineAnnexeAncrage );
        if( longTChar 
             != indAbsLemmeAncrageCou - indAbsLemmeAncrageIni )
        {
             TxtErreur( 164 , 
                        "ModifieChainageLemmeEtReglesAncrage(alinsere)" );
             return _False;
        }
        LireEcrire( indAbsLemmeAncrageIni , longTChar , p_tCharLemmes , 
                    _Ecrire , FicBDicLemmes );
    }
    /* le lemmme d'ancrage n'a pas encore de regle attachee */
    else
    /* le lemmme d'ancrage a deja au moins une regle attachee */
    /* on ajoute la nouvelle en queue afin de ne pas perturber */
    /* la lexicalisation de ses regles si elle n'a pas encore ete faite */
    {
        while( _True )
        {
            indAbsRegleAncrageIni = indAbsRegleAncrageCou;

            if( RestructurerUneRegle( (unsigned long *)&indAbsRegleAncrageCou ,
                                       &ptRegleAncrageCou ) == _Nil )
            {
                DemarqueEtRendRegle( &ptRegleAncrageCou );
                TxtErreur( 82 , 
                           "ModifieChainageLemmeEtReglesAncrage(alinsere)" );
                return _False;
            }
           
            if( ptRegleAncrageCou->PosFicRegleSuivante == _Nil )
            {
                break;
            }
            indAbsRegleAncrageCou = ptRegleAncrageCou->PosFicRegleSuivante;
            DemarqueEtRendRegle( &ptRegleAncrageCou );
        }

        /* modification et insertion de la regle */
        ptRegleAncrageCou->PosFicRegleSuivante = p_indAbsRegleAjoutee;
        longTChar = DestructurerUneRegle( ptRegleAncrageCou , p_tCharRegles );
        if( longTChar 
            != indAbsRegleAncrageCou - indAbsRegleAncrageIni )
        {
             TxtErreur( 82 , "ModifieChainageLemmeEtReglesAncrage(alinsere)" );
             return _False;
        }

        LireEcrire( indAbsRegleAncrageIni , longTChar , p_tCharRegles , 
                     _Ecrire , FicBDicRegles );
        DemarqueEtRendRegle( &ptRegleAncrageCou );
    }

    DemarqueEtRendDag( &(lemmeAncrage.ValTerme.ArcsFils) );
    return _True;
}

/*************************************************************************
           rattachement des regles a d'autres ancrages lexicaux
              (suite a un rattachement optimise, par exemple)
*************************************************************************/
char     RattacheRegles( char *p_nomFicEntree )
{
    char                            nomFicDicRegles[ 256 ];
    char                            nomFicDicLemmes[ 256 ];
    char                            nomFicDicLiens[ 256 ];

    long                            longTChar;
    long                            nbReglesRelevees;

    TabChar                         tCharLemmes;
    TabChar                         tCharRegles;
    
    TabLong                         posLemmeFic;
    unsigned long                   longFicLemme;
    long                            tailTabPosLemmes;
    
    unsigned long                   indAbsLemmeIni;
    unsigned long                   indAbsLemmeCou;
    struct LemmeMem                 lemmeCou;
    ChIdent                         chaineAnnexe;
    long                            refLemmeCou;

    long                            refLemmeAncrage;

    int                             nbReglesLoc;
    unsigned long                   indAbsRegleSui;
    unsigned long                   indAbsRegleCou;
    unsigned long                   indAbsRegleIni;
    TabRegleMem                     ptRegleCou;
    TabRegleMem                     ptRegleIni;
    TabRegleMem                     ptReglePrec;

    FILE                           *ptFicAncrages;
    char                            car;

    /* sauvegarde des fichiers existants 
    if( !SauveArchivesDictionnaire() )
    {
       TxtErreur( 29 , "RattacheRegles(alinsere)" );
       return _False;
    }
    */

    strcpy( nomFicDicRegles , NomFicDico );
    strcat( nomFicDicRegles , ".RUL" );
    strcpy( nomFicDicLemmes , NomFicDico );
    strcat( nomFicDicLemmes , ".LEM" );
    strcpy( nomFicDicLiens , NomFicDico );
    strcat( nomFicDicLiens , ".LIN" );
    ptFicAncrages = fopen( p_nomFicEntree , _FopenRT );
    if( !ptFicAncrages || !FicBDicRegles || !FicBDicLemmes || !FicBDicLiens )
    {
        sprintf( ChAux , "RattacheRegles(alinsere) [%s, %s ou %s]" , 
           p_nomFicEntree , nomFicDicRegles , 
           nomFicDicLemmes );
        TxtErreur( 147 , ChAux );
        return _False;
    }
    
    tCharLemmes = AlloueBuffer( TailleBuffLemmesCourt );
    tCharRegles = AlloueBuffer( TailleBuffReglesCourt );

    /* attention aux mots a racine multiple */
    tailTabPosLemmes = NbLemmesDictionnaire() + 2000L;
    /* LiensTermesFichier contiendra les liens entre les identificateurs
      des lemmes et leur position absolue dans le fichier des lemmes */
    posLemmeFic = (TabLong) _Alloc( (unsigned long)sizeof( long )
                                     * (unsigned long)tailTabPosLemmes );
    if( !tCharLemmes || !tCharRegles || !posLemmeFic )
    {
        TxtErreur( 81 ,"RattacheRegles(alinsere)" );
        return _False;
    }
    
    if( !MapLitPositionsLemmes( posLemmeFic , &longFicLemme , 
                                tailTabPosLemmes ) )
    {
        _Free( posLemmeFic );
        _Free( tCharLemmes );
        _Free( tCharRegles );
        return _False;
    }

    _NouveauDialogueInfo( 128 , _True );
    TxtTitresDialogueInfo( 36 , 37 , 0 );

    car = ' ';
    SautBlanc( ptFicAncrages , &car );
    nbReglesRelevees = 0L;

    /* regles fichier */
    indAbsLemmeCou = (long)( _TailleAmorce + _TailleSeparation );
    while( indAbsLemmeCou < longFicLemme )
    {
        /* on extrait le lemme courant */
        indAbsLemmeIni = indAbsLemmeCou;
        if( RestructurerBDicLemme( &indAbsLemmeCou , lemmeCou.ValTerme.Ident  ,
                                   &lemmeCou , chaineAnnexe ) == _Nil 
            || EgaliteChaine( lemmeCou.ValTerme.Ident , DernLemme ) )
        {
             DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
             break;
        }
        refLemmeCou = lemmeCou.ValTerme.Ref;
        
        /* lemme avec regles */
        if( lemmeCou.NbRegles > 0 )
        {
            /* on extrait toutes les regles du lemme courant (y
               compris celles qui on deja ete rajoutees par
               lexicalisation et qui ne serviront a rien */
            indAbsRegleCou = lemmeCou.ValTerme.PosFicRegle;
            indAbsRegleIni = indAbsRegleCou;
            if( RestructurerLesRegles( &indAbsRegleCou , &ptRegleCou , 
                                       lemmeCou.NbRegles , _False ) == _Nil )
            {
                DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
                DemarqueEtRendListeRegle( &ptRegleCou );
                _FinDialogueInfo();
                _Free( posLemmeFic );
                _Free( tCharLemmes );
                _Free( tCharRegles );
                return _False;
            }

            /* on ne lexicalise que les regles du lemme courant qui
               etaient presentes sur ce lemme avant lexicalisation
               (donc les lemmeCou.NbRegles premieres, puisque les
               regles ajoutees par lexicalisation le sont en queue) */
            nbReglesLoc = 0;
            ptReglePrec = NULL;
            ptRegleIni = ptRegleCou;
            indAbsRegleCou = indAbsRegleIni;
            ptRegleCou = ptRegleIni;
            while( ptRegleCou && nbReglesLoc < lemmeCou.NbRegles )
            {
                  indAbsRegleSui = ptRegleCou->PosFicRegleSuivante;

                  if( nbReglesRelevees++ % 100L == 0 )
                  {
                      sprintf( ChAux , "%ld" , nbReglesRelevees );
                      TxtEcritItemInfo( 1 , ChAux );
                  }
                  nbReglesLoc++;

                  if( car == EOF )
                  {
                       TxtErreur( 140 , "RattacheRegles(alinsere)" );
                       _FinDialogueInfo();
                       _Free( posLemmeFic );
                       _Free( tCharLemmes );
                       _Free( tCharRegles );
                       return _False;
                  }
                  else
                  {
                      refLemmeAncrage = LireEntier( ptFicAncrages , &car );
                  }

                  /* cas ou le lemme d'ancrage n'est pas le lemme courant */
                  if( refLemmeAncrage > _PremRefLemme 
                         && refLemmeAncrage != refLemmeCou )
                  {
                      if( !ModifieChainageLemmeEtReglesAncrage( 
                                          refLemmeAncrage , 
                                          indAbsRegleCou , posLemmeFic , 
                                          tCharLemmes , tCharRegles ) )
                      {
                             _FinDialogueInfo();
                             _Free( posLemmeFic );
                             _Free( tCharLemmes );
                             _Free( tCharRegles );
                             return _False;
                      }

                      /* changement du chainage des regles du lemme courant */
                      /* cas de suppression d'une regle de tete */
                      if( !ptReglePrec )
                      {
                          lemmeCou.ValTerme.PosFicRegle = indAbsRegleSui;
                      }
                      else
                      /* cas de suppression d'une regle de rang >= 1 */
                      {
                          ptReglePrec->PosFicRegleSuivante = indAbsRegleSui;
                      }
                      ptRegleCou->PosFicRegleSuivante = _Nil;
                      
                      /* sauvegard de la regle avec ce nouveau
                         chainage en cas de 2 ancrages successifs sur
                         le meme lemme */
                      longTChar = DestructurerUneRegle( ptRegleCou , 
                                                        tCharRegles );
                      LireEcrire( indAbsRegleCou , longTChar , tCharRegles , 
                                  _Ecrire , FicBDicRegles );
                  }
                  /* cas ou le lemme d'ancrage n'est pas le lemme courant */
                  else
                  {
                       ptReglePrec = ptRegleCou;
                  }

                  indAbsRegleCou = indAbsRegleSui;
                  ptRegleCou = ptRegleCou->RegleFrere;
            }

            if( lemmeCou.NbRegles > 0 && nbReglesLoc < lemmeCou.NbRegles )
            {
                   sprintf( ChAux , 
                      "RattacheRegles(alinsere) attendues %d , trouvees %d\n" ,
                           lemmeCou.NbRegles , nbReglesLoc );
                   TxtErreur( 164 , ChAux );
                  _FinDialogueInfo();
                  _Free( posLemmeFic );
                  _Free( tCharLemmes );
                  _Free( tCharRegles );
                  return _False;
            }

            /* ecriture du lemme courant dont le pointeur de regle est */
            /* eventuellement modifie */
            longTChar = DestructurerLemme( lemmeCou.ValTerme.Ident , 
                                           &lemmeCou ,
                                           tCharLemmes , chaineAnnexe );
            if( (unsigned long)longTChar != indAbsLemmeCou - indAbsLemmeIni )
            {
                 TxtErreur( 164 , "RattacheRegles(alinsere)" );
                 _FinDialogueInfo();
                 _Free( posLemmeFic );
                 _Free( tCharLemmes );
                 _Free( tCharRegles );
                 return _False;
            }
            LireEcrire( indAbsLemmeIni , longTChar , tCharLemmes , _Ecrire , 
                         FicBDicLemmes );

            /* ecriture des lemmeCou.NbRegles premieres regles du
               lemme qui ont eventuellement eu leur chainage modifie */
            ptRegleCou = ptRegleIni;
            indAbsRegleCou = indAbsRegleIni;
            for( nbReglesLoc = 0 ; 
                 nbReglesLoc < lemmeCou.NbRegles ; nbReglesLoc++ )
            {
                  /* reecriture de la regle sur elle-meme avec son
                     nouveau chainage eventuel */
                  longTChar = DestructurerUneRegle( ptRegleCou , tCharRegles );
                  if( longTChar > 0L && longTChar < TailleBuffReglesCourt )
                  {
                        LireEcrire( indAbsRegleCou , longTChar , 
                                    tCharRegles , _Ecrire , 
                                    FicBDicRegles );
                        indAbsRegleCou += (unsigned long)longTChar;
                  }
                  else
                  {
                        TxtErreur( 164 , "RattacheRegles(alinsere)" );
                        _FinDialogueInfo();
                        _Free( posLemmeFic );
                        _Free( tCharLemmes );
                        _Free( tCharRegles );
                        return _False;
                  }

                  ptRegleCou = ptRegleCou->RegleFrere;
            }
            DemarqueEtRendListeRegle( &ptRegleIni );
        }
        /* lemme avec regles */

        DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
    }

    fprintf( FicLog , "\n%ld regles lexicalisees\n\n" , nbReglesRelevees );
    fclose( ptFicAncrages );
    _Free( posLemmeFic );
    _Free( tCharLemmes );
    _Free( tCharRegles );
    _FinDialogueInfo();

    return ClasseRegles( _True );
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            fin de alinsere.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
