/*
 * FASt Term Recognizer
 *
 *  FASTR/almappin.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for mapping a function on the lexical data
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*************************************************************************
      references des termes simples
*************************************************************************/
void     MapReferencesTermesSimples( int p_avecRegles , 
                                     int p_pourLexicalisation )
{

    unsigned long             indAbsCou;
    unsigned long             indAbsRegleCou;
    unsigned long             longFic;

    struct LemmeMem           lemmeCou;
    TabRegleMem               ptRegleCou;
     TabRegleMem               ptRegleAux;
    ChIdent                   chaineAnnexe;
    char                      nomFicIdent[ 256 ];
    long                      nbMots = 0L;
   
    TabCorrIdent_PosFic       tCorrespondance;
    ChIdent                   ident;
    int                       nbReglesLoc;

    _NouveauDialogueInfo( 128 , _True );
    TxtTitresDialogueInfo( 22 , 23 , 0 );
    tCorrespondance = NULL;
    strcpy( nomFicIdent , NomFicDico );
    strcat( nomFicIdent , ".REF" );

    if( FicRefMots ) {
        fclose( FicRefMots );
      }

    if( !OuvreFichier( nomFicIdent , &FicRefMots , _FopenWPT ) )
    {
        return;
    }

    if( p_avecRegles )
    {
        if( !AlloueEtChargeCorrespondances( FicRefIdent , &tCorrespondance ) )
        {
             return;
        }
    }

    indAbsCou = (long)( _TailleAmorce + _TailleSeparation );
    fseek( FicBDicLemmes , 0L , 2 );
    fgetpos( FicBDicLemmes , &longFic );
    *(lemmeCou.ValTerme.Ident) = '\0';
    if( !p_avecRegles )
    {
          fseek( FicRefMots , 0L , 0 );
          fprintf( FicRefMots , "%-10ld\n" , 2L );
    }

    /* parcours de tous les mots du lexique */
    while( indAbsCou < longFic   
           && !EgaliteChaine( lemmeCou.ValTerme.Ident , DernLemme ))
    {
        if( RestructurerBDicLemme( &indAbsCou , lemmeCou.ValTerme.Ident  ,
                                   &lemmeCou , chaineAnnexe ) == _Nil )
        {
              DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
              break;
        }

        if( p_avecRegles )
        {
            fprintf( FicRefMots ,  "%-32s%-10ld%-3d\n" , 
                                 lemmeCou.ValTerme.Ident , 
                                 lemmeCou.ValTerme.Ref , lemmeCou.NbRegles );

            indAbsRegleCou = lemmeCou.ValTerme.PosFicRegle;
            if( RestructurerLesRegles( &indAbsRegleCou , &ptRegleCou , 
                                       lemmeCou.NbRegles , _True ) == _Nil )
            {
                DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
                DemarqueEtRendListeRegle( &ptRegleCou );
                break;
            }

            nbReglesLoc = 0;
            ptRegleAux = ptRegleCou;
            while( ptRegleCou )
            {
                  nbReglesLoc++;
                  if( p_pourLexicalisation )
                  {
                      fprintf( FicRefMots ,  "    %ld\t" ,
                               ptRegleCou->TermeFils->ValTerme.Ref );
                      ParcoursDFArbre( &((ptRegleCou)->TermeFils) , 
                                       EcrireRefTermeMotMem , _PreOrdre );
                      fprintf( FicRefMots , "#\n" );
                  }
                  else
                  {
                      if( Identificateur( ident , 
                                      ptRegleCou->TermeFils->ValTerme.Ref , 
                                      FicRefIdent , tCorrespondance ) )
                      {
                          fprintf( FicRefMots ,  "    [%d]<%s>\n" , 
                                   nbReglesLoc , ident );
                      }
                      else
                      {
                          fprintf( FicRefMots ,  "    [%d]<%ld>\n" , 
                                   nbReglesLoc , 
                                   ptRegleCou->TermeFils->ValTerme.Ref );
                      }
                  }
                  ptRegleCou = ptRegleCou->RegleFrere;
            }
            if( lemmeCou.NbRegles > 0 && lemmeCou.NbRegles != nbReglesLoc )
            {
                   sprintf( ChAux , 
         "MapReferencesTermesSimples(almappin) attendues %d , trouvees %d\n" ,
                           lemmeCou.NbRegles , nbReglesLoc );
                   TxtErreur( 164 , ChAux );
                  _FinDialogueInfo();
                  return;
            }
            DemarqueEtRendListeRegle( &ptRegleAux );
        }
        else
        {
            fprintf( FicRefMots ,  "%-31s,%-10ld\n" , 
                     lemmeCou.ValTerme.Ident , lemmeCou.ValTerme.Ref );
        }

        DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );

        nbMots++;
        if( nbMots % 100L == 0 )
        {
           sprintf( ChAux , "%s" , lemmeCou.ValTerme.Ident );
           TxtEcritItemInfo( 1 , ChAux );
        }

        if( _ClickDansAnnulerInfo() )
        {
             _FinDialogueInfo();
             return;
        }
    }

    if( !p_avecRegles )
    {
          fseek( FicRefMots , 0L , 0 );
          fprintf( FicRefMots , "%-10ld\n" , nbMots );
    }
    else
    {
        if( tCorrespondance )
        {
            _Free( tCorrespondance );
        }
    }
    _FinDialogueInfo();
}

/*************************************************************************
 charge les lien entre les lemmes et leurs positions dans le fichler .LEM
*************************************************************************/
char     MapLitPositionsLemmes(  TabLong p_posLemmeFic , 
                                 unsigned long *p_longFicLemme ,
                                 long p_tailTabPosLemmes )
{
    unsigned long              indAbsLemmeCou;
    unsigned long              indAbsLemmeIni;
    struct LemmeMem            lemmeCou;
    ChIdent                    chaineAnnexe;
    ChIdent                    chAux;
    long                       refLemmeCou;

    _NouveauDialogueInfo( 128 , _True );
    TxtTitresDialogueInfo( 36 , 12 , 0 );
    
    for( refLemmeCou = 0L ; refLemmeCou < p_tailTabPosLemmes ; refLemmeCou++ )
    {
        *(p_posLemmeFic + refLemmeCou) = _Nil;
    }

    indAbsLemmeCou = (long)( _TailleAmorce + _TailleSeparation );

    fseek( FicBDicLemmes , 0L , 2 );
    fgetpos( FicBDicLemmes , p_longFicLemme );
    while( indAbsLemmeCou < *p_longFicLemme  
           && !EgaliteChaine( lemmeCou.ValTerme.Ident , DernLemme ))
    {
           indAbsLemmeIni = indAbsLemmeCou;
           if( RestructurerBDicLemme( &indAbsLemmeCou , 
                                      lemmeCou.ValTerme.Ident  ,
                                      &lemmeCou , chaineAnnexe ) == _Nil )
           {
                   DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
                   break;
           }
           DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );

           refLemmeCou = lemmeCou.ValTerme.Ref;
           if( refLemmeCou >= p_tailTabPosLemmes )
           {
                sprintf( ChAux ,
"MapLitPositionsLemmes(almappin) Ref maxiamle attendue %ld ref courant %ld\n" ,
                        p_tailTabPosLemmes , refLemmeCou );
                TxtErreur( 16 , ChAux );
                _FinDialogueInfo();
                return _False;
           }

           /* les racines annexes ne sont pas le lieu de lexicalisation */
           if( lemmeCou.NbRegles > _RacineAnnexe1 
                || lemmeCou.NbRegles < _RacineAnnexeMax )
           {
                 *(p_posLemmeFic + refLemmeCou) = indAbsLemmeIni;
           }

           if( (refLemmeCou % 1000L) == 0 )
           {
              strcpy( chAux , lemmeCou.ValTerme.Ident );
              TxtEcritItemInfo( 1 , chAux );
           }
           if( _ClickDansAnnulerInfo() )
           {
                _FinDialogueInfo();
                return _False;
           }
    }
    _FinDialogueInfo();
    return _True;
}

/************************************************************************
    PRODUCTION DE TOUTES LES FLEXIONS D'UN MOT
*************************************************************************/
char     MapAfficheLexiques( char *p_listeCatNoFlexion[] , int p_nbCats ,
                             int p_estFormeFlechie , int p_avecDerives ,
                             int p_avecLiensSemantiques , int p_avecRegles )
{
    unsigned long            indAbsCou;
    unsigned long            longFic;
    struct LemmeMem          lemmeCou;
    TabLemmeMem              ptLemmeProduit;
    char                     racineSuffixee[ _MaxLetIdent ];
    char                     chaineNormalisee[ 2 * _MaxLetIdent ];

    /* indice parcours des flexions ou des deriv. d'1 Lemme */
    int                      indFlexion;        
    int                      indDerivation;        

    /* no de flexion ou de derivation determinant le choix des
       suffixes a ajouter */
    int                      noTypeFlexion;
    int                      noTypeDerivation;

    /* nombre de flexions de la cat lex du mot courant */
    int                      nbIndFlexions;
    int                      nbIndDerivations;

    int                      catLex;

    /* description du terme releve dans le dico */
    char                     chaineAnnexe[_MaxLetIdent];

    long                     nbMots = 0L;

    /* tableaux de categories et de l'indice de la flexion souhaitee */
    int                     *categories = NULL;
    int                     *indFlexionNormale = NULL;

    _NouveauDialogueInfo( 128 , _True );
    TxtTitresDialogueInfo( 22 , 23 , 0 );

    if( p_estFormeFlechie && p_nbCats > 0 )
    {
        int      ind;
        int      indReel;

        categories = (int *)_Alloc( p_nbCats * sizeof( int ) );
        indFlexionNormale = (int *)_Alloc( p_nbCats * sizeof( int ) );
        if( !categories || !indFlexionNormale )
        {
            TxtErreur( 19 , "MapAfficheLexiques(aldeflex)" );
             exit(0);
        }

        indReel = 0;
        for( ind = 0 ; ind < p_nbCats ; ind++ )
        {
            if( IdentATraitsCatLex( p_listeCatNoFlexion[ 2 * ind ] , 
                                    _CatLex , categories + indReel ) )
            {
                indFlexionNormale[ indReel ] = 
                      atoi( p_listeCatNoFlexion[ 2 * ind + 1 ] );
                if( indFlexionNormale[ indReel ] 
                    > DicoTraitsCatLex[ categories[ indReel ] ].NbFlexions )
                {
                    TxtErreur( 248 , "MapAfficheLexiques(aldeflex)" );
                    exit(0);
                }
                indReel++;
            }
            else
            {
                TxtErreur( 247 , "MapAfficheLexiques(aldeflex)" );
                exit(0);
            }
        }
        p_nbCats = indReel;
    }

    /* parcours sequentiel du lexique pour generer la flexion
       preferentielle de chaque mot */
    indAbsCou = (long)( _TailleAmorce + _TailleSeparation );
    fseek( FicBDicLemmes , 0L , 2 );
    fgetpos( FicBDicLemmes , &longFic );
    *(lemmeCou.ValTerme.Ident) = '\0';

    /* saut du premier lemme bidon */
    if( RestructurerBDicLemme( &indAbsCou , lemmeCou.ValTerme.Ident  ,
                               &lemmeCou , chaineAnnexe ) == _Nil )
    {
          DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
          if( categories ){ _Free( categories ); }
          if( indFlexionNormale ){ _Free( indFlexionNormale ); }
          return _True;
    }

    /* parcours de tous les mots du lexique */
    while( indAbsCou < longFic )
    {
        if( RestructurerBDicLemme( &indAbsCou , lemmeCou.ValTerme.Ident  ,
                                   &lemmeCou , chaineAnnexe ) == _Nil 
            || EgaliteChaine( lemmeCou.ValTerme.Ident , DernLemme ) )
        {
              DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
              break;
        }
        lemmeCou.LemmeFrere = NULL;

        noTypeFlexion 
              = NoTypeSuffixage( &(lemmeCou.ValTerme.ArcsFils) , 
                                 _True );
        catLex = CategorieLexicale( &(lemmeCou.ValTerme.ArcsFils) );

        /* mot ayant des flexions on note le nb de flexions de
           sa categorie */
        if( catLex == _Nil 
            || noTypeFlexion <= 0
            || noTypeFlexion
               > DicoTraitsCatLex[ catLex ].NbNoTypeFlexion )
        {
            noTypeFlexion = _Nil;
            nbIndFlexions = _Nil;
        }
        else
        {
            nbIndFlexions 
                = DicoTraitsCatLex[ catLex ].NbFlexions;
        }

        /* Les derivations ne sont pas notees, elles sont insignifiantes */
        indDerivation = _Nil;
        noTypeDerivation = _Nil;
        nbIndDerivations = _Nil;

        /* recherche d'une eventuelle specification de la flexion */
        indFlexion = 0;
        if( p_nbCats > 0 && catLex != _Nil  )
        {
            int      ind;
    
            for( ind = 0 ; ind < p_nbCats ; ind++ )
            {
                if( categories[ ind ] == catLex )
                {
                    indFlexion = indFlexionNormale[ ind ] - 1;
                    break;
                }
            }
        }

        /* production de la suffixation choisie */
        if( p_estFormeFlechie
            && ProduireSuffixages( lemmeCou.ValTerme.Ident , 
                    chaineAnnexe , catLex , lemmeCou.NbRegles ,
                    noTypeFlexion , noTypeDerivation , 
                    nbIndFlexions , nbIndDerivations , 
                    &indFlexion , &indDerivation , racineSuffixee ,
                    _True , _False ) != _False )
        {
           if( !AlloueEtGarnitLemme( &lemmeCou , catLex , 
                                     indFlexion , indDerivation , 
                                     &ptLemmeProduit , (int)'N' , 
                                     lemmeCou.ValTerme.Ident ) )
           {
              DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
              if( categories ){ _Free( categories ); }
              if( indFlexionNormale ){ _Free( indFlexionNormale ); }
              return _False;
           }

           NormaliseChaine( racineSuffixee , chaineNormalisee ,
                            (int)'\'' );
           fprintf( stdout , "\n\'%s\'" , chaineNormalisee );
           EcrireLemmeMem( &ptLemmeProduit , stdout , _Sortie , 0 );
           DemarqueEtRendLemme( &ptLemmeProduit );
        }
        /* production de la suffixation choisie */
        else
        /* pas de flexion: on ecrit le lemme */
        {
           TabLemmeMem               ptLemmeAux = &lemmeCou;

           NormaliseChaine( ChaineLemme( &(lemmeCou.ValTerme.ArcsFils) ) , 
                            chaineNormalisee , (int)'\'' );
           fprintf( stdout , "\n\'%s\'" , chaineNormalisee );
           EcrireLemmeMem( &ptLemmeAux , stdout , _Sortie , 0 );
        }
        /* pas de flexion: on ecrit le lemme */

        /* avec les derives et racines */
        if(  p_avecDerives )
        {
           TabLemmeMem                 ptLemmeTeteFamDeriv = NULL;

           if( LemmeADerives( &lemmeCou , &ptLemmeTeteFamDeriv ) )
           {
               EcrireListeLemme( stdout , _Sortie , ptLemmeTeteFamDeriv );
               DemarqueEtRendListeLemme( &ptLemmeTeteFamDeriv );
           }
           else
           {
               return _False;
           }
        }

        /* avec les derives et racines */
        if(  p_avecLiensSemantiques )
        {
           TabLemmeMem                 ptLemmeTeteFamSemantique = NULL;
           TabLemmeMem                 ptLemmeTeteFamilleSemantiqueRec = NULL;

           if( LemmeAFamilleSem( &lemmeCou , NULL , &ptLemmeTeteFamSemantique ,
				 _False ) ) {
               EcrireListeLemme( stdout , _Sortie , ptLemmeTeteFamSemantique );
           }
           else
           {
               return _False;
           }
	   lemmeCou.LemmeFrere = ptLemmeTeteFamSemantique;
 
	   /* on ajoute les liens semantiques reciproques des lemmes auxquels
	      il est semantiquement lie et retrouves dans la boucle
	      precedente */
	   ptLemmeTeteFamilleSemantiqueRec = NULL;
	   if( LemmeAFamilleSem( &lemmeCou , NULL , 
				 &ptLemmeTeteFamilleSemantiqueRec ,
				  _True ) ) {
               EcrireListeLemme( stdout , _Sortie , 
				 ptLemmeTeteFamilleSemantiqueRec );
           }
           else
           {
               return _False;
           }
	   DemarqueEtRendListeLemme( &ptLemmeTeteFamSemantique );
	   DemarqueEtRendListeLemme( &ptLemmeTeteFamilleSemantiqueRec );
        }

        /* avec les regles */
        if( p_avecRegles )
        {
           TabRegleMem                 ptTeteListeRegles = NULL;

           AjouteReglesARetourner( &lemmeCou , &ptTeteListeRegles );
           EcrireListeRegleMem( ptTeteListeRegles , _False , 
                                stdout , _Sortie );
           DemarqueEtRendListeRegle( &ptTeteListeRegles );
        }

        DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );

        nbMots++;
        if( nbMots % 100L == 0 )
        {
           sprintf( ChAux , "%s" , lemmeCou.ValTerme.Ident );
           TxtEcritItemInfo( 1 , ChAux );
        }

        if( _ClickDansAnnulerInfo() )
        {
           _FinDialogueInfo();
           if( categories ){ _Free( categories ); }
           if( indFlexionNormale ){ _Free( indFlexionNormale ); }
           return _True;
        }
    }
    /* parcours de tous les mots du lexique */

    if( categories ){ _Free( categories ); }
    if( indFlexionNormale ){ _Free( indFlexionNormale ); }
    return _True;
}

/*************************************************************************
     passe tous les lemmes du fichiers et effectue des modifications
     au moyen d'une fonction p_fonction
*************************************************************************/
char     MapLitEtModifieLemmes( 
    char                        (*p_fonction)(
    TabLemmeMem                                  p_ptLemme,
    unsigned long                                p_posFicLemme,
    TabChar                                      p_tCharRegles
                                         ) )
{

    unsigned long             indAbsCou;
    unsigned long             indAbsIni;
    unsigned long             longFic;
    long                      longTChar;
    long                      nbLemmesRelevees;

    struct LemmeMem           lemmeCou;
    ChIdent                   chaineAnnexe;
    TabChar                   tCharLemmes = NULL;
    TabChar                   tCharRegles = NULL;

    tCharLemmes = AlloueBuffer( TailleBuffLemmesCourt );
    if( !tCharLemmes )
    {
       TxtErreur( 81 , "MapLitEtModifieLemmes(alinsere)" );
       _FinDialogueInfo();
       return _False;
    }

   /* le buffer de destructuration de la racine */
    tCharRegles = AlloueBuffer( TailleBuffReglesCourt );
    if ( !tCharRegles ) {
      TxtErreur( 81 ,"MapLitEtModifieLemmes(alinsere)" );
      return _False;
    }   

    _NouveauDialogueInfo( 128 , _True );
    TxtTitresDialogueInfo( 27 , 56 , 0 );

    indAbsCou = (long)( _TailleAmorce + _TailleSeparation );
    fseek( FicBDicLemmes , 0L , 2 );
    fgetpos( FicBDicLemmes , &longFic );
    *(lemmeCou.ValTerme.Ident) = '\0';
    nbLemmesRelevees = 0L;

    while( indAbsCou < longFic 
           && !EgaliteChaine( lemmeCou.ValTerme.Ident , DernLemme ) )
    {
        indAbsIni = indAbsCou;

        /* on charge le lemme courant */
        if( RestructurerBDicLemme( &indAbsCou , lemmeCou.ValTerme.Ident  ,
                                   &lemmeCou , chaineAnnexe ) == _Nil )
        {
           DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
           break;
        }

        /* travail sur tous les lemmes sauf le premier et le dernier */
        if( strcmp( lemmeCou.ValTerme.Ident , PremLemme ) != 0
            && strcmp( lemmeCou.ValTerme.Ident , DernLemme ) != 0 ) 
        {
           /* on applique la fonction courante */
           if( !p_fonction( &lemmeCou , indAbsIni , tCharRegles ) )
           {
           }

           /* insertion du lemme et reecriture par dessus le lemme
              initial */
           longTChar = DestructurerLemme( lemmeCou.ValTerme.Ident , 
                                          &lemmeCou ,
                                          tCharLemmes , chaineAnnexe );
           if( (unsigned long)longTChar != indAbsCou - indAbsIni ) {
	     printf("[%ld/%ld]\n"
		    ,(unsigned long)longTChar,indAbsCou - indAbsIni);
	     TxtErreur( 164 , "MapLitEtModifieLemmes(almappin)" );
	     return _False;
           }
           LireEcrire( indAbsIni , longTChar , tCharLemmes , _Ecrire , 
                       FicBDicLemmes );
        }

        DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
        nbLemmesRelevees++;
    }

    if( Trace ) 
    {
        fprintf( FicLog , "\n%ld lemma(s) scanned\n\n" , nbLemmesRelevees );
    }
    _Free( tCharLemmes );
    _Free( tCharRegles );
    _FinDialogueInfo();
    return _True;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            fin de almappin.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
