/*
 * FASt Term Recognizer
 *
 *  FASTR/almotrai.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for assigning features to inflected words
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*****************************************************************************
 passage d'un lemme a son No de flexion, sa categ lexicale ...
*****************************************************************************/
int      CategorieLexicale( TabDagMem *p_hdlDag )
{
     TabDagMem             *hdlDag;
     
     hdlDag = ExisteCheminFils( CheminCategorie , p_hdlDag );
     if( hdlDag && *hdlDag && (*hdlDag)->TypeDag == _AtomeRef
         && ((*hdlDag)->ValDag.Atome.AtomeConjDisj) != _ConjonctionNeg )
     {
          return (int)((*hdlDag)->ValDag.Atome.ValAtome.Ref);
     }
     else
     {
         return (int)_Nil;
     }
}

long     ReferenceLexicale( TabDagMem *p_hdlDag )
{
     TabDagMem             *hdlDag;
     
     hdlDag = ExisteCheminFils( CheminReference , p_hdlDag );
     if( hdlDag && *hdlDag && (*hdlDag)->TypeDag == _AtomeEntier
         && ((*hdlDag)->ValDag.Atome.AtomeConjDisj) != _ConjonctionNeg )
     {
          return (long)((*hdlDag)->ValDag.Atome.ValAtome.Entier);
     }
     else
     {
         return (long)_Nil;
     }
}

/************************************************************************
        RECUPERE TOUTES LES REFERENCES DES MOTS SEMANTIQUEMENT
	LIES A UN MOT DONNE
*************************************************************************/

int      ReferenceLexicaleFamilleSem( TabDagMem *p_hdlDag , 
				      TabLong p_referencesFamilleSem )
{
  TabDagMem                       *hdlDag;
  int                              indCheminSemantique;
  int                              indVoisinSemantique;

  indCheminSemantique = 0;
  indVoisinSemantique = 0;
  /* parcours de tous les liens semantiques possibles */
  while( CheminsSemantique[indCheminSemantique ] ) {
    hdlDag = ExisteCheminFils( CheminsSemantique[ indCheminSemantique ] , 
			       p_hdlDag );

    /* parcours de toutes les listes d'une disjonction */
    while( hdlDag && *hdlDag && (*hdlDag)->TypeDag == _AtomeListe
	   && ((*hdlDag)->ValDag.Atome.AtomeConjDisj) != _ConjonctionNeg 
	   && (*hdlDag)->Fils ) {
      p_referencesFamilleSem[ indVoisinSemantique ] =
	((*hdlDag)->Fils)->ValDag.Atome.ValAtome.Entier;

      /* incremente le nombre de references retournees */
      indVoisinSemantique++;
      if( indVoisinSemantique >= _NbMaxTypesLiensSem ) {
	TxtErreur( 306 , "ReferenceLexicaleFamilleSem(almotrai)" );
	return _False;
      }

      hdlDag = &((*hdlDag)->Frere);
    }

    indCheminSemantique++;
  }

  return indVoisinSemantique;
}

TabChar  ChaineLemme( TabDagMem *p_hdlDag )
{
     TabDagMem                  *hdlDag;
     TabDagMem                   ptDag;
     TabDagMem                  *hdlDagFilsCheminSelf;

     hdlDag = ExisteCheminFils( CheminLemme , p_hdlDag );
     if( hdlDag && *hdlDag && (*hdlDag)->TypeDag == _AtomeChaine
         && ((*hdlDag)->ValDag.Atome.AtomeConjDisj) != _ConjonctionNeg )
     {
          return (TabChar)((*hdlDag)->ValDag.Atome.ValAtome.Chaine);
     }
     else if( ( hdlDagFilsCheminSelf 
		= ExisteCheminFils( CheminSelf , p_hdlDag ) ) ) {
       if( hdlDagFilsCheminSelf && *hdlDagFilsCheminSelf 
	   && (*hdlDagFilsCheminSelf)->TypeDag == _AtomeListe
	   && (ptDag = (*hdlDagFilsCheminSelf)->Fils)
	   && ptDag->TypeDag == _AtomeEntier
	   && (ptDag = ptDag->Frere)
	   && ptDag->TypeDag == _AtomeChaine
	   && (ptDag->ValDag.Atome.AtomeConjDisj) != _ConjonctionNeg ) {
	 return (TabChar)(ptDag->ValDag.Atome.ValAtome.Chaine);
       }
     }
     return (TabChar)NULL;
}

int      NoTypeSuffixage( TabDagMem *p_hdlDag , int p_estFlexion )
{
     TabDagMem             *hdlDag;

     if( p_estFlexion )
     {
          hdlDag = ExisteCheminFils( CheminFlexion , p_hdlDag );
     }
     else
     {
          hdlDag = ExisteCheminFils( CheminDerivation , p_hdlDag );
     }
     if( hdlDag && *hdlDag && (*hdlDag)->TypeDag == _AtomeEntier )
     {
          return (int)((*hdlDag)->ValDag.Atome.ValAtome.Entier);
     }
     else
     {
         return (int)_Nil;
     }
}

void     NormaliseChaine( TabChar p_chaineIn , TabChar p_chaineOut , 
                          int p_separateur )
{
   int    ind;
   int    indAux;

   indAux = 0;

   for( ind = 0 ; ind < _MaxLetIdent - 1 
	  && indAux < _MaxLetIdent - 1 ; ind++ ) {
     if( !(p_chaineIn[ ind ]) ) {
         p_chaineOut[ indAux ] = '\0';
         break;
     }
     else if( p_chaineIn[ ind ] != (char)p_separateur ) {
         p_chaineOut[ indAux ] 
              = p_chaineIn[ ind ];
     }
     else {
         p_chaineOut[ indAux ] 
              = '\\';
         indAux++;
         p_chaineOut[ indAux ] 
              = (char)p_separateur;
     }
     indAux++;
   }
   p_chaineOut[ indAux ] = '\0';
}

/*****************************************************************************
   conception d'une fonction qui permet de passer d'une liste de valeurs
   decrivant un lien dans le lexique vers une autre unite lexicale 
   a une structure de traits contenant ces valeurs. 
   dans le cas ou la liste n'est pas complete, elle est completee
   pour avoir la forme standard (ref,lemme,cat,no flexion). l'ajout
   de la reference se faisant en tete, cela peut modifier la valeur
   de la liste.
   la ref absente est completee a -1, le no flexion absent est
   complete a 1.
   la fonctionrenvoit 1) la nouvelle valeur du pointeur vers
   la structure de traits, 2) les pointeurs vers les differents
   champs (reference,chaine,categorie et no flexion)
*****************************************************************************/

int      LienUniteLexicaleATraits( TabDagMem *p_hdlDagListe , 
				   TabDagMem *p_hdlDagStructTraits , 
				   TabLong *p_hdlReference , 
				   TabChar * p_hdlChaine ,
				   TabLong *p_hdlCategorie , 
				   TabLong *p_hdlNoFlexion ,
				   int p_avecStructureDeTraits )
{
  TabDagMem                 nouvelAtome;
  TabDagMem                 ptDag;
  struct ValAtomeMem        atome;

  ptDag = (*p_hdlDagListe)->Fils;

  /* 1 : la reference */
  if( !(ptDag) ) {
    TxtChaineErreur( 285 , ChErreurCompil );
    return _False;
  }
  else {
    /* pas de reference, on en ajoute une vide en tete de la liste */
    if( (ptDag)->TypeDag != _AtomeEntier ) {
      atome.AtomeConjDisj = _False; 
      atome.ValAtome.Entier = _Nil;	
      if( !(nouvelAtome = AjouteAtomeFils( p_hdlDagListe , _AtomeEntier , 
					   atome , _False ) ) ) {
	printf( "Erreur 1\n" );
	return _False;
      }
      *p_hdlReference = &(nouvelAtome->ValDag.Atome.ValAtome.Entier);
      /* ptDAg n'est pas modifie, il continue de pointer
	 vers la chaine de la racine, qui est desormais le
	 frere de la reference ajoutee a cette racine */
    }
    else {
      atome.AtomeConjDisj = _False; 
      atome.ValAtome.Entier = ptDag->ValDag.Atome.ValAtome.Entier;
      *p_hdlReference = &(ptDag->ValDag.Atome.ValAtome.Entier);
      ptDag = ptDag->Frere;
    }

    /* ajout de la reference de la racine */
    if( p_avecStructureDeTraits ) {
      if( !AjouteCheminEtAtomeFils( p_hdlDagStructTraits ,
				    CheminReference , _AtomeEntier , 
				    atome , _True ) ) {
	printf( "Erreur 2\n" );
	return _False;
      }
    }
  }

  /* 2 : le lemme */
  if( !(ptDag) || (ptDag)->TypeDag != _AtomeChaine ) {
    TxtChaineErreur( 253 , ChErreurCompil );
    printf( "Erreur 3\n" );
    return _False;
  }      
  else {
    /* ajout du lemme de la racine */
    atome.AtomeConjDisj = _False; 
    strcpy( atome.ValAtome.Chaine , ptDag->ValDag.Atome.ValAtome.Chaine );
   *p_hdlChaine = ptDag->ValDag.Atome.ValAtome.Chaine;
    if( p_avecStructureDeTraits ) {
      if( !AjouteCheminEtAtomeFils( p_hdlDagStructTraits ,
				   CheminLemme , _AtomeChaine , 
				   atome , _True ) ) {
	printf( "Erreur 4\n" );
	return _False;
      }
    }
  }

  /* 3 : la categorie */
  ptDag = ptDag->Frere;
  if( !(ptDag) || (ptDag)->TypeDag != _AtomeRef ) {
    TxtChaineErreur( 286 , ChErreurCompil );
    printf( "Erreur 5\n" );
    return _False;
  }      
  else {
    /* ajout de la categorie de la racine */
    atome.AtomeConjDisj = _False; 
    atome.ValAtome.Ref = ptDag->ValDag.Atome.ValAtome.Ref;
    *p_hdlCategorie = &(ptDag->ValDag.Atome.ValAtome.Ref);
    if( p_avecStructureDeTraits ) {
      if( !AjouteCheminEtAtomeFils( p_hdlDagStructTraits ,
				    CheminCategorie , _AtomeRef , 
				    atome , _True ) ) {
	printf( "Erreur 6\n" );
	return _False;
      }
    }
  }

  /* 4 : le numero de flexion */
  ptDag = ptDag->Frere;
  /* si pas de numero de flexion, un est ajoute en queue de
       liste avec la valeur par defaut 1 */
  if( !(ptDag) ) {
    atome.AtomeConjDisj = _False; 
    atome.ValAtome.Entier = _Nil;	
    if( !(nouvelAtome = AjouteAtomeFils( 
		     &((*p_hdlDagListe)->Fils) , _AtomeEntier , 
		     atome , _True ) ) ) {
      printf( "Erreur 7\n" );
      return _False;
    }
    *p_hdlNoFlexion = &(nouvelAtome->ValDag.Atome.ValAtome.Entier);
  }
  else {
    atome.AtomeConjDisj = _False; 
    atome.ValAtome.Entier = ptDag->ValDag.Atome.ValAtome.Entier;
    *p_hdlNoFlexion = &(ptDag->ValDag.Atome.ValAtome.Entier);
    ptDag = ptDag->Frere;
  }
     
  /* ajout du no de flexion de la racine */
  if( p_avecStructureDeTraits ) {
    if( !AjouteCheminEtAtomeFils( p_hdlDagStructTraits ,
				  CheminFlexion , _AtomeEntier , 
				  atome , _True ) ) {
      printf( "Erreur 8\n" );
      return _False;
    }
  }

  /* fin de la liste attendue */
  if( ptDag ) {
    TxtChaineErreur( 287 , ChErreurCompil );
    return _False;
  } 

 return _True;
}

/*****************************************************************************
 passage d'un lemme dont on a trouve une flexion au lemme flechi
 correspondant
*****************************************************************************/
char     AlloueEtGarnitLemme( TabLemmeMem p_ptLemme, int p_catLex , 
                              int p_indFlexion , int p_indDerivation ,
                              TabLemmeMem *p_hdlLemmeMem , int p_fo ,
                              ChIdent p_chLemme )
{
   int                             rangPremFlexion;
   int                             rangPremDerivation;
   struct ValAtomeMem              atome;
   TabDagMem                      *hdlDag;
   TabDagMem                       ptDagAux;
   TabTermeMem                     ptTerme;

   if( !(*p_hdlLemmeMem = NouveauLemme() ) )
   {
       return _False;
   }
   (*p_hdlLemmeMem)->ValTerme  = p_ptLemme->ValTerme;
   (*p_hdlLemmeMem)->NbDerives  = p_ptLemme->NbDerives;
   (*p_hdlLemmeMem)->NbVoisinsSemantiques = p_ptLemme->NbVoisinsSemantiques;
   (*p_hdlLemmeMem)->NbVoisinsSemantiquesRec 
     = p_ptLemme->NbVoisinsSemantiquesRec;
   (*p_hdlLemmeMem)->NbRegles  = p_ptLemme->NbRegles;
   (*p_hdlLemmeMem)->PosFicDerives =  p_ptLemme->PosFicDerives;
   (*p_hdlLemmeMem)->PosFicLiensSemantiques =  
     p_ptLemme->PosFicLiensSemantiques;
   (*p_hdlLemmeMem)->PosFicLiensSemantiquesRec =  
     p_ptLemme->PosFicLiensSemantiquesRec;
   strcpy( (*p_hdlLemmeMem)->ValTerme.Ident , p_chLemme );
   (*p_hdlLemmeMem)->ValTerme.ArcsFils = NULL;

   /* pour le desuffixage uniforme, on ne tient pas compte des infos
      morphologiques pour gerer les cas de transcategorisation */
   if( SuffStripping )
   {
      /* ajout de la chaine du lemme pour gerer la negation par
         unification */
      atome.AtomeConjDisj = _False; 
      strcpy( atome.ValAtome.Chaine , p_chLemme );
      if( !AjouteCheminEtAtomeFils( &((*p_hdlLemmeMem)->ValTerme.ArcsFils) ,
                                    CheminLemme , _AtomeChaine , 
				    atome , _True ) )
      {
          DemarqueEtRendLemme( p_hdlLemmeMem );
          return _False;
      }
      return _True;
   }

   /* En flexion ou en derivation (donc categorie non _Nil), recopie
      des infos du lemme : sa categorie, son no de flexion
      essentiellement */
   if( p_catLex != _Nil )
   {
       /* forme flechie ou forme qui n'est ni flechie ni derivee */
       if( p_indDerivation == _Nil )
       {
           if( !CopieDag( &((*p_hdlLemmeMem)->ValTerme.ArcsFils) , 
                          p_ptLemme->ValTerme.ArcsFils ) )
           {
               DemarqueEtRendLemme( p_hdlLemmeMem );
               return _False;
           }
       }
       /* En derivation, recopie des infos du lemme sur un chemin qui les
          detache du mot flechi */
       else
       {
          if( ( hdlDag
                  = CopieChemin( &((*p_hdlLemmeMem)->ValTerme.ArcsFils) , 
                                 CheminRacine ) ) )
          {
               if( !CopieDag( hdlDag ,
                              p_ptLemme->ValTerme.ArcsFils ) )
               {
                   DemarqueEtRendLemme( p_hdlLemmeMem );
                   return _False;
               }
               /* en derivation, on ne peut pas reprendre la reference
                  de la base, puisqu'il s'agit d'un autre lemme. La
                  reference doit rester celle d'un mot inconnu */
               (*p_hdlLemmeMem)->ValTerme.Ref = _Nil;
               (*p_hdlLemmeMem)->ValTerme.RefRacine = p_ptLemme->ValTerme.Ref;
          }
          else
          {
              DemarqueEtRendLemme( p_hdlLemmeMem );
              return _False;
          }
       }
   }

   /* ajout de la forme du mot et de son trait self */
   if( p_fo != _Nil ) {
      atome.AtomeConjDisj = _False;
      *(atome.ValAtome.Chaine) = (char)p_fo;
      *(atome.ValAtome.Chaine + 1) = '\0';
      if( !AjouteCheminEtAtomeFils( &((*p_hdlLemmeMem)->ValTerme.ArcsFils) ,
                                    CheminForme , _AtomeChaine , 
				    atome , _True ) ) {
          DemarqueEtRendLemme( p_hdlLemmeMem );
          return _False;
      }
   }

   (*p_hdlLemmeMem)->ValTerme.TypeTerme  = _MotFlechi;

   /* recopie du dag de flexion */
   ptDagAux = NULL;
   if( p_catLex != _Nil )
   {
       if( p_indFlexion != _Nil )
       {
           rangPremFlexion 
                 = DicoTraitsCatLex[ p_catLex ].RangPremDagFlexion;

           /* suppression de la categorie du lemme si la flexion redefinit
              la categorie (dans le cas de transcategorisation) */
           if( rangPremFlexion != _Nil )
           {
              hdlDag = ExisteCheminFils( CheminCategorie ,
                &(TabDagsSuffixage[ rangPremFlexion + p_indFlexion - 1 ]) );
              if( hdlDag && *hdlDag && (*hdlDag)->TypeDag == _AtomeRef )
              {
                   SupprimeCheminEtValeur( CheminCategorie ,
                        &((*p_hdlLemmeMem)->ValTerme.ArcsFils) );
              }

              /* copie du dag du lemme (les tags partiels) */
              if( !CopieDag( &ptDagAux , 
                  TabDagsSuffixage[ rangPremFlexion + p_indFlexion - 1 ] ) )
              {
                  return _False;
              }
          }
       }
       else if( p_indDerivation != _Nil )
       {
           rangPremDerivation 
                 = DicoTraitsCatLex[ p_catLex ].RangPremDagDerivation;

           if( rangPremDerivation != _Nil )
           {

               if( !CopieDag( &ptDagAux , 
                   TabDagsSuffixage[ rangPremDerivation 
                                     + p_indDerivation - 1 ] ) )
               {
                  DemarqueEtRendLemme( p_hdlLemmeMem );
                  return _False;
               }
           }
       }

       /* placement du dag du lemme flechi dans un terme pour la
          restauration des pointeurs */
       if( !(ptTerme = NouveauTerme()) )
       {
           return _False;
       }
       ptTerme->ValTerme.ArcsFils = (*p_hdlLemmeMem)->ValTerme.ArcsFils;

       /* les 4 bases de restauration des pointeurs */
       PtOperLogU = NULL; 
       HdlDagU = &ptDagAux; 
       HdlTermeU = &ptTerme;
       PtBlocU = NULL;   

       /* echec d'unification: on elimine le lemme flechi considere */
       if( !Unifier( &ptDagAux , &(ptTerme->ValTerme.ArcsFils) ) )
       {
           /* on coupe le lien entre le lemme et son dag, puisque le
              dag va etre desalloue lors de la desallocation du terme */
           (*p_hdlLemmeMem)->ValTerme.ArcsFils = NULL;
           /* desallocation du terme */
           DemarqueEtRendTerme( &ptTerme );
           /* desallocation du lemme localement alloue */
           DemarqueEtRendLemme( p_hdlLemmeMem );
           return _False;
       }
       else
       /* reussite d'unification: on garde les traits du lemme
          unifie */
       {
           (*p_hdlLemmeMem)->ValTerme.ArcsFils 
                 = ptTerme->ValTerme.ArcsFils;
           /* on coupe le lien entre le lemme et son dag, puisque le
              dag est conserve sur le lemme flechi */
           ptTerme->ValTerme.ArcsFils = NULL;
           /* desallocation du terme */
           DemarqueEtRendTerme( &ptTerme );
       }
   }

   return _True;
}


/*****************************************************************************
     VERIFICATION DE NON PRESENCE DE DOUBLONS DANS UNE LISTE DE LEMMES
*****************************************************************************/
char     LemmeHorsListeLemme( long p_refLemmeDonnee , TabLemmeMem p_ptLemme )
{
    TabLemmeMem       ptLemmeCou;

    ptLemmeCou = p_ptLemme;
    while( ptLemmeCou )
    {
        if( ptLemmeCou->ValTerme.Ref == p_refLemmeDonnee )
        {
            return _False;
        }
        ptLemmeCou = ptLemmeCou->LemmeFrere;
    }
    return _True;
}

/*****************************************************************************
          PASSAGE D'UN LEMME A SA FAMILLE DERIVATIONNELLE
*****************************************************************************/
char     LemmeADerives( TabLemmeMem p_ptLemme , 
			TabLemmeMem *p_hdlLemmeTeteFamDeriv )
{
    long              posFicLemmeRacine;
    long              posFicLemmeDerive;
    long              posFicLienMorphEtSem;

    TabLemmeMem       ptLemmeDerivePrec;
    TabLemmeMem       ptLemmeDeriveCou;
    TabLemmeMem       ptLemmeRacine;
    TabLemmeMem       ptLemmeInitial = p_ptLemme;

    ChIdent           chaineAnnexe;
    TabChar           chaineRacine;
    TabLong           ptRefRacine;
    TabLong           ptCategorie;
    TabLong           ptNoFlexion;  
    TabDagMem        *hdlDag;

    ptLemmeDerivePrec = NULL;
    ptLemmeDeriveCou = NULL;
    (*p_hdlLemmeTeteFamDeriv) = NULL;
    while( p_ptLemme )
    {
      ptLemmeRacine = NULL;
      posFicLemmeRacine = p_ptLemme->PosFicRacine;

      /* le lemme a une racine: recuperation de la racine */
      if( posFicLemmeRacine > 0L )
      {
         if( !(ptLemmeRacine = NouveauLemme() ) )
         {
             return _False;
         }

         /* on recupere le lemme de la racine */
         if( RestructurerBDicLemme( (unsigned long *)&posFicLemmeRacine , 
                                    ptLemmeRacine->ValTerme.Ident  ,
                                    ptLemmeRacine , 
                                    chaineAnnexe )  != _Nil )
         {
             if( LemmeHorsListeLemme( ptLemmeRacine->ValTerme.Ref , 
                                      ptLemmeInitial )
                 && LemmeHorsListeLemme( ptLemmeRacine->ValTerme.Ref , 
                                         (*p_hdlLemmeTeteFamDeriv) ) )
             {
                 ptLemmeRacine->ValTerme.TypeTerme  = _MotRacine;
                 if( ptLemmeDerivePrec )
                 {
                      ptLemmeDerivePrec->LemmeFrere = ptLemmeRacine;
                 }
                 ptLemmeDerivePrec = ptLemmeRacine;
                 if( !(*p_hdlLemmeTeteFamDeriv) )
                 {
                      (*p_hdlLemmeTeteFamDeriv) = ptLemmeRacine;
                 }
             }
             else
             {
                 DemarqueEtRendLemme( &ptLemmeRacine );
             }
         }
         /* on recupere le lemme de la racine */
         else
         /* lemme de la racine non extrait */
         {
              TxtErreur( 267 , "LemmeADerives(almotrai)" );
              DemarqueEtRendLemme( &ptLemmeRacine );
              return _False;;
         }
         /* lemme de la racine non extrait */
      }
      /* le lemme a une racine: recuperation de la racine */
      else if( p_ptLemme->NbDerives > 0 )
      /* le lemme est une racine */
      {
         ptLemmeRacine = p_ptLemme;
      }
      /* le lemme est une racine */
      else
      /* le lemme n'est pas une racine, il n'a pas de reference
         fichier vers une racine, mais il a des traits de racine */
      {
	/* le chemin <root> existe */
	if( (hdlDag = ExisteCheminFils( CheminRacine ,  
					&(p_ptLemme->ValTerme.ArcsFils) ) ) ) {
	  /* les traits de la racine sont obtenus */
	  if( !LienUniteLexicaleATraits( hdlDag , 
					 (TabDagMem *)NULL ,
					 &ptRefRacine , &chaineRacine ,
					 &ptCategorie , &ptNoFlexion , 
					 _False ) ) {
	    TxtErreur( 267 , "LemmeADerives(almotrai)" );
	    return _False;
	  }
	  /* les traits de la racine sont obtenus */

	  /* la racine existe bien */
	  if( chaineRacine && (*ptRefRacine) != _Nil 
	      && ( posFicLemmeRacine = 
		   (long)ChercheLemmeBDic( chaineRacine , (*ptRefRacine) , 
					   &ptLemmeRacine ) ) )  {
	    /* la racine n'est pas encore connue */
	    if( LemmeHorsListeLemme( ptLemmeRacine->ValTerme.Ref , 
				     ptLemmeInitial )
		&& LemmeHorsListeLemme( ptLemmeRacine->ValTerme.Ref , 
					(*p_hdlLemmeTeteFamDeriv) ) ) {
		ptLemmeRacine->ValTerme.TypeTerme  = _MotRacine;
		if( ptLemmeDerivePrec )
		  {
		    ptLemmeDerivePrec->LemmeFrere = ptLemmeRacine;
		  }
		ptLemmeDerivePrec = ptLemmeRacine;
		if( !(*p_hdlLemmeTeteFamDeriv) )
		  {
		    (*p_hdlLemmeTeteFamDeriv) = ptLemmeRacine;
		  }
	    }
	    /* la racine n'est pas encore connue */
	    else {
	      DemarqueEtRendLemme( &ptLemmeRacine );
	    }
	  }
	  /* la racine existe bien */
	  else {
	    DemarqueEtRendLemme( &ptLemmeRacine );
	  }
	}
	/* le chemin <root> existe */
      }
      /* le lemme est une racine */

      /* le lemme a une racine */
      if( ptLemmeRacine )
      {
         TabRegleMem                       ptRegle;
         int                               indDerive;
         TabLong                           ptLong;

         ptRegle = NULL;
         AjouteReglesARetourner( ptLemmeRacine , &ptRegle );
         ReporteReglesSurReglesPhrase( ptRegle );

         /* recuperation des derives de la racine non encore
            presents */
         posFicLienMorphEtSem = ptLemmeRacine->PosFicDerives;
         for( indDerive = 0 ; indDerive < ptLemmeRacine->NbDerives ; 
              indDerive++ )
         {
           /* lemme d'un des derives de la racine non extrait */
           if( posFicLienMorphEtSem < 0L )
           {
                TxtErreur( 268 , "LemmeADerives(almotrai)" );
                return _False;;
	   }

           /* lecture du lien pour ce derive */
           if( !BufferLiensMorphologiquesEtSemantiques )
           {
               posFicLienMorphEtSem += LireEcrire( posFicLienMorphEtSem , 
                                    (unsigned long)sizeof(long) ,
                                    (TabChar)(&posFicLemmeDerive) , 
                                    _Lire , FicBDicLiens );
	   }
           else
           {
               ptLong = (TabLong)(BufferLiensMorphologiquesEtSemantiques
                                  + posFicLienMorphEtSem);
               posFicLemmeDerive = *ptLong;
               posFicLienMorphEtSem += (long)sizeof(long);
	   }

           if( !(ptLemmeDeriveCou = NouveauLemme() ) )
           {
               return _False;
           }
           /* on recupere le lemme d'un des derives de la racine */
           if( RestructurerBDicLemme( 
                        (unsigned long *)&posFicLemmeDerive ,
                        ptLemmeDeriveCou->ValTerme.Ident ,
                        ptLemmeDeriveCou , 
                        chaineAnnexe )  != _Nil )
           {
              if( LemmeHorsListeLemme( ptLemmeDeriveCou->ValTerme.Ref , 
                                       ptLemmeInitial )
                  && LemmeHorsListeLemme( ptLemmeDeriveCou->ValTerme.Ref , 
                                          (*p_hdlLemmeTeteFamDeriv) ) )
              {
                  TabRegleMem          ptRegle;

                  ptLemmeDeriveCou->ValTerme.TypeTerme  = _MotDerive;
                  ptRegle = NULL;
                  AjouteReglesARetourner( ptLemmeDeriveCou , &ptRegle );
                  ReporteReglesSurReglesPhrase( ptRegle );

                  if( !(*p_hdlLemmeTeteFamDeriv) )
                  {
                       (*p_hdlLemmeTeteFamDeriv) = ptLemmeDeriveCou;
                  }
                  if( ptLemmeDerivePrec )
                  {
                      ptLemmeDerivePrec->LemmeFrere = ptLemmeDeriveCou;
                  }
                  ptLemmeDerivePrec = ptLemmeDeriveCou;
              }
              else
              {
                   DemarqueEtRendLemme( &ptLemmeDeriveCou );
              }
               /* on recupere le lemme d'un des derives de la
                  racine */
           }
	   /* pointeur fichier du derive connu */
         }
         /* recuperation des derives de la racine non encore
            presents */
      }

      p_ptLemme = p_ptLemme->LemmeFrere;
    }

    /* le lemme a une racine */
    return _True;
}

/*****************************************************************************
          PASSAGE D'UN LEMME A SA FAMILLE SEMANTIQUE
*****************************************************************************/
char     LemmeAFamilleSem( TabLemmeMem p_ptLemmeBaseDesLiens , 
			   TabLemmeMem p_ptLemmeStopList , 
			   TabLemmeMem *p_hdlLemmeTeteFamilleSem  ,
			   int p_estLienSemantiqueReciproque )
{
    long              posFicLemmeSemant;
    long              posFicLienMorphEtSem;

    TabLemmeMem       ptLemmeSemantPrec;
    TabLemmeMem       ptLemmeSemantCou;
    TabLemmeMem       ptLemmeInitial = p_ptLemmeBaseDesLiens;

    ChIdent           chaineAnnexe;
    int               indLienSem;
    int               nbVoisinsSemantiques;
    TabLong           ptLong;

    ptLemmeSemantPrec = NULL;
    ptLemmeSemantCou = NULL;
    (*p_hdlLemmeTeteFamilleSem) = NULL;
    while( p_ptLemmeBaseDesLiens )
    {
      if( !p_estLienSemantiqueReciproque ) {
	nbVoisinsSemantiques 
	  = p_ptLemmeBaseDesLiens->NbVoisinsSemantiques;
      }
      else {
	nbVoisinsSemantiques 
	  = p_ptLemmeBaseDesLiens->NbVoisinsSemantiquesRec;
      }	
      /* le lemme a des liens semantiques: recuperation de la famille sem */
      if( ( nbVoisinsSemantiques > 0 ) ) {
         /* recuperation des liens semantiques */
	if( !p_estLienSemantiqueReciproque ) {
	  posFicLienMorphEtSem 
	    = p_ptLemmeBaseDesLiens->PosFicLiensSemantiques;
	}
	else {
	  posFicLienMorphEtSem 
	    = p_ptLemmeBaseDesLiens->PosFicLiensSemantiquesRec;
	}	

         for( indLienSem = 0 ; 
	      indLienSem < nbVoisinsSemantiques ; 
              indLienSem++ )
         {
           /* lemme d'un des liens semantiques de la racine non extrait */
           if( posFicLienMorphEtSem < 0L )
           {
                TxtErreur( 268 , "LemmeAFamilleSem(almotrai)" );
                return _False;;
	   }

           /* lecture de ce lien semantique */
           if( !BufferLiensMorphologiquesEtSemantiques )
           {
               posFicLienMorphEtSem += LireEcrire( posFicLienMorphEtSem , 
                                    (unsigned long)sizeof(long) ,
                                    (TabChar)(&posFicLemmeSemant) , 
                                    _Lire , FicBDicLiens );
	   }
           else
           {
               ptLong = (TabLong)(BufferLiensMorphologiquesEtSemantiques
                                  + posFicLienMorphEtSem);
               posFicLemmeSemant = *ptLong;
               posFicLienMorphEtSem += (long)sizeof(long);
	   }

           if( !(ptLemmeSemantCou = NouveauLemme() ) )
           {
               return _False;
           }
           /* on recupere le lemme d'un des liens sem de la racine */
           if( RestructurerBDicLemme( 
                        (unsigned long *)&posFicLemmeSemant ,
                        ptLemmeSemantCou->ValTerme.Ident ,
                        ptLemmeSemantCou , 
                        chaineAnnexe )  != _Nil )
           {
              if( LemmeHorsListeLemme( ptLemmeSemantCou->ValTerme.Ref , 
                                       ptLemmeInitial )
                  && LemmeHorsListeLemme( ptLemmeSemantCou->ValTerme.Ref , 
                                          (*p_hdlLemmeTeteFamilleSem) )
                  && LemmeHorsListeLemme( ptLemmeSemantCou->ValTerme.Ref , 
                                          p_ptLemmeStopList ) )
              {
                  TabRegleMem          ptRegle;

		  if( !p_estLienSemantiqueReciproque ) {
		    ptLemmeSemantCou->ValTerme.TypeTerme  = _MotSemantLie;
		  }
		  else {
		    ptLemmeSemantCou->ValTerme.TypeTerme  = _MotSemantLieRec;
		  }	
                  ptRegle = NULL;
                  AjouteReglesARetourner( ptLemmeSemantCou , &ptRegle );
                  ReporteReglesSurReglesPhrase( ptRegle );

                  if( !(*p_hdlLemmeTeteFamilleSem) )
                  {
                       (*p_hdlLemmeTeteFamilleSem) = ptLemmeSemantCou;
                  }
                  if( ptLemmeSemantPrec )
                  {
                      ptLemmeSemantPrec->LemmeFrere = ptLemmeSemantCou;
                  }
                  ptLemmeSemantPrec = ptLemmeSemantCou;
              }
              else
              {
                   DemarqueEtRendLemme( &ptLemmeSemantCou );
              }
               /* on recupere le lemme d'un des Semants de la
                  racine */
           }
	   /* pointeur fichier du Semant connu */
         }
         /* recuperation des Semants de la racine non encore
            presents */
      }

      p_ptLemmeBaseDesLiens = p_ptLemmeBaseDesLiens->LemmeFrere;
    }

    /* le lemme a une racine */
    return _True;
}

/*****************************************************************************
    en cas d'exploitation automatique des liens derivationnels, on
    recupere la famille derivationnelle associee a chaque mot du texte:
          PASSAGE D'UN MOT A SA FAMILLE DERIVATIONNELLE
          ET PASSAGE D'UN MOT A SA FAMILLE SEMANTIQUE
*****************************************************************************/
char     MotADerivesEtFamilleSem( TabMotMem p_ptMot )
{
    TabLemmeMem       ptLemme;
    TabLemmeMem       ptLemmeTeteFamilleDerivationnelle;
    TabLemmeMem       ptLemmeTeteFamilleSemantique;
    TabLemmeMem       ptLemmeTeteFamilleSemantiqueRec;
    TabLemmeMem       ptLemmeTeteFamilleSemantiqueRecEtendus;
    TabLemmeMem       ptTeteLemmesAjoutes;

    ptLemmeTeteFamilleDerivationnelle = NULL;
    ptLemmeTeteFamilleSemantique = NULL;
    ptLemmeTeteFamilleSemantiqueRec = NULL;
    ptLemmeTeteFamilleSemantiqueRecEtendus = NULL;
    ptTeteLemmesAjoutes = NULL;

    /* pour chaque lemme, on recherche sa famille derivationnelle:
       --- sa racine,
       --- les derives de cette racine (sauf ceux qui sont deja
           dans les lemmes associes au mot courant */
     if( ExploitationLiensMorpho ) {
      if( !LemmeADerives( p_ptMot->PtListeLemmes , 
			  &ptLemmeTeteFamilleDerivationnelle ) ) {
	return _False;
      }
     }

    /* pour chaque lemme, on recherche les lemmes qui lui sont
       directement lies et ceux qui lui sont reciproquement lies */
    if( ExploitationLiensSem ) {
      if( !LemmeAFamilleSem( p_ptMot->PtListeLemmes , 
			     NULL ,
			     &ptLemmeTeteFamilleSemantique ,
			     _False ) ) {
	return _False;
      }
      if( ExploitationLiensSem == _SemantiqueDeClasses ) {
	if( !LemmeAFamilleSem( p_ptMot->PtListeLemmes , 
			       ptLemmeTeteFamilleSemantique ,
			       &ptLemmeTeteFamilleSemantiqueRec ,
			       _True ) ) {
	  return _False;
	}
      }
    }

    /* on concatene les lemmes, la famille derivationnelle,
      la famille semantique  et la famille semantique reciproque */
    ptLemme = p_ptMot->PtListeLemmes;
    if( ptLemmeTeteFamilleDerivationnelle ) {
      while( ptLemme ) {
	if( ptLemme->LemmeFrere ) {
	  ptLemme = ptLemme->LemmeFrere;
	}
	else {
	  ptTeteLemmesAjoutes = ptLemmeTeteFamilleDerivationnelle;
	  ptLemme->LemmeFrere = ptLemmeTeteFamilleDerivationnelle;
	  break;
	}
      }
    }

    if( ptLemmeTeteFamilleSemantique ) {
      while( ptLemme ) {
	if( ptLemme->LemmeFrere ) {
	  ptLemme = ptLemme->LemmeFrere;
	}
	else {
	  if( !ptTeteLemmesAjoutes ) {
	    ptTeteLemmesAjoutes = ptLemmeTeteFamilleSemantique;
	  }
	  ptLemme->LemmeFrere = ptLemmeTeteFamilleSemantique;
	  break;
	}
      }
    }

    if( ptLemmeTeteFamilleSemantiqueRec ) {
      while( ptLemme ) {
	if( ptLemme->LemmeFrere ) {
	  ptLemme = ptLemme->LemmeFrere;
	}
	else {
	  if( !ptTeteLemmesAjoutes ) {
	    ptTeteLemmesAjoutes = ptLemmeTeteFamilleSemantiqueRec;
	  }
	  ptLemme->LemmeFrere = ptLemmeTeteFamilleSemantiqueRec;
	  break;
	}
      }
    }

    /* on ajoute les liens semantiques reciproques des lemmes auxquels
       il est semantiquement lie et retrouves dans la boucle
       precedente */
    if( ExploitationLiensSem == _SemantiqueDeClasses ) {
      if( !LemmeAFamilleSem( ptLemmeTeteFamilleSemantique , 
			     p_ptMot->PtListeLemmes ,
			     &ptLemmeTeteFamilleSemantiqueRecEtendus ,
			     _True ) ) {
	return _False;
      }
    }

    /* on concatene la famille semantique reciproque aux lemmes, la
       famille derivationnelle et la famille semantique */
    if( ptLemmeTeteFamilleSemantiqueRecEtendus ) {
      while( ptLemme ) {
	if( ptLemme->LemmeFrere ) {
	  ptLemme = ptLemme->LemmeFrere;
	}
	else {
	  if( !ptTeteLemmesAjoutes ) {
	    ptTeteLemmesAjoutes = ptLemmeTeteFamilleSemantiqueRecEtendus;
	  }
	  ptLemme->LemmeFrere = ptLemmeTeteFamilleSemantiqueRecEtendus;
	  break;
	}
      }
    }
    return _True;
}


/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                         fin de almotrai.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
