/*
 * FASt Term Recognizer
 *
 *  FASTR/alphramo.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for sentence tokenizing
 */

#include "al-ext.h"
#include "al-var.h"
#include "al-car.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*****************************************************************************
                   rechargement du buffer de stockage
*****************************************************************************

/* rechargement du buffer de stockage du texte dans le cas ou on est a la fin
   du precedent ou si on n'a pas encore lu le texte
   p_ptFicEntree est le pointeur vers le fichier de texte        */

char     LireBuffTexte( FILE *p_ptFicEntree )
{
    unsigned long              posFic;
    long                       nbLus;
    unsigned long              tailleRestanteDuFichier;
    long                       tailleALire;
    TabChar                    ptCou;

    if( EstEntreeSortieBuffer ) {
      IndCar = 1L;
      IndCarM = 0L;
      return EOF;
    }

    if( p_ptFicEntree != stdin) {
        /* on stocke les prochains caracteres dans le buffer */
        fgetpos( p_ptFicEntree , &posFic );
        fseek( p_ptFicEntree , 0L , 2 );
        fgetpos( p_ptFicEntree , &tailleRestanteDuFichier );
        tailleRestanteDuFichier = tailleRestanteDuFichier - posFic;
        tailleALire = _Min( (long)tailleRestanteDuFichier , 
                            (long)TailleBuffStd );
        if( !tailleALire )
        {
             IndCar = 1L;
             IndCarM = 0L;
             return EOF;
        }      

        nbLus = LireEcrire( posFic , tailleALire , BuffTexte , 
                            _Lire , p_ptFicEntree );
        fseek( p_ptFicEntree , posFic + nbLus  , 0 );
        if( !nbLus )
        {
             IndCar = 1L;
             IndCarM = 0L;
             return EOF;
        }
    }
    else {
        nbLus = 0L;
        while( nbLus < (long)TailleBuffStd ) {
            if( (BuffTexte[ nbLus++ ] = getchar()) == EOF ) {
               break;
            }
        }
        if( !nbLus )
        {
             IndCar = 1L;
             IndCarM = 0L;
             return EOF;
        }
    }

    /* suppression d'un eventuel caractere egal a EOF du texte d'orogine */
    ptCou = BuffTexte;
    for( posFic = 0L ; posFic < (unsigned long)nbLus ; (posFic++ , ptCou++ ) )
    {
        if( *ptCou == EOF )
        {
            *ptCou = ' ';
        }
    }

    if( nbLus <= 0L )
    {
        BuffTexte[ 0 ] = EOF;
        nbLus = 1L;
    }
    if( nbLus < (long)TailleBuffStd )
    {
        BuffTexte[ nbLus ] = EOF;
        nbLus++;
    }
    IndCar = 1L;
    IndCarM = nbLus;

    return BuffTexte[ 0L ];
}

/*****************************************************************************
                        etape 1 : phrase ----> mot
******************************************************************************/

void     LireMot( char *p_carlu , TabMotMem p_ptMot , FILE *p_ptFicEntree)
{
    char                 c;
    char                *ptChMot;
    int                  indChar;

    c = *p_carlu;
    /* pointeur vers la chaine du mot a lire */
    ptChMot = p_ptMot->ChaineMot;
    indChar = 0;
    
    /* le programme se charge de la segmentation */
    if( SegmentationAutomatique )
    {
        /* mot commencant par un caractere alphabetique */
        if( _Alpha(c) )
        {
            while( (_Alphanum(c) || strchr( CarSpeciauxMots , c ) )
                    && !strchr( Ponctuation , c )
                    && indChar < (_MaxLetMot - 1) )
            {
                /* saut des caracteres alphanumeriques              */
                ptChMot[ indChar ] = c;
                indChar++;
                c = _LireC( p_ptFicEntree );
            }
            if( indChar == _MaxLetIdent - 1 )
            {
                while( _Alphanum(c) || strchr( CarSpeciauxMots , c ) )
                {
                    c = _LireC( p_ptFicEntree );
                }
            }
        }
        /* mot commencant par un caractere numerique                */
        else if( _Num(c) )
        {
            while( ( _Alphanum(c) || strchr( CarSpeciauxMots , c ) ) 
                   && indChar < (_MaxLetMot - 1) )
            {
                /* saut des caracteres numeriques                   */
                ptChMot[ indChar ] = c;
                indChar++;
                c = _LireC( p_ptFicEntree );
            }
            if( indChar == _MaxLetIdent - 1 )
            {
                while( _Alphanum(c) || strchr( CarSpeciauxMots , c ) )
                {
                    c = _LireC( p_ptFicEntree );
                }
            }
        }
        /* ponctuation                                          */
        else if( strchr( Ponctuation , c ) )
        {
            ptChMot[ indChar ] = c;
            indChar++;
            c = _LireC( p_ptFicEntree );
        }
    }
    /* le texte est suppose deja segmente et les mots sont separes par
       des blancs */
    else
    {
        while( !_CarBlanc( c )
               && indChar < (_MaxLetMot - 1) && c != EOF )
        {
            /* saut des caracteres non blancs */
            ptChMot[ indChar ] = c;
            indChar++;
            c = _LireC( p_ptFicEntree );
        }
        if( indChar == _MaxLetIdent - 1 )
        {
            while( !_CarBlanc( c ) && c != EOF )
            {
                c = _LireC( p_ptFicEntree );
            }
        }
    }
    ptChMot[ indChar ] = '\0';

    /* cas d'un mot debutant par une majuscule  de longueur >= 2 */
    /* on le passe en minuscule */
    if( MinusculisationAutomatique == 1 )
    {
        if( *ptChMot && *(ptChMot + 1)
            && ( *ptChMot >= 'A' && *ptChMot <= 'Z' )
            && !( *(ptChMot + 1) >= 'A' && *(ptChMot + 1) <= 'Z') 
          )
        {
              (*ptChMot) = 'a' + ((*ptChMot) - 'A');
              p_ptMot->Forme = 'M';
        }
    }
    else if( MinusculisationAutomatique == 2 )
    {
      int    ind = 0;
      
      while(  *(ptChMot + ind) ) {
	if( *(ptChMot + ind) >= 'A' && *(ptChMot + ind) <= 'Z' ) {
	  *(ptChMot + ind) = 'a' + (*(ptChMot + ind) - 'A');
	}
	ind++;
      }
      p_ptMot->Forme = 'A';
    }

    /* traitement du caractere suivant un mot                   */
    if (c == '\'')
    {
       p_ptMot->Forme = '\'';
       c = _LireC( p_ptFicEntree );
    }
    else if (c == '-')
    {
       p_ptMot->Forme = '-';
       c = _LireC( p_ptFicEntree );
    }

    *p_carlu = c;
}

/*****************************************************************************
                       lecture de tous les mots d'une phrase
*****************************************************************************/

void     LirePhrase( char *p_carlu , FILE *p_ptFicEntree )
{
    char            estNeg;
    long            nbLignesTotCou;
    int             nbCharCou;
    char            derMot;
    TabLemmeMem     ptLemme;
    TabMotMem       ptMot;
    TabMotMem      *hdlMotPrec;

    if( *p_carlu == EOF )
    {
         return;
    }

    NbMots = 0;
    MotsPhrase = NULL;
    hdlMotPrec = &MotsPhrase;

    /* entree = texte : retour ligne == saut de ligne */
    if( *p_carlu == '\n' 
        && (FormatDEntree == _Txt) )
    {
         NbLignesTot++;
         *p_carlu = _LireC( p_ptFicEntree );
    }

    /* entree = sortie anterieure au format inist */
    /* ligne vide : fin du fichier */
    if( FormatDEntree == _SortieInist && *p_carlu == '\n' )
    {
         *p_carlu = EOF;
         return;
    }
        
    /* entree format inist */
    /* on note le no de ligne et on saute une tab */
    if( (FormatDEntree == _SortieInist || FormatDEntree == _TxtNumerote)
        && (_Num(*p_carlu) || *p_carlu == '-') )
    {
        nbLignesTotCou = 0L;
        nbCharCou = 0;

        /* lecture du numero de ligne */
        estNeg = _False;
        if( *p_carlu == '-' )
        {
            estNeg = _True;
            *p_carlu = _LireC( p_ptFicEntree );
        }
        while( _Num(*p_carlu) )
        {
            nbLignesTotCou = 10L * nbLignesTotCou + (long)(*p_carlu - '0');
            nbCharCou++;
            *p_carlu = _LireC( p_ptFicEntree );
        }
        /* saut d'une tabulation */
        if( *p_carlu == '\t' )
        {
            *p_carlu = _LireC( p_ptFicEntree );
            if( nbCharCou == 6 )
            {
                NbLignesTot = nbLignesTotCou;
                if( estNeg )
                {
                    NbLignesTot = -NbLignesTot;
                }
            }
        }        
    }

    /* entree = sortie anterieure au format inist */
    /* on saute le premier champ : l'identificateur du terme extrait */
    if( FormatDEntree == _SortieInist )
    {
        /* saut de l'identificateur */
        while( *p_carlu != EOF && *p_carlu != '\t' )
        {
            *p_carlu = _LireC( p_ptFicEntree );
        }
        /* saut d'une tabulation */
        if( *p_carlu == '\t' )
        {
            *p_carlu = _LireC( p_ptFicEntree );
        }
    }

    /* lecture des mots jusqu'a la fin de la phrase                 */
    do
    {
        /* entree = sortie anterieure au format inist */
        /* on saute le dernier champ : type de variation */
        if( FormatDEntree == _SortieInist )
        {
            if( *p_carlu == '\t' )
            {
                while( *p_carlu != '\n' && *p_carlu != EOF )
                {
                    *p_carlu = _LireC( p_ptFicEntree );
                }
                if( *p_carlu == '\n' )
                {
                    *p_carlu = _LireC( p_ptFicEntree );
                }
                break;
            }
        }

	
	ptMot = NouveauMot();
	*hdlMotPrec = ptMot;
	hdlMotPrec = &(ptMot->MotFrere);
	derMot = _False;

        if( FormatDEntree != _TxtTaggePartiellement 
	    && FormatDEntree != _TxtFenTaggePartielt )
        {
           /* saut des caracteres non alphanumeriques ou ponctuation
              si la segmentation n'a pas ete faite, lecture des blancs
              intermediares sinon */
           /* puis lecture du mot */
	   while( 
              ( SegmentationAutomatique 
                      && !_Alpha(*p_carlu) && !_Num(*p_carlu) 
                      && !strchr( Ponctuation , *p_carlu ) 
                      && *p_carlu != EOF )
              ||
              ( !SegmentationAutomatique 
                      && _CarBlanc(*p_carlu)
                      && *p_carlu != EOF ) )
           {
              /* si entree numerotee on saute la fin de la ligne */
              if( *p_carlu == '\n' && FormatDEntree == _TxtNumerote )
              {
                  *p_carlu = _LireC( p_ptFicEntree );
                  break;
              }
              *p_carlu = _LireC( p_ptFicEntree );
           }
           if( *p_carlu == EOF )
           {
               break;
           }

           LireMot( p_carlu , ptMot , p_ptFicEntree );
        }
        /* mot flechi portant des etiquettes partielles */
        else
        {
           while( *p_carlu != '\'' && *p_carlu != EOF )
           {
                *p_carlu = LireCar(p_ptFicEntree);
           }
           if( *p_carlu == EOF )
           {
               return;
           }
           if( !LireChaineAvecBalise( p_ptFicEntree , p_carlu , 
                                      ptMot->ChaineMot ,
                                      _MaxLetMot - 1 , '\'' , _True ) )
           {
                printf( "String : [%s]\n" , ptMot->ChaineMot );
                TxtErreur( 193 , "LirePhrase(alphramo)" );
                return;
           }

          /* on lit le dag associe au mot flechi en cas d'etiquetage
              partiel. Il est stocke dans un lemme associe au mot
              flechi courant */
           if( !LireLemmesTagges( p_carlu ,  ptMot , &derMot ,
                                  p_ptFicEntree ) )
           {
                printf( "Lemma : [%s]\n" , ptMot->ChaineMot );
                TxtErreur( 240 , "LirePhrase(alphramo)" );
                return;
           }
	   if( Trace ) {
	     /* fprintf( FicLog , "Tagged lemma loaded\n" ); */
	     /* PhotoMemoire( FicLog ); */
	   }

           /* on fait passer le dag du lemme sur le mot flechi */
           if( (ptLemme = ptMot->PtListeLemmes) )
           {
                if( ptLemme->LemmeFrere )
                {
                   TxtErreur( 244 , "LirePhrase(alphramo)" );
                   return;
                }
                ptMot->ArcsFils = ptLemme->ValTerme.ArcsFils;
                ptLemme->ValTerme.ArcsFils = NULL;
                DemarqueEtRendLemme( &ptLemme );
                ptMot->PtListeLemmes = NULL;
           }
        }

        if( Trace )
        {
            fprintf( FicLog , "Word: [%s]\n\n" , ptMot->ChaineMot );
        }

        NbMots++;
    }
    while( !( *(ptMot->ChaineMot) /* non empty char for strchr */ &&
              strchr( PonctuationFinale , *(ptMot->ChaineMot) ) )
           && *p_carlu != EOF && !derMot );
    /* lecture des mots jusqu'a la fin de la phrase */

    if( FormatDEntree == _Txt 
        || FormatDEntree == _TxtTaggePartiellement  
	|| FormatDEntree == _TxtFenTaggePartielt )
    {
         NbLignesTot++;
    }
}

/************************************************************************
Passage de tous les mots de la phrase a des items lexicaux + releve
              des regles
*************************************************************************/
char     PhraseALemme( void )
{
    TabLemmeMem                     ptLemme;
    TabMotMem                       ptMot;
    int                             indTraitSem;

    ReglesPhrase = NULL;
    ptMot = MotsPhrase;

    if( Trace ) {
      fprintf( FicLog ,"\n---\n\n");
      fprintf( FicLog , "Inflected word: [%s]\n" , 
	       ptMot->ChaineMot );
    }

    while( ptMot )
    {
         /* if( Trace ) {
	   PhotoMemoire( FicLog );
	   fprintf( FicLog , "Inflected word to lemma\n" );
	 } */
        if( Trace ) {
	   fprintf( FicLog , "Inflected word: [%s]\n" , 
		    ptMot->ChaineMot );
	 }

         if( !MotALemme( ptMot ) )
         {
            return _False;
         }

	 /* if( Trace ) {
	   PhotoMemoire( FicLog );
	   fprintf( FicLog , "Addition of lexical featrues\n" );
	 } */
         if( ( FormatDEntree == _TxtTaggePartiellement 
	       || FormatDEntree == _TxtFenTaggePartielt )
             && ptMot->ArcsFils )
         {
            if( !FiltreLemme( ptMot ) )
            {
                return _False;
            }
         }

	 ptLemme = ptMot->PtListeLemmes;
	 /* pour chaque lemme, on ajoute les traits par defaut */
	 while( ptLemme ) {
	   /* addition of the self feature shared by root and semantic
	     feature in case of automatic root and semantics */
	   if( AutomaticSelfMorphologicalRoot || AutomaticSelfSemanticLinks )
	     if( !AddsSelfFeature( &((ptLemme)->ValTerme.ArcsFils) ,
				   ReferenceLexicale (
					   &((ptLemme)->ValTerme.ArcsFils) ) ,
				   ChaineLemme( 
					   &((ptLemme)->ValTerme.ArcsFils) ) ,
				   CategorieLexicale( 
					   &((ptLemme)->ValTerme.ArcsFils) ) ,
				   NoTypeSuffixage( 
					   &((ptLemme)->ValTerme.ArcsFils) ,
						   _True ) ) ) {
	       return _False;
	     }
	  
	   /* la racine n'est pas donnee le lemme est sa propre racine */
	   if( AutomaticSelfMorphologicalRoot )
	     {
	       if( !AddsRootFeature( (ptLemme) ) )
		 {
		   return _False;
		 }
	     }
	  
	   /* semantic links are not provided, they are automaically added */
	   if( AutomaticSelfSemanticLinks ) {
	     /****************** scans all the semantic features ***********/
	     int indChemin = 0;
	     while( CheminsSemantique[ indChemin ] ) {
	      
	       if( !AddsSemanticFeatures( (ptLemme) ,
					  CheminsSemantique[ indChemin ]  ) ) {
		 return _False;
	       }
	       indChemin++;
	     }
	   }

	   /* chemin racine */
	   if( !TraitAReferences( &((ptLemme)->ValTerme.ArcsFils) ,
				  CheminRacine ,
				  &((ptLemme)->ValTerme.RefRacine) ,
				  NULL ,
				  _False ) ) {
	     return _False;
	   }

	   /* parcours de tous les traits semantiques declares dans le
	      header */
	   indTraitSem = 0;
	   while( CheminsSemantique[ indTraitSem ] 
		  && indTraitSem < _NbMaxTypesLiensSem ) {
	     if( !TraitAReferences( &((ptLemme)->ValTerme.ArcsFils) ,
				    CheminsSemantique[ indTraitSem ]  ,
				    NULL ,
				    &((ptLemme)->ValTerme.RefSemantique) ,
				    _True ) ) {
	       return _False;
	     }
	     indTraitSem++;
	   }

	   /* trait self */
	   if( !TraitAReferences( &((ptLemme)->ValTerme.ArcsFils) ,
				  CheminSelf ,
				  NULL ,
				  &((ptLemme)->ValTerme.RefSelf) ,
				  _True ) ) {
	     return _False;
	   }
           ptLemme = (ptLemme)->LemmeFrere;
	 }

	 /* if( Trace ) {
	   fprintf( FicLog , "Loading derived words\n" );
	   PhotoMemoire( FicLog );
	 } */
	 if( !MotADerivesEtFamilleSem( ptMot ) ) {
	   return _False;
	 }

	 /* if( Trace ) {
	   fprintf( FicLog , "Processing 1 word completed\n" );
	   PhotoMemoire( FicLog );
	 } */

	 ptMot = ptMot->MotFrere;
    }
    return _True;
}

/*****************************************************************************
    Un lemme ayant ete lu dans un corpus tagge, on recherche les regles
    liees au lemme du dictionnaire ayant la meme reference
*****************************************************************************/
void     ChercheReglesLieesAuLemmeTagge( TabChar p_chMotCou , 
                                         long p_reference ,
				         int *p_nbRegles , 
                                         TabRegleMem *p_hdlRegle )
{
    char                  chaineUlCou[_MaxLetIdent];    /*la chaine du
                                                        Lemme cour.  */
    unsigned long         indAbsLemme;          /* pointeur absolu
                                                   vers le Lemme courante*/
    char                  egale;                /* ulr de meme racine */
    TabChar               chLemmeCou;           /* pt vers la chaine
                                                   du Lemme cour.  */
    struct LemmeMem       lemmeCou;                   /* Lemme courant */
    char                  chaineAnnexe[_MaxLetIdent];   /*l'autre racine
                                     evntuelle
                                     chaineFlechieUlCou[_MaxLetIdent];
                                     chaine flechie du Lemme ct */

    /* BOUCLE SUR LES LEMMES DE MEME CHAINE                    */
    indAbsLemme = IndiceInsertionDicLemmes( p_chMotCou , &egale );
    while( ( chLemmeCou = BufferLemmesCourant + _IRelLemme( indAbsLemme ) )
             && strcmp( p_chMotCou , chLemmeCou ) == 0 )
    {
        if( RestructurerBDicLemme( &indAbsLemme , chaineUlCou  ,
                                   &lemmeCou , chaineAnnexe ) == _Nil
           || EgaliteChaine( chaineUlCou , DernLemme ) )
        {
                 DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
                 break;
        }

        if( p_reference 
               == ReferenceLexicale( &(lemmeCou.ValTerme.ArcsFils) ) )
        {
           if( Trace )
           {
                fprintf( FicLog , "Matched (regular word): [%s]/%ld\n", 
			 chaineUlCou , p_reference );
            }
            AjouteReglesARetourner( &lemmeCou , p_hdlRegle );
            *p_nbRegles = lemmeCou.NbRegles;
        }
        DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
    }
}

/*****************************************************************************
    Un lemme ayant ete lu dans un corpus tagge, on le recherche dans
    le fichier
*****************************************************************************/
unsigned long ChercheLemmeBDic( TabChar p_chMotCou , long p_reference ,
                                TabLemmeMem  *p_hdlLemme )
{
    char                  chaineUlCou[_MaxLetIdent];    /*la chaine du
                                                        Lemme cour.  */
    unsigned long         indAbsLemmeIni;
    unsigned long         indAbsLemme; 
    char                  egale;                /* ulr de meme racine */
    TabChar               chLemmeCou;           /* pt vers la chaine du Lemme 
                                                   cour.      */
    char                  chaineAnnexe[_MaxLetIdent];   /*l'autre racine
                                     evntuelle chaineFlechieUlCou[_MaxLetIdent];
                                     chaine flechie du Lemme ct */

    /* BOUCLE SUR LES LEMMES DE MEME CHAINE                    */
    (*p_hdlLemme) = NouveauLemme();
    indAbsLemme = IndiceInsertionDicLemmes( p_chMotCou , &egale );
    while( ( chLemmeCou = BufferLemmesCourant + _IRelLemme( indAbsLemme ) )
             && strcmp( p_chMotCou , chLemmeCou ) == 0 )
    {
        indAbsLemmeIni = indAbsLemme;
        if( RestructurerBDicLemme( &indAbsLemme , chaineUlCou  ,
                                   (*p_hdlLemme) , chaineAnnexe ) == _Nil
           || EgaliteChaine( chaineUlCou , DernLemme ) )
        {
                 DemarqueEtRendLemme( p_hdlLemme );
                 return 0L;
        }

        if( p_reference 
                == ReferenceLexicale( &((*p_hdlLemme)->ValTerme.ArcsFils) ) )
        {
           if( Trace )
           {
                fprintf( FicLog , "Matched (regular word): [%s]/%ld\n", 
			 chaineUlCou , p_reference );
           }
           return indAbsLemmeIni;
        }
        DemarqueEtRendDag( &((*p_hdlLemme)->ValTerme.ArcsFils) );
    }
    DemarqueEtRendLemme( p_hdlLemme );
    return 0L;
}

/*****************************************************************************
                        mots ----> lemmes flechi
    en fonction de tags lus dans le corpus pretraite
*****************************************************************************/
char     LireLemmesTagges( char *p_carlu, TabMotMem p_ptMot , char *p_derMot ,
                           FILE *p_ptFicEntree  )
{
    TabDagMem                        operLogFils;

    int                              derives;
    int                              racine;
    int                              syntagmes;
    TabLemmeMem                      ptLemme;

    TabRegleMem                      ptRegle;

    TabChar                          chMotCou;

    p_ptMot->PtListeLemmes = NULL;

    /* boucle sur les lemmes associes a un mot flechi */
    while( *p_carlu != '\'' && *p_carlu != EOF )
    {
       if( !(p_ptMot->PtListeLemmes) )
       {
           p_ptMot->PtListeLemmes = NouveauLemme();
           ptLemme = p_ptMot->PtListeLemmes;
       }
       else
       {
           ptLemme->LemmeFrere = NouveauLemme();
           ptLemme = ptLemme->LemmeFrere;
       }
       if( !ptLemme )
       {
           return _False;
       }

       /* lecture des contraintes sur les dags */
       operLogFils = NULL;
       if( !LireContraintesTerme( p_carlu , p_ptFicEntree , &operLogFils , 
                                  &(ptLemme->ValTerme.ArcsFils) ) )
       {
           while( *p_carlu != '.' && *p_carlu != EOF )
           {
                *p_carlu = _LireC( p_ptFicEntree );
           }
           *p_carlu = _LireC( p_ptFicEntree );
           SautBlanc( p_ptFicEntree , p_carlu );
           return _False;
       }
   
       /* unification des contraintes sur les dags */
       if( !UnifierContraintes( &operLogFils , &(ptLemme->ValTerme.ArcsFils) , 
                                NULL , NULL ) )
       {
           return _False;
       }
     
       if( !ReconnaitreCaractere( p_ptFicEntree , (int)'.' ,
                                  p_carlu , _True ) )
       {
           while( *p_carlu != '\'' && *p_carlu != EOF )
           {
                *p_carlu = _LireC( p_ptFicEntree );
           }
           return _False;
       }
   
       ptLemme->ValTerme.TypeTerme  = _MotFlechi;
       ptLemme->ValTerme.Ref 
                 = ReferenceLexicale( &(ptLemme->ValTerme.ArcsFils) );

       *p_derMot = ( CategorieLexicale( &(ptLemme->ValTerme.ArcsFils) )
   		     == RefPonctuationFinale );

       /* tagging partiel, on ne cherche ni la valeur du lemme ni les
          regles associees ni les derives ou mots semantiquement
          associes. Le lemme peut seulement etre suivi par un
          syntagme. */
       if( FormatDEntree == _TxtTaggePartiellement  
	   || FormatDEntree == _TxtFenTaggePartielt )
       {
           if( *p_carlu != '\'' 
	       && *p_carlu != '(' && *p_carlu != EOF )
           {
	       printf( "Next character [%c]\n" , *p_carlu );
               TxtErreur( 243 , "LirePhrase(alphramo)" );
               while( *p_carlu != '\'' && *p_carlu != EOF )
               {
		 char                   ligne[256];
		 LireLigne( p_ptFicEntree , p_carlu , ligne , 256 );
               }
               return _False;
           }
       }
       else {
	 ptRegle = NULL ;
	 if( (chMotCou = ChaineLemme( &(ptLemme->ValTerme.ArcsFils) ) ) ) {
	   ChercheReglesLieesAuLemmeTagge( chMotCou , ptLemme->ValTerme.Ref ,
					   &(ptLemme->NbRegles) , &ptRegle );
	   ReporteReglesSurReglesPhrase( ptRegle );
	 }
       }

       derives = _False;
       racine = _False;
       syntagmes = 0;
       
       /* boucle sur les derives, la racine ou les syntagmes associes
          a un lemme */
       while( *p_carlu == '{' || *p_carlu == '(' || *p_carlu == '[' 
	      || *p_carlu == '+' )
       {
           /* lecture de l'ouverture ou de la poursuite  de l'enumeration */
           if( derives || racine )
           {
              if( !ReconnaitreCaractere( p_ptFicEntree , (int)'+' , 
                                         p_carlu , _True ) )
              {
                  return _False;
              }
           }
           else
           {
              if( *p_carlu == '{' )
              {
                  if( !ReconnaitreChaine( p_ptFicEntree , "{{" , p_carlu ) )
                  {
                      return _False;
                  }
                  SautBlanc( p_ptFicEntree , p_carlu );
                  racine = _True;
              }
              else if( *p_carlu == '[' )
              {
                  if( !ReconnaitreChaine( p_ptFicEntree , "[[" , p_carlu ) )
                  {
                      return _False;
                  }
                  SautBlanc( p_ptFicEntree , p_carlu );
                  derives = _True;
              }
              else if( *p_carlu == '(' )
              {
		char                   ligne[256];

		/* actuellement les syntagmes imbriques sont traites
                   comme des commentaires dans le corpus. Ils ne
                   donnent pas lieu a une representation par un arbre
                   de mots. */
		if( !ReconnaitreChaine( p_ptFicEntree , "((" , p_carlu ) ) {
		  return _False;
		}
		SautBlanc( p_ptFicEntree , p_carlu );
		syntagmes++;

		while( syntagmes && *p_carlu != EOF ) {
		  LireLigne( p_ptFicEntree , p_carlu , ligne , 256 );

		  while( *p_carlu == '(' ) {
		    if( !ReconnaitreChaine( p_ptFicEntree , "((" , 
					    p_carlu ) ) {
		      return _False;
		    }
		    SautBlanc( p_ptFicEntree , p_carlu );
		    syntagmes++;
		  }
		  while( *p_carlu == ')' ) {
		    if( !ReconnaitreChaine( p_ptFicEntree , "))" , 
					    p_carlu ) ) {
		      return _False;
		    }
		    SautBlanc( p_ptFicEntree , p_carlu );
		    syntagmes--;
		  }
		}  
              }
           }
   
           if( racine || derives ){
	     ptLemme->LemmeFrere = NouveauLemme();
	     ptLemme = ptLemme->LemmeFrere;
	   }
   
           /* lecture des contraintes sur les dags */
           operLogFils = NULL;
           if( racine || derives ){
	     if( !LireContraintesTerme( p_carlu , p_ptFicEntree , 
					&operLogFils , 
					&(ptLemme->ValTerme.ArcsFils) ) ) {
		 return _False;
	     }

	     /* unification des contraintes sur les dags */
	     if( !UnifierContraintes( &operLogFils , 
				      &(ptLemme->ValTerme.ArcsFils) , 
				      NULL , NULL ) ) {
	       return _False;
	     }

	     if( !ReconnaitreCaractere( p_ptFicEntree , (int)'.' , 
					p_carlu , _True ) ) {
	       return _False;
	     }
	   }

           if( derives )
           {
               ptLemme->ValTerme.TypeTerme  = _MotDerive;
           }
           else if( racine )
           {
               ptLemme->ValTerme.TypeTerme  = _MotRacine;
           }

          if( derives || racine )
           {
	     /* recherche de la reference lexicale */
	     ptLemme->ValTerme.Ref 
	       = ReferenceLexicale( &(ptLemme->ValTerme.ArcsFils) );

	     /* recherche du lemme */
	     if( !(chMotCou 
		   = ChaineLemme( &(ptLemme->ValTerme.ArcsFils) ) ) )
	       {
		 TxtErreur( 239 , "LireLemmesTagges(alphramo)" );
		 return _False;
	       }
	   }

           if( derives || racine )
           {
	     /* chargenent des regles du mot derive ou de la racine */
	     ptRegle = NULL ;
	     ChercheReglesLieesAuLemmeTagge( chMotCou , 
					     ptLemme->ValTerme.Ref ,  
					     &(ptLemme->NbRegles) , &ptRegle );
	     ReporteReglesSurReglesPhrase( ptRegle );
	   }

	   /* lecture de la fermeture de l'enumeration */
	   if( derives && *p_carlu == ']'  ) 
           {
               if( !ReconnaitreChaine( p_ptFicEntree , "]]" , p_carlu ) )
	       {
                   return _False;
               }
               SautBlanc( p_ptFicEntree , p_carlu );
               derives = _False;
           }
           else if( racine && *p_carlu == '}'  )
           {
               if( !ReconnaitreChaine( p_ptFicEntree , "}}" , p_carlu ) )
               {
                   return _False;
               }
               SautBlanc( p_ptFicEntree , p_carlu );
               racine = _False;
           }
       }
       /* boucle sur les derives associes a un lemme */
    }
    /* boucle sur les lemmes associes a un mot flechi */

    return _True;
}

/*****************************************************************************
                       lecture de tous les mots d'une phrase
                  Cas d'une phrase taggee dans un corpus pretraite
*****************************************************************************/

char     LirePhraseTagge( char *p_carlu , FILE *p_ptFicEntree )
{
    char            derMot;
    TabMotMem       ptMot;
    TabMotMem      *hdlMotPrec;

    NbMots = 0;
    MotsPhrase = NULL;
    hdlMotPrec = &MotsPhrase;

    derMot = _False;

    SautBlanc( p_ptFicEntree , p_carlu );
    while( *p_carlu != '\'' && *p_carlu != EOF )
    {
         *p_carlu = LireCar(p_ptFicEntree);
    }

    if( *p_carlu == EOF )
    {
         return _True;
    }

    /* lecture des mots jusqu'a la fin de la phrase                 */
    do
    {
        if( *p_carlu == EOF )
        {
            return _True;
        }

	ptMot = NouveauMot();
	*hdlMotPrec = ptMot;
	hdlMotPrec = &(ptMot->MotFrere);

        if( !LireChaineAvecBalise( p_ptFicEntree , p_carlu , 
                                   ptMot->ChaineMot ,
			           _MaxLetMot - 1 , '\'' , _True ) )
	{
             printf( "String: [%s]\n" , ptMot->ChaineMot );
             TxtErreur( 193 , "LirePhraseTagge(alphramo)" );
        }
        if( Trace )
        {
            fprintf( FicLog , "Word: [%s]\n" , ptMot->ChaineMot );
        }

        if( !LireLemmesTagges( p_carlu , ptMot , &derMot ,
                               p_ptFicEntree ) )
        {
             printf( "Lemma: [%s]\n" , ptMot->ChaineMot );
             TxtErreur( 240 , "LirePhraseTagge(alphramo)" );
        }
        NbMots++;
    }
    while( !derMot && *p_carlu != EOF );
    /* lecture des mots jusqu'a la fin de la phrase                 */

    NbLignesTot++;

    return _True;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            fin de alphramo.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
