/*
 * FASt Term Recognizer
 *
 *  FASTR/alppal.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for evaluating the command line
 */

#include "unistd.h"
#include "al-ext.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*************************************************************************
                  ALLOCATIONS DYNAMIQUES DE TERMES
*************************************************************************/
int                         NbMots;
TabMotMem                   MotsPhrase;

long                        NbLignesTot;

                            /* les mots reconnus de la phrase  */
TabMotMem                   PremierMotLibre;

                            /* les Lemmes allouables   */
TabLemmeMem                 PremierLemmeLibre;

                            /* les refs multiples (self & sem) */
TabRefPhraseMem             PremierRefPhraseLibre;

                            /* les Termes allouables   */
TabTermeMem                 PremierTermeLibre;

                            /* les Dag allouables   */
TabDagMem                   PremierDagLibre;

                            /* les metaregles allouables   */
TabRegleMemMeta             PremierRegleLibreMeta;

                            /* les regles allouables   */
TabRegleMem                 PremiereRegleLibre;

TabRegleMem                 ReglesMem;
int                         NbReglesMem;

TabLiensIdentTerme          LienIdentTerme;  /* lien ident/terme en compil*/
int                         NbLiensIdentTerme;

TabLienMemFic               LiensNoeudsMemFic;    
                                       /* lien entre adresse mem et ptfic  */
int                         NbLiensNoeudsMemFic;  
                                       /* nbr liens dags mem / fichier     */

TabLienMemVar               LiensNoeudsMemVar;    
                                       /* lien pt noeuds et No de variable */
int                         NbLiensNoeudsMemVar;  
                                       /* utilise en affich et lecture dag */

TabNoeudSauve               LiensNoeudsCopies;    
                                       /* lien pt noeuds et noeuds copies */
int                         NbLiensNoeudsCopies;  
                                       /* utilise en copie de dags et arbres */

int                         DebutMotsReconnus;        
int                         FinMotsReconnus;        
                                       /* utilise fin d'analyse avant affic */

TabCaracBouc                TBoucEnCours;         
                                       /* boucle en cours ds 1 anal d'ER */
int                         NbBoucEnCours;        
                                       /* nb de boucles imbr en cours    */

/*****************************************************************************
    LA GESTION DES POINTEURS VERS LES DAGS SUPPRIMES PAR UNIFICATION
*****************************************************************************/
TabDagMem                   PtDagJeteU;           
                                /* un dag jete par unification   */
TabDagMem                   PtDagGardeU;          
                                /* le dag conserve a la place de PtDagGardeU */
int                         NbRestaurationsU;     
                                /* nb de m a j de PtDagJeteU en PtDagGardeU */
TabDagMem                   PtOperLogU;           
                                /* liste des operations d'unif a faire */
TabDagMem                  *HdlDagU;              
                                /* dags sur lesquels porte l'unif */
TabTermeMem                *HdlTermeU;            
                                /* arbre des termes portant les dags unifies */
TabBlocInitial              PtBlocU;              
                                /* liste des blocs portant les dags unifies */

/*****************************************************************************
               LE TEXTE LU
*****************************************************************************/
/* 1:le buffer du texte lu                                                  */
/* 2:indice du caractere du fichier texte dans le buffer ( indice relatif ) */
char                        *BuffTexte;
char                        *BuffSortie;
int                         EstEntreeSortieBuffer;
long                        IndCar;
long                        IndCarM;
long                        IndCarSortieM;

/*************************************************************************
                LE DICTIONNAIRE DE TRANSFERT DES CATEGORIES EN
                       GENERATION AUTOMATIQUE DE REGLES
*************************************************************************/
/**                 variable garnie dans le set up                    **/
/* struct EntreeDicoConversion                DicoTransfert[ _DerCatLex + 2 ];*/

/*************************************************************************
                             LE BARBRE
*************************************************************************/
long                       *TPremIndB_Arb;  
                        /* indice de chaque niveau du BArb */
long                       *TNbTermesB_Arb; 
                        /* nombre de termes de chaque niveau du BArbre */

TabPorteSuperieure          B_ArbSup;
TabPorteNormale             B_ArbInf;
ChIdent                     DernLemme;
ChIdent                     PremLemme;

int                         ProfondeurBArbre; 
                        /* nombre de niveaux du B-Arbre */
int                         NbPortesBArbre;   
                        /* nombre de portes a chaque noeud du B-Arbre */
long                        FrequenceMAJBArbre; 
                        /* intervalle de remise a jour du B-Arbre en compil */
long                        TailleBArbInf;
long                        TailleBArbSup;

/*************************************************************************
                             DIVERS
*************************************************************************/
TabRegleMem                 ReglesPhrase;       
                                       /* les regles des mots reconnus       */
struct TeteRegleMeta        Tableaux[ _NbMaxTableau ]; 
                                       /* les tableaux de metaregles */
FILE                        *PtFicRegles;

char                        Reponses[ 2 *(_NbLigneEcran - 6) ],
                            Rubriques[ 2 *(_NbLigneEcran - 6) ] [ _NbLColEcran ],
                            Titre[ 256 ];

/*************************************************************************
                       DICTIONNAIRE BINAIRE
*************************************************************************/
/* gestion du fichier de stockage des caracteristiques des
   elements lexicaux */
/* 1 : le fichier */
/* 2 : l'indice absolu minimal et maximal dans le buffer
   (en dehors desquels il faut aller se servir dans le fichier pour trouver
   l'element). les indices minimaux servent a deduire l'indice relatif en les
   soustrayant a l'indice absolu */
FILE                       *FicBDicRegles;
FILE                       *FicBDicLemmes;
FILE                       *FicBDicLiens;
FILE                       *FicBDicLiensAux;
unsigned long               IndBuffLemmeMin;
unsigned long               IndBuffLemmeMax;
unsigned long               IndBuffRegleMin;
unsigned long               IndBuffRegleMax;
char                        NomFicDico[ _NbMaxLetNomFic ];
char                        NomFicMessages[ _NbMaxLetNomFic ];
char                        NomFicExtUnifier[ _NbMaxLetNomFic ];
char                        NomFicUnification[ _NbMaxLetNomFic ];
char                        NomFicAux[ _NbMaxLetNomFic ];
FILE                       *FicUnification;

/* un buffer pour relever une unite lexicale + regles en une seule lecture */
TabChar                     BufferLemmesCourant;
TabChar                     BufferReglesCourant;
TabChar                     BufferLemmesCourt;
TabChar                     BufferReglesCourt;
TabChar                     BufferLemmesLong;
TabChar                     BufferReglesLong;
TabChar                     BufferLiensMorphologiquesEtSemantiques;
long                        TailleBuffStd;     
                                       /* taille buffer d'ES stdrd (txt)    */
long                        TailleBuffLemmesCourt;  
                                       /* taille buffer d'ES stdrd (lemmes) */
long                        TailleBuffReglesCourt;  
                                       /* taille buffer d'ES stdrd (regles) */
long                        TailleBuffLemmesLong;  
                                       /* taille buffer d'ES stdrd (lemmes) */
long                        TailleBuffReglesLong;  
                                       /* taille buffer d'ES stdrd (regles) */
long                        TailleBuffLemmesCourant;  
                                       /* taille buffer d'ES en cours(lemmes)*/
long                        TailleBuffReglesCourant;  
                                       /* taille buffer d'ES en cours(regles)*/
int                         BDicSurDisque;   
                                       /* booleen faux en cas de chrgt
                                          du dico en memoire */
int                         BDicReduit;  
                                       /* dictionnaire ou sont
                                          supprimes les chemins lemma
                                          et inflection */

/*************************************************************************
        LE DICTIONNAIRE DES MOTS COURTS OU USUELS
*************************************************************************/
struct EntreeDucs           *Ducs;

/*************************************************************************
                FICHIER GLOBAUX ET BOOLEENS DE SETUP
*************************************************************************/
ChIdent                     Versionfastr;      /* version of fastr */
ChIdent                     Datefastr;         /* date of fastr */
char                        NomFicLog[ _NbMaxLetNomFic ];
char                        NomFicHeader[ _NbMaxLetNomFic ];
FILE                       *FicBufferiseCourant; 
                           /* fichier d'entree bufferise*/
FILE                       *FicLog;     
                           /* fichier des messages       */
FILE                       *FicRefIdent;    
                           /* fichier table des identificateurs de
                              regles + ref */
FILE                       *FicRefMots;     
                           /* fichier table des mots + ref */
int                         VerificationAccord;
int                         PretraitementDesNombres;
int                         OuvreAutomMetaregl;
int                         OuvreAutomReglesMem;
int                         AriteMin;  
                           /* arite en-dessous de laquelle les regles
                              ne sont pas tenteees */
int                         AriteMax;  
                           /* arite au-dessus de laquelle les regles
                              ne sont pas tenteees */
int                         FormatDEntree;  
                           /* type des donnees en indexation */
int                         FormatDeSortie;  
                           /* type des donnees en indexation */
int                         TraceTexteEnSortie; 
                           /* ecriture du texte lu en analyse */
int                         HeaderEnSortie; 
                           /* ecriture du texte lu en analyse */
long                        NoPremLigne;  
                           /* no de la premiere notice en sortie Inist */
int                         ArretSurErreurCompil;
int                         ClasstReglesAutom;  
                           /* resequentialisation autom. des regles en
                              fin de compil */
int                         LinkAutom;  
                           /* ed autom de liens en fin de compil */
int                         LexicalisationMultiple;  
                           /* lexicalisation autom d'1 regle sur ses lemmes */
int                         TraitementBatch;  
                           /* insertion dans un shell */
int                         MinusculisationAutomatique;  
                           /* Passage automatique en minuscules */
int                         SegmentationAutomatique;  
                           /* Segmentation automatique des mots */
int                         LexicalisationImperativeDesRegles;
int                         AutomaticSelfMorphologicalRoot;  
                           /* automatic addition of morph. root <self> */
int                         AutomaticSelfSemanticLinks;  
                           /* automatic addition of semantic links to <self> */
char                        ChErreurCompil[ 2048 ];
char                        ChAux[ 2048 ];
int                         Trace;

/*************************************************************************
                            VARIABLES D'ANALYSE
*************************************************************************/
/* un indice, indiquant eventuellement l'indice du mot courant de la
  la sequence en cours de reconnaissance                               */
int                         IndiceMotCou;

/* la pile d'effacement avec son pointeur et l'indice de haut de pile  */
TabChar                     Pile;
unsigned long               HautPile;
unsigned long               TaillePileEffacement;

int                         NbMaxGenerations;

/*************************************************************************
            VARIABLES GLOBALES POUR LA GESTION DES FLEXIONS
*************************************************************************/
/**    LISTES DE SUFFIXES A AJOUTER AUX RACINES POUR AVOIR LES FLEXIONS    **/
struct UnNoeudSuff          NoeudsSuff[ _NbMaxNoeudsSuffixe ];
int                         NbNoeudsSuff;    /* prochainn noeud a allouer   */
/**        LES SUFFIXES PREDEFINIS                                         **/
struct Suffixe              *Suffixes;

/**        LE DICTIONNAIRE DE TRAITS ET CATEGORIES LEXICALES           **/
TabDicoTraitsCatLex         DicoTraitsCatLex;

/**        LES CHEMINS PREDEFINIS                                      **/
TabDagMem                   CheminCategorie;
TabDagMem                   CheminEtiquette;
TabDagMem                   CheminRacinesAnnexes;
TabDagMem                   CheminLexicalisation;
TabDagMem                   CheminLemme;
TabDagMem                   CheminFlexion;
TabDagMem                   CheminDerivation;
TabDagMem                   CheminForme;
TabDagMem                   CheminNombre;
TabDagMem                   CheminReference;
TabDagMem                   CheminEtiquetteMeta;
TabDagMem                   CheminRacine;
TabDagMem                   CheminLeadingExprReg;
TabDagMem                   CheminTrailingExprReg;
TabDagMem                   CheminsSemantique[ _NbMaxTypesLiensSem ];
TabDagMem                   CheminSelf;

/**        LES CATEGORIES PREDEFINIES                                  **/
long                        RefCatMotsInc;
long                        RefDeterminantNumeral;
long                        RefPonctuation;
long                        RefPonctuationFinale;

/**        LES CATEGORIES A DISTINGUER EN MORPHO ROBUSTE               **/
long                        RefMajorCategories[ _MaxMajorCategories ];
int                         NbMajorCategories;
long                        RefCatLexVide; /* la val de la cat lex vide X */

/**        LES ATOMES PREDEFINIS                                       **/
long                        RefSingulier;
long                        RefPluriel;

/**        LES FLEXIONS PREDEFINES                                     **/
TabDagMem                   TabDagsSuffixage[ _NbMaxSuffixages ];
TabDagMem                   TabListeMetaReglesStd[ _NbMaxTableau ];

/**        LES CARACTERES SPECIAUX                                     **/
char                        CarSpeciauxMots[ _MaxCarSpeciaux ];
char                        Ponctuation[ _MaxCarSpeciaux ];
char                        PonctuationFinale[ _MaxCarSpeciaux ];

/*************************************************************************
                              tailles des tableaux
*************************************************************************/
int                         NbMaxReglesMem;  
                           /* nb max de regles memoire          */

long                        NbMaxElemAllouables[ _DerTypeElemAllouable ];
/*int                         NbMaxMotPhrase;  */
/*                           /* nombre maximal de mots lus ds 1 ph*/
/*                             NbMax[ _RefPhrase ]; */
/*                            /* nombre maximal de references de mots */
/* 			      (semantiques) dans une phrase */
/*                             NbMax[ _Regles ];      */
/*                            /* nb max de regles allouables       */
/*                             NbMax[ _Termes ];      */
/*                            /* nb max de termes de regles alloues*/
/*                             NbMax[ _ReglesMeta ];  */
/*                            /* nb max de metaregles allouables   */
/*                             NbMax[ _Lemmes ];      */
/*                            /* nb max de lemmes de regles alloues*/
/*                             NbMax[ _Dag ];         */
/*                            /* nb max de fils de arcs alloues    */

                            /* nb de fois ou on a du reallouer un type donne */
int                         NbReallocations[ _DerTypeElemAllouable ];

int                         NbMaxLiensMemorises; 
                           /* nb max de liens memorises     */

int                         NbMaxTraitsCatLex; 
                           /* n max d'ident de cat lex ou traits*/
int                         NbMaxSuffixes;     
                           /* nb max de suffixes de mots simples*/
int                         NbMaxTermesDansUneRegle; 
                           /* nb max de termes dans une regle (pour
                              ref dans chemins) */
int                         NbMaxChainesTermes;
                           /* nb max de chaines de termes ds la liste
                              de reference*/
int                         NbMaxLiens;        
                           /* nb max de liens acquis par var. */

/*************************************************************************
        VARIABLES GLOBALES DE LA MORPHOLOGIE
*************************************************************************/
int                         SuffStripping;     
                           /* suppression uniforme des suffixes */
int                         TailSuffStripping; 
                           /* taille de la suppression uniforme des suffixes */
int                         MorphoFlex;     
                           /* morphologie flexionnelle interne */
int                         MorphoDer;     
                           /* morphologie derivationnelle interne */
int                         ExploitationLiensMorpho;     
                           /* exploitation des liens morphologiques compiles */
int                         ExploitationLiensSem;     
                           /* exploitation des liens semantiques compiles */

/*************************************************************************
                        FONCTION PPALE
*************************************************************************/
int      main( int argc , char *argv[] )
{
    int                i;
    int                j;
    int                indTypeElemAllouable;
    char               setUpLu = _False;
    int                traceCou;
    int                traceForcee = _False;
    char              *argvLocal[ 16 ];
    char              *chFicConf;

   FicLog = NULL; 
   _IniEcran();

   if( argc > 16 ) {
    fprintf( stderr , "Usage: fastr [-ciIlLmpstVxXz] argts [-C config-file] [-v]\n" );
    exit( 0 );
   }

#ifdef SUN
   for( i = 0 ; i < argc ; i++ ){
     argvLocal[ i ] = (char *)_Alloc( 128 );
     strcpy( argvLocal[ i ] , argv[ i ] );
   }
   if( argc > 1 ) {
     for( i = 0 ; i < argc ; i++ ){
       /* set-up file */
       if( strcmp( argvLocal[ i ] , "-C" ) == 0 ) {
	 if( i < argc - 1 && *(argvLocal[ i + 1 ]) != '-' ) {
	     if( !LireSetUp( argvLocal[ i + 1 ] ) ) {
	       exit( 0 );
	     }
	     setUpLu = _True;
	     for( j = i ; j < argc - 2 ; j++ ) {
	       strcpy( argvLocal[ j ] , argvLocal[ j + 2 ] );
	     }
	     argc = argc - 2;
	     i--;
           }
           else {
	     fprintf( stderr , "Usage: fastr -C config-file\n" );
	     exit( 0 );
           }
        }
	/* trace, verbose mode */
	else if( strcmp( argvLocal[ i ] , "-v" ) == 0 ) {
	  traceForcee = _True;
	  for( j = i ; j < argc - 1 ; j++ ) {
	    strcpy( argvLocal[ j ] , argvLocal[ j + 1 ] );
	  }
	  argc = argc - 1;
	  i--;
	}
      }
    }           
#endif
    if( !setUpLu )
    {
      chFicConf = getenv( "FASTREMPTY" );
      if( chFicConf ) {
	if( !LireSetUp( chFicConf ) ){
	  exit( 0 );
	}
      }
      else {
	if( !LireSetUp( "/etc/fastr.conf-empty" ) ){
	  exit( 0 );
	}
      }
    }

    if( traceForcee ) {
      Trace = _True;
    }

    IniTabIndEtFichiers(); /* a appeler avant IniGrandsTableaux qui 
                                 utilise la, taille du b-arbre */
    IniGrandsTableaux( _False );
    for( indTypeElemAllouable = 0 ; indTypeElemAllouable < _DerTypeElemAllouable ; 
	 indTypeElemAllouable++ ) {
      NbReallocations[ indTypeElemAllouable ] = 0;
      IniElemsAllouables( indTypeElemAllouable );
    }

    traceCou = Trace;
    /* Trace = _False; */
    _EnTete( _True );
    _ChargeFichierDicoSelectionne();
    
    if( !OuvreDictionnaire( _True ) )
    {
         if( !TraitementBatch )
         {
             /* demande le nom du fichier dictionnaire 
                de facon interactive si non trouve    */
             /* en mode batch, garde le nom du set-up */
             TxtLireChaine( "Nom du nouveau fichier dictionnaire" , 0 , 
                            NomFicDico , 24 , _Alphabetique );
        }
         if( !BootstrapDictionnaire( _True , _True ) )
         {
             *NomFicDico = '\0';
             TxtErreur( 22 , "main(alppal)" );
             exit(0);
         }
    }

    LireBArbre();
    if( !LitEnTeteGrammaireEtErreur() )
    {
        TxtAlerte( 15 , 1 );
    }

    Trace = traceCou;
    if( _MachineGraphique )
    {
        _BoucleAttenteEvenement();
    }
    else
    {
        if( !TraitementBatch )
        {
            MenuPpal();
        }
        else
        {
            Choix( argc , argvLocal );
        }
    }
    FermeDictionnaire();
    return( 1 );
}

/*************************************************************************
                    Menu principal
*************************************************************************/
void     MenuPpal( void )
{
     int                  i;

     do
     {
        sprintf( Titre , "Choix [%s]" , NomFicDico );
#ifdef GENERATION
        strcpy( Rubriques[0] ,"[g] Menu generation");
        strcpy( Rubriques[1] ,"[c] Menu acquisition");
        strcpy( Rubriques[2] ,"[l] Menu liens");
        strcpy( Rubriques[3] ,"[a] Menu des applications auxiliaires");
        strcpy( Rubriques[4] ,"[s] Menu statistiques");
        strcpy( Rubriques[5] ,"[m] Menu morphologie\n");
        strcpy( Rubriques[6] ,"[e] Menu Erli\n");
#else
        strcpy( Rubriques[0] ,"");
        strcpy( Rubriques[1] ,"");
        strcpy( Rubriques[2] ,"");
        strcpy( Rubriques[3] ,"");
        strcpy( Rubriques[4] ,"");
        strcpy( Rubriques[5] ,"");
        strcpy( Rubriques[6] ,"");
#endif
        strcpy( Rubriques[7] ,"[t] Menu Travail" );
        strcpy( Rubriques[8] ,"[u] Menu Utilitaires");
        strcpy( Rubriques[9] ,"[,] Fin" );
#ifdef GENERATION
        strcpy( Reponses , "gclasmetu," );
#else
        strcpy( Reponses , "       tu," );
#endif

        i = TxtMenu( Titre , Rubriques , Reponses );
        switch(i)
        {
#ifdef GENERATION
        case 0: MenuGeneration();
            break;
        case 1: MenuAcquisition();
            break;
        case 2: MenuLiens();
            break;
        case 3: MenuApplicationsAux();
            break;
        case 4: MenuStats();
            break;
        case 5: MenuMorpho();
            break;
        case 6: MenuErli();
            break;
#else
        case 0: 
            break;
        case 1: 
            break;
        case 2: 
            break;
        case 3: 
            break;
        case 4: 
            break;
        case 5: 
            break;
        case 6: 
            break;
#endif
        case 7: MenuTravail();
            break;
        case 8: MenuUtilitaires();
            break;
        case 9:
           return;
        default : break;
        }
     }
     while ( _False != _True );
}

/*************************************************************************
                    Menu dit Travail
**************************************************************************/
void     MenuTravail( void )
{
     int                         i;
     int                         nbReglesFausses;
     long                        nbReglesMots;
     long                        nbReglesTermes;
     char                        chaine[ 256 ];
     char                        chaineSortie[ 256 ];

     do
     {
        sprintf( Titre , "Menu Travail [%s]" , NomFicDico );
        strcpy( Rubriques[0] ,"");
        strcpy( Rubriques[1] ,"[s] Compilation regles fichier Sans Insertion");
        strcpy( Rubriques[2] ,"[k] Compilation regles fichier");
        strcpy( Rubriques[3] ,"[c] Classe regles");
        strcpy( Rubriques[4] ,"[l] Edition de liens");
        strcpy( Rubriques[5] ,"[v] Indexation sur variations");
        strcpy( Rubriques[6] ,"[e] Indexation sur variations exclusivement" );
        strcpy( Rubriques[7] ,"[i] Indexation" );
        strcpy( Rubriques[8] ,"[y] Analyse MorphoSyntaxique" );
        strcpy( Rubriques[9] ,"[o] Collocations" );
        strcpy( Rubriques[10] ,"" );
        strcpy( Rubriques[11] ,"[r] Nouveau rattachement des regles" );
        strcpy( Rubriques[12] ,"[t] Etiquetage d'un texte" );
        strcpy( Rubriques[13] ,"[,] Fin" );
        strcpy( Reponses , " skclveiyo rt," );

        i = TxtMenu( Titre , Rubriques , Reponses );
        RAZIndicesBuffersDic();

        switch(i)
        {
            case 1 :
                TxtLireChaine( "Nom du fichier de regles a compiler" , 
                               0 , chaine , 50 , _Alphabetique );
                if( !CompilationRegles( chaine , NULL , NULL , 
                                        _False , _True , 
                                        _False , &nbReglesFausses ,
                                        &nbReglesMots , &nbReglesTermes ) )
                {
                    TxtAlerte( 4 , 1 );
                }
                break;
            case 2 :
                TxtLireChaine( "Nom du fichier de regles fichier a compiler" , 
                               0 , chaine , 50 , _Alphabetique );
                if( !CompilationRegles( chaine , NULL , NULL , 
                                        _True , _False , 
                                        _False , &nbReglesFausses ,
                                        &nbReglesMots , &nbReglesTermes ) 
                    && ArretSurErreurCompil )
                {
                    TxtAlerte( 4 , 1 );
                }
                if( nbReglesMots > 0 )
                {
                    EditionLiensMorphologiquesEtSemantiques();
                }
                if( nbReglesTermes > 0 && ClasstReglesAutom )
                {
                    ClasseRegles( _False );
                }
                if( nbReglesTermes > 0 && LinkAutom )
                {
                    EditionLiensTermesAppel();
                }
                break;
            case 3 :
                ClasseRegles( _False );
                break;
            case 4 :
                EditionLiensTermesAppel();
                break;
            case 5 :
                TxtLireChaine( "Nom du fichier a indexer sur variations" , 0 ,
                               chaine , 50 , _Alphabetique );
                TxtLireChaine( "Nom du fichier de resultats" , 0 ,
                               chaineSortie , 50 , _Alphabetique );
                IndexationOuAnalyse( chaine , chaineSortie , _IndexationVariations );
                break;
            case 6 :
                TxtLireChaine( "Nom du fichier a indexer sur variations exclusivement" , 0 ,
                               chaine , 50 , _Alphabetique );
                TxtLireChaine( "Nom du fichier de resultats" , 0 ,
                               chaineSortie , 50 , _Alphabetique );
                IndexationOuAnalyse( chaine , chaineSortie , _Variations );
                break;
            case 7 :
                TxtLireChaine( "Nom du fichier a indexer" , 0 ,
                               chaine , 50 , _Alphabetique );
                TxtLireChaine( "Nom du fichier de resultats" , 0 ,
                               chaineSortie , 50 , _Alphabetique );
                IndexationOuAnalyse( chaine , chaineSortie , _Indexation );
                break;
            case 8 :
                TxtLireChaine( "Nom du fichier a analyser" , 0 ,
                               chaine , 50 , _Alphabetique );
                TxtLireChaine( "Nom du fichier de resultats" , 0 ,
                               chaineSortie , 50 , _Alphabetique );
                IndexationOuAnalyse( chaine , chaineSortie , _Analyse );
                break;
            case 9 :
                TxtLireChaine( "Nom du fichier ou relever les collocations" , 0 ,
                               chaine , 50 , _Alphabetique );
                TxtLireChaine( "Nom du fichier de resultats" , 0 ,
                               chaineSortie , 50 , _Alphabetique );
                IndexationOuAnalyse( chaine , chaineSortie , _Collocations );
                break;
            case 10 :
                break;
            case 11 :
                strcpy( chaine , NomFicDico );
                strcat( chaine , ".OPT" );
                RattacheRegles( chaine );
                break;
            case 12 :
                TxtLireChaine( "Nom du fichier a etiqueter" , 0 ,
                               chaine , 50 , _Alphabetique );
                TxtLireChaine( "Nom du fichier de resultats" , 0 ,
                               chaineSortie , 50 , _Alphabetique );
                IndexationOuAnalyse( chaine , chaineSortie , _Etiquetage );
                break;
            case 13: return;
            /* case :
                /* TxtLireChaine( "Nom du fichier de non terminaux a generer" , 0 , 
                                  chaine , 50 , _Alphabetique ); */
                /* GenerationParFichier( chaine ); */
                /* GenerationParDialogue(); */
            default : break;
        }
     }
     while (_True != _False);

}

/*************************************************************************
                    Menu dit Utilitaires
**************************************************************************/
void     MenuUtilitaires( void )
{
     int                         i;
     char                        chaine1[ 256 ];
     char                        trouvee;
     ChIdent                     chMot;

     do
     {
        sprintf( Titre , "Menu Utilitaires [%s]" , NomFicDico );
        strcpy( Rubriques[0] ,"[d] RAZ Dictionnaire" );
        sprintf( Rubriques[1] ,"[r] RAZ Regles Memoire" );
        strcpy( Rubriques[2] ,"[e] RAZ Metaregles");
        strcpy( Rubriques[3] ,"");
        strcpy( Rubriques[4] ,"[b] Mise a jour B-Arbre");
        strcpy( Rubriques[5] ,"[h] Recherche Mots Simple (B-Arbre)");
        strcpy( Rubriques[6] ,"[s] Ecriture Mots Simples");
        strcpy( Rubriques[7] ,"[&] Ecriture Mots & Regles");
#ifdef RATTACHEMENT
        strcpy( Rubriques[8] ,"[!] Recherche du Rattachement optimal");
#else
        strcpy( Rubriques[8] ,"[!] Ecriture Mots & Regles pour Rattachement optimal");
#endif
        strcpy( Rubriques[9] ,"[o] Ecriture Regles d'un Mot");
        strcpy( Rubriques[10] ,"[-] Ecriture Meta-Regles" );
        strcpy( Rubriques[11] ,"[c] Changement de Dictionnaire");
        strcpy( Rubriques[12] ,"[n] Nouveau Dictionnaire" );
        strcpy( Rubriques[13] ,"" );
        sprintf( Rubriques[14] ,"[t] Trace (%d)" , Trace );
        strcpy( Rubriques[15] ,"[p] Photo memoire");
        strcpy( Rubriques[16] ,"");
        strcpy( Rubriques[17] ,"[,] Fin" );
        strcpy( Reponses , "dre bhs&!o-cn tp ," );

        i = TxtMenu(Titre , Rubriques , Reponses );
        RAZIndicesBuffersDic();
        switch(i)
        {
            case 0 :
                if( !BootstrapDictionnaire( _False , _False ) )
                {
                    TxtErreur( 22 , "main(alppal)" );
                }
                break;
            case 1 :
                RAZReglesMemoire();
                break;
            case 2 :
                RAZReglesMeta();
                break;
            case 3 :
                break;
            case 4 :
                IniBArbre();
                break;
            case 5 :
                TxtLireChaine( NULL , 50 , chMot, _MaxLetIdent , _Alphabetique );
                RechercheBArbre( chMot , _True );
                IndiceInsertionDicLemmes( chMot , &trouvee );
                if( trouvee )
                {
                     fprintf (FicLog , "Lemme trouve\n\n" );
                }
                else
                {
                     fprintf (FicLog , "Lemme non trouve\n\n" );
                }
                break;
            case 6 :
                MapReferencesTermesSimples( _False , _False );
                break;
            case 7 :
                MapReferencesTermesSimples( _True , _False );
                break;
            case 8 :
                MapReferencesTermesSimples( _True , _True );
#ifdef RATTACHEMENT
                strcpy( chaine1 , NomFicDico );
                strcat( chaine1 , ".REF" );
                strcpy( chaine2 , NomFicDico );
                strcat( chaine2 , ".OPT" );
                BatchRattachement( chaine1 , chaine2 , 'r' );
#endif
                break;
            case  9 :
                TxtAlerte( 14 , 1 );
                TxtLireChaine( NULL , 12 , chMot, _MaxLetIdent , _Alphabetique );
                EcrireReglesDUnMot( chMot , stdout , _Sortie );
                break;
            case 10 :
                EcrireListeRegleMeta();
                break;
            case 11 :
                TxtLireChaine( "Nom du fichier dictionnaire" , 0 , 
                               chaine1 , 50 , _Alphabetique );
                ChangeDictionnaire( chaine1 );
                break;
            case 12 :
                TxtLireChaine( "Nom du nouveau fichier dictionnaire" ,
                               0 , chaine1 , 50 , _Alphabetique );
                if( strcmp( chaine1 , NomFicDico ) == 0 )
                {
                    break;
                }
                strcpy( NomFicDico , chaine1 );
                if( !BootstrapDictionnaire( _True , _False ) )
                {
                    *NomFicDico = '\0';
                    TxtErreur( 22 , "main(alppal)" );
                    exit(0);
                }
                break;
            case 13 :
                break;
            case 14 :
                Trace = 1 - Trace;
                break;
            case 15 :
                PhotoMemoire( FicLog );
                break;
            case 16 :
                break;
            case 17 : return;
            default : break;
        }
     }
     while (_True != _False);

}

/*************************************************************************
                        INTEGRATION UNIX
*************************************************************************/
/* exemples de ligne de commande :
   > al -k REGLES
   > al -l
   > al -v METAL.TXT INDEXATION.TXT
   > al -v stdin stdout
   > al -v METAL.TXT stdout */

void     Choix( int p_argc , char *p_argvLocal[] )
{
  int                         nbReglesFausses;
  long                        nbReglesMots;
  long                        nbReglesTermes;
  
  /* no args */
  if( p_argc == 1 ) {
    fprintf( stderr , "Usage: fastr [-ciIlLmpstVxXz] argts [-C config-file] [-v]\n" );
    return;
  }
  /* version */
  else if( strcmp( p_argvLocal[ 1 ] , "-V" ) == 0 ) { 
    if( p_argc != 2 ) {
      fprintf( stderr , "Usage: fastr -V\n" );
      return;
    }
    printf( 
"fastr 2.04/04-02-01: Term Processing (c)LIMSI-CNRS [jacquemin@limsi.fr]\n" 
	   );
  }
  /* compiling */
  else if( strcmp( p_argvLocal[ 1 ] , "-c" ) == 0 ) { 
    if( p_argc != 3 )
      {
	fprintf( stderr , "Usage: fastr -c rule-file\n" );
	return;
      }
      if( !CompilationRegles( p_argvLocal[ 2 ] , 
			    NULL , NULL , _True , _False , 
			    _False , &nbReglesFausses ,
			    &nbReglesMots , &nbReglesTermes ) 
	&& ArretSurErreurCompil )
      {
	TxtAlerte( 4 , 1 );
      }
      if( nbReglesMots > 0 )
      {
	EditionLiensMorphologiquesEtSemantiques();
      }
      if( nbReglesTermes > 0 && ClasstReglesAutom )
      {
	ClasseRegles( _False );
      }
      if( nbReglesTermes > 0 && LinkAutom )
      {
	EditionLiensTermesAppel();
      }
  }
  /* indexing */
  else if( strcmp( p_argvLocal[ 1 ] , "-i" ) == 0 ) { 
    if( p_argc != 2 && p_argc != 3 )
      {
	fprintf( stderr , "Usage: fastr -i corpus-file?\n" );
	return;
      }
    if( p_argc == 2 ) {
      IndexationOuAnalyse( "stdin" , "stdout" , 
			   _IndexationVariations );
    }
    else {
      IndexationOuAnalyse( p_argvLocal[ 2 ] , "stdout" , 
			   _IndexationVariations );
    }
  }
  /* projecting lexical information on the corpus */
  else if( strcmp( p_argvLocal[ 1 ] , "-p" ) == 0 ) { 
    if( p_argc != 2 && p_argc != 3 )
      {
	fprintf( stderr , "Usage: fastr -p corpus-file?\n" );
	return;
      }
    if( p_argc == 2 ) {
      IndexationOuAnalyse( "stdin" , "stdout" , _EtiquetageFin );
    }
    else {
      IndexationOuAnalyse( p_argvLocal[ 2 ] , "stdout" , 
			   _EtiquetageFin );
    }
  }
  /* indexing (exclusively variants) */
  else if( strcmp( p_argvLocal[ 1 ] , "-X" ) == 0 ) { 
    if( p_argc != 2 && p_argc != 3 )
      {
	fprintf( stderr , "Usage: fastr -X corpus-file?\n" );
	return;
      }
    if( p_argc == 2 ) {
      IndexationOuAnalyse( "stdin" , "stdout" , _Variations );
    }
    else {
      IndexationOuAnalyse( p_argvLocal[ 2 ] , "stdout" , _Variations );
    }
  }
  /* indexing (exclusively terms) */
  else if( strcmp( p_argvLocal[ 1 ] , "-x" ) == 0 ) { 
    if( p_argc != 2 && p_argc != 3 )
      {
	fprintf( stderr , "Usage: fastr -x corpus-file?\n" );
	return;
      }
    if( p_argc == 2 ) {
      IndexationOuAnalyse( "stdin" , "stdout" , _Indexation );
    }
    else {
      IndexationOuAnalyse( p_argvLocal[ 2 ] , "stdout" , _Indexation );
    }
  }
  /* indexing (parsing) */
  else if( strcmp( p_argvLocal[ 1 ] , "-I" ) == 0 ) { 
    if( p_argc != 2 && p_argc != 3 )
      {
	fprintf( stderr , "Usage: fastr -I corpus-file?\n" );
	return;
      }
    if( p_argc == 2 ) {
      IndexationOuAnalyse( "stdin" , "stdout" , _Analyse );
    }
    else {
      IndexationOuAnalyse( p_argvLocal[ 2 ] , "stdout" , _Analyse );
    }
  }
  /* resets dictionaries */
  else if( strcmp( p_argvLocal[ 1 ] , "-z" ) == 0 ) { 
    if( p_argc != 2 )
      {
	fprintf( stderr , "Usage: fastr -z\n" );
	return;
      }
    BootstrapDictionnaire( _True , _False );
  }
  /* replaces term identifiers with strings */
  else if( strcmp( p_argvLocal[ 1 ] , "-s" ) == 0 ) {  
    if( p_argc != 4 )
      {
	fprintf( stderr , "Usage: fastr -s term-file #column_ident\n" );
	return;
      }
    ChainesTermes( "stdin" , p_argvLocal[ 2 ] , "stdout" , 
		   atoi(p_argvLocal[ 2 + 1 ]) , 
		   _True , _False , _False );
  }
  /* displays lexicon (basic lemmas) */
  else if( strcmp( p_argvLocal[ 1 ] , "-l" ) == 0 ) {  
    if( p_argc % 2 == 1 )
      {
	printf( 
	       "Usage: fastr -l (inflected-category #inflection-rank)*\n" );
	return;
      }
    if( p_argc / 2 > 1 )
      {
	MapAfficheLexiques( p_argvLocal + 2 , p_argc / 2 - 1 , 
			    _True , _False , _False , _False );
      }
    else
      {
	MapAfficheLexiques( p_argvLocal + 2 , p_argc / 2 - 1 , 
			    _False , _False , _False , _False );
      }
  }
  /* displays lexicon (basic lemmas + links) */
  else if( strcmp( p_argvLocal[ 1 ] , "-L" ) == 0 ) {  
    if( p_argc != 2 )
      {
	fprintf( stderr , "Usage: fastr -L\n" );
	return;
      }
    MapAfficheLexiques( NULL , 0 , 
			_False , _True , _True , _False );
  }
  /* displays lexicon (basic lemmas + term rules) */
  else if( strcmp( p_argvLocal[ 1 ] , "-t" ) == 0 ) {  
    if( p_argc != 2 )
      {
	fprintf( stderr , "Usage: fastr -t\n" );
	return;
      }
    MapAfficheLexiques( NULL , 0 , 
			_False , _False , _False , _True );
  }
  /* term acquisition */
  else if( strcmp( p_argvLocal[ 1 ] , "-a" ) == 0 ) {  
    if( p_argc != 3 && p_argc != 4  )
      {
	fprintf( stderr , "Usage: fastr -a link-file? [CIP]+\n" );
	return;
      }
    if( p_argc == 3  ) {
      SequencesAcquisition( "stdin" , "stdout" , 
			    p_argvLocal[ 2 ] );
    }
    else {
      SequencesAcquisition( p_argvLocal[ 2 ] , "stdout" , 
			    p_argvLocal[ 2 + 1 ] );
    }	      
  }
  /* grouping of term variations */
  else if( strcmp( p_argvLocal[ 1 ] , "-g" ) == 0 ) {  
    if( p_argc != 4  )
      {
	fprintf( stderr , "Usage: fastr -g link-file? [CIP]+\n" );
	return;
      }
    GroupesVariation( p_argvLocal[ 2 ] , "" , "stdout" , 
		      p_argvLocal[ 2 + 1 ] );
  }
  /* displays metarules */
  else if( strcmp( p_argvLocal[ 1 ] , "-m" ) == 0 ) { 
    if( p_argc != 2 )
      {
	fprintf( stderr , "Usage: fastr -m\n" );
	return;
      }
    EcrireListeRegleMeta();
  }
  else {
    fprintf( stderr , "Usage: fastr [-ciIlLmpstVxXz] argts [-C config-file] [-v]\n" );
  }
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            fin de alppal.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
