/*
 * FASt Term Recognizer
 *
 *  FASTR/alppal.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for evaluating the command line
 */

#include "unistd.h"
#include "al-ext.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*************************************************************************
                  ALLOCATIONS DYNAMIQUES DE TERMES
*************************************************************************/
int                         NbMots;
TabMotMem                   MotsPhrase;

long                        NbLignesTot;

                            /* les mots reconnus de la phrase  */
TabMotMem                   PremierMotLibre;

                            /* les Lemmes allouables   */
TabLemmeMem                 PremierLemmeLibre;

                            /* les refs multiples (self & sem) */
TabRefPhraseMem             PremierRefPhraseLibre;

                            /* les Termes allouables   */
TabTermeMem                 PremierTermeLibre;

                            /* les Dag allouables   */
TabDagMem                   PremierDagLibre;

                            /* les metaregles allouables   */
TabRegleMemMeta             PremierRegleLibreMeta;

                            /* les regles allouables   */
TabRegleMem                 PremiereRegleLibre;

TabRegleMem                 ReglesMem;
int                         NbReglesMem;

TabLiensIdentTerme          LienIdentTerme;  /* lien ident/terme en compil*/
int                         NbLiensIdentTerme;

TabLienMemFic               LiensNoeudsMemFic;    
                                       /* lien entre adresse mem et ptfic  */
int                         NbLiensNoeudsMemFic;  
                                       /* nbr liens dags mem / fichier     */

TabLienMemVar               LiensNoeudsMemVar;    
                                       /* lien pt noeuds et No de variable */
int                         NbLiensNoeudsMemVar;  
                                       /* utilise en affich et lecture dag */

TabNoeudSauve               LiensNoeudsCopies;    
                                       /* lien pt noeuds et noeuds copies */
int                         NbLiensNoeudsCopies;  
                                       /* utilise en copie de dags et arbres */

int                         DebutMotsReconnus;        
int                         FinMotsReconnus;        
                                       /* utilise fin d'analyse avant affic */

TabCaracBouc                TBoucEnCours;         
                                       /* boucle en cours ds 1 anal d'ER */
int                         NbBoucEnCours;        
                                       /* nb de boucles imbr en cours    */

/*****************************************************************************
    LA GESTION DES POINTEURS VERS LES DAGS SUPPRIMES PAR UNIFICATION
*****************************************************************************/
TabDagMem                   PtDagJeteU;           
                                /* un dag jete par unification   */
TabDagMem                   PtDagGardeU;          
                                /* le dag conserve a la place de PtDagGardeU */
int                         NbRestaurationsU;     
                                /* nb de m a j de PtDagJeteU en PtDagGardeU */
TabDagMem                   PtOperLogU;           
                                /* liste des operations d'unif a faire */
TabDagMem                  *HdlDagU;              
                                /* dags sur lesquels porte l'unif */
TabTermeMem                *HdlTermeU;            
                                /* arbre des termes portant les dags unifies */
TabBlocInitial              PtBlocU;              
                                /* liste des blocs portant les dags unifies */

/*****************************************************************************
               LE TEXTE LU
*****************************************************************************/
/* 1:le buffer du texte lu                                                  */
/* 2:indice du caractere du fichier texte dans le buffer ( indice relatif ) */
char                        *BuffTexte;
char                        *BuffSortie;
int                         EstEntreeSortieBuffer;
long                        IndCar;
long                        IndCarM;
long                        IndCarSortieM;

/*************************************************************************
                LE DICTIONNAIRE DE TRANSFERT DES CATEGORIES EN
                       GENERATION AUTOMATIQUE DE REGLES
*************************************************************************/
/**                 variable garnie dans le set up                    **/
/* struct EntreeDicoConversion                DicoTransfert[ _DerCatLex + 2 ];*/

/*************************************************************************
                             LE BARBRE
*************************************************************************/
long                       *TPremIndB_Arb;  
                        /* indice de chaque niveau du BArb */
long                       *TNbTermesB_Arb; 
                        /* nombre de termes de chaque niveau du BArbre */

TabPorteSuperieure          B_ArbSup;
TabPorteNormale             B_ArbInf;
ChIdent                     DernLemme;
ChIdent                     PremLemme;

int                         ProfondeurBArbre; 
                        /* nombre de niveaux du B-Arbre */
int                         NbPortesBArbre;   
                        /* nombre de portes a chaque noeud du B-Arbre */
long                        FrequenceMAJBArbre; 
                        /* intervalle de remise a jour du B-Arbre en compil */
long                        TailleBArbInf;
long                        TailleBArbSup;

/*************************************************************************
                             DIVERS
*************************************************************************/
TabRegleMem                 ReglesPhrase;       
                                       /* les regles des mots reconnus       */
struct TeteRegleMeta        Tableaux[ _NbMaxTableau ]; 
                                       /* les tableaux de metaregles */
FILE                        *PtFicRegles;

char                        Reponses[ 2 *(_NbLigneEcran - 6) ],
                            Rubriques[ 2 *(_NbLigneEcran - 6) ] [ _NbLColEcran ],
                            Titre[ 256 ];

/*************************************************************************
                       DICTIONNAIRE BINAIRE
*************************************************************************/
/* gestion du fichier de stockage des caracteristiques des
   elements lexicaux */
/* 1 : le fichier */
/* 2 : l'indice absolu minimal et maximal dans le buffer
   (en dehors desquels il faut aller se servir dans le fichier pour trouver
   l'element). les indices minimaux servent a deduire l'indice relatif en les
   soustrayant a l'indice absolu */
FILE                       *FicBDicRegles;
FILE                       *FicBDicLemmes;
FILE                       *FicBDicLiens;
FILE                       *FicBDicLiensAux;
unsigned long               IndBuffLemmeMin;
unsigned long               IndBuffLemmeMax;
unsigned long               IndBuffRegleMin;
unsigned long               IndBuffRegleMax;
char                        NomFicDico[ _NbMaxLetNomFic ];
char                        NomFicMessages[ _NbMaxLetNomFic ];
char                        NomFicExtUnifier[ _NbMaxLetNomFic ];
char                        NomFicUnification[ _NbMaxLetNomFic ];
char                        NomFicAux[ _NbMaxLetNomFic ];
FILE                       *FicUnification;

/* un buffer pour relever une unite lexicale + regles en une seule lecture */
TabChar                     BufferLemmesCourant;
TabChar                     BufferReglesCourant;
TabChar                     BufferLemmesCourt;
TabChar                     BufferReglesCourt;
TabChar                     BufferLemmesLong;
TabChar                     BufferReglesLong;
TabChar                     BufferLiensMorphologiquesEtSemantiques;
long                        TailleBuffStd;     
                                       /* taille buffer d'ES stdrd (txt)    */
long                        TailleBuffLemmesCourt;  
                                       /* taille buffer d'ES stdrd (lemmes) */
long                        TailleBuffReglesCourt;  
                                       /* taille buffer d'ES stdrd (regles) */
long                        TailleBuffLemmesLong;  
                                       /* taille buffer d'ES stdrd (lemmes) */
long                        TailleBuffReglesLong;  
                                       /* taille buffer d'ES stdrd (regles) */
long                        TailleBuffLemmesCourant;  
                                       /* taille buffer d'ES en cours(lemmes)*/
long                        TailleBuffReglesCourant;  
                                       /* taille buffer d'ES en cours(regles)*/
int                         BDicSurDisque;   
                                       /* booleen faux en cas de chrgt
                                          du dico en memoire */
int                         BDicReduit;  
                                       /* dictionnaire ou sont
                                          supprimes les chemins lemma
                                          et inflection */

/*************************************************************************
        LE DICTIONNAIRE DES MOTS COURTS OU USUELS
*************************************************************************/
struct EntreeDucs           *Ducs;

/*************************************************************************
                FICHIER GLOBAUX ET BOOLEENS DE SETUP
*************************************************************************/
ChIdent                     Versionfastr;      /* version of fastr */
ChIdent                     Datefastr;         /* date of fastr */
char                        NomFicLog[ _NbMaxLetNomFic ];
char                        NomFicHeader[ _NbMaxLetNomFic ];
FILE                       *FicBufferiseCourant; 
                           /* fichier d'entree bufferise*/
FILE                       *FicLog;     
                           /* fichier des messages       */
FILE                       *FicRefIdent;    
                           /* fichier table des identificateurs de
                              regles + ref */
FILE                       *FicRefMots;     
                           /* fichier table des mots + ref */
int                         VerificationAccord;
int                         PretraitementDesNombres;
int                         OuvreAutomMetaregl;
int                         OuvreAutomReglesMem;
int                         AriteMin;  
                           /* arite en-dessous de laquelle les regles
                              ne sont pas tenteees */
int                         AriteMax;  
                           /* arite au-dessus de laquelle les regles
                              ne sont pas tenteees */
int                         FormatDEntree;  
                           /* type des donnees en indexation */
int                         FormatDeSortie;  
                           /* type des donnees en indexation */
int                         TraceTexteEnSortie; 
                           /* ecriture du texte lu en analyse */
int                         HeaderEnSortie; 
                           /* ecriture du texte lu en analyse */
long                        NoPremLigne;  
                           /* no de la premiere notice en sortie Inist */
int                         ArretSurErreurCompil;
int                         ClasstReglesAutom;  
                           /* resequentialisation autom. des regles en
                              fin de compil */
int                         LinkAutom;  
                           /* ed autom de liens en fin de compil */
int                         LexicalisationMultiple;  
                           /* lexicalisation autom d'1 regle sur ses lemmes */
int                         TraitementBatch;  
                           /* insertion dans un shell */
int                         MinusculisationAutomatique;  
                           /* Passage automatique en minuscules */
int                         SegmentationAutomatique;  
                           /* Segmentation automatique des mots */
int                         LexicalisationImperativeDesRegles;
int                         AutomaticSelfMorphologicalRoot;  
                           /* automatic addition of morph. root <self> */
int                         AutomaticSelfSemanticLinks;  
                           /* automatic addition of semantic links to <self> */
char                        ChErreurCompil[ 2048 ];
char                        ChAux[ 2048 ];
int                         Trace;

/*************************************************************************
                            VARIABLES D'ANALYSE
*************************************************************************/
/* un indice, indiquant eventuellement l'indice du mot courant de la
  la sequence en cours de reconnaissance                               */
int                         IndiceMotCou;

/* la pile d'effacement avec son pointeur et l'indice de haut de pile  */
TabChar                     Pile;
unsigned long               HautPile;
unsigned long               TaillePileEffacement;

int                         NbMaxGenerations;

/*************************************************************************
            VARIABLES GLOBALES POUR LA GESTION DES FLEXIONS
*************************************************************************/
/**    LISTES DE SUFFIXES A AJOUTER AUX RACINES POUR AVOIR LES FLEXIONS    **/
struct UnNoeudSuff          NoeudsSuff[ _NbMaxNoeudsSuffixe ];
int                         NbNoeudsSuff;    /* prochainn noeud a allouer   */
/**        LES SUFFIXES PREDEFINIS                                         **/
struct Suffixe              *Suffixes;

/**        LE DICTIONNAIRE DE TRAITS ET CATEGORIES LEXICALES           **/
TabDicoTraitsCatLex         DicoTraitsCatLex;

/**        LES CHEMINS PREDEFINIS                                      **/
TabDagMem                   CheminCategorie;
TabDagMem                   CheminEtiquette;
TabDagMem                   CheminRacinesAnnexes;
TabDagMem                   CheminLexicalisation;
TabDagMem                   CheminLemme;
TabDagMem                   CheminFlexion;
TabDagMem                   CheminDerivation;
TabDagMem                   CheminForme;
TabDagMem                   CheminNombre;
TabDagMem                   CheminReference;
TabDagMem                   CheminEtiquetteMeta;
TabDagMem                   CheminRacine;
TabDagMem                   CheminLeadingExprReg;
TabDagMem                   CheminTrailingExprReg;
TabDagMem                   CheminsSemantique[ _NbMaxTypesLiensSem ];
TabDagMem                   CheminSelf;

/**        LES CATEGORIES PREDEFINIES                                  **/
long                        RefCatMotsInc;
long                        RefDeterminantNumeral;
long                        RefPonctuation;
long                        RefPonctuationFinale;

/**        LES CATEGORIES A DISTINGUER EN MORPHO ROBUSTE               **/
long                        RefMajorCategories[ _MaxMajorCategories ];
int                         NbMajorCategories;
long                        RefCatLexVide; /* la val de la cat lex vide X */

/**        LES ATOMES PREDEFINIS                                       **/
long                        RefSingulier;
long                        RefPluriel;

/**        LES FLEXIONS PREDEFINES                                     **/
TabDagMem                   TabDagsSuffixage[ _NbMaxSuffixages ];
TabDagMem                   TabListeMetaReglesStd[ _NbMaxTableau ];

/**        LES CARACTERES SPECIAUX                                     **/
char                        CarSpeciauxMots[ _MaxCarSpeciaux ];
char                        Ponctuation[ _MaxCarSpeciaux ];
char                        PonctuationFinale[ _MaxCarSpeciaux ];

/*************************************************************************
                              tailles des tableaux
*************************************************************************/
int                         NbMaxReglesMem;  
                           /* nb max de regles memoire          */

long                        NbMaxElemAllouables[ _DerTypeElemAllouable ];
/*int                         NbMaxMotPhrase;  */
/*                           /* nombre maximal de mots lus ds 1 ph*/
/*                             NbMax[ _RefPhrase ]; */
/*                            /* nombre maximal de references de mots */
/* 			      (semantiques) dans une phrase */
/*                             NbMax[ _Regles ];      */
/*                            /* nb max de regles allouables       */
/*                             NbMax[ _Termes ];      */
/*                            /* nb max de termes de regles alloues*/
/*                             NbMax[ _ReglesMeta ];  */
/*                            /* nb max de metaregles allouables   */
/*                             NbMax[ _Lemmes ];      */
/*                            /* nb max de lemmes de regles alloues*/
/*                             NbMax[ _Dag ];         */
/*                            /* nb max de fils de arcs alloues    */

                            /* nb de fois ou on a du reallouer un type donne */
int                         NbReallocations[ _DerTypeElemAllouable ];

int                         NbMaxLiensMemorises; 
                           /* nb max de liens memorises     */

int                         NbMaxTraitsCatLex; 
                           /* n max d'ident de cat lex ou traits*/
int                         NbMaxSuffixes;     
                           /* nb max de suffixes de mots simples*/
int                         NbMaxTermesDansUneRegle; 
                           /* nb max de termes dans une regle (pour
                              ref dans chemins) */
int                         NbMaxChainesTermes;
                           /* nb max de chaines de termes ds la liste
                              de reference*/
int                         NbMaxLiens;        
                           /* nb max de liens acquis par var. */

/*************************************************************************
        VARIABLES GLOBALES DE LA MORPHOLOGIE
*************************************************************************/
int                         SuffStripping;     
                           /* suppression uniforme des suffixes */
int                         TailSuffStripping; 
                           /* taille de la suppression uniforme des suffixes */
int                         MorphoFlex;     
                           /* morphologie flexionnelle interne */
int                         MorphoDer;     
                           /* morphologie derivationnelle interne */
int                         ExploitationLiensMorpho;     
                           /* exploitation des liens morphologiques compiles */
int                         ExploitationLiensSem;     
                           /* exploitation des liens semantiques compiles */


/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            fin de alppal.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
