/*
 * FASt Term Recognizer
 *
 *  FASTR/alrefere.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for calculating complete references of lemmas
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*************************************************************************
      VARIABLES LOCALES (GRAPHISME)
*************************************************************************/
extern int                             ProfondeurDag;
extern int                             ProfondeurTerme;
extern int                             Largeur;


/*************************************************************************
     ECRITURE DE L'EN-TETE DU FICHIER
*************************************************************************/
void     EcritEnTete( FILE *p_ptFic , long p_nbIdent )
{
    fprintf( p_ptFic , "%-10ld\n" , p_nbIdent );
}

/*************************************************************************
     ECRITURE D'UN IDENTIFICATEUR DANS LE FICHIER
*************************************************************************/
void     EcritIdent( FILE *p_ptFic , char *p_ch , 
                     ChIdent p_ident , long p_ref )
{
    if( p_ptFic )
    {
        fprintf( p_ptFic ,  "%-31s,%-10ld\n" , p_ident , p_ref );
    }
    else if( p_ch )
    {
        sprintf( p_ch ,  "%-31s,%-10ld\n" , p_ident , p_ref );
    }
}

/*************************************************************************
     ECRITURE DE L'EN-TETE DU FICHIER
*************************************************************************/
void     LitEnTete( FILE *p_ptFic , long *p_nbIdent )
{
    fscanf( p_ptFic , "%ld" , p_nbIdent );
}

/*************************************************************************
     ECRITURE D'UN IDENTIFICATEUR DANS LE FICHIER
*************************************************************************/
void     LitIdent( FILE *p_ptFic , ChIdent p_ident , 
                   long *p_ref , char *p_carLu )
{
    int              i;

    for( i = 0 ; 
         (i < _MaxLetIdent && *p_carLu != ' ' && *p_carLu != EOF); 
         i++ )
    {
       p_ident[ i ] = *p_carLu;
       *p_carLu = LireCar( p_ptFic );
    }
    p_ident[ i ] = '\0';
    SautBlanc( p_ptFic , p_carLu );
    ReconnaitreCaractere( p_ptFic , (int)',' , p_carLu , _True );
    *p_ref = LireEntier( p_ptFic , p_carLu );
}

/*************************************************************************
     RECHERCHE DES IDENTIFICATEURS DANS LE FICHIER TEXTE QUI CONTIENT
                             LEURS REFERENCES
*************************************************************************

* Cette recherche n'est pas optimisee, puisqu'elle se fait directement
* dans un fichier texte non charge, mais cette recherche se fait
* uniquement en compilation et non en execution. Il s'agit d'une
* recherche dichotomique dans le fichier dont tous les enregistrements
* ont la meme taille : 31 lettres pour l'identificateur completees
* eventuellement avec des blancs une virgule et 10 lettres pour
* l'entier , complete aussi avec des blancs eventuellement */

long     Recherche( char *p_ident , long *p_numero , 
                    FILE *p_ptFic , int *p_erreur )
{
    long            nbIdent;
    long            derNumero;
    unsigned long   tailleFic;

    ChIdent         identInf,
                    identSup;
    long            numInf;
    long            numSup;
    long            valRet;

    char            carLu;

    *p_erreur = _False;
    fseek( p_ptFic , 0L , 0 );
    LitEnTete( p_ptFic , &derNumero );

    fseek( p_ptFic , 0L , 2 );
    fgetpos( p_ptFic , &tailleFic );

/* 	printf( "tailleFic %ld\n" , tailleFic ); */

    nbIdent = ((long)tailleFic -
        ( (long)_MaxLetLong + (long)_FinLigne ) /* 1ere ligne */ )
        / ( (long)_MaxLetLong + (long)_MaxLetIdent
                      + (long)_FinLigne + 1L /* ',' et '\n' */ );

/* 	printf( "nbIdent %ld\n" , nbIdent ); */

    if( nbIdent != derNumero )
    {
        sprintf( ChAux , "Recherche(alrefere) [%ld,%ld,%ld]",
                         nbIdent , derNumero , tailleFic );
        TxtErreur( 33 , ChAux );
        *p_erreur = _True;
        return _Nil;
    }

    fseek( p_ptFic , (long)(_MaxLetLong + _FinLigne ) , 0 );
    carLu = LireCar( p_ptFic );
    LitIdent( p_ptFic , identInf , &numInf , &carLu );

/* 	printf( "identInf %s\n" , identInf ); */

    fseek( p_ptFic , (long)tailleFic -
           (long)_MaxLetLong - (long)_MaxLetIdent - (long)_FinLigne - 1L , 0 );
    carLu = LireCar( p_ptFic );
    LitIdent( p_ptFic , identSup , &numSup , &carLu );

/* 	printf( "identSup %s\n" , identSup ); */

    valRet = RechercheDicho( (long)(_MaxLetLong + _FinLigne ) ,
                         (long)tailleFic - (long)_MaxLetLong 
                               - (long)_MaxLetIdent - (long)_FinLigne - 1L ,
                         identInf , identSup , p_ident , numInf , numSup , 
                         p_numero , p_ptFic , p_erreur );
    return valRet;
}

/*************************************************************************
     RECHERCHE DES IDENTIFICATEURS DANS LE FICHIER AVEC LEURS REFERENCES
*************************************************************************/
char     AlloueEtChargeCorrespondances( FILE *p_ptFic , 
                    TabCorrIdent_PosFic *p_tCorrespondance )
{
    long                nbIdent;
    long                derNumero;
    unsigned long       tailleFic;
    long                num;
    long                indCou;
    long                posFic;

    ChIdent             ident;
    char                carLu;

    if( !p_ptFic )
    {
        TxtErreur( 150 , "AlloueEtChargeCorrespondances(alrefere)" );
        return _False;
    }
    /* calcul du nombre total de regles */
    fseek( p_ptFic , 0L , 0 );
    LitEnTete( p_ptFic , &derNumero );
    if( !(*p_tCorrespondance) )
    {
        /* p_tCorrespondance contiendra les liens entre les
          identificateurs des regles et leur position absolue dans le
          fichier des identificateur Cette procedure devra etre
          effectuee apres toute recherche rapide dans le fichier des
          identificateurs de regles .  */
        *p_tCorrespondance = (TabCorrIdent_PosFic)
                        _Alloc( (unsigned long)sizeof(struct CorrIdent_PosFic)
                                * (unsigned long)(derNumero + 1L) );
        if ( !(*p_tCorrespondance) )
        {
            TxtErreur( 81 ,"AlloueEtChargeCorrespondances(alrefere)" );
            return _False;
        }
    }
    fseek( p_ptFic , 0L , 2 );
    fgetpos( p_ptFic , &tailleFic );
    if( tailleFic == 0L )
    {
        TxtErreur( 146 , ChAux );
        _Free( (*p_tCorrespondance) );
        return _False;
    }

    nbIdent = ((long)tailleFic -
        ( (long)_MaxLetLong + (long)_FinLigne ) /* 1ere ligne */ )
        / ( (long)_MaxLetLong + (long)_MaxLetIdent
                      + (long)_FinLigne + 1L /* ',' et '\n' */ );

    if( nbIdent != derNumero )
    {
        sprintf( ChAux , 
                 "AlloueEtChargeCorrespondances(alrefere) [%ld,%ld,%ld]",
                  nbIdent , derNumero , tailleFic );
        TxtErreur( 32 , ChAux );
        _Free( (*p_tCorrespondance) );
        return _False;
    }
    for( indCou = 0L ; indCou < nbIdent ; indCou++ )
    {
        posFic = (long)(_MaxLetLong + _FinLigne )
                 + indCou * ( (long)_MaxLetLong + (long)_MaxLetIdent
                 + (long)_FinLigne + 1L /* ',' et '\n' */ );
        fseek( p_ptFic , posFic , 0 );

        carLu = LireCar( p_ptFic );
        LitIdent( p_ptFic , ident , &num , &carLu );
        ((*p_tCorrespondance) + indCou )->RefIdent = num;
        ((*p_tCorrespondance) + indCou )->PosFicIdent = posFic;
        if( indCou % 300L == 0L )
        {
             sprintf( ChAux , "%s" , ident );
             TxtEcritItemInfo( 1 , ChAux );
        }
    }
    return _True;
}

/*************************************************************************
     RECHERCHE DES IDENTIFICATEURS DANS LE FICHIER AVEC LEURS REFERENCES
*************************************************************************/
char     Identificateur( char *p_ident , long p_numero , FILE *p_ptFic ,
                         TabCorrIdent_PosFic p_tCorrespondance )
{
    long                nbIdent;
    long                derNumero;
    unsigned long       tailleFic;
    long                num;
    long                indCou;

    char                carLu;
    int                 traceCou;

    if( (traceCou = Trace) )
    {
         Trace = _False;
    }

    if( !p_ptFic )
    {
        TxtErreur( 150 , "Identificateur(alrefere)" );
        Trace = traceCou; return _False;
    }
    fseek( p_ptFic , 0L , 0 );
    LitEnTete( p_ptFic , &derNumero );

    if( !p_tCorrespondance )
    {
        fseek( p_ptFic , 0L , 2 );
        fgetpos( p_ptFic , &tailleFic );
        if( tailleFic == 0L )
        {
            TxtErreur( 146 , ChAux );
            Trace = traceCou; return _False;
        }

        nbIdent = ((long)tailleFic -
            ( (long)_MaxLetLong + (long)_FinLigne ) /* 1ere ligne */ )
            / ( (long)_MaxLetLong + (long)_MaxLetIdent
                          + (long)_FinLigne + 1L /* ',' et '\n' */ );

        if( nbIdent != derNumero )
        {
            sprintf( ChAux , "Identificateur(alrefere) [%ld,%ld,%ld]",
                             nbIdent , derNumero , tailleFic );
            TxtErreur( 34 , ChAux );
            Trace = traceCou; return _False;
        }
        for( indCou = 0 ; indCou < nbIdent ; indCou++ )
        {
            fseek( p_ptFic , (long)(_MaxLetLong + _FinLigne )
                     + indCou * ( (long)_MaxLetLong + (long)_MaxLetIdent
                     + (long)_FinLigne + 1L /* ',' et '\n' */ ) , 0 );

            carLu = LireCar( p_ptFic );
            LitIdent( p_ptFic , p_ident , &num , &carLu );
            if( num == p_numero )
            {
                Trace = traceCou; return _True;
            }
        }
    }
    else
    {
        for( indCou = 0L ; indCou < derNumero ; indCou++ )
        {
            if( (p_tCorrespondance + indCou)->RefIdent == p_numero )
            {
                fseek( p_ptFic , 
                       (p_tCorrespondance + indCou)->PosFicIdent , 0 );
                carLu = LireCar( p_ptFic );
                LitIdent( p_ptFic , p_ident , &num , &carLu );
                if( num == p_numero )
                {
                    Trace = traceCou; return _True;
                }
            }
        }
    }

    sprintf( ChAux , "Identificateur(alrefere) [%ld]", p_numero );
    TxtErreur( 35 , ChAux );
    TxtAlerte( 1 , 1 );
    *p_ident = '\0';
    Trace = traceCou; return _False;
}

/*************************************************************************
            RECHERCHE DICHOTOMIQUE RECURSIVE DANS LE FICHIER
*************************************************************************/

long     RechercheDicho( long p_indInf , long p_indSup , ChIdent p_identInf ,
             ChIdent p_identSup , ChIdent p_ident , long p_numInf ,
             long p_numSup ,long *p_numero , FILE * p_ptFic , int *p_erreur )
{
    ChIdent         identMedian;

    long            numMedian,

                    indMedian,

                    intervalle,
                    nbIdent;

    char            carLu;

    *p_erreur = _False;

    if( (intervalle = (p_indSup - p_indInf))
      == (long)_MaxLetLong + (long)_MaxLetIdent + (long)_FinLigne + 1L )
    {
        if( EgaliteChaine( p_identInf , p_ident ) )
        {
            *p_numero = p_numInf;
            return( p_indInf );
        }
        else if( EgaliteChaine( p_identSup , p_ident ) )
        {
            *p_numero = p_numSup;
            return( p_indSup );
        }
        else if( InferieurStrict( p_ident , p_identSup )
                 && InferieurStrict( p_identInf , p_ident ) )
        {
            *p_numero = _Nil;
            return( p_indSup ); /* l'indice du 1er superieur */
        }
        else
        {
            sprintf( ChAux , "RechercheDicho(alrefere) [%s,%s,%s] [%d,%d,%d]" ,
                             p_identInf , p_ident , p_identSup ,
                             (int)(*p_identInf) , (int)(*p_ident) , 
                             (int)(*p_identSup) );
            TxtErreur( 36 , ChAux );
            *p_erreur = _True;
            return _Nil;
        }
    }

    /* recherche de l'identificateur et du numero de l'indice median */
    nbIdent = intervalle
        / ( (long)_MaxLetIdent + (long)_MaxLetLong + (long)_FinLigne + 1L );
    nbIdent /= 2;
    indMedian = p_indInf
        + nbIdent
         * ( (long)_MaxLetIdent + (long)_MaxLetLong + (long)_FinLigne + 1L );

/* 	printf( "indMedian %ld\n" , indMedian ); */

    fseek( p_ptFic , indMedian , 0 );
    carLu = LireCar( p_ptFic );
    LitIdent( p_ptFic , identMedian , &numMedian , &carLu );

/* 	printf( "identMedian %s\n" , identMedian ); */

    if( InferieurStrict( identMedian , p_ident ) )
    {
        return( RechercheDicho( indMedian , p_indSup , 
                                identMedian , p_identSup ,
                                p_ident , numMedian , 
                                p_numSup , p_numero , p_ptFic , p_erreur ));
    }
    else if( InferieurStrict( p_ident , identMedian ) )
    {
        return( RechercheDicho( p_indInf , indMedian , 
                                p_identInf , identMedian ,
                                p_ident , p_numInf , 
                                numMedian , p_numero , p_ptFic , p_erreur ));
    }
    else
    {
        *p_numero = numMedian;
        return( indMedian );
    }
}

/*************************************************************************
            INSERTION D'UN IDENTIFICATEUR DANS LE FICHIER
*************************************************************************/

long     InsereIdentFichier( ChIdent p_ident , long *p_numero , 
                             FILE * p_ptFic , int *p_erreur , int *p_trouve )
{
    unsigned long  indRecherche;

    *p_trouve = _False;
    indRecherche = (unsigned long)Recherche( p_ident , p_numero , 
                                             p_ptFic , p_erreur );
    if( *p_erreur )
    {
        return _Nil;
    }

    /* l'identificateur existe deja */
    if( *p_numero != (long)_Nil )
    {
       *p_trouve = _True;
       return (long)indRecherche;
    }

    /* l'identificateur n'est pas encore connu, il doit etre insere */
    fseek( p_ptFic , 0L , 0 );
    LitEnTete( p_ptFic , p_numero );
    fseek( p_ptFic , 0L , 0 );
    EcritEnTete( p_ptFic , *p_numero + 1L );

    /* nouveau terme */
    EcritIdent( NULL , ChAux , p_ident , *p_numero );

    /* on fait la place pour cette ecriture */
    Inserer( indRecherche , indRecherche ,
             (long)_MaxLetLong + (long)_MaxLetIdent
                      + (long)_FinLigne + 1L /* ',' et '\n' */ ,
             (TabChar)ChAux , p_ptFic );

    /* on reecrit la chaine car le \n est mal code si on ne passe par
       printf */
    fseek( p_ptFic , (long)indRecherche , 0 );
    EcritIdent( p_ptFic , NULL , p_ident , *p_numero );

    return (long)indRecherche;
}

/*************************************************************************
        BOOT STRAP D'UN FICHIER DICO D'IDENTIFICATEURS
*************************************************************************/
void     BootStrapFichierIdent( FILE * p_ptFic )
{
    EcritEnTete( p_ptFic , 2L );
    EcritIdent( p_ptFic ,  NULL , PremLemme , 0L );
    EcritIdent( p_ptFic ,  NULL , DernLemme , 1L );
}

/*************************************************************************
            PARCOURS DEAPTH-FIRST D'UN ARBRE DE TERMES
*************************************************************************/
char     ParcoursDFArbre( 
           TabTermeMem       *p_hdlTerme , 
           char              (*p_fonction)(
           TabTermeMem                   *p_hdlTerme
                                          ) ,
           OrdresDeParcours   p_ordre )
{
    TabTermeMem          fils;
    TabTermeMem          frere;

    if( !(*p_hdlTerme) )
    {
        return _True;
    }

    fils = (*p_hdlTerme)->TermeFils;
    frere = (*p_hdlTerme)->TermeFrere;

    if( p_ordre == _PreOrdre && !((*p_fonction)( p_hdlTerme )) )
    {
        return _False;
    }

    ProfondeurTerme++;
    if( !ParcoursDFArbre( &fils , p_fonction , p_ordre ) )
    {
        ProfondeurTerme--;
        return _False;
    }
    ProfondeurTerme--;

    if( p_ordre == _InOrdre && !((*p_fonction)( p_hdlTerme )) )
    {
        return _False;
    }

    Largeur++;
    if( !ParcoursDFArbre( &frere , p_fonction , p_ordre ) )
    {
        Largeur--;
        return _False;
    }
    Largeur--;

    if( p_ordre == _PostOrdre && !((*p_fonction)( p_hdlTerme )) )
    {
        return _False;
    }
    return _True;
}

/*************************************************************************
            PARCOURS DEAPTH-FIRST D'UN GRAPHE DE DAGS
*************************************************************************/
char     ParcoursDFDag( 
           TabDagMem   *p_ptDag , 
           char              (*p_fonctionDag)(
           TabDagMem                          *p_ptFils
                                                    ) ,
           OrdresDeParcours  p_ordre , int  p_arretSiMarque )
{
     TabDagMem  *tabNoeudsParcourus;
     int         nbNoeudsParcourus;
     long        taille;
     char        valRet;

     taille = (unsigned long) sizeof( TabDagMem )
                * (unsigned long) NbMaxLiensMemorises;
     tabNoeudsParcourus = (TabDagMem *)_Alloc( taille );
     if ( !tabNoeudsParcourus )
     {
       TxtErreur( 18 , "ParcoursDFDa(alrefere.c)" );
         exit(0);
     }

     nbNoeudsParcourus = 0;
     valRet = ParcoursDFDagRec( p_ptDag , p_fonctionDag , 
                                p_ordre , p_arretSiMarque ,
                                tabNoeudsParcourus , &nbNoeudsParcourus );
     _Free( tabNoeudsParcourus );
     return valRet;
}

char     ParcoursDFDagRec( 
           TabDagMem   *p_ptDag , 
           char              (*p_fonctionDag)(
           TabDagMem                          *p_ptFils
                                                    ) ,
           OrdresDeParcours  p_ordre , int p_arretSiMarque,
           TabDagMem *p_noeudsVus , int *p_nbNoeudsVus )
{
    TabDagMem            fils;
    TabDagMem            frere;
    TabDagMem            frereSuivant;
    int                  largeurCou;
    int                  nbNoeudsVusCou;
    char                 resul;
    int                  nbFreres = 0;
    
    if( p_arretSiMarque && (*p_ptDag) && (*p_ptDag)->Marque )
    {
       return _True;
    }

    frereSuivant = (*p_ptDag);
    largeurCou = Largeur;
    while( frereSuivant && nbFreres < 10000 )
    {
        /* blocage de la cyclicite */
        for( nbNoeudsVusCou = 0 ; nbNoeudsVusCou < *p_nbNoeudsVus ; 
             nbNoeudsVusCou++ )
        {
            if( frereSuivant == p_noeudsVus[ nbNoeudsVusCou ] )
            {
               if( *p_fonctionDag == EcrireDagMemAvecReentrance
                   || *p_fonctionDag == NoteLiensNoeudsMemVar 
                   || *p_fonctionDag == EcrireDagMem )
               {
                     (*p_fonctionDag)( &frereSuivant );
               }
               return _True;
            }
        }
        if( *p_nbNoeudsVus >= NbMaxLiensMemorises )
        {
            TxtErreur( 94 , "ParcoursDFDag(alrefere)" );
            return _False;
        }
        p_noeudsVus[ *p_nbNoeudsVus ] = frereSuivant;
        (*p_nbNoeudsVus)++;
        /* blocage de la cyclicite */
    
        Largeur++;

        frere = frereSuivant;
        frereSuivant = frere->Frere;

        fils = NULL;
        switch( frere->TypeDag )
        {
        case _AtomeEntier :
        case _AtomeChaine :
        case _AtomeRef :
        case _TypeDagVide :
            break;
        case _AtomeListe :
        case _Arc :
        case _Oper :
            fils = frere->Fils;
            break;
        default :
            sprintf( ChAux , 
                     "ParcoursDFDagrec(alrefere) Unknown type: %d" ,
                     frere->TypeDag );
            TxtErreur( 99 , ChAux );
            break;
        }

/* {ChIdent ch1; */
/*     if( !ChaineDagMem( &frere , ch1 ) ) */
/*     { */
/*        return _False; */
/*     } */
/*     fprintf(FicLog,"%s\n",ch1); */

/* /*     EcritChaineDag( ch1 ); */ 
/* } */
       if( p_ordre == _PreOrdre )
        {
           if( !( resul = (*p_fonctionDag)( &frere )) )
           {
               return _False;
           }
        }
        else
        {
            resul = _True;
        }

        ProfondeurDag++;
        if( fils && resul != _ArretParcoursFils
            && !ParcoursDFDagRec( &fils , p_fonctionDag , p_ordre ,
                              p_arretSiMarque ,
                              p_noeudsVus , p_nbNoeudsVus ) )
        {
            ProfondeurDag--;
            return _False;
        }
        ProfondeurDag--;

        if( p_ordre == _InOrdre || p_ordre == _PostOrdre)
        {
           if( !((*p_fonctionDag)( &frere )) )
           {
               return _False;
           }
        }

        nbFreres++;
    }

    if( nbFreres == 10000 )
    {
       TxtErreur( 251 , "ParcoursDFDagrec(alrefere)" );
       return _False;
    }

    Largeur = largeurCou;

    return _True;
}

/*************************************************************************
            PARCOURS BROAD-FIRST D'UNE RACINE
*************************************************************************/

char     ParcoursListeRegles( 
           TabRegleMem       *p_ptRegle , 
           char              (*p_fonction)(
           TabRegleMem                       *p_hdlTerme
                                          ) )
{
    TabRegleMem                           frere;

    if( !(*p_ptRegle) )
    {
        return _True;
    }

    frere = (*p_ptRegle)->RegleFrere;

    if( !((*p_fonction)( p_ptRegle )) )
    {
        return _False;
    }
    if( !ParcoursListeRegles( &frere , p_fonction ) )
    {
        return _False;
    }

    return _True;
}

/*************************************************************************
 REMPLACEMENT DES REFERENCES PAR DES CHAINES AVANT IMPRESSION D'UN TERME
*************************************************************************/
char     GarnirChaineTermeMem( TabTermeMem *p_hdlTerme )
{
     ChIdent                         ident;
     long                            ref;
     char                            type;

    if( !(*p_hdlTerme) )
    {
        return _False;
    }
    type = (*p_hdlTerme)->ValTerme.TypeTerme;
    if( !(*((*p_hdlTerme)->ValTerme.Ident)) )
    {
       ref = (*p_hdlTerme)->ValTerme.Ref;

       if( type == _UniteLexicale )
       {
           if( ref == _Nil )
           {
               strcpy( (*p_hdlTerme)->ValTerme.Ident , "\'\'" );
           }
           else if( ref == _RefNum )
           {
               strcpy( (*p_hdlTerme)->ValTerme.Ident , "Num" );
           }
           else if( ref < _PremRefLemme )
           {
               if( !IdentificateurDucs( ident , ref ) )
               {
                   Identificateur( ident , ref , FicRefMots , NULL );
               }
               strcpy( (*p_hdlTerme)->ValTerme.Ident , ident );
           }
           else if( ref >= _PremRefLemme )
           {
               Identificateur( ident , ref , FicRefMots , NULL );
               strcpy( (*p_hdlTerme)->ValTerme.Ident , ident );
           }
           else
           {
               sprintf( ChAux , 
                        "Reference : %ld. GarnirChaineTermeMem(alrefere)\n" , 
                        ref );
               TxtErreur( 89 , ChAux );
           }
       }
       else if( type == _Regle )
       {
           if( ref == _Nil )
           {
               strcpy( (*p_hdlTerme)->ValTerme.Ident , "<>" );
           }
           else if( ref > _PremRefLemme )
           {
               Identificateur( ident , ref , FicRefIdent , NULL );
               strcpy( (*p_hdlTerme)->ValTerme.Ident , "<" );
               strcat( (*p_hdlTerme)->ValTerme.Ident , ident );
               strcat( (*p_hdlTerme)->ValTerme.Ident , ">" );
           }
           else
           {
               sprintf( ChAux , 
                        "Reference : %ld. GarnirChaineTermeMem(alrefere)\n" , 
                        ref );
               TxtErreur( 90 , ChAux );
           }
       }
       else
       {
           switch( type )
           {
             case  _Concatenation: strcpy( (*p_hdlTerme)->ValTerme.Ident , 
                                           "." );
                                   break;
             case  _Liste        : strcpy( (*p_hdlTerme)->ValTerme.Ident , 
                                           "+" );
                                   break;
				   /* unknown word */
             case  _Feuille        : strcpy( (*p_hdlTerme)->ValTerme.Ident , 
                                           "[]" );
                                   break;
             default : sprintf( ChAux , 
                             "Type : %d. GarnirChaineTermeMem(alrefere)\n" , 
                                type );
                       TxtErreur( 88 , ChAux );
           }
       }
    }
    return _True;
}

/*************************************************************************
 REMPLACEMENT DES REFERENCES PAR DES CHAINES AVANT IMPRESSION D'UNE REGLE
*************************************************************************/
char     GarnirRefRegleMem( TabRegleMem *p_ptRegle )
{
    if( !(*p_ptRegle) || !((*p_ptRegle)->TermeFils) )
          return _True;

    sprintf( ChAux , "%ld" , (*p_ptRegle)->TermeFils->ValTerme.Ref );
    TxtEcritItemInfo( 1 , ChAux );

    /***        RECHERCHE DES CHAINES ASSOCIEES AUX REFERENCES      ****/
    ProfondeurTerme = 1;
    Largeur = 1;
    ParcoursDFArbre( &((*p_ptRegle)->TermeFils) , 
                     GarnirChaineTermeMem , _PreOrdre );

    if( (*p_ptRegle)->Transformee )
    {
        ParcoursDFArbre( &((*p_ptRegle)->Transformee) , 
                         GarnirChaineTermeMem , _PreOrdre );
    }

    if( _ClickDansAnnulerInfo() )
    {
         return _False;
    }

    return _True;
}

/*************************************************************************
            REFERENCE DES TERMES LEXICAUX
*************************************************************************/
/* renvoit la taille fichier du lemme s'il est trouve, 0 sinon,
   _Nil s'il ne s'agit pas d'un lemme memoire */
long     ReferenceTermeLexical( ChIdent p_ident ,
                                struct ValTerme *p_valTerme ,
                                char *p_estMotCourant , 
                                unsigned long *p_posFic )
{
    unsigned long           indLemme;
    unsigned long           indLemmeIni;
    ChIdent                 chaineAnnexe;
    ChIdent                 chaineLemme;
    char                    estEgale;
    struct LemmeMem         lemmeCou;

    int                     indHomographe;
    struct EntreeDucs      *listePtDucs[ _NbMaxHomographe ];
    int                     noTypeFlexion1;
    int                     noTypeFlexionTerme;
    int                     catLex1;
    int                     catLexTerme;

    *p_estMotCourant = _False;
    *p_posFic = _Nil;

    if( !( *(p_ident) ) )
    {
        p_valTerme->Ref = _Nil;
        return _Nil;
    }

    /* cas particulier d'un nombre */
    if( PretraitementDesNombres && _Num(*(p_ident)) )
    {
        p_valTerme->Ref = _RefNum;
        return _Nil;
    }
    else
    {
      catLexTerme = CategorieLexicale( &(p_valTerme->ArcsFils) );
      noTypeFlexionTerme 
                  = NoTypeSuffixage( &(p_valTerme->ArcsFils), _True );

      /* recherche dans les mots usuels */
      ListePointeursDucsParChaine( p_ident , listePtDucs );

      if( listePtDucs[0] )
      {
         indHomographe = 0;
         while( listePtDucs[ indHomographe ] )
         {
             catLex1 = CategorieLexicale( 
                   &(listePtDucs[ indHomographe ]->Lemme.ValTerme.ArcsFils ));

             if( catLex1 == catLexTerme || catLex1 == _Nil 
		 || catLexTerme == _Nil )
             {
                p_valTerme->Ref 
                         = listePtDucs[ indHomographe ]->Lemme.ValTerme.Ref;
                *p_estMotCourant = _True;
                return _Nil;
             }
             indHomographe++;
         }
      }

      /* recherche dans les mots longs */
      indLemme = IndiceInsertionDicLemmes( p_ident , &estEgale );

      /* BOUCLE SUR LES LEMMES DE MEME RACINE                    */
      while( EgaliteChaine( BufferLemmesCourant + _IRelLemme( indLemme ) , 
                            p_ident ) == _True)
      {
          indLemmeIni = indLemme;
          if( RestructurerBDicLemme( &indLemme , chaineLemme  ,
                                     &lemmeCou , chaineAnnexe ) == _Nil 
              || EgaliteChaine( chaineLemme , DernLemme ) )
          {
                DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
                break;
          }
          catLex1 = CategorieLexicale( &(lemmeCou.ValTerme.ArcsFils) );
          noTypeFlexion1 
                  = NoTypeSuffixage( &(lemmeCou.ValTerme.ArcsFils) , _True );
          DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
          
          /* si on est sur une racine annexe, on ne compare pas avec
             le lemme donne, car une racine annexe est une partie de
             la racine ppale et n'a pas de valeur representative de
             lemme (de plus sa cat lex et sa flexion ne sont pas
             definies */
          if( !(*chaineAnnexe)
              && ( catLex1 == catLexTerme || catLex1 == _Nil 
		   || catLexTerme == _Nil )
              && ( noTypeFlexion1 == noTypeFlexionTerme 
                   || noTypeFlexion1 == _Nil || noTypeFlexionTerme == _Nil ) )
          {
                  p_valTerme->Ref = lemmeCou.ValTerme.Ref;
                  *p_posFic = indLemmeIni;
                  return indLemme - indLemmeIni;
          }
      }

      return 0L;
    }
}

/*************************************************************************
        Ajout d'une reference numerique et d'une chaine a un terme
        qui est un lemme (etiq lemma) ou un non terminal (etiq label)
        Convient pour les regles ou la chaine qui sert de reference 
        lors de la lecture des contraintes peut etre perdue apres
*************************************************************************/
char     AjouteAutomCheminReference( TabTermeMem *p_hdlTerme )
{
    struct ValAtomeMem     atome;

    /* on ajoute un trait CheminReference s'il n'existe pas */
    if(!ExisteCheminFils( CheminReference , 
                          &((*p_hdlTerme)->ValTerme.ArcsFils) ) 
       && (*p_hdlTerme)->ValTerme.Ref != _Nil )
    {
       atome.ValAtome.Entier = (*p_hdlTerme)->ValTerme.Ref;
       atome.AtomeConjDisj = _False;
       if( !AjouteCheminEtAtomeFils( &((*p_hdlTerme)->ValTerme.ArcsFils) ,
                                     CheminReference , _AtomeEntier , 
				     atome , _True ) )
       {
            return _False;
       }         
    }
    return _True;
}

char     GarnirReferenceTermeRegle( TabTermeMem *p_hdlTerme )
{
  if( !GarnirReferenceUnTerme( p_hdlTerme , _True , _False ) ) {
    return _False;
  }

  return AjouteAutomCheminReference( p_hdlTerme );
}

char     GarnirReferenceTermeRegleMem( TabTermeMem *p_hdlTerme )
{
    if( !GarnirReferenceUnTerme( p_hdlTerme , _True , _True ) )
    {
       return _False;
    }

    return AjouteAutomCheminReference( p_hdlTerme );
}

/*************************************************************************
        Ajout d'une reference numerique sans modif de la chaine pour
        un terme qui est un lemme (etiq lemma) ou un non terminal
        (etiq label) Convient pour les metaregle pour lesquelles la
        chaine du terme sert au-dela de la lecture de la metaregle
        pour la transmission des traits de la partie gauche vers la
        partie droite
*************************************************************************/
char     GarnirReferenceTermeMetaRegle( TabTermeMem *p_hdlTerme )
{
    return GarnirReferenceUnTerme( p_hdlTerme , _False , _False );
}

/*************************************************************************
      AJOUT DE TRAITS DE RACINE DANS UN LEMME DE REGLE SI CELUI-CI EN
      A DANS LE FICHIER DES LEMMES
*************************************************************************/
char     GarnirTraitsRacineEtSemantique( TabTermeMem p_ptTerme , 
					 unsigned long p_posFicLemme )
{
  struct LemmeMem                          lemmeCou;
  ChIdent                                  chaineAnnexe;

  TabChar                                  chaineRacine;
  TabLong                                  ptRefRacine;
  TabLong                                  ptCategorie;
  TabLong                                  ptNoFlexion;  
  TabDagMem                               *hdlDag;

  if( RestructurerBDicLemme( &p_posFicLemme , 
			     lemmeCou.ValTerme.Ident  ,
			     &lemmeCou , chaineAnnexe ) == _Nil ) {
    DemarqueEtRendDag( &(lemmeCou.ValTerme.ArcsFils) );
    return _False;
  } 

  /* les 4 bases de restauration des pointeurs */
  PtOperLogU = NULL; 
  HdlDagU = &(lemmeCou.ValTerme.ArcsFils); 
  HdlTermeU = &p_ptTerme;
  PtBlocU = NULL;   
  
  /* echec d'unification: on elimine le lemme flechi considere */
  if( !Unifier( &(lemmeCou.ValTerme.ArcsFils) , 
		&(p_ptTerme->ValTerme.ArcsFils) ) ) {
    TxtErreur( 313 , "GarnirTraitsRacineEtSemantique(alrefere)" );

    return _False;
  }
  else {
   /* si le terme a une racine, on copie la reference de la racine */
   if( (hdlDag = ExisteCheminFils( CheminRacine ,  
				    &(p_ptTerme->ValTerme.ArcsFils) ) ) ) {
       if( !LienUniteLexicaleATraits( hdlDag , 
				      (TabDagMem *)NULL ,
				      &ptRefRacine , &chaineRacine ,
				      &ptCategorie , &ptNoFlexion , 
				      _False ) ) {
	 return _False;
       }
       p_ptTerme->ValTerme.RefRacine = (*ptRefRacine);
   }

   return _True;
  }
}

/*************************************************************************
        Ajout d'une reference numerique (et event d'une chaine) a un terme
        qui est un lemme (etiq lemma) ou un non terminal (etiq label)
*************************************************************************/
char     GarnirReferenceUnTerme( TabTermeMem *p_hdlTerme , int p_avecChaine ,
				 int p_estRegleMem )
{
    struct ValTerme          *ptValTerme;
    long                      valRet;
    int                       erreur;
    int                       trouve;
    char                      ch[ 256 ];
    char                      estMotCourant;
    TabDagMem                *hdlFilsEtiquette;
    TabDagMem                *hdlFilsLemme;
    TabChar                   chLemme;
    TabChar                   chEtiquette;
    unsigned long             posFicLemme;
    long                      resul;


    if( !(*p_hdlTerme) )
    {
        return _True;
    }
    ptValTerme = &((*p_hdlTerme)->ValTerme);

    hdlFilsEtiquette = ExisteCheminFils( CheminEtiquette ,
                                         &(ptValTerme->ArcsFils) );
    hdlFilsLemme = ExisteCheminFils( CheminLemme , 
                                     &(ptValTerme->ArcsFils) );

    if( hdlFilsLemme && hdlFilsEtiquette )
    {
        TxtErreur( 126 , "GarnirReferenceUnTerme(alrefere) [NULL]" );
        return _False;
    }

    if( p_avecChaine )
    {
         *(ptValTerme->Ident) = '\0';
    }

    if( hdlFilsLemme && ptValTerme->TypeTerme != _Feuille )
    {
            TxtErreur( 126 , "GarnirReferenceUnTerme(alrefere) [NULL]" );
            return _False;
    }

    if( hdlFilsLemme && (*hdlFilsLemme) )
    {
        ptValTerme->TypeTerme = _UniteLexicale;
        if( (*hdlFilsLemme)->TypeDag == _AtomeChaine )
        {
            chLemme = (*hdlFilsLemme)->ValDag.Atome.ValAtome.Chaine;

            if( p_avecChaine )
            {
                strcpy( ptValTerme->Ident , chLemme );
            }

            /* cas particulier d'une ponctuation */
            if( FormatDEntree != _TxtTagge
		&& FormatDEntree != _TxtTaggePartiellement 
		&& FormatDEntree != _TxtFenTaggePartielt
		&& strlen( chLemme ) == 1 && strchr( Ponctuation , *chLemme ) )
            {
                 ptValTerme->Ref 
                    = _DebutRefPonctuation 
                    + (long)(strchr( Ponctuation , *chLemme ) - Ponctuation);
            }
            /* cas general d'un lemme fichier ou memoire */
            else if( (resul = ReferenceTermeLexical( chLemme , ptValTerme , 
						     &estMotCourant , 
						     &posFicLemme ) ) )
            {
                /* on recopie egalement la categorie et la reference de
                   la racine eventuelle */
                if( resul != _Nil )
                {
		     AjouteAutomCheminReference( p_hdlTerme );

                     if( !GarnirTraitsRacineEtSemantique( (*p_hdlTerme) , 
							  posFicLemme ) )
                     {
                         return _False;
                     }
                }                
            }
            else 
            {
               /* erreur terminal non retrouve dans le dictionnaire, 
                   la regle ne pourra pas etre inseree */
                TxtChaineErreur( 41 , ch );
                sprintf( ChAux , "%s [%s]\n" , ch , chLemme );
                strcpy( ChErreurCompil , ChAux );
                return _False;
            }
            
            if( (*hdlFilsLemme)->ValDag.Atome.AtomeConjDisj 
		   == _ConjonctionNeg )
            {
                 ptValTerme->Ref = _RefConjonctionNeg;
            }

            if( ptValTerme->Ref == _Nil )
            {
                printf( "\nLemma : [%s]\n\n" , chLemme );
                TxtErreur( 157 , "GarnirReferenceUnTerme(alrefere)" );
                return _False;
            }

            return _True;
        }
        else if( (*hdlFilsLemme)->TypeDag != _TypeDagVide )
        {
            /* un dag de type vide existe dans le cas d'une egalite
               sur des lemmes qui n'est pas garnie par une valeur. Il
               n'y a donc pas lieu dans ce cas de rechercher une
               reference de lemme */
            printf( "Type of dag node [%d]\n" , (*hdlFilsLemme)->TypeDag );
            TxtErreur( 157 , "GarnirReferenceUnTerme(alrefere)" );
            return _False;
        }
    }

    if( hdlFilsEtiquette && (*hdlFilsEtiquette) )
    {
        /* une feuille avec une etiquette est un appel de regle */
        /* sinon, ce doit etre la racine d'une regle */
        if( ptValTerme->TypeTerme == _Feuille )
        {
             ptValTerme->TypeTerme = _Regle;
        }
        if( (*hdlFilsEtiquette)->TypeDag == _AtomeChaine )
        {
            chEtiquette = (*hdlFilsEtiquette)->ValDag.Atome.ValAtome.Chaine;
            
            if( p_avecChaine )
            {
                strcpy( ptValTerme->Ident , chEtiquette );
            }

            if( *(chEtiquette) )
            {
                valRet = InsereIdentFichier( chEtiquette ,
                                             &(ptValTerme->Ref) , 
                                             FicRefIdent , &erreur , &trouve );
                if( valRet == _Nil || erreur )
                {
                    TxtErreur( 40 , "GarnirReferenceUnTerme(alrefere)" );
                    return _False;
                }
                else if( trouve && !p_estRegleMem )
                {
		    /* si la reference du terme existe deja, on reporte
                       une erreur l'Indiquant, sauf si c'est une MR puisque
		       leur reference est deja connue des compils prec */
		    TxtChaineErreur( 252 , ch );
		    sprintf( ChErreurCompil , "%s [%s]\n" , 
			     ch , chEtiquette );
                    return _False;
                }
                else
                {
                    (*hdlFilsEtiquette)->TypeDag = _AtomeRef;
                    (*hdlFilsEtiquette)->ValDag.Atome.ValAtome.Ref 
                                                 = ptValTerme->Ref;
                    return _True;
                }
            }
        }
        else
        {
            TxtErreur( 39 , "GarnirReferenceUnTerme(alrefere)" );
            return _False;
        }
    }

    if( ( hdlFilsLemme && !(*hdlFilsLemme) )
        || ( hdlFilsEtiquette && !(*hdlFilsEtiquette) ) )
    {
        TxtErreur( 40 , "GarnirReferenceUnTerme(alrefere)" );
        return _False;
    }

    return _True ;
}

/*************************************************************************
           REMPLACEMENT D'UNE CHAINE DE RACINE PAR UN POINTEUR
*************************************************************************/
char     GarnirReferenceLemme( TabLemmeMem p_ptLemme )
{
    char                   estMotCourant;
    char                   ch[ 256 ];
    unsigned long          posFicLemme;

    if( ReferenceTermeLexical( p_ptLemme->ValTerme.Ident , 
                               &(p_ptLemme->ValTerme) , &estMotCourant ,
                               &posFicLemme )  != 0L  )
    {
        /* erreur : le terme existe deja, on ne souhaite pas le saisir
           a nouveau */
        if( estMotCourant )
        {
           TxtChaineErreur( 38 , ch );
           sprintf( ChAux , "%s [%s]\n" , 
                    ch , p_ptLemme->ValTerme.Ident );
        }
        else
        {
           TxtChaineErreur( 37 , ch );
           sprintf( ChAux , "%s [%s]\n" , 
                    ch , p_ptLemme->ValTerme.Ident );
        }
        strcpy( ChErreurCompil , ChAux );
        return _False;
    }

    return _True;
}

/*************************************************************************
      COMPLETE LES LIENS SEMANTIQUES (en ajoutant systematiquement
      reference et no de flexion)
*************************************************************************/
char     CompleteTraitsMorphologiquesLemme( TabLemmeMem p_ptLemme ,
					    TabChar p_chLog )
{
    TabDagMem              *hdlDag;
    TabLong                 ptReference;
    TabLong                 ptCategorie;
    TabLong                 ptNoFlexion;  
    TabChar                 ptChaine;
    char                    ch[ 256 ];

    /****************** scans all the semantic features *********************/
    hdlDag = ExisteCheminFils( CheminRacine ,  
                               &(p_ptLemme->ValTerme.ArcsFils) );
    if( !hdlDag )
    {
         /* la racine n'est pas donnee le lemme est sa propre racine */
         return _True;
    }

    /* recherche de la liste des informations decrivant la racine */
    if( !(*hdlDag) || ( (*hdlDag)->TypeDag != _AtomeListe ) ) {
       TxtChaineErreur( 285 , ChErreurCompil );
       return _False;
    }      

    /* loads the list describing the target lexical item */
    if( !LienUniteLexicaleATraits( hdlDag , &(p_ptLemme->ValTerme.ArcsFils) ,
				   &ptReference , &ptChaine ,
				   &ptCategorie , &ptNoFlexion , _False ) ) {
      TxtErreur( 262 , "CompleteTraitsMorphologiquesLemme(alinsere)" );
      return _False;
    }

    sprintf( ch , " => [%s] " , ptChaine  );
    if( strlen( p_chLog ) + strlen( ch ) < _NbMaxLetChOut ) {
      strcat( p_chLog , ch );
    }

    return _True;
}

/*************************************************************************
      COMPLETE LES LIENS SEMANTIQUES (en ajoutant systematiquement
      reference et no de flexion)
*************************************************************************/
char     CompleteTraitsLiensSemantiquesLemme( TabLemmeMem p_ptLemme ,
					      TabChar p_chLog )
{
    TabDagMem              *hdlDag;
    TabLong                 ptReference;
    TabLong                 ptCategorie;
    TabLong                 ptNoFlexion;  
    TabChar                 ptChaine;
    int                     indChemin;
    char                    ch[ 256 ];

    /****************** scans all the semantic features *********************/
    indChemin = 0;
    while( CheminsSemantique[ indChemin ] ) {

      hdlDag = ExisteCheminFils( CheminsSemantique[ indChemin ] ,  
				 &(p_ptLemme->ValTerme.ArcsFils) );

      /****************** semantic links can be disunctions *****************/
      /**************** scans all the links in a disjunction ****************/
      while( hdlDag && *hdlDag ) {

	/* recherche de la liste des informations decrivant la racine */
	if( ( (*hdlDag)->TypeDag != _AtomeListe ) ) {
	  printf( "Type: %d\n" , (*hdlDag)->TypeDag );
	  TxtChaineErreur( 285 , ChErreurCompil );
	  return _False;
	}

	/* loads the list describing the target lexical item */
	if( !LienUniteLexicaleATraits( hdlDag , 
				       &(p_ptLemme->ValTerme.ArcsFils) ,
				       &ptReference , &ptChaine ,
				       &ptCategorie , &ptNoFlexion , 
				       _False ) ) {
	  TxtErreur( 262 , "CompleteTraitsLiensSemantiquesLemme(alinsere)" );
	  return _False;
	}

	sprintf( ch , " ~> [%s]" , ptChaine );
	if( strlen( p_chLog ) + strlen( ch ) < _NbMaxLetChOut ) {
	  strcat( p_chLog , ch );
	}

	hdlDag = &((*hdlDag)->Frere);
      }
      /**************** scans all the links in a disjunction ****************/

      indChemin++;
    }
    /****************** scans all the semantic features *********************/

    return _True;
}

/*************************************************************************
  REMPLACEMENT D'UNE CHAINE D'IDENTIFICATEUR PAR UN POINTEUR
*************************************************************************/
char     GarnirReferenceRegle( TabRegleMem p_ptRegle , int p_estRegleMemoire )
{
  /* garde la memoire des noeuds copies au fur et a mesure de 
     leur copie sur les termes de la regle afin de retrouver
     les reentrances */
  NbLiensNoeudsCopies = 0;

  if( !p_estRegleMemoire ) {
    if( p_ptRegle->TermeFils 
        && !ParcoursDFArbre( &(p_ptRegle->TermeFils) , 
                             GarnirReferenceTermeRegle , _InOrdre ) )
    {
             return _False;
    }
  }
  else {
    if( p_ptRegle->TermeFils 
        && !ParcoursDFArbre( &(p_ptRegle->TermeFils) , 
                             GarnirReferenceTermeRegleMem , _InOrdre ) )
    {
             return _False;
    }
  }

  return _True;
}

/*************************************************************************
  REMPLACEMENT D'UNE CHAINE DE RACINE OU D'IDENTIFICATEUR PAR UN POINTEUR
         POUR LES TERMES DES METAREGLES
*************************************************************************/
char     GarnirReferenceRegleMeta( TabRegleMemMeta p_ptRegleMeta )
{
    if( p_ptRegleMeta->TermeFils )
    {
        if( !ParcoursDFArbre( &(p_ptRegleMeta->TermeFils) ,
                              GarnirReferenceTermeMetaRegle , _InOrdre ) )
         {
             return _False;
         }
    }
    return _True;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            fin de alrefere.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

