/*
 * FASt Term Recognizer
 *
 *  FASTR/alsaisie.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for reading strings
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/***********************************************************
        PROCHAIN CARACTERE EN ENTREE
***********************************************************/
char     LireCar( FILE *p_ptFic )
{
    char    c;

    if( !p_ptFic )
    {
        TxtErreur( 237 , "LireCar(alsaisie)" );
    }

    if( p_ptFic != FicBufferiseCourant )
    {
        c = getc( p_ptFic );
    }
    else
    {
        c = _LireC( p_ptFic );
    }
    if( FicLog && Trace && c != EOF )
    {
         /* fputc( c , FicLog ); */
         fprintf( FicLog , "%c" , c );
    }
    return( c );
}

/***********************************************************
        PROCHAIN CARACTERE NON BLANC EN ENTREE
        SAUTE EGALEMENT LES COMMENTAIRES ENTRE ""
***********************************************************/
void SautBlanc ( FILE *p_ptFic , char *p_c)
{
    while ( _CarBlanc(*p_c) || *p_c == '"' )
    {
        if( *p_c != '"' )
        {
            *p_c = LireCar(p_ptFic);
        }
        else
        {
            *p_c = LireCar(p_ptFic);
            while( *p_c != '"' && *p_c != EOF )
                *p_c = LireCar(p_ptFic);
            if( *p_c == '"' )
                *p_c = LireCar(p_ptFic);
        }
    }
}

/***********************************************************
   LECTURE DE LA CHAINE DE CARACTERES D'UN IDENTIFICATEUR
***********************************************************/
void     LireIdentificateur( FILE *p_ptFic , char *p_c , char *p_chaine , 
                             int p_taille ,
                             int p_avecSautBlanc )
{
    int             i = 0;

    if (_Alphanum(*p_c))
    {
        while ( (_Alphanum(*p_c)) && i < p_taille - 1 )
        {
            p_chaine[i++] = *p_c;
            *p_c = LireCar(p_ptFic);
        }

        if( i == p_taille - 1)
        {
            while (_Alphanum(*p_c))
            {
                *p_c = LireCar(p_ptFic);
            }
        }
        p_chaine[i] = '\0';

        if( p_avecSautBlanc )
        {
            SautBlanc(p_ptFic , p_c);
        }
     }
     else
     {
        p_chaine[0] = '\0';
     }
}

/***********************************************************
   LECTURE D'UNE LIGNE
***********************************************************/
void     LireLigne( FILE *p_ptFic , char *p_c , char *p_chaine , 
                             int p_taille )
{
    int             i = 0;

    while( *p_c != '\n' && *p_c != EOF && i < p_taille - 1 ) {
      p_chaine[i++] = *p_c;
      *p_c = LireCar(p_ptFic);
    }
    if( *p_c == '\n' && i < p_taille - 1  ) {
      p_chaine[i++] = *p_c;
      *p_c = LireCar(p_ptFic);
    }
    else if( i == p_taille - 1) {
      while( *p_c != '\n' && *p_c != EOF ) {
	*p_c = LireCar(p_ptFic);
      }
    }
    p_chaine[i] = '\0';
}

/***********************************************************
   LECTURE DE LA CHAINE DE CARACTERES D'UN IDENTIFICATEUR
     ELLE PEUT COMPORTER CERTAINS CARACT NON ALPHANUM
***********************************************************/
void     LireChaineAvecCarSpec( FILE *p_ptFic , char *p_c , char *p_chaine ,
                                        int p_taille , char *p_lisCarSpec ,
                                        int p_avecSautBlanc )
{
    int             i = 0;

    if( _Alphanum(*p_c) || strchr( p_lisCarSpec , *p_c ) )
    {
        while( ( _Alphanum(*p_c)  || strchr( p_lisCarSpec , *p_c ) ) 
               && i < p_taille - 1 )
        {
            p_chaine[i++] = *p_c;
            *p_c = LireCar(p_ptFic);
        }

        if( i == p_taille - 1)
        {
            while ( _Alphanum(*p_c) || strchr( p_lisCarSpec , *p_c )  )
            {
                *p_c = LireCar(p_ptFic);
            }
        }
        p_chaine[i] = '\0';

        if( p_avecSautBlanc )
        {
            SautBlanc( p_ptFic , p_c );
        }
     }
     else
     {
        p_chaine[0] = '\0';
     }
}

/***********************************************************
   LECTURE DE LA CHAINE DE CARACTERES D'UN IDENTIFICATEUR
     ELLE PEUT COMPORTER CERTAINS CARACT NON ALPHANUM
***********************************************************/
char     LireChaineAvecBalise( FILE *p_ptFic , char *p_c , char *p_chaine ,
                                        int p_taille , int p_balise ,
                                        int p_avecSautBlanc )
{
    int             i = 0;

    if( !ReconnaitreCaractere( p_ptFic , p_balise , p_c , _False ) )
    {
        return _False;
    }
    while( *p_c != (char)p_balise && i < p_taille - 1  && *p_c != EOF )
    {
        /* saut des caracteres inf au blanc
        while( *p_c >= 0 && *p_c < ' '  && *p_c != EOF  
	       && *p_c != p_balise ) {
	     *p_c = LireCar(p_ptFic);
        } */

        if( *p_c == '\\' )
        {
            *p_c = LireCar(p_ptFic);
        }
        if( *p_c == '%' )
        {
             /* pour ne pas laisser de car de formatage dans la chaine */
             p_chaine[i++] = '%';
        }
        p_chaine[i++] = *p_c;
        *p_c = LireCar(p_ptFic);
    }
    if( i == p_taille - 1)
    {
        while( *p_c != (char)p_balise && *p_c != EOF )
        {
             if( *p_c == '\\' )
             {
                 *p_c = LireCar(p_ptFic);
             }
             *p_c = LireCar(p_ptFic);
        }
    }
    p_chaine[i] = '\0';

    if( !ReconnaitreCaractere( p_ptFic , p_balise , 
                               p_c , p_avecSautBlanc ) )
    {
        return _False;
    }

    return _True;
}

/***********************************************************
   LECTURE DE LA CHAINE DE CARACTERES D'UN IDENTIFICATEUR
     ELLE PEUT COMPORTER CERTAINS CARACT NON ALPHANUM
***********************************************************/
void     LireIdentificateurAvecCarSpec( FILE *p_ptFic , char *p_c , 
                                        char *p_chaine ,
                                        int p_taille , char *p_lisCarSpec ,
                                        int p_avecSautBlanc )
{
    int             i = 0;

    if( _Alphanum(*p_c) || strchr( p_lisCarSpec , *p_c )
        || *p_c == '\\' )
    {
        while( ( _Alphanum(*p_c)  || strchr( p_lisCarSpec , *p_c )
                 || *p_c == '\\' )  &&  *p_c != '\'' && i < p_taille - 1 )
        {
            if( *p_c == '\\' )
            {
                *p_c = LireCar(p_ptFic);
            }
            p_chaine[i++] = *p_c;
            *p_c = LireCar(p_ptFic);
        }

        if( i == p_taille - 1)
        {
            while( ( _Alphanum(*p_c) || strchr( p_lisCarSpec , *p_c )
                    || *p_c == '\\' )  &&  *p_c != '\'' )
            {
                *p_c = LireCar(p_ptFic);
            }
        }
        p_chaine[i] = '\0';

        if( p_avecSautBlanc )
        {
            SautBlanc( p_ptFic , p_c );
        }
     }
     else
     {
        p_chaine[0] = '\0';
     }
}

/***********************************************************
        LECTURE DE LA CHAINE DE CARACTERES D'UN SUFFIXE
***********************************************************/

void LireSuffixe ( FILE *p_ptFic , char *p_c , char *p_chaine , int p_taille)
{
        int             i = 0;

    if( !(_Alpha(*p_c) || *p_c == '!'|| *p_c == '*' || *p_c == '?' || *p_c == '0') )
    {
        printf( "\n[%c,%d]\n" , *p_c , *p_c );
        TxtErreur( 145 , "LireSuffixe(alsaisie)" );
    }

    /* prefixe du suffixe                                                            */
    if( *p_c == '?' )                         /* racine annexe        */
    {
            p_chaine[i++] = *p_c;
            *p_c = (unsigned char)LireCar(p_ptFic);
            if( _Num(*p_c) )
            {
                    p_chaine[i++] = *p_c;
                    *p_c = LireCar(p_ptFic);
            }
            else
                    TxtErreur( 110 , "LireSuffixe(alsaisie)" );
    }
    else if( *p_c == '!' || *p_c == '*' )
    /* lettre redoublee ou flexion interdite */
    {
            p_chaine[i++] = *p_c;
            *p_c = LireCar(p_ptFic);

    }
    else if( *p_c == '0' )            /* suffixe vide            */
    {
            p_chaine[0] = '\0';
            *p_c = LireCar(p_ptFic);
            SautBlanc(p_ptFic , p_c);
            return;
    }

    while ( (_Alpha(*p_c)) && i < p_taille - 1)
    {
            p_chaine[i++] = *p_c;
            *p_c = LireCar(p_ptFic);
    }

    if( i == p_taille - 1)
    {
        while (_Alpha(*p_c))
        {
            *p_c = LireCar(p_ptFic);
        }
    }

    SautBlanc(p_ptFic , p_c);
    p_chaine[i] = '\0';
}

/***********************************************************
                LECTURE D'UN ENTIER
***********************************************************/

long     LireEntier( FILE *p_ptFic , char *p_c )
{
        long            i = 0L;
        short           neg = 0;

        if( *p_c == '-')
        {
                neg = 1;
                *p_c = LireCar(p_ptFic);
        }
        if (_Num(*p_c))
        {
            while ( _Num(*p_c) )
            {
                i = i*10L + (long)(*p_c - '0');
                *p_c = LireCar(p_ptFic);
            }

            SautBlanc(p_ptFic , p_c);
            if( neg == 1)
                i = -i;
            return(i);
        }
        else
        {
            return(0L);
        }
}

/***********************************************************
                LECTURE D'UN REEL
***********************************************************/

float    LireReel( FILE *p_ptFic , char *p_c )
{
    float           resul,
                    dec = 0.1;

    resul = (float) LireEntier( p_ptFic , p_c );
    if( *p_c == '.')
    {
       *p_c = LireCar(p_ptFic);

       while ( _Num(*p_c) )
       {
                resul += dec * (*p_c - '0');
                dec *= 0.1;
                *p_c = LireCar(p_ptFic);
       }
    }
    SautBlanc(p_ptFic , p_c);
    return( resul );
}

/***********************************************************
                LECTURE D'UNE CHAINE ATTENDUE
***********************************************************/
char     ReconnaitreChaine( FILE *p_ptFic , char *p_modele , char *p_c )
{
    int             i = 0;

    while( i < (int)strlen( p_modele ) && p_modele[ i ] == *p_c )
    {
        i++;
        *p_c = LireCar(p_ptFic);
    }

    if( i < (int)strlen( p_modele ) )
    {
         TxtChaineErreur( 180 , ChAux );
         sprintf( ChErreurCompil , "[%s] %s %d [%c,%d]",
                           p_modele , ChAux , i , *p_c , *p_c );
         return _False;
    }

    SautBlanc( p_ptFic , p_c );
    return _True;
}

/***********************************************************
                LECTURE D'UN CARACTERE ATTENDU
***********************************************************/
char     ReconnaitreCaractere( FILE *p_ptFic , int p_modele , char  *p_c ,
                               int p_avecSautBlanc )
{
    if( (char)p_modele == *p_c )
    {
         *p_c = LireCar(p_ptFic);
    }
    else
    {
        TxtChaineErreur( 181 , ChAux );
        sprintf( ChErreurCompil , "[%c] %s [%c,%d]" , 
                 (char)p_modele , ChAux , *p_c , *p_c );
        return _False;
    }

    if( p_avecSautBlanc )
    {
        SautBlanc(p_ptFic , p_c);
    }
    return _True;
}

/***********************************************************
                OUVERTURE D'UN FICHIER
***********************************************************/
char     OuvreFichier( char *p_nomFic , FILE **p_ptFic , char *p_mode )
{
      *p_ptFic = fopen( p_nomFic , p_mode );
      if( !(*p_ptFic) )
      {
          sprintf( ChAux , "OuvreFichier(alsaisie) [%s]" , p_nomFic );
          TxtErreur( 13 , ChAux );
          return _False;
      }
      else
      {
          return _True;
      }
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
            fin de alsaisie.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
