/*
 * FASt Term Recognizer
 *
 *  FASTR/althesau.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for managing thesarus links
 */

#include "al-ext.h"
#include "al-var.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/*************************************************************************
   VARIABLES LOCALES DE GESTION DES LIENS ENTRE TERMES
*************************************************************************/
TabTermedeLien                     TabTermesdeLiens;
TabLien                            TabLiens;

/************************************************************************
           LECTURE D'UN FICHIER DE LIENS TRM CONNU / TRM ACQUIS
*************************************************************************/
int      LireLiensTermes( FILE *p_ptFicLiens , long *p_nbLiens , 
                          char *p_listeVar , int p_noEtape , 
			  int p_enLesTriant )
{
     char                    carLu;
     char                    trouve;
     char                    identTermeIni[ _MaxLetIdent +1 ];
     char                    identTermeAcquis[ _MaxLetIdent +1 ];
     long                    idTermeIniCou;
     long                    idTermeAcquisCou;
     char                    variationCou;
     long                    indLienCou;
     long                    indLienAux;
     long                    resul;
     TabLien                 ptLien;
     int                     indCar;
     char                    tailleTerme[ _MaxLetIdent + 1 ];
     int                     noMetaRegleCou;
     char                    identMetaregle[ _MaxLetIdent +1 ];

     carLu = getc( p_ptFicLiens );
     while( carLu != EOF && (*p_nbLiens) < NbMaxLiens )
     {
         /********************/
         /* IDent Trm INI    */
         /********************/
         for( indCar = 0 ; indCar < _MaxLetIdent ; indCar++ )
         {
             if( !_Num( carLu ) && carLu != '\t' )
             {
	       TxtErreur( 270 , "LireLiensTermes(althesau)" );
	       return _False;
             }
             if( carLu == '\t' )
             {
                  break;
             }
             identTermeIni[ indCar ] = carLu;
             carLu = getc( p_ptFicLiens );
         }
         if( carLu != '\t' )
         {
	   TxtErreur( 271 , "LireLiensTermes(althesau)" );
	   return _False;
         }
         identTermeIni[ indCar ] = '\0';
         carLu = getc( p_ptFicLiens );
/* printf("%s\n",identTermeIni); */

         /********************/
         /* IDent Trm Nouv   */
         /********************/
         for( indCar = 0 ; indCar < _MaxLetIdent ; indCar++ )
         {
             if( !_Num( carLu ) && carLu != '\t' )
             {
	       TxtErreur( 272 , "LireLiensTermes(althesau)" );
	       return _False;
             }
             if( carLu == '\t' )
             {
                  break;
             }
             identTermeAcquis[ indCar ] = carLu;
             carLu = getc( p_ptFicLiens );
         }
         if( carLu != '\t' )
         {
	   TxtErreur( 273 , "LireLiensTermes(althesau)" );
	   return _False;
         }
         identTermeAcquis[ indCar ] = '\0';
         carLu = getc( p_ptFicLiens );
/* printf("%s\n",identTermeAcquis); */

         /********************/
         /* METAREGLE        */
         /********************/
         LireIdentificateur( p_ptFicLiens , &carLu , tailleTerme , 
                             _MaxLetIdent , _True );
         if( !ReconnaitreCaractere( p_ptFicLiens , ',' , &carLu ,
                                     _True ) )
         {
             break;
         }
         noMetaRegleCou = (int)LireEntier( p_ptFicLiens, &carLu );
         if( !ReconnaitreCaractere( p_ptFicLiens , ',' , &carLu ,
                                    _True ) )
         {
             break;
         }
         LireIdentificateur( p_ptFicLiens, &carLu , (char *)identMetaregle , 
                             (int)_MaxLetIdent , (char)_False );
	 /* saut jusqu'a la ligne suivante */
	 while( carLu != '\n' && carLu != EOF ) {
	   carLu = getc( p_ptFicLiens );
	 }
	 if( carLu == '\n' ) {
	   carLu = getc( p_ptFicLiens );
	 }

         idTermeIniCou = atol( identTermeIni );
         idTermeAcquisCou = atol( identTermeAcquis );
         if( strcmp( identMetaregle , "Coor" ) == 0 )
         {
             if( strchr( p_listeVar , 'C' ) )
             {
                   variationCou = 'C';
             }
             else
             {
                   variationCou = ' ';
             }
         }
         else if( strcmp( identMetaregle , "Ins" ) == 0
		  || strcmp( identMetaregle , "Modif" ) == 0  )
         {
             if( strchr( p_listeVar , 'I' ) )
             {
                   variationCou = 'I';
             }
             else
             {
                   variationCou = ' ';
             }
         }
         else if( strcmp( identMetaregle , "Perm" ) == 0 )
         {
             if( strchr( p_listeVar , 'P' ) )
             {
                   variationCou = 'P';
             }
             else
             {
                   variationCou = ' ';
             }
         }
         else /* identMetaregle inconnu */
         {
            variationCou = ' ';
            if( !TraitementBatch )
            {
	      sprintf( ChAux , "LireLiensTermes(althesau) [%s]\n" ,
                       identMetaregle );
	      TxtErreur( 274 , ChAux );
	      return _False;
            }
         }
/* printf("[%c]\n",variationCou); */

         /* variation a prendre en compte */
         if( variationCou != ' ' )
         {
             if( p_enLesTriant )
             {
                 trouve = _False;
                 ptLien = TabLiens + (*p_nbLiens);
                 for( indLienCou = (*p_nbLiens) ; 
                      indLienCou > 0L ; 
                      indLienCou-- )
                 {
                     resul = idTermeIniCou - (ptLien - 1)->idTermeIni;
                     if( resul == 0L
                         && idTermeAcquisCou == (ptLien - 1)->idTermeAcquis
                         && variationCou == (ptLien - 1)->variation )
                     {
                          trouve = _True;
                          break;
                     }
                     else if( resul > 0L )
                     {
                          break;
                     }
                     ptLien--;
                 }
                 if( !trouve )
                 {
                     for( indLienAux = (*p_nbLiens) ; 
                          indLienAux > indLienCou ; 
                          indLienAux-- )
                     {
                         (TabLiens + indLienAux)->idTermeIni
                            = (TabLiens + indLienAux - 1)->idTermeIni;
                         (TabLiens + indLienAux)->idTermeAcquis
                            = (TabLiens + indLienAux - 1)->idTermeAcquis;
                         (TabLiens + indLienAux)->variation
                            = (TabLiens + indLienAux - 1)->variation;
                         (TabLiens + indLienAux)->etapeAcquisition
                            = (TabLiens + indLienAux - 1)->etapeAcquisition;
                     }
                     (TabLiens + indLienCou)->idTermeIni = idTermeIniCou;
                     (TabLiens + indLienCou)->idTermeAcquis = idTermeAcquisCou;
                     (TabLiens + indLienCou)->variation = variationCou;
                     (TabLiens + indLienCou)->etapeAcquisition 
		       = (char)p_noEtape;
                     (*p_nbLiens)++;
		     if( *p_nbLiens >= NbMaxLiens - 1 ) {
		       TxtErreur( 155 , "LireLiensTermes(althesau)" );
		       return _False;
		     }
                  }
              }
              else
              {
                 (TabLiens + (*p_nbLiens))->idTermeIni = idTermeIniCou;
                 (TabLiens + (*p_nbLiens))->idTermeAcquis = idTermeAcquisCou;
                 (TabLiens + (*p_nbLiens))->variation = variationCou;
                 (TabLiens + (*p_nbLiens))->etapeAcquisition = (char)p_noEtape;
                 (*p_nbLiens)++;
		 if( *p_nbLiens >= NbMaxLiens - 1 ) {
		   TxtErreur( 155 , "LireLiensTermes(althesau)" );
		   return _False;
		 }
              }

              if( (*p_nbLiens) % 100L == 0 )
              {
                sprintf( ChAux , "%ld" , (*p_nbLiens) );
                TxtEcritItemInfo( 2 , ChAux );
              }
          }
     }

     if( carLu != EOF )
     {
          TxtErreur( 275 , "LireLiensTermes(althesau)" );
	  return _False;
     }
     return _True;
}

/************************************************************************
           MISE EN EVIDENCE DES SEQUENCES D'ACQUISITION
*************************************************************************/
int      LireSerieDeLiens( char *p_nomFicLiensNouveaux , 
			   char *p_nomFicLiensAnciens , 
                           long *p_nbLiens , char *p_listeVar , 
			   int p_enLesTriant )
{
     FILE                   *ptFicLiensAnc;
     FILE                   *ptFicLiensNou;
     int                     noEtape;
     
     TxtTitresDialogueInfo( 43 , 45 , 44 );

     for( noEtape = 1 ; noEtape < 100 ; noEtape++ ) {
       sprintf( ChAux , "%s.%d.link" , p_nomFicLiensNouveaux , noEtape );
       ptFicLiensNou = fopen( ChAux , _FopenRT );
       if( ptFicLiensNou ) {
	 TxtEcritItemInfo( 1 , ChAux );
	 if( !LireLiensTermes( ptFicLiensNou , p_nbLiens , p_listeVar , 
			       noEtape , p_enLesTriant ) ) {
	   fclose( ptFicLiensNou );
	   return _False;
	 }
	 fclose( ptFicLiensNou );
       }
	 
       if( p_nomFicLiensAnciens ) {
	 sprintf( ChAux , "%s.%d.link" , p_nomFicLiensAnciens , noEtape );
	 ptFicLiensAnc = fopen( ChAux , _FopenRT );
	 if( ptFicLiensAnc ) {
	   TxtEcritItemInfo( 1 , ChAux );
	   if( !LireLiensTermes( ptFicLiensAnc , p_nbLiens , p_listeVar , 
				 noEtape , p_enLesTriant ) ) {
	     fclose( ptFicLiensAnc );
	     return _False;
	   }
	   fclose( ptFicLiensAnc );
	 }
       }
	 
       if( !ptFicLiensNou && !ptFicLiensAnc ) {
	 break;
       }
     }

     return _True;
}

/************************************************************************
           ECRITURE DES SEQUENCES D'ACQUISITION
*************************************************************************/
void     EcrireSequenceAcquisition( struct SequenceDeLiens  *p_ptChDeLiens , 
                                    int p_noEtape , long p_nbLiens , 
                                    char *p_listeVar , 
				    FILE *p_ptFicSequences )
{
     long                          indLienCou;
     long                          nouveauEtapePrec;
     TabLien                       ptLien;
     struct SequenceDeLiens        seqDeLienCou;
     TabSequenceDeLiens            ptChDeLienAux;

     /* recherche du lien de l'etape precedente */
     ptChDeLienAux = p_ptChDeLiens;
     while( ptChDeLienAux->lienSui )
     {
         ptChDeLienAux = ptChDeLienAux->lienSui;
     }
     ptChDeLienAux->lienSui = &seqDeLienCou;
     nouveauEtapePrec = (ptChDeLienAux->lien)->idTermeAcquis;
     
     ptLien = TabLiens;
     /* recherche d'une variation poursuivant l'acquisition */
     for( indLienCou = 0 ; indLienCou < p_nbLiens ; indLienCou++ )
     {
         if( ptLien->idTermeIni > nouveauEtapePrec )
         {
            break;
         }
         else if( ptLien->idTermeIni == nouveauEtapePrec
                  && ptLien->etapeAcquisition == p_noEtape
                  && strchr( p_listeVar , ptLien->variation ) )
         {
              seqDeLienCou.lien = ptLien;
              seqDeLienCou.lienSui = NULL;
              EcrireSequenceAcquisition( p_ptChDeLiens , p_noEtape + 1 , 
                                         p_nbLiens , p_listeVar , 
					 p_ptFicSequences );
         }

         ptLien++;
     }
     ptChDeLienAux->lienSui = NULL;

     /* ecriture de la chaine d'acquisition courante 1: variations */
     ptChDeLienAux = p_ptChDeLiens;
     while( ptChDeLienAux )
     {
         fprintf( p_ptFicSequences , "%c" , 
                  (ptChDeLienAux->lien)->variation );
         ptChDeLienAux = ptChDeLienAux->lienSui;
     }
     fprintf( p_ptFicSequences , "\t" );

     /* ecriture de la chaine d'acquisition courante 2: termes */
     ptChDeLienAux = p_ptChDeLiens;
     while( ptChDeLienAux )
     {
         fprintf( p_ptFicSequences , "%06ld\t" , 
                  (ptChDeLienAux->lien)->idTermeIni );
         if( !(ptChDeLienAux->lienSui) )
         {
             fprintf( p_ptFicSequences , "%06ld\n" , 
                      (ptChDeLienAux->lien)->idTermeAcquis );
         }
         ptChDeLienAux = ptChDeLienAux->lienSui;
     }
}

/************************************************************************
           MISE EN EVIDENCE DES SEQUENCES D'ACQUISITION
*************************************************************************/
void     SequencesAcquisition( char *p_nomFicLiensNouveaux , 
			       char *p_nomFicSequences , 
                               char *p_listeVar )
{
     FILE                     *ptFicSequences;
     long                      tailleLiens;
     long                      indLienCou;
     long                      nbLiens;
     TabLien                   ptLien;
     struct SequenceDeLiens    seqDeLiens;
     
     if( strcmp( p_nomFicSequences , "stdout" ) == 0 ) {
       ptFicSequences = stdout;
     }
     else {
       if( !OuvreFichier( p_nomFicSequences , &ptFicSequences , "w" ) ) {
	   return;
	 }
     }

     tailleLiens = (unsigned long) sizeof( struct Lien )
                              * (unsigned long) NbMaxLiens;
     TabLiens = (TabLien)_Alloc( tailleLiens );
     if( !TabLiens )
     {
       TxtErreur( 273 , "SequencesAcquisition(althesau)" );
       fclose( ptFicSequences );
       return;
     }

     nbLiens = 0L;
     LireSerieDeLiens( p_nomFicLiensNouveaux , "" , 
		       &nbLiens , p_listeVar  , _True );

     TxtTitresDialogueInfo( 46 , 44 , 0 );

     ptLien = TabLiens;
     /* ecriture du resultat */
     for( indLienCou = 0 ; indLienCou < nbLiens ; indLienCou++ )
     {
         if( ptLien->etapeAcquisition == 1 )
         {
            seqDeLiens.lien = ptLien;
            seqDeLiens.lienSui = NULL;
            EcrireSequenceAcquisition( &seqDeLiens , 
				       ptLien->etapeAcquisition + 1 , 
                                       nbLiens , p_listeVar , ptFicSequences );
         }
         ptLien++;

         if( indLienCou % 100L == 0 )
         {
             sprintf( ChAux , "%ld" , indLienCou );
             TxtEcritItemInfo( 1 , ChAux );
         }
    }

     _Free( TabLiens );
     fclose( ptFicSequences );
}

/************************************************************************
           ECRITURE DES SEQUENCES QUELCONQUES
*************************************************************************/
void     Ecrire1SeqDUnGroupe( TabSequenceDeLiens p_ptChDeLiens , 
							 int p_profondeur , long p_nbLiens , 
                                     char *p_listeVar , FILE *p_ptFicSequences )
{
     long                           indLienCou;
     long                           nouveauEtapePrec;
     TabLien                        ptLien;
     struct SequenceDeLiens         seqDeLienCou;
     TabSequenceDeLiens             ptChDeLienAux;
     char                           uneEcriture;

     /* recherche du lien de l'etape precedente */
     ptChDeLienAux = p_ptChDeLiens;
     while( ptChDeLienAux->lienSui )
     {
         ptChDeLienAux = ptChDeLienAux->lienSui;
     }
     ptChDeLienAux->lienSui = &seqDeLienCou;
     if( (ptChDeLienAux->lien)->aLEndroit == _True )
     {
         nouveauEtapePrec = (ptChDeLienAux->lien)->idTermeAcquis;
     }
     else
     {
         nouveauEtapePrec = (ptChDeLienAux->lien)->idTermeIni;
     }
     
     /* inutile de verifier que l'on ne parcourt pas des cycles puisque  */
     /* les liens ne sont parcourus qu'une fois                          */

     ptLien = TabLiens;
     /* recherche d'une variation poursuivant l'acquisition */
     for( indLienCou = 0 ; indLienCou < p_nbLiens ; indLienCou++ )
     {
         if( ptLien->idTermeIni > nouveauEtapePrec )
         {
            break;
         }
         else if( !(ptLien->parcouru)
                  && ptLien->idTermeIni == nouveauEtapePrec
                  && strchr( p_listeVar , ptLien->variation ) )
         {
              seqDeLienCou.lien = ptLien;
              seqDeLienCou.lienSui = NULL;
              ptLien->parcouru = _True;
              ptLien->aLEndroit = _True;

              Ecrire1SeqDUnGroupe( p_ptChDeLiens , p_profondeur + 1 , 
				   p_nbLiens , p_listeVar , p_ptFicSequences );
         }
         else if( !(ptLien->parcouru)
                  && ptLien->idTermeAcquis == nouveauEtapePrec
                  && strchr( p_listeVar , ptLien->variation ) )
         {
              seqDeLienCou.lien = ptLien;
              seqDeLienCou.lienSui = NULL;
              ptLien->parcouru = _True;
              ptLien->aLEndroit = _False;

              Ecrire1SeqDUnGroupe( p_ptChDeLiens , p_profondeur + 1 , 
				   p_nbLiens , p_listeVar , p_ptFicSequences );
         }

         ptLien++;
     }
     ptChDeLienAux->lienSui = NULL;

     /* if( p_profondeur > 1 ) */
     {
         /* ecriture de la chaine d'acquisition courante 1: variations */
         uneEcriture = _False;
         ptChDeLienAux = p_ptChDeLiens;
         while( ptChDeLienAux )
         {
             if( (ptChDeLienAux->lien)->parcouru != _Ecrit )
             {
                 uneEcriture = _True;
                 if( (ptChDeLienAux->lien)->aLEndroit )
                 {
                     fprintf( p_ptFicSequences , "%c" , 
                              (ptChDeLienAux->lien)->variation );
                 }
                 else
                 {
                     fprintf( p_ptFicSequences , "%c" , 
                              _ToLower( (ptChDeLienAux->lien)->variation ) );
                 }
             }
             ptChDeLienAux = ptChDeLienAux->lienSui;
         }

         /* ecriture de la chaine d'acquisition courante 2: termes */
         ptChDeLienAux = p_ptChDeLiens;
         while( ptChDeLienAux )
         {
             if( (ptChDeLienAux->lien)->parcouru != _Ecrit )
             {
                 if( (ptChDeLienAux->lien)->aLEndroit )
                 {
                     fprintf( p_ptFicSequences , "\t%06ld" , 
                              (ptChDeLienAux->lien)->idTermeIni );
                 }
                 else
                 {
                     fprintf( p_ptFicSequences , "\t%06ld" , 
                              (ptChDeLienAux->lien)->idTermeAcquis );
                 }
             }
             if( (ptChDeLienAux->lien)->parcouru != _Ecrit
                 && !(ptChDeLienAux->lienSui) )
             {
                 if( (ptChDeLienAux->lien)->aLEndroit )
                 {
                     fprintf( p_ptFicSequences , "\t%06ld" , 
                              (ptChDeLienAux->lien)->idTermeAcquis );
                 }
                 else
                 {
                     fprintf( p_ptFicSequences , "\t%06ld" , 
                              (ptChDeLienAux->lien)->idTermeIni );
                 }
             }
             (ptChDeLienAux->lien)->parcouru = _Ecrit;
             ptChDeLienAux = ptChDeLienAux->lienSui;
         }
         if( uneEcriture )
         {
             fprintf( p_ptFicSequences , "\n" );
         }
     }
}

/************************************************************************
           MISE EN EVIDENCE DES SEQUENCES SUR UNE VARIATION
*************************************************************************/
void     GroupesVariation( char *p_nomFicLiensNouveaux , 
			   char *p_nomFicLiensAnciens , 
			   char *p_nomFicSequences , 
			   char *p_listeVar )
{
     FILE                      *ptFicSequences;
     long                       tailleLiens;
     long                       indLienCou;
     long                       nbLiens;
     TabLien                    ptLien;
     struct SequenceDeLiens     seqDeLiens;
     
     if( strcmp( p_nomFicSequences , "stdout" ) == 0 ) {
       ptFicSequences = stdout;
     }
     else {
       if( !OuvreFichier( p_nomFicSequences , &ptFicSequences , "w" ) ) {
	   return;
	 }
     }

     tailleLiens = (unsigned long) sizeof( struct Lien )
                              * (unsigned long) NbMaxLiens;
     TabLiens = (TabLien)_Alloc( tailleLiens );
     if( !TabLiens )
     {
       TxtErreur( 273 , "GroupesVariation(althesau)" );
       fclose( ptFicSequences );
       return;
     }

     nbLiens = 0L;
     LireSerieDeLiens( p_nomFicLiensNouveaux , p_nomFicLiensAnciens , 
                       &nbLiens , p_listeVar  , _True );
     ptLien = TabLiens;
     for( indLienCou = 0 ; indLienCou < nbLiens ; indLienCou++ )
     {
         ptLien->parcouru = _False;
         ptLien->aLEndroit = _True;
         ptLien++;
     }

     TxtTitresDialogueInfo( 46 , 44 , 0 );

     ptLien = TabLiens;
     /* ecriture du resultat */
     for( indLienCou = 0 ; indLienCou < nbLiens ; indLienCou++ )
     {
         seqDeLiens.lien = ptLien;
         seqDeLiens.lienSui = NULL;
         if( !(ptLien->parcouru) )
         {
             ptLien->parcouru = _True;
             ptLien->aLEndroit = _True;
             Ecrire1SeqDUnGroupe( &seqDeLiens , 1 , nbLiens , 
                                         p_listeVar , ptFicSequences );
             fprintf( ptFicSequences , "\n" );
         }
         ptLien++;

         if( indLienCou % 100L == 0 )
         {
             sprintf( ChAux , "%ld" , indLienCou );
             TxtEcritItemInfo( 1 , ChAux );
         }
    }

     fclose( ptFicSequences );
     _Free( TabLiens );
}

/***************************************************************************
                   lecture des chaines de termes
***************************************************************************/
void     LireIdDesTermes( FILE *p_ptFicTermes , long *p_nbTermes )
{
     long                      idTerme;
     TabTermedeLien            ptTermedeLien;
     long                      indTermesCou;
     long                      indTermesAux;
     long                      valcomp;
     char                      trouve;
     char                      ch[ 256 ];

     while( fgets( ch , 255 , p_ptFicTermes ) && (*ch != '\n') 
            && (*p_nbTermes) < NbMaxChainesTermes )
     {
         sscanf( ch , "%ld" , &idTerme );

         /* recherche du point d'insertion du terme pour */
         /* un classement par ordre croissant */
         trouve = _False;
         ptTermedeLien = TabTermesdeLiens + (*p_nbTermes);
         for( indTermesCou = (*p_nbTermes) ; 
              indTermesCou > 0 ; 
              indTermesCou-- )
         {
             valcomp =  idTerme - (ptTermedeLien - 1)->id;
             if( valcomp == 0 )
             {
                  trouve = _True;
                  break;
             }
             if( valcomp > 0 )
             {
                  break;
             }
             ptTermedeLien--;
         }
         if( !trouve )
         {
             for( indTermesAux = (*p_nbTermes) ; 
                  indTermesAux > indTermesCou ; 
                  indTermesAux-- )
             {
                 (TabTermesdeLiens + indTermesAux)->id 
                             = (TabTermesdeLiens + indTermesAux - 1)->id;
             }
             (TabTermesdeLiens + indTermesCou)->id = idTerme;
             (*p_nbTermes)++;
             if( (*p_nbTermes) % 100L == 0 )
             {
                 sprintf( ChAux , "%ld" , (*p_nbTermes) );
                 TxtEcritItemInfo( 1 , ChAux );
             }
         }
     }
     if( (*p_nbTermes) >= NbMaxChainesTermes )
     {
          TxtErreur( 153 , "LireIdDesTermes(posthesau)" );
     }
     fclose( p_ptFicTermes );
}

/************************************************************************
           VOLUMES D'ACQUISITION SUR SOUS-ENSEMBLE DE L'ENSEMBLE INITIAL
*************************************************************************/
void     AcquisitionSurSousEnsemble( char *p_nomFicTermes , 
				     char *p_nomFicLiensNouveaux , 
				     char *p_nomFicLiensAnciens , 
				     char *p_nomFicAcquisitions ,
				     long  p_tailleSousEnsemble , 
				     char *p_listeVar )
{
     FILE                      *ptFicAcquisitions;
     FILE                      *ptFicTerme;
     long                       tailleLiens;
     long                       tailleTermes;
     long                       indLienCou;
     long                       indTerme;
     long                       indTermeAux;
     long                       indTermeCour;
     long                       nbPas;
     long                       nbLiensTot;
     long                       nbTermesTot;
     long                       nbTermesPrec;
     long                       nbTermesInitiauxMax;
     long                       nbTermesAcquis;
     long                       nbTermesAcquisTot;
     long                       idTermeConnu;
     long                       idTermeIniCou;
     long                       idTermeCandidat;
     TabLien                    ptLien;
     TabTermedeLien             ptTermedeLien;
     TabTermedeLien             ptTermedeLienAux;
     int                        noEtape;
     char                       trouve;
     long                       cumulAcquisParEtape[ 40 ];
     int                        noEtapeMax;
     
     if( !OuvreFichier( p_nomFicAcquisitions , &ptFicAcquisitions , "w" ) )
     {
         return;
     }

     tailleLiens = (unsigned long) sizeof( struct Lien )
                              * (unsigned long) NbMaxLiens;
     TabLiens = (TabLien)_Alloc( tailleLiens );
     tailleTermes = (unsigned long) sizeof( struct TermedeLien )
                              * (unsigned long) NbMaxChainesTermes;
     TabTermesdeLiens = (TabTermedeLien)_Alloc( tailleTermes );
     if( !TabLiens || !TabTermesdeLiens )
     {
          printf( "allocation memoire TabLiens et TabTermesdeLiens impossible\n" );
          TxtAttente();
          return;
     }
     
     fprintf( ptFicAcquisitions , "Variations\t%s\n\n" , p_listeVar );

     /*******************
     lecture des termes 
     ********************/
     ptTermedeLien = TabTermesdeLiens;
     for( indTerme = 0 ; indTerme < NbMaxChainesTermes ; indTerme++ )
     {
         ptTermedeLien->connuAEtape = (char)0;
         ptTermedeLien++;
     }
     
     nbTermesTot = 0;
     nbTermesPrec = 0;
     TxtTitresDialogueInfo( 38 , 39 , 45 );
     for( noEtape = 1 ; noEtape < 100 ; noEtape++ )
     {
         sprintf( ChAux , "%s%02d" , p_nomFicTermes , noEtape );
         ptFicTerme = fopen( ChAux , _FopenRT );
         if( !ptFicTerme )
         {
             break;
         }
         TxtEcritItemInfo( 2 , ChAux );
         LireIdDesTermes( ptFicTerme , &nbTermesTot );
         fclose( ptFicTerme );
         
         if( noEtape == 1 )
         {
             nbTermesInitiauxMax = nbTermesTot;
         }
         fprintf( ptFicAcquisitions , "Etape\t%d\t%ld\n" , noEtape , 
                                      nbTermesTot - nbTermesPrec );
         nbTermesPrec = nbTermesTot;
     }
     fprintf( ptFicAcquisitions , "Total acquisition\t%ld\n" , 
              nbTermesTot - nbTermesInitiauxMax );

     TxtTitresDialogueInfo( 41 , 39 , 0 );

     /*********************************************************
     selection d'un sous-ensemble des termes initiaux 
     **********************************************************/
     indTermeCour = 0L;
     if( p_tailleSousEnsemble > nbTermesInitiauxMax )
     {
          TxtErreur( 156 , "AcquisitionSurSousEnsemble(posthesau)" );
          return;
     }
     for( indTerme = 0L ; indTerme < p_tailleSousEnsemble ; indTerme++ )
     {
         nbPas = (long)(( (double)rand() / (double)(RAND_MAX) ) 
                        * (double) nbTermesInitiauxMax ) % nbTermesInitiauxMax;
                        
         indTermeAux = (indTermeAux + nbPas) % nbTermesInitiauxMax;
         ptTermedeLien = TabTermesdeLiens + indTermeAux;

         while( ptTermedeLien->connuAEtape == (char)1 )
         {
             indTermeAux++;
             if( indTermeAux < nbTermesInitiauxMax )
             {
                 ptTermedeLien++;
             }
             else
             {
                 indTermeAux = 0L;
                 ptTermedeLien = TabTermesdeLiens;
             }
         }
         (TabTermesdeLiens + indTermeAux)->connuAEtape = (char)1;
         indTermeCour = indTermeAux;

         if( indTerme % 100L == 0 )
         {
             sprintf( ChAux , "%ld" , indTerme );
             TxtEcritItemInfo( 1 , ChAux );
         }
     }

     /*******************
        VERIFICATION 
     ********************/
     {
         long nbTermesConnus = 0L;
         long nbTermesInconnus = 0L;

         ptTermedeLien = TabTermesdeLiens;
         for( indTerme = 0 ; indTerme < nbTermesInitiauxMax ; indTerme++ )
         {
             if( ptTermedeLien->connuAEtape == (char)0 )
             {
                   nbTermesInconnus++;
             }
             else if( ptTermedeLien->connuAEtape == (char)1 )
             {
                   nbTermesConnus++;
             }
             else
             {
                   printf( "Erreur d'algorithme\n" );
             }
             ptTermedeLien++;
         }
         if( nbTermesConnus != p_tailleSousEnsemble )
         {
             printf( "Erreur d'algorithme\n" );
         }
         fprintf( ptFicAcquisitions , 
		  "\nTermes connus\t%ld\n" , nbTermesConnus );
         fprintf( ptFicAcquisitions , 
		  "Termes inconnus\t%ld\n\n" , nbTermesInconnus );
     }

     /*******************
     lecture des liens 
     ********************/
     ptLien = TabLiens;
     for( indLienCou = 0 ; indLienCou < NbMaxLiens ; indLienCou++ )
     {
         ptLien->parcouru = _False;
         ptLien++;
     }

     nbLiensTot = 0L;
     LireSerieDeLiens( p_nomFicLiensNouveaux , p_nomFicLiensAnciens , 
                       &nbLiensTot , p_listeVar , _True );

     /*******************
     acquisition incrementale
     ********************/
     TxtTitresDialogueInfo( 53 , 39 , 54 );
 
     noEtape = 1;
     nbTermesAcquis = p_tailleSousEnsemble;
     nbTermesAcquisTot = 0L;
     /* une etape d'acquisition */
     while( nbTermesAcquis != 0L )
     {
         fprintf( ptFicAcquisitions , "Etape\t%d\t%ld\n" , noEtape , 
                                      nbTermesAcquis );
         if( noEtape < 40 )
         {
             cumulAcquisParEtape[ noEtape ] = nbTermesAcquis;
         }
         nbTermesAcquis = 0L;
         noEtape++;
         sprintf( ChAux , "%d" , noEtape );
         TxtEcritItemInfo( 2 , ChAux );
         
         /* parcours de tous les termes acquis a l'etape precedente */
         ptTermedeLien = TabTermesdeLiens;
         for( indTerme = 0 ; indTerme < nbTermesTot ; indTerme++ )
         {
             /* terme connu a l'etape precedente */
             if( ptTermedeLien->connuAEtape == (char)(noEtape - 1) )
             {
                   idTermeConnu = ptTermedeLien->id;

                   /* recherche d'un lien ayant pour terme initial 
		      ce terme connu */
                   ptLien = TabLiens;
                   for( indLienCou = 0L ; indLienCou < nbLiensTot ; 
			indLienCou++ )
                   {
                       /* lien ayant pour terme initial ce terme connu */
                       idTermeIniCou = ptLien->idTermeIni;
                       if( idTermeIniCou == idTermeConnu )
                       {
                           idTermeCandidat = ptLien->idTermeAcquis;

                           /* recherche de ce candidat parmi les termes */
                           trouve = _False;
                           ptTermedeLienAux = TabTermesdeLiens;
                           for( indTermeAux = 0 ; 
                                indTermeAux < nbTermesTot ; 
                                indTermeAux++ )
                           {
                               if( ptTermedeLienAux->id == idTermeCandidat )
                               {
                                  trouve = _True;
                                  break;
                               }  
                               ptTermedeLienAux++;                             
                           }

                           if( !trouve && idTermeCandidat )
                           {
                              printf( "Erreur d'algorithme\n" );
                           }
                           if( ptTermedeLienAux->connuAEtape == 0 )
                           {
                                ptTermedeLienAux->connuAEtape = noEtape;
                                nbTermesAcquis++;
                                if( nbTermesAcquis % 100L == 0 )
                                {
                                    sprintf( ChAux , "%ld" , nbTermesAcquis );
                                    TxtEcritItemInfo( 1 , ChAux );
                                }
                           }
                       }
                       else if( idTermeIniCou > idTermeConnu )
                       {
                           break;
                       }

                       ptLien++;
                       /* lien ayant pour terme initial ce terme connu */
                   }
                   /* recherche d'un lien ayant pour terme initial 
		      ce terme connu */
             }
             /* terme connu a l'etape precedente */

             ptTermedeLien++;
         }
         /* parcours de tous les termes acquis a l'etape precedente */

         nbTermesAcquisTot += nbTermesAcquis;
     }
     /* une etape d'acquisition */
     fprintf( ptFicAcquisitions , "Etape\t%d\t%ld\n" , noEtape , 
                                  nbTermesAcquis );
     fprintf( ptFicAcquisitions , "Total acquisition\t%ld\n" , 
                                  nbTermesAcquisTot );
     fprintf( ptFicAcquisitions , "\n" );
     noEtapeMax = noEtape;

     for( noEtape = 2 ; noEtape < noEtapeMax - 1 ; noEtape++ )
     {
         cumulAcquisParEtape[ noEtape + 1 ] += cumulAcquisParEtape[ noEtape ];
     }
     for( noEtape = 2 ; noEtape < noEtapeMax ; noEtape++ )
     {
         fprintf( ptFicAcquisitions , "%ld\n" , 
		  cumulAcquisParEtape[ noEtape ] );
     }
     fprintf( ptFicAcquisitions , "\n" );

     /*******************
        VERIFICATION 
     ********************/
     {
         long nbTermesAcquisINIST = 0L;
         long nbTermesAcquisNouveaux = 0L;

         ptTermedeLien = TabTermesdeLiens;
         for( indTerme = 0 ; indTerme < nbTermesTot ; indTerme++ )
         {
             if( ptTermedeLien->connuAEtape > (char)1 )
             {
                 if( indTerme < nbTermesInitiauxMax )
                 {
                    nbTermesAcquisINIST++;
                 }
                 else
                 {
                    nbTermesAcquisNouveaux++;
                 }
             }
             ptTermedeLien++;
         }

         fprintf( ptFicAcquisitions , "\nTermes acquis INIST\t%ld\n" , 
                  nbTermesAcquisINIST );
         fprintf( ptFicAcquisitions , "Termes acquis nouveaux\t%ld\n\n" , 
                  nbTermesAcquisNouveaux );
     }

     fclose( ptFicAcquisitions );
     _Free( TabLiens );
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                        fin de althesau.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
