/*
 * FASt Term Recognizer
 *
 *  FASTR/altrfime.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for converting file data into memory data
 */

#include "al-ext.h"
#include "al-var.h"
#include "al-fic.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/************************************************************************
    passage fichier ---> memoire (pour une arborescence de termes)
*************************************************************************/
TabTermeMem RestructurerArbre( long p_indTeteLemme , TabChar p_bufferChar ,
                               long *p_ind , int p_fils )
{
    TabTermeFic                ptTermeFic;
    TabTermeMem                ptTermeMem;

    /* terme existant */
    if( p_fils != _Nil )
    {
        if( p_fils != _RefEnAvant )
        {
           TxtErreur( 56 , 
              "RestructurerArbre(altrfime) non corresponding indexes 1" );
        }
        ptTermeMem = NouveauTerme();
        if( !ptTermeMem )
        {
             return NULL;
        }
    }
    /* terme existant */
    else
    /* terme inexistant */
    {
        return NULL;
    }
    /* terme inexistant */

    /* recopie du terme courant */
    ptTermeFic = (TabTermeFic)(p_bufferChar + *p_ind);
    ptTermeMem->ValTerme.Ref = ptTermeFic->Ref;
    ptTermeMem->ValTerme.RefRacine = ptTermeFic->RefRacine;
    ptTermeMem->ValTerme.RefSemantique = NULL;
    ptTermeMem->ValTerme.RefSelf = NULL;
    ptTermeMem->ValTerme.TypeTerme = ptTermeFic->TypeTerme;
    ptTermeMem->ValTerme.PosFicRegle = ptTermeFic->PosFicRegle;

    *p_ind += (long)sizeof( struct TermeFic );
    while( (*p_ind) % _TailleMot != 0L )
    {
        (*p_ind)++;
    }

    /***************** recopie du dag du terme *********************/
    ptTermeMem->ValTerme.ArcsFils = 
               RestructurerDag( p_indTeteLemme , p_bufferChar , 
                                p_ind , (int)ptTermeFic->ArcsFils );

    /***************** appel recursif sur l'arborescence ***********/
    ptTermeMem->TermeFils
            = RestructurerArbre( p_indTeteLemme , p_bufferChar , p_ind ,
                                 (int)ptTermeFic->TermeFils );
    ptTermeMem->TermeFrere
            = RestructurerArbre( p_indTeteLemme , p_bufferChar , p_ind ,
                                 (int)ptTermeFic->TermeFrere );

    return ptTermeMem;
}

/************************************************************************
    passage fichier ---> memoire (pour un dag)
*************************************************************************/
TabDagMem RestructurerDag( long p_indTeteLemme , TabChar p_bufferChar ,
                           long *p_ind , int p_fils )
{
    TabDagFic               ptNoeudFic;
    struct OperLogFic      *ptOperLogFic;
    TabDagMem               nouvNoeud;
    int                     ind;
    TabChar                 ch;
    int                     longCh;
    long                   *ptRefEnArriere;

    /* dag existant */
    if( p_fils != _Nil )
    {
       /* dag deja en memoire */
       if( p_fils == _RefEnArriere )
       {
          /* en cas de ref en arriere, celle-ci est donnee apres le dag */
          ptRefEnArriere = (long *)(p_bufferChar + *p_ind);
          *p_ind += (long)sizeof(long);
          while( (*p_ind) % _TailleMot != 0L )
          {
              (*p_ind)++;
          }

          /* pteur vers un dag deja passe en memoire */
          /* on retrouve son pteur memoire */
          for( ind = 0 ; ind < NbLiensNoeudsMemFic ; ind++ )
          {
              if( (LiensNoeudsMemFic + ind)->PtFic == *ptRefEnArriere )
              {
                  return (LiensNoeudsMemFic + ind)->PtMem;
              }
          }

          /* pointeur non retrouve */
          TxtErreur( 56 , 
                     "RestructurerDag(altrfime) backward dag not found" );
          printf("\nRef looked for %ld\n\n" , *ptRefEnArriere );
          printf("\nExisting refs (%d) : \n" , NbLiensNoeudsMemFic );
          for( ind = 0 ; ind < NbLiensNoeudsMemFic ; ind++ )
          {
               printf( " --->        %ld\n" , 
                       (LiensNoeudsMemFic + ind)->PtFic );
          }
          return NULL;
       }
       /* dag deja en memoire */
       else
       /* dag a creer en memoire */
       {
           if( p_fils != _RefEnAvant )
           {
               TxtErreur( 56 , 
                     "RestructurerDag(altrfime) Unknown dag type" );
           }

           ptNoeudFic = (TabDagFic)(p_bufferChar + *p_ind);
           nouvNoeud = NouveauDag( (int)(ptNoeudFic->TypeDagFic) );

           if( !nouvNoeud )
           {
               return NULL;
           }
       }
       /* dag a creer en memoire */
    }
    /* dag existant */
    else
    /* dag inexistant */
    {
         return NULL;
    }
    /* dag inexistant */

    /* dag non encore connu, on le copie */
    /* sauvegarde du lien du noeud courant */
    (LiensNoeudsMemFic + NbLiensNoeudsMemFic)->PtMem = nouvNoeud;
    (LiensNoeudsMemFic + NbLiensNoeudsMemFic)->PtFic = *p_ind - p_indTeteLemme;
    NbLiensNoeudsMemFic++;
    if( NbLiensNoeudsMemFic >= NbMaxLiensMemorises )
    {
         TxtErreur( 94 , "RestructurerDag(altrfime)" );
         return NULL;
    }

    /* recopie des informations du noeud courant */
    nouvNoeud->TypeDag = (char)(ptNoeudFic->TypeDagFic);
    *p_ind += (long)sizeof(struct DagFic);
    switch( nouvNoeud->TypeDag )
    {
        case _TypeDagVide : 
                    break;
        case _Arc : nouvNoeud->ValDag.Arc.ValSymb
                           = ptNoeudFic->ValDag.Arc.ValSymb;
                    *(nouvNoeud->ValDag.Arc.Ident) = '\0';
                    break;
        case _AtomeRef : 
                    if( ptNoeudFic->ValDag.Atome.AtomeConjDisj != _Nil )
                    {
                        nouvNoeud->ValDag.Atome.AtomeConjDisj
                              = ptNoeudFic->ValDag.Atome.AtomeConjDisj;
                    }
                    else
                    {
                        nouvNoeud->ValDag.Atome.AtomeConjDisj
                              = _False;
                    }
                    nouvNoeud->ValDag.Atome.ValAtome.Ref
                           = ptNoeudFic->ValDag.Atome.ValAtome.Ref;
                    break;
        case _AtomeEntier : 
                    if( ptNoeudFic->ValDag.Atome.AtomeConjDisj != _Nil )
                    {
                        nouvNoeud->ValDag.Atome.AtomeConjDisj
                              = ptNoeudFic->ValDag.Atome.AtomeConjDisj;
                    }
                    else
                    {
                        nouvNoeud->ValDag.Atome.AtomeConjDisj
                              = _False;
                    }
                    nouvNoeud->ValDag.Atome.ValAtome.Entier
                              = ptNoeudFic->ValDag.Atome.ValAtome.Entier;
                    break;
        case _AtomeChaine : 
                    if( ptNoeudFic->ValDag.Atome.AtomeConjDisj != _Nil )
                    {
                        nouvNoeud->ValDag.Atome.AtomeConjDisj
                              = ptNoeudFic->ValDag.Atome.AtomeConjDisj;
                    }
                    else
                    {
                        nouvNoeud->ValDag.Atome.AtomeConjDisj
                              = _False;
                    }
                    longCh = (int)ptNoeudFic->ValDag.Atome.ValAtome.LongChaine;
                    if( longCh == _Nil )
                    {
                        longCh = 0;
                    }
                    ch = nouvNoeud->ValDag.Atome.ValAtome.Chaine;
                    for( ind = 0 ; ind < longCh ; ind++ )
                    {
                        ch[ ind ] = p_bufferChar[ *p_ind ];
                        (*p_ind)++;
                    }
                    ch[ longCh ] = '\0';
                    break;
        case _AtomeListe :
                    if( ptNoeudFic->ValDag.Atome.AtomeConjDisj != _Nil )
                    {
                        nouvNoeud->ValDag.Atome.AtomeConjDisj
                              = ptNoeudFic->ValDag.Atome.AtomeConjDisj;
                    }
                    else
                    {
                        nouvNoeud->ValDag.Atome.AtomeConjDisj
                              = _False;
                    }
                    break;
        case _Oper :
                    while( (*p_ind) % _TailleMot != 0L )
                    {
                       (*p_ind)++;
                    }
                    ptOperLogFic = (struct OperLogFic *)(p_bufferChar + (*p_ind));
                    *p_ind += (long)sizeof(struct OperLogFic);
                    break;
        default :   sprintf( ChAux , 
                "RestructurerDag(altrfime) Noeud de type inconnu : %d\n" , 
                             nouvNoeud->TypeDag );
                    TxtErreur( 42 , ChAux );
                    break;
    }
    while( (*p_ind) % _TailleMot != 0L )
    {
       (*p_ind)++;
    }

   /***************** appel recursif sur l'arborescence ***********/
    nouvNoeud->Fils = RestructurerDag( p_indTeteLemme , p_bufferChar , 
                                       p_ind , (int)ptNoeudFic->Fils );
    nouvNoeud->Frere = RestructurerDag( p_indTeteLemme , p_bufferChar , 
                                       p_ind , (int)ptNoeudFic->Frere );
 
    return nouvNoeud;
}

/************************************************************************
                       passage fichier ----> memoire (lemmes)
*************************************************************************/
long     RestructurerLemme( TabChar p_bufferChar , long *p_indBufferChar , 
                            char *p_chaine , TabLemmeMem p_ptLemmeMem , 
                            char *p_chaineAnnexe)
{
    long                i;
    long                indTeteLemme;
    long                ind = *p_indBufferChar;
                            /* indice du tableau cible p_bufferChar */
    /* termes temporaires pour passage des donnees fichier aux donnees
       memoire*/
    TabLemmeFic         ptLemmeFic;

    /* recopie de la chaine                                         */
    if( p_bufferChar[ ind ] == '\0')
    {
        TxtErreur( 69 ,"Restructurer(altrfime)" );
    }

    if( Trace )
    {
        fprintf( FicLog , "<--F[%s]\n" , p_bufferChar + ind );
    }

    while( ind % _TailleMot != 0L )
    {
        ind++;
    }
    i = 0L;
    while( (p_chaine[ i ] = p_bufferChar[ ind ]) != '\0')
    {
         ind++;
         i++;
    }
    /* saut de la fin de chaine */
    ind++;
    while( ind % _TailleMot != 0L )
    {
        ind++;
    }

    indTeteLemme = ind;

    /* recopie du Lemme                                          */
    ptLemmeFic = (TabLemmeFic)(p_bufferChar + ind);
    strcpy( p_ptLemmeMem->ValTerme.Ident , p_chaine );
    p_ptLemmeMem->ValTerme.Ref = ptLemmeFic->Ref;
    p_ptLemmeMem->ValTerme.RefRacine = ptLemmeFic->RefRacine;
    p_ptLemmeMem->ValTerme.RefSemantique = NULL;
    p_ptLemmeMem->ValTerme.RefSelf = NULL;
    p_ptLemmeMem->ValTerme.TypeTerme = (char)_Nil;
    p_ptLemmeMem->ValTerme.PosFicRegle = ptLemmeFic->PosFicRegle;
    p_ptLemmeMem->NbRegles = ptLemmeFic->NbRegles;
    p_ptLemmeMem->NbDerives = ptLemmeFic->NbDerives;
    p_ptLemmeMem->PosFicDerives = ptLemmeFic->PosFicDerives;
    p_ptLemmeMem->PosFicRacine = ptLemmeFic->PosFicRacine;
    p_ptLemmeMem->NbVoisinsSemantiques = ptLemmeFic->NbVoisinsSemantiques;
    p_ptLemmeMem->PosFicLiensSemantiques = ptLemmeFic->PosFicLiensSemantiques;
    p_ptLemmeMem->NbVoisinsSemantiquesRec 
      = ptLemmeFic->NbVoisinsSemantiquesRec;
    p_ptLemmeMem->PosFicLiensSemantiquesRec 
      = ptLemmeFic->PosFicLiensSemantiquesRec;
    p_ptLemmeMem->ValTerme.ArcsFils = NULL;
    ind += (long)sizeof(struct LemmeFic);
    while( ind % _TailleMot != 0L )
    {
        ind++;
    }

    NbLiensNoeudsMemFic = 0;
    /***************** recopie du dag du lemme *********************/
    p_ptLemmeMem->ValTerme.ArcsFils
           = RestructurerDag( indTeteLemme , p_bufferChar , 
                              &ind , (int)ptLemmeFic->ArcsFils );

    /***************** recopie des racines annexes eventuelles *****/
    if( p_ptLemmeMem->NbRegles <= _RacineAnnexe1
        && p_ptLemmeMem->NbRegles >= _RacineAnnexeMax )
    {
        strcpy( p_chaineAnnexe , p_chaine );
        /* recopie de la chaine annexe                          */
        if (p_bufferChar[ ind ] == '\0')
        {
            TxtErreur( 73 , "Restructurer(altrfime)" );
        }

        i = 0L;
        while((p_chaine[ i ] = p_bufferChar[ ind ]) != '\0')
        {
             ind++;
             i++;
        }

        /* saut de la fin de chaine */
        ind++;

        /* saut du -1 en fin de chaine
           afin de ne pas inclure la fin de chaine
           comme 1er 0 de la separation a la destructuration      */
        ind++;
        while( ind % _TailleMot != 0L )
        {
           ind++;
        }
     }
     else
     {
         p_chaineAnnexe[ 0 ] = '\0';
     }

    if( p_ptLemmeMem->NbRegles == _Nil )
    {
         p_ptLemmeMem->NbRegles = 0;
    }

    /* saut d'une zone tampon non nulle */
    ind += (long)_TailleSeparation;

    if( ind + (long)_TailleSeparation > (long)TailleBuffLemmesCourant )
    {
          sprintf( ChAux , "Restructurer(altrfime) [%s] [%ld->%ld]",
                           p_bufferChar + *p_indBufferChar , *p_indBufferChar ,
                           ind + (long)_TailleSeparation );
          TxtErreur( 77 , ChAux );
          return _Nil;
    }

    /* saut d'une zone tampon nulle */
    ind += (long)_TailleSeparation;

    *p_indBufferChar = ind;

    return ind;
}

/************************************************************************
                       passage fichier ----> memoire (une regle)
*************************************************************************/
long     RestructurerUneRegle( unsigned long *p_indAbsRegle , 
                               TabRegleMem *p_ptRegleAux )
{
    long                indTeteRegle;
    long                ind;
    long                indAux;
    TabRegleFic         ptRegleFic;

    ind = _IRelRegle( *p_indAbsRegle );
    if( ind < 0L || ind % _TailleMot != 0L )
    {
        return (long)_Nil;
    }

    indTeteRegle = ind;

    /* amorce pour que le debut du tableau ne soit pas nul */
    for( indAux = 0 ; indAux < _TailleMot ; indAux ++ )
    {
       BufferReglesCourant[ ind ] = (char)_Nil;
       ind++;
    }

    (*p_ptRegleAux) = NouvelleRegle();

    /* recopie de la tete de regle No indRegle */
    ptRegleFic = (TabRegleFic)( BufferReglesCourant + ind );
    (*p_ptRegleAux)->RegleFrere = NULL;
    /* (*p_ptRegleAux)->TermeFils est garni lors de
    la recopie des fils */
    (*p_ptRegleAux)->PosFicRegleSuivante = ptRegleFic->PosFicRegleSuivante;

    ind += (long)sizeof(struct RegleFic);
    while( ind % _TailleMot != 0L )
    {
        ind++;
    }
 
    /* termes temporaires pour passage des donnees fichier aux donnees
       memoire il ne peut y avoir de dags communs entre 2 regles
       differentes, donc cet indice doit etre remis a zero entre
       chaque regle */
    NbLiensNoeudsMemFic = 0;

    /***************** recopie des termes *********************/
    (*p_ptRegleAux)->TermeFils
           = RestructurerArbre( indTeteRegle , BufferReglesCourant , &ind ,
                                (int)ptRegleFic->TermeFils );

    if( Trace )
    {
         EcrireRegleMem( &(*p_ptRegleAux) , _False , FicLog , _Sortie );
    }

    /* saut d'une zone tampon non nulle */
    ind += (long)_TailleSeparation;

    /* saut d'une zone tampon nulle */
    ind += (long)_TailleSeparation;

    if( ind > (long)TailleBuffReglesCourant )
    {
          sprintf( ChAux , "Restructurer(altrfime) [%s] [%ld->%ld]",
                           BufferReglesCourant + ind , ind ,
                           ind + (long)_TailleSeparation );
          TxtErreur( 78 , ChAux );
          return _Nil;
    }

    *p_indAbsRegle = IndBuffRegleMin + (unsigned long)ind;
    return ind;
}

/************************************************************************
                       passage fichier ----> memoire (toutes les regles)
*************************************************************************/
long     RestructurerLesRegles( unsigned long *p_indAbsRegle , 
                                TabRegleMem *p_ptRegleCou , int p_nbRegles ,
                                int p_avecVerifNbRegles )
{
    unsigned long       indAbsRegleSuivante;
    int                 indRegle;             /* indice des regles */
    TabRegleMem         ptRegleAux;
    TabRegleMem         ptReglePrec;
    long                ind;

    /***************** recopie des regles eventuelles *********************/
    *p_ptRegleCou = NULL;
    indAbsRegleSuivante = *p_indAbsRegle;

    indRegle = 0;
    /* liste des regles                                                    */
    while ( (long)indAbsRegleSuivante != (long)_Nil )
    {
        if( indRegle > 0 )
        {
           ptReglePrec = ptRegleAux;
        }

        ind =  RestructurerUneRegle( &indAbsRegleSuivante , &ptRegleAux  );
        if( ind == _Nil )
        {
            TxtErreur( 61 , "RestructurerLesRegles(altrfime)" );
            break;
        }

        if( indRegle == 0 )
        {
           *p_ptRegleCou = ptRegleAux;
        }
        else
        {
           ptReglePrec->RegleFrere = ptRegleAux;
        }

        *p_indAbsRegle = indAbsRegleSuivante;
        indAbsRegleSuivante = ptRegleAux->PosFicRegleSuivante;
        indRegle++;
    }
    /* parcours des regles */

    if( p_avecVerifNbRegles && !( ( p_nbRegles > 0 && indRegle == p_nbRegles)
                                 || ( p_nbRegles <= 0 && indRegle == 0 ) )  )
    {
         TxtErreur( 95 , "RestructurerLesRegles(altrfime)" );
         return _Nil;
    }

    return ind;
}


/************************************************************************
                       passage fichier ----> memoire (lemmes)
*************************************************************************/
/* memes fonctions que les precedentes sauf que l'on part d'un indice
   absolu d'un Lemme ou d'une regle et que la fonction renvoit
   l'indice absolu du Lemme ou de la regle suivante */
long     RestructurerBDicLemme( unsigned long *p_indAbsLemme , char *p_chaine,
                                TabLemmeMem p_ptLemmeMem , 
                                char *p_chaineAnnexe )
{
    long                indRelLemmeIni;
    long                indRelLemme;    /* indices relatifs du Lemme */

    indRelLemmeIni = ( indRelLemme = _IRelLemme( *p_indAbsLemme ) );
    if( indRelLemmeIni < 0L )
    {
        return (long)_Nil;
    }

    if( RestructurerLemme( BufferLemmesCourant , &indRelLemme , p_chaine , 
                           p_ptLemmeMem , p_chaineAnnexe) == _Nil )
    {
        return (long)_Nil;
    }

    /* valeur retournee */
    *p_indAbsLemme += ( indRelLemme - indRelLemmeIni );

    return indRelLemme - indRelLemmeIni;
}

/****************************************************************************
    recherche une regle dans le fichier des regles avec sa
 position et son numero de regle dans la liste des regles qui seront lues
****************************************************************************/
char     ReleverUneRegleBDic( unsigned long p_indAbs , 
                              TabRegleMem *p_ptRegle )
{
    if( RestructurerUneRegle( &p_indAbs , p_ptRegle ) == _Nil || !(*p_ptRegle) )
    {
        sprintf( ChAux , "ReleverUneRegleBDic(altrfime) [%ld]",
                           p_indAbs );
        TxtErreur( 136 , ChAux );
        return _False;
    }

    return _True;
}

/****************************************************************************
          lecture des ula + mots composes attaches a un mot simple du
          dictionnaire des mots courts et usuels
****************************************************************************
struct EntreeDucs               *p_ptEntree :     pt dans Ducs */

void     RestructurerDucs( struct EntreeDucs *p_ptEntree ,
                           struct LemmeMem *p_ulaCou,
                           TabRegleMem *p_ptRegleCou )
{
    int                     i;
    int                     nbFilsTerme;
    ChIdent                 p_chaine;
    TabRegleMem             ptRegleAux;
    TabRegleMem             ptReglePrec;

    strcpy( p_chaine , p_ptEntree->Lemme.ValTerme.Ident);
    p_ulaCou = &(p_ptEntree->Lemme);
    if( (nbFilsTerme = p_ptEntree->Lemme.NbRegles) != _Nil )
    {
        p_ulaCou->NbRegles = nbFilsTerme;

        /* recopie des tetes de graphes                   */
        for( i = 0 ; i < nbFilsTerme ; i++)
        {
            if( i == 0 )
            {
               ptRegleAux = NouvelleRegle();
               *p_ptRegleCou = ptRegleAux;
            }
            else
            {
               ptReglePrec = ptRegleAux;
               ptRegleAux = NouvelleRegle();
               ptReglePrec->RegleFrere = ptRegleAux;
           }

           /* recopie des noeuds de graphe                 */
        }
    }
}

/****************************************************************************
                        Parametres d'execution (Lecture)
*****************************************************************************/
char     LireVersionEtDate( FILE *p_ptFic , char *p_carLu ,
			   ChIdent p_versionfastrCou ,
			   ChIdent p_datefastrCou )
{
  int                       indCar;

  SautBlanc( p_ptFic , p_carLu );
  if( !ReconnaitreChaine( p_ptFic , "FASTR" , p_carLu ) ) {
    return _False;
  }
  if( !ReconnaitreChaine( p_ptFic , "Version" , p_carLu ) ) {
    return _False;
  }
  indCar = 0;
  while( indCar < _MaxLetIdent && *p_carLu != '/'  
	&& *p_carLu != EOF ) {
    p_versionfastrCou[ indCar ] = *p_carLu;
    *p_carLu = LireCar( p_ptFic );
    indCar++;
  }
  p_versionfastrCou[ indCar ] = '\0';
  while( *p_carLu != '/'  && *p_carLu != EOF ) {
    *p_carLu = LireCar( p_ptFic );
    indCar++;
  }
  indCar = 0;
  while( indCar < _MaxLetIdent && !_CarBlanc(*p_carLu)  
	&& *p_carLu != EOF ) {
    p_datefastrCou[ indCar ] = *p_carLu;
    *p_carLu = LireCar( p_ptFic );
    indCar++;
  }
  p_datefastrCou[ indCar ] = '\0';
  SautBlanc( p_ptFic , p_carLu );

  return _True;
}     


char     LireSetUp( char *p_nomFicSetUp )
{
     FILE               *ptFicSetUp;
     char               car;
     int                ind = 2;
     char               ok = _True;

     /* lecture des parametres */
      ptFicSetUp = fopen( p_nomFicSetUp , _FopenRT );
      if( !(ptFicSetUp) )
      {
           printf( 
              "fastr: Error 13: File not found [LireSetUp(altrfime) [%s]]\n" , 
              p_nomFicSetUp );
           return _False;
      }

     car = LireCar( ptFicSetUp );
     SautBlanc( ptFicSetUp , &car );

     ok &= LireVersionEtDate( ptFicSetUp , &car , 
			     Versionfastr , Datefastr );

     /******************************************************************
                           NOMS DES FICHIERS
      ******************************************************************/
     ok &= ReconnaitreChaine( ptFicSetUp , "Message file            :" , 
                              &car );
     LireChaineAvecBalise( ptFicSetUp , &car , NomFicMessages ,
                           _NbMaxLetNomFic - 1 , '\'' , _True );
     ExpandEnvironementVariable( NomFicMessages );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     LireChaineAvecBalise( ptFicSetUp , &car , NomFicDico ,
                           _NbMaxLetNomFic - 1 , '\'' , _True );
     ExpandEnvironementVariable( NomFicDico );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     LireChaineAvecBalise( ptFicSetUp , &car , NomFicHeader ,
                           _NbMaxLetNomFic - 1 , '\'' , _True );
     ExpandEnvironementVariable( NomFicHeader );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     LireChaineAvecBalise( ptFicSetUp , &car , NomFicLog ,
                           _NbMaxLetNomFic - 1 , '\'' , _True );
     ExpandEnvironementVariable( NomFicLog );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     LireChaineAvecBalise( ptFicSetUp , &car , NomFicExtUnifier ,
                           _NbMaxLetNomFic - 1 , '\'' , _True );
     ExpandEnvironementVariable( NomFicExtUnifier );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     LireChaineAvecBalise( ptFicSetUp , &car , NomFicUnification ,
                           _NbMaxLetNomFic - 1 , '\'' , _True );
     ExpandEnvironementVariable( NomFicUnification );

     /* n'est actuellement pas dans le fichier de configuration
	pourrait s'y trouver dans le futur */
     strcpy( NomFicAux , NomFicDico );
     strcat( NomFicAux , ".TMP" );

     /******************************************************************
                                  BOOLEENS
      ******************************************************************/
     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     Trace = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     ArretSurErreurCompil = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     VerificationAccord = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     PretraitementDesNombres = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     LexicalisationImperativeDesRegles = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     AriteMin = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     AriteMax = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     BDicSurDisque = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     BDicReduit = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     OuvreAutomReglesMem = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     OuvreAutomMetaregl = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     FormatDEntree = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     FormatDeSortie = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     TraceTexteEnSortie = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     HeaderEnSortie = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NoPremLigne = LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     ClasstReglesAutom = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     LinkAutom = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     LexicalisationMultiple = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     TraitementBatch = (int)LireEntier( ptFicSetUp , &car );


     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     SuffStripping = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     TailSuffStripping = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     MorphoFlex = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     MorphoDer = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     ExploitationLiensMorpho = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     AutomaticSelfMorphologicalRoot = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     ExploitationLiensSem = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     AutomaticSelfSemanticLinks = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     MinusculisationAutomatique = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     SegmentationAutomatique = (int)LireEntier( ptFicSetUp , &car );

     /******************************************************************
                       CARACTERISTIQUES DU B-ARBRE
      ******************************************************************/
     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     ProfondeurBArbre = (int)LireEntier( ptFicSetUp , &car );
     
     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbPortesBArbre = (int)LireEntier( ptFicSetUp , &car );
     
     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     FrequenceMAJBArbre = LireEntier( ptFicSetUp , &car );


     /******************************************************************
                           TAILLE DES BUFFERS
      ******************************************************************/
     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     TailleBuffStd = LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     TailleBuffLemmesCourt = LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     TailleBuffReglesCourt = LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     TaillePileEffacement = (unsigned long)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbMaxElemAllouables[ _TabMots ] 
       = (long)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbMaxElemAllouables[ _TabRefPhrase ] 
       = (long)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbMaxElemAllouables[ _TabRegles ] 
       = (long)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbMaxReglesMem = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbMaxElemAllouables[ _TabTermes ] 
       = (long)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbMaxElemAllouables[ _TabReglesMeta ] 
       = (long)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbMaxElemAllouables[ _TabLemmes ] 
       = (long)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbMaxElemAllouables[ _TabDag ] 
       = (long)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbMaxLiensMemorises = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbMaxTraitsCatLex = (int)LireEntier( ptFicSetUp , &car );

     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbMaxSuffixes = (int)LireEntier( ptFicSetUp , &car );
     
     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbMaxTermesDansUneRegle = (int)LireEntier( ptFicSetUp , &car );
     
     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbMaxChainesTermes = (int)LireEntier( ptFicSetUp , &car );
     
     TxtChaineRsrc( 135 , ind++ , ChAux );
     ok &= ReconnaitreChaine( ptFicSetUp , ChAux , &car );
     NbMaxLiens = (int)LireEntier( ptFicSetUp , &car );
     
     if( !ok )
     {
           printf( 
      "fastr: Error 91: Incorrect set-up file [LireSetUp(altrfime) [%s]]\n" , 
                      p_nomFicSetUp );
       return _False;
     }
     
     fclose( ptFicSetUp );
     return _True;
}

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                  fin de altrfime.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

