/*
 * FASt Term Recognizer
 *
 *  FASTR/altrmefi.c
 *  Version 2.04/04-02-01
 *
 *  Copyright (C) 2004  Christian Jacquemin, LIMSI-CNRS
 *
 *  This file contains code for converting memory data into file data
 */

#include "al-ext.h"
#include "al-var.h"
#include "al-fic.h"

#ifndef WARNINGS
#include "al-head.h"
#endif
#ifdef WARNINGS
#include "al-headP.h"
#endif

/****************************************************************************
                   transfert memoire <--> fichier
                      ou fichier <--> fichier
 ****************************************************************************
lecture ou ecriture, selon le p_mode de p_longueur enregistrements de
taille 1 octet a partir de l'enregistrement p_debut (octet) du fichier
indFic et a partir du debut du buffer p_buff */

long     LireEcrire( unsigned long p_debut, long p_longueur, 
                     TabChar p_buff, int p_mode,
                     FILE *p_ptFic )
{
    long                 nbTotTransferes = 0L;
                                    /* nombre total d'octets transferes     */

    if( p_longueur < 0L )
    {
        sprintf( ChAux , "LireEcrire(altrmefi) [%ld->%ld]" , 
                 p_debut , p_longueur );
        TxtErreur( 129 , ChAux );
        return _Nil;
    }
    if( p_longueur == 0L )
    {
        return 0L;
    }

    fseek( p_ptFic , (long)p_debut , 0 );

#ifndef PCold
    if ( p_mode == _Lire )
    {
       nbTotTransferes
          = (long)fread( (char *)p_buff , 1 , 
                         (unsigned long)p_longueur , p_ptFic );
    }
    else
    {
        nbTotTransferes
          = (long)fwrite( (char *)p_buff , 1 , 
                          (unsigned long)p_longueur , p_ptFic );
    }
#else
    pt = p_buff;

    /* lecture des buffers entiers                                  */
    for( i = 0L ; i < p_longueur / TailleBuffStd ; i++ )
    {
        if ( p_mode == _Lire )
        {
           nbTransferes
              = (long)fread( (char *)pt , 1 , (int)TailleBuffStd , p_ptFic );
        }
        else
        {
            nbTransferes
              = (long)fwrite( (char *)pt , 1 , (int)TailleBuffStd , p_ptFic );
        }

        if ( nbTransferes > 0L )
        {
           nbTotTransferes += nbTransferes;
           pt += nbTransferes / (long)sizeof( *p_buff );
        }
        else
        {
           break;
        }
    }

    if( p_longueur % TailleBuffStd > 0L )
    {
        /* lecture du buffer incomplet final                            */
        if ( p_mode == _Lire )
        {
          nbTransferes
            = (long)fread( (char *)pt , 
                            1 , (int)(p_longueur % (long)TailleBuffStd) , 
                            p_ptFic );
        }
        else
        {
          nbTransferes
             = (long)fwrite( (char *)pt , 
                             1 , (int)(p_longueur % (long)TailleBuffStd) , 
                             p_ptFic );
        }
        if ( nbTransferes > 0L )
        {
                nbTotTransferes += nbTransferes;
        }
    }
#endif

    if( nbTotTransferes != p_longueur )
    {
		if ( p_mode == _Lire ) {
			if( p_ptFic == FicLog ) {
				printf( "Fichier FicLog\n" );
			}
			else if( p_ptFic == PtFicRegles ) {
				printf( "Fichier PtFicRegles\n" );
			}
			else if( p_ptFic == FicBDicRegles ) {
				printf( "Fichier FicBDicRegles\n" );
			}
			else if( p_ptFic == FicBDicLemmes ) {
				printf( "Fichier FicBDicLemmes\n" );
			}
			else if( p_ptFic == FicBDicLiens ) {
				printf( "Fichier FicBDicLiens\n" );
			}
			else if( p_ptFic == FicBDicLiensAux ) {
				printf( "Fichier FicBDicLiensAux\n" );
			}
			else if( p_ptFic == FicRefIdent ) {
				printf( "Fichier FicRefIdent\n" );
			}
			else if( p_ptFic == FicRefMots ) {
				printf( "Fichier FicRefMots\n" );
			}
			sprintf( ChAux , "LireEcrire(altrmefi) Read [%ld->%ld] (%ld) " , 
                        p_debut , p_longueur , nbTotTransferes );
		}
		else {
			sprintf( ChAux , "LireEcrire(altrmefi) Write [%ld->%ld] (%ld) " , 
                        p_debut , p_longueur , nbTotTransferes );
		}
		TxtErreur( 130 , ChAux );
    }

    /* toute ecriture dans le fichier doit s'accompagner d'un
      rechargement du buffer pour en avoir une image exacte */
    if( p_ptFic == FicBDicLemmes && p_mode == _Ecrire )
    {
       IndBuffLemmeMax = 0L;
    }
    if( p_ptFic == FicBDicRegles && p_mode == _Ecrire )
    {
       IndBuffRegleMax = 0L;
    }

    return nbTotTransferes;
}

/* transfert de p_longueur entiers courts d'un fichier ptfic1 a partir
 du numero p_no1 vers le fichier ptfic2 a partir du numero p_no2 */

long     Recopier( unsigned long p_no1, unsigned long p_no2, 
                   unsigned long p_longueur,
                   FILE *p_ptFic1 , FILE *p_ptFic2 )
{

    long              nbTotEcrits = 0L;  /* p_longueur total d'octets 
                                            effectivement recopies  */
    long              nbLus;             /* p_longueur d'octets lus dans 
                                            le fichier de depart */
    long              nbEcrits;          /* p_longueur d'octets ecrits 
                                            dans le fic  d'arrivee */
    long              i;                 /* compteur du p_longueur de buff 
                                            entiers transferes */
    unsigned long     posCour;

    TabChar           buff;              /* buffer de transfert */

    buff = AlloueBuffer( TailleBuffStd );

    /* positionnement sur les 2 fichiers de depart et d'arrivee          */
    fseek( p_ptFic1 , (long)p_no1 , 0 );
    fseek( p_ptFic2 , (long)p_no2 , 0 );

    /* recopie des buffers entiers                                       */
    for( i = 0L ; i < ((long)p_longueur / (long)TailleBuffStd) ; i++ )
    {
         fgetpos( p_ptFic1 , &posCour );
         nbLus = LireEcrire( posCour , (long)TailleBuffStd , 
                             (TabChar)buff , _Lire , p_ptFic1 );

         if (nbLus > 0L)
         {
               fgetpos( p_ptFic2 , &posCour );
               nbEcrits = LireEcrire( posCour , (long)nbLus , 
                                      (TabChar)buff , _Ecrire , p_ptFic2 );
         }
         else
               break;

         if( nbEcrits > 0L )
         {
               nbTotEcrits += nbEcrits;
         }
         else
         {
               break;
         }
     }

     /* recopie du buffer incomplet final                                */
     /* on ne recopie que si la lecture precedente a donne un nombre
           d'octet non nul i .e. si l'on n'est pas a la fin du fichier
           de lecture                                                       */
     if( nbLus > 0L )
     {
        fgetpos( p_ptFic1 , &posCour );
        nbLus = LireEcrire( posCour , (long)p_longueur % (long)TailleBuffStd , 
                            (TabChar)buff , _Lire , p_ptFic1 );
 
        if( nbLus > 0L )
        {
            fgetpos( p_ptFic2 , &posCour );
            nbEcrits = LireEcrire( posCour , (long)nbLus , 
                                   (TabChar)buff , _Ecrire , p_ptFic2 );

           if( nbEcrits > 0L )
           {
              nbTotEcrits += nbEcrits;
           }
        }
    }

    if( p_longueur != (unsigned long)nbTotEcrits )
    {
       sprintf( ChAux , "Recopier(altrmefi) [%ld/%ld]",
                        nbTotEcrits , p_longueur );
       TxtErreur( 54 , ChAux );
    }

    _Free( buff );
    return nbTotEcrits;
}

/****************************************************************************
                insertion dans un fichier par decalage de la fin
 ****************************************************************************
insertion de p_longueur enregistrements (octets) du buffer p_buff dans le fichier
indFic a partir de l'enregistrement numero p_debut
   s'il n'y a pas de Lemme preexistante p_debut = p_debutDecalage
        long    p_debut,                 debut de l'insertion du nouveau buffer
                p_debutDecalage          debut de la zone a repousser
                p_longueur;              taille du buffer a inserer            */

void     Inserer( unsigned long p_debut, unsigned long p_debutDecalage, 
                  long p_longueur,
                  TabChar p_buff, FILE *p_ptFic )
{
long             tailleTotaleARecopier;
                 /* longueur a recopier entre le debut de  
                 l'insertion et la fin du fichier initial     */
long             tailleRecopie;        
                 /* nombre d'octets recopies pour faire le va et vient
                                          de recopie de la fin du
                                          fichier */
unsigned long    departInsertion;      
                 /* debut de l'insertion dans le fichier source
                                          i.e. tout ce qui precede
                                          n'est pas a modifier*/
unsigned long    departCopie;          
                 /* debut de la recopie dans le fichier source
                                          i.e. tout ce qui suit est a
                                          decaler */
long             finPremier ;
long             finDeuxieme;
                                       
                 /* les bornes des 2 sauvegardes dans 2 buffers
                                           auxiliaires permettant de
                                           decaler la fin du fic.*/
unsigned long    tailleInitialeFic;
unsigned long    tailleFinFic;         
                 /* nombre d'octets restant a recopier en fin du
                                          fichier afin de pouvoir
                                          faire l'insertion par
                                          recopie de fin du fichier
                                          sur un fichier auxilliaire
                                          au cas ou la taille de
                                          l'insertion est superieure a
                                          celle du buffer de
                                          sauvegarde */
TabChar          ancien;               
                 /* numero du buffer contenant les donnees les plus
                                          anciennes, donc les 1eres a
                                          transferer sur le fic.*/

TabChar          buff1;                
                 /* le buffer de transfert no 1 */
TabChar          buff2;                
                 /* le buffer de transfert no 2 */

FILE             *ptFicAux;     
                 /* fich. aux de recopie au cas ou la recopie ne se
                                   fait pas directement sur le fichier
                                   initial (si l'insertion est plus
                                   grande que le buffer de copie */

   if ( p_longueur <= 0L )
   {
        return;
   }

   buff1 = AlloueBuffer( TailleBuffStd );
   buff2 = AlloueBuffer( TailleBuffStd );

   fseek( p_ptFic , 0L , 2 );
   fgetpos( p_ptFic , &tailleInitialeFic );

   if ( (p_longueur - (long)(p_debutDecalage - p_debut) ) 
                 <= (long)TailleBuffStd)
   {
        /* initialisations                                              */
        ancien = buff1;
        departInsertion = p_debut;
        departCopie = p_debutDecalage;
        tailleTotaleARecopier = (long)(tailleInitialeFic - p_debutDecalage);

        /* remplissage du 1er buffer de recopie                         */
        tailleRecopie = _Min( (long)TailleBuffStd , tailleTotaleARecopier );
        finPremier = LireEcrire(  departCopie , tailleRecopie ,
                                  buff1 , _Lire , p_ptFic );
        departCopie += (unsigned long)finPremier;
        tailleTotaleARecopier -= (unsigned long)finPremier;

        /* remplissage du 2er buffer de recopie                         */
        tailleRecopie = _Min( (long)TailleBuffStd , tailleTotaleARecopier );
        finDeuxieme = LireEcrire( departCopie , tailleRecopie ,
                                  buff2 , _Lire , p_ptFic );
        departCopie += (unsigned long)finDeuxieme;
        tailleTotaleARecopier -= finDeuxieme;

        departInsertion += (unsigned long)LireEcrire( 
                 departInsertion , p_longueur ,
                 p_buff , _Ecrire , p_ptFic );

        while( finPremier > 0L )
        {
            /* insertion du buffer le plus ancien                   */
            departInsertion += (unsigned long)LireEcrire( 
                 departInsertion , finPremier , ancien , 
                 _Ecrire , p_ptFic );
            finPremier = finDeuxieme;

            /* sauvegarde dans le buffer le plus ancien             */
            tailleRecopie = _Min( TailleBuffStd, tailleTotaleARecopier );
            finDeuxieme = LireEcrire( departCopie , tailleRecopie ,
                                      ancien , _Lire , p_ptFic );
            departCopie += (unsigned long)finDeuxieme;
            tailleTotaleARecopier -= (unsigned long)finDeuxieme;

            /* echange des buffers ancien et nouveau                */
            if( ancien == buff1 )
            {
               ancien = buff2;
            }
            else
            {
               ancien = buff1;
            }
         }/* tant que le buffer a recopier de la memoire vers le
         fichier n'est pas vide */
         if( tailleTotaleARecopier != 0L )
         {
            sprintf( ChAux , "Inserer(altrmefi) [%ld]" , 
                     tailleTotaleARecopier );
            TxtErreur( 131 , ChAux );
         }

   }    /* recopie par buffer sur le fichier lui-meme                   */
   else
   {
        ptFicAux = fopen( NomFicAux , _FopenWPB );
        fseek( p_ptFic , 0L , 2 );
        fgetpos( p_ptFic , &tailleFinFic );

        /* recopie de la fin du 1er fichier a sauvegarder         */
        Recopier( p_debutDecalage , 0L , 
                  tailleFinFic - (unsigned long)p_debutDecalage ,
                  p_ptFic , ptFicAux );

        /* insertion de la partie nouvelle                        */
        LireEcrire( p_debut , p_longueur , p_buff , _Ecrire , p_ptFic );

        /* remise de la partie sauvegardee sur le fichier initial */
        Recopier( 0L , p_debut + (unsigned long)p_longueur , 
                  tailleFinFic - (unsigned long)p_debutDecalage ,
                  ptFicAux , p_ptFic );

        fclose( ptFicAux );
        remove( NomFicAux );
    }   /* fin de recopie par fichier auxiliaire                        */

    fseek( p_ptFic , 0L , 2 );
    fgetpos( p_ptFic , &tailleFinFic );
    if( tailleInitialeFic + p_longueur - (p_debutDecalage - p_debut) 
                 != tailleFinFic )
    {
       sprintf( ChAux , "Inserer(altrmefi) [%ld,%ld,%ld] [%ld,%ld,%ld]" ,
                        p_debut , p_debutDecalage , p_longueur ,
                        tailleInitialeFic ,
                        tailleInitialeFic + p_longueur 
                                - ( p_debutDecalage - p_debut ) ,
                        tailleFinFic );
       TxtErreur( 55 , ChAux );
    }

    /* toute ecriture dans le fichier doit s'accompagner d'un
       rechargement du buffer pour en avoir une image exacte */
    if( p_ptFic == FicBDicLemmes )
    {
         IndBuffLemmeMax = 0L;
    }
    if( p_ptFic == FicBDicRegles )
    {
         IndBuffRegleMax = 0L;
    }
    _Free( buff1 );
    _Free( buff2 );
}

/************************************************************************
    passage memoire ---> fichier (pour une arborescence de termes)
*************************************************************************/
char     DestructurerArbre( TabTermeMem p_noeudCourant ,
                            long p_indTeteLemme , TabChar p_bufferChar , 
                            long *p_ind )
{
     TabTermeFic            ptTermeFic;
     long                   indTermeFic;

    /* terme existant */
    if( p_noeudCourant )
    {
        indTermeFic = *p_ind;
    }
    /* terme existant */
    else
    /* terme inexistant */
    {
        return _Nil;
    }
    /* terme inexistant */

    /* recopie du terme courant */
    ptTermeFic = (TabTermeFic)(p_bufferChar + *p_ind );
    ptTermeFic->Ref = p_noeudCourant->ValTerme.Ref;
    ptTermeFic->RefRacine = p_noeudCourant->ValTerme.RefRacine;
    ptTermeFic->TypeTerme = p_noeudCourant->ValTerme.TypeTerme;
    ptTermeFic->PosFicRegle = p_noeudCourant->ValTerme.PosFicRegle;
    *p_ind += (long)sizeof(struct TermeFic);
    while( (*p_ind) % _TailleMot != 0L )
    {
       p_bufferChar[ *p_ind ] = (char)_Nil;
       (*p_ind)++;
    }
    if( *p_ind > (long)TailleBuffReglesCourt ){
      TxtChaineErreur( 65 , ChErreurCompil );
      return (char)_Nil;
    }

    /***************** recopie du dag du terme *********************/
    ptTermeFic->ArcsFils = DestructurerDag( p_noeudCourant->ValTerme.ArcsFils ,
                                            p_indTeteLemme , p_bufferChar, 
                                            p_ind ,
                                            TailleBuffReglesCourt );

    /***************** appel recursif sur l'arborescence ***********/
    ptTermeFic->TermeFils
            = DestructurerArbre( p_noeudCourant->TermeFils ,
                                 p_indTeteLemme , p_bufferChar , p_ind );
    ptTermeFic->TermeFrere
            = DestructurerArbre( p_noeudCourant->TermeFrere ,
                                 p_indTeteLemme , p_bufferChar , p_ind );

    return _RefEnAvant;
}

/************************************************************************
    passage memoire ---> fichier (pour un dag)
*************************************************************************/
char     DestructurerDag( TabDagMem p_noeudCourant ,
                          long p_indTeteLemme , TabChar p_bufferChar , 
                          long *p_ind ,
                          long p_tailleBuffer )
{
    TabDagFic               ptNoeudFic;
    long                    nouvInd;
    long                    indAux;
    long                   *ptRefEnArriere;
    int                     ind;
    TabChar                 ch;
    int                     longCh;

    /* dag existant */
    if( p_noeudCourant )
    {
       /* dag deja en memoire */
       if( p_noeudCourant->Marque )
       {
          /* pteur vers un dag deja passe sur fichier */
          /* on retrouve son pteur fichier */
          for( ind = 0 ; ind < NbLiensNoeudsMemFic ; ind++ )
          {
              if( (LiensNoeudsMemFic + ind)->PtMem == p_noeudCourant )
              {
                   /* en cas de ref en arriere, celle-ci est donnee
                      apres le dag */
                   ptRefEnArriere = (long *)(p_bufferChar + *p_ind);
                   *ptRefEnArriere = (LiensNoeudsMemFic + ind)->PtFic;
                   *p_ind += (long)sizeof(long);
                   while( (*p_ind) % _TailleMot != 0L )
                   {
                       p_bufferChar[ *p_ind ] = (char)_Nil;
                       (*p_ind)++;
                   }
                   if( *p_ind > (long)p_tailleBuffer)
                   {
                        return (char)_Nil;
                   }
                   return _RefEnArriere;
              }
          }
          TxtErreur( 56 , "DestructurerDag(altrmefi) dag not found 4" );
          return (char)_Nil;
       }
       /* dag deja en memoire */
       else
       /* dag a creer sur fichier */
       {
           nouvInd = *p_ind;
       }
       /* dag a creer en memoire */
    }
    /* dag existant */
    else
    /* dag inexistant */
    {
        return (char)_Nil;
    }
    /* dag inexistant */


    /* dag non encore connu, on le copie */
    /* sauvegarde du lien du noeud du p_noeudCourant courant */
    (LiensNoeudsMemFic + NbLiensNoeudsMemFic)->PtMem 
                                            = p_noeudCourant;
    (LiensNoeudsMemFic + NbLiensNoeudsMemFic)->PtFic 
                                            = nouvInd - p_indTeteLemme;
    NbLiensNoeudsMemFic++;
    if( NbLiensNoeudsMemFic >= NbMaxLiensMemorises )
    {
         TxtErreur( 94 , "RestructurerDag(altrmefi)" );
         return (char)_Nil;
    }

    p_noeudCourant->Marque = _True;
    /* recopie du noeud du p_noeudCourant courant */
    ptNoeudFic = (TabDagFic)(p_bufferChar + nouvInd);
    for( indAux = nouvInd ; 
         indAux < nouvInd + (long)sizeof(struct DagFic) ; indAux++ )
    {
       p_bufferChar[ indAux ] = (char)_Nil;
    }
    ptNoeudFic->TypeDagFic = (char)(p_noeudCourant->TypeDag);
    ptNoeudFic->x1 = _Nil;
    *p_ind += (long)sizeof(struct DagFic);
    switch( p_noeudCourant->TypeDag )
    {
        case _TypeDagVide :
                    break;
        case _Arc : ptNoeudFic->ValDag.Arc.ValSymb
                           = p_noeudCourant->ValDag.Arc.ValSymb;
                    break;
        case _AtomeRef : 
                    if( p_noeudCourant->ValDag.Atome.AtomeConjDisj != _False )
                    {
                        ptNoeudFic->ValDag.Atome.AtomeConjDisj
                              = p_noeudCourant->ValDag.Atome.AtomeConjDisj;
                    }
                    else
                    {
                        ptNoeudFic->ValDag.Atome.AtomeConjDisj
                              = _Nil;
                    }
                    ptNoeudFic->ValDag.Atome.ValAtome.Ref
                           = p_noeudCourant->ValDag.Atome.ValAtome.Ref;
                    break;
        case _AtomeEntier : 
                    if( p_noeudCourant->ValDag.Atome.AtomeConjDisj != _False )
                    {
                        ptNoeudFic->ValDag.Atome.AtomeConjDisj
                              = p_noeudCourant->ValDag.Atome.AtomeConjDisj;
                    }
                    else
                    {
                        ptNoeudFic->ValDag.Atome.AtomeConjDisj
                              = _Nil;
                    }
                    ptNoeudFic->ValDag.Atome.ValAtome.Entier
                              = p_noeudCourant->ValDag.Atome.ValAtome.Entier;
                    break;
        case _AtomeChaine : 
                    if( p_noeudCourant->ValDag.Atome.AtomeConjDisj != _False )
                    {
                        ptNoeudFic->ValDag.Atome.AtomeConjDisj
                              = p_noeudCourant->ValDag.Atome.AtomeConjDisj;
                    }
                    else
                    {
                        ptNoeudFic->ValDag.Atome.AtomeConjDisj
                              = _Nil;
                    }
                    ch = p_noeudCourant->ValDag.Atome.ValAtome.Chaine;
                    longCh = strlen( ch );
                    if( longCh > _MaxLetIdent )
                    {
                        longCh = _MaxLetIdent;
                    }
                    if( longCh == 0 )
                    {
                        longCh = _Nil;
                    }
                    ptNoeudFic->ValDag.Atome.ValAtome.LongChaine = (long)longCh;
                    for( ind = 0 ; ind < longCh ; ind++ )
                    {
                        p_bufferChar[ *p_ind ] = ch[ ind ];
                        (*p_ind)++;
                    }
                    break;
        case _AtomeListe :
                    if( p_noeudCourant->ValDag.Atome.AtomeConjDisj != _False )
                    {
                        ptNoeudFic->ValDag.Atome.AtomeConjDisj
                              = p_noeudCourant->ValDag.Atome.AtomeConjDisj;
                    }
                    else
                    {
                        ptNoeudFic->ValDag.Atome.AtomeConjDisj
                              = _Nil;
                    }
                    break;
        case _Oper :
                    /*
                    while( (*p_ind) % _TailleMot != 0L )
                    {
                       p_bufferChar[ *p_ind ] = (char)_Nil;
                       (*p_ind)++;
                    }
                    ptOperLogFic = (struct OperLogFic *)
                                            (p_bufferChar + (*p_ind));
                    ptOperLogFils->Oper = _Nil
                    *p_ind += (long)sizeof(struct OperLogFic);
                    */
                    TxtErreur( 42 ,
"DestructurerDag(altrmefi) sauvegarde d'operation non encore implementee");
                    break;
        default :   sprintf( ChAux , 
"DestructurerDag(altrmefi) Noeud de type inconnu : %d\n" , 
                             p_noeudCourant->TypeDag );
                    TxtErreur( 42 , ChAux );
                    break;
    }

    /* recopie du p_noeudCourant courant */
    while( (*p_ind) % _TailleMot != 0L )
    {
       p_bufferChar[ *p_ind ] = (char)_Nil;
       (*p_ind)++;
    }
    if( *p_ind > (long)p_tailleBuffer)
    {
          return (char)_Nil;
    }

    /***************** appel recursif sur l'arborescence ***********/
    ptNoeudFic->Fils = DestructurerDag( 
                            p_noeudCourant->Fils , p_indTeteLemme , 
                            p_bufferChar , p_ind , p_tailleBuffer );

    ptNoeudFic->Frere = DestructurerDag( 
                            p_noeudCourant->Frere , p_indTeteLemme , 
                            p_bufferChar , p_ind , p_tailleBuffer );
    return _RefEnAvant;
}

/************************************************************************
    passage memoire ---> fichier (lemmes)
*************************************************************************

   On verifie de plus qu'il n'y a pas de \0\0 entre le debut et la fin
   ce qui creerait une fausse fin pour cet Lemme On retourne la
   longueur totale, y compris le delimiteur de fin */

long     DestructurerLemme( char *p_chaine, TabLemmeMem p_ptLemmeMem,
                       TabChar p_bufferChar , char *p_chaineAnnexe )
{
    long                i;
    long                ind = 0L;
    long                indAux;
    long                indTeteLemme;
                        /* indice du tableau cible p_bufferChar  */

    /* termes temporaires pour passge des donnees memoire aux donnees
       fichier */
    TabLemmeFic         ptLemmeFic;

    NbLiensNoeudsMemFic = 0;

     /* racine annexe */
     if( p_ptLemmeMem->NbRegles <= _RacineAnnexe1
         && p_ptLemmeMem->NbRegles >= _RacineAnnexeMax )
    {
        /* recopie de la chaine annexe                          */
        while((p_bufferChar[ ind ] = p_chaineAnnexe[ ind ]) != '\0')
        {
              ind++;
        }
        ind++;          /* saut de la fin de chaine             */
        while( ind % _TailleMot != 0L )
        {
           p_bufferChar[ ind ] = (char)_Nil;
           ind++;
        }

        indTeteLemme = ind;

        /* recopie du Lemme   */
        ptLemmeFic = (TabLemmeFic)( p_bufferChar + ind );
        ptLemmeFic->Ref = p_ptLemmeMem->ValTerme.Ref;
        ptLemmeFic->RefRacine = p_ptLemmeMem->ValTerme.RefRacine;
        ptLemmeFic->PosFicRegle = p_ptLemmeMem->ValTerme.PosFicRegle;
        ptLemmeFic->NbRegles = p_ptLemmeMem->NbRegles;
        ptLemmeFic->NbDerives = p_ptLemmeMem->NbDerives;
        ptLemmeFic->PosFicRacine = p_ptLemmeMem->PosFicRacine;
        ptLemmeFic->NbVoisinsSemantiques = p_ptLemmeMem->NbVoisinsSemantiques;
        ptLemmeFic->PosFicLiensSemantiques 
	  = p_ptLemmeMem->PosFicLiensSemantiques;
        ptLemmeFic->NbVoisinsSemantiquesRec 
	  = p_ptLemmeMem->NbVoisinsSemantiquesRec;
        ptLemmeFic->PosFicLiensSemantiquesRec 
	  = p_ptLemmeMem->PosFicLiensSemantiquesRec;
        ptLemmeFic->ArcsFils = (char)_Nil;
        ptLemmeFic->x1 = _Nil;
        ptLemmeFic->x2 = _Nil;
        ptLemmeFic->x3 = _Nil;
        ind += (long)sizeof(struct LemmeFic);
        while( ind % _TailleMot != 0L )
        {
           p_bufferChar[ ind ] = (char)_Nil;
           ind++;
        }

        /* recopie de la chaine principale                    */
        if( p_chaine[ 0 ] == '\0')
        {
              TxtErreur( 59 ,"DestructurerLemme(altrmefi)");
        }
        i = 0L;
        while( (p_bufferChar[ ind ] = p_chaine[ i ]) != '\0')
        {
              ind++;
              i++;
        }
        ind++;          /* saut de la fin de chaine             */

        /* incrementation pour ne pas inclure la fin de chaine
           comme 1er 0 de la separation                         */
        p_bufferChar[ ind ] = (char)_Nil;
        ind++;
        while( ind % _TailleMot != 0L )
        {
           p_bufferChar[ ind ] = (char)_Nil;
           ind++;
        }
        if( ind > (long)TailleBuffLemmesCourt ) {
          sprintf( ChAux , "DestructurerLemme(altrmefi) [%ld/%ld]",
                           ind , TailleBuffLemmesCourt );
          TxtErreur( 67 , ChAux );
	  return 0L;
        }
    }
    /* racine annexe */
    else
    /* racine principale */
    {
        if( Trace )
        {
            fprintf( FicLog , "-->F[%s]\n" , p_chaine );
        }
        /* recopie de la chaine principale                       */
        while( (p_bufferChar[ ind ] = p_chaine[ ind ]) != '\0')
        {
            ind++;
        }
        ind++;                  /* saut de la fin de chaine  */
        while( ind % _TailleMot != 0L )
        {
           p_bufferChar[ ind ] = (char)_Nil;
           ind++;
        }

        indTeteLemme = ind;

        /* recopie du Lemme   */
        ptLemmeFic = (TabLemmeFic)( p_bufferChar + ind );
        ptLemmeFic->Ref = p_ptLemmeMem->ValTerme.Ref;
        ptLemmeFic->RefRacine = p_ptLemmeMem->ValTerme.RefRacine;
        ptLemmeFic->PosFicRegle = p_ptLemmeMem->ValTerme.PosFicRegle;
        ptLemmeFic->NbRegles = p_ptLemmeMem->NbRegles;
        ptLemmeFic->NbDerives = p_ptLemmeMem->NbDerives;
        ptLemmeFic->PosFicDerives = p_ptLemmeMem->PosFicDerives;
        ptLemmeFic->PosFicRacine = p_ptLemmeMem->PosFicRacine;
	ptLemmeFic->NbVoisinsSemantiques = p_ptLemmeMem->NbVoisinsSemantiques;
	ptLemmeFic->PosFicLiensSemantiques 
	  = p_ptLemmeMem->PosFicLiensSemantiques;
	ptLemmeFic->NbVoisinsSemantiquesRec 
	  = p_ptLemmeMem->NbVoisinsSemantiquesRec;
	ptLemmeFic->PosFicLiensSemantiquesRec 
	  = p_ptLemmeMem->PosFicLiensSemantiquesRec;
        ptLemmeFic->ArcsFils = (char)_Nil;
        ptLemmeFic->x1 = _Nil;
        ptLemmeFic->x2 = _Nil;
        ptLemmeFic->x3 = _Nil;
        ind += (long)sizeof(struct LemmeFic);
        while( ind % _TailleMot != 0L )
        {
           p_bufferChar[ ind ] = (char)_Nil;
           ind++;
        }

        /************* recopie du dag du lemme ********************/
        DemarqueDagDF( &(p_ptLemmeMem->ValTerme.ArcsFils) );
        ptLemmeFic->ArcsFils = DestructurerDag( 
                            p_ptLemmeMem->ValTerme.ArcsFils ,
                            indTeteLemme , p_bufferChar, &ind ,
                            TailleBuffLemmesCourt );

        if( ind > (long)TailleBuffLemmesCourt ) {
          sprintf( ChAux , "DestructurerLemme(altrmefi) [%ld/%ld]",
                           ind , TailleBuffLemmesCourt );
          TxtErreur( 77 , ChAux );
	  return 0L;
        }
    }
    /* racine principale */

    /* ajout d'une zone tampon non nulle */
    for ( i = 0L ; i < (long)_TailleSeparation ; i++ )
    {
        p_bufferChar[ ind++ ] = _Nil;
    }

    if (p_bufferChar[ ind - 1L ] == '\0')
    {
        sprintf( ChAux , "DestructurerLemme(altrmefi) [%ld,%c]", ind - 1 ,
                         p_bufferChar[ ind - 1 ] );
        TxtErreur( 66 ,ChAux );
        exit( -1 );
    }

    /* ajout d'une zone tampon nulle */
    for ( i = 0L ; i < (long)_TailleSeparation ; i++ )
    {
        p_bufferChar[ ind++ ] = '\0';
    }

    /* verification de l'absence de 6 x \0 a l'interieur              */
    i = 0L;
    for ( indAux = 0 ; indAux < ind - (long)_TailleSeparation ; indAux++ )
    {
      if ( p_bufferChar[ indAux ] == '\0' )
      {
          i++;
      }
      else
      {
          i = 0L;
      }

      if ( i == (long)_TailleSeparation )
      {
         TxtErreur( 68 , "DestructurerLemme(altrmefi)" );
         return 0L;
      }
    }

    return ind;
}

/************************************************************************
    passage memoire ---> fichier (regles)
*************************************************************************

   On verifie de plus qu'il n'y a pas de \0\0 entre le debut et la fin ce qui
   creerait une fausse fin pour cet Lemme
   On retourne la longueur totale, y compris le delimiteur de fin       */

long     DestructurerUneRegle( TabRegleMem p_ptRegleCou ,
                            TabChar p_bufferChar )
{
    long                   i;
    long                   ind;
    long                   indAux;
    long                   indTeteRegle;

    /* termes temporaires pour passge des donnees memoire aux donnees
       fichier */
    TabRegleFic             ptRegleFic;

    /* initialisations */
    NbLiensNoeudsMemFic = 0;
    ind = 0L;
    indTeteRegle = ind;

    /* liste des tetes de regles */
    if( !p_ptRegleCou )
    {
        TxtErreur( 61 , "DestructurerUneRegle(altrmefi)" );
    }

    /* amorce pour que le debut du tableau ne soit pas nul */
    for( indAux = 0 ; indAux < _TailleMot ; indAux++ )
    {
       p_bufferChar[ ind ] = (char)_Nil;
       ind++;
    }

    /* recopie de la tete de regle */
    ptRegleFic = (TabRegleFic)(p_bufferChar + ind );
    /* ptRegleFic->TermeFils est garni lors de la recopie des fils */
    ptRegleFic->PosFicRegleSuivante = p_ptRegleCou->PosFicRegleSuivante;
    ptRegleFic->x1 = _Nil;
    ptRegleFic->x2 = _Nil;
    ptRegleFic->x3 = _Nil;

    ind += (long)sizeof(struct RegleFic);
    while( ind % _TailleMot != 0L )
    {
       p_bufferChar[ ind ] = (char)_Nil;
       ind++;
    }

    /************* recopie des termes et de leurs dags ********************/
    DemarqueArbreDF( &(p_ptRegleCou->TermeFils) );
    ptRegleFic->TermeFils = DestructurerArbre( p_ptRegleCou->TermeFils ,
                                               indTeteRegle , p_bufferChar, 
                                               &ind );

    /* ajout d'une zone tampon non nulle */
    for ( i = 0L ; i < (long)_TailleSeparation ; i++ )
    {
        p_bufferChar[ ind++ ] = _Nil;
    }

    if (p_bufferChar[ ind - 1L ] == '\0')
    {
        sprintf( ChAux , "DestructurerUneRegle(altrmefi) [%ld,%c]", ind - 1 ,
                          p_bufferChar[ ind - 1 ] );
        TxtErreur( 66 ,ChAux );
        exit( -1 );
    }

    /* ajout d'une zone tampon nulle */
    for ( i = 0L ; i < (long)_TailleSeparation ; i++ )
    {
        p_bufferChar[ ind++ ] = '\0';
    }

    if( ind > (long)TailleBuffReglesCourt )   {
      TxtChaineErreur( 78 , ChErreurCompil );
      return 0L;
    }

    /* verification de l'absence de 6 x \0 a l'interieur              */
    i = 0L;
    for ( indAux = 0 ; indAux < ind - (long)_TailleSeparation ; indAux++ )
    {
      if ( p_bufferChar[ indAux ] == '\0' )
      {
          i++;
      }
      else
      {
          i = 0L;
      }

      if ( i == (long)_TailleSeparation )
      {
         TxtErreur( 68 , "DestructurerUneRegle(altrmefi)" );
         return 0L;
      }
    }

    return ind;
}


/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                  fin de altrmefi.c
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/

